
NUMTESTS := 1;

SetVerbose("Selmer", 0);
SetVerbose("ThreeDescent", 0);
SetVerbose("CSAMaximalOrder", 0);

CD := CremonaDatabase();
MC := MonomialCoefficient;
MD := MonomialsOfDegree;

genfile := Open("CrvG1/CrvG1_data","r");

function GetCurve(s)
  ss := Split(s," []\t");
  N := StringToInteger(ss[1]);
  aa := [StringToInteger(a) : a in Split(ss[4],",")];
  E := EllipticCurve(aa);
  assert Conductor(E) eq N;
  rank := StringToInteger(ss[5]);
  return E,rank;
end function;

function GetPoints(s:Curve:=GetCurve(s))
  ss := Split(s," []\t");
  rank := StringToInteger(ss[5]);
  pts := [];
  for i in [1..rank] do
    pp := Split(ss[5+i],":");
    P := Curve![StringToInteger(x): x in pp];
    pts cat:= [P];
  end for;
  return pts;
end function;

// We read in the first 100 elliptic curves E/Q 
// with positive rank and certain 3-torsion types
// "mwtestdata" is an excerpt from Cremona's tables.

GenericExamples := [];
SylowExamples := [];
DihedralExamples := [];
IsogenyExamples :=[];
NonSplitExamples1 :=[];
NonSplitExamples2 :=[];
DiagonalExamples :=[];
SplitExamples :=[];

while true do
  s := Gets(genfile);
  if IsEof(s) then
    break;
  end if;
  E := GetCurve(s);
  P := GetPoints(s)[1];
  case ThreeTorsionType(E):
    when "Generic" : GenericExamples cat:= [<E,P>];
    when "2Sylow" : SylowExamples cat:= [<E,P>];
    when "Dihedral" : DihedralExamples cat:= [<E,P>];
    when "Generic3Isogeny" : IsogenyExamples cat:= [<E,P>];
    when "Z/3Z-nonsplit" : NonSplitExamples1 cat:= [<E,P>];
    when "mu3-nonsplit" : NonSplitExamples2 cat:= [<E,P>];
    when "Diagonal" : DiagonalExamples cat:= [<E,P>];
    when "mu3+Z/3Z" : SplitExamples cat:= [<E,P>];
  end case;
end while;

assert #GenericExamples eq 100;
assert #SylowExamples eq 100;
assert #DihedralExamples eq 100;
assert #IsogenyExamples eq 100;
assert #NonSplitExamples1 eq 100;
assert #NonSplitExamples2 eq 100;
assert #DiagonalExamples eq 100;
assert #SplitExamples eq 100;

threetorsiontypes := [ 
  "Generic", "2Sylow", "Dihedral", "Generic3Isogeny",
  "Z/3Z-nonsplit", "mu3-nonsplit", "Diagonal","mu3+Z/3Z"];

Q := Rationals();
R<x,y,z> := PolynomialRing(Q,3); 

for 
    i in [1 .. NUMTESTS], 
    type in threetorsiontypes,
    dothreedescent in [false,true]
do

  printf "\n//////////////////////////////////////////////\n"*
         "%o, dothreedescent = %o, test #%o\n\n",
          type, dothreedescent, i;

  time0 := Cputime();

  case type :
    when "Dihedral" : data := Random(DihedralExamples);
    when "2Sylow" : data := Random(SylowExamples);
    when "Generic" : data := Random(GenericExamples);
    when "Generic3Isogeny" : data := Random(IsogenyExamples);
    when "Z/3Z-nonsplit" : data := Random(NonSplitExamples1);
    when "mu3-nonsplit" : data := Random(NonSplitExamples2);
    when "Diagonal" : data := Random(DiagonalExamples);
    when "mu3+Z/3Z" : data := Random(SplitExamples);
  end case;

  E := data[1];
  assert ThreeTorsionType(E) eq type;
  
  if dothreedescent then 
    S3,S3map := ThreeSelmerGroup(E);
    vecs := {};
    for v in S3 do
      if -v notin vecs and v ne S3!0 then 
        vecs := vecs join {v};
      end if;
    end for;
    assert #vecs eq (#S3-1)/2;
    alphas := [S3map(v): v in vecs];
  else
    P := data[2];
    D := 2*Divisor(E!0)+Divisor(E!P);
    C := Image(DivisorMap(D));
    R<x,y,z> := PolynomialRing(Rationals(),3);
    cubic := R!Equations(C)[1];
    cubic *:= LCM([Denominator(x): x in Coefficients(cubic)]);
    cubic := Minimise(cubic);
    cubic := Reduce(cubic);
    alpha := ThreeSelmerElement(E,cubic);
    alphas := [alpha];
  end if;

  for alpha in alphas do
    cubics := [];
    for method in [ "HessePencil",
                    Random(["FlexAlgebra","SegreEmbedding"]) ] do
      cc := R!Equation(ThreeDescentCubic(E,alpha:Method:=method));
      cubics cat:= [cc];
      assert IsIsomorphic(Jacobian(cc),E);
    end for;
    flag,g:= IsEquivalent(cubics[1],cubics[2]:E:=E);
    assert flag;

  end for;

  printf "\nTotal time for test: %os\n", Cputime(time0); 

end for;


