freeze;

// This function assumes that E is a field!!!
function identify_field(E)
  F:= BaseRing(E);
  d:= Dimension(E);
  if d eq 1 then return hom< F -> E | >; end if;
  B:= Basis(E);
  P:= PolynomialRing(F);

  // add more if you want
  if Type(F) in {FldRat, FldFunRat, FldFun} or ISA(Type(F), FldAlg) then
    rand:= func< F, r | Random(F, r) >;
  elif Type(F) eq FldFin then
    rand:= func< F, r | Random(F) >;
  else
    error "Base field is not supported";
  end if;

  repeat
    x:= &+[ rand(F, 10) * b : b in B];
    m:= P ! MinimalPolynomial(x);
    if m mod P.1 eq 0 then
      m div:= P.1;
    end if;
  until Degree(m) eq d;
  K:= ext< F | m >;
  return hom< K -> E | x >;
end function;

// This function assumes that A is simple
function algebra_over_center(A)
  mats:= Type(A) eq AlgMat;
  B:= Basis(A);
  C:= Center(A);
  F:= BaseRing(A);
  h:= identify_field(C); K:= Domain(h);
  CB:= [ x @ h : x in Basis(K) ];
  m:= #CB;
 
  BB:= [ ];
  BK:= [];
  V:= VectorSpace(F, mats select Degree(A)^2 else Dimension(A) );
  W:= sub< V | >;
  i:= 0;
  while #BB ne #B do
    repeat
      i+:= 1;
      x:= Vector(F, Eltseq(B[i]));
     until x notin W;
    Append(~BK, B[i]);
    BB cat:= [ b*B[i]: b in CB ];
    W:= sub< V | [Eltseq(x) : x in BB] >;
  end while;

  // get structure constants over K
  T:= Matrix(F, [ Coordinates(A, b) : b in BB ] )^-1;
  SC:= [ [ [ K ! x : x in Partition(Eltseq(Vector(F, Coordinates(A, c*b)) * T), m) ] : b in BK ] : c in BK ];

  AA:= AssociativeAlgebra< K, #SC | SC >;
  hh:= map< A -> AA | x:-> [ (K ! P[i]) : i in [1..#P] ] where P:= Partition(Eltseq(Vector(F, Eltseq(Coordinates(A, x))) * T), m),
                      y:-> &+ [ BB[i] * e[i] : i in [1..#BB] ] where e:= &cat [ Eltseq(x) : x in Eltseq(y) ] >;

  return AA, hh;
end function;

intrinsic AlgebraOverCenter( A :: Alg ) -> AlgAss, Map
{Given a simple algebra A with center K, returns a K-isomorphic K-algebra B and a K-isomorphism from A to B}
  F:= BaseRing(A);
  T:=  Type(A);
  require T eq AlgMat or ISA(T, AlgAss) : "The algebra must be of type AlgAss, AlgMat or AlgClff";
  require IsField(F) : "Base ring must be a field";
  require #CentralIdempotents(A) eq 1 : "The algebra is not simple";
  if Dimension(Center(A)) eq 1 then 
    if T eq AlgAss then
      return A, hom< A -> A | x:-> x, y:-> y >;
    elif T eq AlgMat then
      return Algebra(A);
    else
      B:= AssociativeAlgebra(A);
      return B, Coercion(A, B);
    end if;
  end if;
  return algebra_over_center(A); 
end intrinsic;
