freeze;

intrinsic ConstaCyclicCode(n::RngIntElt, f::RngUPolElt, c::FldFinElt) -> Code
{Return the length n 
code generated by consta-cyclic shifts by c of the coefficients
of f};

//check for c in the right field, (with f) and n being longer than f
    F := BaseRing(f);
    require Type(F) eq FldFin : "f must be over a finite field";
    require Parent(c) eq F: "f must be over the same finite field as c";
    require c ne 0: "The constant c must be non-zero";

	//reduce f mod x^n - c
    f mod:= (Parent(f).1)^n - c;
 
    coeffs := Coefficients(f);
    coeffs cat:= [0 : i in [1.. n - #coeffs]];
        //cyclic shifter
    gen := Matrix(F,1,n,coeffs);
 
    newcoeffs := coeffs;
    repeat
        newcoeffs := [c*newcoeffs[n]] cat newcoeffs[1..n-1];
        gen := VerticalJoin(gen,Matrix(F,1,n,newcoeffs));
    until newcoeffs eq coeffs;
 
    return LinearCode(gen);
 
end intrinsic;

