//getdata.m 5/14/09

freeze;

import "hecke.m":
    hecke_from_mats,
    hecke_mats,
    TpMatrix,
    Tpp_mats;

import "cohomology.m":
    HomologyData;

import "hecke.m":
    Tab_mats;

function cuspidal_polynomial_part(f,E)
    //twists seem to show up, so have to take off any factor that is root of
    //root of unity times E
    P<x>:=Parent(f);
    R<z>:=PolynomialRing(SplittingField(f));
    g:=R!f;
    L:=Roots(g);
    for a in L do
	m:=a[2];
	c:=a[1];
	if IsRootOfUnity(c/E) then
	    g:=ExactQuotient(g, (z-c)^m);
	end if;
    end for;
    
    /*
    while IsDivisibleBy(g,x-E) do
	g:=ExactQuotient(g, x-E);
    end while;
    */

    /*
    //Can we loop over all eigenvalues of norm = N(p) + 1?
    //Do this until we figure out something better.
    for i in [1..Degree(f)] do
	h:=GCD(g,x^i - E^i);
	while IsDivisibleBy(g,h) and h ne 1 do
	    g:=ExactQuotient(g, h);
	end while;
    end for;
    */    
    return P!g;
end function;

function is_square(p)
    //returns boolean true if p is a square in the class group. 
    C,f:=ClassGroup(Order(p));
    two:=hom<C->C | [2*g:g in Generators(C)]>;
    return p@@f in Image(two);
end function;

/*
intrinsic NumberOfCusps(M:ModFrmBianchi)->RngIntElt
    {Given space of cusp forms, computes the number of cusps}
    level:=Level(M);
    K:=BaseField(M);
    h:=ClassNumber(K);
    D:=Divisors(level);
    MM:=[d + level*d^(-1):d in D];
    ans:=0;
    for m in MM do 
	R,mu:=quo<Integers(K)|m>;
	U:={mu(a):a in M`ModFrmData`torsion_units};
	Rs,mus:=UnitGroup(R);
	RsU:=quo<Rs|[a@@mus:a in U]>;
	ans:=ans + #RsU;
    end for;
    return h* ans;
end intrinsic;
*/

procedure set_cuspidal_space(M)
    //looks to see if Tp for principal prime has been computed.
    //If yes, then use it.
    HomologyData(M);
    F:=BaseField(M);
    if not assigned M`basis_matrix then
	if FullDimension(M) eq 0 then //only for fields
	    H:=M`LevelData`homology`H;
	    cuspspace,cm:=sub<H|H>;
	else
	    if #M`EHecke ne 0 then 
		tup:=Representative(M`EHecke);
		P:=tup[1];
		Tp:=tup[2];
	    else
		
		//Find principal prime not dividing level.
		//Should be written better since it assumes there is a
		//principal prime of norm less than 200 that is prime to
		//the level.
		t:=1;
		PP:=[];
		while #PP eq 0 do 
		    PP:=PrincipalPrimesUpTo(200*t,BaseField(M):coprime_to:=Level(M));
		    t:=t+1;
		end while;
		
		_,i:=Minimum([Norm(a): a in PP]);
		P:=PP[i];
		Tp:=TpMatrix(M,P);
	    end if;
	    f:=CharacteristicPolynomial(Tp);
	    g:=cuspidal_polynomial_part(f,Norm(P)+1);
	    ET:=Evaluate(g,Tp);
	    H:=M`LevelData`homology`H;
	    cuspspace,cm:=sub<H|Nullspace(ET)>;
	    	    
	    vprintf Bianchi,1:"Cuspidal space is %o.\n",cuspspace;
	end if;
	bm:=BasisMatrix(cuspspace);
	M`basis_matrix := BasisMatrix(cuspspace);
	I:=IdentityMatrix(CoefficientRing(bm),NumberOfRows(bm)); 
	M`basis_matrix_inv:=
	    Transpose(Solution(Transpose(M`basis_matrix),I));
	//M`LevelData`homology`cuspidal_space_map:=cm;  Don't need this anymore.
    end if;
end procedure;

forward TaaTpMatrix;


function heckematrixprincipalprime(M,n)
    //Computes Hecke matrix Tn restricted to cuspidal space.  Assumes p is
    //principal, and either prime or the product of two primes.
    
    for tup in M`Hecke do
	    if n eq tup[1] then return tup[2]; end if;
    end for;
    
    // TO DO: when possible, factor n as a product of coprime principal ideals,
    // and call recursively

    b:=M`basis_matrix;
    bi:=M`basis_matrix_inv;
    //if we have computed a hecke matrix for the whole
    //cohomology space, use this instead.
    for tup in M`EHecke do
	if n eq tup[1] then
	    Tp:=tup[2];
	    L:=[];
	    for i in [1..NumberOfRows(b)] do
		Append(~L,b[i]*Tp);
	    end for;
	    T:=Matrix(L) * bi;
	    Append(~M`Hecke,<n,T>);
	    return T;
	end if;
    end for;
    //if we get here, we must compute.
    T:=hecke_from_mats(M,hecke_mats(M,n),b) * bi;
    Append(~M`Hecke,<n,T>);
    return T;
end function;

intrinsic HeckeMatrixBianchi(M::ModFrmBianchi,n::RngOrdIdl)->Mtrx
    {Computes Hecke matrix Tn restricted to cuspidal space.  Assumes p is
    principal, and either prime or the product of two primes.  Alternatively,
    if p is prime and is a square in the class group, computes T_a,a*Tp for
    suitable a.  This will return T_p, since T_a,a is trivial on the space we
    are considering.  See Cremona preprint.}
    set_cuspidal_space(M);
    d:=Dimension(M);
   
    error if GCD(n, Level(M)) ne 1*Order(n), 
	"The second argument must be an ideal coprime to the level of the space";
    if IsPrincipal(n) then
	T:=heckematrixprincipalprime(M,n);
    elif is_square(n) then
	T:=TaaTpMatrix(M,n);
    else
	error "HeckeMatrixBianchi not implemented for this type of ideal.  Ideal is not principal or square in the class group.";
    end if;
    
    return T;
end intrinsic;


intrinsic TppMatrix(M::ModFrmBianchi,p::RngOrdIdl)->Mtrx
    {Computes Hecke matrix Tp,p restricted to cuspidal space.  Assumes p
    is non-principal, and prime with p^2 principal.}
    //if we have computed a hecke matrix for the whole
    //cohomology space, use this instead.
    error if not IsPrincipal(p*p), "The second argument must square to a principal ideal.";
    b:=M`basis_matrix;
    bi:=M`basis_matrix_inv;
    T:=hecke_from_mats(M,Tpp_mats(M,p),b) * bi;
    Append(~M`Hecke_pp,<p,T>);
    return T;
end intrinsic;

function get_good_a(M,p)
    //Finds an ideal a, prime to the level such that a^2*p is principal.
    //Assumes that class group always chooses the same isomorphism
    
    C,f:=ClassGroup(BaseField(M));
    g:=p@@f;
    if IsDefined(M`LevelData`good_a,Eltseq(g)) then 
	return M`LevelData`good_a[Eltseq(g)];
    end if;
    bound:=100;
    N:=Level(M);
    while true do
	//currently cand_list just uses primes.  It might be better to
	//range over composite as will to try to find smaller a.
	cand_list:=PrimesUpTo(bound,BaseField(M));
	for a in cand_list do
	    if &and[(a*a*p)@@f eq C!0, GCD(a,N) eq 1*Order(a)] then
		return a;
	    end if;
	end for;
	bound:=bound+100;
    end while;
end function;

function TaaTpMatrix(M,p)
    //TaaTpMatrix(M::ModFrmBianchi,p::RngOrdIdl)->.
    //Given a prime p, whose ideal class is a square,
    //returns T_a,aT_p, where a is chosen
    //so that a^2p is principal.  This should give the eigenvalues
    //for T_p, I think.  Check Cremona preprint.

    for tup in M`Hecke_ap do
	if p eq tup[1] then return tup[2]; end if;
    end for;
    a:=get_good_a(M,p);
    mats:=Tab_mats(M,a*p,a);
    b:=M`basis_matrix;
    bi:=M`basis_matrix_inv;
    vprintf Bianchi,3: "matrices are %o.\n", mats;
    T:=hecke_from_mats(M,mats,b) * bi;
    Append(~M`Hecke_ap,<p,T>);
    return T;
end function;


