//findcells.m

freeze;

import "defs.m": CellRec;
import "findperfectform.m":
    ipm,
    IsUniqueForm,
    IsPerfectForm,
    ToLat;
import "setup.m":
    act_on_cone,
    StabilizerOfO2Vectors;

import "sharbly.m":
    unimodularize;


function IsGLConj(M,S,T)
    //Given 2 lists of O2 vectors returns a g which takes S to T,
    //up to order and torsion units
    Tor:=M`ModFrmData`torsion_units;
    F:=BaseField(M);
    O:=MaximalOrder(F);
    
    SS:=S;
    TT:={};
    
    for e in Tor, v in T do
	if e*v[1] in O and e*v[2] in O then
	    Include(~TT,Vector(O,[e*v[1],e*v[2]]));
	end if;
    end for;
    //TT:={e*v:e in Tor, v in T};
   
    unitf:=exists(A){A:a,b in SS|IsInvertible(A)
	where A:=ChangeRing(Matrix([a,b]),F)};
    in_G:=M`ModFrmData`in_G;
    Stostd:=A^(-1);
    for a in TT do
	for b in TT diff {a,-a} do 
	    g:=Stostd*B where B:=Matrix([a,b]);
	    if in_G(g) then
		g:=ChangeRing(g,O);
		if {v*g:v in SS} subset TT then
		    return true,Transpose(g);
		end if;
	    end if;
	end for;
    end for;
    return false,_;
    //if it gets here then not GL conjugate.
end function;


function g_on_cone(M,g)
    //given a 2x2 matrix g, returns the 4x4 matrix representing the
    //action of g on the cone.
    B:=Basis(M`ModFrmData`cone_space);//basis of the cone
    L:=[act_on_cone(M,g,x):x in B];
    G:=Matrix(L);
    assert Determinant(G) eq 1;
    return G;
end function;


function GetSign(F,g,L,NT)
    //given g and two lists of O2 vectors, returns the sign of
    //the permutation induced on the vectors.  Assumes the lists are size 2.
    NL:=[ToBianchiCone(F,v*Transpose(g)):v in L];
    g:=Sym(#L)![Position(NL,a):a in NT];
    return Sign(g);
end function;

function Resort(cell)
    //resorts cell cone_vertices and O2_vertices so that the ordering matches
    //the ordering of the vertices of the polytope.
    newcell:=cell;
    cv:=cell`cone_vertices;
    ov:=cell`O2_vertices;
    v:=[Eltseq(a):a in Vertices(cell`polytope)];
    
    ncv:=[cv[Position(cv,v[i])]:i in [1..#v]];
    nov:=[ov[Position(cv,v[i])]:i in [1..#v]];
    
    newcell`cone_vertices:=ncv;
    newcell`O2_vertices:=nov;
    return newcell;
end function;


function MakeCell(M,minvecs,O2minvecs)
    O2uni:=[unimodularize(M,a):a in O2minvecs];
    cell:=rec<CellRec|
	cone_vertices:=minvecs,
	//O2_vertices:=O2minvecs>;
    O2_vertices:=O2uni>;
    cell`polytope:=Polytope([Eltseq(a):a in minvecs]:areVertices:=true);
    //cell`polytope:=Polytope([Eltseq(a):a in minvecs]);
    cell:=Resort(cell);
    return cell;
end function;

function GetStabilizer(M,cell)
    O2minvecs:=cell`O2_vertices;
    G:=StabilizerOfO2Vectors(M,O2minvecs);
    return G;
end function;

function GetStabilizerPlus(M,cell)
    O2minvecs:=cell`O2_vertices;
    G:=cell`stabilizer;
    //L:=[g_on_cone(M,g):g in G];
    Gplus:=[g:g in G|
	GetSign(BaseField(M),g,O2minvecs,cell`cone_vertices) eq 1];
    return Gplus;
end function;

function IsPositiveForm(M,form)
    Q:=ipm(M,form);
    return IsPositiveDefinite(Q);
end function;

    
function vIJ(M,I,E,phiI)
    V:=M`ModFrmData`cone_space;
    B,newform:=IsUniqueForm(M,E);
    vcand:=V!B[1];
    tflist:=[(V!vcand,V!x) ge 0:x in I];
    if &and tflist then
	v:=vcand;
    else v:=-vcand;
    end if;
    return v;
end function;

function Findphiell(M,phiI,rho,v)
    //Given a field F and a perfect form, return the phi_l vector
    //as described by Gunnells
    step:=1;
    F:=BaseField(M);
    O:=MaximalOrder(F);
    B:=Basis(O);
    BB:=BasisMatrix(O);
    L:=ToLat(M,phiI);
    lower:=Minimum(L);
    C:=M`ModFrmData`cone_space;
    while true do
	//Check if this is necesary to do
	P:=ShortVectorsProcess(L,lower,lower+step);
	norm:=lower;
	//Skips norm a and skips the vectors already done
	while norm eq lower do 
	    nv,norm:=NextVector(P);
	end while;
	lower+:=step;
	while norm ne -1 do
	    O2temp:=[O|nv[1]*B[1]+nv[2]*B[2], nv[3]*B[1]+ nv[4]*B[2]];
	    a:=O2temp[1]*F`Conjugate(O2temp[1]);
	    c:=O2temp[2]*F`Conjugate(O2temp[2]);
	    b:=Vector(Eltseq(F!(O2temp[1]*F`Conjugate(O2temp[2]))) )*BB^(-1);
	    temp:= [RationalField()|a,b[1],b[2],c];
	    iptemp:=(C!v,C!temp);
	    if iptemp lt 0 then
		rhox:=rho(temp);
		templ:=phiI+rhox*v;
		if IsPositiveForm(M,templ) then
		    //templ is the correct choice
		    return templ;
		end if;
	    end if;
	    nv,norm:=NextVector(P);
	end while;
    end while;
end function;

function GetNeighbor(M,cell,facet)
    //Given F, cell record cell, and list of vertices defining a facet,
    //compute the neighboring cell and return as a a cell record;
    F:=BaseField(M);
    vprint Bianchi, 3: "Finding neighbor across facet:",facet;
    V:=M`ModFrmData`cone_space;
    I:=cell`cone_vertices;
    phiI:=cell`form;
    E:=[cell`cone_vertices[i]:i in facet];
    v:=vIJ(M,I,E,phiI);
    
    rho:=func<x|(1-(V!x,V!phiI))/(V!x,V!v)>;
    
    phiell:=Findphiell(M,phiI,rho,v);
    Lat:=ToLat(M,phiell);
    TO2:=ShortVectors(Lat,1);
    
    B:=Basis(Integers(F));
    O2templist:=[[nv[1,1]*B[1]+nv[1,2]*B[2], nv[1,3]*B[1]+ nv[1,4]*B[2]]:
	nv in TO2];
    T:=[ToBianchiCone(F,v):v in O2templist];
    T:=[z:z in T|not z in E];
    
    min:=Minimum([rho(z): z in T]);
    Z:=[z:z in T|rho(z) eq min];
    L:=E cat Z;
    tf,neighborform,neighborminvecs,neighborO2minvecs:=IsPerfectForm(M,L);
    assert tf;
    neighbor:=MakeCell(M,neighborminvecs,neighborO2minvecs);
    neighbor`form:=neighborform;
    
    return neighbor;
end function;

function f(F,x)
    ans:=0;
    L:=x`cone_vertices;
    e1:=[1,0,0,0];
    e2:=[0,0,0,1];
    e3:=[1,1,0,1];
    if e1 in L then ans:=100;
	if e2 in L then ans:=ans+50;
	    if e3 in L then ans:=ans+25;
	    end if;
	end if;
    end if;
    return ans;
end function;

function get_mover(M,stdtypes,neighborvecs)
    //given stdtype and neighborvecs as O2vectors finds the GL type
    //of neighborvecs and returns a gamma which sends neighborvecs to stdvecs

    tf:=exists(type){i:i in [1..#stdtypes]|
	IsGLConj(M,neighborvecs,stdtypes[i])};
    assert tf;
    _,gamma:=IsGLConj(M,neighborvecs,stdtypes[type]);
    return <type,gamma>;
end function;

    
function get_neighbor_movers(M,three_poly)
    //Attaches neighbors_mover list <i,gamma>, where i tells the GL type
    //of the cell and gamma tells the mover to standard_i
    three_poly_copy:=[];
    std_types:=[std`O2_vertices:std in three_poly];
    for std in three_poly do
	newstd:=std;
	newstd`neighbors_mover:=[get_mover(M,std_types,neighbor`O2_vertices):
	    neighbor in std`neighbors];
	Append(~three_poly_copy,newstd);
    end for;
    return three_poly_copy;
end function;

function Neighbors(M,cell)
    //Assumes top cone
    facetlist:=FacetIndices(cell`polytope);
    neighborlist:=[GetNeighbor(M,cell,facet):facet in facetlist];
    return neighborlist;
end function;

function get_edges(p)
    //assumes p is 3 dimensional polytope.
    f:=Facets(p);
    V:=Vertices(p);
    edges:={};
    for face in f do
	e:=Facets(face);
	newedges:={Vertices(a):a in e};
	edges:=edges join newedges;
    end for;
    return {[Position(V,a[1]),Position(V,a[2])]:a in edges};
    
end function;

    
function graph_of_edges(p)
    edges:=get_edges(p);
    G:=Graph<#Vertices(p)|>;
    v:=Vertices(G);
    for e in edges do
	AddEdge(~G,v[e[1]],v[e[2]]);
    end for;
    return G;
end function;

function get_graphs(three_poly)
    //attaches graphs of edges to the standard 3 dimensional polytopes.
    three_poly_copy:=[];
    for p in three_poly do
	pnew:=p;
	pnew`graph:=graph_of_edges(p`polytope);
	Append(~three_poly_copy,pnew);
    end for;
    return three_poly_copy;
end function;

    
procedure MakeThreePoly(M)
    //Given F, attached ModFrmData of three_poly, starting with the
    //initial perfect form
    vprint Bianchi, 1: "Finding 3-dimensional cells.";
    phi:=M`ModFrmData`initial_form;
    F:=BaseField(M);
    initialcell:=MakeCell(M,phi`minimal_vectors,phi`O2minimal_vectors);
    initialcell`form:=phi`form;
    todo:=[initialcell];
    three_poly:=[];
    
    while #todo ne 0 do
	central:=todo[1];
	tf:=exists(cell){c:c in three_poly|
	    IsGLConj(M,c`O2_vertices,central`O2_vertices)};
	if not tf then
	    central`stabilizer:=GetStabilizer(M,central);
	    central`stabilizer_plus:=GetStabilizerPlus(M,central);
	    N:=Neighbors(M,central);
	    central`neighbors:=N;
	    Append(~three_poly,central);
	    todo:=todo cat N;
	end if;
	todo:=todo[2..#todo];
	vprintf Bianchi, 3:
	    "     There are %o left top cones left to check.\n",#todo;
	vprintf Bianchi, 3:
	    "       Found %o GL-classes of top cones so far.\n",#three_poly;
    end while;
    three_poly:=get_neighbor_movers(M,three_poly);
    three_poly:=get_graphs(three_poly);
    M`ModFrmData`three_poly:=three_poly;
    vprintf Bianchi, 1: "  Found %o 3-dimensional cells.\n",#three_poly;
end procedure;
    
function AllTwoPoly(M)
    //Given F, computes all of the two dimensional polytopes that arise as
    //the facets of the three polytopes.  This later gets fed to a GLconj
    //class sorter
    threepoly:=M`ModFrmData`three_poly;
    F:=BaseField(M);
    all:=[];
    for cell in threepoly do
	cv:=cell`cone_vertices;
	ov:=cell`O2_vertices;
	poly:=cell`polytope;
	for a in FacetIndices(poly) do
	    newvecs:=[cv[i]:i in a];
	    newovecs:=[ov[i]:i in a];
	    newface:=MakeCell(M,newvecs,newovecs);
	    
	    Append(~all,newface);
	end for;
    end for;
    vprintf Bianchi,3 :"     There is at most %o 2-dimensional cells.\n",#all;
    Sort(~all,func<x,y | f(F,y)-f(F,x)>);
    return all;
end function;

forward OrientedFacets;

procedure MakeTwoPoly(M)
    //Given F, attach ModFrmData of two_poly, three_poly already computed
    vprint Bianchi, 1: "Finding 2-dimensional cells.";
    F:=BaseField(M);
    twopoly:=[];
    todo:=AllTwoPoly(M);
    while #todo ne 0 do
	cell:=todo[1];
	if not exists(c){c:c in twopoly|
	    IsGLConj(M,cell`O2_vertices,c`O2_vertices)} then
	    cell`stabilizer:=GetStabilizer(M,cell);
	    cell`stabilizer_plus:=GetStabilizerPlus(M,cell);
	    cell`oriented_facets:=OrientedFacets(cell`polytope);
	    
	    Append(~twopoly,cell);
	end if;
	todo:=todo[2..#todo];
    end while;
    M`ModFrmData`two_poly:=twopoly;
    vprintf Bianchi, 1: "  Found %o 2-dimensional cells.\n",#twopoly;
end procedure;

function AllOnePoly(M)
    //Given F, computes all of the one dimensional polytopes that arise
    //as the facets of the two polytopes.  This later gets fed to a GLconj
    //class sorter
    twopoly:=M`ModFrmData`two_poly;
    all:=[];
    for cell in twopoly do
	cv:=cell`cone_vertices;
	ov:=cell`O2_vertices;
	poly:=cell`polytope;
	for a in FacetIndices(poly) do
	    newvecs:=[cv[i]:i in a];
	    newovecs:=[ov[i]:i in a];
	    newface:=MakeCell(M,newvecs,newovecs);
	    //don't need this.  already done in MakeCell
	    /*
	    newface`polytope:=Polytope([Eltseq(a):a in newvecs]:
	    areVertices:=true);
	    //newface`polytope:=Polytope([Eltseq(a):a in newvecs]);
	    */
	    Append(~all,newface);
	end for;
    end for;
    vprintf Bianchi, 3:
	"     There is at most %o 1-dimensional cells.\n",#all;
    F:=BaseField(M);
    Sort(~all,func<x,y | f(F,y)-f(F,x)>);
    return all;
end function;

procedure MakeOnePoly(M)
    //Given M, attach ModFrmData of one_poly, two_poly already computed
    F:=BaseField(M);
    vprint Bianchi, 1: "Finding 1-dimensional cells.";
    onepoly:=[];
    todo:=AllOnePoly(M);
    while #todo ne 0 do
	cell:=todo[1];
	if not exists(c){c:c in onepoly|
	    IsGLConj(M,cell`O2_vertices,c`O2_vertices)} then
	    cell`stabilizer:=GetStabilizer(M,cell);
	    cell`stabilizer_plus:=GetStabilizerPlus(M,cell);
	    
	    Append(~onepoly,cell);
	end if;
	Remove(~todo,1);
    end while;
    M`ModFrmData`one_poly:=onepoly;
    vprintf Bianchi, 1: "  Found %o 1-dimensional cells.\n",#onepoly;
end procedure;

function GLtype(M,GLclasses,fvecs,cvecs)
    //g sends std to fvecs
    for type in [1..#GLclasses] do
	L:=GLclasses[type];
	
	tf,g:=IsGLConj(M,L,fvecs);
	if tf then 
	    sgn:=GetSign(BaseField(M),g,L,cvecs);
	    return <type,sgn,g>;
	end if;
    end for;
    error "No type found.";
end function;

function GLFacetTypes(M,cell)  
    //list of <a,sgn,g>, where a gives the GL class of edge that this edge
    //is conjugate to, and sgn is the orientation with respect to the choice
    //made for standard. g sends std to cell
    facettypes:=[];
    OF:=cell`oriented_facets;
    O2vecs:=cell`O2_vertices;
    GLclasses:=[a`O2_vertices:a in M`ModFrmData`one_poly];
    for f in OF do
	fvecs:=[O2vecs[a]:a in f];
	//O2vectors corresponding to facet f
	cvecs:=[cell`cone_vertices[a]:a in f];
	//O2vectors corresponding to facet f	
	Append(~facettypes,GLtype(M,GLclasses,fvecs,cvecs));
	//Appends <type,+/-,g>
    end for;
    return facettypes;
end function;


function OrientedFacets(poly)
    //Given a 2-dimensional polytope poly, returns a list of lists
    //representing the facets.  The ordering on the vertices is counter-
    //clockwise, as viewed from the outside.
    //Does this make sense?
    //Doesn't matter whether it is correct or reversed, since we only
    //want relations.
    if not assigned poly`oriented_facets then 
	ver:=Vertices(poly);
	facets:=[Setseq(a):a in FacetIndices(poly)];
	todo:=facets;
	ofacets:=[todo[1]];
	Remove(~todo,1);
	while #todo ne 0 do
	    current:=ofacets[#ofacets];
	    tf:=exists(nextfacet){t:t in todo| current[2] in t};
	    Exclude(~todo,nextfacet);
	    assert tf;
	    if nextfacet[1] eq current[2] then  
		Append(~ofacets,nextfacet);
	    else
		Append(~ofacets,[nextfacet[2],nextfacet[1]]);
	    end if;
	end while;
	poly`oriented_facets:=ofacets;
    end if;
    
    return poly`oriented_facets;
end function;

	

procedure FindCells(M)
    //Given F, computes the polytopes up to GL.
    F:=BaseField(M);
    if not assigned M`ModFrmData`three_poly then 
	MakeThreePoly(M);
	MakeTwoPoly(M);
	MakeOnePoly(M);
	for c  in [1..#M`ModFrmData`two_poly] do
	    cell:=	M`ModFrmData`two_poly[c];
	    M`ModFrmData`two_poly[c]`GL_facet_types:=GLFacetTypes(M,cell);
	end for;
	L:=&join{{e*Determinant(Matrix([a,b])):
	    a,b in cell`O2_vertices,e in M`ModFrmData`torsion_units}:
	    cell in M`ModFrmData`three_poly};
	M`ModFrmData`dets:=L;
    end if;
end procedure;
