freeze;

forward LowIndexSubgroupsH, ExtendLowIndexSubgroupsH;
import "extend1sub.m":ExtendOneSubgroupH;

LowIndexSubgroupsRecursive := function(G,maxind :
                                           Presentation:=false, Print:=0)
/* Should only be called on TF-group with composition factors not in database.
 */
  local maxsi, si, stabpt, stabsubs;

  maxsi := 1;
  stabpt := 1;
  for o in {Representative(o): o in Orbits(G)} do
    si := Index(G,Stabiliser(G,o));
    if si gt maxsi then
      maxsi := si;
      stabpt := o;
    end if;
  end for;

  if Print gt 0 then
    print
       "Composition factor not in database. Trying stabilizer of point",stabpt;
  end if;

  stabsubs := LowIndexSubgroupsH(Stabiliser(G,stabpt),maxind: Print:=Print);

  if Print gt 0 then
    print "Found",#stabsubs,"subgroups of stabiliser";
  end if;

  return ExtendLowIndexSubgroupsH(G,stabpt,maxind:
                   Presentation:=Presentation, Print:=Print, ss:=stabsubs);

end function;


LowIndexSubgroupsH := function(G,maxind: Presentation:=false, Print:=0)
/* Subgroup records are returned.
 * We do it first for the radical quotient, and then use lifting.
 * If presentation is true, then presentations are computed.
 */ 
  local lisubs, lisubsl, H, Hind, ind, new, calcpres, ct,
        L, Q, pQ, oQ, S, M, MM, mm, s, Res, f, phi, nms, RF;

  L:=ElementaryAbelianSeries(G);
  if Print ge 1 then
     print "Elementary abelian series of length:",#L-1;
     print "Layer sizes:", [Index(L[i],L[i+1]): i in [1..#L-1]];
  end if;

  /* First get subgroups of radical quotient */
  
  Q, pQ := RadicalQuotient(G);
  oQ := #Q;
  calcpres := #L eq 1 select Presentation else true;

  if L[1] eq G then // group is soluble
    lisubs := Subgroups(Q:Presentation:=calcpres);
  elif oQ le 100000 then
    M := Subgroups(Q : Presentation:=calcpres, IndexLimit := maxind);
    assert forall{m : m in M | (oQ div m`order) le maxind};
    lisubs := [m : m in M | (oQ div m`order) le maxind];
    Sort(~lisubs,func<x,y|y`order-x`order>);
  elif HasComputableSubgroups(Q) then 
    M := MaximalSubgroupsH(Q,sub<Q|>:Presentation:=calcpres, Print:=Print);
    lisubs := [m : m in M | (oQ div m`order) le maxind];
    /* Need to append Q itself */
    RF := Format(M[1]);
    mm := rec <RF | subgroup:=Q, length:=1, order := #Q >;
/*
    if calcpres then
        f,phi := FPGroup(Q:StrongGenerators:=true);
        mm`presentation := f;
        mm`subgroup := sub<Q | [phi(f.i) : i in [1..Ngens(f)] ]>;
    end if;
*/
    Append(~lisubs,mm);
    Sort(~lisubs,func<x,y|y`order-x`order>);
    ct := 2;
    while ct le #lisubs do
      H := lisubs[ct]`subgroup;
      if Print gt 1 then
        print ct, "Index:", Index(Q,H);
      end if;
      if 2*Index(Q,H) gt maxind then
        break;
      end if;
      M := MaximalSubgroupsH(H,sub<H|>:Presentation:=calcpres, Print:=Print);
      for i in [1..#M] do
        S := M[i]`subgroup;
        ind := Index(Q,S);
        if ind le maxind then
          new :=true;
          for j in [ct+1..#lisubs] do
            K := lisubs[j]`subgroup;
            if IsConjugate(Q,S,K) then
              new:=false;
              break;
            end if;
          end for;
          if new then
            M[i]`length := Index(Q,Normaliser(Q,S));
            Append(~lisubs,M[i]);
            if Print gt 1 then
              print "  New: index",ind;
            end if;
            if ind lt Index(Q,lisubs[#lisubs-1]`subgroup) then
               Sort(~lisubs,func<x,y|y`order-x`order>);
            end if;
          end if;
        end if;
      end for;
      ct := ct+1;
    end while;
  else lisubs := LowIndexSubgroupsRecursive(Q,maxind:
                                        Print:=Print, Presentation:=calcpres);
  end if;

  //Presentation option of Maximal Subgroups is not always working!
  if calcpres then
    for i in [1..#lisubs] do
      if not assigned lisubs[i]`presentation then
        f,phi := FPGroupStrong(lisubs[i]`subgroup);
        lisubs[i]`presentation := f;
        lisubs[i]`subgroup := sub<Q | [phi(f.i) : i in [1..Ngens(f)] ]>;
      end if;
    end for;
  end if;

  RF := recformat<order, length, subgroup, presentation, index>;
  lisubsl := [];
  for i in [1..#lisubs] do
    newrec := rec< RF | >;
    newrec`order := lisubs[i]`order;
    newrec`length := lisubs[i]`length;
    if calcpres then
      newrec`presentation := lisubs[i]`presentation;
    end if;
    S := lisubs[i]`subgroup;
    newrec`subgroup :=
                      sub< G | [(S.i) @@ pQ : i in [1..Ngens(S)] ] >; 
    newrec`index := Index(G, sub<G | newrec`subgroup, L[1] > );
    Append(~lisubsl, newrec);
  end for;
  lisubs := lisubsl;

  lisubsfinal := [];
  for i in [1..#L-1] do
    calcpres := i eq #L-1 select Presentation else true;
    lisubsl := [];
    if Print ge 1 then
      print "Extending through layer size ",#L[i]/#L[i+1];
      print #lisubs,"subgroups";
      print "orders:", [subrec`order : subrec in lisubs];
    end if;
    ct:=0;
    for subrec in lisubs do
      ct +:= 1;
      if Print gt 1 then
        print "Lifting subgroup number",ct,"of",#lisubs,"of order",
                subrec`order;
      end if;
      lisubse :=
        ExtendOneSubgroupH(G,L[i],L[i+1],subrec,maxind,calcpres:Print:=Print);
      for subrece in lisubse do
        cps := subrece;
        //cps`subgroup := sub< G | subrece`subgroup, L[i+1] >; 
        //if Index(G, cps`subgroup) gt maxind div 2 then
        if cps`index gt maxind div 2 then
          cps`subgroup := sub< G | subrece`subgroup, L[i+1] >; 
          //cps`order := #(cps`subgroup);
          cps`order := #G div cps`index;
          Append(~lisubsfinal, cps);
        else
          Append(~lisubsl, subrece);
        end if;
      end for;
/*
      lisubsl cat:=
         ExtendOneSubgroupH(G,L[i],L[i+1],subrec,maxind,calcpres:Print:=Print);
      if Print gt 2 then print "   ",#lisubsl; end if;
*/
    end for;
    lisubs := lisubsl;
  end for;

  lisubs := lisubsfinal cat lisubsl;
  Sort(~lisubs,func<x,y|y`order-x`order>);
  for i in [1..#lisubs] do
    AssertAttribute(lisubs[i]`subgroup,"Order",lisubs[i]`order);
  end for;

  return lisubs;
end function;

CheckInt := function(G,K,stabpt,g)
/* K should be a subgroup of Stabiliser(G,stabpt), where G is a transitive
 * permutation group, and g an element of G. We want to test whether
 * K = Stabiliser(<K,g>,stabpt). This function does some quick tests and
 * returns false if it finds that this is not the case, and true
 * if it fails to come to a conclusion.
 */
 local L, O, o, gg;
 o := Order(g);
 if o eq 1 then
   return true;
 end if;
 L := sub<G|K,g>;
 O := Orbit(L,stabpt);
 for i in [1..10] do
   gg := g^Random([1..o-1])*Random(K)*g^Random([1..o-1])*Random(K);
         //g^Random([1..o-1])*Random(K);
   _,h := IsConjugate(L,stabpt,stabpt^gg);
   if not gg*h^-1 in K then
      return false;
   end if;
 end for;
 return true;
end function;

ExtendLowIndexSubgroupsH := function(G,stabpt,maxind:
           Print:=0, Presentation:=false, ss:=[], all:=false, attempt:=false)
/* For the case when G is too large to apply MaximalSubgroups directly,
 * but H = Stabiliser(G,stabpt) for some stabpt is not too large, then apply
 * LowIndexSubgroupsH first to H, and then see which subgroups * extend to G.
 * If the sequence ss is nonempty then it is used as the list of subgroups of
 * the stabiliser
 */
  local H, lisubsH, lisubs, orbreps, g, h, new, L, LS, ind, ca, ci, ext, ct, N,
        RF, subct, lisubsHnew, NG, F, phi, gct;
  H := Stabiliser(G,stabpt);
  if #ss ne 0 then
    lisubsH := ss;
  else
    lisubsH := LowIndexSubgroupsH(H,maxind:Print:=Print);
    if attempt then
       // Can we exclude subgroups conjugate under the action of G?
      lisubsHnew:=[];
      for s in lisubsH do
        new:=true;
        for t in lisubsHnew do
          if IsConjugate(G,s`subgroup,t`subgroup) then
            new:=false;
            break;
          end if;
        end for;
        if new then Append(~lisubsHnew,s); end if;
      end for;
      lisubsH := lisubsHnew;
    end if;
  end if;
  RF := Format(lisubsH[1]);
  lisubs := [rec<RF|subgroup:=G, order:=#G,length:=1>];

  if Print gt 1 then
    print #lisubsH, "SUBGROUPS TO EXTEND!";
  end if;

  subct:=0;
  for LF in lisubsH do
    L := LF`subgroup;
    subct+:=1;
    if Print gt 1 then
      print subct, "Stab sub index", #H div LF`order;
    end if;
    ca, ci := CosetAction(H,L);
    ind := Index(G,L);
    if all or ind le maxind then
        new :=true;
        for K in lisubs do
          if IsConjugate(G,L,K`subgroup) then
            new:=false;
            break;
          end if;
        end for;
        if new then
           Append(~lisubs,
             rec<RF|subgroup:=L, order:=#L,length:=Index(G,Normaliser(G,L))> );
           if Print gt 1 then
             print "  New: index",ind;
           end if;
           if ind lt Index(G,lisubs[#lisubs-1]`subgroup) then
              Sort(~lisubs,func<x,y|y`order-x`order>);
           end if;
        end if;
    end if;
    ext := [];
      // this will be the list of proper subgroups of G with stabiliser L
/* Theory of algorithm below:
    We are looking at subgroups <L,g> for g in G - we try to
    restrict the g that we need to consider.
    (i) Since <L,g> = <L,gl> for any l in L, choose one g for each
        orbit of L, such that 1^g = x for x orbit rep. of L.
        For each such g in (i), we need to consider hg for h in H, and
        we try to restrict the h that we need consider.
    (ii) Clearly we can restrict to right coset reps. of L in H.
   (iii) Better - for any l in (H meet gLg^-1), <L, hlg> = <L, hg>,
         so we can use orbit reps. of (H meet gLg^-1) in the right
         coset action of H on the cosets of L.
*/
    orbreps := [Representative(o) : o in Orbits(L) | o ne {stabpt} ];
    // restrict to stabpt^G
    orbreps := [o: o in orbreps | o in Orbit(G,stabpt)];
    gct := 0;
    for o in orbreps do
      // restriction (i) above
      _, g := IsConjugate(G,stabpt,o);
      LS := Stabiliser(L,o); // = L meet g^-1 H g
      Lorbreps := [Representative(lo) : lo in Orbits(ca(LS^(g^-1))) ];
      // LS^(g^-1)) = (H meet gLg^-1) - see (iii) above.

      for lo in Lorbreps do
        _, h := IsConjugate(ci,1,lo);
        h := h@@ca;
        N := sub< G | L, h*g >;
        gct +:= 1;
        if CheckInt(G,L,stabpt,h*g) and Stabiliser(N,stabpt) eq L then
          ind := Index(G,N);
          if Print gt 1 then
            print "  Extends to index in G:",ind;
          end if;
          if not N in ext then
            Append(~ext,N);
          end if;
        end if;
      end for;
    end for;
    if Print gt 1 then
      print "Tried",gct,"elements to adjoin to K";
    end if;

    if Print gt 1 then
      print #ext,"extensions for this subgroup of the stabiliser";
    end if;
    /* now see if any pair of groups in ext generates a larger group */
    ct := 2;
    while ct le #ext do
      for ct2 in [1..ct-1] do
        N := sub< G | ext[ct], ext[ct2] >;
        if CheckInt(G,ext[ct],stabpt,Random(ext[ct2])) and
                                               Stabiliser(N,stabpt) eq L then
          ind := Index(G,N);
          if Print gt 1 then
            print " Larger extension of index in G:",ind;
          end if;
          if not N in ext then
            Append(~ext,N);
          end if;
        end if;
      end for;
      ct := ct+1;
    end while;

    /* Now see which extensions are new subgroups of G */
    for N in ext do
      ind := Index(G,N);
      if all or ind le maxind then
        new :=true;
        for K in lisubs do
          if IsConjugate(G,N,K`subgroup) then
            new:=false;
            break;
          end if;
        end for;
        if new then
          Append(~lisubs,
             rec<RF|subgroup:=N, order:=#N,length:=Index(G,Normaliser(G,N))> );
          if Print gt 1 then
            print "  New: index",ind;
          end if;
          if ind lt Index(G,lisubs[#lisubs-1]`subgroup) then
            Sort(~lisubs,func<x,y|y`order-x`order>);
          end if;
        end if;
      end if;
    end for;

  end for; //loop over subgroups of H

  if Presentation then
    //simply have to compute them!
    if Print gt 1 then
      print "Calculating presentations";
    end if;
    for i in [1..#lisubs] do
      N := lisubs[i]`subgroup;
      if #N le 5000 then
        F,phi := FPGroup(N);
      else
        F,phi := FPGroupStrong(N);
        lisubs[i]`subgroup := sub<G | [phi(F.i) : i in [1..Ngens(F)] ]>;
      end if;
      lisubs[i]`presentation := F;
    end for;
  end if; 

  return lisubs;
end function;
