/******************************************************************************
 *
 *    solve.m   Finding a non-trivial stabiliser element in Sz(q)
 *
 *    File      : $HeadURL:: https://subversion.sfac.auckland.ac.nz/svn/prj_#$:
 *    Author    : Henrik Brnhielm
 *    Dev start : 2004-10-21
 *
 *    Version   : $Revision:: 2226                                           $:
 *    Date      : $Date:: 2011-03-19 10:29:05 +1300 (Sat, 19 Mar 2011)       $:
 *    Last edit : $Author:: jbaa004                                          $:
 *
 *    $Id:: solve.m 2226 2011-03-18 21:29:05Z jbaa004                      $:
 *
 *****************************************************************************/

freeze;
 
/*****************************************************************************/
/* DECLARATIONS                                                              */
/*****************************************************************************/

declare verbose SuzukiTrick, 10;

// Return codes
STATUS_SUCCESS      := 0;
STATUS_ZERO_ENTRIES := 1;
STATUS_NO_ROOTS     := 2;
STATUS_NO_SOLUTION  := 3;
STATUS_NO_BETA      := 4;
STATUS_ZERO_DENOM   := 5;

forward solveMagicPolynomial, findCyclicGroup, findMappingElements;

import "../../../util/basics.m" : MatSLP;

/*****************************************************************************/
/* TEST FUNCTIONS                                                            */
/*****************************************************************************/

/*****************************************************************************/
/* AUXILIARY FUNCTIONS                                                       */
/*****************************************************************************/

function solveMagicPolynomial(h, q, r, m)
    local b, s, polys, poly, roots, sols, eqns, t, c, d, badRoots,
	denominators, beta, alphat, F;

    // Crucial exponents
    t := 2^(m + 1);
    s := 2^m;
    F := GF(2^(2 * m + 1));
    
    // Matrix elements as a list, for easier indexing
    b := ElementToSequence(h);

    // All elements of our point must be non-zero
    if exists{r[i] : i in [1 .. 4] | IsZero(r[i])} then
	return {}, STATUS_ZERO_ENTRIES;
    end if;

    // We want to solve a univariate polynomial in alpha
    polys<alpha> := PolynomialAlgebra(F);

    // Coefficients in our 4 equations of alpha^(it) for i = 1..4 
    c := [0, r[1]*r[3]^s*alpha^4*q[1]^s*q[1]*b[2]^s*b[2]+r[1]*r[3]^s*alpha^3*q[1]*b[2]*q[2]^s*b[6]^s-r[2]^s*r[2]*alpha^4*q[1]^s*q[1]*b[3]^s*b[1]-r[2]^s*r[2]*alpha^3*q[2]^s*b[7]^s*q[1]*b[1],
	r[1]*r[3]^s*alpha^3*q[1]*b[2]*q[3]^s*b[10]^s+r[1]*r[3]^s*alpha^2*q[1]*b[2]*q[4]^s*b[14]^s+r[1]*r[3]^s*alpha^4*q[2]*b[6]*q[1]^s*b[2]^s+r[1]*r[3]^s*alpha^3*q[2]^s*q[2]*b[6]^s*b[6]+r[1]*r[3]^s*alpha^2*q[3]*b[10]*q[1]^s*b[2]^s+r[1]*r[3]^s*alpha*q[3]*b[10]*q[2]^s*b[6]^s-r[2]^s*r[2]*alpha^4*q[1]^s*b[3]^s*q[2]*b[5]-r[2]^s*r[2]*alpha^2*q[1]^s*b[3]^s*q[3]*b[9]-r[2]^s*r[2]*alpha^3*q[2]^s*q[2]*b[7]^s*b[5]-r[2]^s*r[2]*alpha*q[2]^s*b[7]^s*q[3]*b[9]-r[2]^s*r[2]*alpha^3*q[3]^s*b[11]^s*q[1]*b[1]-r[2]^s*r[2]*alpha^2*q[4]^s*b[15]^s*q[1]*b[1],
	r[1]*r[3]^s*alpha^3*q[2]*b[6]*q[3]^s*b[10]^s+r[1]*r[3]^s*alpha^2*q[2]*b[6]*q[4]^s*b[14]^s+r[1]*r[3]^s*alpha*q[3]^s*q[3]*b[10]^s*b[10]+r[1]*r[3]^s*q[3]*b[10]*q[4]^s*b[14]^s+r[1]*r[3]^s*alpha^2*q[4]*b[14]*q[1]^s*b[2]^s+r[1]*r[3]^s*alpha*q[4]*b[14]*q[2]^s*b[6]^s-r[2]^s*r[2]*alpha^2*q[1]^s*b[3]^s*q[4]*b[13]-r[2]^s*r[2]*alpha*q[2]^s*b[7]^s*q[4]*b[13]-r[2]^s*r[2]*alpha^3*q[3]^s*b[11]^s*q[2]*b[5]-r[2]^s*r[2]*alpha*q[3]^s*q[3]*b[11]^s*b[9]-r[2]^s*r[2]*alpha^2*q[4]^s*b[15]^s*q[2]*b[5]-r[2]^s*r[2]*q[4]^s*b[15]^s*q[3]*b[9],
	r[2]*r[3]*q[1]^2*alpha^4*b[1]*b[4]-r[4]*r[1]*q[1]^2*alpha^4*b[2]*b[3],
	r[2]*r[3]*q[1]*alpha^4*b[1]*q[2]*b[8]+r[2]*r[3]*q[1]*alpha^2*b[1]*q[3]*b[12]+r[2]*r[3]*q[2]*alpha^4*b[5]*q[1]*b[4]+r[2]*r[3]*q[3]*b[9]*q[1]*alpha^2*b[4]-r[4]*r[1]*q[1]*alpha^4*b[2]*q[2]*b[7]-r[4]*r[1]*q[1]*alpha^2*b[2]*q[3]*b[11]-r[4]*r[1]*q[2]*alpha^4*b[6]*q[1]*b[3]-r[4]*r[1]*q[3]*b[10]*q[1]*alpha^2*b[3],
	r[2]*r[3]*q[1]*alpha^2*b[1]*q[4]*b[16]+r[2]*r[3]*q[2]^2*alpha^4*b[5]*b[8]+r[2]*r[3]*q[2]*alpha^2*b[5]*q[3]*b[12]+r[2]*r[3]*q[3]*b[9]*q[2]*alpha^2*b[8]+r[2]*r[3]*q[3]^2*b[9]*b[12]+r[2]*r[3]*q[4]*b[13]*q[1]*alpha^2*b[4]-r[4]*r[1]*q[1]*alpha^2*b[2]*q[4]*b[15]-r[4]*r[1]*q[2]^2*alpha^4*b[6]*b[7]-r[4]*r[1]*q[2]*alpha^2*b[6]*q[3]*b[11]-r[4]*r[1]*q[3]*b[10]*q[2]*alpha^2*b[7]-r[4]*r[1]*q[3]^2*b[10]*b[11]-r[4]*r[1]*q[4]*b[14]*q[1]*alpha^2*b[3],
	r[2]*r[3]*q[2]*alpha^2*b[5]*q[4]*b[16]+r[2]*r[3]*q[3]*b[9]*q[4]*b[16]+r[2]*r[3]*q[4]*b[13]*q[2]*alpha^2*b[8]+r[2]*r[3]*q[4]*b[13]*q[3]*b[12]-r[4]*r[1]*q[2]*alpha^2*b[6]*q[4]*b[15]-r[4]*r[1]*q[3]*b[10]*q[4]*b[15]-r[4]*r[1]*q[4]*b[14]*q[2]*alpha^2*b[7]-r[4]*r[1]*q[4]*b[14]*q[3]*b[11],
	r[3]^s*r[3]*alpha^4*q[1]^s*q[1]*b[4]*b[2]^s+r[3]^s*r[3]*alpha^3*q[1]*b[4]*q[2]^s*b[6]^s-r[4]*r[2]^s*alpha^4*q[1]^s*q[1]*b[3]^s*b[3]-r[4]*r[2]^s*alpha^3*q[2]^s*b[7]^s*q[1]*b[3],
	r[3]^s*r[3]*alpha^3*q[1]*b[4]*q[3]^s*b[10]^s+r[3]^s*r[3]*alpha^2*q[1]*b[4]*q[4]^s*b[14]^s+r[3]^s*r[3]*alpha^4*q[2]*b[8]*q[1]^s*b[2]^s+r[3]^s*r[3]*alpha^3*q[2]^s*q[2]*b[8]*b[6]^s+r[3]^s*r[3]*alpha^2*q[3]*b[12]*q[1]^s*b[2]^s+r[3]^s*r[3]*alpha*q[3]*b[12]*q[2]^s*b[6]^s-r[4]*r[2]^s*alpha^4*q[1]^s*b[3]^s*q[2]*b[7]-r[4]*r[2]^s*alpha^2*q[1]^s*b[3]^s*q[3]*b[11]-r[4]*r[2]^s*alpha^3*q[2]^s*q[2]*b[7]^s*b[7]-r[4]*r[2]^s*alpha*q[2]^s*b[7]^s*q[3]*b[11]-r[4]*r[2]^s*alpha^3*q[3]^s*b[11]^s*q[1]*b[3]-r[4]*r[2]^s*alpha^2*q[4]^s*b[15]^s*q[1]*b[3],
	r[3]^s*r[3]*alpha^3*q[2]*b[8]*q[3]^s*b[10]^s+r[3]^s*r[3]*alpha^2*q[2]*b[8]*q[4]^s*b[14]^s+r[3]^s*r[3]*alpha*q[3]^s*q[3]*b[12]*b[10]^s+r[3]^s*r[3]*q[3]*b[12]*q[4]^s*b[14]^s+r[3]^s*r[3]*alpha^2*q[4]*b[16]*q[1]^s*b[2]^s+r[3]^s*r[3]*alpha*q[4]*b[16]*q[2]^s*b[6]^s-r[4]*r[2]^s*alpha^2*q[1]^s*b[3]^s*q[4]*b[15]-r[4]*r[2]^s*alpha*q[2]^s*b[7]^s*q[4]*b[15]-r[4]*r[2]^s*alpha^3*q[3]^s*b[11]^s*q[2]*b[7]-r[4]*r[2]^s*alpha*q[3]^s*q[3]*b[11]^s*b[11]-r[4]*r[2]^s*alpha^2*q[4]^s*b[15]^s*q[2]*b[7]-r[4]*r[2]^s*q[4]^s*b[15]^s*q[3]*b[11],
	r[3]^s*r[3]*alpha*q[4]*b[16]*q[3]^s*b[10]^s+r[3]^s*r[3]*q[4]^s*q[4]*b[16]*b[14]^s-r[4]*r[2]^s*alpha*q[3]^s*b[11]^s*q[4]*b[15]-r[4]*r[2]^s*q[4]^s*q[4]*b[15]^s*b[15],
	alpha^6*r[3]^t*r[3]*q[1]^t*b[4]^t*q[1]*b[2]+alpha^4*r[3]^t*r[3]*q[2]^t*b[8]^t*q[1]*b[2]-alpha^6*r[2]*r[4]^t*q[1]^t*b[3]^t*q[1]*b[3]-alpha^4*r[2]*r[4]^t*q[2]^t*b[7]^t*q[1]*b[3],
	alpha^6*r[3]^t*r[3]*q[1]^t*b[4]^t*q[2]*b[6]+alpha^4*r[3]^t*r[3]*q[1]^t*b[4]^t*q[3]*b[10]+alpha^4*r[3]^t*r[3]*q[2]^t*b[8]^t*q[2]*b[6]+alpha^2*r[3]^t*r[3]*q[2]^t*b[8]^t*q[3]*b[10]-alpha^6*r[2]*r[4]^t*q[1]^t*b[3]^t*q[2]*b[7]-alpha^4*r[2]*r[4]^t*q[1]^t*b[3]^t*q[3]*b[11]-alpha^4*r[2]*r[4]^t*q[2]^t*b[7]^t*q[2]*b[7]-alpha^2*r[2]*r[4]^t*q[2]^t*b[7]^t*q[3]*b[11],
	alpha^4*r[3]^t*r[3]*q[1]^t*b[4]^t*q[4]*b[14]+alpha^2*r[3]^t*r[3]*q[2]^t*b[8]^t*q[4]*b[14]+alpha^4*r[3]^t*r[3]*q[3]^t*b[12]^t*q[1]*b[2]+alpha^2*r[3]^t*r[3]*q[4]^t*b[16]^t*q[1]*b[2]-alpha^4*r[2]*r[4]^t*q[1]^t*b[3]^t*q[4]*b[15]-alpha^2*r[2]*r[4]^t*q[2]^t*b[7]^t*q[4]*b[15]-alpha^4*r[2]*r[4]^t*q[3]^t*b[11]^t*q[1]*b[3]-alpha^2*r[2]*r[4]^t*q[4]^t*b[15]^t*q[1]*b[3],
	alpha^4*r[3]^t*r[3]*q[3]^t*b[12]^t*q[2]*b[6]+alpha^2*r[3]^t*r[3]*q[3]^t*b[12]^t*q[3]*b[10]+alpha^2*r[3]^t*r[3]*q[4]^t*b[16]^t*q[2]*b[6]+r[3]^t*r[3]*q[4]^t*b[16]^t*q[3]*b[10]-alpha^4*r[2]*r[4]^t*q[3]^t*b[11]^t*q[2]*b[7]-alpha^2*r[2]*r[4]^t*q[3]^t*b[11]^t*q[3]*b[11]-alpha^2*r[2]*r[4]^t*q[4]^t*b[15]^t*q[2]*b[7]-r[2]*r[4]^t*q[4]^t*b[15]^t*q[3]*b[11]];

    // right hand side of our equations, terms that are independent of alpha^t
    d := [r[2]^s*r[2]*alpha*q[3]^s*b[11]^s*q[4]*b[13]+r[2]^s*r[2]*q[4]^s*q[4]*b[15]^s*b[13]-r[1]*r[3]^s*alpha*q[4]*b[14]*q[3]^s*b[10]^s-r[1]*r[3]^s*q[4]^s*q[4]*b[14]^s*b[14],
	r[4]*r[1]*q[4]^2*b[14]*b[15]-r[2]*r[3]*q[4]^2*b[13]*b[16], 0,
	alpha^2*r[2]*r[4]^t*q[3]^t*b[11]^t*q[4]*b[15]+r[2]*r[4]^t*q[4]^t*b[15]^t*q[4]*b[15]-alpha^2*r[3]^t*r[3]*q[3]^t*b[12]^t*q[4]*b[14]-r[3]^t*r[3]*q[4]^t*b[16]^t*q[4]*b[14]];

    // Denominators occuring when solving for alpha^t
    denominators := {c[13], c[9]*c[14]-c[10]*c[13],
	c[1]*c[14]*c[11]-c[1]*c[15]*c[10]+c[2]*c[9]*c[15]-
	c[2]*c[11]*c[13]-c[3]*c[9]*c[14]+c[3]*c[10]*c[13],
	c[14]*c[9]*c[8]*c[3]-c[14]*c[9]*c[7]*c[4]-c[14]*c[5]*c[12]*c[3]+
	c[14]*c[5]*c[11]*c[4]-c[14]*c[8]*c[1]*c[11]+c[14]*c[7]*c[1]*c[12]+
	c[9]*c[6]*c[15]*c[4]-c[9]*c[6]*c[16]*c[3]-c[9]*c[8]*c[2]*c[15]+
	c[9]*c[7]*c[2]*c[16]+c[5]*c[16]*c[3]*c[10]-c[6]*c[11]*c[13]*c[4]+
	c[6]*c[12]*c[13]*c[3]+c[6]*c[16]*c[1]*c[11]-c[7]*c[1]*c[16]*c[10]-
	c[7]*c[2]*c[12]*c[13]+c[7]*c[4]*c[10]*c[13]+c[8]*c[1]*c[15]*c[10]+
	c[8]*c[2]*c[11]*c[13]-c[6]*c[15]*c[1]*c[12]-c[8]*c[3]*c[10]*c[13]+
	c[5]*c[15]*c[2]*c[12]-c[5]*c[15]*c[4]*c[10]-c[5]*c[16]*c[2]*c[11]}
	diff {polys!0};

    // All 4 denominators must be non-zero
    if #denominators lt 4 then
	return {}, STATUS_ZERO_DENOM;
    end if;

    // alpha^t expressed as a polynomial in alpha
    alphat := (-c[9]*c[7]*d[1]*c[14]+c[9]*c[14]*d[2]*c[3]+
	c[9]*c[6]*c[15]*d[1]+c[9]*c[7]*c[2]*d[4]-c[9]*c[6]*d[4]*c[3]-
	c[9]*d[2]*c[2]*c[15]+c[14]*c[5]*c[11]*d[1]-c[14]*d[2]*c[1]*c[11]-
	c[14]*c[5]*d[3]*c[3]+c[14]*c[7]*c[1]*d[3]-c[5]*c[11]*c[2]*d[4]+
	c[5]*d[3]*c[2]*c[15]+c[6]*d[4]*c[1]*c[11]+c[6]*c[13]*d[3]*c[3]+
	d[2]*c[2]*c[13]*c[11]+c[5]*c[10]*d[4]*c[3]-c[6]*c[15]*c[1]*d[3]+
	d[2]*c[1]*c[10]*c[15]+c[7]*d[1]*c[13]*c[10]-c[6]*c[13]*c[11]*d[1]-
	c[5]*c[10]*c[15]*d[1]-d[2]*c[3]*c[13]*c[10]-c[7]*c[1]*c[10]*d[4]-
	c[7]*c[2]*c[13]*d[3]) /
	(-c[9]*c[7]*c[4]*c[14]+c[9]*c[14]*c[8]*c[3]-c[9]*c[6]*c[16]*c[3]+
	c[9]*c[6]*c[15]*c[4]+c[9]*c[7]*c[2]*c[16]-c[9]*c[8]*c[2]*c[15]-
	c[14]*c[5]*c[12]*c[3]+c[14]*c[5]*c[11]*c[4]-c[14]*c[8]*c[1]*c[11]+
	c[14]*c[7]*c[1]*c[12]-c[6]*c[13]*c[11]*c[4]+c[5]*c[10]*c[16]*c[3]+
	c[7]*c[4]*c[13]*c[10]-c[5]*c[10]*c[15]*c[4]+c[6]*c[16]*c[1]*c[11]+
	c[6]*c[13]*c[12]*c[3]-c[7]*c[2]*c[13]*c[12]-c[6]*c[15]*c[1]*c[12]-
	c[8]*c[3]*c[13]*c[10]-c[7]*c[1]*c[10]*c[16]+c[8]*c[1]*c[10]*c[15]+
	c[8]*c[2]*c[13]*c[11]+c[5]*c[12]*c[2]*c[15]-c[5]*c[11]*c[2]*c[16]);

    // Our main polynomials
    poly := [alphat^4 * c[1] + alphat^3 * c[2] + alphat^2 * c[3] +
	alphat * c[4] - d[1],
	alphat^4 * c[5] + alphat^3 * c[6] + alphat^2 * c[7] +
	alphat * c[8] - d[2],
	alphat^4 * c[9] + alphat^3 * c[10] + alphat^2 * c[11] +
	alphat * c[12] - d[3],
	alphat^4 * c[13] + alphat^3 * c[14] + alphat^2 * c[15] +
	alphat * c[16] - d[4]];

    vprint SuzukiTrick, 4: poly[1];
    
    roots := {};
    for p in poly do
	if not p eq polys!0 then

	    vprint SuzukiTrick, 7: Category(p);

	    // Check if we have a polynomial or a rational function
	    if Category(p) eq RngUPol then
		pol := p;
	    else
		pol := Numerator(p);
	    end if;
	    
	    vprint SuzukiTrick, 5: Degree(pol), #Terms(pol), Valuation(pol);
	    rot := Roots(pol);
	    vprint SuzukiTrick, 5: #rot;
	    vprint SuzukiTrick, 5: Factorization(pol);
		
	    // Remove multiplicites 
	    roots join:= {x[1] : x in rot |
		x[1] in F and not IsZero(x[1])};
	end if;
    end for;
    if IsEmpty(roots) then
	return {}, STATUS_NO_ROOTS;
    end if;

    vprint SuzukiTrick, 5: #roots;
    
    sols := [];
    for alpha in roots do

	// Parts of original equations, for computation of beta
	eqns := [(q[1]*alpha^(t+1)*b[1]+q[2]*alpha*b[5]+
	    q[3]*b[9]*alpha^(-1)+q[4]*b[13]*alpha^(-t-1)),
	(q[1]*alpha^(t+1)*b[2]+q[2]*alpha*b[6]+
	    q[3]*b[10]*alpha^(-1)+q[4]*b[14]*alpha^(-t-1)),
	(q[1]*alpha^(t+1)*b[3]+q[2]*alpha*b[7]+
	    q[3]*b[11]*alpha^(-1)+q[4]*b[15]*alpha^(-t-1)),
	(q[1]*alpha^(t+1)*b[4]+q[2]*alpha*b[8]+
	    q[3]*b[12]*alpha^(-1)+q[4]*b[16]*alpha^(-t-1))];
	
	// Equations must be non-zero since q[1] .. q[4] are nonzero 
	if exists{eqn : eqn in eqns | IsZero(eqn)} then
	    continue;
	end if;
	
	// Compute beta from alpha
	beta := Sqrt(r[3]^(-1) * r[2] * L^(-1) * M) where L is eqns[2]
	    where M is eqns[3];
	if not IsZero(beta) then
	    Append(~sols, <alpha, beta, gamma> where gamma is
		L * beta * r[2]^(-1) where L is eqns[2]);
	end if;
    end for;
    if IsEmpty(sols) then
	return {}, STATUS_NO_BETA;
    end if;

    // Check that solutions satisfy our equations
    sols := {sol : sol in sols | K * beta^(t + 1) eq r[1] * C and
	L * beta eq r[2] * C and M * beta^(-1) eq C * r[3] and
	N * beta^(-t - 1) eq r[4] * C where C is sol[3] where
	beta is sol[2] where K is eqns[1] where L is eqns[2]
	where M is eqns[3] where N is eqns[4] where eqns is
	[(q[1]*alpha^(t+1)*b[1]+q[2]*alpha*b[5]+
	q[3]*b[9]*alpha^(-1)+q[4]*b[13]*alpha^(-t-1)),
	(q[1]*alpha^(t+1)*b[2]+q[2]*alpha*b[6]+
	q[3]*b[10]*alpha^(-1)+q[4]*b[14]*alpha^(-t-1)),
	(q[1]*alpha^(t+1)*b[3]+q[2]*alpha*b[7]+
	q[3]*b[11]*alpha^(-1)+q[4]*b[15]*alpha^(-t-1)),
	(q[1]*alpha^(t+1)*b[4]+q[2]*alpha*b[8]+
	q[3]*b[12]*alpha^(-1)+q[4]*b[16]*alpha^(-t-1))] where alpha is sol[1]};

    if IsEmpty(sols) then
	return {}, STATUS_NO_SOLUTION;
    end if;
    
    return sols, STATUS_SUCCESS;
end function;

// Diagonalise an element of order q-1 in standard Suzuki group
function diagonaliseSuzukiCyclicGroup(g, m)
    local G, eigenvalues, t, q, lambda;

    F := GF(2, 2 * m + 1);
    q := #F;
    t := 2^(m + 1);
    G := GL(4, F);

    eigenvalues, eigenvectors := SuzukiCyclicEigenvalues(g);
    lambda := eigenvalues[2];
    
    // 4 x 4 conjugating matrix with eigenvectors
    x := G ! &cat[ElementToSequence(f) : f in eigenvectors];

    // 4 x 4 diagonal matrix with eigenvalues
    d := G ! DiagonalMatrix(F, 4, eigenvalues);
    assert d^x eq g;

    return d, x, lambda;
end function;

// Find element of order q-1 and diagonalise it
function findCyclicGroup(R, m)
    local a, lambda, eigenvalues, x, d, t, F, word,  G, q;

    t := 2^(m + 1);
    q := 2^(2 * m + 1);
    
    repeat
	a, word := Random(R);
    until Order(a : Proof := false) eq q - 1;
    
    d, x, lambda := diagonaliseSuzukiCyclicGroup(a, m);
    return a, word, d, x, lambda;
end function;

// Find element in Sz(q) that maps P to Q, where P and Q are points in the
// standard ovoid.
function findMappingElements(R, P, Q, m : Benchmark := false,
    UseDiscreteLog := true)
	local words, t, h, a, d, eigenvectors, lambda, points, mappings, coset,
	q, sols, status, logTime, exponent, mappers;
    
    assert P ne Q;
    words := [];
    t := 2^(m + 1);

    // Get random matrix
    h, words[1] := Random(R);
    
    // Find random element of order q - 1 and diagonalise it
    // so move to another coset
    a, words[2], d, eigenvectors, lambda := findCyclicGroup(R, m);

    // Perform some benchmark if necessary
    points := {};
    mappings := {};
    if Benchmark then
	coset := {};
	q := 2^(2 * m + 1);
	for i in [1 .. q - 1] do
	    for j in [1 .. q - 1] do
		Include(~coset, a^i * h * a^j);
		Include(~points, P * a^i * h * a^j);
		if Normalise(P * a^i * h * a^j) eq Normalise(Q) then
		    Include(~mappings, <i, j>);
		end if;
	    end for;
	end for;
	if #coset eq q - 1 then
	    vprint SuzukiTrick, 1: "h normalises <a>";
	end if;
    end if;
    
    // Find element in standard coset that maps one point to another
    sols, status := solveMagicPolynomial(h^(eigenvectors^(-1)),
	P * eigenvectors^(-1), Q * eigenvectors^(-1), m);
    
    if status eq STATUS_SUCCESS then
	if UseDiscreteLog then
	    // Express solutions in terms of the generator of the cyclic groups
	    // by computing discrete logarithms
	    logTime := Cputime();
	    exponent := {<Log(lambda, s[1]), -Log(lambda, s[2])> : s in sols};
	    logTime := Cputime(logTime);
	    
	    // Compute stabiliser elements
	    mappers := {<word, p> : exp in exponent |
		Normalise(P * p) eq Normalise(Q) where p is a^j * h * a^(-i)
		where word is words[2]^j * words[1] * words[2]^(-i)
		where j is exp[1] where i is exp[2]};
	else
	    // Compute stabiliser elements
	    mappers := {<Identity(Parent(words[1])), p> : sol in sols |
		Normalise(P * p) eq Normalise(Q)
		where p is x^eigenvectors * h * y^eigenvectors
		where x is Generic(Parent(a)) !
		DiagonalMatrix([sol[1]^(t + 1), sol[1],
		sol[1]^(-1), sol[1]^(-t - 1)])
		where y is Generic(Parent(a)) !
		DiagonalMatrix([sol[2]^(t + 1), sol[2],
		sol[2]^(-1), sol[2]^(-t - 1)])};
	    assert not IsEmpty(mappers);
	    logTime := 0;
	end if;
    else
	mappers := {};
	logTime := 0;
    end if;

    return mappers, status, logTime, #points, #mappings;
end function;

