ClassicalPermutationGroup := function( type, dim, q )
    
    gens := ClassicalStandardGenerators( type, dim, q );
    q2 := case< type | "SU": q^2, default: q >;
    V := VectorSpace( GF( q2 ), dim );
    S := {@ sub< V | v > : v in V @};
    deg := #S;
    
    pgens := [];
    
    for gen in gens do
        Append( ~pgens, Sym( deg )!{@ Position( S, s^gen ) : s in S @});
    end for;
    
    return sub< Sym( deg ) | pgens >, pgens;
end function;
    
    
TestClassicalRewrite := function( type, dim, q : nr := 100 )
    print "# Testing ", type, "(", dim, ",", q, ").Z( GL( ", dim, ",", 
      q, ") in natural";
    
    q2 := case< type | "SU": q^2, default: q >;
    gl := GL( dim, q2 );
    
    
    print "# first testing  natural rep by Elliot";
    gens := ClassicalStandardGenerators( type, dim, q );
    g := sub< gl | gens >;
    
    for i in [1..nr] do
        x := Random( g );
        v, p := ClassicalRewrite( g, gens, type, dim, q, x );
        if not v or Evaluate( p, gens ) ne x then
            return false;
        end if;
    end for;    
    
    tw := One( GL( dim, q2 )); //Random( GL( dim, q2 ));
    gens := [ x^tw : x in gens ];
    
    if type eq "SL" then
        form := Zero( MatrixAlgebra( GF( q2 ), dim ));
    elif type eq "Sp" then
        _, form := SymplecticForm( sub< gl | gens > );
    elif type eq "SU" then
        _, form := UnitaryForm( sub< gl | gens > );
    elif type in { "Omega+", "Omega-", "Omega" } then
        _, _, form := OrthogonalForm( sub< gl | gens > );
    end if;    
    
    gens1 := gens;
    Append( ~gens1, ScalarMatrix( dim, PrimitiveElement( GF( q2 ))));
    g := sub< gl | gens >;
    g1 := sub< gl | gens1 >;
    
//    return g, g1, gens, gens1;
    
    print "#...first using unipotent stabilizers";
    
    // first we test SL(n,q)Z(GL(n,q))
      
    ct := 0; cf := 0;
    for i in [1..nr] do
        x := Random( g1 );
        xq := gl![ z^q : z in Eltseq( x )];
        ising := Determinant( x ) eq 1 and x*form*Transpose( xq ) eq form;
        if type in { "Omega+", "Omega-" } then
            ising := ising and SpinorNorm( x, form ) eq 0;
        end if;
        
        //if not ising then return g, gens, x; end if;
        v, p := ClassicalRewrite( g, gens, type, dim, q, x : 
                        Method := "CharP" ); 
        if v then
            ct := ct+1;
        else
            cf := cf+1;
        end if;
        
        x1 := Evaluate( p, gens );
        if not v eq ising then
            print "v", v, "ising", ising;
            print 1;
            return false, g, gens, x;
            print 2;
        elif v and x1 ne x then
            print 3;
            return false, g, gens, x;
        elif not v and { (x1*x^-1,z) : z in Generators( g )} ne {One(g)} then
            print 4;
            return false, g, g1, gens, gens1, x;
        end if;
    end for;
    
    print "# ( true, false ) = ( ", ct, ", ", cf, " )";

    
    print "#...then black-box";
    
    if not type in { "Omega", "Omega+", "Omega-" } then
    
    ct := 0; cf := 0;
    for i in [1..nr] do
        x := Random( g1 );
        xq := gl![ z^q : z in Eltseq( x )];
        ising := Determinant( x ) eq 1 and x*form*Transpose( xq ) eq form;
        v, p := ClassicalRewrite( g, gens, type, dim, q, x : 
                        Method := "BB" ); 
        if v then
            ct := ct+1;
        else
            cf := cf+1;
        end if;
        
        x1 := Evaluate( p, gens );
        
        if not v eq ising then
            print ising;
            print "fail";
            return false, g, g1, gens, gens1, x;
        elif v and x1 ne x then
            print 1;
            return false, g, g1, gens, gens1, x;
        elif not v and { (x1*x^-1,z) : z in Generators( g )} ne {One(g)} then
            return false, g, g1, gens, gens1, x;
        end if;

    end for;
    
    print "# ( true, false ) = ( ", ct, ", ", cf, " )";
    
    end if;
        
    //then we test GX( n, q )
    
      print "# Testing GX(", dim, ",", q, ") in natural";
    
        
    d := One( MatrixAlgebra( GF( q2 ), dim ));
    
    if type eq "SL" then
        
        d[1,1] := PrimitiveElement( GF( q ));
        
    elif type eq "Sp" then
        
        for i in [1..dim-1 by 2] do
            d[i,i] := PrimitiveElement( GF( q2 ));
        end for;
        
    elif type eq "SU" then
        
        for i in [1..dim-1 by 2] do
            d[i,i] := PrimitiveElement( GF( q2 ))^(i*(q-1));
            d[i+1,i+1] := PrimitiveElement( GF( q2 ))^(i*(q-1));
        end for;
        if IsOdd( dim ) then
            d[dim,dim] := PrimitiveElement( GF( q2 ))^(q-1);
        end if;
        
    end if;

    
    gens1 := gens;
    Append( ~gens1, (gl!d)^tw );
    
    g := sub< gl | gens >;
    g1 := sub< gl | gens1 >;
    
    print "#...first using unipotent stabilizers";
    
    ct := 0; cf := 0;
    for i in [1..nr] do
        x := Random( g1 );
        xq := gl![ z^q : z in Eltseq( x )];
        ising := Determinant( x ) eq 1 and x*form*Transpose( xq ) eq form;
        v, p := ClassicalRewrite( g, gens, type, dim, q, x : 
                        Method := "CharP" ); 
        if v then
            ct := ct+1;
        else
            cf := cf+1;
        end if;
        
        if v ne ising then 
            return false;
        end if;
        
        if v and Evaluate( p, gens ) ne x then
            return false;
        end if;
        
    end for;
    
    print "# ( true, false ) = ( ", ct, ", ", cf, " )";
    print "#...then black-box";
    
    if not type in { "Omega", "Omega+", "Omega-" } then
    
    ct := 0; cf := 0;
    for i in [1..nr] do
        x := Random( g1 );
        xq := gl![ z^q : z in Eltseq( x )];
        ising := Determinant( x ) eq 1 and x*form*Transpose( xq ) eq form;
        v, p := ClassicalRewrite( g, gens, type, dim, q, x : 
                        Method := "BB" ); 
        if v then
            ct := ct+1;
        else
            cf := cf+1;
        end if;
        
        if v ne ising then 
            return false;
        end if;
        
        if v and Evaluate( p, gens ) ne x then
            return false;
        end if;
        
    end for;
    
    print "# ( true, false ) = ( ", ct, ", ", cf, " )";
    
    end if;
    
    if not type in { "Omega", "Omega+", "Omega-" } then
            
    print "# Testing ", type, "(", dim, ",", q, ") x P";
     
    P := sub< GL( 2, q2 ) | [1,1,0,1] >;
    
    h2 := DirectProduct( sub< g | One( g ) >, P);
    h1 := DirectProduct( g, sub< GL( 2, q2 )| One( GL( 2, q2 ))>);
    
    h := sub< GL( dim+2, q2 ) | 
              Generators( h1 ) join Generators( h2 )>;
    
    gensh1 := [ GL( dim+2,q2)!DiagonalJoin( x, 
                      One( GL( 2, q2 ))) : x in gens ];
    
    ct := 0; cf := 0; 
    
    for i in [1..nr] do
        x := Random( h );
        
        d := dim;
        
        isinh1 := (x[d+1,d+1] eq 1 and x[d+1,d+2] eq 0 and 
                   x[d+2,d+1] eq 0 and x[d+2,d+2] eq 1);
        v, prog := ClassicalRewrite( h1, gensh1, type, dim, q, x ); 
        
        if v then
            ct := ct+1;
        else
            cf := cf+1;
        end if;
        
        if v then x1 := Evaluate( prog, gensh1 ); end if;
        if not v eq isinh1 then return false; end if;
        if v and x1 ne x then 
            return false;
        end if;
    end for;
    
    print "# ( true, false ) = ( ", ct, ", ", cf, " )";
    
    if (q2^dim-1)/(q2-1) le 5000 then
        print "# testing perm rep.";
        g, gens := ClassicalPermutationGroup( type, dim, q );
        for i in [1..nr] do
            x := Random( g );
            v, p := ClassicalRewrite( g, gens, type, dim, q, x );
            if not v or Evaluate( p, gens ) ne x then
                return false;
            end if;
        end for;
    end if;
    
    end if;
    
    return true;
end function;


ExteriorSquareGroup := function( type, dim, q : Twist := false )
    
    gens := [ ExteriorSquare( x ) : 
              x in ClassicalStandardGenerators( type, dim, q )];
    q2 := case< type | "SU": q^2, default: q >;
    g := sub< GL( Binomial( dim, 2 ), q2 ) | gens >;
    
    M := GModule( g );
    mods := Sort( MaximalSubmodules( M ), 
                  func< u, v | Dimension ( u ) - Dimension( v )>);
    
    if #mods eq 1 then
      
      if Twist then 
        tw := Random( g );
        gens := [ x^tw : x in gens ];    
        g := sub< GL( Binomial( dim, 2 ), q2 ) | gens >;
      end if;
      
      return g, gens;
      
  end if;
    
    W, f := M/mods[1];
    
    
    
    newgens := [];
    for gen in gens do
        newgen := [];
        for b in Basis( W ) do
            Append( ~newgen, Eltseq( ((b@@f)^gen)@f));
        end for;
        
        Append( ~newgens, GL( Dimension( W ), q2 )!newgen );
    end for;
    
    if Twist then 
        tw := Random( GL( Dimension( W ), q2 ));
        newgens := [ x^tw : x in newgens ];    
    end if;
    
    return sub< GL( Dimension( W ), q2 ) | newgens >, newgens;
    
end function;
    

ConjugateClassicalGroup := function( type, d, q )
    
    gens := ClassicalStandardGenerators( type, d, q );
    
    if type in { "SU", "Omega" } and IsOdd( d ) then
        list := [d];
        last := d-1;
    else
        list := [];
        last := d;
    end if;
    
    q2 := case< type | "SU": q^2, default: q >;
    
    for i in [1..d-1 by 2] do
        Append( ~list, i );
    end for;
    
    for i in [last..2 by -2] do
        Append( ~list, i );
    end for;
    
    p := GL( d,q2 )!PermutationMatrix( GF( q2 ), Sym( d )!list );
    
    gens := [ x^(p^-1) : x in gens ];
    g := sub< GL( d, q2) | gens >;

return g, gens;
end function;
    
    
TestAgainstElliot := function( type, dim, q : count := 10 )
        
    
    g, gens := ExteriorSquareGroup( type, dim, q );
    
    x := Random( g );
    
    t := Cputime();
    v, p := ClassicalRewrite( g, gens, type, dim, q, x );

    print "######## Testing ", type, "( ", dim, ", ", q, ") ############";
    print "Time of first ClassicalRewrite: ", Cputime( t );
    
    t := Cputime();
    for i in [1..count] do
        x := Random( g ); 
        v, p := ClassicalRewrite( g, gens, type, dim, q, x : Method := 
                        "CharP" );
        if not v or Evaluate( p, gens ) ne x then
            return false;
        end if;
    end for;
    
    tcs := Cputime(t)/count;
    print "Average CPU time of ClassicalRewrite: ", tcs;
    
    ElliotProgram := case< type | "SL": int_SLAltRepWordInGen, 
                     "Sp": int_SpAltRepWordInGen, 
                     "SU": int_SUAltRepWordInGen, 
                     "Omega+": int_OmegaPlusAltRepWordInGen,
                     "Omega-": int_OmegaMinusAltRepWordInGen,
                     "Omega": int_OmegaCircleAltRepWordInGen,

                     default: false >;
    
    q2 := case< type| "SU": q^2, default: q >;
    t := Cputime();
    v, p := ElliotProgram( g, gens, x, dim, q2 );
    print "Time of first ElliotProgram: ", Cputime( t );
    
    t := Cputime();
    for i in [1..count] do
        x := Random( g ); 
        v, p := ElliotProgram( g, gens, x, dim, q2 );
        if not v or Evaluate( p, gens ) ne x then
            return false;
        end if;
    end for;
    te := Cputime(t)/count;
    print "Average CPU time of ElliotProgram: ", te;
    
    return te-tcs;
    
end function;
    
    
        
            
           
        
    
    
    
