freeze;

PCOSG := function(G, f, pt)
/* Polycyclic orbit-stabiliser calcualtion.
 * f is a group-action function S x G -> S, with G of type GrpPC, and pt in S.
 * Return orbit of pt, stabiliser of pt as subgroup of G, and
 * generator numbers in G that were used to build up orbit.
 */
  local ng,ords,orb,stabgens,orbnos,im,posmg,lo,pow,g,h;
  ng := NPCgens(G);
  ords := PCPrimes(G);
  orb := {@ pt @};
  stabgens := [G|];
  orbnos := [];
  for i in [ng .. 1 by -1] do
    im := f(<pt , G.i>);
    pos := Position(orb,im);
    if pos eq 0 then
      orb join:= {@ f(<o, G.i^k>) : o in orb, k in [1..ords[i]-1] @};
      Append(~orbnos,i);
    else
      g := G.i;
      lo := #orb;
      for t in Reverse(orbnos) do
        if pos eq 1 then break; end if;
        lo := lo div ords[t];
        pow := (pos-1) div lo;
        h := G.t^-pow;
        g *:= h;
        im := f(< im, h >);
        pos := Position(orb,im);
      end for;
      assert f(<pt,g>) eq pt;
      Append(~stabgens, g);
    end if;
  end for;
  return orb, sub< G | stabgens>, orbnos;
end function;

PCActGIsCon := function(G,f,orb,orbnos,impt) 
//orb, orbnos should be as returned by PCOSG on (G,f,pt)
//Determine if impt is in orb, and if so return conjugating element in G
local ng, g, lo, pos, pow, h;
      pos := Position(orb,impt);
      if pos eq 0 then return false,_; end if;
      ng := NPCgens(G);
      g := Id(G);
      lo := #orb;
      ords := PCPrimes(G);
      for t in Reverse(orbnos) do
        if pos eq 1 then break; end if;
        lo := lo div ords[t];
        pow := (pos-1) div lo;
        h := G.t^-pow;
        g *:= h;
        impt := f(<impt, h>);
        pos := Position(orb,impt);
      end for;
      assert pos eq 1;
      return true, g^-1;
end function;

EltCent := function(G,g:so:=256)
  local r,rpc,rad,rtopc,rq,Gtorq,rm,gmacts,img,imc,cgens,x,cpc,icgens,cs,m,ptom,
        rep,p,d,qpc,pctoqpc,cqpc,cmats,col,pg,ig,mat,V,fact,rado,radst,radonos,
        igenmats,lo,fullgp,subgp,oa,oi,suborbs,ba,bi,rqact,rqst,fullo,deg,pt,
        perms,vec,imvec,pos,gorbim,stg,istg,ords,iwmimc,slp,smat,pow,y;
  r := G`LMGNode;
  rpc := r`radPC;
  rad := r`rad;
  rtopc := r`radtoPC;
  rq := r`radquot;
  Gtorq := r`Gtoradquot;
  rm := r`radmods;
  gmacts := r`rqradmodactions;

  img := Gtorq(g);
  imc := Centraliser( rq, img );
  //try to minimise generators
  cgens := [img];
  while sub< imc | cgens > ne imc do
    repeat x:=Random(imc); until not x in sub< imc | cgens >;
    Append(~cgens,x);
  end while;
  imc := sub< imc | cgens >;
  //centraliser will be stored as intersection of centraliser in rpc
  //with extra gens outside of radquot
  cpc := sub< rpc | >;
  icgens := [g] cat [cgens[i] @@ Gtorq : i in [2..#cgens]];
  cs := r`cseriesrad;

  for k in [#cs-1..1 by -1] do
    m := rm[k][1];
    ptom := rm[k][2];
    rep := rm[k][3];
    p := #BaseRing(m); d := Dimension(m);
    vprint LMG,1: "Lifting through layer",k,"of size",p,"^",d;
    //get action of generators of imc on layer
    //start with action of rpc
    qpc, pctoqpc := quo< rpc | cs[k] >;
    cqpc := sub< qpc | pctoqpc(cs[k+1]), pctoqpc(cpc) >;
    cmats := [];
    col := Matrix(GF(p),d+1,1,[0: i in [1..d]] cat [1]);
    for i in [1..NPCgens(cqpc)] do
      pg := cqpc.i @@ pctoqpc; ig := pg @@ rtopc;
      mat := HorizontalJoin(
         VerticalJoin( rep(pg), Matrix((g,ig) @ rtopc @ ptom)), col );
      Append(~cmats,  mat);
    end for;
    cqpcact := hom< cqpc->GL(d+1,p) | cmats >;
    V := VectorSpace(GF(p),d+1);
    fact := func<  t | t[1] * cqpcact(t[2]) >;
    rado,radst,radonos := PCOSG(cqpc,fact,V.(d+1));
    vprint LMG,2: "  #rado:",#rado,"#radst",#radst;

    //now action of imc on orbits
    igenmats := [ GL(d+1,p) | HorizontalJoin(
         VerticalJoin( gmacts[k](icgens[i]:im:=imc.i),
                         Matrix((g,icgens[i]) @ rtopc @ ptom)), col ):
                             i in [1..#icgens] ];
    //need two strategies for action of rq
    lo := #rado;
    if lo le so then
      vprint LMG,2: "  small orbit calculation";
      fullgp := sub< GL(d+1,p) | cmats cat igenmats >;
      subgp := sub< GL(d+1,p) | cmats >;
      oa, oi := OrbitAction(fullgp,V.(d+1));
      suborbs := Orbits(oa(subgp));
      ba, bi := BlocksAction(oi,suborbs[1]);
      rqact := hom< imc->bi | [igenmats[i] @ oa @ ba : i in [1..#icgens]] >;
      rqst := Stabiliser(bi,1) @@ rqact;
      vprint LMG,2: "  degree oi:",Degree(oi),"#rqst",#rqst;
    else
      vprint LMG,2: "  non-small orbit calculation";
      fullo := rado; deg := 1;
      perms := [ [] : i in [1..#igenmats] ];
      pt := 1;
      while pt le deg do
        vec := fullo[(pt-1)*lo+1];
        for i in [1..#igenmats] do
          imvec := vec*igenmats[i];
          pos := Position(fullo,imvec);
          if pos eq 0 then //new orbit
            fullo join:=  PCOSG(cqpc,fact,imvec);
            deg +:= 1;
            perms[i][pt] := deg;
          else perms[i][pt] := (pos-1) div lo + 1;
          end if;
        end for;
        pt +:= 1;
      end while;
      ChangeUniverse(~perms, Sym(deg));
      gorbim := sub< Sym(deg) | perms >;
      rqact := hom< imc->gorbim | perms >;
      rqst := Stabiliser(gorbim,1) @@ rqact;
      vprint LMG,2: "  #fullo:",#fullo,"#rqst",#rqst;
    end if;
    assert  imc.1 in rqst;
    //try to minimise generators
    stg := [imc.1];
    while sub< rqst | stg > ne rqst do
      repeat x:=Random(rqst); until not x in sub< rqst | stg >;
      Append(~stg,x);
    end while;
    rqst := sub< rqst | stg >;
    istg := [Generic(G)|g];
    //ords := [ #quo<cs[i]|cs[i+1]> where cs:=CompositionSeries(cqpc) :
    //              i in [1..NPCgens(cqpc)] ]; //should be a standard function!
    iwmimc := InverseWordMap(imc);
    for l in [2..#stg] do
      slp := iwmimc(stg[l]);
      smat := Evaluate(slp, igenmats);
      vec := V.(d+1) * smat;
      isc,x := PCActGIsCon(cqpc,fact,rado,radonos,vec); assert isc;
      istg[l] := Evaluate(slp,icgens) * ((x^-1) @@ pctoqpc @@ rtopc);
    end for;
    cpc := sub< rpc | [ radst.i @@ pctoqpc : i in [1..NPCgens(radst)] ] >;
    icgens := istg;
    imc := rqst;
  end for; //for k in [#cs-1..1 by -1] 
  return sub<Generic(G) | [cpc.i @@ rtopc : i in [1..Ngens(cpc)]] cat icgens >;
end function;

IsConj := function(G,g,h:so:=256)
  local r,rpc,rad,rtopc,rq,Gtorq,rm,gmacts,img,imc,cgens,x,cpc,icgens,cs,m,ptom,
        rep,p,d,qpc,pctoqpc,cqpc,cmats,col,pg,ig,mat,V,fact,rado,radst,radonos,
        igenmats,lo,fullgp,subgp,oa,oi,suborbs,ba,bi,rqact,rqst,fullo,deg,pt,
        perms,vec,imvec,pos,gorbim,stg,istg,ords,iwmimc,slp,smat,pow,y,ans,imh,
        orb,convec,convecpos,isc;
  r := G`LMGNode;
  rpc := r`radPC;
  rad := r`rad;
  rtopc := r`radtoPC;
  rq := r`radquot;
  Gtorq := r`Gtoradquot;
  img := Gtorq(g); imh := Gtorq(h);
  isc, x := IsConjugate( rq, img, imh);
  if not isc then return false, _; end if;
  vprint LMG,2: "  Initialising conjugating element";
  ans := x @@ Gtorq;  g := g^ans;

  rm := r`radmods;
  gmacts := r`rqradmodactions;

  imc := Centraliser( rq, imh );
  //try to minimise generators
  cgens := [imh];
  while sub< imc | cgens > ne imc do
    repeat x:=Random(imc); until not x in sub< imc | cgens >;
    Append(~cgens,x);
  end while;
  imc := sub< imc | cgens >;
  //centraliser will be stored as intersection of centraliser in rpc
  //with extra gens outside of radquot
  cpc := sub< rpc | >;
  icgens := [g] cat [cgens[i] @@ Gtorq : i in [2..#cgens]];
  cs := r`cseriesrad;

  for k in [#cs-1..1 by -1] do
    m := rm[k][1];
    ptom := rm[k][2];
    rep := rm[k][3];
    p := #BaseRing(m); d := Dimension(m);
    vprint LMG,1: "Lifting through layer",k,"of size",p,"^",d;
    //get action of generators of imc on layer
    qpc, pctoqpc := quo< rpc | cs[k] >;
    V := VectorSpace(GF(p),d+1);
    convec := V![ Eltseq( (h^-1*g) @ rtopc @ ptom ) cat [1] ];
    //this must be in the orbit if g,h are conjugate 

    //start with action of rpc
    cqpc := sub< qpc | pctoqpc(cs[k+1]), pctoqpc(cpc) >;
    cmats := [];
    col := Matrix(GF(p),d+1,1,[0: i in [1..d]] cat [1]);
    for i in [1..NPCgens(cqpc)] do
      pg := cqpc.i @@ pctoqpc; ig := pg @@ rtopc;
      mat := HorizontalJoin(
         VerticalJoin( rep(pg), Matrix((h,ig) @ rtopc @ ptom)), col );
      Append(~cmats,  mat);
    end for;
    cqpcact := hom< cqpc->GL(d+1,p) | cmats >;
    fact := func<  t | t[1] * cqpcact(t[2]) >;
    rado,radst,radonos := PCOSG(cqpc,fact,V.(d+1));
    vprint LMG,2: "  #rado:",#rado,"#radst",#radst;

    //now action of imc on orbits
    igenmats := [ GL(d+1,p) | HorizontalJoin(
         VerticalJoin( gmacts[k](icgens[i]:im:=imc.i),
                         Matrix((h,icgens[i]) @ rtopc @ ptom)), col ):
                             i in [1..#icgens] ];
    //need two strategies for action of rq
    lo := #rado;
    if lo le so then
      vprint LMG,2: "  small orbit calculation";
      fullgp := sub< GL(d+1,p) | cmats cat igenmats >;
      subgp := sub< GL(d+1,p) | cmats >;
      if not convec in Orbit(fullgp,V.(d+1)) then return false,_; end if;
      oa, oi := OrbitAction(fullgp,V.(d+1));
      isc, x := IsConjugate(oi, oa(V.(d+1)), oa(convec)); assert isc;
      suborbs := Orbits(oa(subgp));
      ba, bi := BlocksAction(oi,suborbs[1]);
      rqact := hom< imc->bi | [igenmats[i] @ oa @ ba : i in [1..#icgens]] >;
      x := x @ ba @@ rqact @@ Gtorq;
      vprint LMG,2: "  Adjusting conjugating element";
      ans *:= x^-1; g := g^(x^-1);
      rqst := Stabiliser(bi,1) @@ rqact;
      vprint LMG,2: "  degree oi:",Degree(oi),"#rqst",#rqst;
    else
      vprint LMG,2: "  non-small orbit calculation";
      fullo := rado; deg := 1;
      perms := [ [] : i in [1..#igenmats] ];
      pt := 1;
      while pt le deg do
        vec := fullo[(pt-1)*lo+1];
        for i in [1..#igenmats] do
          imvec := vec*igenmats[i];
          pos := Position(fullo,imvec);
          if pos eq 0 then //new orbit
            fullo join:=  PCOSG(cqpc,fact,imvec);
            deg +:= 1;
            perms[i][pt] := deg;
          else perms[i][pt] := (pos-1) div lo + 1;
          end if;
        end for;
        pt +:= 1;
      end while;
      convecpos := Position(fullo,convec); 
      if convecpos eq 0 then return false,_; end if;
      convecpos := (convecpos-1) div lo + 1;
      ChangeUniverse(~perms, Sym(deg));
      gorbim := sub< Sym(deg) | perms >;
      rqact := hom< imc->gorbim | perms >;
      isc, x := IsConjugate(gorbim, 1, convecpos); assert isc;
      x := x @@ rqact @@ Gtorq;
      vprint LMG,2: "  Adjusting conjugating element";
      ans *:= x^-1; g := g^(x^-1);
      rqst := Stabiliser(gorbim,1) @@ rqact;
      vprint LMG,2: "  #fullo:",#fullo,"#rqst",#rqst;
    end if;
    assert  imc.1 in rqst;
    //we still need to conjugate  g to h in cqpc
    convec := V![ Eltseq( (h^-1*g) @ rtopc @ ptom ) cat [1] ];
    isc,x := PCActGIsCon(cqpc,fact,rado,radonos,convec); assert isc;
    vprint LMG,2: "  Adjusting conjugating element in radical";
    x := x @@ pctoqpc @@ rtopc;
    ans *:= x^-1; g := g^(x^-1);

    //try to minimise generators
    stg := [imc.1];
    while sub< rqst | stg > ne rqst do
      repeat x:=Random(rqst); until not x in sub< rqst | stg >;
      Append(~stg,x);
    end while;
    rqst := sub< rqst | stg >;
    istg := [Generic(G)|g];
    ords := [ #quo<cs[i]|cs[i+1]> where cs:=CompositionSeries(cqpc) :
                    i in [1..NPCgens(cqpc)] ]; //should be a standard function!
    iwmimc := InverseWordMap(imc);
    for l in [2..#stg] do
      slp := iwmimc(stg[l]);
      smat := Evaluate(slp, igenmats);
      vec := V.(d+1) * smat;
      isc,x := PCActGIsCon(cqpc,fact,rado,radonos,vec); assert isc;
      istg[l] := Evaluate(slp,icgens) * ((x^-1) @@ pctoqpc @@ rtopc);
    end for;
    cpc := sub< rpc | [ radst.i @@ pctoqpc : i in [1..NPCgens(radst)] ] >;
    icgens := istg;
    imc := rqst;
  end for; //for k in [#cs-1..1 by -1] 

  return true, ans;
end function;

ConjClasses := function(G:so:=256)
  local r,rpc,rad,rtopc,rq,Gtorq,rm,gmacts,img,imc,cgens,x,cpc,icgens,cs,m,ptom,
        rep,p,d,qpc,pctoqpc,cqpc,cmats,col,pg,ig,mat,V,fact,rado,radst,radonos,
        igenmats,lo,fullgp,subgp,oa,oi,suborbs,ba,bi,rqact,rqst,fullo,deg,pt,
        perms,vec,imvec,pos,gorbim,stg,istg,ords,iwmimc,slp,smat,pow,y,rqcl,
        list,orbs,v,w,ans,g,h,ncpc,orbrep;
  r := G`LMGNode;
  rpc := r`radPC;
  rad := r`rad;
  rtopc := r`radtoPC;
  rq := r`radquot;
  Gtorq := r`Gtoradquot;
  rm := r`radmods;
  gmacts := r`rqradmodactions;
  cs := r`cseriesrad;

  rqcl := Classes(rq);
  vprint LMG, 1: #rqcl, "classes in radical quotient";
  //ans will contain returned answer,
  //list will contain elements and their centralisers to be lifted
  list := [car<Generic(G),PowerStructure(GrpPC),PowerStructure(GrpPerm),
                                     PowerSequence(Generic(G))>| ];
  tct:=0;
  for rqc in rqcl do
    tct+:=1; vprint LMG,2:"  Radical quotient class",tct;
    img := rqc[3]; g := img @@ Gtorq;
    imc := Centraliser(rq, img);
    cgens := [img];
    while sub< imc | cgens > ne imc do
      repeat x:=Random(imc); until not x in sub< imc | cgens >;
      Append(~cgens,x);
    end while;
    imc := sub< imc | cgens >;
    cpc := sub< rpc | >;
    icgens := [g] cat [cgens[i] @@ Gtorq : i in [2..#cgens]];
    Append(~list, <g, cpc, imc, icgens > );
  end for;
  vprint LMG,1: #list,"classes in radical quotient";

  for k in [#cs-1..1 by -1] do

    //lift through layer cs[len]/cs[len-1]
    newlist := [car<Generic(G),PowerStructure(GrpPC),PowerStructure(GrpPerm),
                                     PowerSequence(Generic(G))>| ];
    m := rm[k][1];
    ptom := rm[k][2];
    rep := rm[k][3];
    p := #BaseRing(m); d := Dimension(m);
    vprint LMG,1: "Lifting through layer",k,"of size",p,"^",d;
    qpc, pctoqpc := quo< rpc | cs[k] >;

    tct:=0;
    for t in list do
      //get action of generators of C on layer
      tct+:=1; vprint LMG,2:"  Class number",tct;
      g := t[1]; cpc := t[2]; imc := t[3]; icgens := t[4];
      cqpc := sub< qpc | pctoqpc(cs[k+1]), pctoqpc(cpc) >;
      cmats := [];
      col := Matrix(GF(p),d+1,1,[0: i in [1..d]] cat [1]);
      for i in [1..NPCgens(cqpc)] do
        pg := cqpc.i @@ pctoqpc; ig := pg @@ rtopc;
        mat := HorizontalJoin(
           VerticalJoin( rep(pg), Matrix((g,ig) @ rtopc @ ptom)), col );
        Append(~cmats,  mat);
      end for;
      cqpcact := hom< cqpc->GL(d+1,p) | cmats >;
      V := VectorSpace(GF(p),d+1);
      //igenmats := [ GL(d+1,p) | HorizontalJoin(
      //   VerticalJoin( gmacts[k](x), Matrix((g,x) @ rtopc @ ptom)), col ):
      //                       x in icgens ];
      igenmats := [ GL(d+1,p) | HorizontalJoin(
         VerticalJoin( gmacts[k](icgens[i]:im:=imc.i),
                       Matrix((g,icgens[i]) @ rtopc @ ptom)), col ):
                             i in [1..#icgens] ];
      fullgp := sub< GL(d+1,p) | cmats cat igenmats >;
      subgp := sub< GL(d+1,p) | cmats >;
      if p eq 2 then
        orbs := Orbits(fullgp); //only want those with nontrivial d+1 entry
        orbs := [ o: o in orbs | Rep(o)[d+1] ne 0 ];
      else
        orbs := LineOrbits(fullgp); //only want those with nontrivial d+1 entry
        orbs := [ o: o in orbs | (Rep(o).1)[d+1] ne 0 ];
      end if;
      vprint LMG,2: "  ",#orbs,"orbits";
      for orb in orbs do
        orbrep := orb[1];
        vprint LMG,3: "    Doing next orbit computation";
        vvec := p eq 2 select orbrep else (vec/vec[d+1] where vec := orbrep.1);
        h := g * (m!Prune(Eltseq(vvec))) @@ ptom @@ rtopc;
        fact := func<  t | t[1] ^ cqpcact(t[2]) >;
        rado, radst, radonos := PCOSG(cqpc, fact, orbrep);
        vprint LMG,3: "    #rado:",#rado,"#radst",#radst;
        //two strategies for action of rq
        lo := #rado;
        if forall{m : m in cmats cat igenmats | orbrep ^ (GL(d+1,p)!m) in rado}
         then
         //easiest case: rado fixed by all gens
           vprint LMG,3:"    rado fixed by rq generators";
           rqst := imc;
           stg := [ imc.i : i in [1..Ngens(imc)] ];
        elif lo le so then
          vprint LMG,3: "    small orbit calculation";
          oa, oi := OrbitAction(fullgp, orbrep);
          suborbs := Orbits(oa(subgp));
          ba, bi := BlocksAction(oi,suborbs[1]);
          rqact := hom< imc->bi | [igenmats[i] @ oa @ ba : i in [1..#icgens]] >;
          rqst := Stabiliser(bi,1) @@ rqact;
          vprint LMG,3: "    degree oi:",Degree(oi),"#rqst",#rqst;
          stg := [imc.1];
          //minimise generators
          while sub< rqst | stg > ne rqst do
            repeat x:=Random(rqst); until not x in sub< rqst | stg >;
            Append(~stg,x);
          end while;
          rqst := sub< rqst | stg >;
        else
          vprint LMG,3: "    non-small orbit calculation";
          fullo := rado; deg := 1;
          perms := [ [] : i in [1..#igenmats] ];
          pt := 1;
          while pt le deg do
            vec := fullo[(pt-1)*lo+1];
            for i in [1..#igenmats] do
              imvec := vec ^ GL(d+1,p)!igenmats[i];
              pos := Position(fullo,imvec);
              if pos eq 0 then //new orbit
                fullo join:=  PCOSG(cqpc,fact,imvec);
                deg +:= 1;
                perms[i][pt] := deg;
              else perms[i][pt] := (pos-1) div lo + 1;
              end if;
            end for;
            pt +:= 1;
          end while;
          ChangeUniverse(~perms, Sym(deg));
          gorbim := sub< Sym(deg) | perms >;
          rqact := hom< imc->gorbim | perms >;
          rqst := Stabiliser(gorbim,1) @@ rqact;
          vprint LMG,3: "    #fullo:",#fullo,"#rqst",#rqst;
          stg := [imc.1];
          //minimise generators
          while sub< rqst | stg > ne rqst do
            repeat x:=Random(rqst); until not x in sub< rqst | stg >;
            Append(~stg,x);
          end while;
          rqst := sub< rqst | stg >;
        end if;
        istg := [Generic(G)|h];
        iwmimc := InverseWordMap(imc);
        for l in [2..#stg] do
          slp := iwmimc(stg[l]);
          smat := Evaluate(slp, igenmats);
          w := orbrep ^ smat;
          isc,x := PCActGIsCon(cqpc,fact,rado,radonos,w); assert isc;
          istg[l] := Evaluate(slp,icgens) * ((x^-1) @@ pctoqpc @@ rtopc);
        end for;
        ncpc := sub< rpc | [ radst.i @@ pctoqpc : i in [1..NPCgens(radst)] ] >;
        Append(~newlist, <h,ncpc,rqst,istg> );
      end for;
    end for;
    list := newlist;
     vprint LMG,1: #list,"classes in this quotient";
  end for;

  ans := [car<IntegerRing(), IntegerRing(), Generic(G)> | ];
  for t in list do
    g := t[1]; cpc := t[2]; imc := t[3];
    Append(~ans, <Order(g), LMGOrder(G) div (#cpc * #imc), g> );
  end for;
  return ans;
end function;
