freeze;

/*
Miscellaneous LLL functions.
*/

intrinsic LLL(Q::[Mtrx]: Delta := 0.75, Sort := false, Proof) -> []
{Apply LLL to the matrices in the sequence Q}

    if #Q eq 0 then
	return 0;
    end if;

    B := Matrix([Eltseq(x): x in Q]);
    if 1 eq 1 then
	_, T := LLLGram(
	    B*Transpose(B): Delta := Delta, Sort := Sort, Proof := Proof
	);
	B := Matrix(BaseRing(Parent(B)), T)*B;
	delete T;
    else
	B := LLL(B: Delta := Delta, Sort := Sort, Proof := Proof);
    end if;

    if ISA(Type(Q[1]), ModTupRngElt) then
	return Rows(B);
    end if;

    nr := Nrows(Q[1]);
    nc := Ncols(Q[1]);
    return [Matrix(nr, nc, Eltseq(B[i])): i in [1 .. Nrows(B)]];
end intrinsic;

intrinsic QuickLLL(X::Mtrx: Delta := 0.75, UseGram := 0) -> Mtrx
{Quick LLL applied to matrix X (no proof, sort rows)}

    if UseGram cmpeq 0 and Ncols(X) gt 10*Nrows(X) then
	UseGram := true;
    end if;

    if UseGram cmpne 0 then
	//return LLL(X: Proof := false, InitialSort, Sort, UseGram := UseGram);
//time "Get Gram";
	time F := X*Transpose(X);
//time "Do LLLGram";
	_, T := LLLGram(F: Proof := false, InitialSort, Sort);
	delete F;
//"Gram T:", Parent(T);
	return Matrix(BaseRing(Parent(X)), T)*X, T;
    else
	return LLL(X: Proof := false, InitialSort, Sort, Delta := Delta);
    end if;

end intrinsic;

intrinsic QuickLLLGram(F::Mtrx: Delta := 0.75) -> Mtrx
{Quick LLL applied to Gram matrix F (no proof, sort rows)}

    return LLLGram(F: Proof := false, InitialSort, Sort, Delta := Delta);

end intrinsic;
