freeze;

/////////////////////////////////////////////////////////////////////////
// dim3_canonical.m
/////////////////////////////////////////////////////////////////////////
// $Revision: 38043 $
// $Date: 2012-04-01 01:05:07 +1100 (Sun, 01 Apr 2012) $
// $LastChangedBy: kasprzyk $
/////////////////////////////////////////////////////////////////////////
// Author: Alexander Kasprzyk
/////////////////////////////////////////////////////////////////////////
// Reverse look-up tables for obtaining the database ID of a
// three-dimensional canonical polytope.
/////////////////////////////////////////////////////////////////////////
// Note: The tables were automatically generated; do NOT edit!
// For details on how this data was constructed, see:
//  nonexport/package/ToricGeom/reverse/canonical3/
/////////////////////////////////////////////////////////////////////////

import "dim3_canonical_facet.m": canonical_dim3_facet_to_id;
import "../read.m": open_database_block;

/////////////////////////////////////////////////////////////////////////
// Look-up tables
/////////////////////////////////////////////////////////////////////////

// Returns the index in the D array, as given by the V array, or 0.
function index_V(P,V)
    v:=Volume(P);
    for i in [1..#V by 2] do
        if V[i] eq v then return V[i + 1]; end if;
    end for;
    return 0;
end function;

// Returns the stard and finish indices in the I array, as given by the D array,
// or 0.
function index_D(P,D,i)
    d:=Volume(Dual(P));
    start:=D[i];
    i +:= 1;
    while D[i] ne 0 do
        if D[i] eq d then
            return start,start + D[i + 1] - 1;
        else
            start +:= D[i + 1];
            i +:= 2;
        end if;
    end while;
    return 0,_;
end function;

// Unpacks the candidate ids for the given polygon. Returns true followed by the
// possible ids on success, false otherwise.
function unpack_ids(P,V,D,I)
    idx:=index_V(P,V);
    if idx eq 0 then return false,_; end if;
    start,finish:=index_D(P,D,idx);
    if start eq 0 then return false,_; end if;
    return true,I[start..finish];
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 0 mod 47.
function cached_ids_0(P,hash1,hash2,hash3)
    if hash1 eq 1175 and hash2 eq 3387 and hash3 eq 1526 then
        V:=[23,1,24,5,25,13,26,23,27,33,28,39];
        D:=[1,39/2,1,0,2,18,1,59/3,1,20,1,0,5,33/2,1,107/6,1,109/6,1,233/12,1,0,
            9,16,1,49/3,5,50/3,1,56/3,1,0,17,89/6,1,101/6,1,0,19,40/3,1,0];
        I:=[247331,247386,247333,247022,247176,247400,247389,246818,247129,
            246905,247163,247172,247317,247318,247309,246568,246629,246492,
            246622];
    elif hash1 eq 2914 and hash2 eq 3072 and hash3 eq 105 then
        V:=[23,1,24,5,25,9,26,23,27,33,28,41,29,49,30,53,32,61];
        D:=[1,151/6,1,0,2,64/3,1,0,3,79/4,1,119/6,1,121/6,1,131/6,1,167/6,1,
            203/6,1,0,9,55/3,1,239/12,1,263/12,1,335/12,1,0,13,183/10,1,75/4,2,
            41/2,1,0,17,223/15,2,76/5,1,50/3,1,0,21,509/30,2,0,23,51/5,1,31/3,1,
            147/10,1,0,26,347/35,1,211/21,1,151/15,1,51/5,1,0];
        I:=[537285,537286,537345,537314,537332,537291,537327,537316,537315,
            537341,537288,537324,537353,537340,537343,537322,537306,537334,
            537344,537290,537283,537311,537273,537313,537302,537305,537338,
            537300,537335];
    elif hash1 eq 658 and hash2 eq 2338 and hash3 eq 407 then
        V:=[24,1,25,11,26,21,27,25,29,33,30,41];
        D:=[1,82/3,1,30,1,94/3,1,95/3,1,0,5,87/4,1,131/6,1,157/6,2,167/6,1,0,10,
            43/2,1,0,11,197/12,1,101/6,1,599/30,1,0,14,44/3,1,91/6,1,205/12,1,0,
            17,787/60,1,203/15,1,0];
        I:=[418696,416749,418489,418759,416168,418527,417857,418694,417855,
            417755,416617,418338,415556,416511,418521,416175,416613,417726];
    elif hash1 eq 1551 and hash2 eq 14 and hash3 eq 3302 then
        V:=[21,1,22,5,24,19,25,25,26,29];
        D:=[1,53/2,2,0,3,21,1,23,2,74/3,1,80/3,1,27,2,29,1,0,11,20,1,65/3,1,0,
            13,121/6,2,0,15,49/3,2,263/12,1,0];
        I:=[248835,248868,248869,248806,248983,249142,249146,248802,248836,
            249211,248834,247898,247853,247906,248249,248251,247547];
    elif hash1 eq 893 and hash2 eq 294 and hash3 eq 842 then
        V:=[21,1,22,5,23,13,24,27,25,31,26,37,27,43,28,59,29,63];
        D:=[1,73/2,1,0,2,25,1,80/3,1,92/3,1,0,5,127/6,1,43/2,1,55/2,1,115/4,1,
            175/6,2,59/2,2,0,13,77/3,1,0,14,71/4,1,37/2,1,0,16,55/3,1,101/5,1,0,
            18,173/12,1,29/2,1,311/20,1,197/12,1,101/6,1,103/6,1,551/30,1,0,25,
            72/5,1,0,26,38/3,1,0];
        I:=[411581,410538,410249,412187,410364,410518,411889,411544,410528,
            412209,411862,411890,410199,411090,411791,410797,410208,410874,
            410529,409655,411076,411854,411021,410206,410969,412268];
    elif hash1 eq 3055 and hash2 eq 3396 and hash3 eq 2366 then
        V:=[26,1,28,7,29,13,30,17,32,23];
        D:=[1,21,1,68/3,1,0,3,16,1,53/3,2,0,6,97/6,2,0,8,57/4,1,43/3,7,0,16,
            34/3,1,0];
        I:=[667247,667177,667197,665600,667490,665195,665271,665278,665938,
            666010,666850,667083,667096,667212,667291,665599];
    elif hash1 eq 1880 and hash2 eq 790 and hash3 eq 22 then
        V:=[17,1,18,7,19,13,20,21,21,27,22,31,23,37,26,43,27,47];
        D:=[1,45/2,1,49/2,1,0,3,31,1,98/3,1,0,5,83/4,1,91/4,1,139/6,1,0,8,116/5,
            1,156/5,1,0,10,257/12,1,0,11,50/3,1,55/3,1,0,13,217/12,1,37/2,1,0,
            15,183/10,1,0,16,239/20,1,0];
        I:=[516731,516448,516579,516736,516549,516179,516593,516121,515819,
            516307,515941,516192,515799,515816,515622,515993];
    elif hash1 eq 3431 and hash2 eq 197 and hash3 eq 1049 then
        V:=[17,1,18,5,19,13,20,19,21,27,22,39,23,49];
        D:=[1,61/2,2,0,3,25,1,27,3,86/3,1,0,8,47/2,3,151/6,3,0,14,65/3,3,24,1,
            101/4,1,0,19,121/6,3,41/2,1,87/4,1,131/6,1,133/6,1,0,26,55/3,2,56/3,
            2,101/5,1,62/3,1,0,32,101/6,2,0];
        I:=[254106,254107,254104,254082,254102,254105,254099,254075,254100,
            254101,254059,254066,254090,254062,254091,254092,254041,254044,
            254068,254096,254097,254084,254048,254053,254054,254050,254055,
            254047,254051,254030,254032,254029,254093];
    elif hash1 eq 282 and hash2 eq 945 and hash3 eq 2652 then
        V:=[26,1,27,5,28,11,29,19,30,25,32,31];
        D:=[1,25,1,0,2,43/2,1,151/6,2,0,5,53/3,1,116/5,1,71/3,1,0,8,109/6,1,
            121/6,2,0,11,167/12,1,215/12,1,0,13,151/12,1,13,1,40/3,1,15,1,0];
        I:=[232000,233975,233215,234155,232197,230147,233309,232324,230142,
            232075,234033,233280,234001,232239,232565,229025];
    elif hash1 eq 0 and hash2 eq 875 and hash3 eq 608 then
        V:=[23,1,24,9,25,19,26,27,27,31,28,43,29,47,30,51,31,57];
        D:=[1,175/6,1,59/2,1,187/6,1,0,4,24,1,89/3,1,94/3,1,95/3,1,0,8,145/6,1,
            157/6,1,181/6,1,0,11,67/3,1,0,12,101/6,2,37/2,1,113/6,1,115/6,1,
            39/2,1,0,18,58/3,1,0,19,67/4,1,0,20,178/15,2,143/12,1,0,23,347/30,1,
            697/60,3,0];
        I:=[502639,502617,502170,502168,502118,502734,502165,502784,502731,
            502781,502735,502163,502641,502116,502737,502683,502774,502771,
            502780,502164,502779,502123,502749,502171,502736,502778];
    elif hash1 eq 2867 and hash2 eq 662 and hash3 eq 1526 then
        V:=[14,1,15,5,16,13,17,23,18,33,19,41,20,49,21,57,22,61];
        D:=[1,38,1,0,2,61/2,1,65/2,1,73/2,1,0,5,23,1,74/3,1,35,1,116/3,1,0,9,
            43/2,1,163/6,1,55/2,1,175/6,1,0,13,59/3,1,64/3,2,26,1,0,17,109/6,1,
            217/10,1,111/4,1,0,20,298/15,1,23,1,105/4,1,0,23,101/6,1,0,24,
            248/15,1,0];
        I:=[518098,518102,518107,518113,518128,518086,518039,518035,518063,
            518047,518114,518068,518090,518079,518093,518106,518045,517976,
            517795,517985,518043,517801,518116,517878];
    elif hash1 eq 94 and hash2 eq 3110 and hash3 eq 3412 then
        V:=[14,1,15,19,16,29,17,41,18,53,19,59,20,69];
        D:=[1,26,1,30,1,32,1,34,3,36,2,38,1,40,1,44,1,0,12,57/2,1,69/2,1,73/2,1,
            81/2,1,0,16,23,2,25,1,27,1,92/3,1,104/3,2,0,23,127/6,3,151/6,2,51/2,
            1,163/6,1,71/2,1,0,31,59/3,1,83/3,1,0,33,233/12,1,119/6,1,133/6,1,
            169/6,1,0,37,68/3,1,0];
        I:=[429050,429049,428993,429042,429043,429061,429055,429059,428968,
            429026,429028,429047,429018,429032,428908,429051,429053,429007,
            428938,429013,429009,429029,428984,429014,429030,428914,428918,
            429000,428925,428763,428892,428939,428944,428934,428809,428658,
            428774];
    elif hash1 eq 2350 and hash2 eq 3378 and hash3 eq 415 then
        V:=[25,1,26,5,27,19,29,23,30,31,31,39,32,53];
        D:=[1,167/6,1,0,2,358/15,1,24,1,73/3,1,131/5,1,113/4,1,91/3,1,0,8,
            289/12,1,0,9,511/30,1,521/30,2,35/2,2,0,14,883/60,1,233/15,1,187/12,
            1,0,17,40/3,1,59/4,1,91/6,1,1079/70,1,667/42,1,159/10,1,0,23,82/7,1,
            0];
        I:=[417792,416896,416653,416646,416073,417252,416751,417510,417058,
            416894,417104,417094,417609,415573,416110,416664,416488,414927,
            415861,415024,415008,415002,415369];
    elif hash1 eq 987 and hash2 eq 2491 and hash3 eq 2059 then
        V:=[18,1,19,5,21,9,22,17,23,31,26,37,27,43,28,47,29,51];
        D:=[1,186/5,1,0,2,287/6,1,0,3,283/14,1,311/14,1,289/12,1,0,6,182/9,1,
            76/3,1,398/15,1,166/5,1,133/4,1,523/15,1,0,12,119/6,1,1129/42,1,0,
            14,557/36,1,337/21,1,0,16,243/20,2,0,18,616/45,1,0,19,2191/180,1,0];
        I:=[543770,543669,543665,543592,543667,543594,543613,543717,543614,
            543596,543751,543617,543643,543593,543650,543752,543760,543591,
            543595];
    elif hash1 eq 1081 and hash2 eq 2981 and hash3 eq 2145 then
        V:=[22,1,23,7,24,15,25,21,26,29,27,33];
        D:=[1,23,2,74/3,1,0,4,127/6,2,43/2,1,139/6,2,0,9,59/3,4,20,2,0,15,71/4,
            1,107/6,1,109/6,3,0,20,49/3,5,0,25,89/6,1,0];
        I:=[673571,673573,673663,673566,673567,673410,673407,673497,673379,
            673482,673593,673615,673389,673432,673537,673557,673242,673384,
            673617,673266,673292,673377,673385,673397,673296];
    elif hash1 eq 3055 and hash2 eq 1094 and hash3 eq 1468 then
        V:=[31,1,33,11,34,27,35,31];
        D:=[1,95/6,2,97/6,1,33/2,1,107/6,2,0,7,79/6,1,169/12,1,431/30,2,173/12,
            1,29/2,1,147/10,1,89/6,1,0,15,193/15,2,0,17,341/30,2,23/2,1,0];
        I:=[660673,662497,662456,662457,659860,662050,659991,660822,659820,
            659821,660784,660513,659803,660516,659990,660791,661469,661973,
            660844];
    elif hash1 eq 1974 and hash2 eq 2379 and hash3 eq 342 then
        V:=[22,1,23,11,24,21,26,31,29,37];
        D:=[1,21,2,68/3,1,25,1,29,1,0,6,35/2,1,115/6,1,39/2,2,127/6,3,0,13,16,1,
            52/3,1,96/5,1,59/3,1,0,17,57/4,1,43/3,1,0,19,117/10,1,0];
        I:=[411804,412144,412104,411851,410372,411716,411943,410900,411819,
            410930,410955,412281,410957,412043,410625,410358,412022,410974,
            411126];
    elif hash1 eq 1222 and hash2 eq 3251 and hash3 eq 834 then
        V:=[24,1,25,9,26,13,27,23];
        D:=[1,17,1,19,3,25,1,0,6,103/6,4,0,10,47/3,3,16,1,69/4,1,52/3,1,0,16,
            209/12,1,37/2,1,0];
        I:=[245680,245263,245831,245832,243197,244970,245169,245757,245791,
            243033,245267,245766,243968,245491,242955,244317,244391];
    elif hash1 eq 799 and hash2 eq 2650 and hash3 eq 2642 then
        V:=[24,1,25,7,26,11,27,17,28,25,29,31,30,39,31,51];
        D:=[1,82/3,1,89/3,1,0,3,143/6,1,0,4,298/15,1,287/12,1,0,6,221/12,1,
            125/6,1,127/6,1,0,9,103/7,1,223/15,1,0,11,59/4,1,50/3,1,17,1,0,14,
            274/21,1,787/60,1,79/6,1,203/15,2,41/3,1,0,20,117/10,1,47/4,1,0];
        I:=[416833,416318,416436,416366,416166,415439,416446,417371,415372,
            415462,414985,415456,415359,415401,415377,415465,417376,417513,
            416544,416424,416435];
    elif hash1 eq 3102 and hash2 eq 3130 and hash3 eq 2276 then
        V:=[17,1,18,11,19,23,20,33,21,45,22,63,23,69,24,75,25,87];
        D:=[1,49/2,1,53/2,1,57/2,1,61/2,1,0,5,19,1,27,1,29,2,98/3,1,110/3,1,0,
            11,127/6,1,43/2,1,163/6,2,175/6,1,0,16,53/3,1,18,1,65/3,1,93/4,1,
            136/5,1,0,21,107/6,2,197/10,1,121/6,1,87/4,1,131/6,1,237/10,1,145/6,
            1,49/2,1,0,30,50/3,1,101/5,1,0,32,33/2,1,551/30,1,0,34,76/5,1,46/3,
            1,248/15,1,253/15,1,17,2,0,40,213/14,1,157/10,1,95/6,1,0];
        I:=[515568,515548,515562,515572,515536,515565,515227,515241,515594,
            515222,515573,515538,515231,515560,515250,515466,515185,515543,
            515177,514550,515193,515535,514271,515593,515237,515199,514764,
            514880,515322,515203,515008,515559,514301,514548,514785,515002,
            514519,514881,514897,514080,514077,514047];
    elif hash1 eq 2961 and hash2 eq 2844 and hash3 eq 1908 then
        V:=[21,1,22,7,23,23,24,37,25,53,26,63,27,71,28,79];
        D:=[1,175/6,1,259/6,1,0,3,77/3,2,82/3,1,83/3,2,89/3,1,30,1,94/3,1,106/3,
            1,0,12,45/2,1,257/10,1,155/6,1,179/6,1,181/6,1,135/4,1,0,18,67/3,1,
            121/5,2,113/4,1,86/3,1,448/15,1,91/3,1,92/3,1,0,26,75/4,1,113/6,1,
            671/30,1,137/6,1,0,30,223/12,2,85/4,1,64/3,1,0,34,451/30,1,91/6,1,
            177/10,1,0,37,87/5,1,263/15,3,268/15,2,0];
        I:=[508718,508541,508584,508639,508758,508500,508701,508475,508547,
            508577,508474,508715,508589,508559,508756,508497,508467,508576,
            508477,508501,508463,508641,508571,508597,508757,508665,508719,
            508565,508615,508555,508613,508460,508679,508609,508601,508505,
            508596,508515,508625,508659,508663,508667];
    elif hash1 eq 1786 and hash2 eq 845 and hash3 eq 337 then
        V:=[25,1,26,9,27,21,28,29,29,43,30,55,31,63];
        D:=[1,43/2,2,47/2,2,151/6,1,0,6,77/4,1,58/3,1,59/3,2,20,2,65/3,1,0,13,
            97/6,1,109/6,2,41/2,1,0,17,44/3,2,191/12,1,81/5,2,49/3,1,50/3,2,
            73/4,3,0,28,431/30,1,29/2,1,91/6,1,197/12,1,33/2,1,0,33,155/12,4,13,
            1,15,1,0,39,137/12,1,23/2,1,0];
        I:=[240907,240912,240628,240889,240809,240384,241322,240544,241148,
            240910,241394,240265,240596,239974,240075,240624,239219,240223,
            240466,239835,240126,239855,240405,240940,239588,239995,239996,
            239194,240574,239378,238087,238634,238664,238841,239208,240192,
            239221,238154,239782,240438];
    elif hash1 eq 1504 and hash2 eq 1197 and hash3 eq 195 then
        V:=[21,1,22,11,23,19,24,27,25,33,26,39];
        D:=[1,45/2,1,53/2,1,57/2,1,65/2,1,0,5,21,1,23,1,25,1,0,8,35/2,1,127/6,1,
            43/2,1,0,11,47/3,1,22,1,0,13,185/12,1,95/6,1,0,15,57/4,2,44/3,1,0];
        I:=[412121,412059,411952,411762,412282,412220,411592,411747,412141,
            411987,412028,410827,411750,412091,411675,412246,411729];
    elif hash1 eq 752 and hash2 eq 1718 and hash3 eq 647 then
        V:=[14,1,15,5,16,9,17,13,18,17,19,23,20,27,21,37,22,41,23,45,24,49,25,
            53,26,57];
        D:=[1,40,1,0,2,93/2,1,0,3,146/3,1,0,4,199/6,1,0,5,100/3,1,36,1,0,7,
            181/6,1,0,8,311/12,1,478/15,1,181/5,1,145/4,1,0,12,519/20,1,0,13,
            409/20,1,0,14,1237/60,1,0,15,263/15,1,0,16,95/6,1,0,17,238/15,1,
            191/12,1,0];
        I:=[518749,518748,518734,518733,518742,518688,518692,518455,518563,
            518540,518525,518452,518413,518554,518369,518368,518367,518362];
    elif hash1 eq 2256 and hash2 eq 1515 and hash3 eq 2274 then
        V:=[22,1,23,9,24,23,25,33,26,43,28,49];
        D:=[1,80/3,1,92/3,1,31,1,0,4,139/6,1,47/2,1,151/6,2,51/2,1,163/6,1,
            175/6,1,0,11,59/3,1,65/3,1,71/3,1,76/3,1,0,15,217/10,1,131/6,1,
            133/6,3,45/2,2,0,22,56/3,1,61/3,1,0,24,218/15,1,175/12,2,15,6,
            253/15,1,17,4,0];
        I:=[248794,249087,249503,249127,248807,248778,248795,248697,248817,
            248710,248984,249151,249027,248685,247983,248690,247977,248717,
            248774,248784,249500,248780,249026,248486,248031,248556,247844,
            248026,248703,248714,248779,248907,247616,247623,247624,247639,
            247711];
    elif hash1 eq 3243 and hash2 eq 1551 and hash3 eq 2617 then
        V:=[24,1,25,9,28,17,29,23,30,35,31,45];
        D:=[1,77/3,1,88/3,1,30,1,0,4,217/10,1,133/6,1,145/6,1,0,7,253/15,1,52/3,
            2,0,10,439/30,1,15,1,451/30,2,181/12,1,91/6,1,0,16,53/4,1,317/21,1,
            907/60,2,61/4,1,0,21,677/60,2,34/3,1,0];
        I:=[417619,418375,418342,416946,417538,418240,417100,418261,418341,
            415234,416201,416950,417176,416823,416497,416655,415140,415237,
            415522,416208,416578,416654,416519];
    elif hash1 eq 47 and hash2 eq 1353 and hash3 eq 3167 then
        V:=[19,1,20,9,21,19,22,27,23,35,24,45,26,49];
        D:=[1,53/2,1,61/2,1,69/2,1,0,4,25,1,80/3,2,27,1,92/3,1,0,9,139/6,1,99/4,
            1,163/6,2,0,13,58/3,1,59/3,1,101/4,1,0,16,233/12,1,79/4,1,119/6,1,
            121/6,1,0,20,239/12,1,0,21,15,1,0];
        I:=[254002,254001,253952,253898,253779,253893,253968,253710,253888,
            253750,253662,253790,253973,253935,253492,253745,253769,253778,
            253718,253510,253525];
    elif hash1 eq 141 and hash2 eq 3393 and hash3 eq 126 then
        V:=[14,1,15,7,16,13,17,29,18,59,19,87,20,119,21,139,22,159,23,183,24,
            221,25,229,26,233];
        D:=[1,32,1,36,1,0,3,69/2,2,77/2,1,0,6,33,1,35,2,37,2,116/3,1,39,1,41,1,
            128/3,1,0,15,139/6,1,51/2,1,55/2,1,175/6,1,59/2,2,123/4,1,187/6,2,
            199/6,2,67/2,1,139/4,1,211/6,2,75/2,1,235/6,1,79/2,1,0,33,101/4,1,
            26,1,88/3,3,89/3,2,30,1,94/3,4,95/3,2,32,1,100/3,1,101/3,2,141/4,1,
            106/3,1,119/3,1,0,54,213/10,1,143/6,1,103/4,1,155/6,2,157/6,2,
            277/10,2,167/6,1,169/6,1,179/6,1,377/12,1,317/10,2,127/4,1,191/6,1,
            193/6,2,425/12,1,0,74,43/2,1,328/15,1,22,1,358/15,1,121/5,1,51/2,1,
            388/15,2,131/5,1,79/3,2,0,85,103/6,1,221/12,1,113/6,1,391/20,1,
            241/12,1,125/6,1,22,1,45/2,2,479/20,1,0,95,199/12,1,253/15,1,17,1,
            369/20,2,37/2,1,227/12,2,19,1,58/3,3,313/15,2,21,2,275/12,1,0,112,
            461/30,1,331/20,1,499/30,1,50/3,2,67/4,2,103/6,1,781/42,2,56/3,1,
            75/4,1,189/10,2,265/14,1,571/30,1,229/12,1,115/6,2,581/30,3,233/12,
            1,221/10,1,451/20,1,0,137,233/15,1,187/12,2,467/28,1,0,141,653/42,1,
            0,142,181/15,1,0];
        I:=[428415,428509,428443,428503,428416,428260,428350,428498,428487,
            428502,428476,428533,428531,428259,428261,428400,428352,428522,
            428309,428397,428212,428275,428445,428320,428420,428347,428360,
            428361,428459,428349,428247,428461,428197,428315,428290,428340,
            428515,428090,428480,428080,428381,428439,428469,428538,428307,
            428505,428485,428517,428427,428473,428109,428342,428010,427938,
            428379,428204,427903,428378,427888,428251,427856,428075,428339,
            428319,427760,428295,428031,428032,427768,427965,428065,428311,
            428130,427919,428079,428056,427935,428009,428296,427842,428025,
            427922,427758,428387,428030,428357,428428,428027,428016,428541,
            428133,427752,427781,427793,428059,428011,428058,427779,427923,
            427762,428220,428345,427749,427746,428331,428471,427940,428125,
            427899,427993,427754,427932,428043,427956,428085,428243,427791,
            428038,428245,427628,427678,427741,427672,427990,428104,427631,
            428140,427635,427753,427772,427751,427996,428006,427759,427693,
            427684,428061,427951,428052,427685,427661,428015];
    elif hash1 eq 423 and hash2 eq 1063 and hash3 eq 1436 then
        V:=[23,1,24,7,25,15,26,23,27,29,29,37,30,41];
        D:=[1,59/2,1,247/6,1,0,3,89/3,1,95/3,1,166/5,1,0,6,237/10,1,257/10,1,
            103/4,1,0,9,81/5,1,215/12,1,0,11,37/2,1,207/10,1,125/6,1,0,14,
            205/12,1,0,15,363/28,1,787/60,1,203/15,1,0];
        I:=[411259,410121,411366,411367,410088,411502,410477,410147,409298,
            411435,409117,410122,410087,409108,409137,409273,411449];
    elif hash1 eq 1504 and hash2 eq 2144 and hash3 eq 2318 then
        V:=[19,1,20,11,21,27,22,41,23,49,24,59,25,69,26,85];
        D:=[1,53/2,1,57/2,1,65/2,1,73/2,1,0,5,25,3,80/3,1,86/3,1,29,1,92/3,1,31,
            1,33,1,0,14,139/6,3,47/2,1,151/6,3,51/2,2,163/6,4,55/2,2,0,29,64/3,
            2,65/3,1,71/3,2,0,34,107/6,1,233/12,1,257/12,1,143/6,1,0,38,215/12,
            2,55/3,1,61/3,1,328/15,1,0,43,479/30,1,16,1,33/2,1,101/6,2,221/12,1,
            37/2,1,22,1,0,51,223/15,1,46/3,1,0];
        I:=[253250,253245,253228,253193,253221,253222,253251,253287,253292,
            253179,252974,253254,253225,252877,253288,253289,253255,252893,
            252950,253206,253208,253253,252853,252870,252929,252975,252849,
            253195,252921,253282,252954,252850,252869,252953,252919,252668,
            252549,252925,252935,253213,252575,252552,252606,252622,252670,
            252879,252968,252554,252538,252492,252660,252648];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 1 mod 47.
function cached_ids_1(P,hash1,hash2,hash3)
    if hash1 eq 1364 and hash2 eq 544 and hash3 eq 1109 then
        V:=[25,1,26,7,27,15,28,21,29,27,30,31,31,35,32,39];
        D:=[1,127/6,1,63/2,2,0,4,53/3,1,58/3,3,89/3,1,0,9,71/4,1,109/6,3,0,13,
            65/4,1,268/15,1,0,15,147/10,1,0,16,182/15,1,0,17,51/4,1,0,18,53/4,1,
            0];
        I:=[498581,498743,498824,498479,498458,498546,498593,498623,498797,
            498243,498773,498823,498788,498439,498441,498044,496384,495539];
    elif hash1 eq 2633 and hash2 eq 541 and hash3 eq 3283 then
        V:=[18,1,19,5,20,13,21,19,22,33,23,45,24,51,25,55,26,63];
        D:=[1,122/3,1,0,2,175/6,1,223/6,1,75/2,1,0,5,89/3,1,101/3,1,0,7,157/6,1,
            167/6,1,169/6,1,297/10,1,179/6,1,191/6,1,0,13,358/15,1,24,1,121/5,1,
            26,1,105/4,1,0,18,671/30,1,269/12,1,0,20,19,1,0,21,997/60,1,521/30,
            1,229/12,2,0,25,89/6,1,61/4,1,233/15,1,47/3,1,0];
        I:=[427328,427190,426988,427601,427362,426850,427192,427202,426864,
            426852,426862,427329,426866,427185,426868,427333,426483,426507,
            426508,427210,426550,426867,426486,426569,426691,426695,426857,
            426873];
    elif hash1 eq 330 and hash2 eq 244 and hash3 eq 2961 then
        V:=[24,1,25,5,26,13,29,21,30,25];
        D:=[1,33,1,0,2,163/6,4,55/2,1,59/2,1,0,8,70/3,1,71/3,2,24,1,0,12,101/6,
            3,0,15,15,2,0];
        I:=[671742,671297,671797,671798,672218,672246,672434,671267,671728,
            672328,671758,671268,671545,672243,671260,671801];
    elif hash1 eq 706 and hash2 eq 1076 and hash3 eq 835 then
        V:=[21,1,22,5,23,15,24,25,25,33,26,37,27,53,28,63,29,67,30,77];
        D:=[1,223/6,1,0,2,82/3,1,100/3,1,101/3,1,125/3,1,0,6,277/10,1,191/6,1,
            337/10,1,203/6,1,0,10,73/3,1,129/5,1,80/3,1,0,13,49/2,1,0,14,131/7,
            1,283/15,1,19,1,21,1,106/5,1,64/3,1,338/15,1,0,21,103/6,1,181/10,1,
            559/30,1,569/30,1,0,25,233/15,1,0,26,137/10,1,83/6,1,46/3,1,185/12,
            1,0,30,117/10,1,253/21,1,181/15,1,0];
        I:=[423305,423307,422861,423217,423716,422991,423719,422939,422989,
            423309,422607,422986,423234,422371,422389,423356,422972,422930,
            423352,422555,423063,422519,422662,422593,422853,422985,423360,
            422431,422430,422374,422390,422445];
    elif hash1 eq 189 and hash2 eq 2208 and hash3 eq 1934 then
        V:=[26,1,27,11,28,19,29,27,30,31,31,41];
        D:=[1,65/3,2,70/3,1,76/3,2,77/3,1,0,7,107/6,1,131/6,1,143/6,1,0,10,
            191/12,1,91/5,1,101/5,1,0,13,89/6,1,0,14,179/12,1,15,1,248/15,1,
            253/15,1,0,18,259/20,1,391/30,2,157/12,1,181/12,1,0];
        I:=[242506,244238,245445,244369,244376,245660,243019,243683,244075,
            244175,241964,242206,243703,244333,242436,241948,241946,244275,
            243237,243649,244173,241968];
    elif hash1 eq 2351 and hash2 eq 1187 and hash3 eq 1051 then
        V:=[25,1,26,7,29,11,30,23];
        D:=[1,51/2,1,163/6,1,0,3,71/3,6,0,9,33/2,1,67/4,1,101/6,1,103/6,2,35/2,
            1,0,15,15,2,46/3,1,0];
        I:=[670414,670564,670502,670544,670895,670902,670916,670952,669993,
            669979,670866,670937,670983,670891,670419,670665,670864];
    elif hash1 eq 3056 and hash2 eq 473 and hash3 eq 1358 then
        V:=[19,1,20,9,21,15,22,27,23,31,24,39,25,47,26,55,27,59];
        D:=[1,49/2,1,65/2,1,73/2,1,0,4,31,1,98/3,1,0,6,151/6,1,163/6,1,59/2,1,
            187/6,1,211/6,1,0,11,82/3,1,0,12,119/6,1,133/6,1,155/6,1,0,15,
            208/15,1,215/12,1,62/3,1,0,18,217/12,1,269/12,1,91/4,1,0,21,175/12,
            2,0,23,299/20,1,451/30,1,0];
        I:=[509010,508982,509002,509066,509035,508983,508980,508913,508976,
            508916,509155,508932,508936,508923,509333,509145,508973,508962,
            508473,508486,508917,509074,508471,508472];
    elif hash1 eq 1740 and hash2 eq 2981 and hash3 eq 2899 then
        V:=[8,1,9,13,10,23,11,35,12,45,13,55,14,69,15,79,16,93,17,107,19,121,21,
            137,22,151];
        D:=[1,44,1,46,1,48,2,50,2,52,1,0,8,85/2,1,93/2,2,97/2,2,113/2,1,0,14,37,
            1,122/3,1,41,1,43,1,146/3,1,0,19,223/6,1,247/6,1,259/6,1,87/2,1,0,
            23,101/3,1,107/3,1,36,1,119/3,1,0,27,181/6,1,215/6,4,217/6,2,437/10,
            1,263/6,1,265/6,1,0,37,28,1,91/3,1,538/15,1,181/5,1,0,41,911/30,2,
            307/10,1,185/6,1,1091/30,1,437/12,1,73/2,1,0,48,70/3,1,24,1,319/12,
            1,569/20,1,428/15,1,343/12,1,0,54,1357/60,1,459/20,1,691/30,1,
            277/12,1,701/30,2,47/2,1,0,61,107/6,1,1591/70,1,1919/84,1,1373/60,1,
            649/28,1,1393/60,1,70/3,1,0,68,3377/210,1,1355/84,1,323/20,1,227/14,
            1,487/30,1,977/60,1,0,74,191/15,2,0];
        I:=[520135,520137,520138,520140,520127,520136,520139,520129,520128,
            520131,520130,520132,520118,520126,520111,520120,520125,520112,
            520124,520122,520108,520119,520107,520109,520133,520103,520106,
            520104,520110,520114,520117,520116,520121,520090,520098,520086,
            520115,520123,520082,520078,520093,520094,520080,520097,520101,
            520089,520088,520105,520134,520102,520087,520084,520083,520091,
            520095,520081,520100,520079,520085,520099,520113,520070,520077,
            520069,520073,520072,520067,520076,520071,520074,520066,520075,
            520068,520092,520096];
    elif hash1 eq 1881 and hash2 eq 882 and hash3 eq 1607 then
        V:=[16,1,17,9,18,23,19,45,20,69,21,89,22,109,23,131,24,143];
        D:=[1,26,2,28,2,30,1,0,6,49/2,1,53/2,3,57/2,5,61/2,1,65/2,3,73/2,1,0,20,
            68/3,2,23,2,74/3,1,25,6,80/3,2,86/3,3,29,2,92/3,1,33,1,104/3,1,0,41,
            127/6,1,43/2,4,91/4,1,139/6,7,47/2,1,151/6,2,51/2,1,163/6,2,55/2,2,
            175/6,1,59/2,1,0,64,58/3,2,59/3,8,64/3,2,65/3,5,22,1,93/4,1,71/3,2,
            77/3,1,83/3,1,0,87,107/6,1,109/6,1,197/10,1,119/6,3,121/6,1,257/12,
            1,217/10,1,131/6,1,133/6,5,0,102,50/3,1,268/15,1,215/12,3,91/5,1,
            55/3,2,298/15,2,61/3,2,43/2,1,263/12,1,89/4,1,0,117,89/6,4,491/30,3,
            33/2,5,101/6,2,265/12,1,0,132,223/15,1,76/5,1,0];
        I:=[254764,254766,254736,254738,254742,254763,254753,427359,427521,
            254737,254741,254767,427384,427445,427406,254739,254756,427594,
            427374,427367,427506,254769,427465,426997,254754,427004,427026,
            427099,427199,427340,427028,427488,427033,427100,427482,427452,
            427461,427021,427544,427291,427436,254740,427071,427388,427395,
            427029,427084,427088,427214,427460,427468,427509,427599,427380,
            427008,427540,427110,427022,427293,426800,427103,427228,426705,
            427494,427516,427065,427138,427164,427344,427389,427407,427453,
            427501,426996,427087,427066,427078,427111,427383,427527,427543,
            427213,426821,427197,426711,426699,427390,427094,426792,426685,
            427142,427542,426861,427179,426733,426712,426662,426682,426831,
            427058,427145,426833,426799,426654,427143,427148,426676,426823,
            426994,426479,426731,426510,426679,426642,426764,426915,426998,
            427052,427055,427219,426578,426819,426840,426791,426793,426874,
            427405,427505,426752,426763,426447,426655,426627];
    elif hash1 eq 1176 and hash2 eq 1615 and hash3 eq 626 then
        V:=[18,1,19,11,20,21,21,35,22,41,23,49,24,53,25,61];
        D:=[1,24,1,26,1,28,1,30,1,0,5,37/2,1,41/2,1,65/2,1,73/2,1,0,9,21,1,68/3,
            1,23,1,74/3,1,80/3,1,29,1,0,15,91/4,1,55/2,1,0,17,53/3,1,65/3,3,
            77/3,1,0,22,97/6,1,0,23,49/3,1,268/15,1,55/3,1,0,26,147/10,1,0];
        I:=[420459,420534,420461,420475,420564,420462,420472,419996,420474,
            420457,420463,420473,420095,420130,420224,419302,420042,419319,
            420198,420211,419067,419983,419981,419113,419100,419264];
    elif hash1 eq 1881 and hash2 eq 1979 and hash3 eq 2496 then
        V:=[20,1,21,7,22,13,23,19,24,25,25,31,26,37,27,43];
        D:=[1,23,1,74/3,1,0,3,139/6,2,175/6,1,0,6,59/3,2,65/3,1,0,9,107/6,3,
            197/10,1,0,13,91/5,1,101/5,1,0,15,89/6,1,491/30,1,0,17,175/12,1,
            344/21,1,0,19,401/30,2,0];
        I:=[424166,424155,424196,424247,423123,423044,424201,423152,423331,
            424011,424083,423338,422384,423366,423158,423369,424203,422436,
            422381,423332];
    elif hash1 eq 2398 and hash2 eq 1734 and hash3 eq 2927 then
        V:=[28,1,29,13,30,23,31,27,32,31,33,35];
        D:=[1,70/3,1,71/3,3,77/3,5,26,1,83/3,1,0,12,131/6,2,133/6,1,237/10,1,
            143/6,1,0,17,55/3,1,0,18,101/6,3,0,21,175/12,1,0,22,27/2,4,0];
        I:=[667591,667509,667570,667637,666295,666603,667516,667776,667781,
            667771,667782,666328,667659,665381,665463,665461,666327,666680,
            667593,667702,666318,665469,667508,667645,667752];
    elif hash1 eq 2022 and hash2 eq 2044 and hash3 eq 355 then
        V:=[26,1,27,5,28,11,29,23,30,29];
        D:=[1,17,1,0,2,103/6,1,35/2,2,0,5,61/4,1,47/3,3,52/3,2,53/3,1,106/5,1,0,
            13,29/2,4,97/6,2,0,19,43/3,2,44/3,1,0];
        I:=[385096,384715,385043,385270,385088,384939,385118,385147,385133,
            385151,382216,383126,382031,382245,384984,385126,384480,384550,
            382538,383267,384344];
    elif hash1 eq 3432 and hash2 eq 1140 and hash3 eq 2089 then
        V:=[18,1,19,5,20,13,21,17,22,23,23,27];
        D:=[1,34,1,0,2,53/2,1,57/2,2,65/2,2,0,7,29,2,0,9,139/6,1,151/6,3,0,13,
            22,1,0,14,131/6,1,133/6,2,0];
        I:=[674586,674594,674588,674590,674550,674593,674555,674582,674589,
            674551,674556,674591,674557,674546,674547,674558];
    elif hash1 eq 1458 and hash2 eq 1919 and hash3 eq 1470 then
        V:=[20,1,21,13,22,27,23,35,24,45,25,53];
        D:=[1,25,2,80/3,1,86/3,1,92/3,2,104/3,1,0,8,115/6,1,127/6,2,43/2,1,
            151/6,2,63/2,3,211/6,1,0,18,59/3,1,64/3,1,71/3,1,0,21,173/10,1,
            107/6,3,119/6,2,145/6,1,0,28,18,1,56/3,2,111/5,1,0,32,311/20,1,
            479/30,1,16,1,359/20,1,18,1,269/12,1,0];
        I:=[419939,419940,420580,419936,420145,420575,420426,420570,420363,
            420414,420615,419941,420416,419517,419674,420362,419700,420577,
            419571,419668,419695,419694,420306,420424,419667,420251,419142,
            419747,419521,420423,419132,419544,419728,420382,419133,419135,
            419028];
    elif hash1 eq 800 and hash2 eq 1104 and hash3 eq 2830 then
        V:=[16,1,17,5,18,9,19,19,21,23,24,33,27,37,28,41];
        D:=[1,56/3,1,0,2,59/2,1,0,3,69/4,1,36,1,186/5,1,112/3,1,0,7,157/6,1,0,8,
            193/12,1,339/14,1,327/10,1,209/6,1,0,12,147/10,2,0,14,1857/140,2,0,
            16,734/63,1,1037/84,1,0];
        I:=[543767,543713,543768,543673,543715,543685,543719,543763,543657,
            543714,543689,543749,543764,543630,543671,543590,543586];
    elif hash1 eq 1082 and hash2 eq 522 and hash3 eq 2084 then
        V:=[21,1,22,7,23,25,24,31,26,41,27,57,28,77];
        D:=[1,187/6,1,67/2,1,0,3,24,1,76/3,1,77/3,1,82/3,1,83/3,2,28,1,88/3,1,
            89/3,2,0,13,133/6,1,49/2,1,0,15,263/12,1,121/5,1,73/3,1,129/5,1,0,
            19,253/15,3,17,1,52/3,2,369/20,1,277/15,1,223/12,1,227/12,2,0,30,
            27/2,1,59/4,1,449/30,1,15,1,181/12,4,91/6,5,31/2,2,95/6,2,559/30,1,
            0,48,53/4,1,317/21,1,907/60,1,0];
        I:=[422284,422134,422107,422271,422226,422180,421835,422082,422086,
            422165,422231,422278,422228,422106,421327,421686,421307,421204,
            421297,421326,421782,421465,421919,422250,421298,421181,421304,
            421255,421269,421316,421776,421510,422032,421680,421970,422034,
            422085,422245,422246,422249,422282,422289,421303,422156,421834,
            421981,421109,421549,421172,421175];
    elif hash1 eq 2492 and hash2 eq 1290 and hash3 eq 1052 then
        V:=[23,1,24,11,25,15,26,21,27,33,28,41,29,49];
        D:=[1,115/6,1,139/6,1,47/2,1,151/6,1,0,5,18,1,0,6,107/6,1,119/6,1,0,8,
            16,2,50/3,2,215/12,1,55/3,1,56/3,1,0,15,77/6,1,79/6,1,173/12,1,0,18,
            188/15,2,13,1,103/7,1,0,22,341/30,1,0];
        I:=[507948,507942,507985,507981,507952,508037,508024,507965,508002,
            507960,507961,507936,507967,507955,507996,507940,508008,507946,
            508029,507951,507874,508005];
    elif hash1 eq 1552 and hash2 eq 2710 and hash3 eq 2580 then
        V:=[20,1,21,5,22,9,23,13,24,17,25,27,26,33,27,37,28,49];
        D:=[1,31,1,0,2,63/2,1,0,3,77/3,1,0,4,53/2,1,0,5,263/12,1,67/3,1,68/3,1,
            73/3,1,0,9,113/6,1,245/12,1,0,11,223/12,1,0,12,181/12,1,91/6,2,
            185/12,1,511/30,1,103/6,2,0,19,187/12,1,0];
        I:=[422256,422297,422227,421845,421742,421791,421818,421339,422260,
            421709,421325,421768,422230,422232,421319,421234,421228,421229,
            421235];
    elif hash1 eq 612 and hash2 eq 2560 and hash3 eq 40 then
        V:=[13,1,14,5,15,11,16,15,17,19,18,25,19,29,20,35,22,43,23,47,25,51];
        D:=[1,117/2,1,0,2,122/3,1,158/3,1,0,4,271/6,1,0,5,137/3,1,0,6,227/6,1,
            397/10,1,0,8,38,1,0,9,361/12,1,971/30,2,0,12,122/5,2,428/15,1,181/6,
            1,0,16,1363/60,1,0,17,683/30,1,0,18,427/30,1,238/15,1,0];
        I:=[519422,519474,519416,519523,519415,519527,519464,519420,519488,
            519511,519530,519429,519430,519479,519439,519385,519348,519456,
            519346];
    elif hash1 eq 1928 and hash2 eq 217 and hash3 eq 696 then
        V:=[20,1,21,5,22,13,23,41,24,51,25,61,26,79,27,91,28,107,29,113];
        D:=[1,110/3,1,0,2,67/2,1,223/6,1,75/2,1,0,5,64/3,1,65/3,2,24,2,101/4,1,
            77/3,1,26,1,83/3,1,28,1,117/4,1,89/3,7,95/3,1,133/4,1,101/3,1,0,26,
            197/10,1,143/6,1,145/6,1,155/6,2,0,31,61/3,1,62/3,1,121/5,1,80/3,1,
            0,35,167/10,1,35/2,1,18,1,551/30,2,221/12,1,37/2,2,113/6,1,611/30,2,
            0,46,17,1,86/5,1,53/3,1,278/15,2,307/15,1,0,52,157/12,5,625/42,1,
            299/20,1,181/12,2,31/2,1,67/4,1,205/12,1,0,64,77/6,1,53/4,1,0,66,
            347/30,1,499/42,1,119/10,1,137/10,1,55/4,1,0];
        I:=[510537,510434,510158,510418,509744,509672,509695,510312,510313,
            509814,510461,510116,510305,510449,510373,510307,510308,510331,
            510337,510392,510401,510430,510391,510025,509867,509686,510154,
            510446,509885,509899,510024,510083,509477,508378,509694,510314,
            510290,510054,510128,510422,508869,510397,510086,509930,510021,
            510334,510114,510330,509897,510022,508561,510443,510457,510607,
            510636,510666,508856,508781,508387,508444,509933,508858,508788,
            510500,508850,509715,508864,508797,508433,508430];
    elif hash1 eq 1787 and hash2 eq 2377 and hash3 eq 159 then
        V:=[18,1,19,5,20,9,21,21,22,33,23,39,24,45,26,49];
        D:=[1,27,1,0,2,99/4,1,0,3,85/4,1,65/3,2,100/3,1,101/3,1,34,1,0,9,329/12,
            1,111/4,1,377/12,1,337/10,1,203/6,2,0,15,388/15,2,192/7,1,0,18,
            335/14,1,293/12,1,0,20,337/15,1,0,21,167/9,1,1123/60,1,0];
        I:=[542488,542430,542427,542426,542485,542434,542487,542486,542433,
            542489,541937,542367,542370,542372,542371,542379,541936,541941,
            542360,541920,541793,541794];
    elif hash1 eq 2210 and hash2 eq 1505 and hash3 eq 899 then
        V:=[17,1,19,5,20,11,21,21,22,31,23,39,24,43,25,53,26,57];
        D:=[1,69/2,1,0,2,175/6,1,67/2,1,0,4,65/3,1,77/3,1,82/3,1,89/3,1,0,8,
            143/6,1,155/6,1,167/6,1,61/2,1,0,12,91/5,1,61/3,1,67/3,1,0,15,
            599/30,1,0,16,179/12,1,15,1,121/6,1,449/20,1,0,20,181/12,1,0,21,
            233/15,1,0];
        I:=[426113,426139,426143,426134,426145,426184,425908,426114,425881,
            425975,425625,425889,425970,425893,425710,425895,425896,425642,
            425639,426152,425632];
    elif hash1 eq 1881 and hash2 eq 715 and hash3 eq 2554 then
        V:=[25,1,26,7,27,13,28,19,29,25,30,31,31,35];
        D:=[1,39/2,1,43/2,2,0,4,59/3,1,20,1,0,6,71/4,2,107/6,2,0,10,191/12,1,
            268/15,2,0,13,173/12,2,29/2,1,0,16,13,1,0,17,23/2,1,0];
        I:=[672055,671951,672427,671564,672026,671560,671594,671843,672152,
            671613,671135,671505,671363,672049,671187,671624,671630];
    elif hash1 eq 3150 and hash2 eq 2689 and hash3 eq 313 then
        V:=[21,1,22,5,24,9,25,17,26,23,27,31,28,39,29,45];
        D:=[1,41/2,1,0,2,19,1,0,3,52/3,2,59/3,2,106/5,1,0,8,177/10,1,37/2,1,0,
            10,31/2,1,49/3,2,262/15,1,0,14,27/2,1,91/6,1,37/2,2,0,18,72/5,1,
            223/15,1,0,20,391/30,1,0];
        I:=[506754,506623,506626,506629,504366,506363,506401,504367,506259,
            506781,504379,506402,505700,506585,503881,505582,505834,506160,
            506279,504461];
    elif hash1 eq 95 and hash2 eq 903 and hash3 eq 393 then
        V:=[17,1,18,13,19,31,20,45,21,67,22,83,23,99,24,109,25,117,26,123];
        D:=[1,49/2,1,53/2,1,57/2,2,61/2,1,65/2,1,0,7,23,5,74/3,2,80/3,1,27,1,
            86/3,3,29,1,31,1,98/3,1,0,22,127/6,1,139/6,5,47/2,3,151/6,1,163/6,1,
            55/2,1,0,34,59/3,1,20,1,85/4,2,64/3,3,65/3,4,93/4,1,71/3,1,24,2,
            126/5,1,83/3,1,0,51,37/2,1,119/6,3,121/6,4,41/2,1,257/12,1,217/10,2,
            87/4,1,0,64,73/4,1,55/3,4,56/3,1,298/15,1,101/5,1,62/3,1,322/15,1,0,
            74,33/2,1,101/6,4,37/2,1,265/12,1,0,81,44/3,2,61/4,1,47/3,1,0,85,
            461/30,1,157/10,1,0,87,37/3,1,0];
        I:=[426407,426411,426392,426418,426408,426393,426242,426261,426300,
            426412,426417,426225,426281,426308,426410,426217,426224,426317,
            426423,426090,426089,426214,426240,426247,426254,426255,426269,
            426083,426394,426409,426091,425860,426034,426245,426260,426058,
            426075,425997,426070,426382,426086,426244,426267,426379,426014,
            425839,425808,426096,425756,425865,426093,425772,425817,426068,
            425752,425868,426025,426044,426035,426109,425765,425795,426180,
            425848,425797,426009,426062,426204,426043,425853,425773,425856,
            425746,426288,425779,425803,426026,426085,425692,425660,425994,
            426003,425743,425818,425655,425653,425728];
    elif hash1 eq 2539 and hash2 eq 1622 and hash3 eq 2796 then
        V:=[25,1,26,5,27,11,28,19,29,25,30,35];
        D:=[1,151/6,1,0,2,71/3,1,83/3,1,0,4,119/6,1,87/4,1,237/10,1,0,7,101/5,1,
            62/3,1,0,9,197/12,2,33/2,1,101/6,1,217/12,1,0,14,76/5,1,46/3,1,
            247/15,1,0];
        I:=[240147,239521,239915,239643,239822,239329,240481,239328,239293,
            240325,239475,239766,239943,239852,239833,238162];
    elif hash1 eq 2351 and hash2 eq 2615 and hash3 eq 2377 then
        V:=[23,1,24,7,25,11,27,17,28,23,29,31,30,35];
        D:=[1,151/6,2,175/6,1,0,4,71/3,1,0,5,119/6,1,103/4,1,0,7,187/10,1,113/6,
            3,0,11,175/12,1,15,1,223/12,1,0,14,391/30,1,0,15,139/12,1,0];
        I:=[249600,249691,249375,248390,247998,247517,247681,247601,247657,
            247658,249449,249539,247500,248182,247805];
    elif hash1 eq 1176 and hash2 eq 2252 and hash3 eq 2063 then
        V:=[27,1,29,5,30,13,31,25,32,29,33,41];
        D:=[1,329/12,1,0,2,37/2,1,245/12,1,207/10,1,0,5,50/3,1,86/5,1,283/15,1,
            227/12,1,313/15,1,0,10,451/30,1,0,11,66/5,1,203/15,1,41/3,1,208/15,
            1,89/6,1,0,16,677/60,1,34/3,1,697/60,1,35/3,1,117/10,1,47/4,2,0];
        I:=[408018,407003,408116,406657,406408,405491,406935,406808,407107,
            405494,405471,405495,406641,405780,404498,408120,406717,408030,
            406716,407006,406271,406349];
    elif hash1 eq 3244 and hash2 eq 1062 and hash3 eq 2214 then
        V:=[22,1,23,9,24,17,25,25,27,29,28,33,29,41,30,49];
        D:=[1,149/4,1,113/3,1,38,1,0,4,257/10,1,297/10,1,127/4,1,0,7,111/5,1,
            121/5,1,74/3,1,0,10,245/12,1,0,11,209/12,1,0,12,44/3,1,301/20,1,
            76/5,1,0,15,181/12,1,303/20,1,153/10,2,0,19,341/30,1,253/21,1,
            1145/84,1,273/20,1,0];
        I:=[509625,509743,510605,509991,509913,510203,510001,510135,510496,
            509711,510503,508523,509313,510065,508447,508522,509219,509304,
            509721,508532,508421,508436];
    elif hash1 eq 471 and hash2 eq 615 and hash3 eq 580 then
        V:=[21,1,22,9,23,21,24,31,25,39,26,51,27,57];
        D:=[1,53/2,1,57/2,1,61/2,2,0,5,23,5,74/3,2,25,3,80/3,1,27,2,0,18,127/6,
            5,43/2,1,139/6,3,47/2,2,0,29,58/3,3,59/3,2,20,2,0,36,71/4,1,109/6,3,
            37/2,1,119/6,1,121/6,1,0,43,49/3,1,50/3,2,0,46,147/10,1,59/4,1,89/6,
            1,91/6,1,0];
        I:=[673690,673687,673671,673703,673647,673676,673677,673683,673711,
            673623,673624,673635,673686,673707,673491,673364,673719,673370,
            673633,673649,673714,673734,673395,673414,673462,673470,673460,
            673504,673374,673403,673701,673401,673441,673413,673665,673485,
            673356,673632,673650,673710,673269,673309,673310,673445,673448,
            673288,673248,673245,673315];
    elif hash1 eq 377 and hash2 eq 1288 and hash3 eq 2700 then
        V:=[12,1,13,9,14,17,15,23,16,27,17,41,18,51,19,61,20,69,22,73,24,79,25,
            83];
        D:=[1,42,1,50,1,54,1,0,4,85/2,1,93/2,1,101/2,1,0,7,122/3,1,43,1,0,9,
            211/6,1,0,10,146/5,1,89/3,1,101/3,1,106/3,1,107/3,1,131/3,1,0,16,
            103/4,2,277/10,1,401/12,2,215/6,1,0,22,328/15,1,111/5,1,335/12,1,
            431/12,1,0,26,731/30,1,519/20,1,28,1,0,29,23,1,0,30,187/12,1,347/15,
            1,0,32,1313/84,1,0,33,121/10,1,0];
        I:=[519550,519549,519568,519603,519588,519558,519604,519569,519560,
            519477,519515,519598,519557,519581,519493,519541,519542,519528,
            519582,519597,519503,519463,519455,519492,519510,519525,519428,
            519386,519517,519543,519367,519358,519448];
    elif hash1 eq 2727 and hash2 eq 1141 and hash3 eq 2153 then
        V:=[27,1,28,9,29,15,30,19,31,27,32,43];
        D:=[1,119/6,1,121/6,1,133/6,1,0,4,215/12,1,56/3,1,0,6,33/2,2,0,8,179/12,
            1,47/3,1,203/12,1,0,11,71/6,1,391/30,1,79/6,1,27/2,3,59/4,1,181/12,
            1,31/2,1,0,20,173/15,1,163/12,1,0];
        I:=[403570,403572,403537,403617,403472,403560,403564,403534,402904,
            402949,403643,403466,403464,403621,403623,403625,402776,402770,
            402827,403679,402686];
    elif hash1 eq 941 and hash2 eq 2267 and hash3 eq 2838 then
        V:=[22,1,23,5,24,13,25,29,26,41,27,47,28,59];
        D:=[1,17,1,0,2,103/6,1,35/2,1,107/4,1,0,5,47/3,1,52/3,1,59/3,1,20,1,
            106/5,1,22,1,116/5,1,0,12,157/10,2,95/6,2,97/6,1,177/10,1,193/10,1,
            0,19,238/15,1,81/5,1,0,21,431/30,1,29/2,1,89/6,2,227/14,1,67/4,1,0,
            27,193/15,1,302/21,1,218/15,1,0];
        I:=[503408,503510,503582,503635,503561,502685,502973,502463,502720,
            502414,502564,503104,503399,503280,503508,503564,502972,502638,
            502976,501976,503117,501094,502522,502545,502016,501575,503303,
            501831,501963];
    elif hash1 eq 3385 and hash2 eq 1984 and hash3 eq 811 then
        V:=[18,1,19,7,20,15,21,27,22,35,23,45,24,57,25,69,26,79];
        D:=[1,31,1,98/3,1,0,3,163/6,2,55/2,2,59/2,1,0,8,71/3,1,76/3,1,77/3,3,
            83/3,2,89/3,1,0,16,237/10,1,95/4,1,257/10,1,0,19,61/3,3,62/3,2,22,1,
            382/15,1,0,26,187/10,1,20,1,611/30,1,207/10,2,125/6,1,0,32,15,1,
            46/3,1,253/15,1,17,1,52/3,1,0,37,181/12,1,91/6,3,709/42,1,521/30,1,
            0,43,203/15,1,47/3,1,0];
        I:=[426194,426154,426131,426178,426148,426171,426179,426116,426173,
            426111,426140,426177,425890,426176,425898,425873,426133,425892,
            426175,426181,426183,425911,426126,425963,425712,425637,425701,
            425966,425884,425960,425967,425875,426108,425952,425950,425905,
            426172,426110,426112,426163,425626,425638,425641,425630];
    elif hash1 eq 3291 and hash2 eq 234 and hash3 eq 1242 then
        V:=[22,1,23,7,24,13,25,21,26,33,29,39];
        D:=[1,83/3,1,95/3,1,0,3,155/6,1,191/6,1,0,5,328/15,1,287/12,1,359/12,1,
            0,8,611/30,2,659/30,1,269/12,1,335/14,1,293/12,1,0,14,401/20,1,
            163/8,1,0,16,377/28,1,83/6,2,0];
        I:=[419449,419278,419184,419408,419186,419435,418981,419209,419464,
            419254,419438,418979,418982,419323,418934,418970,419860,419894];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 2 mod 47.
function cached_ids_2(P,hash1,hash2,hash3)
    if hash1 eq 1506 and hash2 eq 623 and hash3 eq 1786 then
        V:=[15,1,16,7,17,17,18,27,19,31,20,41,21,49,22,55];
        D:=[1,65/2,2,77/2,1,0,4,86/3,1,31,1,98/3,1,37,1,0,8,175/6,1,59/2,1,63/2,
            1,223/6,1,0,12,95/3,3,0,15,143/6,3,145/6,1,155/6,1,191/6,1,0,21,
            287/12,2,121/5,1,74/3,1,0,25,265/12,2,269/12,1,0,28,278/15,1,0];
        I:=[255569,255612,255653,255553,255650,255478,255595,255624,255641,
            255635,255415,255398,255468,255620,255583,255604,255621,255646,
            255557,255447,255444,255464,255414,255404,255388,255410,255389,
            255380];
    elif hash1 eq 3104 and hash2 eq 1130 and hash3 eq 2681 then
        V:=[23,1,24,9,25,19,28,31];
        D:=[1,145/6,1,257/10,1,155/6,1,0,4,61/3,1,68/3,1,131/5,1,80/3,1,0,8,
            101/6,1,551/30,1,113/6,2,599/30,1,45/2,1,0,14,203/15,2,41/3,1,
            208/15,3,0];
        I:=[421403,421431,421425,421379,421420,421141,421210,421590,421360,
            421366,421419,421434,421244,421205,421593,421588,421371,421422,
            421435];
    elif hash1 eq 284 and hash2 eq 3050 and hash3 eq 1934 then
        V:=[24,1,25,5,26,13,27,31,28,41,29,55,30,61,31,67,32,75];
        D:=[1,31,1,0,2,39/2,1,151/6,1,107/4,1,0,5,59/3,1,70/3,1,71/3,2,101/4,1,
            76/3,2,77/3,1,109/4,1,83/3,1,0,15,107/6,1,109/6,1,133/6,1,253/10,1,
            0,19,215/12,1,18,1,55/3,2,101/5,1,61/3,1,328/15,1,0,26,16,1,103/6,1,
            0,28,175/12,1,223/15,1,0,30,391/30,1,27/2,4,91/6,1,0,36,82/7,1,0];
        I:=[244489,244053,243205,243151,244056,244034,242348,243727,244108,
            242497,244248,244286,244312,244109,243172,243211,243380,241844,
            243111,244434,244124,244480,244215,244488,242190,242875,242804,
            243339,242892,244243,244042,244304,244406,244595,242119,241920];
    elif hash1 eq 472 and hash2 eq 1516 and hash3 eq 1967 then
        V:=[27,1,28,9,29,15];
        D:=[1,103/6,1,35/2,3,115/6,2,0,7,47/3,6,16,1,0,14,85/6,3,29/2,1,0];
        I:=[664409,664329,664356,664480,664216,664235,663571,664366,664542,
            664549,664550,664607,664340,663805,664518,664623,664357];
    elif hash1 eq 1553 and hash2 eq 3125 and hash3 eq 2516 then
        V:=[21,1,22,5,23,13,24,19,25,31,27,35];
        D:=[1,53/2,1,0,2,74/3,1,25,1,27,1,0,5,139/6,2,47/2,1,0,8,53/3,1,59/3,1,
            20,1,85/4,1,65/3,3,0,15,107/6,1,0,16,89/6,2,0];
        I:=[673547,673525,673529,673508,673511,673519,673510,673327,673550,
            673509,673325,673328,673333,673336,673330,673522,673523];
    elif hash1 eq 2258 and hash2 eq 659 and hash3 eq 394 then
        V:=[20,1,21,7,22,15,23,25,24,33,25,45,26,53,27,65,28,77,30,87];
        D:=[1,125/3,1,130/3,1,0,3,111/4,1,215/6,1,229/6,1,0,6,105/4,1,418/15,1,
            85/3,2,448/15,1,0,11,45/2,1,911/30,1,365/12,1,0,14,277/15,1,302/15,
            1,343/15,2,227/8,1,29,1,0,20,35/2,1,379/20,1,115/6,1,0,23,233/15,1,
            187/12,1,16,1,86/5,1,1267/60,1,0,28,169/12,1,103/6,1,527/30,1,
            1583/84,1,115/6,2,0,34,823/60,1,385/24,1,81/5,1,1577/84,1,0,38,
            541/45,1,49/4,1,0];
        I:=[542366,542266,542347,542154,542313,542357,542344,542161,542356,
            542327,542158,542300,542359,542286,542293,542317,542318,541831,
            541834,542105,542298,541786,542319,542353,542128,542358,542301,
            542310,542323,542325,541914,541909,541919,542303,541825,541829,
            541783,541774,541799];
    elif hash1 eq 425 and hash2 eq 2725 and hash3 eq 307 then
        V:=[24,1,25,5,26,13,27,19,29,23,30,29,31,39];
        D:=[1,117/4,1,0,2,143/6,1,277/10,1,167/6,1,0,5,328/15,1,80/3,1,0,7,
            221/12,1,0,8,157/10,1,205/12,1,0,10,12,1,203/15,2,41/3,1,233/15,1,0,
            15,117/10,2,0];
        I:=[413954,414269,414045,414121,413964,414334,413639,413498,413580,
            414368,414038,414403,414082,413500,413947,414366];
    elif hash1 eq 1271 and hash2 eq 1774 and hash3 eq 1697 then
        V:=[24,1,25,11,26,23,27,37,28,53,29,67,30,75];
        D:=[1,65/3,3,71/3,1,24,1,76/3,1,0,7,197/10,1,79/4,1,119/6,1,217/10,1,
            237/10,1,0,12,49/3,1,55/3,1,81/4,2,61/3,2,62/3,1,263/12,1,0,20,89/6,
            1,91/6,1,67/4,1,103/6,1,551/30,1,221/12,1,75/4,1,0,27,44/3,1,223/15,
            1,15,1,76/5,1,46/3,1,199/12,1,0,33,401/30,1,161/12,2,27/2,1,0,37,
            37/3,1,203/15,1,0];
        I:=[414487,414493,414692,414501,414754,414626,414230,414726,414444,
            414175,414178,414680,414288,413877,414398,414428,414514,414725,
            413819,414362,414378,414533,414495,414430,413497,413841,414563,
            414356,414497,413922,414431,413700,413931,413721,413842,414191,
            413481,413435];
    elif hash1 eq 1506 and hash2 eq 1658 and hash3 eq 1341 then
        V:=[20,1,21,7,22,17,23,25];
        D:=[1,26,1,28,1,0,3,45/2,1,49/2,2,53/2,3,61/2,1,0,10,68/3,3,23,3,25,1,0,
            17,127/6,2,0];
        I:=[674224,674238,674211,674210,674226,674180,674188,674239,674169,
            674168,674184,674219,674227,674231,674240,674182,674189,674221];
    elif hash1 eq 3198 and hash2 eq 2411 and hash3 eq 1375 then
        V:=[25,1,26,7,27,11,28,19,29,25,30,33];
        D:=[1,47/2,1,151/6,1,0,3,18,1,0,4,97/6,1,109/6,1,121/6,1,0,7,91/5,1,
            55/3,1,0,9,77/6,2,431/30,1,197/12,1,0,13,188/15,2,151/12,1,0];
        I:=[407648,408708,408634,407051,408506,408556,405606,405090,406938,
            407091,406947,405622,406908,407080,406833];
    elif hash1 eq 2681 and hash2 eq 92 and hash3 eq 869 then
        V:=[8,1,9,9,10,13,11,19,12,27,13,31,14,35,15,41,17,49];
        D:=[1,46,2,48,1,54,1,0,5,81/2,1,0,6,41,1,49,1,0,8,247/6,1,83/2,1,259/6,
            1,0,11,101/3,1,0,12,215/6,1,0,13,85/3,1,109/3,1,0,15,337/12,1,
            341/12,1,57/2,1,0,18,691/30,2,0];
        I:=[520157,520159,520158,520154,520156,520155,520152,520147,520149,
            520150,520148,520151,520153,520143,520144,520145,520146,520141,
            520142];
    elif hash1 eq 1741 and hash2 eq 2052 and hash3 eq 2861 then
        V:=[20,1,21,11,22,21,23,29,24,37,25,43,26,49];
        D:=[1,62/3,1,23,1,27,1,86/3,2,0,6,35/2,1,127/6,2,43/2,1,163/6,1,0,11,
            53/3,2,59/3,1,24,1,0,15,33/2,2,197/10,1,121/6,1,0,19,49/3,1,91/5,1,
            0,21,33/2,1,101/6,2,0,24,76/5,1,0];
        I:=[424657,424620,424652,424158,424246,424576,424619,424665,424663,
            423223,423963,424659,424656,424255,423887,424234,422760,422777,
            423927,422614,422779,423173,423256,422414];
    elif hash1 eq 143 and hash2 eq 381 and hash3 eq 2589 then
        V:=[21,1,22,11,24,15,25,23,27,29];
        D:=[1,69/2,2,357/10,1,215/6,1,393/10,1,0,6,86/3,1,0,7,44/3,1,46/3,1,
            50/3,2,0,11,97/6,1,115/6,1,0,13,85/6,1,481/30,1,97/6,2,113/6,1,0];
        I:=[542094,542159,542349,542163,541977,542092,542495,542164,542060,
            542494,542160,541796,542033,542122,542034,542043,541777];
    elif hash1 eq 3010 and hash2 eq 2131 and hash3 eq 806 then
        V:=[17,1,18,7,19,13,20,19,21,25,22,31,23,43,24,51,27,57,28,61];
        D:=[1,53/2,1,73/2,1,0,3,31,1,128/3,1,0,5,139/6,1,175/6,1,0,7,95/3,1,
            165/4,1,0,9,131/6,1,193/6,2,0,12,358/15,1,121/5,1,388/15,1,79/3,1,
            80/3,1,0,17,241/12,1,125/6,1,247/10,1,0,20,82/5,2,278/15,1,0,23,
            407/30,2,0,25,82/7,1,0];
        I:=[516007,516001,516004,516143,516130,516070,516149,516189,516170,
            516080,516178,516275,516273,515793,516035,516265,516156,516229,
            515874,515995,516294,516175,515671,516317,515766];
    elif hash1 eq 1835 and hash2 eq 2655 and hash3 eq 908 then
        V:=[18,1,19,11,20,19,21,27,22,33,23,41];
        D:=[1,30,1,34,1,36,1,40,1,0,5,49/2,2,57/2,1,65/2,1,0,9,68/3,1,80/3,1,29,
            1,0,12,151/6,2,51/2,1,0,15,59/3,1,64/3,1,82/3,1,0,18,71/4,1,0];
        I:=[420086,420107,420039,420093,420030,420078,420082,420181,420028,
            420001,420197,419926,420212,420121,420199,420193,419164,420054];
    elif hash1 eq 1083 and hash2 eq 338 and hash3 eq 131 then
        V:=[22,1,23,9,25,15,26,21,27,29,28,33];
        D:=[1,29,1,98/3,1,33,1,0,4,91/4,1,63/2,2,0,7,97/6,1,121/6,1,0,9,191/12,
            1,49/3,1,61/3,1,0,12,169/12,1,0,13,44/3,1,247/15,1,199/12,1,0];
        I:=[411705,411659,411655,411187,410380,411493,410579,410776,411504,
            411232,410727,410555,411684,409450,409410];
    elif hash1 eq 3386 and hash2 eq 436 and hash3 eq 2481 then
        V:=[29,1,30,7,31,17,32,21,33,27];
        D:=[1,87/4,1,131/6,1,0,3,73/4,1,55/3,1,56/3,2,239/12,1,0,8,101/6,5,0,13,
            179/12,2,15,4,0,19,461/30,1,0];
        I:=[663255,663965,663283,663690,663249,663250,663908,663609,663622,
            663856,663866,663970,663254,663595,663245,663246,663689,663966,
            663127];
    elif hash1 eq 942 and hash2 eq 2158 and hash3 eq 1411 then
        V:=[13,1,14,7,15,15,16,21,17,35,18,43,19,63,20,79,21,89,22,99,25,109];
        D:=[1,77/2,1,89/2,1,0,3,98/3,1,110/3,2,37,1,0,7,211/6,1,223/6,2,0,10,
            89/3,1,95/3,1,186/5,1,113/3,2,119/3,2,173/4,1,0,18,179/6,1,317/10,1,
            191/6,4,0,24,24,1,74/3,1,26,1,131/5,1,79/3,1,359/12,1,383/12,1,32,1,
            98/3,1,0,33,24,1,289/12,1,731/30,1,247/10,1,28,1,337/12,1,385/12,1,
            0,40,133/6,1,271/12,1,362/15,1,157/6,1,0,44,1117/60,1,571/30,1,
            229/12,1,581/30,1,0,48,523/28,1,113/6,1,286/15,1,20,1,0,52,737/60,2,
            0];
        I:=[429874,429858,429879,429792,429900,429917,429878,429848,429919,
            429914,429903,429666,429668,429712,429708,429790,429639,429845,
            429702,429683,429765,429901,429921,429889,429838,429884,429705,
            429883,429713,429810,429766,429667,429751,429754,429669,429671,
            429715,429633,429592,429637,429680,429616,429582,429687,429678,
            429752,429672,429590,429588,429690,429696,429642,429682];
    elif hash1 eq 1741 and hash2 eq 3049 and hash3 eq 1021 then
        V:=[17,1,18,5,20,9,21,15,22,25,23,29,24,35,25,39,26,43,27,47,28,51,29,
            55];
        D:=[1,271/6,1,0,2,136/3,1,0,3,508/15,1,211/5,1,0,5,271/8,1,1031/30,1,
            457/12,1,79/2,1,0,9,632/21,1,0,10,1033/42,1,313/12,1,0,12,505/24,1,
            0,13,183/10,1,0,14,1816/105,1,0,15,1889/120,1,0,16,171/14,1,0,17,
            110/9,1,0];
        I:=[518482,518515,518384,518473,518195,518194,518294,518293,518297,
            518306,518305,518196,518284,518283,518198,518289,518187];
    elif hash1 eq 378 and hash2 eq 6 and hash3 eq 898 then
        V:=[17,1,18,5,19,15,20,23,21,35,22,43,23,53,24,59,25,67,27,73];
        D:=[1,61/2,1,0,2,74/3,1,80/3,1,98/3,1,110/3,1,0,6,175/6,1,63/2,1,235/6,
            2,0,10,71/3,1,136/5,1,125/4,1,95/3,2,112/3,1,0,16,237/10,1,143/6,1,
            193/6,1,0,19,67/3,1,121/5,1,79/3,1,359/12,1,0,23,551/30,1,113/6,1,0,
            25,278/15,1,283/15,1,247/12,1,0,28,1093/60,1,679/30,1,0,30,569/42,1,
            76/5,1,0];
        I:=[516008,516009,516092,516052,516270,516099,516267,515982,516346,
            516091,516053,516100,516123,516286,516331,516269,516083,516323,
            516057,516125,515830,516337,516122,516282,516108,516215,516040,
            515864,515776,515725,515755];
    elif hash1 eq 2916 and hash2 eq 581 and hash3 eq 741 then
        V:=[22,1,24,5,25,17,26,25,29,35];
        D:=[1,74/3,1,0,2,59/3,1,85/4,1,65/3,1,77/3,1,26,1,0,7,107/6,1,237/10,1,
            95/4,1,0,10,99/5,1,239/12,1,61/3,2,62/3,1,0,15,129/10,1,27/2,1,
            451/30,1,461/30,1,0];
        I:=[415713,415684,415910,415714,415779,415736,415762,415738,415631,
            415725,415819,415304,415632,415728,415295,415326,415309,415297];
    elif hash1 eq 566 and hash2 eq 333 and hash3 eq 3079 then
        V:=[18,1,19,5,20,13,21,17,23,27];
        D:=[1,34,1,0,2,49/2,1,57/2,1,65/2,3,0,7,80/3,1,0,8,127/6,2,43/2,1,139/6,
            1,151/6,6,0,18,257/12,1,131/6,1,133/6,1,0];
        I:=[253625,253650,253633,253631,253637,253646,253640,253549,253638,
            253644,253658,253630,253632,253636,253645,253649,253652,253562,
            253558,253563];
    elif hash1 eq 2117 and hash2 eq 195 and hash3 eq 1434 then
        V:=[22,1,23,5,24,11,25,19,26,29,27,33,28,39,30,47];
        D:=[1,33,1,0,2,151/6,1,163/6,1,0,4,65/3,1,71/3,2,24,1,0,8,197/10,1,
            119/6,1,121/6,3,41/2,1,0,14,55/3,1,0,15,101/6,1,221/12,1,0,17,
            175/12,1,179/12,1,15,2,0,21,173/15,1,0];
        I:=[249485,249254,248719,248346,248325,248408,249461,248397,248088,
            248185,249250,249445,249460,248476,248184,247694,248183,248177,
            249301,249459,248188];
    elif hash1 eq 2634 and hash2 eq 518 and hash3 eq 2755 then
        V:=[17,1,18,5,19,9,20,13,21,23,22,39,25,49,26,59,28,65];
        D:=[1,81/2,1,0,2,39,1,0,3,223/6,2,0,5,88/3,1,89/3,1,94/3,1,101/3,1,0,9,
            305/12,1,277/10,1,167/6,1,179/6,1,127/4,1,191/6,4,193/6,1,0,19,
            358/15,1,388/15,1,311/12,2,79/3,1,0,24,379/20,1,19,2,229/12,1,39/2,
            1,0,29,143/10,1,91/6,1,0,31,176/15,1,0];
        I:=[425382,425170,425071,425248,424979,424977,425037,424945,424950,
            424961,425163,424885,424982,424980,425246,425252,425253,425065,
            424970,424884,424883,425062,424882,424919,424877,424917,424921,
            425063,425067,425012,424958];
    elif hash1 eq 49 and hash2 eq 1531 and hash3 eq 2617 then
        V:=[22,1,23,9,24,13,25,21,26,29,27,35,28,39,30,43];
        D:=[1,33,1,35,1,37,1,0,4,123/4,1,0,5,101/4,1,88/3,1,89/3,1,0,8,103/4,1,
            157/6,1,277/10,1,0,11,268/15,1,51/2,1,0,13,197/12,1,0,14,227/12,2,0,
            16,787/60,1,66/5,1,0];
        I:=[411476,411160,411256,411285,411477,411386,411153,411467,411470,
            411489,411027,409433,411301,411377,411399,411309,411028];
    elif hash1 eq 1788 and hash2 eq 2843 and hash3 eq 1038 then
        V:=[22,1,23,9,24,15,25,23,26,33,27,37,28,43,30,47];
        D:=[1,20,1,71/3,1,101/4,1,0,4,121/6,1,273/10,1,0,6,56/3,1,298/15,1,22,1,
            0,9,167/10,1,551/30,1,115/6,1,283/14,1,0,13,56/3,1,0,14,149/10,1,
            1117/60,1,0,16,203/15,1,0,17,253/21,1,0];
        I:=[510157,510138,510185,510183,510144,509847,510165,509182,510153,
            510143,510176,508511,509185,508885,508508,509817,508303];
    elif hash1 eq 284 and hash2 eq 45 and hash3 eq 2651 then
        V:=[27,1,28,7,29,15,30,25,31,31,32,41];
        D:=[1,35/2,1,115/6,2,0,4,46/3,1,47/3,1,16,1,0,7,85/6,2,157/10,1,97/6,1,
            177/10,1,0,12,208/15,1,167/12,1,0,14,149/12,1,25/2,1,127/10,1,77/6,
            1,0,18,45/4,1,0];
        I:=[234570,234545,234751,234446,234328,234720,231855,232007,232288,
            234260,232767,233061,231444,232708,231018,230792,232623,233456];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 3 mod 47.
function cached_ids_3(P,hash1,hash2,hash3)
    if hash1 eq 1037 and hash2 eq 1782 and hash3 eq 1082 then
        V:=[21,1,22,9,23,19,24,27,26,31,27,37];
        D:=[1,49/2,1,53/2,2,57/2,1,0,5,23,1,25,1,80/3,1,29,1,0,9,75/4,1,115/6,1,
            127/6,2,0,13,53/3,1,0,14,16,1,49/3,1,0,16,169/12,1,0];
        I:=[503571,503555,503685,503569,503668,503409,503572,503542,503556,
            503407,502815,503574,503531,503274,503580,503567];
    elif hash1 eq 426 and hash2 eq 1010 and hash3 eq 800 then
        V:=[24,1,25,5,26,11,27,15,28,25,29,33,30,39,31,43];
        D:=[1,125/4,1,0,2,107/6,1,257/12,1,0,4,67/3,1,0,5,221/12,1,245/12,1,
            41/2,1,125/6,1,0,9,74/5,1,199/12,1,18,1,0,12,59/4,1,75/4,1,0,14,
            261/20,1,0,15,499/42,1,0];
        I:=[415753,417627,417248,416326,417249,416036,417246,416419,417405,
            417617,415032,417255,415413,417383,415423];
    elif hash1 eq 3434 and hash2 eq 2989 and hash3 eq 1568 then
        V:=[21,1,22,15,23,29,24,37,25,51,26,57,27,65,28,69];
        D:=[1,139/6,1,51/2,1,59/2,1,187/6,1,147/4,1,223/6,1,0,7,64/3,1,77/3,1,
            156/5,1,95/3,1,32,1,34,1,0,13,257/12,1,103/4,1,169/6,1,0,16,298/15,
            1,101/5,1,81/4,1,61/3,1,119/5,1,287/12,1,0,22,359/20,1,125/6,1,0,24,
            15,1,17,1,56/3,1,0,27,83/6,1,0,28,316/21,1,227/15,2,0];
        I:=[510595,510678,510689,510586,510559,510675,510684,510671,509606,
            510670,510676,510677,510571,510213,510239,509548,510224,510687,
            510682,509441,508906,510223,510648,510679,509437,509275,509442,
            508897,508899,508900];
    elif hash1 eq 2964 and hash2 eq 1430 and hash3 eq 2943 then
        V:=[15,1,16,5,17,11,18,15,19,19,21,25,23,37,24,43,26,49];
        D:=[1,105/2,1,0,2,110/3,1,45,1,0,4,271/6,1,0,5,196/5,1,0,6,317/10,1,
            227/6,1,0,8,289/12,1,731/30,3,49/2,2,149/6,1,361/12,1,0,16,631/30,1,
            1333/60,1,0,18,334/15,1,1363/60,1,0,20,157/10,1,0];
        I:=[428814,428776,428677,428823,428732,428695,428821,428799,428665,
            428670,428687,428668,428688,428693,428719,428710,428644,428603,
            428595,428608];
    elif hash1 eq 1178 and hash2 eq 1323 and hash3 eq 2200 then
        V:=[21,1,22,5,23,11,24,19,25,29,26,33,27,39,28,43,29,49,30,57];
        D:=[1,199/6,1,0,2,89/3,1,156/5,1,0,4,329/12,1,317/10,1,191/6,3,0,9,
            263/12,1,358/15,2,388/15,1,104/3,1,0,14,161/6,1,0,15,223/12,1,112/5,
            1,0,17,189/10,1,0,18,238/15,2,69/4,1,0,21,361/30,2,137/10,2,157/10,
            1,0,26,176/15,2,707/60,1,0];
        I:=[509826,509871,510046,509870,510039,509595,510035,510190,508818,
            509725,510043,509872,509491,509471,508817,509351,509367,508776,
            508785,508520,509683,509772,509709,510010,508782,509726,509733,
            509771];
    elif hash1 eq 1789 and hash2 eq 456 and hash3 eq 2336 then
        V:=[23,1,24,11,25,21,26,29,27,35,28,41,29,49];
        D:=[1,103/4,1,179/6,1,377/12,1,203/6,1,0,5,67/3,1,388/15,1,28,1,121/4,1,
            0,9,221/12,1,227/10,1,57/2,1,0,12,272/15,1,227/12,1,0,14,511/30,1,
            253/12,1,0,16,233/15,1,257/15,1,287/15,1,0,19,797/60,1,55/4,2,
            373/24,1,467/30,1,1271/70,1,0];
        I:=[510015,509116,510221,509118,509141,510069,509700,510007,509462,
            509448,508404,509938,510101,510006,509520,510063,509958,508730,
            509460,509535,510215,508408,508382,508308];
    elif hash1 eq 896 and hash2 eq 2812 and hash3 eq 2265 then
        V:=[20,1,22,5,23,11,24,21,25,29,26,41,27,45,30,49];
        D:=[1,104/3,1,0,2,89/3,1,95/3,1,0,4,143/6,1,155/6,1,191/6,1,135/4,1,0,8,
            111/5,2,358/15,1,79/3,1,0,12,241/12,1,245/12,1,41/2,1,265/12,2,
            311/14,1,0,18,278/15,1,0,19,559/30,1,0,20,253/21,1,0];
        I:=[420397,419507,420315,420316,420447,419776,420393,419390,419424,
            419841,420422,420194,420195,419735,420270,420285,419010,419548,
            419175,418973];
    elif hash1 eq 2259 and hash2 eq 631 and hash3 eq 1308 then
        V:=[23,1,25,5,26,15,27,23,28,29,29,35];
        D:=[1,47/2,1,0,2,95/4,1,143/6,1,155/6,1,179/6,1,0,6,328/15,1,97/4,1,
            388/15,1,0,9,75/4,1,41/2,1,0,11,386/21,1,106/5,1,0,13,509/30,1,35/2,
            1,569/30,1,0];
        I:=[502159,502634,502486,502468,502441,502072,502654,502074,502516,
            501163,501402,502397,502228,501841,501790];
    elif hash1 eq 3011 and hash2 eq 2383 and hash3 eq 961 then
        V:=[15,1,18,5,19,13,20,23,21,33,24,41,25,53,27,63];
        D:=[1,93/2,1,0,2,149/4,1,113/3,1,118/3,1,0,5,377/12,1,337/10,1,205/6,1,
            377/10,1,0,9,359/12,1,32,1,129/4,1,97/3,1,0,13,731/30,1,26,1,791/30,
            1,0,16,381/20,1,1147/60,1,115/6,1,293/15,1,235/12,1,0,21,1031/70,1,
            443/30,1,74/5,1,893/60,1,0,25,709/60,1,0];
        I:=[428992,428764,428818,428728,428765,428751,428750,428812,428626,
            428950,428790,428806,428802,428722,428627,428589,428641,428610,
            428810,428794,428596,428749,428700,428708,428773];
    elif hash1 eq 2776 and hash2 eq 1840 and hash3 eq 2894 then
        V:=[20,1,21,5,22,13,23,21,24,35,25,43,26,49,27,69,29,73];
        D:=[1,104/3,1,0,2,151/6,1,175/6,1,199/6,1,0,5,83/3,2,89/3,1,95/3,1,0,9,
            87/4,1,281/12,1,143/6,1,257/10,1,155/6,1,157/6,1,0,15,298/15,1,
            101/5,1,311/12,1,0,18,217/12,1,221/12,1,0,20,248/15,1,253/15,1,17,1,
            223/12,1,283/15,1,227/12,1,19,3,96/5,1,58/3,3,0,33,56/3,1,0,34,47/4,
            1,0];
        I:=[252900,253052,252656,253204,252600,253220,252531,253077,252661,
            252627,252641,252561,252920,252532,252609,252655,252490,252663,
            252516,252666,252654,252591,252555,252539,252541,252521,252543,
            252669,252558,252544,252559,252684,252484,252667];
    elif hash1 eq 2071 and hash2 eq 3107 and hash3 eq 550 then
        V:=[22,1,23,17,24,29,25,45,26,61,27,69];
        D:=[1,19,2,62/3,1,21,2,68/3,2,23,1,74/3,2,27,1,0,12,115/6,1,39/2,1,
            127/6,1,43/2,2,139/6,1,0,18,52/3,1,53/3,6,18,2,77/4,2,58/3,2,59/3,2,
            20,1,0,34,95/6,1,97/6,3,33/2,1,209/12,2,177/10,1,107/6,1,109/6,3,0,
            46,191/12,2,16,4,55/3,1,0,53,431/30,2,0];
        I:=[251500,251503,251545,251102,251239,251088,251430,250943,251233,
            251587,250422,251458,251472,251089,250587,251569,250449,251226,
            250744,251084,251343,251429,251442,251449,251238,251432,250135,
            250335,250166,251339,250101,251353,250718,250197,250127,250218,
            251376,250137,250333,250858,250387,251154,250441,250589,250811,
            250124,250850,249981,250115,251246,251455,249834,250355,250547];
    elif hash1 eq 1648 and hash2 eq 1202 and hash3 eq 1808 then
        V:=[24,1,25,7,26,13,27,25,28,29,29,33,30,41,31,47];
        D:=[1,77/3,1,89/3,1,0,3,133/6,1,157/6,1,0,5,215/12,1,61/3,1,62/3,1,
            328/15,2,67/3,1,0,11,217/12,1,0,12,46/3,2,0,14,451/30,1,157/10,1,
            339/20,2,0,18,787/60,1,163/12,1,0,20,361/30,1,0];
        I:=[417263,417299,417112,417160,417713,416787,417140,417006,417056,
            416382,416621,417040,417115,417031,416037,415864,416631,417335,
            417408,415378];
    elif hash1 eq 97 and hash2 eq 2751 and hash3 eq 143 then
        V:=[21,1,22,9,23,21,24,29,25,39,26,51,27,59];
        D:=[1,49/2,1,53/2,1,65/2,1,0,4,21,1,68/3,2,23,4,74/3,1,27,1,0,13,115/6,
            4,139/6,4,47/2,1,0,22,53/3,1,59/3,1,65/3,2,71/3,1,0,27,185/12,2,
            63/4,1,95/6,2,121/6,1,41/2,2,0,35,43/3,1,49/3,3,55/3,2,0,41,169/12,
            3,431/30,2,29/2,1,89/6,1,0];
        I:=[418409,417878,418449,418425,417909,417914,417928,418418,418428,
            418455,418437,418466,417874,418387,418411,418630,416719,417890,
            417913,418458,418535,417871,417920,416470,418539,415978,418479,
            418481,417892,418417,418424,417633,417891,418565,418488,417906,
            418314,418453,415975,417626,418366,418367,418487,416583,416742,
            417877,416778];
    elif hash1 eq 1272 and hash2 eq 640 and hash3 eq 1409 then
        V:=[25,1,26,9,27,19,28,23,29,29,30,37,31,41];
        D:=[1,353/12,1,119/4,1,179/6,1,0,4,287/12,1,26,1,335/12,1,91/3,1,0,8,
            731/30,1,0,9,313/15,1,338/15,1,0,11,521/30,1,35/2,1,445/24,1,0,14,
            257/15,1,0,15,91/6,1,467/30,1,0];
        I:=[416707,415748,416373,417769,417458,416704,417251,417554,415858,
            414956,416918,414917,414977,415355,415349,415850];
    elif hash1 eq 426 and hash2 eq 931 and hash3 eq 2704 then
        V:=[13,1,14,7,15,15,16,25,17,33,18,49,19,71,20,81,21,97,22,117,24,133,
            25,143];
        D:=[1,73/2,1,85/2,1,0,3,116/3,1,43,1,140/3,1,0,6,131/4,1,75/2,1,235/6,1,
            79/2,1,0,10,89/3,1,95/3,1,101/3,1,0,13,119/4,1,179/6,3,127/4,1,
            191/6,5,193/6,1,65/2,1,203/6,2,0,27,388/15,1,311/12,1,26,2,79/3,1,
            80/3,1,121/4,1,383/12,1,161/5,1,97/3,1,431/12,1,0,38,149/6,1,26,1,
            313/12,2,53/2,1,0,43,133/6,1,45/2,1,338/15,1,271/12,2,275/12,1,
            548/21,1,319/12,1,0,51,571/30,1,229/12,1,581/30,1,233/12,1,39/2,4,
            119/6,2,1333/60,1,89/4,1,47/2,1,0,64,1123/60,1,97/5,2,293/15,4,
            298/15,2,113/5,1,1363/60,1,461/20,1,0,76,1313/84,1,157/10,1,947/60,
            1,377/20,1,0,80,737/60,1,0];
        I:=[430329,430314,430324,430334,430297,430254,430348,430323,430328,
            430288,430337,430289,430228,430266,430292,430312,430252,430301,
            430305,430333,430338,430341,430339,430344,430302,430309,430189,
            430256,430308,430317,430185,430325,430188,430224,430180,430195,
            430203,430196,430268,430216,430257,430186,430219,430184,430163,
            430187,430206,430202,430128,430129,430177,430199,430193,430250,
            430174,430233,430279,430294,430304,430330,430142,430151,430105,
            430111,430205,430208,430159,430160,430171,430179,430204,430210,
            430106,430120,430118,430121,430127,430125,430104,430166];
    elif hash1 eq 191 and hash2 eq 3441 and hash3 eq 2622 then
        V:=[23,1,24,9,25,17,28,21,29,31];
        D:=[1,139/6,1,151/6,1,51/2,1,0,4,65/3,2,70/3,2,76/3,1,0,9,45/2,2,0,11,
            223/15,2,15,3,76/5,1,46/3,2,0,19,31/2,1,0];
        I:=[249752,249761,249728,249486,249714,249650,249651,249484,248798,
            248812,248564,248591,248547,248566,249743,248096,248046,248097,
            247588];
    elif hash1 eq 473 and hash2 eq 599 and hash3 eq 2907 then
        V:=[20,1,21,5,22,11,23,23,24,29,26,37,27,41];
        D:=[1,23,1,0,2,151/6,1,51/2,2,0,5,93/4,1,70/3,1,71/3,1,101/4,1,83/3,1,0,
            10,131/6,1,143/6,1,0,12,215/12,1,322/15,1,263/12,1,0,15,15,3,0,18,
            299/20,1,0];
        I:=[422214,422208,422191,422281,422137,422309,421423,422148,422299,
            421586,422049,422077,421502,421266,422205,422212,422294,421188];
    elif hash1 eq 3340 and hash2 eq 3296 and hash3 eq 147 then
        V:=[16,1,17,13,18,33,19,55,20,87,21,107,22,137,23,167,24,205,25,219,26,
            231];
        D:=[1,98/3,1,35,1,116/3,1,39,1,41,1,0,6,151/6,1,115/4,1,59/2,1,187/6,2,
            67/2,1,211/6,3,71/2,1,223/6,1,235/6,1,0,18,77/3,2,83/3,2,89/3,1,30,
            1,125/4,1,94/3,3,95/3,2,100/3,1,176/5,1,107/3,1,0,33,87/4,1,131/6,1,
            133/6,1,145/6,1,257/10,1,155/6,1,157/6,2,53/2,2,167/6,1,169/6,1,
            179/6,1,317/10,1,127/4,2,193/6,2,143/4,1,0,52,62/3,1,263/12,1,68/3,
            1,73/3,1,74/3,1,388/15,1,311/12,2,131/5,1,79/3,7,0,68,221/12,1,37/2,
            1,113/6,1,115/6,1,599/30,1,241/12,3,41/2,1,125/6,3,439/20,1,22,1,
            671/30,2,269/12,2,363/14,1,317/12,1,0,88,203/12,1,17,1,278/15,1,
            223/12,1,283/15,1,227/12,2,19,1,58/3,7,401/20,1,313/15,4,441/20,1,
            464/21,1,133/6,1,23,1,0,112,31/2,1,67/4,1,205/12,1,181/10,1,1093/60,
            2,73/4,1,559/30,1,189/10,2,571/30,3,115/6,2,581/30,4,233/12,2,39/2,
            2,197/10,1,79/4,1,1333/60,1,679/30,1,689/30,1,0,140,91/6,1,76/5,2,
            187/12,1,101/6,1,73/4,1,75/4,1,0,147,433/28,1,937/60,1,47/3,1,63/4,
            1,94/5,1,0,152,727/60,2,0];
        I:=[429511,429492,429527,429474,429429,429478,429384,429427,429404,
            429515,429473,429512,429514,429532,429350,429468,429560,429499,
            429502,429397,429476,429405,429374,429376,429396,429510,429522,
            429395,429537,429551,429211,429208,429453,429421,429477,429380,
            429277,429414,429311,429493,429378,429401,429505,429400,429164,
            429324,429171,429407,429357,429402,429289,429438,429352,429467,
            429440,429250,429266,429389,429431,429231,429143,429220,429268,
            429308,429342,429385,429442,429262,429469,429498,429430,429292,
            429304,429434,429443,429525,429471,429533,429546,429166,429172,
            429221,429334,429302,429332,429104,429113,429294,429299,429348,
            429330,429343,429300,429370,429534,429159,429213,429354,429444,
            429448,429463,429518,429295,429263,429264,429291,429336,429331,
            429137,429275,429088,429339,429298,429450,429335,429284,429297,
            429301,429126,429267,429321,429274,429318,429346,429142,429175,
            429155,429296,429315,429319,429149,429168,429145,429158,429252,
            429150,429092,429131,429122,429329,429325,429347,429293,429103,
            429176,429091,429129,429133,429141,429139,429085,429287,429314];
    elif hash1 eq 2212 and hash2 eq 3042 and hash3 eq 2180 then
        V:=[25,1,26,5,27,15,28,23,29,29,30,33,31,39,32,43,33,49];
        D:=[1,199/6,1,0,2,89/3,1,156/5,1,95/3,1,34,1,0,6,217/10,1,305/12,1,
            203/6,1,0,9,67/3,1,74/3,1,0,11,611/30,1,0,12,44/3,1,15,1,0,14,59/4,
            1,0,15,203/15,1,233/15,1,0,17,41/3,1,0];
        I:=[396657,396414,396340,396718,395970,396063,396023,396781,396341,
            396625,396508,396645,396336,395612,395613,395615,396753];
    elif hash1 eq 614 and hash2 eq 2508 and hash3 eq 1634 then
        V:=[29,1,30,7,31,17,32,27];
        D:=[1,115/6,1,39/2,2,0,4,47/3,1,53/3,3,18,1,58/3,1,0,10,157/10,1,95/6,2,
            97/6,1,209/12,2,0,16,71/5,1,43/3,3,0];
        I:=[59571,59458,59562,59428,58784,59453,59694,59536,58987,58968,59649,
            59700,59054,58682,58766,58986,58770,58891,59134];
    elif hash1 eq 2118 and hash2 eq 2261 and hash3 eq 84 then
        V:=[25,1,26,5,27,13,28,21,29,25,30,33];
        D:=[1,139/6,1,0,2,59/3,2,20,1,64/3,1,0,6,97/6,2,33/2,1,109/6,2,0,11,
            50/3,2,0,13,173/12,1,29/2,1,89/6,1,0,16,13,3,0];
        I:=[670664,670509,670920,670595,670514,670370,670940,670948,670918,
            670968,669988,670301,670494,670572,670351,670260,670294,670295];
    elif hash1 eq 567 and hash2 eq 313 and hash3 eq 1910 then
        V:=[22,1,23,9,24,19,25,29,26,35,27,41];
        D:=[1,80/3,1,27,1,86/3,1,0,4,139/6,1,47/2,1,51/2,2,175/6,2,0,10,85/4,1,
            116/5,1,71/3,1,26,1,0,14,119/6,2,257/12,1,0,17,91/5,1,61/3,1,0,19,
            29/2,1,197/12,1,0];
        I:=[250993,251253,251127,251165,251189,250652,251148,250408,251248,
            250520,250440,250564,251235,250988,251206,249957,250450,250554,
            250666,250585];
    elif hash1 eq 1178 and hash2 eq 581 and hash3 eq 2040 then
        V:=[29,1,30,9,31,15,33,23];
        D:=[1,233/12,1,79/4,1,119/6,1,0,4,55/3,1,298/15,1,0,6,167/10,1,101/6,1,
            551/30,1,0,9,379/30,1,389/30,1,157/12,5,0];
        I:=[665640,665145,666078,666225,665156,666283,665979,664946,665793,
            665589,664998,665000,665602,665959,666136];
    elif hash1 eq 144 and hash2 eq 976 and hash3 eq 260 then
        V:=[17,1,18,5,19,13,20,29,21,47,22,67,23,81,24,89,25,97,26,103];
        D:=[1,53/2,1,0,2,25,1,80/3,1,92/3,1,0,5,139/6,1,47/2,2,151/6,1,51/2,1,
            175/6,1,131/4,2,199/6,1,0,14,20,1,65/3,1,22,1,126/5,1,101/4,2,77/3,
            2,146/5,1,89/3,1,0,24,109/6,2,233/12,1,197/10,1,119/6,4,217/10,2,
            131/6,2,133/6,1,273/10,1,169/6,1,0,39,268/15,1,91/5,1,239/12,1,62/3,
            1,263/12,1,111/5,1,0,45,101/6,1,255/14,1,551/30,1,0,48,179/12,1,15,
            1,86/5,1,0,51,451/30,1,461/28,1,0,53,281/21,1,269/20,2,203/15,1,
            233/15,1,0];
        I:=[515093,515269,515286,515332,515020,515192,515331,514630,514788,
            515288,515114,515262,515258,515179,515276,514732,514771,515293,
            515314,515275,515373,514842,514983,514510,515259,515285,514524,
            514835,515034,515140,515180,514095,514863,514838,515273,515260,
            514231,514657,514751,514877,514839,515375,515232,514878,514876,
            514020,514180,515313,514861,514841,513987,514168,514069,514107,
            514577,514067,514078];
    elif hash1 eq 2071 and hash2 eq 1693 and hash3 eq 343 then
        V:=[25,1,26,7,27,19,28,27,30,35];
        D:=[1,163/6,1,175/6,1,0,3,65/3,1,70/3,1,71/3,1,83/3,1,89/3,1,0,8,41/2,1,
            131/6,1,133/6,1,0,11,55/3,2,61/3,1,62/3,1,0,15,52/3,1,0];
        I:=[238183,238174,238066,238298,238099,238225,238048,238184,238094,
            238173,238122,238222,238095,238224,238078];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 4 mod 47.
function cached_ids_4(P,hash1,hash2,hash3)
    if hash1 eq 1132 and hash2 eq 3296 and hash3 eq 3250 then
        V:=[24,1,25,11,27,15,28,23,29,31,30,39,31,43];
        D:=[1,24,1,28,1,95/3,2,101/3,1,0,6,145/6,1,0,7,33/2,1,101/6,2,113/6,1,0,
            11,199/12,2,17,1,283/15,2,0,16,451/30,1,181/12,1,499/30,1,0,19,91/6,
            1,0,20,361/30,1,0];
        I:=[418294,418331,418670,418874,418853,418328,418868,416645,418875,
            418854,416662,417835,418813,416547,416642,416528,417773,415533,
            416190,417839];
    elif hash1 eq 3153 and hash2 eq 2347 and hash3 eq 3140 then
        V:=[21,1,22,7,23,13,24,21,26,27,27,33,28,37];
        D:=[1,115/4,1,67/2,1,0,3,22,1,26,2,0,6,197/10,2,119/6,1,257/10,1,0,10,
            101/5,1,111/5,1,0,12,46/3,1,82/5,1,0,14,157/12,1,0,15,66/5,1,0];
        I:=[512269,512560,512616,512029,512564,511106,511921,512030,512044,
            512106,512200,510921,512272,512742,511188];
    elif hash1 eq 803 and hash2 eq 2526 and hash3 eq 729 then
        V:=[26,1,27,13,28,19,29,25,30,33,31,37,32,43];
        D:=[1,52/3,1,85/4,1,70/3,1,71/3,1,24,1,0,6,79/4,1,119/6,1,0,8,73/4,1,
            55/3,1,0,10,479/30,1,33/2,2,539/30,1,0,14,175/12,1,0,15,391/30,1,
            27/2,3,0,19,239/21,1,139/12,1,0];
        I:=[243641,244219,243191,243893,243826,242683,243001,243475,243086,
            242380,242557,243530,241960,244520,242986,243434,243699,243822,
            242084,242023];
    elif hash1 eq 2072 and hash2 eq 2709 and hash3 eq 1286 then
        V:=[19,1,20,9,21,13,22,25,23,33,24,37,25,43,27,49,28,55];
        D:=[1,103/6,1,47/2,1,95/2,1,0,4,146/5,1,0,5,63/4,1,45/2,1,237/10,1,
            143/6,1,357/10,1,0,10,191/12,1,56/3,1,388/15,1,0,13,163/6,1,0,14,
            175/12,1,64/3,1,0,16,877/60,1,233/12,1,0,18,66/5,2,421/30,1,0,21,
            185/14,1,0];
        I:=[543198,542955,542793,543025,543190,542869,542954,542889,543070,
            543197,542956,543072,542794,543239,542855,543215,542832,543212,
            543236,542800,542734];
    elif hash1 eq 3059 and hash2 eq 645 and hash3 eq 1406 then
        V:=[25,1,26,7,27,15,28,23,29,27,31,33];
        D:=[1,51/2,1,59/2,1,0,3,101/4,1,77/3,3,83/3,1,0,8,131/6,1,143/6,1,145/6,
            1,0,11,61/3,1,0,12,221/12,2,75/4,1,0,15,181/12,1,91/6,1,0];
        I:=[671199,671664,671163,671198,671200,671249,671235,671236,671550,
            671201,671247,671122,671144,671165,671103,671102];
    elif hash1 eq 286 and hash2 eq 2611 and hash3 eq 3017 then
        V:=[16,1,17,15,18,25,19,35,20,39];
        D:=[1,26,1,28,2,30,2,32,4,34,1,36,1,0,12,53/2,3,57/2,2,61/2,1,69/2,2,0,
            20,23,1,25,1,80/3,2,27,1,0,25,163/6,1,0,26,59/3,1,0];
        I:=[61969,61968,61970,61967,61976,61971,61978,61979,254701,61982,61983,
            61977,254704,254708,61973,254703,254684,254660,254705,254694,254706,
            254659,254712,254664,254665,254663];
    elif hash1 eq 1038 and hash2 eq 2266 and hash3 eq 1518 then
        V:=[21,1,22,7,23,15,24,21,25,31,26,37,27,41,28,49,29,53];
        D:=[1,61/2,1,65/2,1,0,3,25,1,29,3,98/3,1,0,8,163/6,2,187/6,1,0,11,85/4,
            1,64/3,1,93/4,1,70/3,1,0,15,119/6,1,145/6,1,0,17,49/3,1,0,18,359/20,
            1,18,1,551/30,1,0,21,29/2,1,0,22,181/12,1,0];
        I:=[411824,411813,412031,411801,411870,412267,410603,410238,410642,
            411945,411800,411803,411815,412234,411820,410216,411775,410051,
            409616,409959,410958,409151];
    elif hash1 eq 1273 and hash2 eq 2410 and hash3 eq 959 then
        V:=[24,1,27,11,29,15,30,19,31,25];
        D:=[1,609/20,1,488/15,1,518/15,1,104/3,1,0,5,905/42,1,0,6,311/20,1,0,7,
            1647/140,2,361/30,1,0,10,211/18,1,2119/180,1,991/84,1,2011/168,1,
            289/24,1,5059/420,1,0];
        I:=[543154,542909,543125,543012,542758,542823,542759,542824,543126,
            542539,542540,542538,542583,542576,542581];
    elif hash1 eq 2213 and hash2 eq 1544 and hash3 eq 1254 then
        V:=[20,1,22,5,24,13,25,17,28,25,29,33,30,37];
        D:=[1,116/3,1,0,2,95/3,1,166/5,1,101/3,1,0,5,104/3,1,0,6,551/30,2,
            241/12,1,245/12,1,0,10,14,1,16,1,286/15,2,0,14,119/10,1,0,15,137/10,
            1,0];
        I:=[541346,541248,541232,541310,541296,541252,541350,540885,541280,
            541226,540933,540844,540899,541349,540625];
    elif hash1 eq 2730 and hash2 eq 267 and hash3 eq 609 then
        V:=[25,1,26,13,27,19,28,29,29,33,30,39,31,43];
        D:=[1,127/6,2,43/2,1,139/6,2,47/2,1,99/4,1,0,8,96/5,2,71/3,2,0,12,63/4,
            1,97/6,1,217/10,1,131/6,1,0,16,44/3,2,0,18,431/30,1,173/12,1,0,20,
            38/3,1,0,21,341/30,1,0];
        I:=[245215,245411,245286,244825,245083,244860,243596,243413,243415,
            243661,243796,243588,243411,242803,242798,242496,242537,242493,
            243695,242672,243153];
    elif hash1 eq 1602 and hash2 eq 313 and hash3 eq 2295 then
        V:=[24,1,25,5,26,13,27,21,28,29,31,37];
        D:=[1,74/3,1,0,2,127/6,2,43/2,1,139/6,3,0,8,58/3,1,59/3,2,106/5,1,0,12,
            177/10,1,107/6,1,109/6,1,0,15,43/3,1,81/5,1,49/3,3,0,20,23/2,1,0];
        I:=[671943,671337,671818,672015,671333,671416,671909,671995,671812,
            671919,671411,671393,671329,671338,671397,671428,671339,671477,
            671917,671332];
    elif hash1 eq 2166 and hash2 eq 2293 and hash3 eq 1240 then
        V:=[24,1,25,7,26,13,27,27,28,35,29,47,30,59,32,69];
        D:=[1,59/3,1,83/3,1,0,3,281/12,1,145/6,1,0,5,73/4,1,61/3,1,328/15,1,
            263/12,1,22,1,111/5,1,0,11,539/30,1,551/30,1,113/6,1,0,14,179/12,1,
            46/3,1,203/12,1,223/12,1,409/20,1,0,19,451/30,2,91/6,1,997/60,1,
            781/42,1,56/3,1,0,25,197/15,1,66/5,1,163/12,1,123/8,1,0,29,335/28,1,
            0];
        I:=[417684,417305,417630,417309,417171,417172,417057,417688,417544,
            416448,416371,417310,417232,417168,415840,417146,415939,416867,
            417218,417225,415838,415455,415315,415408,417028,417410,417207,
            414948,414920];
    elif hash1 eq 3153 and hash2 eq 2361 and hash3 eq 2490 then
        V:=[20,1,21,9,22,17,23,23,24,29,25,39];
        D:=[1,23,1,92/3,1,31,1,0,4,51/2,1,107/4,1,163/6,1,0,7,93/4,1,77/3,1,0,9,
            281/12,1,143/6,1,0,11,55/3,1,99/5,1,81/4,1,328/15,1,0,15,241/12,1,
            0];
        I:=[424605,424687,424704,424697,424110,424288,424093,423215,423027,
            424303,423990,423416,423417,422664,422626];
    elif hash1 eq 1508 and hash2 eq 2174 and hash3 eq 292 then
        V:=[22,1,23,7,24,11,25,21,26,27,27,31,28,37];
        D:=[1,25,1,29,1,0,3,163/6,2,0,5,59/3,1,64/3,1,65/3,2,83/3,1,0,10,197/10,
            1,45/2,1,0,12,91/5,1,0,13,91/6,1,491/30,1,0,15,46/3,1,253/15,1,0];
        I:=[250928,250764,250916,250923,250736,251243,250660,251172,250152,
            250653,250891,250185,250893,250546,250028,249966];
    elif hash1 eq 756 and hash2 eq 1968 and hash3 eq 1147 then
        V:=[26,1,27,5,28,13,29,17,30,27,31,33];
        D:=[1,101/4,1,0,2,257/12,1,131/6,1,133/6,1,0,5,73/4,1,0,6,173/12,1,29/2,
            1,89/6,1,20,1,0,10,175/12,1,44/3,1,0,12,259/20,1,391/30,4,79/6,1,
            625/42,1,0];
        I:=[408222,406706,406473,408215,408225,407005,408191,407170,405413,
            406762,405656,406705,406835,408093,408094,408140,408076,404664];
    elif hash1 eq 2777 and hash2 eq 1545 and hash3 eq 2001 then
        V:=[23,1,24,11,25,21,26,27,27,31];
        D:=[1,45/2,2,49/2,1,53/2,1,57/2,1,0,6,62/3,1,21,1,68/3,1,74/3,1,0,10,
            39/2,1,127/6,3,0,14,59/3,1,0,15,97/6,2,0];
        I:=[672184,672189,672083,672180,672038,672093,671860,672047,671206,
            672214,671214,671590,671617,671612,671342,671459];
    elif hash1 eq 2589 and hash2 eq 638 and hash3 eq 3350 then
        V:=[23,1,24,7,25,25,26,31,27,37,28,43,29,51,30,55];
        D:=[1,223/6,1,75/2,1,0,3,76/3,1,109/4,1,144/5,1,156/5,1,95/3,1,32,1,
            100/3,1,101/3,1,0,11,131/6,1,167/6,1,0,13,62/3,1,67/3,1,0,15,611/30,
            1,41/2,1,0,17,179/12,1,15,1,50/3,1,0,20,229/12,1,0,21,41/3,2,0];
        I:=[409671,411319,410134,410080,409535,409683,409672,410276,410927,
            410790,410302,410316,410279,409994,409999,410275,410906,409618,
            409132,409386,410285,410492];
    elif hash1 eq 1367 and hash2 eq 2167 and hash3 eq 1930 then
        V:=[21,1,22,13,23,21,24,27,25,51,26,63,27,75];
        D:=[1,43/2,2,47/2,2,51/2,3,163/6,1,123/4,1,0,10,22,1,93/4,1,136/5,1,0,
            13,197/10,1,121/6,1,0,15,191/12,1,16,1,49/3,1,50/3,1,91/5,1,73/4,1,
            81/4,1,61/3,1,62/3,1,22,1,89/4,2,0,27,89/6,1,91/6,1,33/2,1,551/30,1,
            113/6,1,0,32,179/12,1,15,3,46/3,1,117/7,1,203/12,1,0,39,401/30,1,
            161/12,1,27/2,1,83/6,1,625/42,1,0];
        I:=[513916,513917,513788,513789,513822,513828,513881,513874,513885,
            513824,513894,513704,513765,513884,513873,513872,513875,513830,
            513729,513531,513808,513878,513886,513508,513515,513572,513879,
            513751,513876,513764,513556,513699,513313,513817,513864,513829,
            513458,513408,513760,513732,513575,513645,513441];
    elif hash1 eq 2354 and hash2 eq 2905 and hash3 eq 665 then
        V:=[22,1,23,7,24,15,25,25,26,35,27,41];
        D:=[1,21,1,29,1,0,3,103/6,1,35/2,1,51/2,1,0,6,52/3,1,53/3,2,18,1,59/3,1,
            0,11,95/6,2,107/6,2,119/6,1,131/6,1,0,17,215/12,1,18,1,0,19,18,1,
            217/12,1,0];
        I:=[417934,416340,418839,418861,417622,418850,418293,418632,416540,
            417692,418829,418849,416416,418629,416684,415504,417608,417542,
            415968,416052];
    elif hash1 eq 3388 and hash2 eq 386 and hash3 eq 3238 then
        V:=[27,1,28,7,29,15,30,23];
        D:=[1,97/6,2,107/6,1,0,4,65/4,1,49/3,1,50/3,1,0,7,173/12,1,29/2,4,89/6,
            5,0,17,155/12,1,13,1,0];
        I:=[240748,241758,241362,238651,241572,239167,239141,240555,240577,
            241480,241767,240570,241157,241370,241477,241546,239132,238949];
    elif hash1 eq 756 and hash2 eq 119 and hash3 eq 1202 then
        V:=[23,1,24,13,25,29,26,41,27,49,29,53,30,61];
        D:=[1,55/2,1,175/6,1,187/6,1,139/4,1,211/6,2,0,7,64/3,1,82/3,1,89/3,1,
            30,1,156/5,1,95/3,2,101/3,1,0,15,237/10,1,95/4,1,305/12,1,277/10,1,
            167/6,1,0,20,39/2,1,239/12,1,328/15,1,0,23,221/12,1,0,24,451/30,1,
            331/20,1,50/3,1,0,27,163/12,1,0];
        I:=[411727,411730,411648,411173,411608,412301,410916,412156,411158,
            411492,411536,411557,412017,411664,411163,411415,411343,410118,
            411611,409532,410896,409466,410003,411457,409448,409283,411368];
    elif hash1 eq 2965 and hash2 eq 2396 and hash3 eq 233 then
        V:=[23,1,25,7,26,17,27,21,28,27,29,33,30,39];
        D:=[1,425/12,1,143/4,1,0,3,731/30,1,341/12,1,911/30,1,61/2,2,0,8,367/15,
            1,0,9,183/8,1,691/30,1,0,11,263/15,3,457/24,1,0,15,63/4,1,103/6,1,0,
            17,803/60,1,289/21,1,827/60,1,76/5,1,377/24,1,947/60,1,0];
        I:=[509626,509212,509952,510173,509894,509590,510172,508518,508358,
            508354,509934,510080,510140,508392,509820,508844,510124,508526,
            508764,508842,508365,508366];
    elif hash1 eq 1461 and hash2 eq 2253 and hash3 eq 2217 then
        V:=[23,1,24,13,25,23,26,29,30,41];
        D:=[1,43/2,1,47/2,1,51/2,1,55/2,1,131/4,1,0,6,16,1,69/4,1,53/3,1,126/5,
            1,0,10,197/10,1,145/6,1,0,12,167/12,1,238/15,1,16,1,101/5,1,81/4,1,
            0,17,61/5,1,0];
        I:=[505392,505320,505390,505436,506057,505138,505913,506592,505398,
            505312,504441,506607,505115,506530,506169,504485,504443];
    elif hash1 eq 3435 and hash2 eq 1912 and hash3 eq 1145 then
        V:=[27,1,28,7,29,17,30,27,31,35,32,41];
        D:=[1,133/6,1,143/6,1,0,3,49/3,1,56/3,1,20,1,61/3,1,0,7,103/6,1,221/12,
            2,187/10,1,113/6,1,0,12,44/3,1,223/15,1,46/3,1,0,15,27/2,2,181/12,1,
            0,18,12,1,0];
        I:=[237256,237390,237112,237433,237114,237162,237232,237068,237069,
            237399,237528,237106,237133,237435,237404,237408,236863,236850];
    elif hash1 eq 1273 and hash2 eq 351 and hash3 eq 539 then
        V:=[21,1,22,9,23,17,24,27,25,33];
        D:=[1,45/2,2,49/2,1,53/2,1,0,5,21,4,23,1,25,1,0,11,103/6,1,35/2,2,127/6,
            1,43/2,2,0,17,59/3,1,64/3,1,0,19,95/6,3,97/6,1,0];
        I:=[249639,249812,249643,249717,249279,249640,249718,249731,248756,
            248001,249550,249606,249776,248893,247944,249352,248089,248110,
            249528,249642,249656,248833];
    elif hash1 eq 427 and hash2 eq 2065 and hash3 eq 1799 then
        V:=[19,1,20,5,21,9,22,15,23,23,24,27,25,33,26,39,27,45,29,53];
        D:=[1,211/6,1,0,2,173/4,1,0,3,377/12,1,227/6,1,0,5,418/15,1,91/3,1,
            532/15,1,0,8,551/20,1,0,9,489/20,1,331/12,1,0,11,573/28,1,709/30,1,
            0,13,263/15,1,722/35,1,0,15,421/24,1,527/30,1,1411/70,1,0,18,
            1217/90,1,2873/210,1,0];
        I:=[542209,542207,542193,542223,542195,542203,542189,542074,542009,
            542073,541964,542098,542224,541965,541806,541822,541850,541792,
            541795];
    elif hash1 eq 3247 and hash2 eq 3205 and hash3 eq 2197 then
        V:=[25,1,26,5,27,15,28,27,29,39,30,43,31,51];
        D:=[1,115/6,1,0,2,65/3,1,22,1,93/4,1,70/3,1,0,6,177/10,1,107/6,1,121/6,
            2,217/10,1,131/6,2,0,13,49/3,1,89/5,1,268/15,1,215/12,1,91/5,1,0,18,
            91/6,1,0,19,103/7,1,223/15,3,15,1,0,24,391/30,1,0];
        I:=[243526,243384,243885,243104,243090,243857,244203,243371,243911,
            243812,243169,243620,242874,242373,243335,242985,242014,243814,
            242015,242013,243108,243295,241897,243263];
    elif hash1 eq 286 and hash2 eq 3398 and hash3 eq 2973 then
        V:=[24,1,25,5,26,11,27,19,29,25,30,33];
        D:=[1,101/4,1,0,2,87/4,1,155/6,2,0,5,328/15,1,111/5,1,85/3,1,0,8,539/30,
            1,221/12,2,0,11,451/30,1,461/30,1,499/30,1,0,14,163/12,1,61/4,1,0];
        I:=[416944,417143,415929,416883,417297,417499,414966,417017,416893,
            417582,417163,417201,416890,416438,415440];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 5 mod 47.
function cached_ids_5(P,hash1,hash2,hash3)
    if hash1 eq 2778 and hash2 eq 1 and hash3 eq 2903 then
        V:=[12,1,13,13,14,31,15,47,16,61,17,77,18,91,19,103,21,119];
        D:=[1,32,1,38,1,40,1,42,2,44,2,0,8,57/2,1,61/2,1,65/2,1,69/2,1,73/2,2,
            77/2,1,81/2,1,85/2,1,0,17,86/3,1,29,1,92/3,1,33,1,104/3,1,35,2,39,1,
            0,25,151/6,1,59/2,1,187/6,1,211/6,2,71/2,4,235/6,1,0,35,77/3,1,82/3,
            3,83/3,3,88/3,2,89/3,1,30,1,107/3,1,0,47,41/2,1,131/6,1,133/6,3,
            167/6,1,169/6,1,181/6,1,0,55,20,1,61/3,2,111/5,1,79/3,1,335/12,1,0,
            61,611/30,1,671/30,2,227/10,1,137/6,2,139/6,1,47/2,1,317/12,1,0,70,
            511/30,1,103/6,1,0];
        I:=[429953,429943,429954,429935,429952,429937,429939,429958,429941,
            429934,429944,429875,429936,429946,429866,429908,429828,429951,
            429923,429870,429911,429867,429904,429868,429912,429787,429905,
            429853,429891,429755,429803,429872,429890,429788,429770,429851,
            429880,429888,429747,429789,429862,429768,429772,429795,429798,
            429631,429726,429886,429780,429829,429833,429657,429645,429648,
            429812,429796,429893,429614,429710,429607,429698,429704,429707,
            429652,429655,429656,429703,429794,429608,429694,429895];
    elif hash1 eq 2026 and hash2 eq 2406 and hash3 eq 471 then
        V:=[20,1,21,7,22,17,23,35,24,41,25,51,27,61,28,69,29,77];
        D:=[1,28,1,88/3,1,0,3,63/4,1,133/6,1,357/10,1,215/6,1,0,7,191/12,1,16,1,
            166/7,1,388/15,1,26,1,442/15,1,98/3,1,104/3,1,0,15,247/10,2,161/6,1,
            0,18,175/12,1,46/3,1,47/3,1,159/7,1,0,22,877/60,3,44/3,2,793/42,1,
            45/2,1,0,29,66/5,2,161/12,1,917/60,1,0,33,1384/105,2,185/14,1,53/4,
            2,0,38,163/14,1,515/42,1,37/3,1,0];
        I:=[542198,542206,542256,542101,542197,542221,542236,542235,541951,
            542102,542211,542271,542196,542180,542096,542100,542172,542255,
            542108,542226,542008,542258,542261,542384,542259,542368,541946,
            541902,542260,542381,542373,541899,541872,542015,541873,541895,
            542014,541773,541790,541789];
    elif hash1 eq 992 and hash2 eq 1928 and hash3 eq 2111 then
        V:=[20,1,22,5,24,13,26,25,28,45];
        D:=[1,21,1,0,2,58/3,1,59/3,1,146/5,1,0,5,17,1,35/2,1,268/15,1,19,1,23,1,
            0,10,53/4,1,41/3,2,61/4,1,82/5,1,50/3,2,253/15,1,278/15,1,223/12,3,
            247/12,1,0,23,323/21,1,0];
        I:=[540303,540305,540172,540237,540325,540275,540174,540327,540295,
            540274,540329,540330,539711,540179,540149,540152,540261,540155,
            540058,540101,540162,540035,539922];
    elif hash1 eq 2120 and hash2 eq 1325 and hash3 eq 3459 then
        V:=[25,1,26,7,27,13,28,21,29,33];
        D:=[1,139/6,2,47/2,3,0,6,59/3,2,64/3,2,0,10,97/6,1,107/6,1,109/6,1,0,13,
            16,1,49/3,1,50/3,2,17,2,55/3,3,0,22,173/12,1,29/2,1,0];
        I:=[670454,670953,670553,670570,670666,670273,670667,670286,670380,
            670492,670520,670875,670668,670482,670466,670982,670124,670401,
            669907,670341,670359,670004,669949];
    elif hash1 eq 3248 and hash2 eq 65 and hash3 eq 211 then
        V:=[24,1,25,5,26,13,27,21,28,31,30,35];
        D:=[1,117/4,1,0,2,131/6,1,95/4,1,143/6,2,0,6,298/15,1,61/3,2,89/4,1,0,
            10,551/30,1,599/30,1,245/12,1,83/4,1,0,14,253/15,1,0,15,269/20,1,
            203/15,1,41/3,3,0];
        I:=[248471,248667,248250,248484,248666,248646,248520,248650,248253,
            248648,248461,247699,247680,248444,247528,248554,247962,248194,
            248281];
    elif hash1 eq 2355 and hash2 eq 1847 and hash3 eq 749 then
        V:=[21,1,22,13,23,27,24,37,25,49,26,57,27,61];
        D:=[1,45/2,1,49/2,3,53/2,4,57/2,3,61/2,1,0,13,21,1,23,1,74/3,1,25,5,27,
            3,86/3,1,0,25,83/4,1,127/6,2,43/2,2,139/6,2,0,32,69/4,1,52/3,1,58/3,
            1,59/3,4,65/3,2,0,41,209/12,2,107/6,5,109/6,1,0,49,49/3,1,0,50,
            431/30,1,173/12,1,89/6,1,0];
        I:=[674108,674037,674058,674095,674041,674060,674109,674113,674044,
            674110,674112,673973,674017,674023,673886,673925,674024,674027,
            674035,674042,673899,673979,673989,673837,673982,673897,674025,
            673993,674115,673814,673980,673965,673994,673967,673831,673969,
            673991,673995,673795,673812,673957,673997,673888,673954,673955,
            674032,674114,673916,673792,673839,673813,673840];
    elif hash1 eq 2167 and hash2 eq 641 and hash3 eq 972 then
        V:=[22,1,23,11,24,21,25,35,26,45];
        D:=[1,19,1,21,1,23,1,74/3,2,0,6,115/6,5,39/2,1,43/2,1,91/4,1,0,14,69/4,
            1,18,1,96/5,1,58/3,1,59/3,1,64/3,1,0,20,97/6,4,209/12,2,177/10,1,
            109/6,2,0,29,191/12,1,35/2,1,0];
        I:=[251549,251292,251099,251013,251454,250721,250773,250885,251369,
            251576,251185,251106,250606,251131,250839,250695,250219,251122,
            249903,250139,250225,250774,251347,250539,250845,250690,250849,
            251024,250134,250394];
    elif hash1 eq 3201 and hash2 eq 545 and hash3 eq 1983 then
        V:=[24,1,25,5,26,19,27,29,28,41,29,51,30,59];
        D:=[1,86/3,1,0,2,115/6,2,127/6,1,43/2,1,91/4,1,139/6,2,47/2,2,0,11,18,2,
            59/3,6,93/4,1,71/3,1,0,21,177/10,1,71/4,1,107/6,1,119/6,2,131/6,1,0,
            27,238/15,1,50/3,1,268/15,1,215/12,2,0,32,29/2,1,491/30,1,33/2,2,0,
            36,223/15,1,76/5,1,0];
        I:=[244741,244794,245262,244690,245054,244388,245088,245579,244783,
            244797,244795,245048,243659,244142,244740,245513,245576,245631,
            242170,244269,243534,242967,245516,244428,244673,242802,243550,
            242204,244481,242191,244436,245476,242289,242184,242199,241833,
            241835];
    elif hash1 eq 569 and hash2 eq 2792 and hash3 eq 1689 then
        V:=[18,1,19,9,20,25,21,43,22,67,23,79,24,91,25,107];
        D:=[1,27,1,98/3,1,33,1,0,4,127/6,1,43/2,2,91/4,1,151/6,1,163/6,2,115/4,
            1,63/2,1,0,13,53/3,1,65/3,1,93/4,1,70/3,1,101/4,1,76/3,1,109/4,1,28,
            1,0,21,107/6,2,37/2,1,119/6,3,121/6,1,217/10,1,87/4,1,131/6,1,133/6,
            1,281/12,1,237/10,1,145/6,2,0,36,50/3,2,61/3,1,109/5,1,263/12,1,22,
            1,0,42,33/2,3,35/2,1,551/30,1,37/2,1,307/14,1,0,49,46/3,1,329/20,2,
            248/15,1,199/12,2,253/15,1,203/12,1,52/3,4,0,61,149/10,1,461/30,1,
            185/12,1,0];
        I:=[517579,517589,517582,517546,517535,517536,517446,517592,517580,
            517583,517510,517373,517537,517570,517427,517581,517509,517576,
            517117,517464,517534,517554,517103,517105,517248,517456,517585,
            517238,517442,517549,517551,517519,517282,517383,517470,517223,
            517395,517140,517203,517233,517107,517376,517471,517552,517118,
            517113,517106,517047,517231,517136,517339,517345,517143,517244,
            517334,517351,517104,517112,517116,517119,517069,517059,517033];
    elif hash1 eq 616 and hash2 eq 2040 and hash3 eq 1766 then
        V:=[24,1,25,5,26,13,27,21,28,27,29,33,30,37,31,41];
        D:=[1,86/3,1,0,2,51/2,1,187/6,1,199/6,1,0,5,22,1,126/5,1,82/3,1,0,8,
            177/10,1,145/6,1,0,10,101/5,1,62/3,1,0,12,173/12,1,0,13,19,1,0,14,
            391/30,1,27/2,1,0];
        I:=[398864,398395,398686,398396,398139,396843,398924,398904,397926,
            396891,396942,398357,396802,398907,399062];
    elif hash1 eq 1509 and hash2 eq 408 and hash3 eq 2756 then
        V:=[14,1,15,7,16,15,17,29,18,43,19,61,20,75,21,91,22,95];
        D:=[1,32,1,36,1,0,3,57/2,1,65/2,3,69/2,1,0,8,80/3,1,29,3,92/3,1,31,2,33,
            2,35,2,0,19,47/2,1,51/2,2,163/6,3,55/2,1,175/6,1,59/2,2,0,29,65/3,2,
            22,1,70/3,1,71/3,1,76/3,1,77/3,2,83/3,1,89/3,1,0,39,121/6,1,131/6,2,
            133/6,2,45/2,1,155/6,2,157/6,2,0,49,101/5,1,61/3,1,111/5,2,67/3,1,
            68/3,2,23,1,311/12,1,0,58,113/6,1,0,59,253/15,1,17,3,0];
        I:=[255339,255334,255333,255304,255327,255346,255324,255214,255259,
            255284,255331,255251,255329,255332,255263,255330,255128,255272,
            255325,255141,255299,255260,255296,255298,255273,255125,255115,
            255281,255261,255271,255262,255279,255265,255116,255124,255288,
            255083,255089,255294,255137,255150,255166,255268,255302,255086,
            255161,255095,255134,255049,255269,255073,255081,255035,255077,
            255092,255167,255017,255122,255079,255255,255287,255301];
    elif hash1 eq 1838 and hash2 eq 1955 and hash3 eq 1400 then
        V:=[25,1,26,5,27,13,28,17,29,25,30,31,31,35,32,39];
        D:=[1,175/6,1,0,2,22,1,71/3,1,26,1,0,5,133/6,1,0,6,50/3,1,56/3,1,19,1,0,
            9,33/2,1,115/6,1,0,11,199/12,1,0,12,71/6,1,0,13,173/15,2,139/12,3,
            0];
        I:=[405551,405550,405878,406591,405557,405547,406587,406603,405546,
            406586,406287,405875,405701,405874,405873,406551,406561];
    elif hash1 eq 3060 and hash2 eq 3068 and hash3 eq 2940 then
        V:=[18,1,21,5,22,13,23,25,24,31,26,35,28,41,29,45];
        D:=[1,152/3,1,0,2,191/6,1,203/6,1,393/10,1,0,5,442/15,1,151/5,1,478/15,
            1,383/12,1,407/12,1,0,10,479/20,1,313/12,1,0,12,470/21,1,0,13,
            1123/60,1,287/15,1,0,15,827/60,1,0,16,2537/210,1,0];
        I:=[516353,516452,516355,515844,515843,515661,516018,516021,516451,
            515900,516405,515665,515855,515845,515897,515685];
    elif hash1 eq 2261 and hash2 eq 1428 and hash3 eq 2016 then
        V:=[19,1,20,9,21,25,22,37,23,63,24,97,25,123,26,141,27,155,28,173,29,
            177];
        D:=[1,51/2,1,163/6,1,175/6,1,0,4,65/3,1,71/3,1,83/3,1,28,1,125/4,1,94/3,
            1,107/3,1,0,11,133/6,1,155/6,1,157/6,2,353/12,1,193/6,1,0,17,61/3,1,
            263/12,1,67/3,1,68/3,1,97/4,1,73/3,1,74/3,1,388/15,1,26,1,79/3,1,
            139/5,1,149/5,1,0,29,551/30,1,113/6,2,611/30,2,245/12,1,41/2,1,
            207/10,1,83/4,1,125/6,1,265/12,1,227/10,1,137/6,1,293/12,1,149/6,1,
            779/30,1,367/14,1,161/6,1,0,47,253/15,1,223/12,1,283/15,2,19,2,
            401/20,1,247/12,1,145/7,1,313/15,2,251/12,1,21,1,68/3,2,275/12,1,0,
            63,339/20,1,511/30,2,209/12,2,559/30,1,75/4,1,229/12,2,233/12,1,
            45/2,1,0,74,43/3,2,475/28,1,358/21,1,1027/60,1,401/21,1,1147/60,2,0,
            82,407/30,1,137/10,3,421/30,7,85/6,4,917/60,1,46/3,1,937/60,1,63/4,
            1,0,101,61/5,1,0,102,1019/84,1,737/60,1,0];
        I:=[517304,517217,517308,517302,517279,517242,517193,517178,517155,
            517207,517195,517196,517180,517291,517324,517168,517270,517176,
            517292,517210,517185,517172,517202,517158,517216,517188,517137,
            517274,517318,517190,517297,517271,517290,517189,517353,517240,
            517204,517212,517277,517198,517090,517091,517144,517283,517013,
            517053,517320,517171,517138,517208,517165,517183,517141,517192,
            517043,517072,517269,517214,517084,517092,517099,517098,517063,
            517358,517359,517186,517278,517235,517319,517260,517296,517031,
            517061,517170,517199,517051,517070,517071,517035,517025,517066,
            517343,517313,517321,517322,517139,517175,517181,517182,517220,
            517221,517367,517036,517163,517174,517215,517034,517054,517055,
            517052,517006,516980,516984];
    elif hash1 eq 851 and hash2 eq 927 and hash3 eq 664 then
        V:=[24,1,25,11,26,19,27,27,28,33];
        D:=[1,19,2,21,1,68/3,1,23,4,0,9,35/2,1,115/6,3,39/2,1,0,14,53/3,1,18,1,
            77/4,1,0,17,95/6,1,97/6,5,0,23,57/4,1,43/3,1,44/3,1,0];
        I:=[670675,670829,670606,670591,670719,670812,670833,670834,670653,
            670689,670753,670853,670650,670079,670670,670030,670710,670040,
            670053,670704,670730,670772,670678,670720,670824];
    elif hash1 eq 1885 and hash2 eq 2814 and hash3 eq 891 then
        V:=[24,1,25,9,26,15,28,21];
        D:=[1,62/3,1,21,3,80/3,1,0,6,75/4,1,115/6,3,0,10,69/4,1,53/3,3,0,14,
            43/3,1,49/3,1,0];
        I:=[672383,672407,672411,672412,671422,671968,671842,671971,672406,
            671977,671383,671992,671997,671379,671419];
    elif hash1 eq 2872 and hash2 eq 3354 and hash3 eq 495 then
        V:=[22,1,23,5,24,13,25,25,26,33,27,41,28,55,29,65,30,69];
        D:=[1,92/3,1,0,2,163/6,1,175/6,1,59/2,1,0,5,93/4,1,71/3,4,24,1,101/4,1,
            76/3,1,0,13,109/6,1,37/2,1,145/6,1,0,16,215/12,1,55/3,1,61/3,1,0,19,
            173/12,1,197/12,1,167/10,1,101/6,1,103/6,1,37/2,1,0,25,13,1,223/15,
            2,321/20,1,17,1,0,30,51/4,1,0,31,173/15,2,0];
        I:=[417902,416400,417908,418031,417466,417286,417970,418024,418029,
            418049,416912,417320,417432,416908,417467,417893,415348,416767,
            416429,416406,416790,416543,418039,416423,418506,416921,416980,
            415509,415342,418505,416516,417331];
    elif hash1 eq 1227 and hash2 eq 1394 and hash3 eq 997 then
        V:=[12,1,13,5,14,11,16,19,17,23,18,27,19,31,20,35,21,45,22,51,25,55];
        D:=[1,36,1,0,2,73/2,1,105/2,1,0,4,110/3,1,116/3,1,134/3,1,0,7,196/5,1,0,
            8,377/12,1,0,9,32,1,0,10,385/12,1,0,11,253/15,1,278/15,1,121/6,1,
            145/6,1,0,15,571/30,1,1453/60,1,0,17,394/21,1,0,18,243/20,1,0];
        I:=[519633,519638,519556,519635,519629,519605,519529,519465,519412,
            519344,519432,519505,519445,519342,519502,519324,519364,519437];
    elif hash1 eq 99 and hash2 eq 386 and hash3 eq 2613 then
        V:=[27,1,28,7,29,15];
        D:=[1,39/2,1,47/2,1,0,3,53/3,4,18,1,59/3,2,0,10,97/6,2,71/4,1,107/6,2,
            109/6,1,0];
        I:=[227290,227787,226174,226281,226339,227296,226334,226180,227627,
            227463,227494,227777,227710,227784,227763];
    elif hash1 eq 334 and hash2 eq 1316 and hash3 eq 821 then
        V:=[15,1,16,9,17,21,18,29,19,37,20,43,21,57,22,65,24,71,25,77,26,83];
        D:=[1,53/2,1,61/2,1,65/2,1,0,4,74/3,1,104/3,1,35,2,37,1,41,1,0,10,55/2,
            1,67/2,1,235/6,1,0,13,146/5,1,133/4,1,186/5,1,0,16,119/6,1,121/6,1,
            0,18,55/3,2,298/15,1,239/12,1,61/3,1,67/3,1,508/15,1,0,25,551/30,1,
            611/30,1,41/2,1,0,28,386/21,1,401/20,1,0,30,61/4,1,579/28,1,0,32,
            939/70,1,917/60,1,0,34,181/15,1,0];
        I:=[518628,518619,518672,518657,518620,518575,518577,518576,518378,
            518607,518668,518490,518508,518603,518474,518617,518673,518581,
            518584,518438,518609,518399,518313,518342,518461,518259,518358,
            518254,518344,518201,518204,518287,518286,518415];
    elif hash1 eq 1133 and hash2 eq 3409 and hash3 eq 2696 then
        V:=[17,1,18,5,19,11,20,19,21,31,22,35,24,41,25,47];
        D:=[1,65/2,1,0,2,74/3,1,86/3,1,0,4,187/6,1,199/6,1,211/6,1,0,7,53/3,1,
            76/3,1,136/5,1,125/4,1,94/3,1,0,12,237/10,1,0,13,49/3,1,263/12,1,0,
            15,129/7,1,223/12,1,0,17,91/6,1,50/3,1,0];
        I:=[515585,515467,515161,514974,515580,515345,515521,515578,514967,
            515346,515107,514919,515581,514671,514037,514151,515574,514559];
    elif hash1 eq 1462 and hash2 eq 370 and hash3 eq 3215 then
        V:=[26,1,27,5,28,11,29,17,30,21,31,29,32,39,33,47];
        D:=[1,109/4,1,0,2,119/6,1,111/4,1,0,4,263/12,1,67/3,1,0,6,125/6,1,0,7,
            253/15,1,86/5,1,52/3,2,0,11,391/30,1,27/2,1,91/6,1,185/12,1,0,15,
            139/12,2,323/21,1,233/15,1,0,19,677/60,2,0];
        I:=[399159,398791,397833,397839,396519,395366,396472,399185,398630,
            398649,395289,399191,396544,397822,398891,399311,395312,395375,
            399706,399945];
    elif hash1 eq 1509 and hash2 eq 228 and hash3 eq 3230 then
        V:=[23,1,24,9,25,15,26,19,27,25,28,33,29,41];
        D:=[1,139/6,1,151/6,1,187/6,1,0,4,126/5,1,28,1,0,6,197/10,1,0,7,101/5,1,
            328/15,1,0,9,59/4,1,187/10,1,115/6,1,0,12,175/12,1,44/3,2,15,1,0,16,
            259/20,1,0];
        I:=[418433,418403,418373,417301,417955,416671,416210,416257,416191,
            415781,418545,417345,417876,418523,418640,415993];
    elif hash1 eq 1509 and hash2 eq 3369 and hash3 eq 151 then
        V:=[18,1,19,9,20,27,21,49,22,67,23,87,24,97,25,117,26,129,27,133,28,
            151];
        D:=[1,29,1,33,1,104/3,1,0,4,163/6,1,63/2,1,71/2,1,223/6,1,75/2,1,235/6,
            1,79/2,1,163/4,1,0,12,71/3,1,77/3,3,83/3,1,88/3,1,89/3,1,156/5,1,
            94/3,2,95/3,2,34,1,107/3,1,0,26,155/6,1,157/6,1,53/2,1,111/4,1,
            353/12,1,193/6,2,203/6,1,205/6,1,0,35,22,1,67/3,2,68/3,1,121/5,3,
            74/3,1,79/3,2,27,1,418/15,1,335/12,1,0,48,551/30,1,115/6,2,45/2,1,
            731/30,1,0,53,253/15,2,283/15,1,19,2,58/3,1,121/6,1,251/12,1,64/3,1,
            337/15,1,343/15,1,0,64,511/30,1,35/2,1,56/3,1,229/12,1,115/6,1,0,69,
            203/15,1,0,70,939/70,1,377/28,1,407/30,1,817/60,1,421/30,2,937/60,1,
            47/3,1,63/4,1,0,79,707/60,1,0];
        I:=[514603,514588,514676,514402,514909,514924,514677,514701,514890,
            514774,514616,514374,514743,514925,514969,514710,514442,514581,
            514656,514479,514918,514624,514910,514686,514619,514797,514667,
            514913,514609,514628,514443,514970,514827,514568,514654,514730,
            514795,514612,514678,514720,514971,514660,514472,514830,514766,
            514608,514769,514644,514911,514996,514820,514260,514638,514714,
            514505,514814,515006,514540,514575,514817,514576,514280,514251,
            514740,514611,514544,514449,514811,514094,514021,514109,514379,
            514092,514674,514945,514119,514079,514071,514105];
    elif hash1 eq 3389 and hash2 eq 1384 and hash3 eq 3151 then
        V:=[23,1,25,5,26,15,27,21,28,29,29,33];
        D:=[1,163/6,3,0,4,217/10,1,95/4,1,143/6,1,155/6,1,0,8,101/5,1,61/3,1,0,
            10,539/30,1,37/2,1,207/10,1,0,13,361/20,1,0,14,625/42,1,451/30,2,0];
        I:=[250432,250444,250536,250697,250590,250154,250577,250627,250599,
            249880,250670,250582,249866,249846,249851,250688];
    elif hash1 eq 2496 and hash2 eq 1815 and hash3 eq 3146 then
        V:=[22,1,23,5,24,13,25,19,26,27,27,35,28,41];
        D:=[1,19,1,0,2,115/6,2,47/2,1,163/6,1,0,6,69/4,1,77/4,3,0,10,95/6,1,
            209/12,1,119/6,1,0,13,262/15,1,18,1,73/4,1,0,16,29/2,1,197/12,1,0,
            18,289/20,1,0];
        I:=[418783,418795,418822,418125,418881,418099,418298,418495,418542,
            416612,418651,416142,417258,416018,416179,418789,415499,415542];
    elif hash1 eq 2073 and hash2 eq 447 and hash3 eq 1029 then
        V:=[22,1,23,7,24,13,25,21,26,27,27,37,29,43];
        D:=[1,27,1,31,1,0,3,43/2,1,163/6,1,0,5,64/3,1,65/3,1,26,1,0,8,119/6,1,
            133/6,1,0,10,55/3,2,56/3,1,61/3,1,62/3,1,0,15,37/2,1,113/6,1,0,17,
            185/12,1,0];
        I:=[248370,248352,248394,247938,248103,248141,247829,248136,247889,
            247913,247933,247927,248142,247872,247867,247992,247558];
    elif hash1 eq 945 and hash2 eq 2299 and hash3 eq 1126 then
        V:=[31,1,32,5,33,9,35,21,36,31];
        D:=[1,131/6,1,0,2,298/15,1,0,3,491/30,1,197/12,2,33/2,1,167/10,1,101/6,
            3,0,11,391/30,2,79/6,1,401/30,1,27/2,1,0,16,173/15,1,0];
        I:=[659384,660546,659793,659797,659805,659800,659984,659987,660565,
            660908,659707,659798,660520,659986,660535,659670];
    elif hash1 eq 663 and hash2 eq 1715 and hash3 eq 508 then
        V:=[25,1,26,9,27,19,28,25,29,29,30,35,31,41];
        D:=[1,115/6,1,55/2,2,175/6,1,0,5,71/3,2,76/3,1,77/3,1,83/3,1,0,10,107/6,
            1,79/4,1,0,12,49/3,1,0,13,33/2,2,221/12,1,0,16,15,2,46/3,1,0,19,
            391/30,1,161/12,1,0];
        I:=[243053,243048,243467,243047,242913,244163,244474,244242,242201,
            244178,244345,243310,242924,244473,242776,244450,244510,242314,
            243284,244438];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 6 mod 47.
function cached_ids_6(P,hash1,hash2,hash3)
    if hash1 eq 2215 and hash2 eq 1837 and hash3 eq 1902 then
        V:=[23,1,24,5,25,11,26,21,27,31];
        D:=[1,39/2,1,0,2,53/3,1,58/3,1,0,4,33/2,1,71/4,1,107/6,1,109/6,3,0,10,
            49/3,2,50/3,1,89/5,1,268/15,1,0,15,59/4,1,89/6,5,91/6,3,167/10,2,
            101/6,1,0];
        I:=[414892,414896,414875,414687,414755,414885,414436,414682,414714,
            414742,414898,414748,413923,414322,414415,413949,414649,414700,
            414721,414732,414610,414619,414796,413660,414347,414392];
    elif hash1 eq 3343 and hash2 eq 1520 and hash3 eq 1811 then
        V:=[27,1,28,5,29,11,30,19,31,27,32,33];
        D:=[1,139/6,1,0,2,64/3,1,65/3,2,0,5,107/6,1,109/6,1,119/6,1,0,8,49/3,2,
            50/3,1,215/12,1,0,12,89/6,2,91/6,1,0,15,193/15,1,13,1,175/12,1,0];
        I:=[665994,666465,666428,666567,666134,666651,665456,666005,666821,
            666654,665391,666432,666741,665242,665736,666731,665067];
    elif hash1 eq 523 and hash2 eq 595 and hash3 eq 2912 then
        V:=[28,1,29,5,30,13,31,29,32,33,33,41];
        D:=[1,18,1,0,2,109/6,1,119/6,1,121/6,1,0,5,81/5,1,65/4,1,49/3,1,50/3,2,
            17,1,73/4,1,61/3,1,0,13,197/12,2,0,15,155/12,1,13,1,53/4,1,0,18,
            71/6,1,0];
        I:=[664042,663578,664103,663927,663573,663952,663761,663558,664089,
            664061,663407,664158,663253,664009,664110,664109,664018,663137];
    elif hash1 eq 3014 and hash2 eq 685 and hash3 eq 1189 then
        V:=[27,1,28,7,29,19,30,27,31,31,32,35,33,43];
        D:=[1,157/6,1,167/6,1,0,3,62/3,1,263/12,1,67/3,2,121/5,1,74/3,1,0,9,
            611/30,2,245/12,1,41/2,1,0,13,253/15,1,0,14,56/3,1,0,15,203/15,2,14,
            1,883/60,2,0,20,935/84,1,673/60,1,677/60,2,34/3,2,0];
        I:=[396870,399483,397348,397879,399488,399701,396897,396858,397207,
            398942,398394,397901,396932,395690,396748,397840,397427,395942,
            396025,396292,396034,397434,398943,397977,397979];
    elif hash1 eq 194 and hash2 eq 3418 and hash3 eq 352 then
        V:=[20,1,21,5,22,13,23,29,24,49,25,65,26,83,28,97,29,109];
        D:=[1,29,1,0,2,51/2,1,55/2,2,187/6,4,0,9,65/3,2,76/3,1,26,1,109/4,1,
            117/4,1,32,1,101/3,1,0,17,131/6,1,133/6,1,45/2,1,281/12,1,237/10,1,
            145/6,4,157/6,2,111/4,1,167/6,1,0,30,56/3,2,239/12,1,101/5,1,61/3,1,
            62/3,1,67/3,1,358/15,1,0,38,491/30,2,551/30,2,37/2,2,113/6,1,115/6,
            1,611/30,1,41/2,1,659/30,1,0,49,253/15,1,17,1,52/3,1,37/2,1,223/12,
            2,19,1,0,56,67/5,3,203/15,1,208/15,1,14,1,309/20,1,0,63,117/10,1,0];
        I:=[423728,423058,422728,423089,422797,423140,423216,423218,422734,
            423124,422721,423131,423273,422856,423126,422873,423298,423283,
            422787,423151,423119,422716,422725,422862,423002,422783,422904,
            423371,422850,422798,422800,423285,422717,423292,423301,422947,
            422609,423107,423109,423045,423059,423120,423297,423249,422909,
            423023,423275,422638,423203,423284,423302,422911,423380,423581,
            423276,422452,423723,423724,423104,422790,422859,422395,422379];
    elif hash1 eq 241 and hash2 eq 451 and hash3 eq 990 then
        V:=[25,1,26,7,27,13,29,17,30,29,32,35];
        D:=[1,139/6,1,175/6,1,0,3,58/3,1,77/3,1,0,5,131/6,1,0,6,77/6,1,33/2,2,
            101/6,1,539/30,1,37/2,1,0,12,253/15,1,203/12,2,0,15,173/15,2,0];
        I:=[405564,406435,406262,405467,405804,405840,405574,406261,405623,
            405299,405741,405728,405729,405765,405739,405855];
    elif hash1 eq 3155 and hash2 eq 626 and hash3 eq 1578 then
        V:=[24,1,25,13,26,27,27,39,28,49,29,57,30,67];
        D:=[1,46/3,1,52/3,1,93/4,1,126/5,1,117/4,1,0,6,85/6,1,209/12,2,107/6,2,
            197/10,1,41/2,1,53/2,1,0,14,238/15,1,215/12,1,19,1,62/3,1,25,1,0,19,
            149/12,1,79/6,1,147/10,1,89/6,2,0,24,13,1,179/12,1,248/15,1,0,27,
            529/42,1,757/60,1,157/12,1,461/30,1,0,31,166/15,1,229/20,1,0];
        I:=[537706,537658,537678,537714,537643,537597,537594,537651,537645,
            537653,537674,537740,537637,537611,537667,537737,537662,537630,
            537632,537735,537727,537647,537660,537715,537485,537673,537303,
            537277,537668,537634,537699,537631];
    elif hash1 eq 2450 and hash2 eq 2175 and hash3 eq 803 then
        V:=[22,1,23,5,24,13,25,19,26,27,27,35,28,41];
        D:=[1,37,2,0,3,127/6,1,163/6,1,187/6,1,0,6,24,1,26,1,0,8,177/10,2,107/6,
            1,217/10,1,0,12,81/5,1,298/15,1,239/12,1,0,15,221/12,1,113/6,1,0,17,
            199/12,1,50/3,1,0];
        I:=[410432,412240,410669,411731,410453,411067,412088,411052,411112,
            411016,410063,410500,410064,410024,409923,410021,409133,409284];
    elif hash1 eq 993 and hash2 eq 1949 and hash3 eq 483 then
        V:=[20,1,21,7,23,17,24,23,26,27];
        D:=[1,80/3,1,29,1,0,3,127/6,1,139/6,1,151/6,1,163/6,2,0,8,119/6,4,121/6,
            2,0,14,239/12,1,0,15,15,2,0];
        I:=[253001,253005,253045,252996,253011,252765,253009,252766,252775,
            252998,253008,252769,253010,252505,252994,253007];
    elif hash1 eq 899 and hash2 eq 1725 and hash3 eq 2101 then
        V:=[18,1,19,7,20,13,21,21,22,25,23,33,24,39,25,43];
        D:=[1,29,1,31,1,0,3,151/6,1,199/6,2,0,6,76/3,2,77/3,1,83/3,1,0,10,121/6,
            1,0,11,298/15,1,101/5,1,61/3,1,0,14,399/20,1,611/30,1,0,16,253/15,1,
            0,17,91/6,1,511/30,1,0];
        I:=[427308,427290,427235,426917,427281,427237,427296,427283,426908,
            426912,426887,426963,426945,426566,426968,426966,427240,426437];
    elif hash1 eq 1698 and hash2 eq 2284 and hash3 eq 323 then
        V:=[25,1,26,11,27,17,28,23,30,31,31,37];
        D:=[1,139/6,1,47/2,1,151/6,1,59/2,1,0,5,71/3,2,76/3,1,0,8,119/6,1,121/6,
            1,0,10,238/15,1,239/12,1,62/3,1,0,13,41/3,1,44/3,1,0,15,44/3,1,0];
        I:=[399579,399602,399682,399381,398492,399073,399683,399150,399085,
            399304,398927,399186,399377,395500,395830];
    elif hash1 eq 2638 and hash2 eq 1237 and hash3 eq 507 then
        V:=[24,1,25,9,26,15,27,21,28,29,29,37];
        D:=[1,23,1,80/3,1,27,1,0,4,127/6,2,43/2,3,0,9,59/3,5,20,1,0,15,71/4,1,
            107/6,4,109/6,4,0,24,81/5,1,49/3,1,91/5,1,0,27,173/12,1,0];
        I:=[61675,61562,61678,61578,61625,61573,61586,61672,61536,61541,61577,
            61598,61682,61709,61531,61543,61568,61575,61576,61534,61559,61560,
            61591,61535,61528,61527,61533];
    elif hash1 eq 1087 and hash2 eq 2508 and hash3 eq 1934 then
        V:=[14,1,15,7,16,19,17,25,18,35,19,43,20,53,26,61];
        D:=[1,38,1,48,1,0,3,69/2,1,73/2,1,81/2,1,85/2,1,93/2,1,0,8,98/3,1,116/3,
            1,0,10,175/6,1,59/2,1,211/6,1,259/6,1,0,14,77/3,1,83/3,2,100/3,2,0,
            19,217/10,2,305/12,1,155/6,1,329/12,1,0,24,298/15,1,55/2,1,335/12,2,
            0,28,181/15,1,0];
        I:=[428928,428935,428826,428897,428856,428936,428932,428859,428848,
            428906,428852,428921,428707,428866,428676,428837,428835,428858,
            428755,428783,428864,428845,428796,428689,428645,428649,428686,
            428771];
    elif hash1 eq 1557 and hash2 eq 1498 and hash3 eq 3135 then
        V:=[27,1,28,9,29,15,30,21];
        D:=[1,39/2,1,127/6,2,43/2,2,0,6,53/3,2,20,1,0,9,71/4,1,107/6,1,0,11,
            43/3,2,49/3,5,0];
        I:=[664470,664457,664784,664223,664573,664266,664615,664177,664124,
            663703,664523,664612,663379,663381,663382,663385,663397];
    elif hash1 eq 1557 and hash2 eq 3409 and hash3 eq 1429 then
        V:=[23,1,24,7,25,15,26,21,28,33,29,41,30,45];
        D:=[1,163/6,1,175/6,1,0,3,71/3,1,24,1,77/3,1,0,6,121/6,1,145/6,1,0,8,
            215/12,1,55/3,1,56/3,1,19,1,62/3,1,0,13,44/3,1,179/12,1,50/3,1,0,16,
            181/12,2,0,18,37/3,1,0];
        I:=[414861,414863,414656,414663,414631,414864,414818,414341,414857,
            414654,414685,414817,414816,414305,413892,413664,413678,414441];
    elif hash1 eq 1369 and hash2 eq 754 and hash3 eq 451 then
        V:=[14,1,15,9,17,17,18,27,19,33,20,37,22,41,23,45,24,49];
        D:=[1,34,2,40,1,46,1,0,5,45/2,1,57/2,1,73/2,1,0,8,139/6,1,151/6,1,71/2,
            1,223/6,2,0,13,24,1,107/3,1,0,15,277/10,1,0,16,18,1,0,17,23,1,0,18,
            35/2,1,0,19,191/12,1,0];
        I:=[518711,518730,518708,518715,518719,518583,518750,518736,518636,
            518468,518621,518631,518590,518429,518347,518592,518277,518345,
            518301];
    elif hash1 eq 2121 and hash2 eq 1551 and hash3 eq 3360 then
        V:=[27,1,28,5,29,13,30,23,31,37,32,45];
        D:=[1,109/6,1,0,2,49/3,1,50/3,2,55/3,1,0,6,173/12,1,89/6,2,193/12,1,
            491/30,2,0,12,218/15,1,44/3,1,223/15,1,179/12,1,15,1,199/12,1,0,18,
            757/60,1,157/12,1,161/12,1,0,21,12,1,0];
        I:=[407428,408022,407565,407612,407996,406939,407478,407834,406916,
            406708,407375,407801,406266,406488,406265,407396,407174,407112,
            407883,406765,404642];
    elif hash1 eq 2356 and hash2 eq 1744 and hash3 eq 1554 then
        V:=[19,1,20,7,21,27,22,47,23,69,24,87,25,107,26,121];
        D:=[1,57/2,2,61/2,1,0,4,68/3,1,25,2,80/3,1,27,2,86/3,2,29,5,92/3,1,31,1,
            33,1,0,20,127/6,2,43/2,1,139/6,2,47/2,2,99/4,1,151/6,6,163/6,3,
            187/6,1,63/2,1,0,39,20,1,64/3,3,65/3,6,22,1,93/4,1,71/3,2,24,2,76/3,
            1,26,1,83/3,1,0,58,107/6,2,109/6,1,37/2,1,79/4,1,119/6,3,121/6,2,
            131/6,2,133/6,2,0,72,49/3,3,215/12,2,91/5,2,73/4,2,55/3,7,56/3,5,19,
            2,239/12,1,61/3,1,0,97,16,1,193/12,1,101/6,1,37/2,1,113/6,1,115/6,1,
            0,103,175/12,1,223/15,2,0];
        I:=[253324,253366,253244,253325,253348,253401,253342,253301,253367,
            253060,253361,252902,253063,253162,253278,253389,253132,253363,
            253227,253188,253352,253356,252966,253360,253156,253295,253088,
            252807,253036,253166,253322,253393,253399,252680,252915,253412,
            252732,252761,252901,253101,253396,253405,252653,252677,252785,
            252946,253015,253230,253279,252743,252712,253093,252644,252872,
            253100,252871,252642,253143,253357,253346,253345,252730,252734,
            253090,253094,253190,253359,252519,252614,252813,252866,252675,
            252715,252733,252509,252819,252739,252783,252592,253017,252617,
            252787,252884,252962,253013,253037,253205,252657,252751,252753,
            252851,252967,253110,253266,252630,252713,253091,252906,253123,
            252506,252557,252518,252763,252640,252833];
    elif hash1 eq 2356 and hash2 eq 1758 and hash3 eq 2545 then
        V:=[20,1,21,7,22,13,23,23,24,31,26,37,27,45];
        D:=[1,37,1,39,1,0,3,139/6,1,75/2,1,0,5,106/5,1,77/3,1,89/3,1,94/3,2,0,
            10,197/10,1,281/12,1,157/6,1,0,13,328/15,1,263/12,1,0,15,369/20,1,
            227/12,1,58/3,1,0,18,67/4,1,0];
        I:=[512735,512740,512216,512212,512223,512739,511527,511579,512734,
            512389,511824,511538,511534,511543,511237,511478,511387,511241];
    elif hash1 eq 2403 and hash2 eq 305 and hash3 eq 248 then
        V:=[27,1,28,5,29,9,30,15,31,19,33,23,34,31];
        D:=[1,57/2,1,0,2,271/12,1,0,3,589/24,1,751/30,1,0,5,46/3,1,0,6,46/3,1,0,
            7,239/20,1,367/30,1,371/30,1,0,10,1069/105,1,619/60,4,1421/120,1,0];
        I:=[500982,501571,501114,501309,501010,501890,501981,502966,503064,
            501086,502877,503063,503145,503151,501082];
    elif hash1 eq 2967 and hash2 eq 1883 and hash3 eq 2572 then
        V:=[29,1,30,7,31,15,32,23,33,27,34,33];
        D:=[1,131/6,1,95/4,1,0,3,298/15,1,239/12,1,101/5,2,0,7,33/2,1,101/6,1,
            217/12,1,0,10,248/15,1,0,11,27/2,2,83/6,1,0,14,11,1,667/60,2,67/6,3,
            56/5,1,45/4,1,0];
        I:=[233070,231321,233231,231954,229331,230256,231677,230945,231340,
            229343,231795,232468,230972,229922,229634,233042,229961,233607,
            233622,232513,231022];
    elif hash1 eq 1980 and hash2 eq 3312 and hash3 eq 794 then
        V:=[22,1,23,7,24,13,25,25,26,35,27,39,28,47];
        D:=[1,86/3,1,33,1,0,3,163/6,1,55/2,1,0,5,59/3,1,64/3,1,65/3,2,71/3,1,
            76/3,1,0,11,121/6,1,257/12,1,87/4,1,133/6,2,0,16,55/3,1,0,17,197/12,
            1,101/6,1,113/6,2,0,21,223/15,1,0];
        I:=[418464,418594,417577,418353,417556,416469,417873,418601,418528,
            418310,418467,416047,416182,416019,416195,418554,416130,417573,
            415972,416194,416308];
    elif hash1 eq 1886 and hash2 eq 3019 and hash3 eq 1235 then
        V:=[19,1,20,5,21,11,22,27,23,43,25,49,26,59,28,63];
        D:=[1,235/6,1,0,2,89/3,1,101/3,1,0,4,157/6,1,277/10,1,167/6,2,179/6,1,
            377/12,1,127/4,1,191/6,1,0,12,263/12,1,358/15,1,287/12,1,388/15,1,
            311/12,1,79/3,1,80/3,1,0,19,611/30,2,26,1,0,22,229/12,1,581/30,1,
            233/12,1,39/2,3,0,28,1123/60,1,0,29,707/60,1,0];
        I:=[426637,426648,427440,426636,426572,426599,426604,426551,426632,
            426503,427063,426623,426601,426619,426498,426497,426548,426647,
            426593,426594,426446,426512,426506,426505,426502,426517,426742,
            426426,426739];
    elif hash1 eq 946 and hash2 eq 2432 and hash3 eq 2011 then
        V:=[26,1,27,5,28,15,29,25,30,33,31,43];
        D:=[1,64/3,1,0,2,63/4,1,95/6,2,97/6,1,197/10,1,0,7,89/5,1,91/5,1,73/4,1,
            55/3,2,0,12,431/30,1,29/2,2,193/12,1,0,16,188/15,1,151/12,1,40/3,1,
            46/3,1,0,20,73/6,1,0];
        I:=[406193,407575,406323,408358,408720,408133,405285,407474,404465,
            407742,407874,406329,408490,408620,404497,407412,408552,406152,
            407875,404499];
    elif hash1 eq 1557 and hash2 eq 1857 and hash3 eq 650 then
        V:=[18,1,19,5,20,13,21,21,22,31,26,39];
        D:=[1,80/3,1,0,2,187/6,1,199/6,1,139/4,1,0,5,109/4,1,156/5,1,141/4,1,0,
            8,281/12,1,305/12,1,329/12,1,377/12,1,0,12,49/3,1,178/7,1,311/12,1,
            0,15,187/12,1,0];
        I:=[515514,515589,515584,514851,515377,514993,514534,515108,515175,
            514836,514153,515583,514124,514150,514066];
    elif hash1 eq 523 and hash2 eq 177 and hash3 eq 1730 then
        V:=[23,1,24,11,25,19,26,27];
        D:=[1,41/2,1,45/2,1,49/2,1,53/2,1,0,5,62/3,1,21,1,23,3,0,10,115/6,1,
            39/2,3,127/6,1,0,15,58/3,1,0];
        I:=[672068,672142,671944,671942,671956,671923,671597,671922,672066,
            671955,671414,671796,672101,671629,671628];
    elif hash1 eq 2873 and hash2 eq 1813 and hash3 eq 687 then
        V:=[25,1,26,7,27,17,28,23,29,27];
        D:=[1,121/6,2,87/4,1,0,4,18,2,73/4,1,55/3,1,239/12,1,0,9,167/10,1,67/4,
            1,0,11,253/15,1,0,12,27/2,1,83/6,3,0];
        I:=[246754,246796,246782,246651,246683,246697,246652,246686,246820,
            246684,246462,246596,246654,246673,246696];
    elif hash1 eq 476 and hash2 eq 1151 and hash3 eq 1808 then
        V:=[18,1,19,7,20,13,21,17];
        D:=[1,26,2,28,1,0,4,45/2,1,49/2,6,0,11,23,2,0,13,127/6,1,43/2,2,0];
        I:=[674600,674602,674597,674560,674562,674563,674567,674601,674603,
            674605,674574,674609,674564,674549,674561];
    elif hash1 eq 758 and hash2 eq 2720 and hash3 eq 1985 then
        V:=[24,1,26,5,28,11,29,23,30,31,31,43];
        D:=[1,65/3,1,0,2,298/15,1,97/3,1,0,4,121/6,1,247/12,1,251/12,1,332/15,1,
            173/7,1,0,9,91/6,1,169/10,1,631/30,1,0,12,907/60,1,46/3,1,101/6,1,
            382/21,1,394/21,1,0,17,817/60,1,86/5,1,0];
        I:=[504918,504963,504641,503958,504330,504636,504010,504202,504620,
            503875,504700,504920,505009,503986,503890,504053,504928,503843];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 7 mod 47.
function cached_ids_7(P,hash1,hash2,hash3)
    if hash1 eq 712 and hash2 eq 2651 and hash3 eq 1317 then
        V:=[27,1,28,7,29,17,30,25,31,33,32,39];
        D:=[1,131/6,1,95/4,1,0,3,16,1,215/12,1,239/12,1,20,1,0,7,193/12,1,
            197/12,1,221/12,1,0,10,217/15,1,175/12,1,223/12,1,0,13,389/30,1,
            391/30,1,0,15,178/15,1,0];
        I:=[240291,238813,240068,239568,238491,239566,240065,240351,238799,
            240436,239569,238164,240467,240074,238064];
    elif hash1 eq 2780 and hash2 eq 1181 and hash3 eq 1737 then
        V:=[23,1,24,13,25,23,26,31,27,41,29,51];
        D:=[1,139/6,1,47/2,1,151/6,1,51/2,1,175/6,1,0,6,59/3,1,65/3,1,24,2,77/3,
            1,0,11,131/6,1,237/10,1,145/6,2,0,15,55/3,4,239/12,2,61/3,1,62/3,1,
            0,23,89/6,1,103/6,1,37/2,1,113/6,1,0,27,401/30,1,27/2,1,0];
        I:=[414266,414280,414275,414044,414049,414283,413904,414061,414394,
            414101,414242,414344,413991,414319,413944,414103,414272,414421,
            414308,414333,414338,413942,414342,414039,414249,414413,414080,
            414089];
    elif hash1 eq 2827 and hash2 eq 2919 and hash3 eq 2329 then
        V:=[27,1,28,13,29,21,30,35,31,47];
        D:=[1,107/6,1,109/6,1,41/2,1,131/6,1,133/6,1,0,6,81/5,1,50/3,2,268/15,1,
            0,10,431/30,1,29/2,1,89/6,1,33/2,1,67/4,1,113/6,1,0,16,179/12,1,
            82/5,1,199/12,2,253/15,1,86/5,1,0,22,401/30,1,0];
        I:=[405043,405048,405200,404882,405354,404950,404865,405022,404938,
            405222,405324,405054,405013,404692,405100,404982,405247,404701,
            405236,405087,405091,404806];
    elif hash1 eq 1323 and hash2 eq 2455 and hash3 eq 1744 then
        V:=[25,1,26,5,27,9,28,17,29,27,30,33,32,39];
        D:=[1,121/6,2,0,3,73/4,1,0,4,103/6,1,221/12,1,37/2,1,0,7,179/12,1,
            242/15,1,248/15,1,253/15,1,0,11,67/4,1,339/20,1,0,13,202/15,1,
            167/12,1,0,15,12,1,0];
        I:=[508113,508143,508106,508131,508111,508028,508105,508139,508121,
            508109,508052,508144,508135,508114,507718];
    elif hash1 eq 1652 and hash2 eq 165 and hash3 eq 2689 then
        V:=[21,1,22,5,23,11,24,19,25,23,26,27,27,37,28,45];
        D:=[1,187/6,1,0,2,76/3,1,117/4,1,0,4,131/6,1,133/6,1,143/6,1,0,7,239/12,
            1,0,8,551/30,1,0,9,321/20,1,97/6,1,253/15,1,17,1,0,13,27/2,2,193/12,
            1,997/60,1,0,17,66/5,1,53/4,1,0];
        I:=[424792,424492,422608,424408,424666,424794,422562,422892,423188,
            423177,423522,423568,423237,423518,422534,422579,423516,423190];
    elif hash1 eq 2122 and hash2 eq 2907 and hash3 eq 1640 then
        V:=[15,1,16,5,17,11,19,17,21,23,22,27,23,35,24,39,26,43];
        D:=[1,101/2,1,0,2,110/3,1,146/3,1,0,4,223/6,1,271/6,1,0,6,179/6,1,227/6,
            1,0,8,361/12,1,0,9,92/5,1,283/15,2,271/12,1,0,13,169/10,1,0,14,
            1363/60,1,0,15,206/15,1,0];
        I:=[518494,518504,518486,518493,518505,518483,518492,518487,518544,
            518412,518481,518457,518414,518258,518463];
    elif hash1 eq 900 and hash2 eq 1017 and hash3 eq 3098 then
        V:=[14,1,15,15,16,31,17,49,18,67,19,85,20,95,21,113,22,117];
        D:=[1,28,1,32,2,34,1,36,4,38,1,40,2,0,12,53/2,1,57/2,2,61/2,1,65/2,2,
            69/2,3,73/2,1,77/2,2,0,24,74/3,1,25,3,27,1,29,1,31,5,104/3,1,35,4,
            39,1,0,41,139/6,2,151/6,1,51/2,1,163/6,8,175/6,2,59/2,3,187/6,1,
            71/2,1,0,60,65/3,2,22,1,70/3,2,76/3,1,77/3,1,26,1,83/3,1,30,1,0,70,
            119/6,2,121/6,2,133/6,1,167/6,1,0,76,55/3,1,56/3,1,61/3,1,328/15,1,
            67/3,1,68/3,2,23,2,79/3,1,0,86,37/2,1,0,87,253/15,1,17,2,0];
        I:=[255713,255693,255726,255698,255547,255695,255703,255714,255699,
            255691,255697,255696,255574,255700,255725,255607,255610,255632,
            255689,255701,255709,255559,255639,255599,255634,255702,255719,
            255577,255615,255578,255590,255636,255637,255722,255627,255507,
            255575,255581,255618,255472,255560,255562,255550,255558,255452,
            255467,255549,255566,255611,255616,255629,255630,255481,255564,
            255473,255487,255556,255469,255374,255463,255588,255482,255565,
            255631,255460,255480,255492,255443,255372,255474,255579,255451,
            255608,255419,255358,255484,255471,255391,255386,255366,255400,
            255402,255399,255476,255357,255489,255382,255371,255572];
    elif hash1 eq 1511 and hash2 eq 3133 and hash3 eq 1350 then
        V:=[16,1,17,5,18,11,19,21,20,31,21,35,22,39,23,47];
        D:=[1,24,2,0,3,45/2,2,49/2,2,0,7,62/3,1,21,1,92/3,2,35,1,0,12,91/4,1,
            139/6,1,47/2,1,71/2,1,0,16,77/4,1,0,17,79/4,1,0,18,81/5,1,50/3,1,
            55/3,1,0,21,217/12,1,45/2,1,0];
        I:=[516693,516786,516696,516788,516445,516695,516683,516694,516511,
            516517,516420,516384,516395,516373,516300,516554,515968,515999,
            515890,516306,515871,515768];
    elif hash1 eq 1652 and hash2 eq 2548 and hash3 eq 136 then
        V:=[25,1,26,7,27,11,28,21,29,31,30,35];
        D:=[1,127/6,3,43/2,1,0,5,59/3,1,0,6,95/6,1,97/6,1,107/6,1,109/6,2,0,11,
            16,1,81/5,1,49/3,1,268/15,1,0,15,173/12,1,0,16,13,1,0];
        I:=[672109,672333,672345,671895,671764,672137,671862,672119,671606,
            672253,672157,671618,672126,671136,671490,671637];
    elif hash1 eq 1605 and hash2 eq 3309 and hash3 eq 2084 then
        V:=[19,1,20,13,21,27,22,41,23,51,24,57,25,73,26,79];
        D:=[1,49/2,3,53/2,2,57/2,1,61/2,4,65/2,1,0,12,68/3,1,74/3,1,25,1,80/3,3,
            27,1,31,3,0,22,43/2,1,139/6,3,47/2,1,151/6,3,51/2,2,175/6,1,0,33,
            106/5,1,64/3,1,65/3,4,116/5,1,0,40,119/6,2,131/6,1,0,43,50/3,1,
            268/15,3,215/12,2,18,2,91/5,1,55/3,2,56/3,2,0,56,197/12,1,37/2,2,0,
            59,218/15,1,0];
        I:=[253841,253849,253860,253850,253961,253825,253863,253895,253916,
            253917,253942,253822,253657,253881,253711,253885,253903,253921,
            253767,253934,253937,253680,253751,253768,253940,253901,253713,
            253772,253852,253566,253628,253577,253561,253677,253670,253681,
            253737,253887,253575,253564,253709,253567,253514,253476,253557,
            253576,253672,253708,253491,253613,253581,253478,253574,253515,
            253738,253494,253454,253470,253488];
    elif hash1 eq 1182 and hash2 eq 1806 and hash3 eq 3067 then
        V:=[27,1,28,7,29,11,30,15,31,23,32,35];
        D:=[1,121/6,1,49/2,1,0,3,62/3,1,0,4,167/10,1,0,5,41/3,2,175/12,2,15,1,0,
            10,391/30,1,157/12,1,401/30,1,27/2,8,137/10,1,0,22,45/4,1,163/12,1,
            0];
        I:=[403390,403834,403846,403482,403794,403850,403454,403650,403837,
            403372,403710,403638,403380,403696,403701,403712,403728,403731,
            403836,403838,403517,403501,402689];
    elif hash1 eq 2968 and hash2 eq 1259 and hash3 eq 1285 then
        V:=[30,1,31,5,32,13,34,21,35,31];
        D:=[1,65/3,1,0,2,107/6,2,109/6,1,37/2,1,0,6,43/3,1,44/3,1,49/3,1,0,9,
            151/12,1,193/15,2,13,2,40/3,1,0,15,67/6,1,0];
        I:=[658578,655723,657100,656706,658065,656933,658995,656245,657968,
            656885,657649,658688,658700,657650,658030];
    elif hash1 eq 1323 and hash2 eq 1230 and hash3 eq 90 then
        V:=[23,1,24,9,25,19,26,31,27,39,28,43];
        D:=[1,41/2,2,45/2,3,49/2,2,0,8,56/3,1,19,1,62/3,2,21,2,0,14,103/6,1,
            35/2,1,115/6,1,39/2,1,127/6,1,0,19,61/4,1,46/3,3,47/3,2,0,25,85/6,1,
            0,26,208/15,1,167/12,1,14,1,43/3,1,0];
        I:=[244997,245082,245029,245259,245433,245060,245322,245328,245097,
            244967,245077,245223,245279,245066,244993,245422,245153,245238,
            245302,244753,245016,245288,245141,245360,245408,243957,245121,
            245074,245075];
    elif hash1 eq 2404 and hash2 eq 37 and hash3 eq 986 then
        V:=[23,1,24,7,25,17,26,23,27,31,28,43];
        D:=[1,131/6,1,45/2,1,0,3,56/3,1,61/3,1,21,1,263/12,1,0,7,103/6,1,611/30,
            1,0,9,47/3,1,253/15,1,369/20,1,0,12,91/6,2,461/30,1,185/12,1,31/2,1,
            95/6,1,0,18,233/15,1,187/12,1,0];
        I:=[422298,422047,421911,422020,422093,421563,421910,421581,422060,
            421621,421618,421276,422306,421756,421279,421545,422154,421262,
            421189];
    elif hash1 eq 1934 and hash2 eq 1893 and hash3 eq 2198 then
        V:=[14,1,15,7,16,11,17,15,18,19,19,25,20,31,21,43,22,51,24,55,25,63,26,
            67];
        D:=[1,36,1,44,1,0,3,69/2,1,0,4,110/3,1,0,5,271/6,1,0,6,141/4,1,36,1,0,8,
            181/6,1,449/12,1,0,10,79/3,1,478/15,1,502/15,1,34,1,104/3,1,0,15,
            339/14,1,591/20,1,423/14,1,0,18,271/12,1,0,19,263/15,1,531/28,1,
            332/15,1,0,22,809/42,1,0,23,71/5,1,0];
        I:=[518411,518488,518402,518410,518556,518391,518533,518530,518558,
            518353,518559,518375,518531,518232,518308,518422,518237,518491,
            518363,518231,518279,518274,518529];
    elif hash1 eq 1793 and hash2 eq 446 and hash3 eq 129 then
        V:=[18,1,19,5,20,11,21,23,22,29,23,35,24,41,25,49,26,55,27,65,28,69];
        D:=[1,122/3,1,0,2,175/6,1,79/2,1,0,4,28,1,30,1,95/3,1,34,1,118/3,2,0,10,
            217/10,2,143/6,1,0,13,138/7,1,151/5,1,0,15,125/6,1,163/6,1,0,17,
            92/5,1,56/3,1,308/15,2,0,21,161/12,1,185/12,1,0,23,79/6,1,419/28,1,
            91/6,1,69/4,1,0,27,797/60,1,0,28,417/35,1,827/60,1,0];
        I:=[543252,543105,543372,543330,543328,543371,543204,543114,543135,
            542816,543118,542796,542829,543134,543205,542627,543103,542828,
            543238,543254,543289,542563,543373,542809,542633,542803,542799,
            542784,542573];
    elif hash1 eq 1417 and hash2 eq 816 and hash3 eq 536 then
        V:=[24,1,25,9,26,19,27,29,28,39,29,45];
        D:=[1,74/3,2,25,1,80/3,1,0,5,115/6,1,127/6,1,47/2,1,151/6,1,0,9,52/3,1,
            77/4,1,64/3,1,71/3,1,0,13,109/6,3,37/2,1,119/6,1,121/6,1,0,19,44/3,
            2,81/5,1,0,22,173/12,1,29/2,1,491/30,1,33/2,1,0];
        I:=[244990,245042,244691,245492,245090,245506,244288,245626,245299,
            244418,245575,244467,242933,244579,245508,244291,243341,244482,
            245558,245620,244292,244002,245256,242281,242835];
    elif hash1 eq 1887 and hash2 eq 2161 and hash3 eq 3380 then
        V:=[16,1,17,5,18,17,19,27,20,39,21,47,22,57,23,67,24,75,25,81];
        D:=[1,36,1,0,2,61/2,1,65/2,1,69/2,1,73/2,1,81/2,1,0,7,92/3,1,31,1,33,2,
            35,1,0,12,163/6,1,59/2,1,187/6,2,199/6,2,211/6,1,0,19,77/3,2,82/3,2,
            83/3,1,0,24,109/6,1,233/12,2,119/6,1,155/6,1,0,29,215/12,1,67/3,2,
            68/3,2,311/12,1,0,35,125/6,1,269/12,1,91/4,1,0,38,179/12,1,15,1,0,
            40,877/60,1,59/4,2,299/20,1,451/30,1,0];
        I:=[515145,515100,515329,515023,515104,515330,515126,515022,515136,
            515188,515051,515113,515096,515102,515415,515130,515287,515224,
            515205,515249,514772,515184,515158,515086,515037,515110,515095,
            514765,515033,514520,514775,514767,515244,514294,514648,514297,
            514289,515418,515043,514578,515026,515039,514547,514553];
    elif hash1 eq 430 and hash2 eq 256 and hash3 eq 2723 then
        V:=[23,1,24,7,25,15,26,23,27,29,28,37,30,43];
        D:=[1,163/6,1,187/6,1,0,3,109/4,1,82/3,1,89/3,1,0,6,281/12,1,155/6,1,
            329/12,1,0,9,298/15,1,73/3,1,0,11,599/30,1,41/2,1,471/20,1,0,14,
            253/15,1,106/5,1,0,16,907/60,1,0];
        I:=[502493,502169,502489,502637,502527,502506,502621,502162,502501,
            502628,502515,502620,502119,502454,501862,501816];
    elif hash1 eq 2404 and hash2 eq 2794 and hash3 eq 3243 then
        V:=[22,1,23,9,24,19,25,29,26,39,27,47,29,67];
        D:=[1,23,1,92/3,1,31,1,0,4,43/2,1,151/6,1,55/2,1,63/2,1,0,8,59/3,2,65/3,
            2,22,1,83/3,1,0,14,109/6,1,37/2,1,121/6,1,41/2,1,0,18,55/3,2,20,1,
            61/3,2,0,23,431/30,2,89/6,1,491/30,1,33/2,1,103/6,1,551/30,1,221/12,
            2,37/2,2,113/6,1,0,35,391/30,3,401/30,1,0];
        I:=[417494,417594,416360,416924,417362,416964,416963,417534,417659,
            416353,417495,417189,417333,417237,416977,416943,416938,416888,
            417605,416738,417223,417610,417624,417625,417574,417564,417673,
            416979,417669,416557,417607,416534,417674,416387,417632,417705,
            417706,416895];
    elif hash1 eq 242 and hash2 eq 2397 and hash3 eq 3286 then
        V:=[20,1,21,7,22,11,24,29,25,35,26,45];
        D:=[1,74/3,1,80/3,1,0,3,139/6,1,0,4,53/3,1,59/3,2,20,1,85/4,1,64/3,1,
            116/5,1,71/3,1,24,1,0,13,49/3,4,91/5,1,0,18,91/6,1,491/30,1,167/10,
            1,101/6,1,0,22,47/3,1,0];
        I:=[422202,422203,422200,422295,422206,422211,422213,421991,422140,
            421560,421559,422028,421978,422198,422199,422303,421287,421514,
            421436,421575,421573,421162];
    elif hash1 eq 2451 and hash2 eq 219 and hash3 eq 1379 then
        V:=[20,1,21,15,22,25,23,33,24,45,25,57,26,71,27,75,28,79];
        D:=[1,23,1,25,1,27,1,86/3,1,29,2,31,2,0,9,47/2,1,151/6,5,51/2,2,163/6,1,
            0,18,64/3,1,22,1,71/3,1,0,21,71/4,1,109/6,1,119/6,2,121/6,2,257/12,
            1,0,28,73/4,3,55/3,3,56/3,6,20,1,81/4,1,0,42,193/12,1,197/12,1,33/2,
            1,101/6,1,113/6,1,269/12,1,0,48,15,1,0,49,31/2,1,0,50,12,1,0];
        I:=[252447,252331,252262,252449,252328,252382,252113,252271,252283,
            252059,252263,252337,252442,252444,252364,252431,252342,252208,
            252110,252137,252131,252330,252112,252126,252070,252084,252082,
            252061,252129,252215,252209,252254,252445,252064,252072,252274,
            252333,252403,252456,252013,251997,252130,251996,252121,252085,
            251867,251832,251969,251837,252009];
    elif hash1 eq 2686 and hash2 eq 3145 and hash3 eq 1903 then
        V:=[20,1,21,11,22,23,23,39,24,57,25,69,26,79,27,83,28,95];
        D:=[1,23,1,74/3,1,86/3,2,29,1,0,6,39/2,2,43/2,3,139/6,2,99/4,1,107/4,1,
            0,15,53/3,1,18,1,58/3,1,59/3,5,106/5,1,65/3,1,77/3,1,0,26,97/6,2,
            177/10,1,109/6,1,197/10,1,121/6,1,87/4,1,45/2,1,143/6,1,0,35,238/15,
            1,191/12,1,49/3,1,18,3,99/5,1,0,42,173/12,1,147/10,1,89/6,1,167/10,
            1,0,46,15,2,0,48,185/14,1,401/30,1,161/12,1,137/10,1,451/30,1,0,53,
            178/15,2,0];
        I:=[510376,510291,510247,510484,510415,510270,510286,509798,509801,
            510249,509579,510622,510628,510263,510281,510665,510277,509850,
            510032,510051,510251,510664,509818,508823,510432,509649,510645,
            509698,510490,508724,509096,510459,508384,509060,509780,510262,
            509633,509075,509144,509207,509400,510629,509755,509796,509650,
            508432,510381,508801,508815,508822,510105,508847,509463,509514];
    elif hash1 eq 1088 and hash2 eq 830 and hash3 eq 2801 then
        V:=[10,1,11,5,13,11,14,21,16,27,17,37,18,49,19,53,20,63,22,81,23,89];
        D:=[1,46,1,0,2,89/2,1,121/2,1,0,4,107/4,1,115/4,1,235/6,1,179/4,1,0,8,
            134/5,1,101/3,1,0,10,32,1,98/3,1,598/15,2,40,2,0,16,269/12,1,289/12,
            1,791/30,1,559/14,1,481/12,1,0,21,441/20,1,0,22,521/30,1,1117/60,1,
            677/28,1,97/4,1,0,26,235/12,1,20,1,304/15,1,316/15,2,223/10,1,70/3,
            1,848/35,1,364/15,1,0,35,1774/105,1,1103/60,1,569/30,1,0,38,331/18,
            1,683/36,1,19,1,0];
        I:=[544227,544226,544214,544197,544190,544222,544194,544179,544210,
            544131,544211,544149,544181,544126,544145,544174,544134,544153,
            544115,544118,544185,544182,544168,544121,544088,544140,544143,
            544186,544173,544178,544087,544080,544092,544086,544107,544127,
            544096,544078,544081,544077];
    elif hash1 eq 1934 and hash2 eq 318 and hash3 eq 2936 then
        V:=[19,1,20,9,21,19,22,25,23,37,24,41,25,49];
        D:=[1,59/2,1,63/2,1,71/2,2,0,5,20,1,22,1,71/3,1,164/5,1,0,9,41/2,1,53/2,
            1,0,11,73/4,1,56/3,1,81/4,1,61/3,1,97/4,1,0,16,103/6,1,0,17,33/2,1,
            37/2,1,61/3,1,0,20,95/6,1,0];
        I:=[515390,515291,515274,515386,515410,515613,515455,514468,515411,
            515309,515451,515311,515456,515310,514233,515416,515450,514234,
            514253,515413];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 8 mod 47.
function cached_ids_8(P,hash1,hash2,hash3)
    if hash1 eq 1042 and hash2 eq 1419 and hash3 eq 1868 then
        V:=[23,1,24,9,25,15,26,23,27,27];
        D:=[1,41/2,1,45/2,2,49/2,2,0,6,62/3,3,21,1,0,10,67/4,1,103/6,2,35/2,2,0,
            15,53/3,1,0,16,95/6,1,0];
        I:=[672897,672645,672898,672774,672896,672708,672776,672788,672863,
            672780,672671,672875,672657,672815,672706,672587];
    elif hash1 eq 2969 and hash2 eq 1761 and hash3 eq 3139 then
        V:=[21,1,22,11,23,21,24,29,25,35,26,47,27,59];
        D:=[1,103/6,1,39/2,1,55/2,1,63/2,1,0,5,53/3,1,65/3,1,93/4,1,146/5,1,0,9,
            107/6,2,79/4,1,131/6,1,0,13,99/5,1,239/12,2,0,16,89/6,1,491/30,1,
            251/14,1,539/30,1,307/14,1,0,21,44/3,1,97/6,1,361/20,1,272/15,1,
            109/6,1,0,26,401/30,1,56/3,1,0];
        I:=[510760,510712,510372,510050,510698,509450,509115,509124,510423,
            510709,509810,510749,510029,509812,509821,510403,509997,508831,
            510000,508450,510384,509950,509097,509332,509335,509891,508296];
    elif hash1 eq 995 and hash2 eq 1417 and hash3 eq 154 then
        V:=[21,1,22,7,23,23,24,29,25,37,26,53,27,65,28,73];
        D:=[1,139/6,1,151/6,1,0,3,59/3,2,85/4,1,65/3,2,22,1,70/3,1,71/3,1,24,1,
            0,12,79/4,1,121/6,2,0,15,50/3,1,215/12,3,55/3,1,0,20,197/12,1,101/6,
            1,539/30,1,551/30,1,37/2,2,187/10,1,75/4,2,0,29,15,1,61/4,1,50/3,1,
            253/15,1,17,1,0,34,27/2,1,299/20,1,181/12,1,0,37,143/12,1,0];
        I:=[422314,422296,422279,422308,422104,422073,422311,422222,422300,
            422304,422122,422155,422103,422318,422092,421611,422108,422116,
            421295,421317,422098,421508,421537,421265,421268,421637,421278,
            421280,421515,421617,421289,421259,421536,421544,421216,421120,
            421787];
    elif hash1 eq 1230 and hash2 eq 1410 and hash3 eq 2761 then
        V:=[25,1,26,5,27,9,28,17,29,27,30,39,31,45,32,53];
        D:=[1,123/4,1,0,2,76/3,1,0,3,97/6,1,109/6,2,197/10,1,0,7,81/5,1,49/3,1,
            262/15,1,62/3,1,0,11,431/30,2,173/12,1,147/10,1,89/6,1,33/2,1,0,17,
            218/15,1,329/20,1,0,19,757/60,1,541/42,1,27/2,1,0,22,239/21,1,0];
        I:=[398737,398557,400612,398777,399907,396965,398807,400244,396939,
            400205,398711,398894,398709,396716,399908,398135,398025,395747,
            398652,396730,400002,396779];
    elif hash1 eq 3392 and hash2 eq 2709 and hash3 eq 3392 then
        V:=[23,1,24,7,25,15,26,27,27,33,28,41];
        D:=[1,103/6,1,115/6,1,0,3,47/3,1,106/5,1,116/5,1,0,6,95/6,2,33/2,1,
            177/10,1,71/4,2,133/6,1,0,13,43/3,2,238/15,1,0,16,431/30,2,89/6,2,
            223/14,1,0,21,193/15,1,302/21,1,0];
        I:=[503583,503549,503406,502109,502102,503587,503698,503565,502812,
            502669,503119,503687,503410,503568,503149,502983,503223,503579,
            503619,501973,503121,502049];
    elif hash1 eq 1089 and hash2 eq 1702 and hash3 eq 2883 then
        V:=[24,1,25,7,26,11,27,17,28,27];
        D:=[1,18,1,59/3,1,0,3,109/6,1,0,4,49/3,2,50/3,1,0,7,89/6,1,91/6,3,
            491/30,2,33/2,1,0,14,53/4,1,40/3,1,0];
        I:=[414891,414902,414774,414878,414890,414778,414771,414634,414695,
            414746,414350,414353,413893,414438,414408];
    elif hash1 eq 2076 and hash2 eq 3359 and hash3 eq 1513 then
        V:=[21,1,22,7,23,17,24,21,25,27,26,31];
        D:=[1,49/2,1,53/2,1,0,3,68/3,1,23,1,74/3,1,25,1,0,7,127/6,3,0,10,53/3,1,
            64/3,1,0,12,209/12,1,0,13,191/12,2,49/3,1,0];
        I:=[674039,674107,674061,673932,673887,674059,673961,673962,674065,
            673926,673794,673928,673849,673963,673791];
    elif hash1 eq 2640 and hash2 eq 976 and hash3 eq 1640 then
        V:=[23,1,24,7,26,11,27,15,28,19,29,25,30,31,31,35];
        D:=[1,163/6,1,199/6,1,0,3,88/3,2,0,5,43/3,1,0,6,221/12,2,0,8,193/15,1,
            19,1,0,10,511/30,2,205/12,1,0,13,223/12,1,0,14,347/30,1,117/10,2,0];
        I:=[502213,502503,502138,502215,502157,502108,502110,502100,502055,
            502073,502133,502264,501048,502117,502125,502240];
    elif hash1 eq 1653 and hash2 eq 2874 and hash3 eq 888 then
        V:=[17,1,18,7,19,17,20,23,21,33,23,43,24,49,26,53,27,57];
        D:=[1,65/2,1,81/2,1,0,3,27,2,86/3,1,92/3,1,31,1,0,8,127/6,1,59/2,4,0,13,
            85/4,2,93/4,1,71/3,1,77/3,1,0,18,233/12,1,121/6,1,257/12,1,281/12,1,
            0,22,101/6,1,551/30,2,0,25,199/12,1,0,26,907/60,1,0,27,361/30,1,0];
        I:=[425306,425323,425140,425316,425266,425083,425147,425185,424991,
            425035,425115,425131,424968,425153,425207,425177,425347,425098,
            424971,425047,425186,424998,424896,425032,425043,424869,425041];
    elif hash1 eq 1794 and hash2 eq 2084 and hash3 eq 1757 then
        V:=[23,1,24,5,25,11,26,17,27,25,28,33,29,39,30,47];
        D:=[1,199/6,1,0,2,85/4,1,71/3,1,0,4,119/6,1,95/4,1,0,6,215/12,1,55/3,1,
            62/3,1,0,9,33/2,1,67/4,2,183/10,1,0,13,44/3,1,15,2,0,16,79/6,1,
            401/30,1,161/12,1,0,19,203/15,1,0];
        I:=[416362,418656,418083,418286,416046,416247,416442,418075,415947,
            416143,416188,416248,418368,416139,416375,415948,417273,415931,
            415000];
    elif hash1 eq 1324 and hash2 eq 2738 and hash3 eq 216 then
        V:=[24,1,25,7,26,13,27,25,28,31];
        D:=[1,25,1,80/3,1,0,3,39/2,4,43/2,1,0,8,47/3,2,53/3,1,18,1,59/3,1,20,2,
            0,15,85/6,2,109/6,2,0,19,167/12,1,71/5,1,0];
        I:=[408727,408717,408409,408684,408740,408741,408763,408207,408227,
            408386,408559,408267,408241,408458,408239,408255,406399,407001,
            408238,408036];
    elif hash1 eq 1559 and hash2 eq 2069 and hash3 eq 642 then
        V:=[14,1,15,11,16,23,17,41,18,67,19,91,20,119,21,145,22,171,23,185,24,
            197,25,213];
        D:=[1,28,1,30,2,32,1,34,1,0,6,61/2,1,65/2,1,69/2,1,73/2,1,81/2,1,0,11,
            25,2,86/3,2,29,1,92/3,1,33,3,35,2,110/3,1,39,1,0,24,139/6,3,47/2,2,
            99/4,1,151/6,2,51/2,2,163/6,2,55/2,1,175/6,7,59/2,1,187/6,1,199/6,2,
            139/4,1,0,49,64/3,1,65/3,1,22,1,70/3,2,71/3,4,24,1,136/5,2,82/3,1,
            83/3,3,28,2,117/4,1,0,68,119/6,1,121/6,3,257/12,1,87/4,1,131/6,3,
            133/6,2,237/10,1,95/4,3,145/6,2,305/12,1,155/6,1,277/10,1,111/4,1,0,
            89,55/3,1,56/3,2,298/15,1,61/3,2,62/3,3,21,1,263/12,1,111/5,1,89/4,
            1,67/3,1,24,1,105/4,1,0,105,551/30,2,221/12,2,37/2,1,113/6,1,611/30,
            1,245/12,1,125/6,1,659/30,2,265/12,1,223/10,1,45/2,1,137/6,1,0,120,
            15,2,50/3,2,17,2,52/3,2,56/3,1,464/21,1,0,130,461/30,1,185/12,2,
            157/10,1,63/4,1,95/6,1,0,136,323/21,2,77/5,1,233/15,1,187/12,1,
            110/7,1,238/15,1,16,1,0,144,371/30,2,0];
        I:=[429573,429572,429574,429564,429566,429569,429578,429580,429563,
            429571,429535,429556,429528,429554,429567,429555,429422,429561,
            429576,429435,429579,429519,429454,429529,429540,429547,429570,
            429575,429418,429545,429559,429420,429568,429521,429543,429565,
            429419,429426,429489,429520,429526,429550,429558,429445,429312,
            429408,429487,429394,429542,429539,429464,429428,429451,429449,
            429524,429541,429544,429577,429281,429327,429503,429424,429437,
            429459,429452,429455,429432,429549,429306,429538,429548,429447,
            429406,429466,429470,429482,429333,429496,429349,429413,429416,
            429423,429337,429446,429425,429170,429344,429169,429536,429317,
            429457,429345,429313,429409,429417,429433,429436,429562,429375,
            429283,429410,429465,429162,429173,429320,429322,429307,429439,
            429338,429461,429280,429163,429282,429191,429340,429341,429197,
            429167,429174,429411,429412,429552,429553,429326,429460,429456,
            429462,429165,429132,429303,429290,429309,429316,429285,429310,
            429120,429123,429124,429134,429127,429136,429121,429125,429286,
            429288];
    elif hash1 eq 3392 and hash2 eq 274 and hash3 eq 2727 then
        V:=[14,1,15,5,16,11,17,19,18,31,19,41,20,55,21,59,22,73,23,79,24,91];
        D:=[1,34,1,0,2,81/2,1,85/2,1,0,4,31,1,35,1,39,1,0,7,199/6,1,67/2,1,
            211/6,1,223/6,1,235/6,1,0,12,82/3,1,89/3,1,125/4,1,101/3,1,0,16,
            121/6,1,143/6,1,155/6,1,329/12,1,167/6,1,333/10,1,0,22,26,1,0,23,
            221/12,1,20,2,245/12,1,41/2,1,24,1,293/12,1,0,30,401/20,1,247/12,1,
            0,32,103/6,1,1117/60,1,241/12,1,81/4,1,41/2,1,0,37,907/60,1,233/15,
            1,0];
        I:=[428562,428557,428556,428566,428563,428558,428540,428254,428464,
            428304,428256,428525,428460,428344,428250,428258,428298,428468,
            428372,428287,428088,427918,428249,428294,428300,428264,428248,
            427796,427790,428036,427927,428246,427739,427774,427740,427764,
            427926,427925];
    elif hash1 eq 1042 and hash2 eq 1189 and hash3 eq 2144 then
        V:=[24,1,25,13,26,23,27,27,29,35];
        D:=[1,21,1,68/3,1,25,1,80/3,1,27,1,0,6,127/6,1,43/2,1,139/6,1,151/6,2,0,
            11,59/3,2,0,13,95/6,2,97/6,2,107/6,1,0,18,173/12,1,89/6,1,0];
        I:=[672646,672778,672623,672577,672618,672775,672674,672673,672526,
            672638,672616,672634,672585,672627,672481,672621,672661,672528,
            672628];
    elif hash1 eq 1230 and hash2 eq 2889 and hash3 eq 378 then
        V:=[23,1,24,5,25,17,26,27,27,37,28,49];
        D:=[1,49/2,1,0,2,19,1,62/3,3,21,3,68/3,1,23,3,0,13,103/6,3,115/6,2,39/2,
            1,127/6,2,0,21,52/3,1,53/3,4,18,1,59/3,1,0,28,185/12,1,63/4,1,95/6,
            3,97/6,2,33/2,1,0,36,71/5,1,57/4,1,0];
        I:=[672285,672320,672096,672309,672341,671873,672261,672308,672140,
            671864,672175,672414,671784,672144,672318,671691,671793,671792,
            671563,671678,671981,671193,671984,671991,672078,671967,671985,
            671715,671488,671533,671556,671749,671698,672217,672007,671553,
            671204];
    elif hash1 eq 1935 and hash2 eq 400 and hash3 eq 3174 then
        V:=[21,1,22,7,23,11,24,19,25,27,26,33,27,37,28,43,30,49];
        D:=[1,57/2,1,65/2,1,0,3,110/3,1,0,4,163/6,1,59/2,1,187/6,1,0,7,46/3,1,
            93/4,1,101/4,1,0,10,85/6,1,133/6,1,0,12,167/12,2,0,14,75/4,1,115/6,
            1,0,16,155/12,1,13,1,0,18,166/15,1,0];
        I:=[505388,505235,505153,505448,505541,505387,505461,505291,505469,
            505655,505046,505240,505471,504631,505353,505050,505649,505437];
    elif hash1 eq 1042 and hash2 eq 2256 and hash3 eq 434 then
        V:=[24,1,25,15,27,27,28,31,29,37,30,41];
        D:=[1,59/3,1,71/3,1,82/3,1,28,1,89/3,1,94/3,1,0,7,95/4,2,143/6,1,145/6,
            1,253/10,1,305/12,1,0,13,33/2,1,0,14,321/20,1,199/12,1,0,16,401/30,
            1,0,17,66/5,2,163/12,1,0];
        I:=[411577,411561,411643,412303,412300,411554,409506,411338,411631,
            411558,409491,411616,409495,409530,411617,411513,411548,411549,
            411374];
    elif hash1 eq 196 and hash2 eq 1966 and hash3 eq 3135 then
        V:=[26,1,27,5,28,13,29,21,30,31];
        D:=[1,59/3,2,0,3,95/6,1,97/6,2,109/6,1,0,7,15,3,49/3,2,50/3,2,0,14,
            173/12,1,29/2,3,89/6,4,197/12,1,0,23,155/12,1,13,1,0];
        I:=[241476,241504,241467,240587,241696,241778,240763,240794,241390,
            241300,241468,239032,241775,238980,241252,241253,241777,240762,
            241419,241486,241548,238058,239130,240201];
    elif hash1 eq 2734 and hash2 eq 209 and hash3 eq 2489 then
        V:=[25,1,26,7,27,15,28,19,30,27];
        D:=[1,49/2,1,53/2,2,0,4,23,3,74/3,2,25,1,0,10,127/6,2,0,12,59/3,1,20,1,
            64/3,1,0,15,50/3,2,0];
        I:=[666713,666530,666845,666393,666424,666576,666725,666745,666737,
            666705,666768,666427,666564,665325,665344,665348];
    elif hash1 eq 3345 and hash2 eq 336 and hash3 eq 856 then
        V:=[25,1,26,7,27,17,28,29,29,37];
        D:=[1,365/12,1,197/6,1,0,3,373/15,1,185/7,1,529/20,1,319/12,1,0,7,
            1357/60,1,321/14,1,277/12,1,1045/42,1,249/10,1,0,12,1147/60,1,
            293/15,1,317/15,1,0,15,725/42,1,0];
        I:=[508530,508908,509404,508557,509103,509126,509132,508554,509435,
            508732,508722,509088,509438,508745,508638];
    elif hash1 eq 901 and hash2 eq 2583 and hash3 eq 561 then
        V:=[20,1,21,11,22,19,24,27,25,37];
        D:=[1,27,1,86/3,1,29,1,31,1,0,5,139/6,1,151/6,1,163/6,1,0,8,59/3,1,70/3,
            2,71/3,1,0,12,56/3,1,99/5,1,239/12,1,111/5,1,0,16,33/2,1,101/6,1,0];
        I:=[252267,252314,252468,252469,252327,252312,252101,252266,252097,
            252103,252108,252323,252046,252047,251851,252252,252105];
    elif hash1 eq 1418 and hash2 eq 1048 and hash3 eq 941 then
        V:=[15,1,16,9,17,15,18,23,19,27,20,33,23,37,24,41];
        D:=[1,57/2,1,69/2,1,81/2,1,0,4,23,1,37,1,0,6,115/6,1,59/2,1,211/6,1,0,9,
            176/5,1,0,10,109/6,1,95/4,1,0,12,298/15,1,0,13,461/30,1,0,14,323/21,
            1,238/15,1,0];
        I:=[518709,518707,518716,518728,518599,518722,518600,518602,518423,
            518635,518591,518355,518432,518300,518288];
    elif hash1 eq 854 and hash2 eq 1772 and hash3 eq 449 then
        V:=[20,1,21,23,22,49,23,77,24,101,25,125,26,153,27,161];
        D:=[1,21,1,25,2,80/3,1,27,2,86/3,2,29,2,92/3,3,31,2,98/3,2,33,1,0,19,
            115/6,1,39/2,1,127/6,4,43/2,2,47/2,1,151/6,3,51/2,1,107/4,1,163/6,2,
            55/2,1,175/6,1,187/6,1,0,38,58/3,1,59/3,2,20,1,106/5,1,64/3,1,65/3,
            1,22,1,116/5,1,70/3,1,71/3,1,126/5,1,76/3,1,77/3,3,0,54,71/4,1,
            107/6,6,109/6,5,37/2,1,193/10,1,233/12,3,197/10,1,119/6,2,121/6,3,
            87/4,1,237/10,1,0,79,50/3,3,35/2,1,215/12,1,18,1,91/5,1,73/4,2,55/3,
            2,56/3,2,101/5,1,61/3,1,62/3,1,0,95,479/30,1,16,1,193/12,2,197/12,2,
            33/2,3,101/6,1,351/20,1,251/14,1,221/12,1,37/2,1,187/10,3,113/6,2,
            115/6,1,0,115,41/3,1,179/12,1,46/3,2,0,119,401/30,4,31/2,1,0];
        I:=[424835,423866,424111,424601,424198,424594,424512,424572,423867,
            423871,423884,423954,424701,423910,424740,423874,424821,424811,
            424765,424808,423842,424268,424411,424837,423883,424729,424557,
            422929,424516,424543,424464,423052,424728,424832,422828,424149,
            423606,424449,424684,424766,424692,422827,424489,423850,424459,
            423738,424129,423623,423562,422631,422854,423381,424379,424137,
            424225,424424,424545,424568,424569,424726,423925,424265,424297,
            424563,424589,424514,423597,423426,423596,424377,423610,424332,
            424819,423946,424292,424764,423537,422570,423553,423977,424456,
            422872,423567,423511,423619,422625,423441,423096,424281,424342,
            424352,423735,422924,423725,423585,424256,423535,423987,422841,
            423424,422622,423584,424736,424306,423734,422399,422402,422462,
            422619,423687,423721,422545,423067,423561,423642,423446,423437,
            423628,423455,423633,423634,423635,422486];
    elif hash1 eq 666 and hash2 eq 813 and hash3 eq 3247 then
        V:=[20,1,21,7,22,15,23,21,24,27,25,31,26,35,28,39];
        D:=[1,30,1,32,1,0,3,61/2,1,65/2,1,73/2,1,0,6,29,1,33,1,0,8,47/2,1,175/6,
            3,0,12,71/3,1,0,13,131/6,1,0,14,263/12,1,0,15,17,1,0];
        I:=[411509,411409,411550,411280,411417,411429,411551,411530,411412,
            411522,411525,411314,411426,409475,410145];
    elif hash1 eq 854 and hash2 eq 2673 and hash3 eq 3171 then
        V:=[20,1,21,5,22,19,23,39,24,53,25,61,26,67,27,85,28,111,29,115];
        D:=[1,37,1,0,2,55/2,1,175/6,1,187/6,1,67/2,1,139/4,1,211/6,1,0,8,76/3,1,
            82/3,1,83/3,1,28,1,146/5,1,117/4,1,89/3,1,30,1,94/3,1,0,17,131/6,1,
            237/10,1,145/6,2,103/4,2,277/10,1,167/6,1,0,25,215/12,1,55/3,1,62/3,
            1,0,28,33/2,2,35/2,1,0,31,15,1,47/3,1,329/20,1,253/15,1,17,1,52/3,2,
            94/5,1,227/12,1,0,40,27/2,1,291/20,1,439/30,1,44/3,1,59/4,1,91/6,1,
            461/30,4,31/2,1,331/20,1,697/42,1,997/60,1,67/4,1,0,55,66/5,2,0,57,
            697/60,1,0];
        I:=[424785,424109,424732,424474,423974,423498,424547,424373,424699,
            424476,423955,423014,422663,423051,424805,424362,424450,423586,
            422701,424522,423482,423531,423617,423536,422988,423351,423005,
            423513,424307,424364,423521,423973,422558,423704,423697,423093,
            424775,422401,423481,423595,423300,422992,423009,423548,424822,
            422883,423480,423693,423699,423989,422495,422514,422551,422425,
            423640,423672,423458];
    elif hash1 eq 1042 and hash2 eq 3344 and hash3 eq 2889 then
        V:=[22,1,23,5,24,13,25,23,26,29,27,35];
        D:=[1,27,1,0,2,115/6,1,39/2,2,139/6,1,0,6,53/3,1,18,1,77/4,1,59/3,1,0,
            10,97/6,1,177/10,1,0,12,16,1,81/5,1,0,14,29/2,2,89/6,2,0];
        I:=[249158,249397,249055,249165,248660,248964,248606,249107,248449,
            249160,248506,249173,248599,249051,249090,248503,249271];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 9 mod 47.
function cached_ids_9(P,hash1,hash2,hash3)
    if hash1 eq 1278 and hash2 eq 903 and hash3 eq 605 then
        V:=[24,1,25,5,26,9,27,23,28,27,29,43,30,65,31,73];
        D:=[1,82/3,1,0,2,145/6,2,0,4,55/3,1,263/12,1,111/5,1,67/3,1,68/3,1,73/3,
            1,0,10,611/30,2,0,12,199/12,1,52/3,1,53/3,1,277/15,1,283/15,1,19,1,
            308/15,1,0,19,451/30,1,181/12,1,91/6,1,461/30,1,185/12,1,31/2,1,
            95/6,1,50/3,1,521/30,2,37/2,1,0,30,787/60,1,53/4,1,91/6,1,0,33,73/6,
            1,0];
        I:=[414372,413998,414085,414052,414259,414062,414376,414304,414009,
            413666,414096,414238,414298,413999,413816,414020,414371,414026,
            414246,414364,414031,414065,413475,414375,414365,414296,414027,
            414427,413434,414292,414309,413579,413510];
    elif hash1 eq 808 and hash2 eq 546 and hash3 eq 1240 then
        V:=[24,1,25,11,26,23,27,31,28,35];
        D:=[1,62/3,1,21,1,68/3,2,23,3,0,8,35/2,1,75/4,1,115/6,2,39/2,1,127/6,1,
            0,14,69/4,1,52/3,1,53/3,1,0,17,95/6,1,0,18,44/3,1,0];
        I:=[672335,672330,671876,671881,671838,671945,672431,671782,671697,
            672095,672213,671839,671688,671777,671810,672114,672143,671540];
    elif hash1 eq 2406 and hash2 eq 244 and hash3 eq 860 then
        V:=[21,1,22,9,23,13,24,19,25,31,26,39,28,43,29,51,30,55];
        D:=[1,175/6,1,223/6,3,235/6,1,0,6,125/4,1,0,7,257/10,2,191/6,2,0,11,
            358/15,1,287/12,1,26,2,141/5,1,86/3,1,0,17,611/30,1,207/10,1,779/30,
            1,0,20,361/20,1,0,21,419/28,1,316/21,1,907/60,1,0,24,137/10,1,0,25,
            977/84,1,0];
        I:=[419467,419267,419348,419492,419395,419239,418987,419419,419837,
            419843,419842,419444,419279,419848,419304,419315,419445,419434,
            419058,419114,419002,418998,418977,419407,418999];
    elif hash1 eq 1231 and hash2 eq 985 and hash3 eq 1052 then
        V:=[22,1,23,7,24,21,25,27,27,35,28,41];
        D:=[1,19,1,25,2,0,4,127/6,2,43/2,1,91/4,1,139/6,2,151/6,1,163/6,1,0,12,
            16,1,59/3,1,0,14,107/6,2,109/6,1,257/12,1,0,18,89/6,1,197/12,1,0,20,
            46/3,1,0];
        I:=[249796,249723,249799,249789,249797,249804,249016,249266,249775,
            249017,248411,249788,249162,248269,249092,249103,248589,249746,
            248604,247524];
    elif hash1 eq 1889 and hash2 eq 2386 and hash3 eq 1603 then
        V:=[18,1,19,7,20,15,21,21,22,29,23,37,24,41,25,47,26,51];
        D:=[1,32,1,36,1,0,3,61/2,1,65/2,1,89/2,1,0,6,98/3,1,35,1,0,8,175/6,1,
            59/2,2,223/6,1,0,12,77/3,1,156/5,1,100/3,2,0,16,353/12,1,0,17,
            263/12,1,67/3,1,0,19,611/30,1,0,20,52/3,1,441/20,1,0];
        I:=[512161,512481,512158,512174,512480,511937,512235,511855,511838,
            511850,512435,512381,512315,511885,512209,512484,512437,512208,
            512331,511820,511315];
    elif hash1 eq 714 and hash2 eq 2568 and hash3 eq 3150 then
        V:=[24,1,25,9,26,21,27,27,28,35];
        D:=[1,19,1,23,1,25,1,0,4,103/6,1,35/2,1,115/6,1,127/6,1,43/2,1,0,9,53/3,
            2,59/3,1,0,12,85/6,1,95/6,1,97/6,1,0,15,43/3,1,44/3,1,0];
        I:=[241206,240929,239384,240545,240715,240706,239315,239386,239380,
            240675,240033,239941,239446,239667,241278,239732];
    elif hash1 eq 761 and hash2 eq 1347 and hash3 eq 2113 then
        V:=[12,1,13,5,14,13,15,21,16,35,17,43,18,63,19,89,20,99,21,111,22,129,
            23,143,24,153,25,159];
        D:=[1,44,1,0,2,81/2,1,85/2,1,93/2,2,0,6,122/3,1,128/3,2,47,1,0,10,199/6,
            1,223/6,2,75/2,1,155/4,1,235/6,1,259/6,1,0,17,117/4,1,101/3,2,118/3,
            1,0,21,353/12,1,179/6,2,181/6,3,377/12,1,191/6,3,193/6,1,203/6,1,
            143/4,1,215/6,2,0,36,388/15,1,311/12,1,26,1,105/4,1,79/3,1,28,1,
            359/12,1,91/3,1,478/15,1,161/5,1,97/3,1,431/12,1,0,48,671/30,1,
            269/12,1,731/30,1,791/30,1,0,52,449/20,1,271/12,1,275/12,1,23,1,
            70/3,1,0,57,56/3,1,379/20,1,19,1,229/12,3,39/2,1,68/3,1,23,1,277/12,
            1,0,67,394/21,1,113/6,1,401/21,1,77/4,1,1363/60,1,461/20,1,0,73,
            917/60,1,937/60,1,157/10,1,63/4,1,0,77,1313/84,1,110/7,1,0,79,
            367/30,2,0];
        I:=[429859,429822,429855,429818,429821,429820,429823,429898,429824,
            429720,429819,429909,429830,429776,429885,429738,429777,429742,
            429852,429744,429729,429692,429815,429693,429782,429871,429811,
            429850,429854,429873,429857,429816,429663,429748,429749,429691,
            429784,429827,429778,429779,429735,429649,429613,429701,429709,
            429699,429612,429706,429781,429700,429621,429624,429617,429662,
            429665,429722,429725,429661,429797,429743,429775,429809,429826,
            429601,429606,429605,429598,429581,429597,429596,429586,429585,
            429629,429615,429619,429646,429584,429583,429635,429643];
    elif hash1 eq 103 and hash2 eq 3135 and hash3 eq 2433 then
        V:=[31,1,33,7,34,21,35,33];
        D:=[1,207/10,2,671/30,1,0,4,451/30,1,185/12,1,50/3,1,17,2,511/30,1,
            521/30,1,0,11,787/60,1,163/12,1,907/60,1,76/5,1,61/4,2,0,17,27/2,1,
            0];
        I:=[390534,390541,390792,390562,390537,390619,390471,390473,390738,
            390542,390712,390470,390631,390601,390602,390630,390681];
    elif hash1 eq 244 and hash2 eq 376 and hash3 eq 2140 then
        V:=[23,1,24,5,25,11,26,19,27,27,28,39,29,45,30,49];
        D:=[1,115/6,1,0,2,58/3,1,85/4,1,0,4,33/2,1,209/12,1,107/6,1,0,7,238/15,
            1,50/3,1,109/5,1,0,10,79/6,1,27/2,1,491/30,1,37/2,1,241/12,1,0,15,
            155/12,1,223/15,1,0,17,541/42,1,0,18,173/15,2,0];
        I:=[503524,503065,503633,503473,503526,503341,503342,503439,502565,
            503657,503468,503066,502449,502404,503411,502975,501967,503321,
            503322];
    elif hash1 eq 1560 and hash2 eq 1630 and hash3 eq 1307 then
        V:=[22,1,23,5,24,11,25,19,26,29,27,37,28,45,32,51];
        D:=[1,94/3,1,0,2,133/6,1,145/6,1,0,4,67/3,1,74/3,1,113/4,1,0,7,75/4,1,
            207/10,1,125/6,1,671/30,1,0,11,278/15,1,223/12,1,251/12,1,0,14,91/6,
            1,521/30,1,107/6,1,0,17,256/15,1,87/5,1,0,19,136/15,1,0];
        I:=[513373,513399,513384,513405,513377,513343,513429,513374,513342,
            513378,513382,513404,513335,513402,513370,513368,513403,513397,
            513430];
    elif hash1 eq 573 and hash2 eq 2875 and hash3 eq 3157 then
        V:=[24,1,25,7,26,11,27,17,30,23,31,27,32,33];
        D:=[1,35,1,110/3,1,0,3,187/6,1,0,4,116/5,1,28,1,0,6,119/6,1,237/10,1,0,
            8,46/3,1,0,9,341/30,3,251/20,1,0,13,166/15,3,0];
        I:=[535275,535253,535255,535254,535276,535252,535265,535274,535245,
            535246,535271,534919,535244,535269,535270];
    elif hash1 eq 3440 and hash2 eq 2805 and hash3 eq 891 then
        V:=[19,1,20,7,21,19,22,27,23,31,24,39,25,43,26,53];
        D:=[1,61/2,1,69/2,1,0,3,25,1,29,1,92/3,2,31,1,98/3,1,0,9,99/4,1,151/6,1,
            63/2,1,0,12,65/3,1,0,13,63/4,1,107/6,1,233/12,1,0,16,56/3,1,0,17,
            279/20,1,14,1,193/12,1,37/2,1,0,21,175/12,1,0];
        I:=[510654,510347,510318,510319,510362,510365,510351,510352,510354,
            510283,510489,510386,509055,510409,509807,510368,509277,509067,
            509632,509208,510323];
    elif hash1 eq 1983 and hash2 eq 3230 and hash3 eq 951 then
        V:=[25,1,26,7,27,15,28,21,29,33,30,47];
        D:=[1,139/6,5,47/2,1,0,7,20,1,85/4,1,71/3,1,0,10,97/6,1,109/6,1,0,12,
            49/3,1,50/3,3,17,1,215/12,1,73/4,2,0,20,173/12,1,29/2,1,59/4,1,89/6,
            2,197/12,1,33/2,1,0,27,13,2,0];
        I:=[670437,670455,670560,670931,670967,670898,670807,670327,670009,
            670759,670480,670345,670195,670453,670545,670906,670361,669898,
            669915,669977,670339,669989,670360,670984,669927,669923,670281,
            670293];
    elif hash1 eq 2312 and hash2 eq 735 and hash3 eq 2031 then
        V:=[28,1,29,7,30,19,31,27,32,39];
        D:=[1,58/3,2,59/3,1,0,4,97/6,3,209/12,1,71/4,1,107/6,1,119/6,1,0,11,
            191/12,1,49/3,1,268/15,1,0,14,29/2,3,147/10,2,59/4,2,197/12,2,33/2,
            1,0,24,13,2,0];
        I:=[232005,233095,233729,233108,233955,234074,234200,233791,232118,
            233768,232130,233177,230580,230406,230542,233687,231544,231571,
            230405,230493,233548,233806,230047,233043,234081];
    elif hash1 eq 1654 and hash2 eq 1217 and hash3 eq 1448 then
        V:=[22,1,23,5,24,13,25,19,26,31,27,41,28,69];
        D:=[1,77/3,1,0,2,131/6,1,281/12,1,95/4,1,0,5,287/12,1,98/3,1,0,7,245/12,
            1,41/2,1,659/30,2,137/6,1,247/10,1,0,13,313/15,1,64/3,1,441/20,1,
            343/15,1,0,17,781/42,2,559/30,1,75/4,2,151/8,1,571/30,1,229/12,1,
            39/2,1,241/12,1,121/6,1,865/42,1,631/30,1,127/6,1,43/2,1,0,32,
            226/15,1,302/15,1,0];
        I:=[508805,508653,508607,508582,508832,508656,508591,508651,508544,
            508558,508504,508540,508634,508647,508662,508685,508546,508548,
            508633,508536,508573,508399,508341,508534,508347,508690,508686,
            508600,508676,508711,508706,508374,508708];
    elif hash1 eq 2453 and hash2 eq 2147 and hash3 eq 2741 then
        V:=[15,1,16,5,17,9,18,23,19,35,20,43,22,49,23,55,24,63];
        D:=[1,61/2,2,0,3,37,1,0,4,39/2,1,151/6,1,59/2,1,187/6,1,223/6,1,75/2,1,
            0,10,82/3,1,83/3,1,89/3,1,95/3,1,107/3,1,0,15,109/6,1,155/6,1,
            277/10,2,0,19,55/3,2,101/5,1,0,22,17,1,283/15,2,0,25,511/30,1,103/6,
            1,107/6,1,0,28,323/21,1,233/15,1,238/15,1,16,1,0];
        I:=[428409,428446,428404,428399,428406,428452,428437,428419,428435,
            428231,428408,428432,427814,427885,428421,428240,427810,427892,
            428455,428546,427730,428456,427717,427719,427908,427813,427716,
            427644,427639,427657,427641];
    elif hash1 eq 3346 and hash2 eq 74 and hash3 eq 930 then
        V:=[21,1,22,5,23,15,24,29,25,41,26,59,27,81,28,87,29,95];
        D:=[1,63/2,1,0,2,71/3,1,76/3,1,77/3,2,89/3,1,0,7,121/6,1,237/10,1,145/6,
            1,257/10,1,155/6,1,157/6,1,0,13,215/12,1,56/3,1,239/12,1,101/5,1,
            121/5,1,0,18,33/2,2,101/6,1,217/12,1,551/30,2,113/6,1,611/30,2,41/2,
            1,125/6,1,0,29,15,1,76/5,1,203/12,1,86/5,1,52/3,1,53/3,1,129/7,1,
            94/5,1,227/12,1,22,1,0,39,461/30,1,205/12,1,0,41,203/15,1,163/12,1,
            61/4,1,0,44,697/60,2,25/2,1,325/24,2,407/30,1,137/10,1,0];
        I:=[513896,513839,513827,513791,513820,513833,513867,513671,513866,
            513702,513794,513805,513773,513836,513897,513750,513709,513754,
            513850,513793,513687,513693,513753,513806,513655,513752,513299,
            513803,513712,513711,513914,513766,513718,513831,513420,513553,
            513436,513309,513742,513549,513767,513654,513400,513738,513756,
            513425,513306,513317,513308,513307];
    elif hash1 eq 902 and hash2 eq 33 and hash3 eq 1216 then
        V:=[23,1,24,11,25,21,26,33,27,39,28,43];
        D:=[1,39/2,3,43/2,3,91/4,1,139/6,1,0,9,53/3,1,18,1,58/3,1,106/5,1,0,13,
            95/6,3,97/6,1,177/10,1,109/6,2,131/6,1,0,21,16,1,50/3,1,0,23,227/14,
            1,0,24,66/5,2,40/3,1,0];
        I:=[418122,418123,418145,418173,418199,418211,418426,416597,418070,
            417445,416810,417300,417336,418445,418482,418157,417646,415069,
            415813,415675,416379,417590,415426,415540,416584,415527];
    elif hash1 eq 3252 and hash2 eq 3218 and hash3 eq 3379 then
        V:=[25,1,26,7,27,17,28,27];
        D:=[1,35/2,2,83/4,1,0,4,69/4,1,53/3,2,59/3,1,20,1,0,9,63/4,1,95/6,3,
            107/6,1,87/4,1,0,15,43/3,1,44/3,2,0];
        I:=[240638,241634,241147,240788,241664,241673,240868,240866,240437,
            240641,240870,241256,238536,238292,241235,240182,241120];
    elif hash1 eq 3346 and hash2 eq 78 and hash3 eq 2402 then
        V:=[21,1,22,9,23,19,24,25,25,35,26,45,27,53,28,59,29,63];
        D:=[1,163/6,1,55/2,1,175/6,2,0,5,71/3,1,77/3,1,109/4,1,117/4,1,0,9,
            237/10,1,155/6,1,0,11,239/12,1,61/3,1,322/15,1,121/5,1,0,15,551/30,
            1,37/2,1,187/10,1,611/30,1,0,19,253/15,2,203/12,1,386/21,1,0,23,
            451/30,1,697/42,1,0,25,163/12,1,0,26,117/10,1,0];
        I:=[423785,424163,423970,424020,424023,424103,423032,423984,423220,
            423134,424181,423782,422604,422397,423098,423160,423031,422365,
            422993,423038,424212,422364,422345,422378,423193,423156];
    elif hash1 eq 2547 and hash2 eq 3304 and hash3 eq 2917 then
        V:=[18,1,19,5,20,15,21,23,22,35,23,39,24,47];
        D:=[1,22,1,0,2,41/2,3,45/2,2,49/2,1,65/2,1,0,9,19,1,68/3,1,27,1,0,12,
            75/4,1,115/6,3,127/6,1,43/2,1,47/2,1,0,19,22,1,0,20,209/12,1,109/6,
            1,37/2,1,0,23,16,1,49/3,1,0];
        I:=[420620,420592,420593,420627,420420,420518,420330,420544,420268,
            420366,420313,419949,420084,420407,420502,420284,420554,420176,
            419651,420290,419633,420179,420581,419675];
    elif hash1 eq 1278 and hash2 eq 1244 and hash3 eq 2462 then
        V:=[21,1,22,7,23,15,24,25,27,33];
        D:=[1,155/6,1,119/4,1,0,3,131/5,1,79/3,1,80/3,1,0,6,269/12,1,45/2,1,
            137/6,1,779/30,1,0,10,313/15,1,21,1,64/3,1,0,13,421/30,1,85/6,1,
            63/4,1,0];
        I:=[425517,425525,425512,425526,425504,425522,425519,425511,425505,
            425514,425513,425501,425523,425508,425470];
    elif hash1 eq 1983 and hash2 eq 3226 and hash3 eq 1966 then
        V:=[21,1,22,7,23,15,24,23,25,27];
        D:=[1,45/2,2,49/2,1,0,4,62/3,1,21,3,23,1,0,9,115/6,2,39/2,1,127/6,1,0,
            13,53/3,4,0,17,97/6,1,0];
        I:=[673613,673747,673568,673563,673591,673657,673673,673431,673386,
            673592,673436,673534,673380,673408,673451,673653,673666];
    elif hash1 eq 3111 and hash2 eq 1827 and hash3 eq 546 then
        V:=[21,1,22,5,23,9,24,15,25,27,27,31];
        D:=[1,127/6,2,0,3,26,1,0,4,95/4,1,155/6,2,0,7,239/12,1,81/4,1,61/3,2,
            62/3,2,287/12,2,0,15,659/30,1,0,16,371/20,1,0];
        I:=[513783,513821,513782,513785,513595,513610,513795,513787,513784,
            513832,513777,513781,513426,513668,513423,513302];
    elif hash1 eq 3440 and hash2 eq 2784 and hash3 eq 2879 then
        V:=[19,1,20,5,21,19,22,31,23,43,24,51];
        D:=[1,45/2,2,0,3,62/3,1,21,3,23,2,74/3,1,25,1,29,1,0,12,115/6,1,39/2,2,
            83/4,1,127/6,2,47/2,2,0,20,53/3,1,58/3,1,59/3,2,20,1,64/3,1,0,26,
            33/2,1,177/10,1,109/6,1,0,29,49/3,1,50/3,1,0];
        I:=[424788,424851,424833,424472,424812,424843,424387,424561,424667,
            424850,423911,424767,424381,424517,424515,424466,424781,423449,
            424030,424784,424742,423604,423629,423679,424260,423587,423566,
            424067,423439,424446];
    elif hash1 eq 573 and hash2 eq 1581 and hash3 eq 2154 then
        V:=[14,1,15,9,16,19,17,29,18,45,19,51,20,65,21,77,22,87,23,93,25,103,26,
            107];
        D:=[1,32,1,42,1,44,1,0,4,69/2,1,77/2,2,85/2,1,93/2,1,0,9,92/3,1,110/3,1,
            39,1,128/3,1,0,13,91/4,1,175/6,1,187/6,1,199/6,1,155/4,1,235/6,1,
            259/6,1,0,20,146/5,1,94/3,3,0,24,143/6,1,329/12,1,377/12,1,191/6,1,
            193/6,1,215/6,1,0,30,263/12,1,358/15,1,287/12,1,26,1,105/4,1,0,35,
            167/10,1,217/12,1,241/12,1,41/2,1,0,39,248/15,1,283/15,1,0,41,
            509/28,1,56/3,1,229/12,2,277/12,1,0,46,63/4,1,0,47,181/15,1,0];
        I:=[428969,429019,429035,429017,428955,429021,428991,428873,428967,
            429037,428915,428900,428872,429011,428857,428909,428929,428885,
            428752,428679,428807,429008,429033,428661,428762,428941,428656,
            428919,428724,428727,428660,428655,428648,428723,428711,428652,
            428811,428876,428778,428760,428588,428597,428613,428651,428601,
            428600,428767];
    elif hash1 eq 3346 and hash2 eq 1818 and hash3 eq 3054 then
        V:=[18,1,19,7,20,11,21,19,22,27,23,37,24,43,25,47,26,53];
        D:=[1,80/3,1,37,1,0,3,63/2,1,0,4,93/4,1,88/3,1,125/4,1,0,7,121/6,1,
            133/6,1,155/6,1,0,10,215/12,1,73/4,1,20,1,119/5,1,0,14,113/6,2,
            611/30,1,0,17,283/15,1,0,18,83/6,1,67/4,2,0,21,419/28,1,91/6,1,
            183/10,1,0];
        I:=[515540,515255,515302,515300,514325,514311,515270,515301,514312,
            514884,514847,514246,514405,514310,514313,514885,514383,514822,
            514175,514241,514023,514026,513957];
    elif hash1 eq 3205 and hash2 eq 1396 and hash3 eq 3366 then
        V:=[27,1,28,5,29,19,30,35,31,43,32,55];
        D:=[1,139/6,1,0,2,53/3,1,58/3,1,59/3,1,20,2,65/3,1,93/4,1,0,9,97/6,1,
            33/2,1,107/6,1,109/6,2,37/2,1,119/6,2,121/6,1,0,18,16,1,49/3,2,
            215/12,1,0,22,173/12,2,29/2,1,147/10,1,89/6,1,193/12,1,0,28,193/15,
            1,40/3,1,223/15,1,0];
        I:=[667600,667100,667244,667620,665969,667023,666487,666439,666900,
            667277,667226,667225,667558,666913,665382,666468,666915,667080,
            666347,667607,666686,665712,666668,666374,666675,667594,665333,
            665739,666690,665073];
    elif hash1 eq 197 and hash2 eq 1234 and hash3 eq 2877 then
        V:=[21,1,22,11,23,23,24,35,25,43];
        D:=[1,45/2,1,53/2,2,57/2,4,65/2,2,0,10,21,1,68/3,1,23,1,25,3,80/3,1,0,
            17,127/6,1,43/2,1,139/6,3,151/6,1,163/6,1,0,24,59/3,1,64/3,1,65/3,1,
            0,27,119/6,2,0];
        I:=[250906,250760,251094,250732,250741,251096,251365,250745,250915,
            250918,250734,251097,250853,251098,251352,250245,251186,250405,
            250180,250376,251129,250090,250362,250409,250347,250375,250258,
            250339];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 10 mod 47.
function cached_ids_10(P,hash1,hash2,hash3)
    if hash1 eq 3394 and hash2 eq 727 and hash3 eq 1617 then
        V:=[24,1,25,5,26,15,27,21,28,27,29,33,30,37,31,43];
        D:=[1,31,1,0,2,127/6,1,43/2,1,151/6,2,51/2,1,0,7,85/4,1,65/3,1,0,9,71/4,
            1,121/6,1,0,11,49/3,2,56/3,1,0,14,77/6,1,0,15,151/12,2,38/3,1,0,18,
            331/30,1,0];
        I:=[406434,406131,406328,406153,406154,405915,406183,406280,405953,
            405963,406030,406350,406142,405457,405952,406279,405959,405981];
    elif hash1 eq 10 and hash2 eq 901 and hash3 eq 1083 then
        V:=[28,1,30,5,31,15,32,25,33,31];
        D:=[1,24,1,0,2,167/12,1,14,1,43/3,1,65/4,1,0,6,79/6,2,29/2,1,89/6,1,
            103/6,1,0,11,151/12,1,13,3,0,15,161/12,1,0];
        I:=[394732,394065,394553,394850,393431,394601,394610,394547,394064,
            391972,394867,393848,394716,394724,390548];
    elif hash1 eq 2736 and hash2 eq 1996 and hash3 eq 2977 then
        V:=[26,1,27,5,28,13,29,17,30,27,32,33];
        D:=[1,25,1,0,2,115/6,1,39/2,2,127/6,1,0,6,52/3,1,0,7,63/4,1,97/6,2,33/2,
            1,109/6,3,0,14,14,1,43/3,2,0,17,34/3,2,0];
        I:=[667276,667742,666899,667718,666243,666054,666326,666186,667763,
            666068,666552,666553,666685,665522,666122,667579,665756,665963];
    elif hash1 eq 2078 and hash2 eq 1605 and hash3 eq 1270 then
        V:=[20,1,21,7,22,13,23,21,24,29,25,35,26,47];
        D:=[1,31,1,33,1,0,3,55/2,2,115/4,1,0,6,20,1,24,1,26,1,0,9,109/6,1,79/4,
            1,95/4,1,0,12,56/3,1,19,1,0,14,197/12,2,33/2,2,101/6,1,221/12,1,
            119/6,1,0,21,223/15,1,47/3,2,0];
        I:=[422325,422321,422146,422328,421825,422324,422112,422042,421837,
            421812,421224,422183,422326,421793,421811,422162,422177,422143,
            421221,421226,421785,421839,422147];
    elif hash1 eq 2172 and hash2 eq 2752 and hash3 eq 2325 then
        V:=[28,1,29,11,31,19,33,27];
        D:=[1,93/4,1,70/3,1,101/4,1,76/3,1,0,5,121/6,1,237/10,1,143/6,1,0,8,
            167/10,1,101/6,3,103/6,1,0,13,38/3,1,157/12,1,79/6,2,0];
        I:=[666804,666411,665375,666302,666833,665414,665415,666801,665337,
            666305,666453,666835,666024,666597,666310,666577];
    elif hash1 eq 104 and hash2 eq 249 and hash3 eq 3118 then
        V:=[20,1,21,5,22,19,23,35,24,49,25,57,26,61,27,71,28,77,29,81];
        D:=[1,98/3,1,0,2,151/6,1,175/6,1,123/4,1,187/6,2,67/2,2,223/6,2,0,11,
            65/3,1,76/3,1,109/4,1,82/3,2,146/5,1,88/3,1,125/4,1,0,19,79/4,1,
            121/6,1,257/12,1,237/10,1,145/6,1,53/2,1,0,25,298/15,1,61/3,1,73/3,
            1,0,28,551/30,1,0,29,44/3,1,253/15,1,369/20,1,77/4,2,0,34,397/28,1,
            50/3,1,0,36,61/4,1,0,37,697/60,1,0];
        I:=[420048,419261,419370,419526,420076,420115,419388,420170,419752,
            420216,419412,420168,419868,419736,419819,419523,419272,419150,
            419307,419487,419223,419489,419312,420180,419273,419483,419406,
            419211,419400,419732,418990,419091,419856,419054,418966,418965,
            419605];
    elif hash1 eq 1326 and hash2 eq 154 and hash3 eq 3235 then
        V:=[22,1,23,5,24,13,25,25,26,37,27,47];
        D:=[1,23,1,0,2,39/2,1,127/6,1,43/2,2,0,6,52/3,1,53/3,8,58/3,1,59/3,2,22,
            1,0,19,97/6,1,177/10,1,71/4,1,107/6,2,109/6,2,0,26,191/12,1,215/12,
            3,18,1,56/3,2,0,33,173/12,1,29/2,1,0];
        I:=[249665,248865,249441,249414,249667,249674,248961,249218,249369,
            249652,249709,249733,249744,249754,249311,249210,249444,248402,
            249354,248309,248468,249088,249217,248019,248342,248037,248077,
            248225,248226,248234,247687,247700,248036,248157];
    elif hash1 eq 2125 and hash2 eq 679 and hash3 eq 205 then
        V:=[31,1,33,5,34,15];
        D:=[1,97/6,1,0,2,79/6,1,29/2,2,147/10,1,89/6,1,0,7,13,5,40/3,1,29/2,2,
            223/15,2,0];
        I:=[658819,658465,657035,658522,658254,656429,655331,656232,656428,
            658330,658664,656018,655160,655317,655327,655330];
    elif hash1 eq 2595 and hash2 eq 1172 and hash3 eq 993 then
        V:=[20,1,21,7,22,19,23,41,24,51,25,67,26,77,27,95,28,103,29,107];
        D:=[1,31,1,37,2,0,4,163/6,1,175/6,2,63/2,1,199/6,1,223/6,1,0,10,76/3,1,
            77/3,1,26,1,82/3,2,83/3,3,146/5,1,117/4,1,156/5,1,125/4,1,95/3,2,0,
            24,143/6,1,145/6,1,157/6,1,169/6,1,0,28,55/3,1,56/3,1,101/5,1,61/3,
            3,62/3,1,43/2,1,263/12,1,0,37,551/30,1,221/12,2,113/6,1,125/6,1,0,
            42,175/12,1,203/12,1,17,2,277/15,1,223/12,1,283/15,1,227/12,3,96/5,
            1,0,53,181/12,1,91/6,3,67/4,1,0,58,46/3,1,0,59,697/60,1,117/10,2,0];
        I:=[423994,424213,424389,423821,423858,424046,424015,424298,423678,
            423347,422903,422833,422864,423815,423971,424279,424461,422857,
            423491,423068,422695,422891,423247,424220,423026,422836,422987,
            422983,422866,423010,423357,423444,424278,424217,422577,423412,
            422957,423373,423743,423746,422973,422974,423375,423349,423359,
            422515,422376,423069,422617,423050,423066,422698,422755,424194,
            424294,424320,422424,422429,423377,423398,423408];
    elif hash1 eq 10 and hash2 eq 757 and hash3 eq 2342 then
        V:=[19,1,20,9,21,29,22,49,23,73,24,91,25,111,26,133,27,141,28,145];
        D:=[1,49/2,2,57/2,3,61/2,1,0,7,74/3,1,25,5,80/3,1,27,4,86/3,1,29,1,92/3,
            3,31,1,33,1,0,25,127/6,2,43/2,3,139/6,1,47/2,1,151/6,4,51/2,1,163/6,
            2,55/2,3,59/2,1,0,43,58/3,1,59/3,2,65/3,1,22,1,116/5,1,93/4,1,71/3,
            2,24,1,126/5,1,77/3,2,82/3,1,0,57,97/6,1,71/4,1,107/6,4,109/6,2,
            119/6,4,121/6,4,41/2,1,103/4,1,0,75,238/15,1,49/3,1,50/3,1,215/12,1,
            91/5,2,55/3,2,101/5,1,61/3,1,62/3,1,0,86,173/12,1,29/2,1,89/6,1,
            91/6,1,479/30,1,491/30,2,33/2,4,101/6,1,37/2,1,439/20,1,0,100,15,1,
            46/3,1,97/6,1,0,103,461/30,1,0,104,178/15,3,0];
        I:=[424040,424148,424104,424123,424170,424405,424142,423813,423820,
            423919,424061,424135,424039,423817,423964,424060,424264,423882,
            424036,423770,423857,424355,424470,423847,423008,423836,423822,
            423870,424094,423837,424237,422963,423845,424087,424134,424269,
            422922,423891,423541,423811,423816,424236,424112,424132,424270,
            422982,423918,422951,423445,422877,423934,422773,422867,422832,
            423219,422927,423137,424238,423843,423923,424133,424266,423895,
            424048,422914,423001,423838,423967,422865,422908,423907,424422,
            423255,422565,423210,422868,423324,423021,422894,423017,423599,
            423962,423115,423603,423916,422967,423174,424089,424016,422612,
            422962,423150,422916,423353,423933,424226,424263,422935,422440,
            423854,422757,422461,422391,422932,423253,423474];
    elif hash1 eq 2078 and hash2 eq 3197 and hash3 eq 3312 then
        V:=[17,1,18,5,19,9,20,15,21,25,22,33,23,37,24,47,25,53,27,57];
        D:=[1,73/2,1,0,2,45,1,0,3,175/6,1,211/6,1,0,5,88/3,1,89/3,1,95/3,1,
            113/3,1,0,9,155/6,1,167/6,1,179/6,1,0,12,359/12,1,0,13,551/30,2,
            113/6,1,269/12,2,313/12,1,0,19,253/15,1,247/12,1,0,21,451/20,1,0,22,
            407/30,1,137/10,2,0];
        I:=[514683,514664,514665,514584,514590,514723,514688,514853,514703,
            514695,514666,514583,514643,514731,514673,514727,514761,514704,
            514640,514525,514137,514759,514632,514649];
    elif hash1 eq 1138 and hash2 eq 2054 and hash3 eq 2863 then
        V:=[20,1,21,13,22,23,23,33,24,43,25,59,26,73];
        D:=[1,21,1,74/3,1,27,1,86/3,1,35,1,0,6,115/6,1,127/6,1,43/2,1,151/6,2,0,
            11,18,2,64/3,1,77/3,1,26,1,0,16,107/6,3,119/6,1,217/10,1,155/6,1,0,
            22,65/4,1,50/3,1,268/15,1,215/12,1,55/3,1,56/3,2,239/12,1,0,30,89/6,
            1,18,1,217/12,1,37/2,1,115/6,2,265/12,1,0,37,29/2,1,218/15,2,109/6,
            1,0];
        I:=[424606,424136,424595,423834,423205,424604,424113,424150,423899,
            423959,423965,424005,423859,422829,423072,423873,424678,424712,
            423637,422624,422406,423655,423221,423186,422842,423392,422885,
            423650,422849,423291,423525,422840,422460,422613,422643,422341,
            423649,423288,423293,422620];
    elif hash1 eq 3300 and hash2 eq 1165 and hash3 eq 1118 then
        V:=[25,1,26,7,27,15,28,19,29,33,30,37];
        D:=[1,139/6,2,47/2,4,0,7,52/3,1,18,1,71/3,1,0,10,71/4,1,0,11,57/4,1,
            65/4,1,49/3,1,50/3,3,17,1,91/5,1,0,19,431/30,2,0,21,151/12,1,0];
        I:=[239209,239594,240932,241020,241031,241033,239484,240189,239273,
            239462,240233,239325,241041,239394,240518,241051,241021,239713,
            239453,239740,240604];
    elif hash1 eq 950 and hash2 eq 2058 and hash3 eq 1131 then
        V:=[25,1,26,15,27,19,29,31,30,37,31,41];
        D:=[1,103/6,1,115/6,1,39/2,1,127/6,2,91/4,1,51/2,1,0,8,69/4,1,0,9,
            185/12,1,97/6,1,33/2,1,107/6,1,87/4,1,0,14,79/6,3,89/6,1,0,18,
            249/20,1,0,19,341/30,2,0];
        I:=[400218,398760,399127,399280,400152,398786,399494,399609,400209,
            399384,399371,400525,397239,399140,399298,400441,399352,398599,
            398805,399805];
    elif hash1 eq 1467 and hash2 eq 2239 and hash3 eq 543 then
        V:=[23,1,24,5,25,9,26,21,27,29,29,35];
        D:=[1,139/6,1,0,2,26,1,0,3,121/6,1,217/10,1,131/6,2,133/6,2,103/4,1,0,
            10,55/3,3,56/3,1,68/3,1,0,15,491/30,1,33/2,1,0,17,137/10,1,181/12,1,
            0];
        I:=[246633,246733,246648,246713,246717,246791,246714,246834,246858,
            246708,246715,246790,246705,246529,246827,246647,246830,246470];
    elif hash1 eq 1655 and hash2 eq 2997 and hash3 eq 2016 then
        V:=[21,1,22,13,23,29,24,49,25,59,26,75,27,85];
        D:=[1,115/6,2,83/4,1,127/6,1,59/2,1,211/6,1,0,7,77/4,1,58/3,1,59/3,3,
            85/4,1,126/5,1,101/4,1,77/3,2,0,17,95/6,1,173/10,1,209/12,1,109/6,1,
            193/10,1,233/12,2,79/4,2,121/6,2,145/6,1,0,29,238/15,1,50/3,1,
            268/15,2,111/5,1,0,34,89/6,1,197/12,2,351/20,1,251/14,1,37/2,1,75/4,
            1,223/10,1,0,42,44/3,1,46/3,1,115/7,1,211/12,1,0,46,401/30,1,181/12,
            1,0];
        I:=[510762,510763,510267,510276,510515,510748,510661,510703,510699,
            510754,510761,509736,510132,509119,510379,510450,510759,509813,
            510505,510719,509637,510524,510663,510602,510611,510727,510729,
            509674,509842,510424,509799,510014,509634,510322,509714,510495,
            510201,508841,509356,509355,508751,510651,510566,508839,509279,
            509890,508772];
    elif hash1 eq 2877 and hash2 eq 1429 and hash3 eq 2554 then
        V:=[24,1,26,5,27,11,28,25,29,31,30,39,31,49,33,55,34,61];
        D:=[1,98/3,1,0,2,71/3,1,101/4,1,0,4,119/6,1,133/6,1,237/10,1,167/6,1,
            179/6,1,191/6,1,0,10,239/12,1,358/15,1,0,12,491/30,1,399/20,1,
            245/12,1,0,15,321/20,1,253/15,2,86/5,1,369/20,1,0,20,451/30,1,
            511/30,1,0,22,119/10,1,361/30,2,0,25,809/84,2,193/20,1,0];
        I:=[534979,534953,534956,534954,534978,534984,534976,534975,534969,
            534926,534990,534916,534966,534965,534927,534967,534972,534992,
            534971,534986,534988,534946,534963,534985,534937,534973,534918];
    elif hash1 eq 1749 and hash2 eq 466 and hash3 eq 37 then
        V:=[17,1,18,5,19,17,20,27,21,43,22,53,23,65,24,79];
        D:=[1,65/2,1,0,2,74/3,2,25,1,86/3,1,92/3,1,31,1,0,8,139/6,1,51/2,1,
            175/6,3,187/6,5,0,18,65/3,2,70/3,2,126/5,1,76/3,1,77/3,1,95/3,2,32,
            2,0,29,131/6,1,237/10,1,95/4,2,49/2,1,0,34,18,1,239/12,1,61/3,1,
            263/12,1,26,1,0,39,33/2,1,101/6,1,18,1,217/12,1,551/30,3,113/6,1,0,
            47,227/12,1,19,1,58/3,2,59/3,2,133/6,1,0];
        I:=[427597,427539,427596,427598,427313,427557,427526,427534,427258,
            427311,427321,427449,426591,426894,427315,427522,427600,427319,
            427556,427317,427495,426890,427558,426896,426480,427123,426903,
            427267,426804,426803,427222,427307,426902,427554,426777,427541,
            426629,426463,426778,426717,426835,426789,426782,426906,426965,
            426941,426553,426555,426558,426681,426568,426753,426427];
    elif hash1 eq 2313 and hash2 eq 1907 and hash3 eq 1744 then
        V:=[27,1,29,5,30,9,31,21,32,25,33,37];
        D:=[1,43/2,1,0,2,29/2,1,0,3,14,1,44/3,1,15,1,81/5,1,50/3,1,0,8,77/6,1,0,
            9,73/6,1,188/15,1,155/12,1,13,3,212/15,1,0,16,341/30,1,0];
        I:=[383542,384291,385141,385059,384343,382057,383677,384546,384799,
            384239,382115,385060,385108,385146,378790,384457];
    elif hash1 eq 1373 and hash2 eq 420 and hash3 eq 127 then
        V:=[18,1,19,7,20,15,21,23,22,31,23,37,24,41,25,45];
        D:=[1,28,1,32,1,0,3,57/2,2,65/2,2,69/2,1,0,8,92/3,2,31,1,33,1,0,12,43/2,
            1,163/6,2,211/6,1,0,16,65/3,1,70/3,1,0,18,119/6,1,0,19,61/3,1,0,20,
            269/12,1,0];
        I:=[420318,420277,420319,420329,420074,420296,420380,420043,420375,
            420346,420297,420340,419564,419805,419543,419701,419636,419437,
            419846,419030];
    elif hash1 eq 1467 and hash2 eq 908 and hash3 eq 1365 then
        V:=[27,1,28,5,29,13,30,19,31,31];
        D:=[1,39/2,1,0,2,47/3,1,16,2,53/3,3,0,8,95/6,2,107/6,1,0,11,167/12,1,
            57/4,2,43/3,8,44/3,3,81/5,1,0,26,51/4,1,77/6,3,0];
        I:=[666961,667546,667221,667821,666964,667039,667766,666916,667661,
            666365,667053,666722,666825,666506,666700,666701,666860,666882,
            666919,667027,667911,666818,666967,666980,665436,666757,666208,
            666526,666684];
    elif hash1 eq 1467 and hash2 eq 1767 and hash3 eq 258 then
        V:=[13,1,14,9,15,21,16,35,17,57,18,75,19,103,20,125,21,149,22,177,23,
            209,24,227,25,241];
        D:=[1,65/2,1,81/2,1,85/2,1,0,4,80/3,1,116/3,1,39,1,128/3,1,43,1,0,9,
            187/6,1,63/2,1,131/4,1,147/4,1,79/2,1,247/6,1,0,15,124/5,1,77/3,1,
            89/3,2,95/3,2,32,1,100/3,1,38,1,157/4,1,119/3,3,173/4,1,0,29,157/6,
            1,317/10,1,191/6,1,65/2,1,203/6,1,357/10,1,397/10,1,239/6,4,0,40,
            263/12,1,388/15,1,131/5,1,80/3,1,85/3,1,30,1,478/15,2,383/12,3,
            161/5,1,97/3,3,98/3,1,431/12,1,479/12,1,0,58,245/12,1,659/30,1,22,1,
            265/12,2,479/20,1,24,1,791/30,4,317/12,1,53/2,1,447/14,1,0,72,52/3,
            1,369/20,1,223/12,1,19,1,59/3,1,401/20,1,121/6,1,21,1,506/21,1,
            185/7,1,319/12,1,0,83,511/30,2,205/12,1,1117/60,1,39/2,2,119/6,1,
            201/10,1,1213/60,2,209/10,2,631/30,4,621/28,1,89/4,1,139/6,2,
            1453/60,1,0,104,233/15,1,638/35,1,274/15,1,183/10,1,94/5,1,286/15,2,
            77/4,1,389/20,1,39/2,1,59/3,2,298/15,3,239/12,1,778/35,1,334/15,1,
            347/15,1,0,123,47/3,1,1415/84,1,1013/60,1,183/10,1,55/3,1,367/20,1,
            94/5,1,227/12,1,0,131,1313/84,1,95/6,1,826/45,1,677/36,1,113/6,1,
            569/30,1,0,137,243/20,2,0];
        I:=[519275,519179,519279,519177,519278,519276,519260,519196,519288,
            519195,519217,519220,519176,519302,519044,519301,519149,519293,
            519212,519272,519218,519150,519006,519216,519050,519170,519259,
            519037,519154,519004,518960,519213,519284,519088,519039,518962,
            518983,519036,519146,519203,519130,519058,519155,519148,518959,
            518943,519063,518961,519205,519223,519003,519159,519160,519239,
            519085,518988,518922,519025,519043,519040,518938,519030,518942,
            518950,518992,519038,519106,519138,518991,518969,518930,519018,
            519074,519124,519285,519153,519042,519169,519263,518887,518916,
            518872,519012,519016,519024,518993,518954,518967,518964,519041,
            519045,519053,519009,519075,519010,519020,519066,519126,518870,
            518858,518842,518843,518855,519049,518879,518999,519002,518886,
            519022,519073,518955,518979,518957,518968,518981,518972,519026,
            519048,518974,518854,518852,518884,519029,518911,518896,518934,
            518958,518944,518865,518863,518882,518873,518840,518850,518844,
            518846,518987,518996];
    elif hash1 eq 57 and hash2 eq 885 and hash3 eq 2280 then
        V:=[27,1,28,9,29,23,30,39,31,45,33,49,34,55];
        D:=[1,151/6,1,55/2,1,175/6,1,0,4,85/4,1,64/3,1,65/3,2,71/3,1,77/3,2,
            82/3,1,0,12,197/10,1,79/4,1,119/6,2,121/6,2,87/4,1,133/6,1,95/4,2,0,
            22,55/3,1,101/5,1,0,24,197/12,1,0,25,13,1,79/6,2,0,28,239/21,1,0];
        I:=[668619,668575,668592,668139,668567,668006,668599,668626,668091,
            668489,668574,668063,668587,668624,668664,668005,668594,668794,
            668576,667982,668549,668600,667935,667956,668518,668593,668897,
            667965];
    elif hash1 eq 1937 and hash2 eq 2834 and hash3 eq 1236 then
        V:=[18,1,19,5,20,13,21,25,23,29,24,35,25,43];
        D:=[1,74/3,1,0,2,163/6,2,175/6,2,187/6,1,0,7,71/3,2,24,1,76/3,1,136/5,1,
            83/3,1,0,13,109/6,1,0,14,101/6,4,187/10,2,0,20,253/15,1,17,1,53/3,1,
            0,23,461/30,1,95/6,1,0];
        I:=[426330,426336,426381,426344,426390,425926,426332,426378,426362,
            426132,425977,425959,426386,426311,426369,426372,426385,425697,
            425711,425885,425874,425699,425646,425644];
    elif hash1 eq 2360 and hash2 eq 28 and hash3 eq 2627 then
        V:=[24,1,25,7,26,15,27,27,28,37,30,41,31,47];
        D:=[1,83/3,1,32,1,0,3,133/6,2,143/6,1,103/4,2,0,8,239/12,1,101/5,1,
            263/12,2,67/3,2,73/3,1,0,15,217/12,1,221/12,1,37/2,1,241/12,1,0,19,
            277/15,1,0,20,203/15,1,41/3,1,0,22,361/30,1,0];
        I:=[248381,248791,248388,248399,248905,247806,248891,249010,248208,
            248684,248859,248698,248707,248882,249037,249042,247724,248687,
            247584,248091,248111,247505];
    elif hash1 eq 950 and hash2 eq 289 and hash3 eq 3113 then
        V:=[17,1,18,13,19,23,20,29,21,33,22,37,23,45];
        D:=[1,57/2,1,61/2,1,69/2,1,73/2,1,77/2,1,0,6,86/3,3,29,3,92/3,1,31,1,0,
            14,175/6,1,187/6,1,0,16,71/3,3,0,19,131/6,1,0,20,263/12,1,22,1,67/3,
            2,0,24,217/12,1,0];
        I:=[254690,254699,254693,254734,254733,254667,254702,254714,254685,
            254700,254707,254695,254691,254649,254668,254686,254698,254709,
            254666,254647,254651,254648,254655,254653];
    elif hash1 eq 2642 and hash2 eq 788 and hash3 eq 3205 then
        V:=[25,1,26,11,27,17,28,23,29,33,30,43];
        D:=[1,127/6,1,43/2,1,139/6,2,47/2,1,0,6,58/3,1,59/3,2,0,9,97/6,1,107/6,
            2,0,12,81/5,1,65/4,1,49/3,3,50/3,2,0,19,173/12,2,29/2,1,89/6,9,91/6,
            3,0,34,13,3,0];
        I:=[671037,671070,670609,670941,670808,671064,670548,670929,670703,
            670528,671062,670298,669921,670602,670614,670943,670112,670683,
            670346,670495,670627,670247,670256,670259,670348,670613,670963,
            670974,671050,671069,670117,670539,670682,670280,670297,670340];
    elif hash1 eq 2971 and hash2 eq 687 and hash3 eq 721 then
        V:=[25,1,26,7,27,13,28,25,29,33,30,55];
        D:=[1,35/2,1,43/2,1,0,3,58/3,3,59/3,2,0,8,95/6,1,97/6,2,33/2,1,107/6,2,
            119/6,1,0,15,44/3,1,49/3,2,268/15,1,0,19,79/6,1,419/30,1,14,1,
            431/30,3,29/2,2,147/10,2,89/6,3,91/6,3,197/12,1,33/2,1,0,37,13,2,0];
        I:=[245741,244747,244793,245443,245481,244456,245444,244763,244726,
            244779,244720,244698,244699,244457,244101,244725,245614,243359,
            244207,243007,244712,244419,244517,244518,244073,244661,243486,
            243488,244502,245569,245593,243456,244421,245665,242137,243673,
            243358,244447];
    elif hash1 eq 2454 and hash2 eq 1904 and hash3 eq 1250 then
        V:=[18,1,19,5,20,9,21,15,22,23,24,27,25,31,26,35,28,39];
        D:=[1,110/3,1,0,2,187/6,1,0,3,117/4,2,119/3,2,0,7,257/10,1,157/6,1,
            317/10,1,0,10,383/12,1,0,11,223/12,2,0,13,511/30,1,0,14,113/6,1,0,
            15,707/60,1,0];
        I:=[425194,425134,425034,425039,424892,424897,425016,425313,424886,
            424875,424899,425033,425021,424860,425005];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 11 mod 47.
function cached_ids_11(P,hash1,hash2,hash3)
    if hash1 eq 2643 and hash2 eq 1361 and hash3 eq 852 then
        V:=[22,1,23,7,24,11,25,23,26,27,27,31];
        D:=[1,19,3,68/3,1,0,5,103/6,1,0,6,16,1,69/4,2,52/3,2,20,1,64/3,1,0,13,
            95/6,1,0,14,49/3,2,0,16,91/6,1,0];
        I:=[418764,418776,418782,418106,418631,418635,418109,418626,418553,
            418785,418230,416183,417997,417003,417401,415409];
    elif hash1 eq 1656 and hash2 eq 3275 and hash3 eq 1745 then
        V:=[24,1,25,5,26,13,27,25,28,35,29,43,30,49,31,57,32,65];
        D:=[1,95/3,1,0,2,145/6,1,277/10,1,205/6,1,0,5,328/15,1,111/5,1,358/15,1,
            131/5,1,79/3,1,0,10,45/2,1,731/30,1,149/6,1,519/20,1,0,14,56/3,1,
            96/5,1,58/3,1,0,17,59/4,1,571/30,1,0,19,1027/60,1,268/15,1,215/12,1,
            0,22,137/10,2,421/30,2,157/10,1,0,27,289/21,1,827/60,1,0];
        I:=[501690,501846,501489,501749,501415,501509,501659,501445,501750,
            500987,501547,501764,501629,500988,501512,501625,501685,501854,
            501796,501507,501485,501492,501668,501480,501505,501481,501615,
            501808];
    elif hash1 eq 904 and hash2 eq 2444 and hash3 eq 1474 then
        V:=[12,1,13,9,14,21,15,43,16,71,17,97,18,125,19,163,20,201,21,229,22,
            245,23,255,24,273];
        D:=[1,32,1,36,3,40,2,0,7,57/2,1,69/2,1,73/2,3,77/2,1,81/2,1,0,14,29,1,
            92/3,1,31,3,98/3,2,33,2,35,1,110/3,1,37,2,41,1,43,1,0,29,51/2,1,
            107/4,1,163/6,1,175/6,1,59/2,2,187/6,3,63/2,3,199/6,3,67/2,2,211/6,
            6,71/2,1,235/6,1,259/6,1,0,55,77/3,2,26,1,82/3,3,83/3,3,88/3,1,89/3,
            2,30,1,94/3,1,95/3,1,176/5,1,141/4,1,36,1,0,73,131/6,1,133/6,2,95/4,
            1,143/6,1,145/6,2,155/6,3,53/2,1,277/10,2,111/4,1,167/6,1,169/6,3,
            181/6,1,217/6,1,0,93,61/3,1,62/3,1,67/3,3,68/3,1,23,2,287/12,3,
            121/5,1,73/3,2,388/15,1,311/12,1,26,1,131/5,1,79/3,1,418/15,1,
            335/12,1,141/5,1,113/4,1,85/3,1,0,117,113/6,1,115/6,2,611/30,2,
            245/12,1,41/2,3,125/6,2,439/20,1,22,2,265/12,1,269/12,3,45/2,1,
            227/10,2,91/4,1,137/6,2,139/6,3,519/20,1,26,1,317/12,1,0,147,52/3,1,
            223/12,1,19,2,308/15,1,313/15,2,21,1,441/20,1,332/15,1,133/6,1,
            337/15,1,271/12,1,68/3,1,23,1,0,162,511/30,2,103/6,4,621/28,1,
            949/42,1,679/30,1,91/4,1,277/12,1,0,173,187/12,1,238/15,1,191/12,1,
            16,1,0,177,653/42,1,467/30,1,937/60,1,667/42,2,159/10,1,481/30,2,
            193/12,1,97/6,1,0,187,188/15,3,0];
        I:=[430357,430351,430358,430360,430367,430369,430326,430368,430352,
            430363,430371,430356,430370,430313,430347,430350,430364,430366,
            430295,430296,430286,430359,430365,430315,430361,430372,430327,
            430238,430320,430215,430277,430319,430284,430354,430285,430290,
            430321,430267,430353,430362,430322,430340,430346,430236,430355,
            430274,430282,430293,430311,430343,430349,430336,430269,430172,
            430247,430287,430251,430230,430278,430316,430306,430331,430335,
            430342,430241,430318,430248,430232,430194,430207,430275,430223,
            430303,430283,430345,430217,430291,430307,430332,430225,430245,
            430255,430246,430150,430198,430276,430263,430239,430240,430244,
            430146,430145,430310,430140,430139,430259,430273,430280,430261,
            430264,430214,430220,430234,430175,430154,430242,430190,430229,
            430136,430173,430218,430164,430130,430192,430144,430209,430231,
            430227,430249,430138,430183,430271,430141,430182,430226,430243,
            430281,430167,430262,430265,430237,430149,430222,430270,430260,
            430161,430170,430152,430158,430176,430235,430253,430272,430131,
            430134,430137,430143,430155,430221,430258,430168,430169,430197,
            430156,430191,430200,430178,430201,430135,430132,430133,430162,
            430181,430157,430298,430299,430300,430119,430112,430107,430113,
            430110,430211,430165,430213,430212,430117,430124,430115,430108,
            430123,430126,430114,430122,430109,430116,430147,430148,430153];
    elif hash1 eq 58 and hash2 eq 167 and hash3 eq 3015 then
        V:=[15,1,16,7,17,11,18,19,19,29,20,37,22,41];
        D:=[1,61/2,1,65/2,1,0,3,29,2,0,5,51/2,1,175/6,1,59/2,1,0,8,24,1,76/3,1,
            77/3,3,89/3,1,0,14,133/6,2,45/2,1,155/6,1,0,18,67/3,2,0,20,52/3,3,
            0];
        I:=[62013,62010,62011,62014,62012,61997,62000,62008,62002,61999,62004,
            62005,61993,61998,62003,62009,61996,61994,61995,62001,62006,62007];
    elif hash1 eq 1280 and hash2 eq 3025 and hash3 eq 862 then
        V:=[19,1,20,15,21,27,22,37,23,43,24,53,25,63];
        D:=[1,45/2,1,49/2,2,61/2,1,69/2,1,73/2,2,77/2,1,0,9,62/3,1,80/3,1,31,1,
            98/3,1,33,1,0,14,43/2,1,151/6,2,163/6,1,55/2,1,0,19,77/4,1,71/3,1,0,
            21,107/6,3,197/10,1,119/6,2,121/6,1,0,28,238/15,1,81/5,1,268/15,1,
            39/2,1,0,32,431/30,1,0];
        I:=[420011,419948,420026,420324,419986,419998,420173,420398,420103,
            420191,419754,420331,419988,419201,420017,420214,419957,419301,
            420025,419237,419441,420123,420174,419269,419318,419451,420031,
            419423,419422,419425,419102,419721];
    elif hash1 eq 810 and hash2 eq 1471 and hash3 eq 3007 then
        V:=[19,1,20,5,21,13,22,21,23,33,24,49,27,57];
        D:=[1,67/2,1,0,2,88/3,1,89/3,1,94/3,1,0,5,157/6,1,119/4,1,179/6,1,0,8,
            73/3,1,311/12,2,131/5,1,80/3,1,85/3,1,0,14,611/30,1,671/30,2,45/2,2,
            137/6,1,731/30,1,49/2,1,26,1,0,23,283/15,1,19,1,337/15,1,0,26,
            139/10,1,421/30,3,85/6,1,63/4,1,0];
        I:=[425715,425913,425933,425900,425917,425944,425769,425924,425872,
            425934,425903,425937,425943,425914,425919,425945,425897,425940,
            425888,425946,425942,425738,425906,425936,425713,425730,425901,
            425902,425923,425938,425643];
    elif hash1 eq 2878 and hash2 eq 2269 and hash3 eq 1988 then
        V:=[24,1,26,5,27,13,28,17,29,29,30,37];
        D:=[1,70/3,1,0,2,61/3,1,335/12,1,129/4,1,0,5,45/2,1,0,6,223/12,1,227/12,
            1,251/12,1,449/20,1,23,1,0,11,1117/60,1,559/30,1,865/42,1,0,14,
            1003/60,1,52/3,1,0];
        I:=[501693,501694,501504,501418,501464,501691,501712,501447,501814,
            501835,501715,501798,501297,501815,501842];
    elif hash1 eq 1562 and hash2 eq 1290 and hash3 eq 2390 then
        V:=[28,1,30,9,32,17,33,27];
        D:=[1,61/3,1,263/12,1,311/12,1,0,4,253/15,1,52/3,1,19,1,0,7,67/5,1,
            203/15,1,41/3,1,14,1,0,11,34/3,1,23/2,1,697/60,2,349/30,1,35/3,1,
            47/4,1,0];
        I:=[405476,406997,406676,406659,405474,406895,406284,405508,405472,
            407852,406711,404471,407591,408134,404564,406990,406872];
    elif hash1 eq 1327 and hash2 eq 2292 and hash3 eq 3079 then
        V:=[25,1,26,9,27,13,28,19,29,23,31,29,32,43];
        D:=[1,163/6,1,175/6,1,131/4,1,0,4,126/5,1,0,5,97/6,1,133/6,1,0,7,16,1,0,
            8,89/6,1,221/12,1,0,10,25/2,1,541/42,1,391/30,1,157/12,1,79/6,1,
            451/30,1,0,16,67/5,1,0];
        I:=[397138,398306,399839,397443,398812,400076,400293,399431,398129,
            395780,396009,396012,396711,397444,395742,395351];
    elif hash1 eq 2173 and hash2 eq 541 and hash3 eq 1923 then
        V:=[19,1,20,5,21,15,22,27,23,33,24,41,25,47,26,55,27,61,31,65];
        D:=[1,199/6,1,0,2,26,1,83/3,1,95/3,1,130/3,1,0,6,277/10,1,167/6,1,181/6,
            1,229/6,1,167/4,1,0,11,68/3,1,141/5,1,0,13,611/30,1,247/10,1,115/4,
            1,0,16,227/12,1,338/15,1,0,18,461/30,1,31/2,1,257/12,1,0,21,76/5,1,
            268/15,1,0,23,527/30,1,0,24,1907/210,2,0];
        I:=[541952,541945,541975,542007,542010,541955,541953,542012,542006,
            541991,541943,542003,541957,542005,541992,541910,541995,541968,
            541980,542019,541978,541988,541918,541949,541950];
    elif hash1 eq 1233 and hash2 eq 1602 and hash3 eq 1072 then
        V:=[14,1,15,5,16,11,17,17,19,23,20,31,21,41,22,49,23,53,24,57,25,61,26,
            65];
        D:=[1,46,1,0,2,69/2,2,93/2,2,0,6,104/3,2,116/3,1,0,9,163/6,1,259/6,1,0,
            11,97/6,1,121/6,1,425/12,1,0,14,238/15,2,73/4,1,68/3,1,335/12,1,0,
            19,147/10,2,399/20,2,241/12,2,0,25,179/12,1,0,26,689/30,1,0,27,
            203/15,1,0,28,1271/84,1,0,29,58/5,1,0];
        I:=[543680,543712,543736,543684,543728,543681,543735,543704,543726,
            543676,543701,543732,543687,543675,543747,543734,543729,543692,
            543721,543748,543688,543696,543707,543722,543727,543634,543703,
            543632,543686];
    elif hash1 eq 1515 and hash2 eq 392 and hash3 eq 226 then
        V:=[10,1,11,9,12,13,13,19,14,27,15,35,16,41,17,53,18,61,20,65,21,71,22,
            83,23,87,24,91];
        D:=[1,50,1,54,1,58,1,0,4,101/2,1,0,5,140/3,1,51,1,0,7,259/6,1,187/4,1,
            283/6,1,0,10,133/4,1,157/4,1,131/3,1,0,13,473/12,1,263/6,1,0,15,
            448/15,1,151/5,1,32,1,431/12,1,36,1,0,20,317/12,1,971/30,1,433/12,1,
            0,23,529/20,1,0,24,115/6,1,139/6,1,0,26,1583/84,1,537/28,1,809/42,1,
            58/3,1,1393/60,1,0,31,947/60,1,0,32,1327/84,1,0,33,184/15,1,0];
        I:=[520026,520020,520019,520018,520017,520015,520036,519984,519991,
            519990,520013,519985,519992,519954,519973,519972,520022,519966,
            519996,520003,519982,519960,519959,519993,519946,519936,519938,
            519941,519937,519935,519957,519933,519969];
    elif hash1 eq 1703 and hash2 eq 3199 and hash3 eq 2213 then
        V:=[22,1,23,5,24,15,25,29,26,43,27,55,28,61];
        D:=[1,23,1,0,2,39/2,1,139/6,1,47/2,1,151/6,1,0,6,53/3,1,18,2,59/3,2,20,
            3,85/4,1,65/3,2,0,17,33/2,2,71/4,1,107/6,1,109/6,2,37/2,1,79/4,1,0,
            25,191/12,1,16,1,49/3,12,50/3,3,56/3,2,0,44,89/6,1,91/6,1,0,46,40/3,
            2,46/3,1,0];
        I:=[247411,247428,247467,247193,246535,247460,247226,247412,247450,
            247462,247016,247198,247325,247227,246551,247297,247255,247282,
            247106,247214,246932,247239,247242,246545,246962,247251,246942,
            246943,246995,247098,247099,247192,247231,247240,247372,247385,
            247387,247390,246986,247139,247356,246543,246888,246917,246920,
            246878,246895,246483];
    elif hash1 eq 1891 and hash2 eq 1830 and hash3 eq 672 then
        V:=[22,1,23,7,24,13,25,23,26,29,27,35];
        D:=[1,62/3,2,68/3,1,0,4,103/6,1,115/6,2,0,7,47/3,1,53/3,4,59/3,1,85/4,1,
            0,14,95/6,1,131/6,1,0,16,15,1,50/3,1,0,18,31/2,1,0];
        I:=[417861,418763,418177,418275,416077,417976,418661,417865,417982,
            418798,418800,416185,417236,418030,415506,416329,416071,415339];
    elif hash1 eq 2737 and hash2 eq 2248 and hash3 eq 1753 then
        V:=[26,1,27,5,28,11,29,17,30,27,31,39];
        D:=[1,53/3,2,0,3,121/6,1,49/2,1,0,5,44/3,1,191/12,2,0,8,77/6,1,431/30,1,
            29/2,1,491/30,1,0,12,38/3,1,193/15,1,155/12,1,13,1,76/5,1,0,17,
            331/30,1,137/12,1,0];
        I:=[500724,500726,500729,500720,500743,500732,500817,500788,500721,
            500749,500722,500790,500809,500747,500739,500723,500770,500751];
    elif hash1 eq 1844 and hash2 eq 1569 and hash3 eq 1665 then
        V:=[23,1,24,5,25,11,26,19,27,25,28,33,29,47];
        D:=[1,163/6,2,0,3,70/3,1,101/4,1,0,5,119/6,1,121/6,1,133/6,1,0,8,268/15,
            1,61/3,1,0,10,59/4,1,359/20,1,37/2,1,0,13,40/3,3,29/2,1,175/12,1,
            223/15,3,15,2,199/12,1,0,24,391/30,1,157/12,1,0];
        I:=[248326,248718,248810,247691,248926,249608,248347,248348,247989,
            247991,247997,247948,247986,247994,248278,248969,248099,247813,
            247816,248058,247901,249595,247619,248333,248376];
    elif hash1 eq 1562 and hash2 eq 438 and hash3 eq 1353 then
        V:=[22,1,23,5,24,9,25,19,26,27,27,33,28,37];
        D:=[1,92/3,1,0,2,115/6,1,0,3,18,1,59/3,1,20,1,64/3,1,0,7,85/6,2,177/10,
            1,109/6,1,0,11,238/15,1,215/12,1,0,13,77/6,1,0,14,188/15,1,13,1,0];
        I:=[502568,502233,502431,502425,502469,502552,502206,502412,502403,
            502560,502610,502254,502583,502413,502606];
    elif hash1 eq 1703 and hash2 eq 2341 and hash3 eq 3235 then
        V:=[23,1,24,7,26,15,27,19,28,31,29,39];
        D:=[1,163/6,1,55/2,1,0,3,70/3,1,71/3,2,24,2,0,8,55/3,1,0,9,31/2,1,
            167/10,1,101/6,3,103/6,2,221/12,1,0,17,44/3,1,15,2,46/3,1,0,21,79/6,
            1,0];
        I:=[247383,246959,247391,247256,247278,246963,247444,247287,246969,
            246929,246815,246856,247435,247246,247451,246893,246871,247476,
            247477,247197,246915];
    elif hash1 eq 1421 and hash2 eq 1269 and hash3 eq 1213 then
        V:=[20,1,21,5,22,15,23,23,24,31,25,37,26,43];
        D:=[1,31,1,0,2,151/6,2,51/2,1,163/6,1,55/2,1,0,7,22,1,71/3,1,24,1,0,10,
            121/6,3,131/6,1,133/6,1,0,15,56/3,3,89/4,1,0,19,75/4,1,113/6,1,0,21,
            46/3,1,0];
        I:=[251805,251811,251812,251794,251781,251743,251816,251741,251815,
            251750,251782,251822,251714,251776,251771,251784,251785,251703,
            251702,251735,251732];
    elif hash1 eq 1562 and hash2 eq 1952 and hash3 eq 2601 then
        V:=[18,1,19,9,20,19,21,27,22,31,23,41,24,49,25,59];
        D:=[1,27,1,104/3,1,35,1,0,4,163/6,1,55/2,1,59/2,1,63/2,1,0,8,70/3,1,
            77/3,3,83/3,1,0,13,281/12,1,0,14,55/3,3,239/12,1,61/3,2,287/12,1,0,
            21,37/2,1,113/6,1,241/12,1,0,24,199/12,1,50/3,1,17,1,20,1,0,28,
            451/30,1,461/30,1,0];
        I:=[426414,426298,426415,426296,426413,426419,426421,426221,426018,
            426234,426248,425989,426019,426057,426294,426391,426017,426021,
            426023,425671,426104,426232,425740,426107,426056,426011,425674,
            425827,425828];
    elif hash1 eq 904 and hash2 eq 863 and hash3 eq 3169 then
        V:=[20,1,23,5,24,15,25,23,28,33,29,41];
        D:=[1,80/3,1,0,2,119/6,1,87/4,1,143/6,1,167/6,1,0,6,97/4,1,388/15,1,
            79/3,1,0,9,599/30,1,125/6,1,191/8,1,149/6,1,0,13,281/21,1,14,1,
            226/15,1,0,16,241/18,1,139/10,1,0];
        I:=[509052,509134,509114,508951,508349,508953,508375,508666,509073,
            509005,508381,508350,508594,508578,508621,508323,508317];
    elif hash1 eq 152 and hash2 eq 734 and hash3 eq 1669 then
        V:=[21,1,22,5,23,17,24,23,25,31];
        D:=[1,45/2,1,0,2,56/3,1,62/3,1,23,1,74/3,1,25,3,0,9,127/6,4,163/6,1,0,
            14,96/5,1,59/3,2,20,1,0,18,71/4,1,107/6,1,0];
        I:=[251403,251464,251125,251397,251381,251419,251422,251426,250481,
            250983,251402,251406,250407,250485,251417,251424,251451,250985,
            251109];
    elif hash1 eq 575 and hash2 eq 3273 and hash3 eq 2679 then
        V:=[23,1,24,7,25,17,26,29,27,43,28,47,29,57];
        D:=[1,139/6,1,151/6,1,0,3,20,1,65/3,1,93/4,1,71/3,2,0,8,119/6,1,217/10,
            1,131/6,1,133/6,1,143/6,1,0,13,18,1,55/3,1,56/3,2,298/15,1,239/12,1,
            61/3,1,0,20,91/6,2,0,22,15,2,76/5,1,46/3,1,253/15,1,0,27,299/20,1,
            451/30,1,461/30,1,185/12,1,0];
        I:=[249214,249502,248823,248733,249040,248062,249504,249047,248043,
            247876,247757,247725,248724,249505,247863,248890,248087,249019,
            248000,248699,249488,248696,248884,248017,248045,247637,247565,
            247556,247557,247549];
    elif hash1 eq 1609 and hash2 eq 1962 and hash3 eq 1201 then
        V:=[25,1,26,7,27,13,28,23,29,33,30,41];
        D:=[1,43/2,1,99/4,1,0,3,59/3,1,106/5,1,0,5,97/6,1,121/6,2,41/2,1,131/6,
            1,0,10,44/3,1,49/3,1,91/5,1,56/3,1,0,14,173/12,1,29/2,5,89/6,1,0,21,
            155/12,1,0];
        I:=[241412,240315,241164,239607,240777,239232,240470,241130,238393,
            239789,240178,240361,238897,238648,239190,240469,240575,240792,
            241254,240586,238670];
    elif hash1 eq 1562 and hash2 eq 3297 and hash3 eq 210 then
        V:=[19,1,20,9,21,15,22,19,24,29,25,37];
        D:=[1,57/2,1,61/2,1,77/2,1,0,4,80/3,1,104/3,1,0,6,115/6,1,0,7,58/3,1,
            59/3,1,65/3,2,109/4,1,0,12,292/15,1,61/3,1,111/5,1,0,15,391/20,1,
            599/30,1,0];
        I:=[420547,420548,420563,420511,420105,420541,420196,420185,420528,
            420533,419391,419461,419394,419065,419096,419078];
    elif hash1 eq 763 and hash2 eq 449 and hash3 eq 1200 then
        V:=[21,1,22,5,23,17,28,25,29,35];
        D:=[1,135/4,1,0,2,151/5,1,91/3,1,383/12,1,97/3,1,407/12,1,0,7,313/12,2,
            337/12,1,911/30,1,0,11,827/60,1,208/15,1,167/12,1,461/30,1,0,15,
            367/30,1,0];
        I:=[514429,514724,515447,514356,515448,515225,515181,515183,514474,
            514129,514410,514425,514361,514015,514144];
    elif hash1 eq 904 and hash2 eq 1344 and hash3 eq 66 then
        V:=[20,1,21,7,22,13,23,21,24,31,26,35];
        D:=[1,24,1,26,1,0,3,41/2,3,45/2,4,0,10,19,5,62/3,2,21,3,0,20,35/2,1,
            115/6,2,39/2,1,43/2,1,0,25,52/3,2,0,27,43/3,1,0];
        I:=[674141,674142,674022,674140,674144,674085,674126,674132,674143,
            674028,674043,674064,674075,674098,673919,674066,673996,674091,
            674128,673999,673906,674030,673875,673817,674005,674089,673985];
    elif hash1 eq 1985 and hash2 eq 2918 and hash3 eq 1609 then
        V:=[21,1,22,9,23,19,24,27,27,39];
        D:=[1,53/2,1,57/2,1,61/2,1,0,4,68/3,1,25,4,80/3,1,27,2,0,12,127/6,1,
            139/6,3,47/2,1,0,17,58/3,1,59/3,3,64/3,1,65/3,3,76/3,1,0,26,173/12,
            3,29/2,2,0];
        I:=[249443,249093,249359,249223,249039,249117,249274,249358,249059,
            249122,249219,249009,249043,249393,249432,249012,249038,249004,
            249083,249215,248420,248240,249034,249120,247656,248231,249069,
            249224,248450,249099];
    elif hash1 eq 1985 and hash2 eq 3252 and hash3 eq 928 then
        V:=[12,1,13,5,14,9,15,13,16,17,18,23,19,33,22,39,23,47,24,53];
        D:=[1,54,1,0,2,85/2,1,0,3,98/3,1,0,4,175/6,1,0,5,146/5,1,131/3,1,0,7,
            67/3,1,141/5,1,407/12,1,36,1,0,11,63/2,1,1019/30,1,0,13,907/60,1,
            1243/60,1,1387/60,1,0,16,121/8,1,83/4,1,0,18,947/60,1,0];
        I:=[519584,519601,519596,519590,519526,519504,519585,519518,519520,
            519513,519330,519409,519506,519369,519373,519332,519331,519377];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 12 mod 47.
function cached_ids_12(P,hash1,hash2,hash3)
    if hash1 eq 1939 and hash2 eq 1484 and hash3 eq 1848 then
        V:=[18,1,19,5,20,13,21,29,22,43,23,57,24,71,26,79,27,85,28,89];
        D:=[1,26,1,0,2,57/2,3,61/2,1,65/2,1,0,7,68/3,1,74/3,1,80/3,1,27,1,86/3,
            2,98/3,3,33,1,0,17,127/6,1,43/2,1,47/2,1,151/6,3,51/2,4,55/2,1,0,28,
            53/3,1,96/5,2,59/3,2,20,1,65/3,1,109/4,1,0,36,209/12,1,177/10,2,
            107/6,3,79/4,1,257/12,1,131/6,1,0,45,215/12,2,298/15,1,239/12,2,0,
            50,218/15,1,175/12,1,0,52,299/20,1,0,53,178/15,1,0];
        I:=[420106,420009,420075,420087,420044,420020,420085,420029,419912,
            419207,420006,420033,420007,420182,420429,419327,420080,419922,
            419906,420002,420098,420127,419916,419917,419920,419928,419235,
            420172,419418,419420,419206,420116,419240,420053,419156,419951,
            419335,419432,420008,420021,420068,419225,419234,419321,419101,
            419389,419476,419087,419553,419491,419932,419020,419646];
    elif hash1 eq 3067 and hash2 eq 3121 and hash3 eq 9 then
        V:=[24,1,25,13,26,23,27,27,28,31,29,37];
        D:=[1,74/3,1,25,1,80/3,1,27,5,29,2,0,11,43/2,1,139/6,1,47/2,4,151/6,2,0,
            19,59/3,1,0,20,109/6,1,0,21,49/3,2,50/3,1,0,24,89/6,1,91/6,1,0];
        I:=[671719,672327,671714,671693,671759,671773,672250,672339,671799,
            672258,671743,672171,671634,671635,671724,671725,671237,672172,
            672169,671727,671717,672176,671800,671282,671772];
    elif hash1 eq 811 and hash2 eq 631 and hash3 eq 41 then
        V:=[25,1,26,9,27,17,30,25,31,37,32,41];
        D:=[1,151/6,2,163/6,1,59/2,1,0,5,22,1,93/4,1,71/3,1,0,8,119/6,2,121/6,2,
            41/2,1,0,13,175/12,1,44/3,1,223/15,1,15,1,46/3,1,0,18,59/4,1,0,19,
            45/4,1,0];
        I:=[672190,672235,672188,672259,671696,671622,672336,671601,672173,
            671303,671863,671726,671151,671794,671639,671791,671760,671101,
            671641];
    elif hash1 eq 2362 and hash2 eq 1719 and hash3 eq 757 then
        V:=[27,1,28,5,29,13,30,17];
        D:=[1,193/6,1,0,2,68/3,1,74/3,1,79/3,1,0,5,269/12,1,0,6,223/12,1,283/15,
            2,227/12,1,96/5,2,58/3,3,313/15,1,0];
        I:=[406757,407047,406811,407046,406900,406923,404548,406568,406905,
            406758,407061,406820,406907,407028,406840];
    elif hash1 eq 717 and hash2 eq 657 and hash3 eq 1868 then
        V:=[22,1,23,7,24,11,25,17,26,23,28,29];
        D:=[1,29,1,37,3,0,5,187/6,1,0,6,70/3,1,95/3,1,0,8,143/6,2,157/6,1,0,11,
            101/5,1,322/15,1,0,13,361/20,1,277/15,2,0];
        I:=[245877,245868,246014,246018,245887,245885,245867,245869,246011,
            245876,245891,245851,245846,245845,245849];
    elif hash1 eq 2268 and hash2 eq 344 and hash3 eq 1287 then
        V:=[12,1,13,11,14,27,15,55,16,79,17,105,18,137,19,163,20,181,21,195,22,
            209,23,215];
        D:=[1,32,1,34,1,40,2,44,1,0,6,61/2,2,65/2,3,69/2,2,73/2,1,77/2,4,81/2,2,
            89/2,1,0,21,80/3,1,86/3,1,29,1,92/3,1,31,2,98/3,1,33,1,104/3,1,35,3,
            37,1,116/3,1,39,1,122/3,1,0,37,151/6,1,51/2,1,163/6,1,55/2,1,175/6,
            2,187/6,3,199/6,2,67/2,1,211/6,7,235/6,1,87/2,1,0,58,71/3,2,101/4,1,
            76/3,1,77/3,2,26,1,82/3,2,83/3,2,89/3,2,30,1,94/3,1,106/3,1,107/3,2,
            0,76,131/6,1,133/6,1,45/2,1,281/12,1,143/6,4,155/6,1,329/12,1,111/4,
            1,167/6,2,169/6,1,297/10,2,181/6,1,61/2,1,215/6,1,217/6,1,0,96,20,1,
            61/3,1,67/3,3,68/3,1,23,1,47/2,1,358/15,1,287/12,1,73/3,1,418/15,1,
            335/12,1,85/3,1,0,110,611/30,1,245/12,1,41/2,1,671/30,4,269/12,3,
            137/6,1,139/6,2,791/30,1,0,124,86/5,1,52/3,1,369/20,1,278/15,1,45/2,
            1,271/12,1,0,130,511/30,2,103/6,2,1333/60,1,68/3,1,91/4,1,277/12,1,
            0,138,309/20,1,233/15,1,0,140,319/20,1,481/30,1,0];
        I:=[429933,429942,429899,429927,429956,429938,429955,429856,429929,
            429950,429869,429928,429940,429932,429945,429948,429957,429864,
            429947,429844,429897,429769,429931,429876,429861,429949,429877,
            429839,429865,429785,429920,429930,429860,429847,429835,429910,
            429849,429840,429907,429724,429814,429916,429817,429863,429913,
            429730,429767,429799,429783,429808,429836,429892,429896,429902,
            429918,429762,429638,429834,429906,429740,429761,429774,429915,
            429805,429721,429922,429731,429843,429771,429881,429760,429804,
            429807,429681,429756,429846,429800,429758,429718,429773,429832,
            429842,429882,429802,429806,429791,429684,429753,429759,429647,
            429714,429653,429660,429644,429654,429887,429837,429623,429764,
            429841,429763,429786,429719,429650,429741,429793,429609,429632,
            429634,429674,429641,429675,429659,429679,429686,429689,429685,
            429750,429757,429711,429670,429801,429610,429640,429658,429651,
            429695,429673,429664,429676,429688,429697,429894,429587,429589,
            429595,429591,429636,429677,429593,429594];
    elif hash1 eq 999 and hash2 eq 1393 and hash3 eq 505 then
        V:=[24,1,25,15,26,41,27,51,28,65,29,81,30,101,31,115];
        D:=[1,59/3,1,77/3,1,88/3,1,89/3,2,30,1,95/3,1,0,8,257/12,1,217/10,1,
            145/6,2,305/12,1,257/10,1,155/6,3,157/6,1,329/12,1,111/4,1,167/6,1,
            297/10,1,179/6,1,0,23,56/3,1,62/3,1,67/3,2,121/5,1,0,28,221/12,1,
            37/2,1,113/6,1,611/30,2,41/2,1,125/6,1,0,35,15,1,253/15,1,52/3,1,
            277/15,1,278/15,1,223/12,2,313/15,1,0,43,83/6,1,59/4,2,91/6,3,
            461/30,1,185/12,1,31/2,1,997/60,1,511/30,1,379/20,2,0,56,79/6,1,
            66/5,1,27/2,1,163/12,1,41/3,3,61/4,1,0,64,35/3,1,499/42,1,361/30,1,
            0];
        I:=[416703,416560,416418,416494,416929,417123,416911,416690,417121,
            417091,417136,416702,416493,415831,416358,417055,417126,416685,
            415752,416731,417041,417809,417054,416942,415878,416529,416076,
            417469,416313,415448,416973,416975,417521,417570,417066,416177,
            417067,416075,416038,415953,415954,416713,416475,416524,416695,
            415572,415811,417097,416196,416016,416464,416197,415340,415055,
            415131,417359,416474,416523,417008,416471,416486,416624,415520,
            415279,415385,415438];
    elif hash1 eq 2268 and hash2 eq 1056 and hash3 eq 1927 then
        V:=[24,1,25,7,26,11,27,17,28,27,29,35];
        D:=[1,65/3,1,24,1,0,3,121/6,1,0,4,55/3,1,56/3,1,0,6,491/30,1,101/6,3,
            103/6,1,221/12,1,0,12,175/12,1,179/12,2,15,9,0,24,79/6,1,0];
        I:=[247388,247185,247194,247380,247307,246780,246879,247195,247426,
            246931,246897,246632,246936,247200,246901,246990,247189,247359,
            247367,247369,247370,247384,247422,246747];
    elif hash1 eq 482 and hash2 eq 2458 and hash3 eq 2049 then
        V:=[20,1,21,7,22,11,23,19,24,23];
        D:=[1,23,4,25,1,0,6,43/2,3,0,9,59/3,1,20,2,65/3,2,0,14,109/6,2,0,16,
            50/3,1,0];
        I:=[674372,674373,674375,674376,674326,674357,674360,674379,674358,
            674303,674353,674292,674294,674288,674300,674349];
    elif hash1 eq 2362 and hash2 eq 2893 and hash3 eq 1110 then
        V:=[21,1,22,7,23,15,24,23,26,31,27,41,29,47];
        D:=[1,99/4,1,199/6,1,0,3,83/3,2,89/3,1,95/3,1,0,7,95/4,1,257/10,1,155/6,
            1,0,10,215/12,1,18,1,55/3,1,0,13,175/12,1,361/20,1,283/15,2,19,1,0,
            18,853/60,1,91/6,1,0,20,347/30,1,0];
        I:=[419865,420241,419251,420206,419314,419785,419442,419242,419255,
            419300,420067,420190,419249,419013,418963,419436,419057,419012,
            420204,419293];
    elif hash1 eq 905 and hash2 eq 3343 and hash3 eq 2852 then
        V:=[22,1,23,5,24,11,25,21,26,33,27,37,28,53,29,73,31,85];
        D:=[1,94/3,1,0,2,169/6,1,181/6,1,0,4,89/4,1,73/3,1,79/3,2,113/4,1,0,9,
            41/2,1,671/30,4,269/12,1,137/6,1,293/12,1,0,17,272/15,1,0,18,185/12,
            1,511/30,1,103/6,2,209/12,1,107/6,1,571/30,1,1237/60,1,0,26,167/12,
            1,76/5,1,233/15,1,47/3,2,191/12,1,1003/60,1,17,1,87/5,1,19,2,0,37,
            409/30,1,55/4,3,46/3,1,481/30,1,521/28,1,0,44,841/70,1,73/6,1,0];
        I:=[513771,513661,513690,513674,513696,513554,513664,513513,513565,
            513636,513637,513641,513688,513514,513563,513673,513615,513666,
            513701,513581,513612,513577,513675,513621,513667,513689,513643,
            513697,513535,513546,513512,513743,513625,513627,513388,513418,
            513631,513391,513647,513648,513387,513318,513292,513294,513297];
    elif hash1 eq 2973 and hash2 eq 1730 and hash3 eq 1136 then
        V:=[25,1,27,5,28,11,29,15,30,21,31,33];
        D:=[1,139/6,1,0,2,49/2,1,193/6,1,0,4,73/3,1,0,5,221/12,1,39/2,1,0,7,
            129/7,1,131/7,1,227/12,1,96/5,1,106/5,1,0,12,877/60,1,625/42,1,
            149/10,1,451/30,1,499/30,1,0];
        I:=[397999,397097,398323,396972,397527,395796,396762,396681,397847,
            397828,396841,395476,395722,395717,397358,397528];
    elif hash1 eq 1751 and hash2 eq 2547 and hash3 eq 1844 then
        V:=[25,1,26,7,27,13,28,21,29,27,30,35];
        D:=[1,139/6,1,99/4,1,0,3,18,1,20,1,0,5,107/6,1,213/10,1,87/4,1,0,8,50/3,
            1,55/3,2,0,11,29/2,1,491/30,1,33/2,1,0,14,41/3,1,15,1,0];
        I:=[240323,240137,239065,239044,240062,238366,238396,238374,238423,
            239174,238950,238418,238373,238934,238314];
    elif hash1 eq 999 and hash2 eq 647 and hash3 eq 1237 then
        V:=[10,1,11,5,13,9,14,13,15,17,16,21,17,27,18,35,20,39,21,45,22,53];
        D:=[1,54,1,0,2,101/2,1,0,3,247/6,1,0,4,131/3,1,0,5,135/4,1,0,6,407/12,1,
            36,1,0,8,63/2,1,1019/30,1,409/12,1,0,11,392/15,1,0,12,1243/60,1,
            1387/60,1,0,14,61/3,1,83/4,1,104/5,1,0,17,947/60,1,0];
        I:=[520044,520045,520016,519986,519989,519983,519987,519988,519953,
            519951,519952,519961,519944,519948,519945,519947,519962];
    elif hash1 eq 153 and hash2 eq 2042 and hash3 eq 2393 then
        V:=[16,1,19,7,20,11,21,17,23,23,24,27,26,31,27,37,28,43];
        D:=[1,152/3,1,164/3,1,0,3,437/10,1,0,4,250/7,1,127/3,1,0,6,639/20,2,
            1031/30,1,0,9,961/42,1,0,10,58/3,1,0,11,1313/84,1,94/5,1,0,13,
            279/20,1,113/6,1,0,15,1286/105,1,0];
        I:=[518498,518535,518253,518252,518567,518524,518565,518376,518245,
            518181,518261,518218,518459,518190,518273];
    elif hash1 eq 153 and hash2 eq 1525 and hash3 eq 1074 then
        V:=[22,1,23,11,24,21,25,41,26,45,27,53,28,61];
        D:=[1,62/3,1,21,1,74/3,2,29,1,0,6,67/4,1,115/6,1,39/2,1,151/6,1,0,10,
            46/3,1,53/3,1,94/5,1,58/3,1,59/3,2,20,1,106/5,1,85/4,1,116/5,1,0,20,
            97/6,1,0,21,191/12,2,55/3,1,56/3,1,0,25,29/2,1,89/6,2,193/12,1,0,29,
            46/3,1,0];
        I:=[412191,411822,411860,412120,410250,411969,411983,412308,409789,
            412345,412309,410899,410845,412071,412125,412395,409801,411097,
            409628,410774,410240,411779,409281,409594,412098,411024,411119,
            410166,409238];
    elif hash1 eq 482 and hash2 eq 1692 and hash3 eq 1496 then
        V:=[21,1,22,11,23,19,24,27,25,35,26,41];
        D:=[1,41/2,2,45/2,1,61/2,2,65/2,1,0,7,56/3,1,74/3,1,27,1,0,10,115/6,1,
            47/2,2,51/2,1,0,14,53/3,2,59/3,2,65/3,1,0,19,97/6,1,107/6,1,0,21,
            208/15,1,71/5,1,16,1,0];
        I:=[412090,412118,412086,410534,412076,412157,412152,412132,411891,
            412107,410362,411063,410436,410213,412130,410672,412036,411094,
            410363,411046,410581,410649,411055];
    elif hash1 eq 1281 and hash2 eq 1848 and hash3 eq 1503 then
        V:=[16,1,18,7,20,21,22,27,24,35];
        D:=[1,104/3,1,39,1,0,3,65/3,1,70/3,2,77/3,1,94/3,1,32,1,119/3,2,0,11,
            292/15,1,62/3,1,0,13,98/5,1,102/5,1,313/15,2,0,17,208/15,1,0];
        I:=[518754,518674,518691,518545,518596,518705,518351,518699,518470,
            518512,518478,518677,518541,518560,518427,518549,518520];
    elif hash1 eq 2785 and hash2 eq 1212 and hash3 eq 3419 then
        V:=[24,1,25,7,26,11,27,15,28,23,29,33];
        D:=[1,68/3,1,74/3,1,0,3,39/2,1,0,4,53/3,2,0,6,95/6,1,97/6,1,71/4,1,0,9,
            13,1,14,1,71/5,1,57/4,1,0,13,371/30,1,25/2,2,77/6,2,0];
        I:=[406007,406035,406187,406026,406061,405975,406083,405531,406009,
            406033,406012,405982,405884,405969,406016,406008,406082];
    elif hash1 eq 1375 and hash2 eq 3311 and hash3 eq 3389 then
        V:=[20,1,21,7,22,15,23,21,24,27,25,33,26,41,28,45];
        D:=[1,29,1,33,1,0,3,163/6,1,175/6,1,63/2,1,0,6,77/3,1,83/3,1,0,8,143/6,
            2,145/6,1,0,11,263/12,2,67/3,1,0,14,491/30,1,101/6,1,245/12,1,0,17,
            441/20,1,0,18,67/5,1,203/15,3,0];
        I:=[423130,423080,423133,422996,423129,423081,423207,423043,423094,
            423046,423182,423264,423033,423108,423037,422899,422370,423250,
            423101,423103,423117];
    elif hash1 eq 3443 and hash2 eq 1533 and hash3 eq 1671 then
        V:=[21,1,22,7,23,19,24,27,25,39,26,43,28,49];
        D:=[1,57/2,1,61/2,1,0,3,74/3,1,25,2,27,1,31,3,35,1,0,11,139/6,1,163/6,6,
            59/2,1,0,19,64/3,1,65/3,1,22,1,77/3,1,26,1,0,24,133/6,2,0,26,268/15,
            1,55/3,2,0,29,289/20,2,223/15,1,15,2,253/15,1,17,1,0];
        I:=[251141,250821,251153,251139,251188,250731,250759,250761,250767,
            250748,251377,250757,250876,250909,250917,251174,251250,250191,
            250902,251011,250895,250877,250386,250038,250160,250310,250020,
            251320,250269,250433,250019,250751,250752,249949,249948];
    elif hash1 eq 2879 and hash2 eq 1762 and hash3 eq 2811 then
        V:=[25,1,26,7,27,13,28,17,29,21,30,31,31,39,32,45,33,51];
        D:=[1,329/12,1,135/4,1,0,3,105/4,1,151/5,1,0,5,227/10,1,0,6,470/21,1,0,
            7,149/10,1,461/30,1,997/60,1,521/30,1,0,11,53/4,1,77/5,1,268/15,1,0,
            14,40/3,1,55/4,1,0,16,158/15,1,23/2,1,0,18,617/60,1,729/70,1,293/28,
            1,317/30,5,283/24,1,0];
        I:=[501898,502934,502935,503179,501807,501673,502955,503184,501662,
            503182,501510,502954,501833,501499,501954,501978,501462,501053,
            501887,501988,502985,503028,503046,503047,503107,501066];
    elif hash1 eq 1751 and hash2 eq 2223 and hash3 eq 822 then
        V:=[23,1,24,5,25,13,27,21,29,29];
        D:=[1,53/2,1,0,2,68/3,1,23,2,74/3,1,0,6,35/2,1,127/6,3,43/2,2,0,12,63/4,
            1,95/6,2,107/6,3,0,18,14,1,169/12,1,0];
        I:=[672824,672832,672814,672822,672504,672710,672595,672658,672915,
            672811,672821,672593,672576,672781,672509,672542,672783,672483,
            672682];
    elif hash1 eq 2926 and hash2 eq 703 and hash3 eq 266 then
        V:=[18,1,20,11,22,19,24,25,26,29];
        D:=[1,27,1,86/3,1,39,1,45,1,0,5,65/3,1,101/3,1,196/5,1,0,8,89/4,1,24,1,
            0,10,19,1,0,11,41/3,1,208/15,1,43/3,1,96/5,2,0];
        I:=[543250,543276,543277,543127,543112,543258,543080,543194,543088,
            543096,542691,543101,543163,542740,542831];
    elif hash1 eq 1469 and hash2 eq 2760 and hash3 eq 1364 then
        V:=[14,1,15,7,16,13,17,25,18,37,19,43,20,47,21,55,22,61,23,67];
        D:=[1,38,1,42,1,0,3,73/2,3,77/2,1,0,7,86/3,1,98/3,1,110/3,1,37,1,116/3,
            2,0,13,175/6,2,187/6,1,199/6,1,235/6,1,259/6,1,0,19,146/5,1,89/3,2,
            0,22,143/6,2,0,24,22,1,67/3,1,28,1,0,27,269/12,1,45/2,2,0,30,109/6,
            1,283/15,1,0,32,56/3,1,0];
        I:=[254851,254857,254826,254847,254865,254839,254850,254823,254870,
            254862,254798,254866,254832,254846,254852,254817,254813,254786,
            254779,254778,254872,254842,254845,254807,254804,254785,254781,
            254780,254784,254805,254783,254775];
    elif hash1 eq 3067 and hash2 eq 2946 and hash3 eq 1816 then
        V:=[20,1,21,7,22,11,24,23,25,29];
        D:=[1,25,1,27,1,0,3,55/2,1,0,4,59/3,1,20,1,65/3,2,93/4,1,71/3,1,0,10,
            49/3,1,55/3,1,0,12,197/12,2,103/6,2,0];
        I:=[421085,421080,421071,421104,421087,421017,421092,421033,421068,
            421103,420953,420973,420983,420951,420952];
    elif hash1 eq 2127 and hash2 eq 2570 and hash3 eq 3300 then
        V:=[23,1,24,5,25,11,26,17,27,27,28,35,29,49,30,67];
        D:=[1,187/6,1,0,2,59/3,1,83/3,2,0,5,95/4,1,169/6,2,0,8,18,1,111/5,1,
            67/3,2,68/3,1,0,13,197/12,1,551/30,1,37/2,1,0,16,46/3,1,97/6,1,
            199/12,1,253/15,1,17,1,52/3,1,0,22,451/30,1,181/12,1,213/14,1,
            461/30,1,31/2,1,95/6,2,997/60,1,511/30,2,0,32,53/4,1,223/12,1,0];
        I:=[416465,416691,416937,417314,416882,416172,417672,416732,415331,
            416167,417414,415491,416781,416793,414908,416972,416786,416668,
            417535,417498,416164,417015,417515,415117,415332,415477,416020,
            417545,416788,416720,416892,416784,414921];
    elif hash1 eq 1798 and hash2 eq 23 and hash3 eq 1932 then
        V:=[25,1,26,5,27,11,28,17,29,29,31,39];
        D:=[1,47/2,1,0,2,18,2,59/3,1,0,5,107/6,1,121/6,1,0,7,191/12,1,81/5,1,
            49/3,4,91/5,1,73/4,1,0,15,431/30,1,29/2,1,89/6,1,197/12,1,0,19,
            137/12,2,0];
        I:=[241059,241009,241166,240826,240619,239602,239555,239647,240199,
            240627,240841,241168,239768,239599,239626,239921,239597,238665,
            239773,240227];
    elif hash1 eq 811 and hash2 eq 30 and hash3 eq 136 then
        V:=[22,1,23,5,24,13,25,21,26,33,27,43,29,49];
        D:=[1,27,2,0,3,43/2,1,139/6,2,47/2,1,0,7,59/3,5,20,2,65/3,2,0,16,107/6,
            1,109/6,1,119/6,2,121/6,1,41/2,1,0,22,81/5,1,49/3,2,50/3,1,56/3,1,0,
            27,173/12,2,89/6,1,0,30,71/6,1,0];
        I:=[247029,247031,247024,246947,246948,247073,246937,246945,246983,
            247071,247165,247035,247054,246732,246991,246797,247052,246548,
            246743,247159,246607,246792,246661,247127,247184,246601,247125,
            247170,246955,246538];
    elif hash1 eq 3302 and hash2 eq 1764 and hash3 eq 642 then
        V:=[22,1,23,5,24,11,25,23,26,29,27,37];
        D:=[1,59/3,1,0,2,37/2,1,119/6,1,0,4,50/3,1,89/5,1,215/12,1,55/3,1,22,1,
            0,9,31/2,3,101/6,1,0,13,15,2,253/15,1,18,1,0,17,161/12,1,0];
        I:=[513938,513895,513939,513891,513768,513893,513888,513295,513859,
            513900,513901,513594,513882,513883,513758,513591,513759];
    elif hash1 eq 3396 and hash2 eq 1700 and hash3 eq 1784 then
        V:=[20,1,21,9,22,21,23,29,24,41,25,47,26,57,28,63];
        D:=[1,80/3,1,27,5,86/3,1,0,8,139/6,1,151/6,2,51/2,1,55/2,1,175/6,1,0,14,
            71/3,2,24,1,101/4,1,0,18,109/6,1,119/6,1,121/6,3,41/2,1,87/4,1,0,25,
            61/3,1,62/3,1,0,27,91/6,1,33/2,1,167/10,1,101/6,2,0,32,179/12,1,15,
            3,0,36,12,1,0];
        I:=[252193,252150,252157,252168,252171,252210,252219,252153,252169,
            252189,252207,252163,252119,251921,252178,252177,251938,251964,
            251962,252158,252164,252181,252211,251942,251913,251919,252174,
            251956,251941,251927,251940,252173,252146,252172,252203,251931];
    elif hash1 eq 2738 and hash2 eq 168 and hash3 eq 1697 then
        V:=[15,1,16,7,17,19,18,29,19,39,20,51,21,67,22,73,23,79,24,85,25,91,26,
            95];
        D:=[1,61/2,1,81/2,1,0,3,92/3,1,31,1,110/3,2,116/3,1,39,1,0,9,123/4,1,
            63/2,2,67/2,1,223/6,1,0,14,85/4,1,93/4,1,71/3,1,24,1,0,18,257/12,2,
            133/6,1,281/12,1,143/6,1,317/10,2,0,25,61/3,1,322/15,1,263/12,2,
            47/2,1,73/3,1,131/5,1,55/2,1,0,33,37/2,1,731/30,1,0,35,329/20,1,
            253/15,1,0,37,75/4,1,115/6,2,0,40,907/60,1,1003/60,1,0,42,1271/84,1,
            0,43,181/15,1,0];
        I:=[428989,428994,428975,428963,428931,429004,428961,428987,428905,
            428895,428930,428927,428879,428922,428701,428997,428937,428782,
            428867,428801,428846,428940,428654,428758,428804,428824,428759,
            428893,428883,428891,428709,428766,428657,428638,428757,428742,
            428643,428609,428647,428721,428594,428599,428720];
    elif hash1 eq 2315 and hash2 eq 1912 and hash3 eq 2168 then
        V:=[25,1,26,7,27,15,28,21,29,29,30,35];
        D:=[1,139/6,1,47/2,1,0,3,18,1,59/3,4,85/4,1,0,9,107/6,1,109/6,1,0,11,
            191/12,1,49/3,1,50/3,2,0,15,89/6,1,33/2,1,0,17,38/3,3,0];
        I:=[240745,241071,240964,239580,241037,241047,241097,239459,241131,
            239686,240678,239792,241000,241095,239669,238153,240603,241357,
            241364];
    elif hash1 eq 717 and hash2 eq 1551 and hash3 eq 1478 then
        V:=[29,1,30,9];
        D:=[1,95/6,1,97/6,8,33/2,1,0,11,43/3,3,44/3,5,15,1,49/3,2,0];
        I:=[664708,664702,664704,664717,664783,664860,664918,664924,664929,
            664173,664186,664187,664912,664200,664201,664249,664590,664613,
            664690,663375,663980];
    elif hash1 eq 1704 and hash2 eq 2227 and hash3 eq 25 then
        V:=[24,1,25,7,26,15,27,23,28,29,29,35,30,39];
        D:=[1,83/3,1,89/3,4,0,6,103/4,1,155/6,1,157/6,2,0,10,239/12,1,89/4,1,
            67/3,2,0,14,37/2,1,41/2,1,0,16,223/12,1,227/12,1,0,18,181/12,1,0,19,
            269/20,1,203/15,1,163/12,1,0];
        I:=[248677,248673,248808,248822,248850,247734,248688,247722,248736,
            247749,247733,247817,249487,249134,247726,247729,247837,247794,
            248073,247746,247748];
    elif hash1 eq 1939 and hash2 eq 3034 and hash3 eq 75 then
        V:=[23,1,24,7,25,25,26,33,27,41,28,49,29,59,30,65];
        D:=[1,107/4,1,163/6,1,0,3,59/3,2,20,2,22,1,93/4,1,71/3,2,24,1,101/4,1,
            77/3,3,0,16,119/6,1,121/6,1,41/2,1,0,19,17,1,91/5,1,55/3,3,0,24,16,
            1,491/30,2,167/10,2,0,29,13,1,223/15,2,15,1,46/3,1,0,34,51/4,4,79/6,
            1,0,39,173/15,1,0];
        I:=[416610,416875,416378,416428,418079,418080,416986,418148,417963,
            418172,418233,417281,416831,416834,416870,417298,418231,418249,
            417975,416991,416951,418194,418227,418537,417139,417307,416887,
            416976,418248,415441,417686,414926,416997,418085,418315,418316,
            418498,415432,416798];
    elif hash1 eq 2080 and hash2 eq 492 and hash3 eq 2321 then
        V:=[17,1,18,9,19,15,20,27,21,37,22,47,23,59,24,65,25,73];
        D:=[1,61/2,1,73/2,1,77/2,1,0,4,104/3,1,116/3,1,0,6,151/6,1,163/6,1,
            175/6,1,123/4,1,187/6,1,0,11,70/3,1,71/3,2,77/3,1,109/4,1,0,16,
            131/6,2,133/6,1,143/6,1,253/10,1,0,21,73/4,1,39/2,1,239/12,1,20,1,
            61/3,2,0,27,391/20,1,599/30,1,0,29,17,1,369/20,1,278/15,1,0,32,
            181/12,1,31/2,1,0];
        I:=[254762,254755,254760,427349,427043,427418,427408,427209,427010,
            427189,427042,427184,427335,427193,427165,427425,427508,427420,
            427223,426786,427047,427035,427198,427188,427041,427049,426634,
            426729,427053,426536,426514,426724,426725];
    elif hash1 eq 2268 and hash2 eq 3078 and hash3 eq 899 then
        V:=[29,1,30,5,32,9,33,21,34,27,35,33];
        D:=[1,103/4,1,0,2,24,1,0,3,199/12,1,253/15,1,17,3,131/7,1,283/15,1,0,10,
            451/30,1,181/12,1,0,12,66/5,2,53/4,1,0,15,45/4,1,34/3,2,137/12,1,0];
        I:=[230454,233906,230463,230995,231723,232191,233277,228993,228961,
            229711,230455,229198,232526,234026,231528,232993,233816,233822];
    elif hash1 eq 3396 and hash2 eq 3400 and hash3 eq 993 then
        V:=[30,1,31,7,32,19,33,23];
        D:=[1,59/3,4,85/4,1,0,6,177/10,1,71/4,1,107/6,3,109/6,2,197/10,1,0,14,
            49/3,2,0,16,173/12,1,0];
        I:=[59543,59642,59651,59717,58895,58846,58893,59445,59592,59698,59601,
            59670,58666,58674,59634,58815];
    elif hash1 eq 1422 and hash2 eq 2993 and hash3 eq 3232 then
        V:=[18,1,19,9,20,21,21,37,22,51,23,65,24,71,25,79,26,87];
        D:=[1,25,2,27,3,92/3,1,0,7,43/2,1,139/6,1,151/6,3,163/6,1,55/2,1,0,14,
            20,1,65/3,2,22,1,70/3,1,71/3,1,24,2,101/4,2,0,24,37/2,1,121/6,2,
            41/2,3,87/4,1,131/6,1,103/4,1,0,33,55/3,4,56/3,1,101/5,1,61/3,2,
            109/5,1,67/3,1,0,43,33/2,1,103/6,1,0,45,44/3,1,46/3,1,47/3,1,0,48,
            461/30,1,31/2,1,95/6,1,0,51,37/3,1,0];
        I:=[425619,425622,425617,425620,425621,425605,425580,425588,425585,
            425594,425599,425612,425543,425583,425591,425598,425623,425614,
            425541,425542,425544,425534,425550,425529,425587,425592,425540,
            425616,425618,425539,425536,425482,425495,425530,425577,425582,
            425538,425491,425488,425496,425509,425481,425581,425545,425532,
            425492,425493,425475,425474,425476,425485];
    elif hash1 eq 2691 and hash2 eq 1593 and hash3 eq 44 then
        V:=[17,1,18,7,19,15,20,21,21,25,22,29,23,33,25,39,26,43];
        D:=[1,53/2,1,65/2,1,0,3,80/3,1,33,2,116/3,1,0,7,175/6,1,199/6,2,0,10,
            77/3,1,0,11,257/10,1,0,12,49/3,1,0,13,491/30,1,221/12,1,0,15,451/30,
            2,0,17,316/21,1,0];
        I:=[516749,516559,516735,516562,516747,516630,516754,516617,516746,
            516646,516206,516733,516214,516634,516188,516201,515701];
    elif hash1 eq 3020 and hash2 eq 857 and hash3 eq 929 then
        V:=[23,1,24,5,25,11,26,23];
        D:=[1,47/2,1,0,2,53/3,1,65/3,1,0,4,107/6,1,109/6,5,37/2,1,41/2,1,87/4,1,
            0,13,49/3,1,50/3,1,56/3,1,0];
        I:=[247000,247440,247269,247271,247258,247456,247464,247465,247484,
            247471,247432,246532,247211,246997,246892];
    elif hash1 eq 106 and hash2 eq 3380 and hash3 eq 1767 then
        V:=[23,1,24,7,25,19,26,33,27,37,28,43,29,47,30,53,31,57];
        D:=[1,139/4,1,211/6,1,0,3,71/3,2,88/3,1,95/3,1,32,2,101/3,1,0,10,107/6,
            1,109/6,1,237/10,1,155/6,1,329/12,1,277/10,1,0,16,268/15,1,0,17,
            491/30,1,41/2,1,0,19,344/21,1,0,20,401/30,1,973/60,1,0,22,41/3,1,0,
            23,499/42,1,0];
        I:=[410513,411582,411578,411685,411688,411293,411701,412075,412074,
            411568,411574,411534,409902,411519,409332,411195,410684,411589,
            409314,411200,409529,411406,409172];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 13 mod 47.
function cached_ids_13(P,hash1,hash2,hash3)
    if hash1 eq 3256 and hash2 eq 2605 and hash3 eq 2419 then
        V:=[20,1,23,5,24,13,25,23,26,27,27,33,28,37,29,43];
        D:=[1,35,1,0,2,119/4,2,181/6,1,191/6,1,0,6,358/15,1,287/12,1,24,1,73/3,
            1,0,10,51/2,1,0,11,308/15,1,251/12,1,0,13,521/30,1,0,14,359/21,1,
            69/4,1,0,16,817/60,2,137/10,3,55/4,1,0];
        I:=[423048,423065,423757,423755,423756,422767,423378,423233,423718,
            422442,422353,422655,423711,422444,422443,422526,423376,423268,
            423355,423709,422715];
    elif hash1 eq 1282 and hash2 eq 3268 and hash3 eq 1892 then
        V:=[21,1,22,7,23,17,24,25,25,31,26,35,27,41];
        D:=[1,57/2,2,69/2,1,0,4,86/3,1,92/3,1,31,1,37,1,0,8,51/2,2,163/6,1,
            199/6,2,0,13,22,2,88/3,1,0,16,133/6,1,0,17,239/12,1,61/3,2,0,20,
            551/30,1,22,1,0];
        I:=[410536,411350,410865,411541,411109,411306,411335,410812,411537,
            411545,410484,410619,410590,411263,411337,410670,411349,410693,
            410983,411103,409435];
    elif hash1 eq 1705 and hash2 eq 875 and hash3 eq 2967 then
        V:=[24,1,25,11,26,21,27,29,28,39,29,47];
        D:=[1,23,1,74/3,4,25,3,29,1,0,10,103/6,1,115/6,1,139/6,2,151/6,2,0,16,
            52/3,1,18,1,20,1,0,19,157/10,2,95/6,1,97/6,1,107/6,2,0,25,232/15,1,
            50/3,1,91/5,1,0,28,359/30,1,12,1,77/6,1,431/30,1,319/20,2,16,1,33/2,
            1,0];
        I:=[400534,399614,399680,400238,400458,398903,400454,400533,399029,
            400256,399867,398600,399515,399299,400444,399740,400437,398593,
            399612,399656,399864,400435,395879,399592,397333,399027,398061,
            397121,397174,400232,398484,396086,397938,398017,395877];
    elif hash1 eq 1470 and hash2 eq 1367 and hash3 eq 1397 then
        V:=[19,1,20,7,21,21,22,33,23,43,24,53,26,59];
        D:=[1,65/2,1,73/2,1,0,3,80/3,1,27,1,86/3,2,29,1,31,1,33,1,0,10,43/2,1,
            151/6,1,51/2,1,55/2,1,59/2,1,0,15,85/4,1,70/3,1,71/3,1,83/3,1,0,19,
            177/10,1,107/6,1,121/6,1,133/6,3,0,25,215/12,1,55/3,1,0,27,179/12,1,
            15,1,0];
        I:=[253002,252985,253022,252990,252991,253006,253016,253014,253012,
            253023,252978,253003,252989,252776,253019,252986,253021,252788,
            252777,252770,252992,252764,252768,252806,252808,252784,252993,
            252984];
    elif hash1 eq 248 and hash2 eq 1874 and hash3 eq 3385 then
        V:=[16,1,17,5,18,11,19,23,20,31,21,39,22,45,23,49,24,55,26,63];
        D:=[1,28,1,0,2,53/2,1,69/2,1,0,4,86/3,1,29,1,92/3,1,98/3,1,41,1,0,9,
            47/2,1,151/6,1,51/2,1,0,12,22,1,26,2,83/3,1,0,16,71/4,2,197/10,1,0,
            19,298/15,1,0,20,193/12,1,359/20,2,0,23,223/15,1,179/12,1,283/15,1,
            0,26,419/28,1,187/12,1,0];
        I:=[516544,516548,516545,516496,516112,516526,516615,516106,516660,
            516578,516659,516113,516190,516664,515946,516394,516647,516128,
            516142,515873,515806,516293,515923,516416,516296,515771,515689];
    elif hash1 eq 2222 and hash2 eq 163 and hash3 eq 1091 then
        V:=[19,1,20,9,21,15,22,21,23,25,24,29,25,33,26,41];
        D:=[1,41/2,1,49/2,1,53/2,1,0,4,80/3,1,29,1,0,6,51/2,2,211/6,1,0,9,18,1,
            0,10,109/6,1,0,11,18,1,0,12,279/20,1,431/30,1,671/30,1,0,15,242/15,
            1,0];
        I:=[512685,512810,512675,512680,512633,511912,512125,511964,512837,
            512128,511996,512190,511998,511207,511196];
    elif hash1 eq 2833 and hash2 eq 1534 and hash3 eq 3457 then
        V:=[21,1,22,7,23,17,24,25,25,33,26,41];
        D:=[1,139/6,1,47/2,1,0,3,64/3,1,65/3,1,22,2,70/3,1,0,8,119/6,1,121/6,2,
            133/6,1,0,12,73/4,1,55/3,5,56/3,4,0,22,197/12,1,33/2,1,187/10,2,0,
            26,15,1,0];
        I:=[252348,252378,252353,252394,252188,252200,252187,252380,252186,
            252201,251936,252199,252143,252198,252258,252260,252392,251937,
            252161,252185,252190,252026,251986,251930,251983,252029];
    elif hash1 eq 2598 and hash2 eq 332 and hash3 eq 1401 then
        V:=[19,1,20,9,21,21,22,31,23,47,24,63,25,69];
        D:=[1,49/2,2,53/2,3,61/2,1,0,7,21,1,23,5,74/3,3,25,2,27,2,0,20,39/2,4,
            127/6,6,43/2,2,151/6,2,0,34,18,1,58/3,1,59/3,4,20,1,85/4,3,64/3,1,
            65/3,5,0,50,71/4,1,107/6,3,109/6,3,197/10,1,121/6,2,41/2,1,87/4,1,0,
            62,49/3,7,50/3,2,0,71,29/2,2,89/6,1,91/6,1,0];
        I:=[252470,252476,252464,252474,252475,252381,252463,252275,252423,
            252459,252472,252473,252309,252385,252417,252286,252448,252124,
            252373,252276,252301,252317,252410,252231,252298,252303,252310,
            252358,252367,252371,252429,252010,252074,252078,252127,252134,
            252308,252355,252375,252471,252075,252194,252222,252162,251880,
            251973,251976,252083,252089,252056,252073,252232,252261,251979,
            252014,252255,251993,251884,252004,252012,251944,251978,252025,
            252039,252307,252401,252413,252426,252079,252095,252067,252087,
            251985,252023];
    elif hash1 eq 2598 and hash2 eq 1938 and hash3 eq 3144 then
        V:=[22,1,23,9,24,13,25,19,27,29];
        D:=[1,74/3,1,25,2,27,1,0,5,139/6,1,0,6,59/3,2,65/3,1,0,9,83/6,1,107/6,2,
            121/6,1,41/2,1,0,14,173/12,1,29/2,1,0];
        I:=[415654,415643,415653,415649,415670,415645,415823,415584,415700,
            415624,415735,415608,415596,415619,415586];
    elif hash1 eq 154 and hash2 eq 483 and hash3 eq 783 then
        V:=[19,1,20,9,21,23,22,37,23,49,24,71,25,83];
        D:=[1,45/2,3,53/2,1,65/2,1,0,6,62/3,1,68/3,3,23,2,25,1,27,2,86/3,1,0,16,
            115/6,6,39/2,4,83/4,1,127/6,2,43/2,2,55/2,1,0,32,52/3,1,18,2,77/4,1,
            59/3,1,20,1,0,38,97/6,1,209/12,1,177/10,1,71/4,2,107/6,2,109/6,5,
            197/10,1,121/6,1,41/2,1,87/4,1,0,54,238/15,1,49/3,3,50/3,1,18,1,22,
            1,0,61,147/10,1,91/6,1,0];
        I:=[424772,424836,424848,424852,424602,424818,423868,423872,424350,
            424506,424809,424018,423476,424483,424145,424502,424607,424761,
            424791,424807,424826,424407,424598,424771,424849,424371,424567,
            424698,424107,424793,423574,424762,424143,424219,424513,424671,
            424505,423430,424045,423418,423609,423877,424443,424672,423459,
            423614,424224,424240,424328,422567,422566,424262,423077,423428,
            423442,423505,424479,424495,422559,422518,423429,423460];
    elif hash1 eq 2598 and hash2 eq 1692 and hash3 eq 30 then
        V:=[24,1,25,13,26,21,27,27,28,31];
        D:=[1,18,1,59/3,3,20,1,64/3,1,65/3,1,0,8,33/2,1,107/6,2,121/6,1,0,12,
            49/3,1,50/3,2,0,15,29/2,1,0,16,40/3,1,0];
        I:=[247013,246641,246976,247030,247079,246965,246585,247168,246998,
            247116,246636,246593,246699,247045,246974,246846];
    elif hash1 eq 2927 and hash2 eq 1882 and hash3 eq 2780 then
        V:=[16,1,17,7,18,17,19,31,20,45,21,57,22,67,23,85,26,97];
        D:=[1,104/3,1,39,1,0,3,175/6,1,187/6,2,63/2,1,199/6,1,0,8,77/3,1,83/3,1,
            88/3,1,89/3,2,95/3,4,32,1,0,18,49/2,1,257/10,1,155/6,7,157/6,1,
            119/4,1,193/6,1,0,30,22,2,73/3,2,311/12,1,26,1,335/12,1,0,37,22,1,
            265/12,1,269/12,1,313/12,1,0,41,223/12,1,283/15,2,227/12,1,19,3,
            96/5,1,58/3,4,59/3,1,133/6,1,0,55,56/3,1,189/10,1,571/30,2,39/2,2,
            119/6,1,0,62,49/4,1,0];
        I:=[255293,255220,255232,255212,255254,255309,255246,255282,255285,
            255224,255139,255245,255217,255230,255297,255311,255313,255244,
            255069,255052,255120,255223,255227,255248,255257,255277,255060,
            255074,255044,255286,255292,255041,255289,255131,255119,255090,
            255080,255133,255029,255009,255098,255032,255038,255164,255093,
            255109,255234,255034,255033,255094,255157,255229,255221,255065,
            255012,255045,255036,255056,255050,255062,255031,255066];
    elif hash1 eq 2974 and hash2 eq 2921 and hash3 eq 1767 then
        V:=[24,1,25,7,26,11,27,17,28,33,29,41];
        D:=[1,27,1,29,1,0,3,39/2,3,0,6,59/3,1,20,1,0,8,107/6,2,109/6,4,37/2,1,
            197/10,1,79/4,1,121/6,1,41/2,1,0,19,49/3,2,50/3,1,18,1,0,23,431/30,
            1,173/12,1,29/2,4,491/30,1,33/2,3,0];
        I:=[241414,241581,240681,240765,241354,241488,239819,241143,241503,
            240102,240696,240699,241500,241608,239152,239163,239148,241319,
            238921,239226,238384,240420,239216,239116,240667,241413,241482,
            241536,238412,238388,238391,238392];
    elif hash1 eq 2739 and hash2 eq 2229 and hash3 eq 2299 then
        V:=[20,1,21,11,22,23,23,41,24,59,25,71,26,91,27,101,28,105];
        D:=[1,27,1,29,1,92/3,1,31,1,0,5,139/6,1,47/2,1,151/6,3,163/6,1,55/2,1,0,
            12,85/4,1,64/3,1,65/3,2,22,1,70/3,2,71/3,5,24,1,77/3,1,0,26,109/6,1,
            79/4,1,119/6,1,121/6,5,41/2,1,257/12,1,217/10,1,133/6,1,0,38,73/4,2,
            55/3,7,56/3,2,101/5,1,61/3,1,0,51,91/6,1,197/12,2,33/2,4,101/6,1,
            551/30,2,221/12,1,37/2,2,187/10,1,113/6,2,0,67,223/15,1,179/12,1,15,
            3,46/3,1,0,73,461/30,1,0,74,12,1,0];
        I:=[252349,252389,252428,252291,252300,252270,252386,252421,252424,
            252245,252225,252149,252384,252362,252446,252236,252230,252247,
            252088,252239,252299,252372,252409,252240,251947,252111,252125,
            252077,252243,252282,252356,252374,252441,252284,252058,251948,
            251891,252183,252220,251883,251889,252242,252293,252294,252365,
            252387,252223,252316,252020,251914,251873,252086,252138,252041,
            252123,252142,252227,252250,251975,251981,251849,251860,251861,
            251845,251974,251987,251888,252034,251971,252032,252050,252048,
            251831,252011];
    elif hash1 eq 1047 and hash2 eq 2729 and hash3 eq 2071 then
        V:=[21,1,22,7,23,21,24,29,25,37,27,47];
        D:=[1,45/2,1,57/2,1,0,3,56/3,1,23,1,74/3,1,25,4,92/3,1,31,1,0,12,83/4,1,
            127/6,2,163/6,1,0,16,47/3,1,77/4,1,71/3,1,0,19,177/10,1,107/6,1,
            109/6,3,133/6,1,0,25,431/30,2,29/2,1,33/2,1,0];
        I:=[251427,251003,251412,251431,251138,251000,251428,251435,251437,
            250637,251143,251053,251135,251462,250605,251456,251050,251204,
            250635,251040,250580,251237,251244,249886,250528,250532,250575,
            250475];
    elif hash1 eq 1047 and hash2 eq 1571 and hash3 eq 1204 then
        V:=[27,1,28,7,29,13,30,21,31,29];
        D:=[1,79/4,2,121/6,1,0,4,49/3,2,55/3,2,0,8,59/4,1,197/12,1,101/6,1,0,11,
            175/12,1,223/15,2,15,2,0,16,161/12,1,27/2,1,461/30,1,0];
        I:=[239581,239584,239655,239525,239936,238826,239515,239543,238816,
            239407,239656,239437,239585,238883,239496,239737,238108,238021];
    elif hash1 eq 577 and hash2 eq 1186 and hash3 eq 3084 then
        V:=[19,1,20,5,21,15,22,25,23,31,24,37];
        D:=[1,69/2,2,0,3,23,1,74/3,1,25,2,92/3,1,0,8,39/2,1,151/6,1,51/2,2,55/2,
            1,0,13,65/3,1,101/4,1,0,15,107/6,1,109/6,1,0,17,49/3,1,56/3,2,0];
        I:=[424786,424803,424706,424668,423863,424802,424798,424696,424679,
            424140,424804,423435,423589,424372,424820,424768,422870,423878,
            424795];
    elif hash1 eq 3115 and hash2 eq 88 and hash3 eq 2680 then
        V:=[21,1,22,7,23,13,24,21,25,29,26,35,27,39,28,45];
        D:=[1,57/2,1,81/2,1,0,3,86/3,2,29,1,0,6,163/6,1,175/6,1,59/2,1,0,9,64/3,
            1,70/3,2,82/3,1,0,13,233/12,1,145/6,1,0,15,35/2,1,0,16,33/2,1,101/6,
            1,0,18,209/20,1,21/2,1,85/6,2,0];
        I:=[537746,537642,537650,537663,537731,537649,537666,537664,537736,
            537657,537739,537623,537720,537661,537671,537639,537675,537629,
            537636,537672,537747];
    elif hash1 eq 3068 and hash2 eq 764 and hash3 eq 3264 then
        V:=[23,1,24,5,25,11,26,19,27,25,28,29,29,35];
        D:=[1,187/6,1,0,2,65/3,1,77/3,1,0,4,131/6,1,133/6,1,181/6,1,0,7,328/15,
            1,67/3,3,0,11,221/12,1,0,12,223/15,1,15,2,0,15,169/10,1,511/30,1,0];
        I:=[415146,415184,415200,415206,415148,415486,415155,415166,415179,
            415228,415236,415284,415147,415192,415167,415294];
    elif hash1 eq 295 and hash2 eq 2439 and hash3 eq 3382 then
        V:=[22,1,23,5,24,13,25,29,26,33,27,41,29,49];
        D:=[1,25,1,0,2,127/6,1,139/6,1,47/2,1,0,5,53/3,1,18,2,96/5,1,64/3,1,24,
            1,76/3,1,77/3,1,0,13,257/12,1,0,14,268/15,1,55/3,1,239/12,2,0,18,
            431/30,1,89/6,1,101/6,1,0,21,541/42,1,401/30,1,27/2,1,83/6,1,0];
        I:=[411211,411216,410674,410232,410414,410396,410413,410252,410427,
            410263,410311,409686,410591,410061,409552,410818,410999,410373,
            410200,410712,409348,410954,409029,410848];
    elif hash1 eq 13 and hash2 eq 2522 and hash3 eq 2989 then
        V:=[21,1,22,7,23,13,24,21,25,25,26,35];
        D:=[1,53/2,1,57/2,1,0,3,68/3,1,25,3,0,7,35/2,1,39/2,2,127/6,2,0,12,59/3,
            1,0,13,177/10,1,107/6,1,109/6,1,37/2,2,0,18,16,1,49/3,1,0];
        I:=[251468,250796,250795,250597,250798,250899,250870,251488,251565,
            250803,250968,250727,249884,250512,249881,250378,250887,250713,
            250822];
    elif hash1 eq 2410 and hash2 eq 3281 and hash3 eq 786 then
        V:=[24,1,25,9,26,19,27,23,28,33,29,45,30,53];
        D:=[1,21,1,68/3,1,86/3,1,0,4,115/6,2,39/2,1,151/6,1,51/2,1,0,9,53/3,2,0,
            11,209/12,1,109/6,1,79/4,2,41/2,2,0,17,44/3,1,238/15,1,81/5,1,50/3,
            1,262/15,1,0,22,29/2,1,491/30,1,33/2,1,0,25,40/3,1,223/15,1,0];
        I:=[244051,243720,244395,243026,243960,244401,242372,244294,244402,
            244610,243112,242587,243345,244122,244359,244407,244170,243024,
            243517,242167,242186,244165,242179,242780,243230,241864];
    elif hash1 eq 389 and hash2 eq 1064 and hash3 eq 2420 then
        V:=[17,1,18,5,19,15,20,21,21,31,22,35,23,43,24,47,25,51];
        D:=[1,77/2,1,0,2,29,2,92/3,1,98/3,1,33,2,0,8,151/6,1,175/6,1,0,10,101/4,
            1,76/3,1,77/3,2,83/3,1,0,15,103/4,1,0,16,239/12,1,61/3,1,67/3,2,0,
            20,241/12,1,0,21,302/15,1,0,22,451/30,1,181/12,1,91/6,1,0];
        I:=[254597,254643,254644,254604,254412,254608,254609,254488,254425,
            254428,254431,254417,254430,254416,254381,254414,254429,254426,
            254440,254380,254366,254373,254409,254379];
    elif hash1 eq 2269 and hash2 eq 2149 and hash3 eq 3107 then
        V:=[25,1,26,11,27,19,28,27,30,33];
        D:=[1,35/2,1,115/6,3,39/2,1,127/6,1,0,7,53/3,1,59/3,1,20,1,0,10,97/6,1,
            177/10,1,233/12,1,0,13,238/15,1,49/3,1,0,15,193/15,2,0];
        I:=[243741,243776,243847,243871,243532,243578,243050,243107,243862,
            243124,243318,243955,243079,243653,241909,243579];
    elif hash1 eq 1658 and hash2 eq 2281 and hash3 eq 2207 then
        V:=[13,1,15,5,16,9,17,17,18,27,19,35,20,47,21,59,22,69,23,75,25,87,27,
            101];
        D:=[1,89/2,1,0,2,223/6,1,0,3,119/3,1,136/3,1,137/3,1,0,6,203/6,1,397/10,
            1,241/6,1,477/10,1,0,10,97/3,1,455/12,2,278/7,1,0,14,361/12,1,
            971/30,5,65/2,1,1031/30,1,69/2,1,0,23,122/5,1,398/15,1,136/5,1,
            601/20,1,452/15,1,0,28,1573/60,1,1129/42,1,381/14,1,821/30,1,0,32,
            293/15,1,323/15,1,0,34,545/28,1,821/42,1,1177/60,1,679/30,1,91/4,1,
            0,39,199/14,1,427/30,3,431/30,2,238/15,1,319/20,1,16,1,0,48,443/36,
            1,517/42,1,260/21,1,0];
        I:=[519109,519107,519121,519105,519084,519122,519129,519140,518921,
            519101,519083,519096,518924,519114,519090,519100,519110,519113,
            519142,519081,519091,519123,519060,519099,519133,519094,519076,
            519055,518933,518926,518915,519068,519070,518913,518901,518868,
            518914,518912,518929,518906,519069,519134,519111,519141,518905,
            518909,518932,518848,518847,518849];
    elif hash1 eq 2316 and hash2 eq 1238 and hash3 eq 2663 then
        V:=[21,1,22,9,23,17,24,29,25,41,26,51,27,65];
        D:=[1,39/2,1,127/6,2,63/2,1,0,5,77/3,2,82/3,1,117/4,1,0,9,71/4,1,79/4,2,
            121/6,2,237/10,1,143/6,1,0,16,50/3,1,55/3,1,99/5,1,239/12,1,61/3,1,
            0,21,101/6,1,539/30,1,18,1,659/30,1,0,25,44/3,1,179/12,1,46/3,1,
            199/12,1,18,1,37/2,1,0,31,27/2,1,0];
        I:=[513927,513920,513932,513844,513800,513865,513941,513507,513857,
            513835,513842,513918,513926,513762,513588,513813,513685,513748,
            513624,513919,513929,513749,513733,513311,513924,513724,513792,
            513431,513520,513304,513649];
    elif hash1 eq 1329 and hash2 eq 225 and hash3 eq 1784 then
        V:=[23,1,24,5,25,13,26,17,27,21,28,25];
        D:=[1,53/2,2,0,3,19,2,68/3,1,23,1,0,7,115/6,1,0,8,20,1,0,9,97/6,3,0,12,
            167/12,1,14,2,43/3,3,0];
        I:=[241333,241335,240717,240728,241353,240729,240760,240751,240644,
            240663,240716,240670,241205,241274,239817,239858,240642];
    elif hash1 eq 2175 and hash2 eq 29 and hash3 eq 1014 then
        V:=[22,1,23,7,24,15,25,33,26,39,27,43];
        D:=[1,25,1,31,1,0,3,163/6,1,55/2,1,67/2,1,0,6,59/3,1,64/3,1,65/3,2,22,1,
            71/3,1,77/3,1,82/3,1,83/3,1,0,15,131/6,1,281/12,1,0,17,239/12,1,0,
            18,101/6,1,221/12,1,0];
        I:=[502662,502768,502185,502487,502443,502226,502614,502658,502775,
            502084,502518,502427,502523,502661,502086,502142,502723,502225,
            502719];
    elif hash1 eq 2034 and hash2 eq 798 and hash3 eq 1298 then
        V:=[27,1,28,11,31,21,33,29];
        D:=[1,163/6,1,175/6,1,59/2,1,187/6,1,0,5,70/3,1,71/3,3,24,1,77/3,1,0,11,
            167/10,1,67/4,2,101/6,2,0,16,79/6,3,0];
        I:=[668774,668188,668776,668182,668564,668179,668186,668778,668180,
            668775,668101,667980,668548,668128,668773,668505,668560,668833];
    elif hash1 eq 530 and hash2 eq 470 and hash3 eq 2641 then
        V:=[19,1,20,7,21,17,22,23,23,27,24,33,25,37];
        D:=[1,49/2,1,53/2,1,0,3,23,2,74/3,1,25,1,98/3,1,0,8,43/2,2,151/6,1,0,11,
            53/3,1,0,12,209/12,1,121/6,2,0,15,50/3,1,0,16,147/10,1,319/20,2,0];
        I:=[420091,419955,419931,419990,419963,419991,420132,419497,419498,
            420131,419972,420133,419294,419907,419212,419731,419088,419559];
    elif hash1 eq 342 and hash2 eq 1592 and hash3 eq 3149 then
        V:=[25,1,26,5,27,17,28,31,29,37,30,43,31,51,32,57];
        D:=[1,163/6,1,0,2,59/3,1,93/4,1,70/3,1,24,2,77/3,1,0,8,107/6,2,197/10,1,
            121/6,1,131/6,1,133/6,1,281/12,1,0,15,56/3,1,19,1,0,17,193/12,1,
            491/30,1,0,19,40/3,1,218/15,1,179/12,1,0,22,391/30,1,27/2,1,0,24,
            178/15,1,0];
        I:=[245577,245397,244289,245644,245199,245459,245198,244162,245119,
            243724,245052,242495,241875,244509,245013,245452,245122,241898,
            244028,243120,245632,243260,244441,241932];
    elif hash1 eq 859 and hash2 eq 533 and hash3 eq 3271 then
        V:=[24,1,25,11,26,21,27,31,28,41,29,49,30,53];
        D:=[1,62/3,1,21,2,68/3,2,23,3,0,9,35/2,1,115/6,2,39/2,1,127/6,2,0,15,16,
            2,53/3,2,18,1,59/3,1,0,21,157/10,1,95/6,3,97/6,4,177/10,1,0,30,71/5,
            1,57/4,1,44/3,2,0,34,77/6,1,0,35,35/3,1,0];
        I:=[240629,241224,241331,240712,240817,240718,240930,241279,241013,
            240723,241186,241374,239433,239436,241040,241149,239827,240124,
            239940,241359,239935,239926,240543,240879,239636,241171,241197,
            241284,239916,240156,239550,239726,241049,240146,239335];
    elif hash1 eq 60 and hash2 eq 2621 and hash3 eq 565 then
        V:=[23,1,24,7,25,19,26,25,27,35,29,41,30,53,31,67];
        D:=[1,163/6,1,59/2,1,0,3,77/3,2,146/5,1,89/3,1,30,1,100/3,1,0,9,145/6,1,
            69/2,1,0,11,268/15,1,55/3,1,239/12,2,20,1,0,16,491/30,1,221/12,1,0,
            18,83/6,2,181/12,1,95/6,2,205/12,1,189/10,2,0,26,178/15,2,66/5,2,
            41/3,2,89/6,1,61/4,1,47/3,1,0,35,347/30,1,817/60,1,41/3,1,0];
        I:=[502837,503241,502945,503205,503123,503266,503262,502924,503236,
            502909,503387,503176,501666,502890,503212,503175,503240,502959,
            502962,503152,501937,501939,501881,501695,501803,503285,503316,
            503023,503203,502879,502900,502051,501739,501440,503288,501267,
            501273];
    elif hash1 eq 1658 and hash2 eq 1669 and hash3 eq 3407 then
        V:=[22,1,23,5,24,19,25,27,26,35,27,45,28,55,29,65];
        D:=[1,80/3,1,0,2,151/6,1,55/2,1,175/6,1,59/2,1,187/6,1,63/2,1,0,8,93/4,
            1,70/3,1,71/3,1,0,11,107/6,1,121/6,1,145/6,1,0,14,81/5,2,49/3,1,
            215/12,1,18,1,0,19,16,1,33/2,1,103/6,1,35/2,1,0,23,155/12,2,289/20,
            1,179/12,1,401/20,1,0,28,38/3,1,259/20,1,391/30,1,0];
        I:=[418327,416926,418225,416881,417985,417875,418020,416848,416868,
            418242,416916,417308,416727,417419,417640,417373,417295,417064,
            416604,418190,416847,418228,418050,418617,415444,414980,415021,
            418257,417514,416440];
    elif hash1 eq 201 and hash2 eq 1451 and hash3 eq 97 then
        V:=[12,1,13,9,14,17,15,29,16,39,17,49,18,55,19,61,21,69,22,75,24,83,25,
            87];
        D:=[1,42,1,44,1,50,1,0,4,85/2,1,89/2,1,101/2,1,0,7,98/3,1,110/3,1,37,2,
            116/3,1,140/3,1,0,13,131/4,1,199/6,1,235/6,2,259/6,1,0,18,89/3,1,
            157/4,1,130/3,1,131/3,1,0,22,297/10,3,425/12,1,0,26,28,1,36,1,0,28,
            45/2,1,317/12,1,337/12,1,0,31,56/3,1,571/30,1,0,33,523/28,1,115/6,1,
            139/6,1,0,36,947/60,1,0,37,367/30,1,0];
        I:=[430054,430034,430027,430028,430035,430033,430029,430045,430044,
            430069,430050,429992,429998,430049,430010,430032,429989,430042,
            429994,429968,429973,429978,429979,429984,429985,429991,429967,
            429990,429972,429974,429999,429980,429962,429959,429961,429960,
            429986];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 14 mod 47.
function cached_ids_14(P,hash1,hash2,hash3)
    if hash1 eq 1894 and hash2 eq 2457 and hash3 eq 2523 then
        V:=[25,1,26,5,27,15,28,25,29,29,30,39];
        D:=[1,169/6,1,0,2,263/12,2,111/5,1,68/3,1,287/12,1,0,7,221/12,1,37/2,1,
            245/12,1,41/2,1,0,11,17,1,0,12,461/30,1,31/2,1,157/10,1,95/6,1,0,16,
            163/12,1,91/6,1,0];
        I:=[246571,246669,246701,246825,246777,246703,246692,246672,246690,
            246751,246670,246812,246774,246635,246583,246606,246459];
    elif hash1 eq 2552 and hash2 eq 1327 and hash3 eq 79 then
        V:=[18,1,19,9,20,17,21,25,23,31];
        D:=[1,28,1,30,1,32,1,0,4,49/2,3,53/2,1,57/2,1,0,9,62/3,1,21,1,25,2,0,13,
            127/6,1,43/2,3,0,17,109/6,1,0];
        I:=[674530,674529,674543,674496,674534,674544,674541,674484,674495,
            674490,674412,674482,674416,674418,674483,674504,674414];
    elif hash1 eq 766 and hash2 eq 339 and hash3 eq 1996 then
        V:=[20,1,21,13,22,21,23,29,24,33];
        D:=[1,25,1,80/3,3,27,1,29,1,92/3,1,0,8,127/6,1,47/2,1,51/2,1,0,11,77/4,
            1,58/3,1,22,1,0,14,233/12,1,0,15,50/3,1,91/5,1,56/3,1,0];
        I:=[253676,253629,253726,253897,253892,253747,253777,253703,253663,
            253891,253739,253623,253707,253542,253500,253493,253505];
    elif hash1 eq 3210 and hash2 eq 144 and hash3 eq 2301 then
        V:=[25,1,26,11,27,17,28,23,29,33];
        D:=[1,127/6,1,43/2,1,139/6,1,47/2,1,0,5,58/3,1,65/3,1,0,7,107/6,2,121/6,
            1,0,10,65/4,1,49/3,3,50/3,2,73/4,1,0,17,173/12,1,0];
        I:=[670641,670669,670838,670707,670850,670078,670110,670174,670052,
            670055,670150,670657,670713,670091,670692,670023,670228];
    elif hash1 eq 1894 and hash2 eq 91 and hash3 eq 147 then
        V:=[24,1,25,11,26,23,27,27,28,33,29,43];
        D:=[1,23,1,74/3,1,25,3,80/3,1,0,7,83/4,1,127/6,1,43/2,1,139/6,1,151/6,1,
            0,12,59/3,1,0,13,95/6,1,109/6,1,0,15,81/5,2,50/3,2,91/5,1,73/4,3,0,
            23,29/2,3,89/6,1,0];
        I:=[245413,245426,244688,244737,245250,245251,243513,244732,245321,
            242882,245420,245318,245421,245670,242416,242463,242413,242434,
            243002,242485,243602,243615,242421,242534,245291,243611];
    elif hash1 eq 2223 and hash2 eq 933 and hash3 eq 345 then
        V:=[17,1,18,9,19,15,20,19,21,29,22,33,23,37,24,43,25,49];
        D:=[1,65/2,1,69/2,1,73/2,1,0,4,122/3,1,43,1,0,6,63/2,1,0,7,89/3,1,101/3,
            1,186/5,1,124/3,1,0,11,155/6,1,0,12,68/3,1,0,13,187/10,2,115/6,1,0,
            16,278/15,1,308/15,1,0,18,149/10,2,0];
        I:=[543159,543151,543155,543142,543157,543260,543257,543244,543137,
            543144,543146,543100,543108,543249,543233,543173,543181,543171,
            543175];
    elif hash1 eq 155 and hash2 eq 2659 and hash3 eq 660 then
        V:=[16,1,17,11,18,21,19,29,21,35,22,39];
        D:=[1,28,1,32,2,34,1,38,1,0,6,49/2,1,57/2,1,61/2,1,69/2,1,0,10,68/3,1,
            23,1,29,2,0,14,151/6,1,55/2,1,0,16,131/6,1,0,17,55/3,1,0];
        I:=[254611,254572,254610,254612,254575,254632,254521,254630,254515,
            254576,254629,254462,254570,254442,254455,254390,254458];
    elif hash1 eq 2834 and hash2 eq 2385 and hash3 eq 3042 then
        V:=[24,1,25,5,26,13,27,19,28,23,29,31,31,37];
        D:=[1,27,1,0,2,43/2,1,139/6,1,47/2,1,0,5,59/3,2,20,1,0,8,107/6,3,0,11,
            49/3,1,50/3,1,73/4,1,0,14,173/12,1,89/6,1,0,16,23/2,1,0];
        I:=[670577,670763,670680,670783,670087,670760,670698,670035,670062,
            670097,670813,670072,669902,670175,670785,670032];
    elif hash1 eq 2740 and hash2 eq 1713 and hash3 eq 2714 then
        V:=[23,1,24,7,25,23,26,33,27,45,28,49,29,53,30,61];
        D:=[1,175/6,1,211/6,1,0,3,77/3,1,136/5,1,88/3,1,89/3,1,94/3,1,100/3,1,
            141/4,1,0,10,133/6,1,305/12,2,111/4,1,169/6,1,0,15,268/15,1,215/12,
            1,67/3,1,121/5,1,73/3,1,0,20,551/30,1,0,21,223/15,1,0,22,181/12,2,
            91/6,1,511/30,1,0,26,61/4,1,0];
        I:=[409640,410105,411342,410110,411686,411665,411662,411330,411427,
            409658,411176,411524,411469,410112,410724,411322,410496,410094,
            409660,409116,410495,411445,411446,409991,409106,409138];
    elif hash1 eq 2364 and hash2 eq 1166 and hash3 eq 480 then
        V:=[19,1,20,7,21,11,22,25,23,39,24,53,25,59,26,79,27,95];
        D:=[1,175/6,1,199/6,1,0,3,101/3,1,0,4,155/6,1,157/6,1,167/6,3,169/6,1,
            179/6,1,191/6,1,0,12,358/15,1,73/3,2,388/15,1,131/5,1,105/4,1,79/3,
            1,0,19,125/6,2,127/6,1,671/30,1,227/10,1,137/6,1,779/30,1,0,26,
            223/12,2,19,1,0,29,511/30,1,521/30,1,35/2,1,371/20,1,559/30,1,
            379/20,2,571/30,1,229/12,3,233/12,2,0,42,89/6,1,227/15,1,91/6,1,
            61/4,3,233/15,2,47/3,3,1123/60,1,0,54,641/42,1,917/60,1,0];
        I:=[426102,426231,426211,426227,426155,426169,426230,426345,426215,
            426278,426168,425741,425792,426331,425790,425781,425694,425802,
            425767,426342,426249,425750,425793,425804,425652,425807,425824,
            426066,425791,425806,425775,425651,425648,425742,425783,425825,
            425684,425689,425852,425687,425851,425819,425820,426006,425821,
            426050,426054,425770,425809,426228,426256,426259,425624,425647,
            425659];
    elif hash1 eq 1142 and hash2 eq 2125 and hash3 eq 2973 then
        V:=[26,1,27,7,28,17,29,21,30,35,31,45,32,51];
        D:=[1,65/3,1,71/3,3,0,5,119/6,1,121/6,1,87/4,1,131/6,2,0,10,56/3,1,0,11,
            89/6,1,491/30,1,197/12,1,33/2,2,101/6,1,103/6,2,0,19,175/12,1,
            179/12,1,15,4,46/3,2,0,27,79/6,3,161/12,1,0,31,229/20,1,0];
        I:=[239476,239454,239708,239873,239864,239909,240282,238729,239332,
            238050,239753,238563,239722,239811,239923,238525,239716,239992,
            239186,240276,238726,239710,239772,240353,238470,239877,239345,
            239729,240488,239775,238110];
    elif hash1 eq 1753 and hash2 eq 720 and hash3 eq 2857 then
        V:=[19,1,20,5,21,15,22,23,23,29,24,37];
        D:=[1,57/2,1,0,2,25,2,80/3,1,27,3,31,1,0,9,47/2,1,151/6,4,51/2,3,0,17,
            65/3,8,71/3,1,0,26,119/6,2,87/4,1,131/6,3,0,32,215/12,2,91/5,1,55/3,
            7,56/3,3,0];
        I:=[674459,674456,674501,674498,674461,674470,674494,674502,674503,
            674406,674415,674465,674500,674457,674480,674481,674389,674411,
            674417,674458,674460,674471,674475,674505,674388,674405,674407,
            674383,674386,674396,674408,674410,674424,674385,674382,674384,
            674399,674402,674413,674466,674477,674390,674422,674462];
    elif hash1 eq 1518 and hash2 eq 1091 and hash3 eq 1037 then
        V:=[25,1,26,7,27,15,28,23,29,27,30,43,31,51];
        D:=[1,281/12,1,253/10,1,0,3,15,1,18,1,21,1,0,6,173/12,2,103/6,1,539/30,
            1,0,10,155/12,1,0,11,259/20,1,13,1,79/6,1,137/10,1,85/6,2,877/60,1,
            461/30,1,0,19,13,2,197/15,2,53/4,2,0,25,145/12,1,0];
        I:=[508140,508155,507953,507973,507959,508055,508098,507976,508162,
            508136,508132,508077,508067,508043,507956,508006,507780,508165,
            507873,507891,507781,507875,507995,508001,507714];
    elif hash1 eq 2834 and hash2 eq 28 and hash3 eq 1471 then
        V:=[14,1,15,7,16,15,17,33,18,47,19,67,20,81,21,95,22,105,23,113,24,117,
            25,123];
        D:=[1,26,1,28,1,0,3,53/2,1,57/2,1,65/2,1,0,6,23,1,25,2,80/3,1,27,1,29,1,
            31,1,98/3,2,116/3,1,0,16,139/6,2,47/2,2,151/6,4,55/2,1,63/2,2,71/2,
            1,0,28,65/3,2,93/4,1,70/3,1,71/3,1,24,1,76/3,1,26,1,109/4,2,107/3,1,
            0,39,121/6,1,41/2,1,257/12,1,131/6,2,237/10,1,257/10,1,0,46,56/3,1,
            298/15,1,101/5,1,81/4,1,61/3,3,328/15,1,0,54,103/6,2,37/2,1,611/30,
            2,431/20,1,0,60,15,2,223/12,1,271/12,1,0,64,95/6,1,0,65,309/20,1,
            238/15,1,0,67,371/30,1,0];
        I:=[428575,428500,428573,428523,428574,428578,428458,428496,428506,
            428547,428358,428354,428252,428449,428271,428310,428543,428316,
            428356,428302,428374,428508,428542,428286,428045,428368,428274,
            428391,428536,428266,428380,428501,428292,428434,428385,428165,
            428370,427994,428289,428029,428255,427921,428143,427999,428001,
            428376,428142,427988,427937,427955,428037,428364,428002,428000,
            428024,427915,427761,428106,428020,428272,428384,427798,427697,
            428118,427680,427688,428121];
    elif hash1 eq 3116 and hash2 eq 2402 and hash3 eq 1523 then
        V:=[22,1,23,5,25,15,26,27,27,35,28,41,29,49];
        D:=[1,109/4,1,0,2,119/6,1,103/4,1,167/6,1,169/6,1,0,6,217/12,1,241/12,1,
            671/30,1,851/30,1,173/6,1,0,11,248/15,1,369/20,1,19,1,0,14,185/12,1,
            973/60,1,0,16,65/4,1,50/3,1,256/15,1,0,19,481/30,1,97/6,1,0];
        I:=[509990,510159,510002,509999,509973,509972,509975,510102,508355,
            508400,509800,509961,508326,509886,510111,508826,510096,509957,
            508367,508362];
    elif hash1 eq 2881 and hash2 eq 2271 and hash3 eq 490 then
        V:=[25,1,26,11,27,19,30,23,31,31];
        D:=[1,55/2,1,175/6,1,187/6,1,63/2,1,0,5,71/3,1,77/3,1,83/3,2,0,9,131/6,
            1,0,10,155/12,1,223/15,1,46/3,1,0,13,757/60,1,38/3,2,51/4,1,0];
        I:=[245232,245326,242917,245382,245655,243025,242287,242775,242218,
            242531,242439,242430,242501,242502,242528,243681];
    elif hash1 eq 2411 and hash2 eq 1555 and hash3 eq 3378 then
        V:=[23,1,24,7,25,11,26,19,27,27];
        D:=[1,151/6,1,163/6,2,0,4,77/3,1,0,5,119/6,1,121/6,3,131/6,1,0,10,
            215/12,1,55/3,1,56/3,1,0,13,491/30,1,101/6,1,37/2,2,187/10,1,113/6,
            2,0];
        I:=[673477,673394,673610,673481,673255,673253,673627,673636,673392,
            673249,673468,673430,673256,673257,673236,673251,673218,673224,
            673285];
    elif hash1 eq 672 and hash2 eq 2625 and hash3 eq 231 then
        V:=[25,1,26,15,27,27,29,35,31,45,32,51];
        D:=[1,151/6,1,163/6,2,175/6,4,59/2,1,123/4,1,187/6,1,0,11,22,1,116/5,2,
            93/4,1,70/3,1,71/3,4,0,20,79/4,1,119/6,1,121/6,1,0,23,491/30,1,
            167/10,1,101/6,2,103/6,2,0,29,757/60,1,157/12,1,0,31,667/60,1,0];
        I:=[244919,243440,244364,242880,243689,244719,245297,245316,242991,
            243552,243040,243159,243894,242982,245290,243133,243435,244771,
            244883,243222,243441,243678,242425,242428,242422,242424,245155,
            245538,242637,243669,243676];
    elif hash1 eq 954 and hash2 eq 1812 and hash3 eq 2080 then
        V:=[20,1,21,5,22,23,23,31,24,45,25,59,26,77,28,85];
        D:=[1,25,1,0,2,127/6,1,51/2,1,163/6,1,175/6,1,123/4,1,187/6,1,63/2,1,
            71/2,1,0,10,70/3,1,77/3,1,88/3,1,0,13,119/6,1,131/6,1,95/4,1,157/6,
            2,53/2,1,119/4,1,0,20,55/3,1,56/3,1,19,1,101/5,2,73/3,1,311/12,1,0,
            27,217/12,1,551/30,1,37/2,1,187/10,1,113/6,2,115/6,1,241/12,1,137/6,
            1,0,36,248/15,1,272/15,1,337/15,1,0,39,281/21,1,208/15,1,907/60,1,
            91/6,1,0];
        I:=[423028,423136,422765,422825,423546,422969,422772,423138,423325,
            422966,423319,423457,423211,422942,422906,423281,423530,423383,
            423715,423092,422762,422781,423180,423213,422628,423702,422956,
            422915,422763,422981,423242,423280,423254,422910,422550,422869,
            422548,422520,422394,423030,422513,422400];
    elif hash1 eq 1800 and hash2 eq 3241 and hash3 eq 2612 then
        V:=[27,1,28,5,29,11,30,25,32,33];
        D:=[1,47/2,2,0,3,58/3,1,71/3,1,0,5,107/6,2,109/6,1,119/6,1,121/6,4,
            217/10,1,131/6,1,0,15,65/4,1,49/3,2,17,2,0,20,193/15,1,13,1,40/3,4,
            0];
        I:=[666802,666831,666814,666291,666390,666843,666080,666389,666303,
            666457,666620,666834,665418,665350,666546,665165,666307,666314,
            666799,666560,666369,666306,666308,666822,666823];
    elif hash1 eq 3398 and hash2 eq 1352 and hash3 eq 1498 then
        V:=[19,1,20,13,22,21,23,25,25,29,26,39];
        D:=[1,851/30,1,239/8,1,911/30,1,1031/30,1,347/10,1,0,6,187/7,1,227/8,1,
            722/21,2,0,10,601/24,1,0,11,907/42,1,0,12,253/18,1,2957/210,1,
            199/14,1,949/60,1,0,16,634/45,1,509/36,1,0];
        I:=[543957,543902,543911,543958,543921,543938,543909,543916,543923,
            543907,543929,543886,543925,543883,543932,543891,543893];
    elif hash1 eq 2928 and hash2 eq 1935 and hash3 eq 586 then
        V:=[15,1,16,15,17,33,18,43,19,61,20,85,21,107,22,115,23,131,24,141,25,
            147,26,153];
        D:=[1,65/2,1,69/2,1,73/2,2,81/2,1,85/2,2,93/2,1,0,9,92/3,1,98/3,1,33,1,
            104/3,1,35,1,110/3,1,39,1,128/3,1,0,17,115/4,1,175/6,2,59/2,2,187/6,
            1,0,23,77/3,1,117/4,1,89/3,5,125/4,1,94/3,1,95/3,4,101/3,1,141/4,1,
            0,38,143/6,1,145/6,1,257/10,1,155/6,4,157/6,4,167/6,3,353/12,1,
            179/6,1,181/6,1,193/6,1,215/6,1,0,57,328/15,1,263/12,1,22,2,111/5,1,
            67/3,1,358/15,1,24,1,121/5,1,79/3,1,335/12,1,0,68,265/12,1,269/12,2,
            137/6,2,0,73,369/20,1,37/2,1,278/15,1,223/12,4,283/15,2,227/12,3,19,
            3,0,88,781/42,1,229/12,1,68/3,1,23,1,0,92,907/60,1,76/5,1,0,94,
            937/60,1,157/10,1,0,96,61/5,2,0];
        I:=[255678,255573,255597,255668,255673,255669,255677,255663,255675,
            255548,255658,255570,255567,255600,255679,255537,255491,255448,
            255676,255591,255642,255554,255657,255541,255370,255403,255500,
            255502,255617,255475,255446,255571,255576,255592,255640,255681,
            255373,255602,255594,255390,255407,255499,255551,255655,255459,
            255552,255603,255651,255494,255495,255498,255378,255417,255421,
            255393,255439,255408,255503,255665,255674,255387,255496,255396,
            255536,255394,255422,255416,255438,255383,255434,255406,255526,
            255381,255504,255401,255485,255490,255533,255542,255375,255397,
            255367,255457,255466,255364,255450,255660,255352,255351,255353,
            255354,255431,255430,255356,255355,255376,255377];
    elif hash1 eq 3210 and hash2 eq 863 and hash3 eq 2122 then
        V:=[23,1,24,13,25,25,26,39,27,43,28,51,29,61,30,65];
        D:=[1,43/2,1,139/6,2,99/4,1,151/6,1,163/6,1,0,7,20,1,64/3,1,65/3,2,22,1,
            101/4,1,0,13,177/10,1,107/6,1,197/10,1,121/6,1,41/2,1,217/10,1,0,19,
            16,1,0,20,173/12,1,16,1,551/30,2,0,24,40/3,1,175/12,4,15,2,46/3,1,0,
            32,391/30,3,0,35,173/15,1,139/12,1,0];
        I:=[251054,250770,251335,250290,251564,250228,251140,251446,250212,
            250905,251521,250371,250284,250555,250117,251517,250447,249896,
            250095,250561,250345,249926,250232,250235,249829,250226,250427,
            250451,250288,251513,250369,250224,250236,250237,250337,250373];
    elif hash1 eq 2505 and hash2 eq 2098 and hash3 eq 632 then
        V:=[10,1,11,9,12,17,13,25,14,31,15,41,16,45,17,55,18,63];
        D:=[1,40,1,46,1,48,1,0,4,65/2,2,77/2,1,97/2,1,0,8,98/3,1,33,1,41,1,0,11,
            211/6,1,247/6,1,0,13,28,1,89/3,1,106/3,1,36,1,0,17,217/6,1,0,18,
            61/3,1,62/3,1,67/3,1,85/3,1,0,22,41/2,1,671/30,1,341/12,1,0,25,
            343/15,1,0];
        I:=[520050,520058,520057,520046,520053,520052,520035,520021,520059,
            520039,520042,520008,520004,520012,520040,520006,519965,519998,
            519999,520024,519980,520007,519956,519949,519971];
    elif hash1 eq 1659 and hash2 eq 1181 and hash3 eq 2838 then
        V:=[21,1,22,7,23,19,24,29,25,39,26,43,27,51];
        D:=[1,103/6,1,115/6,1,0,3,53/3,1,22,1,70/3,1,134/5,1,146/5,1,0,8,33/2,1,
            107/6,2,119/6,1,253/10,1,0,13,81/5,1,268/15,1,56/3,1,62/3,1,0,17,
            227/14,1,0,18,218/15,2,46/3,1,69/4,1,0,22,451/30,1,91/6,1,0];
        I:=[510282,510475,509790,509331,509334,509298,508939,509312,509310,
            510469,510133,509242,509829,510134,508837,508959,508843,508449,
            508451,508966,508480,508891,508792];
    elif hash1 eq 2411 and hash2 eq 2485 and hash3 eq 1168 then
        V:=[22,1,23,5,24,13,25,23,26,31,27,39];
        D:=[1,25,1,0,2,127/6,1,139/6,2,47/2,1,0,6,58/3,1,59/3,1,20,1,65/3,1,0,
            10,107/6,1,109/6,3,37/2,2,0,16,65/4,1,49/3,1,55/3,1,0,19,89/6,2,0];
        I:=[673120,673129,673094,673197,673167,673132,673133,673149,672992,
            673138,673079,673146,673163,673084,673195,673088,673089,673024,
            672994,673003];
    elif hash1 eq 3351 and hash2 eq 1656 and hash3 eq 3270 then
        V:=[24,1,25,7,26,17,27,25];
        D:=[1,21,1,23,2,0,4,35/2,1,115/6,1,39/2,3,43/2,1,0,10,53/3,8,18,1,59/3,
            3,0,22,95/6,1,97/6,5,109/6,1,0];
        I:=[670992,670947,671076,670913,670932,670565,670792,670985,670554,
            670410,670497,670579,670873,670892,670930,670942,670975,670993,
            670271,670774,670787,670871,670399,670501,670571,670791,670796,
            669924];
    elif hash1 eq 1706 and hash2 eq 684 and hash3 eq 1191 then
        V:=[19,1,20,13,21,21,22,33,23,41,24,53,25,61,26,67,27,73,28,77];
        D:=[1,61/2,1,65/2,2,69/2,1,77/2,1,89/2,1,0,7,92/3,1,35,1,122/3,1,0,10,
            151/6,2,51/2,1,187/6,1,67/2,1,211/6,1,0,16,76/3,1,77/3,1,83/3,1,0,
            19,131/6,1,45/2,1,281/12,1,143/6,1,169/6,1,0,24,89/4,1,47/2,1,
            287/12,1,0,27,359/20,2,611/30,1,0,30,52/3,1,278/15,1,0,32,511/30,1,
            0,33,309/20,1,233/15,1,0];
        I:=[419929,419944,420341,420265,420139,420240,419923,420295,420162,
            420140,420163,419960,420261,419716,420235,419397,419411,420337,
            420141,420440,419619,420428,419562,419781,419137,419139,419332,
            419557,419626,419125,419629,419149,419017,419038];
    elif hash1 eq 531 and hash2 eq 1402 and hash3 eq 2826 then
        V:=[23,1,24,5,25,17,26,29,27,35,28,51,29,59,30,63,31,69];
        D:=[1,163/6,1,0,2,76/3,1,77/3,1,83/3,1,88/3,1,89/3,1,0,7,145/6,1,155/6,
            1,157/6,1,277/10,1,169/6,1,0,12,62/3,1,263/12,1,0,14,551/30,1,
            187/10,2,113/6,1,41/2,1,207/10,1,125/6,2,671/30,1,0,23,50/3,1,
            253/15,1,17,2,0,27,299/20,1,0,28,203/15,1,41/3,2,0,31,697/60,1,0];
        I:=[248130,248537,248387,248382,248508,248169,248560,247852,248173,
            248572,248571,248378,248522,248524,247536,248391,248243,248148,
            248575,248199,248587,247636,248131,248527,248155,248551,247538,
            248176,248196,248246,247510];
    elif hash1 eq 1095 and hash2 eq 1786 and hash3 eq 2035 then
        V:=[15,1,16,5,17,15,18,27,19,43,20,57,21,67,22,79,23,89,24,97,25,111];
        D:=[1,73/2,1,0,2,31,1,33,1,122/3,1,41,1,0,6,175/6,1,187/6,1,199/6,1,
            211/6,1,247/6,1,0,11,76/3,1,77/3,1,83/3,2,166/5,1,100/3,3,101/3,2,
            113/3,1,0,22,305/12,1,103/4,1,155/6,1,157/6,1,277/10,1,167/6,1,0,28,
            101/5,1,111/5,1,287/12,1,412/15,1,0,32,399/20,1,241/12,1,45/2,1,
            511/20,1,519/20,1,0,37,283/15,1,41/2,1,308/15,1,313/15,2,0,42,
            451/30,1,31/2,1,511/30,1,0,45,76/5,1,61/4,1,46/3,1,358/21,1,263/15,
            1,708/35,1,0,51,569/42,1,407/30,1,157/10,1,63/4,1,0];
        I:=[428211,428169,428234,428216,428172,428239,428193,427874,428232,
            427904,428187,428161,428170,428194,427805,428155,428160,428191,
            427865,428242,427850,428156,428229,428176,428235,427878,427875,
            427800,427621,428241,427795,427883,427794,427816,427721,427733,
            427624,427722,427836,427808,427884,427811,428233,427806,427610,
            428230,428182,427654,427620,427612,427640,427645,427658,427648];
    elif hash1 eq 672 and hash2 eq 1937 and hash3 eq 2999 then
        V:=[20,1,21,9,22,23,23,27,24,35,25,39,26,43];
        D:=[1,27,1,29,1,31,1,0,4,139/6,2,151/6,1,51/2,1,163/6,1,55/2,1,175/6,1,
            0,11,76/3,1,0,12,119/6,1,121/6,2,133/6,2,0,17,89/4,1,0,18,89/6,1,0,
            19,175/12,1,44/3,1,223/15,1,0];
        I:=[421974,421950,422089,421931,421961,421915,422012,422070,421927,
            422132,421620,422010,421921,421928,421576,422007,421261,421939,
            421946,421935,421595];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 15 mod 47.
function cached_ids_15(P,hash1,hash2,hash3)
    if hash1 eq 579 and hash2 eq 1750 and hash3 eq 2786 then
        V:=[25,1,26,5,27,17,28,33,29,37];
        D:=[1,109/6,1,0,2,49/3,1,91/5,1,73/4,1,56/3,1,20,1,0,7,89/6,1,91/6,1,
            491/30,1,103/6,1,183/10,1,221/12,1,37/2,2,0,15,179/12,1,0,16,401/30,
            1,27/2,3,0];
        I:=[414156,414126,414117,414162,414149,413976,414211,414202,414119,
            414161,414422,414419,413716,413845,414135,414148,414076,414153,
            414187];
    elif hash1 eq 1002 and hash2 eq 2398 and hash3 eq 723 then
        V:=[22,1,23,11,24,15,25,21,26,31,27,37];
        D:=[1,94/5,1,26,1,133/4,1,101/3,1,0,5,167/6,1,0,6,99/5,1,352/15,1,0,8,
            279/14,1,41/2,2,207/10,2,671/30,1,0,14,361/20,1,109/6,1,0,16,325/18,
            1,0];
        I:=[540265,540291,540078,540293,540147,540204,539800,539969,540021,
            540148,540208,540213,540257,540146,539817,539708];
    elif hash1 eq 1284 and hash2 eq 1862 and hash3 eq 1689 then
        V:=[23,1,24,5,25,13,26,25,27,29,28,35,29,45];
        D:=[1,199/6,1,0,2,70/3,1,77/3,1,117/4,1,0,5,143/6,1,257/10,1,155/6,1,
            277/10,1,169/6,1,0,10,89/4,1,0,11,207/10,1,127/6,1,0,13,199/12,1,17,
            1,361/20,1,223/12,2,0,18,461/30,2,31/2,1,499/30,1,0];
        I:=[416314,416947,416556,416638,417185,416468,416461,417693,417588,
            416640,416555,416531,416660,416953,415367,416501,417485,417555,
            417589,417584,415532];
    elif hash1 eq 2083 and hash2 eq 3357 and hash3 eq 3187 then
        V:=[22,1,23,9,24,21,25,29,26,35,27,43,28,49,29,61,30,67];
        D:=[1,29,1,92/3,1,33,2,0,5,47/2,1,151/6,1,51/2,1,107/4,1,163/6,2,0,11,
            65/3,1,116/5,2,24,1,0,15,119/6,2,121/6,3,0,20,191/12,1,49/3,1,18,1,
            0,23,101/6,1,551/30,1,0,25,281/20,1,223/15,1,15,1,329/20,1,33/2,2,0,
            31,259/20,1,391/30,1,0,33,173/15,1,0];
        I:=[251578,251281,250712,251319,251004,251286,251037,250319,251001,
            251305,250802,250222,250272,250698,251308,251314,250600,250898,
            251360,250092,250096,251041,250384,250167,250239,250357,250927,
            249921,249860,249918,250052,250234,250085];
    elif hash1 eq 485 and hash2 eq 4 and hash3 eq 623 then
        V:=[25,1,26,5,27,23,29,33,30,45];
        D:=[1,217/10,1,0,2,57/4,1,43/3,2,44/3,1,15,1,18,1,56/3,1,322/15,1,109/5,
            1,0,11,29/2,1,167/10,1,103/6,1,75/4,1,0,15,391/30,2,157/12,3,83/6,1,
            299/20,1,181/12,1,0,23,787/60,1,0];
        I:=[415580,417679,417084,417368,415730,415662,415667,415583,416116,
            416000,417032,415769,415581,415518,417024,417111,417022,417071,
            417085,415519,415482,415031,415123];
    elif hash1 eq 1566 and hash2 eq 726 and hash3 eq 3199 then
        V:=[22,1,23,7,24,17,25,31,26,43,27,55];
        D:=[1,74/3,1,25,1,0,3,39/2,1,91/4,1,139/6,2,47/2,1,0,8,18,1,77/4,1,58/3,
            1,59/3,1,20,1,64/3,1,0,14,33/2,1,107/6,1,109/6,1,193/10,1,119/6,1,0,
            19,191/12,1,16,1,49/3,3,50/3,1,91/5,1,0,26,89/6,1,0];
        I:=[249523,249818,249819,248811,248771,249340,249383,248924,248757,
            248753,248706,249338,248920,249294,248913,249251,247868,249249,
            248739,249579,248837,248923,249512,249246,247626,248025];
    elif hash1 eq 485 and hash2 eq 331 and hash3 eq 368 then
        V:=[24,1,25,5,26,11,28,21,29,29];
        D:=[1,83/3,1,0,2,237/10,1,95/4,1,0,4,49/3,1,101/5,1,62/3,1,21,1,0,8,
            179/12,1,15,4,277/15,1,0,14,83/6,1,157/10,1,0];
        I:=[414165,414176,413869,414876,414054,413726,413884,414686,413866,
            414836,414853,414866,413467,413882,413856];
    elif hash1 eq 1660 and hash2 eq 1889 and hash3 eq 3052 then
        V:=[20,1,21,7,22,15,23,23,24,35,25,45,26,53,27,61,28,65];
        D:=[1,27,1,86/3,1,0,3,151/6,1,115/4,1,175/6,1,0,6,71/3,1,83/3,1,88/3,1,
            0,9,217/10,1,95/4,1,103/4,1,155/6,1,157/6,1,0,14,101/5,1,263/12,1,
            97/4,1,74/3,1,0,18,283/14,1,41/2,1,125/6,1,0,21,86/5,1,369/20,1,
            259/12,1,0,24,511/30,1,0,25,208/15,2,14,1,76/5,1,0];
        I:=[422907,422813,422824,422801,422814,422821,422831,422830,422822,
            422806,422621,422871,423121,422823,422964,422648,423669,422446,
            422623,422657,422529,422360,422439,422661,422449,422952,422506,
            422359];
    elif hash1 eq 1519 and hash2 eq 271 and hash3 eq 3216 then
        V:=[27,1,28,7,29,15,30,25,31,35,32,41,33,47];
        D:=[1,143/6,1,135/4,1,0,3,79/3,1,80/3,1,86/3,1,0,6,113/6,1,791/30,1,
            317/12,1,57/2,1,0,10,56/3,1,19,1,106/5,1,74/3,1,0,14,205/12,1,
            209/12,2,0,17,69/4,1,268/15,1,0,19,469/30,3,0];
        I:=[492359,492416,492412,492378,492371,492369,492379,492417,492415,
            492368,492413,492376,492396,492426,492428,492432,492429,492365,
            492408,492409,492410];
    elif hash1 eq 1472 and hash2 eq 2309 and hash3 eq 2014 then
        V:=[23,1,24,9,25,15,27,23,28,29];
        D:=[1,49/2,2,53/2,2,57/2,1,0,6,21,1,23,1,0,8,103/6,1,35/2,1,115/6,1,0,
            11,95/6,3,97/6,1,0,15,167/12,1,14,1,0];
        I:=[408355,408625,408609,408657,408650,408693,408688,408376,408479,
            407127,407162,408739,408742,408583,408565,408706];
    elif hash1 eq 1566 and hash2 eq 769 and hash3 eq 395 then
        V:=[27,1,28,5,29,9,30,19,31,33,33,39];
        D:=[1,107/6,1,0,2,50/3,1,0,3,173/12,1,59/4,1,491/30,1,33/2,1,0,7,212/15,
            1,218/15,1,44/3,1,179/12,1,16,1,82/5,1,0,13,389/30,1,157/12,1,0,15,
            239/20,1,0];
        I:=[407177,407613,407585,407411,407376,407572,407593,407541,405567,
            407415,407769,407790,407752,407640,404008];
    elif hash1 eq 2835 and hash2 eq 207 and hash3 eq 1535 then
        V:=[18,1,19,5,20,23,21,39,22,47,23,65,24,89,25,113,26,137,27,145];
        D:=[1,27,1,0,2,43/2,1,99/4,1,163/6,1,55/2,1,175/6,1,187/6,1,63/2,1,
            199/6,1,0,10,64/3,1,65/3,2,77/3,1,82/3,1,83/3,1,146/5,1,30,1,0,18,
            119/6,3,133/6,1,257/10,1,0,23,73/4,1,55/3,2,56/3,2,99/5,1,239/12,2,
            101/5,1,61/3,1,328/15,1,0,34,101/6,4,539/30,1,18,1,217/12,2,551/30,
            1,221/12,1,599/30,1,207/10,1,125/6,1,307/14,1,45/2,1,0,49,46/3,2,17,
            3,52/3,1,361/20,1,272/15,1,109/6,1,92/5,1,278/15,1,19,3,77/4,2,
            332/15,1,0,66,83/6,2,451/30,2,185/12,1,67/4,1,169/10,1,339/20,1,17,
            1,511/30,2,217/12,1,109/6,1,56/3,1,0,80,203/15,2,163/12,1,187/12,1,
            0,84,239/20,1,0];
        I:=[517569,517539,517431,517545,517571,517548,517568,517425,517538,
            517377,517533,517577,517372,517555,517578,517285,517414,517543,
            517566,517572,517387,517323,517332,517396,517553,517385,517542,
            517299,517474,517525,517350,517597,517284,517275,517379,517394,
            517402,517219,517448,517309,517462,517131,517127,517129,517289,
            517556,517048,516978,517126,517228,517080,517558,517561,517408,
            517224,517256,517259,517331,517341,517096,517102,517110,517109,
            517114,516989,517230,517328,517225,517226,517300,517074,517349,
            517120,517076,517241,517249,517078,517097,516991,517229,517340,
            517251,517004,517132];
    elif hash1 eq 1519 and hash2 eq 551 and hash3 eq 949 then
        V:=[19,1,20,9,21,19,22,31,23,39,24,47,25,67,26,75,27,81,28,85];
        D:=[1,47/2,1,187/6,1,87/2,1,0,4,156/5,1,95/3,2,100/3,1,186/5,1,0,9,
            281/12,1,143/6,1,145/6,1,337/10,1,205/6,1,0,14,62/3,1,222/7,1,97/3,
            1,0,17,101/6,1,217/12,1,187/10,1,0,20,16,1,321/20,1,92/5,1,278/15,1,
            308/15,1,62/3,1,145/7,1,251/12,1,24,1,0,29,461/30,1,521/30,1,781/42,
            1,0,32,50/3,1,69/4,1,0,34,587/36,1,0,35,260/21,1,188/15,1,543/40,2,
            827/60,1,983/60,1,0];
        I:=[542407,542483,542419,542362,542454,542469,542470,542351,542418,
            542071,542449,542316,542440,542388,541976,541900,542245,542272,
            542249,542072,542067,542263,542233,542241,541791,541894,541905,
            541851,542239,542273,541916,541983,542166,541784,541932,541933,
            541802,541844,541853,542511];
    elif hash1 eq 62 and hash2 eq 1971 and hash3 eq 2690 then
        V:=[23,1,24,13,25,19,26,31,27,47,28,57,29,69];
        D:=[1,127/6,2,43/2,1,151/6,1,107/4,1,187/6,1,0,7,59/3,2,126/5,1,0,10,
            107/6,1,197/10,1,79/4,2,119/6,2,121/6,1,0,17,268/15,1,215/12,1,91/5,
            1,56/3,1,19,1,239/12,1,62/3,1,0,24,89/6,2,491/30,1,101/6,1,539/30,1,
            0,29,41/3,1,44/3,1,179/12,2,248/15,1,50/3,1,0,35,27/2,1,625/42,1,
            451/30,1,181/12,1,0];
        I:=[418707,418708,418715,417882,418699,417668,417879,418422,417675,
            418706,417586,418663,418742,418599,418665,417714,417727,417895,
            416687,418754,418673,418703,418550,418709,418710,416677,416288,
            416298,417740,417904,417712,417767,416281,416290,416481,415555,
            415554,415552];
    elif hash1 eq 3305 and hash2 eq 2688 and hash3 eq 2315 then
        V:=[19,1,20,7,21,19,22,27,23,37,24,45,25,57,26,65,27,75,28,81];
        D:=[1,175/6,1,131/4,1,0,3,30,2,32,1,34,1,112/3,1,113/3,1,0,9,131/6,2,
            157/6,1,169/6,1,0,13,298/15,2,131/5,1,151/5,1,502/15,1,0,18,611/30,
            1,207/10,1,107/4,1,0,21,253/15,1,18,1,37/2,1,56/3,1,283/15,1,0,26,
            161/12,1,59/4,1,185/12,1,0,29,79/6,3,91/6,1,233/15,1,247/12,1,0,35,
            1103/84,1,797/60,1,0,37,176/15,1,0];
        I:=[542230,542400,542428,542432,542431,542466,542365,542437,542244,
            542267,542438,542467,542268,542277,542307,542311,542146,542343,
            542289,541841,542278,542429,542444,542475,542287,542397,542264,
            541862,542396,542404,542406,541860,542334,541855,541904,542275,
            542228];
    elif hash1 eq 2036 and hash2 eq 2172 and hash3 eq 1482 then
        V:=[19,1,20,9,21,19,22,31,23,41,24,45,26,49,27,53];
        D:=[1,45/2,1,49/2,1,53/2,2,0,5,62/3,2,92/3,1,33,3,104/3,1,0,12,139/6,1,
            107/4,1,163/6,2,187/6,1,211/6,1,0,18,106/5,1,65/3,1,93/4,1,82/3,1,0,
            22,95/6,1,0,23,268/15,1,0,24,199/12,1,0,25,31/2,1,0];
        I:=[420455,420456,420458,420538,420508,420552,420525,420532,420555,
            420557,419398,419188,420221,420100,420536,420065,419248,419187,
            419902,420073,419443,420539,419320,419410,419000];
    elif hash1 eq 908 and hash2 eq 1046 and hash3 eq 221 then
        V:=[26,1,27,5,28,15,29,21,30,29,31,35,32,39,33,49];
        D:=[1,34,1,0,2,237/10,1,145/6,1,111/4,1,205/6,1,0,6,61/3,1,74/3,1,0,8,
            113/6,1,41/2,1,247/10,1,0,11,86/5,1,52/3,1,0,13,59/4,1,0,14,173/15,
            1,139/12,1,66/5,1,76/5,1,0,18,677/60,3,0];
        I:=[493789,493743,493788,493900,493683,494075,493792,494064,493902,
            493982,494361,493791,494070,493793,493806,493751,493849,493663,
            493665,493800];
    elif hash1 eq 2694 and hash2 eq 1539 and hash3 eq 1975 then
        V:=[19,1,20,7,21,25,22,41,23,49,24,61,25,67,26,79,27,87,28,91];
        D:=[1,223/6,1,75/2,1,0,3,83/3,1,88/3,1,89/3,1,125/4,1,133/4,1,100/3,1,
            101/3,1,186/5,1,0,11,305/12,1,157/6,2,277/10,1,111/4,1,313/10,1,
            191/6,1,203/6,1,0,19,287/12,1,121/5,1,388/15,2,0,23,611/30,1,245/12,
            1,431/20,1,339/14,1,731/30,1,0,28,428/21,1,409/20,1,0,30,511/30,1,
            517/28,1,1117/60,1,569/30,1,193/10,1,0,35,907/60,1,61/4,1,352/21,1,
            0,38,817/60,1,0,39,176/15,1,260/21,1,62/5,1,206/15,1,827/60,1,0];
        I:=[515047,515099,515364,514886,514834,515295,514580,515341,515115,
            514592,515163,514626,514629,514711,515292,514318,514870,514579,
            515247,514721,514791,514794,514634,515204,514316,514132,514116,
            514157,514163,514641,514164,514177,514306,514307,513976,514908,
            514145,514895,514883,514050,514072,513978,513982];
    elif hash1 eq 861 and hash2 eq 2501 and hash3 eq 1249 then
        V:=[21,1,22,7,23,13,24,17,25,25,26,31,27,35,30,41];
        D:=[1,45/2,1,57/2,2,0,4,74/3,2,110/3,1,0,7,127/6,1,0,8,93/4,1,126/5,1,
            117/4,1,0,11,209/12,1,197/10,1,0,13,238/15,1,0,14,371/30,1,169/12,1,
            0,16,166/15,1,0];
        I:=[538709,538755,538782,538738,538752,538648,538647,538774,538718,
            538729,538777,538772,538775,538698,538793,538814];
    elif hash1 eq 532 and hash2 eq 237 and hash3 eq 1703 then
        V:=[28,1,29,7,30,13,31,25,32,31,33,37];
        D:=[1,59/3,3,20,1,0,5,71/4,1,107/6,1,0,7,79/5,1,191/12,1,16,1,49/3,2,
            50/3,1,0,13,173/12,1,29/2,1,0,15,13,1,40/3,3,0,19,341/30,1,23/2,1,
            0];
        I:=[665661,665693,665862,665860,666081,665126,665302,665141,666227,
            665160,666053,665849,666234,665120,665544,665856,666025,666180,
            664963,665028];
    elif hash1 eq 3446 and hash2 eq 175 and hash3 eq 2667 then
        V:=[13,1,15,7,16,13,17,23,18,31,19,37,20,47,23,53,24,59];
        D:=[1,77/2,1,97/2,1,0,3,235/6,1,247/6,1,0,5,89/3,1,100/3,1,101/3,1,
            206/5,1,0,9,297/10,1,179/6,1,401/12,1,0,12,311/12,1,418/15,1,0,14,
            611/30,1,671/30,1,551/20,1,679/20,1,0,18,409/20,1,247/12,1,0,20,
            917/60,1,1451/70,1,0,22,1426/105,1,947/60,1,0];
        I:=[519589,519602,519587,519524,519594,519583,519592,519519,519531,
            519476,519591,519600,519536,519534,519374,519390,519414,519509,
            519392,519328,519329,519378,519376];
    elif hash1 eq 2835 and hash2 eq 1513 and hash3 eq 1348 then
        V:=[23,1,24,5,25,11,26,17,28,25,29,37];
        D:=[1,187/6,1,0,2,93/4,1,28,1,0,4,131/6,1,49/2,1,0,6,43/3,2,49/3,1,
            215/12,1,0,10,218/15,1,175/12,1,44/3,1,179/12,2,15,4,0,19,391/30,1,
            95/6,2,559/30,1,0];
        I:=[418219,417493,415847,415907,415849,418772,418805,418259,417578,
            415893,415510,418867,416820,418189,415885,416163,418096,418280,
            417063,415854,415950,414918];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 16 mod 47.
function cached_ids_16(P,hash1,hash2,hash3)
    if hash1 eq 2554 and hash2 eq 1971 and hash3 eq 2815 then
        V:=[22,1,23,5,24,9,25,21,26,37,28,49,29,55,30,59];
        D:=[1,92/3,1,0,2,115/4,1,0,3,77/3,1,117/4,1,89/3,3,95/3,1,101/3,1,0,10,
            131/6,1,143/6,2,257/10,1,155/6,3,157/6,2,329/12,2,167/6,1,0,22,
            298/15,1,61/3,1,22,2,111/5,1,24,1,0,28,223/12,5,227/12,2,0,35,
            181/12,2,0,37,907/60,1,76/5,1,0];
        I:=[250832,250304,251285,249974,250227,250262,251274,250738,250791,
            250733,250161,250766,250142,250699,250814,251313,250812,251331,
            250300,250332,250044,250202,251252,250889,251291,250141,250122,
            249923,249950,250011,250328,250393,249954,250341,250317,250377,
            249855,249856];
    elif hash1 eq 909 and hash2 eq 1595 and hash3 eq 3118 then
        V:=[31,1,32,7,33,15,34,19];
        D:=[1,95/6,3,97/6,3,0,7,43/3,1,44/3,4,191/12,2,0,14,77/6,4,0,18,35/3,1,
            0];
        I:=[656911,656994,658202,656848,656909,657984,657059,656023,656240,
            657756,658170,656220,656696,656359,656495,657754,657878,655168];
    elif hash1 eq 157 and hash2 eq 661 and hash3 eq 379 then
        V:=[23,1,24,7,25,21,26,35,28,49,29,55,30,61,31,67];
        D:=[1,163/6,1,123/4,1,0,3,77/3,2,109/4,1,83/3,3,28,1,88/3,1,89/3,2,0,13,
            197/10,1,217/10,1,95/4,1,143/6,1,145/6,1,103/4,1,0,19,268/15,1,55/3,
            2,20,1,101/5,1,24,1,129/5,1,0,26,46/3,1,199/12,1,0,28,451/30,2,
            181/12,1,0,31,66/5,2,163/12,1,0,34,321/28,1,0];
        I:=[410663,410650,409334,410700,411392,410019,410680,411015,410226,
            410738,410211,410626,410998,409023,411407,410229,410924,410377,
            411194,411065,411192,410664,409568,409098,409443,410171,410641,
            410539,411054,411320,410463,410956,410165,409237];
    elif hash1 eq 1238 and hash2 eq 2570 and hash3 eq 229 then
        V:=[24,1,25,11,26,15,27,23,28,31,29,51];
        D:=[1,53/3,1,18,1,20,1,65/3,1,0,5,119/6,1,0,6,49/3,1,18,1,55/3,1,0,9,
            91/6,2,491/30,1,167/10,1,0,13,40/3,1,29/2,1,175/12,1,44/3,3,76/5,1,
            46/3,1,115/7,1,247/15,1,199/12,2,0,25,157/12,1,0];
        I:=[418912,418901,418490,418608,418677,418879,418649,418916,417927,
            418571,417738,417775,416320,418504,417777,418522,418808,418887,
            417756,418674,415485,415563,415487,417443,416623];
    elif hash1 eq 3118 and hash2 eq 3309 and hash3 eq 1511 then
        V:=[27,1,28,9,29,13,30,21,31,25,32,35];
        D:=[1,109/6,1,133/6,1,157/6,1,0,4,55/3,1,0,5,173/12,1,29/2,1,89/6,1,0,8,
            223/15,1,0,9,259/20,1,391/30,3,157/12,1,27/2,1,0,15,173/15,1,0];
        I:=[408214,408403,407864,408216,408042,408413,404659,407029,404647,
            406701,406715,407225,406988,408209,404660];
    elif hash1 eq 63 and hash2 eq 3295 and hash3 eq 543 then
        V:=[6,1,7,9,8,15,9,23,10,29,12,33,13,45,14,49,15,59,16,63,18,69,20,81,
            21,91];
        D:=[1,54,2,56,1,62,1,0,5,101/2,1,109/2,1,0,7,49,1,170/3,1,57,1,0,10,
            247/6,1,295/6,1,0,12,131/3,1,0,13,36,1,109/3,1,658/15,1,221/5,1,
            133/3,1,0,18,1091/30,1,0,19,463/15,1,729/20,1,548/15,1,439/12,1,0,
            23,1717/60,1,0,24,1387/60,1,353/15,1,0,26,18,1,2404/105,1,813/35,1,
            1961/84,1,1403/60,1,0,31,2263/140,1,1706/105,1,1369/84,1,979/60,1,0,
            35,383/30,1,0];
        I:=[544390,544392,544391,544387,544389,544388,544386,544379,544381,
            544384,544383,544378,544380,544382,544375,544371,544377,544372,
            544376,544367,544368,544374,544370,544366,544373,544385,544364,
            544365,544361,544363,544358,544360,544362,544359,544369];
    elif hash1 eq 1567 and hash2 eq 1233 and hash3 eq 590 then
        V:=[20,1,21,5,22,17,23,29,24,43,25,51,26,59,27,77,29,93];
        D:=[1,98/3,1,0,2,151/6,1,51/2,1,175/6,2,59/2,1,67/2,1,0,8,101/4,1,77/3,
            4,83/3,1,28,1,89/3,1,0,16,131/6,2,237/10,1,143/6,2,145/6,3,257/10,1,
            155/6,1,0,26,61/3,2,263/12,1,111/5,2,0,31,221/12,2,37/2,1,75/4,1,0,
            35,33/2,1,199/12,1,253/15,1,277/15,1,278/15,1,223/12,3,283/15,2,
            227/12,1,0,46,291/20,1,439/30,1,44/3,1,59/4,1,181/12,1,91/6,1,31/2,
            1,0,53,47/4,1,0];
        I:=[253371,252972,253182,253133,253152,253172,253177,252755,252612,
            252961,253198,253419,252918,253272,252706,253070,253191,252704,
            253069,253171,252695,252707,253424,252616,252891,252710,252963,
            252812,252512,252629,252721,252825,253092,252722,252523,252508,
            252711,252493,252535,252494,252527,252747,252511,252687,252533,
            252835,252620,252838,252899,252709,252608,253364,252714];
    elif hash1 eq 298 and hash2 eq 2789 and hash3 eq 3285 then
        V:=[18,1,19,7,20,19,21,31,22,39,23,51,24,57];
        D:=[1,24,1,26,1,0,3,49/2,1,53/2,2,57/2,3,65/2,1,69/2,1,0,11,62/3,1,21,1,
            23,1,27,1,29,1,0,16,115/6,1,43/2,2,47/2,1,0,20,52/3,2,53/3,1,59/3,2,
            65/3,1,22,2,0,28,131/6,1,133/6,1,0,30,16,2,49/3,1,50/3,1,0];
        I:=[424846,424844,423865,424695,424847,424686,424693,424813,424588,
            424157,424763,424592,424685,423605,424465,424800,423875,424727,
            424363,424603,424817,424669,423869,424670,424444,423653,423876,
            423436,423114,424700,424778,424480,424504];
    elif hash1 eq 392 and hash2 eq 2735 and hash3 eq 1773 then
        V:=[19,1,20,5,21,13,22,19,23,27,24,37,26,43,27,47];
        D:=[1,73/2,1,0,2,68/3,1,25,1,29,1,0,5,39/2,1,43/2,2,0,8,53/3,1,58/3,1,
            28,1,0,11,63/4,1,233/12,1,217/10,1,145/6,2,0,16,16,1,268/15,1,0,18,
            44/3,3,0,21,259/20,1,181/12,1,0];
        I:=[512807,512846,512285,512000,512711,512277,512787,512664,512819,
            512352,512693,512355,512410,511476,511681,511566,511394,511590,
            511746,511759,511837,511210];
    elif hash1 eq 768 and hash2 eq 458 and hash3 eq 742 then
        V:=[21,1,22,5,23,15,24,33,25,43,26,51,27,55,28,61,29,69];
        D:=[1,83/2,1,0,2,89/3,1,101/3,1,106/3,2,194/5,1,0,7,133/6,1,143/6,1,
            277/10,1,167/6,1,169/6,1,317/10,1,127/4,1,203/6,1,0,15,101/5,1,81/4,
            1,152/7,1,263/12,1,0,19,551/30,1,671/30,1,49/2,1,0,22,278/15,1,0,23,
            181/12,1,31/2,1,0,25,14,1,86/5,1,1147/60,2,0,29,55/4,1,0];
        I:=[512318,511252,512001,511486,512075,511808,512074,512414,511536,
            511730,512042,511455,512419,511939,512313,511926,510957,511987,
            511914,511360,511212,511537,511927,511408,511915,511206,510979,
            511149,512099];
    elif hash1 eq 1144 and hash2 eq 927 and hash3 eq 1115 then
        V:=[19,1,21,5,22,9,23,21,24,27,25,35,27,41];
        D:=[1,199/6,1,0,2,157/6,1,0,3,24,1,73/3,1,131/5,1,79/3,2,80/3,1,0,9,
            269/12,1,267/10,1,0,11,283/15,1,19,1,58/3,2,0,15,379/20,1,265/12,1,
            0,17,421/30,2,85/6,1,0];
        I:=[425954,425953,425947,425962,425951,425910,425956,425915,425894,
            425629,425949,425955,425927,425971,425631,425645,425627,425961,
            425968];
    elif hash1 eq 1332 and hash2 eq 795 and hash3 eq 2332 then
        V:=[16,1,17,9,18,17,20,23];
        D:=[1,28,2,30,2,32,1,0,6,53/2,1,57/2,3,61/2,1,0,11,74/3,1,25,3,0,15,
            64/3,1,65/3,1,22,2,0];
        I:=[674679,674680,674673,674678,674677,674672,674665,674670,674674,
            674667,674669,674666,674668,674675,674671,674663,674664,674676];
    elif hash1 eq 1285 and hash2 eq 1333 and hash3 eq 2431 then
        V:=[18,1,19,13,20,25,21,35];
        D:=[1,24,2,26,1,28,2,32,3,34,1,0,10,41/2,2,45/2,1,49/2,1,53/2,4,57/2,2,
            0,20,21,2,68/3,1,23,3,25,1,0,27,115/6,1,39/2,1,127/6,1,0];
        I:=[253994,253997,253844,253840,253996,253828,253949,253992,253915,
            253882,254000,253855,253839,253771,253827,253833,253936,253746,
            253950,253911,253993,253868,253835,253838,253907,253626,253741,
            253627,253618];
    elif hash1 eq 2930 and hash2 eq 710 and hash3 eq 1171 then
        V:=[28,1,29,5,30,9,31,17,32,25];
        D:=[1,59/3,2,0,3,121/6,1,0,4,49/3,2,91/5,2,55/3,1,0,9,89/6,4,163/10,1,
            33/2,1,0,15,155/12,1,40/3,1,0];
        I:=[227944,228014,227736,227872,228613,227771,227788,227811,227281,
            227282,227468,228596,224420,228632,226545,227890];
    elif hash1 eq 674 and hash2 eq 1050 and hash3 eq 1031 then
        V:=[20,1,21,9,22,23,23,37,24,49,25,61,26,77,28,87];
        D:=[1,27,1,29,1,98/3,1,0,4,139/6,1,59/2,3,187/6,2,63/2,1,67/2,1,223/6,1,
            0,13,65/3,1,71/3,2,76/3,1,26,1,89/3,1,100/3,1,0,20,131/6,1,133/6,1,
            45/2,1,145/6,1,167/6,1,0,25,91/5,1,56/3,2,263/12,1,67/3,1,121/5,1,0,
            31,167/10,1,551/30,2,37/2,1,113/6,1,241/12,1,611/30,1,41/2,2,0,40,
            223/15,3,52/3,1,223/12,1,19,1,0,46,203/15,2,0];
        I:=[509601,509532,509603,509530,509557,510018,510038,509536,509992,
            509663,509978,509479,510037,509602,510074,509487,509988,509974,
            510177,509596,509453,509447,509989,509782,509461,509993,510044,
            509544,509454,509613,510034,509554,509614,509611,509945,509619,
            509604,509549,509612,509564,509565,509609,510081,509551,509586,
            509969,510098];
    elif hash1 eq 1097 and hash2 eq 1381 and hash3 eq 1491 then
        V:=[10,1,11,21,12,37,13,63,14,87,15,115,16,143,17,175,18,209,19,235,20,
            259,21,279,22,289,23,311];
        D:=[1,36,1,38,2,40,1,42,1,44,1,46,2,48,1,50,1,52,1,0,12,65/2,1,69/2,1,
            77/2,2,81/2,1,85/2,2,89/2,1,97/2,1,0,21,86/3,1,29,1,92/3,1,98/3,1,
            33,1,35,1,37,1,116/3,1,39,3,122/3,2,43,1,134/3,1,0,36,115/4,1,175/6,
            1,187/6,1,199/6,2,67/2,1,71/2,1,223/6,2,235/6,1,79/2,1,247/6,2,87/2,
            2,0,51,77/3,1,109/4,1,83/3,1,30,1,94/3,1,95/3,1,100/3,1,34,1,106/3,
            3,107/3,5,118/3,2,216/5,1,131/3,2,0,72,103/4,1,155/6,1,157/6,2,
            167/6,3,179/6,2,61/2,1,317/10,1,191/6,1,425/12,1,357/10,1,143/4,1,
            215/6,1,217/6,2,0,90,22,1,67/3,2,68/3,2,287/12,1,73/3,1,418/15,2,
            335/12,1,28,1,141/5,1,85/3,2,92/3,1,538/15,1,181/5,1,145/4,1,109/3,
            1,0,109,41/2,1,207/10,1,125/6,1,671/30,2,45/2,1,47/2,1,24,2,731/30,
            1,293/12,2,313/12,1,791/30,1,53/2,1,559/20,2,28,1,851/30,1,341/12,2,
            0,130,58/3,1,308/15,3,247/12,2,449/20,1,45/2,1,271/12,1,343/15,2,
            275/12,1,70/3,1,71/3,1,529/20,1,53/2,1,0,146,521/30,1,35/2,1,
            1117/60,1,631/30,1,1333/60,1,451/20,1,679/30,1,68/3,1,91/4,1,277/12,
            1,139/6,1,0,157,86/5,2,52/3,1,778/35,1,635/28,1,478/21,1,114/5,1,
            485/21,1,347/15,1,93/4,1,0,167,937/60,1,319/20,1,481/30,1,193/12,1,
            0,171,1636/105,1,1313/84,1,313/20,1,557/35,1,447/28,1,337/21,2,
            241/15,1,967/60,1,81/5,1,65/4,1,0,182,377/30,3,0];
        I:=[519918,519912,519917,519928,519916,519921,519911,519920,519932,
            519913,519925,519924,519903,519919,519926,519886,519922,519930,
            519931,519882,519873,519883,519848,519858,519906,519915,519929,
            519910,519914,519923,519927,519871,519891,519867,519889,519810,
            519859,519862,519887,519888,519803,519826,519861,519899,519866,
            519816,519799,519869,519837,519849,519870,519800,519801,519845,
            519860,519821,519838,519841,519877,519878,519907,519814,519829,
            519895,519896,519905,519820,519851,519777,519778,519786,519809,
            519855,519830,519843,519804,519850,519893,519835,519844,519856,
            519764,519788,519825,519750,519813,519828,519812,519817,519880,
            519864,519909,519720,519836,519808,519892,519751,519779,519827,
            519709,519748,519846,519847,519780,519736,519733,519734,519774,
            519876,519719,519722,519721,519724,519725,519824,519802,519805,
            519783,519811,519832,519714,519784,519716,519717,519744,519710,
            519791,519708,519746,519852,519723,519793,519794,519775,519822,
            519729,519823,519840,519753,519756,519731,519768,519831,519718,
            519713,519749,519789,519728,519755,519792,519797,519761,519765,
            519711,519715,519712,519754,519795,519884,519685,519701,519707,
            519684,519689,519688,519683,519732,519743,519738,519772,519690,
            519681,519697,519705,519693,519680,519698,519696,519694,519699,
            519679,519740,519741,519742];
    elif hash1 eq 2178 and hash2 eq 2550 and hash3 eq 1378 then
        V:=[24,1,25,13,26,17,27,25,28,37,29,45,31,53];
        D:=[1,23,1,74/3,1,86/3,1,29,1,31,2,0,7,151/6,1,0,8,53/3,1,20,2,64/3,1,0,
            12,185/12,1,95/6,1,97/6,1,71/4,1,121/6,1,0,17,167/12,1,73/4,1,56/3,
            1,0,20,29/2,2,193/12,1,197/12,1,0,24,341/30,1,0];
        I:=[245648,245530,244103,242899,244186,245252,244598,244470,242897,
            245253,243275,244032,242862,244055,244081,244225,242857,241887,
            244041,245590,245618,242133,242123,243462];
    elif hash1 eq 1943 and hash2 eq 1081 and hash3 eq 3378 then
        V:=[16,1,17,11,18,19,19,29,20,37,21,47,22,51,23,55,25,61];
        D:=[1,34,1,38,1,40,1,42,1,0,5,65/2,1,73/2,2,77/2,1,0,9,29,1,33,2,35,1,
            39,1,0,14,175/6,1,187/6,3,211/6,2,0,20,77/4,1,101/4,1,76/3,1,82/3,3,
            0,26,131/6,1,0,27,61/3,1,0,28,221/12,1,137/6,1,0,30,185/12,1,0];
        I:=[425367,425362,425274,425280,425368,425277,425298,425328,425259,
            425109,425167,425172,425171,425183,425161,425260,425304,425036,
            425164,425166,425144,425294,425038,425076,425078,424976,425162,
            425099,424898,425011];
    elif hash1 eq 2037 and hash2 eq 2712 and hash3 eq 1714 then
        V:=[23,1,24,23,25,39,26,53,27,69,28,87,29,101,30,111];
        D:=[1,47/2,1,151/6,2,51/2,2,163/6,1,55/2,1,115/4,1,175/6,1,59/2,2,187/6,
            2,67/2,1,0,15,20,1,65/3,1,24,2,126/5,2,101/4,1,77/3,1,136/5,1,0,24,
            71/4,1,109/6,1,197/10,1,119/6,1,121/6,2,145/6,1,0,31,49/3,3,50/3,2,
            215/12,1,55/3,2,101/5,2,61/3,1,62/3,1,0,43,59/4,1,89/6,3,491/30,2,
            101/6,1,183/10,1,221/12,1,75/4,1,113/6,1,0,54,218/15,1,175/12,1,
            44/3,1,223/15,1,76/5,1,199/12,1,0,60,259/20,2,13,2,157/12,1,91/6,1,
            0,66,229/20,1,0];
        I:=[418354,418365,418474,418221,418462,418440,417657,417612,416348,
            418183,418509,418090,418496,418507,418414,418361,417658,418237,
            416349,417158,417598,417949,417116,417502,416561,416818,418349,
            417680,417864,417132,416507,416808,418336,416321,418307,417337,
            416187,416508,416211,416970,416692,418034,416169,416473,416596,
            417319,416392,416680,418218,415449,417117,416689,416514,416694,
            416495,418347,417682,416679,415469,415810,416426,415835,415995,
            416622,415521,416797];
    elif hash1 eq 2883 and hash2 eq 1473 and hash3 eq 2501 then
        V:=[20,1,21,11,22,25,23,39,24,57,25,65,26,71];
        D:=[1,21,1,23,1,80/3,3,27,1,0,7,115/6,1,39/2,3,127/6,8,43/2,1,139/6,1,
            47/2,1,0,22,53/3,1,58/3,1,59/3,1,20,2,85/4,1,65/3,2,0,30,71/4,2,
            107/6,3,109/6,7,37/2,1,233/12,1,79/4,1,121/6,3,131/6,1,0,49,16,1,
            49/3,2,50/3,2,0,54,59/4,1,31/2,1,0,56,46/3,1,47/3,1,0];
        I:=[422169,421826,421955,421958,422254,421823,422188,422167,422322,
            422323,421806,421810,421830,422163,422175,422237,422243,422253,
            422327,421780,421790,422320,421813,422238,421801,422331,422067,
            421990,421999,421799,422159,422160,422182,422286,421300,421804,
            421824,421832,422079,422168,422247,422021,421996,421288,421779,
            422006,422018,421227,422179,421807,422173,421800,422037,421313,
            421310,421123,421122];
    elif hash1 eq 2789 and hash2 eq 1446 and hash3 eq 581 then
        V:=[25,1,26,13,27,29,28,49,29,63,30,73,31,79];
        D:=[1,39/2,1,127/6,1,139/6,2,47/2,2,151/6,3,0,10,53/3,1,18,1,58/3,1,
            59/3,3,20,3,85/4,1,64/3,1,0,21,95/6,1,97/6,1,71/4,1,107/6,2,109/6,3,
            37/2,1,197/10,1,119/6,1,121/6,1,0,33,81/5,1,49/3,4,50/3,1,17,1,89/5,
            1,55/3,1,0,42,431/30,1,173/12,1,29/2,4,89/6,1,0,49,193/15,1,155/12,
            1,0,51,23/2,1,0];
        I:=[240684,241192,240904,241178,240443,241079,239535,240580,240605,
            240746,240977,239878,240020,240581,240595,239333,240095,241030,
            239818,239239,241159,240632,240340,239338,240703,240698,240780,
            240886,241295,239666,239388,238676,240278,239863,239884,240222,
            240987,239371,239991,238101,239324,240091,238434,239387,239881,
            240666,241170,239382,240261,238724,239803];
    elif hash1 eq 63 and hash2 eq 2573 and hash3 eq 1646 then
        V:=[21,1,22,7,23,15,24,21,25,29,26,33,28,41,29,45];
        D:=[1,57/2,1,61/2,1,0,3,25,1,98/3,1,33,1,0,6,163/6,1,175/6,1,0,8,24,1,
            26,1,28,1,0,11,133/6,1,0,12,62/3,1,43/2,1,263/12,1,0,15,17,1,0,16,
            185/12,1,31/2,1,0];
        I:=[250636,250601,250460,250470,250214,250282,250017,250062,250031,
            250053,250037,250051,249943,249924,249916,249863,249857];
    elif hash1 eq 2601 and hash2 eq 2252 and hash3 eq 2307 then
        V:=[22,1,23,5,24,17,25,23,26,29,27,39,28,53,29,69];
        D:=[1,27,1,0,2,151/6,1,51/2,1,163/6,1,55/2,2,175/6,2,0,9,22,1,71/3,3,0,
            13,119/6,1,131/6,1,0,15,49/3,1,215/12,1,73/4,2,55/3,1,0,20,197/12,1,
            33/2,1,101/6,2,103/6,1,551/30,3,221/12,1,0,29,217/15,2,29/2,2,
            175/12,3,223/15,5,15,2,46/3,3,47/3,1,0,47,391/30,1,0];
        I:=[249571,249770,249698,249587,248768,249124,249133,249610,249655,
            247731,248106,248991,248997,247916,249022,247925,247761,247921,
            248908,247697,248886,249130,249564,248512,247698,247811,248457,
            247537,247953,248477,248561,248867,248107,248472,248899,247814,
            247828,247855,247907,247918,248028,249772,248467,248883,249129,
            248942,248038];
    elif hash1 eq 1238 and hash2 eq 790 and hash3 eq 1172 then
        V:=[20,1,21,13,22,23,23,41,24,53,25,67,26,81,27,85];
        D:=[1,29,1,31,2,35,1,110/3,1,37,1,0,7,163/6,1,187/6,2,63/2,1,211/6,2,0,
            13,70/3,1,76/3,1,26,1,82/3,1,83/3,2,89/3,1,30,1,95/3,1,0,22,133/6,1,
            95/4,1,155/6,1,53/2,1,277/10,1,0,27,111/5,1,89/4,1,67/3,1,68/3,2,
            121/5,1,73/3,1,0,34,551/30,1,221/12,2,113/6,1,245/12,1,41/2,2,125/6,
            2,0,43,52/3,1,0,44,451/30,1,181/12,1,91/6,3,511/30,1,103/6,4,0];
        I:=[423993,423945,424467,423909,424064,424385,424485,424302,424344,
            424384,424401,424497,423976,424454,424442,424374,424049,424423,
            424402,424432,423622,424301,423701,424361,423544,423618,423674,
            423502,424447,424406,424496,423621,423593,423271,423306,423583,
            424409,423736,423358,423688,423639,424282,422527,423670,423484,
            424418,424419,424426,422590,422531,422573,422576,422584];
    elif hash1 eq 2648 and hash2 eq 1672 and hash3 eq 468 then
        V:=[23,1,24,5,25,13,26,23,27,35,28,41,29,45,30,51];
        D:=[1,115/4,1,0,2,59/3,1,125/4,1,133/4,1,0,5,109/6,1,119/6,1,305/12,1,
            167/6,1,0,9,215/12,1,73/4,1,263/12,1,68/3,1,358/15,1,0,14,491/30,1,
            207/10,1,0,16,321/20,1,0,17,439/30,1,193/12,1,0,19,41/3,1,0];
        I:=[410214,410952,411302,411494,411121,411127,410618,410378,410227,
            410241,409306,411104,408932,410950,411181,410168,410170,409130,
            411283];
    elif hash1 eq 2836 and hash2 eq 434 and hash3 eq 2879 then
        V:=[26,1,27,7,28,15,29,25,30,29,31,35,33,43];
        D:=[1,83/3,1,101/3,1,0,3,133/6,1,143/6,1,157/6,1,0,6,328/15,1,121/5,1,
            73/3,1,74/3,1,0,10,115/6,1,0,11,283/15,1,19,1,0,13,511/30,1,205/12,
            1,107/6,1,0,16,55/4,1,0];
        I:=[404576,405197,404688,404651,404555,404697,405068,405241,405208,
            405094,404693,404594,405401,404532,405098,404447];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 17 mod 47.
function cached_ids_17(P,hash1,hash2,hash3)
    if hash1 eq 1756 and hash2 eq 2786 and hash3 eq 1673 then
        V:=[22,1,23,5,25,9,26,19,27,27,28,33,30,37];
        D:=[1,110/3,1,0,2,199/6,1,0,3,143/6,3,103/4,1,155/6,1,169/6,1,0,9,
            298/15,1,22,1,24,1,0,12,18,1,41/2,1,0,14,227/12,1,0,15,203/15,1,
            91/6,1,0];
        I:=[250251,250057,250178,250200,250206,249975,250040,250058,250201,
            250169,250192,249945,250083,249959,250150,249854];
    elif hash1 eq 2649 and hash2 eq 1850 and hash3 eq 252 then
        V:=[25,1,26,5,27,13,28,21,29,35];
        D:=[1,121/6,1,0,2,239/12,1,322/15,1,74/3,1,0,5,33/2,1,41/2,1,83/4,1,0,8,
            248/15,1,380/21,1,77/4,1,58/3,1,20,1,21,1,0,14,451/30,1,461/30,1,
            265/14,1,0];
        I:=[501709,501456,501663,501634,501438,502019,501475,501809,501430,
            501640,501437,501810,501466,501184,501713,501258];
    elif hash1 eq 675 and hash2 eq 1843 and hash3 eq 104 then
        V:=[19,1,20,7,21,15,22,23,23,31,24,39];
        D:=[1,53/2,2,57/2,1,0,4,23,2,74/3,2,25,1,0,9,127/6,3,43/2,1,151/6,1,0,
            14,59/3,3,20,1,22,1,0,19,107/6,1,109/6,3,37/2,1,0,24,50/3,4,0];
        I:=[674374,674377,674351,674350,674378,674343,674344,674324,674320,
            674346,674355,674325,674293,674317,674321,674338,674352,674291,
            674299,674298,674304,674305,674301,674282,674314,674315,674359];
    elif hash1 eq 2602 and hash2 eq 1412 and hash3 eq 1216 then
        V:=[20,1,21,9,22,15,23,25,24,33,25,43,26,55];
        D:=[1,21,1,74/3,1,80/3,1,0,4,139/6,1,99/4,1,0,6,53/3,1,85/4,1,64/3,1,
            101/4,1,0,10,119/6,1,213/10,1,131/6,1,0,13,49/3,2,91/5,1,55/3,1,
            263/12,1,0,18,91/6,1,491/30,1,33/2,1,67/4,1,103/6,1,0,23,223/15,1,
            46/3,2,0];
        I:=[422312,422266,422190,422218,421859,422262,422130,422288,421263,
            421407,421349,421283,422152,422265,421284,421248,421184,421523,
            421773,421533,421518,421285,421200,421190,421195];
    elif hash1 eq 158 and hash2 eq 2097 and hash3 eq 1104 then
        V:=[26,1,27,9,28,15,29,21];
        D:=[1,21,2,23,3,74/3,2,0,8,35/2,1,115/6,3,0,12,47/3,1,53/3,2,0,15,97/6,
            1,0];
        I:=[234781,234786,234784,234823,234889,234495,234770,234884,234521,
            234522,234629,232429,234345,234794,232751];
    elif hash1 eq 2367 and hash2 eq 2892 and hash3 eq 2719 then
        V:=[23,1,24,5,25,11,26,17,27,21,28,29,29,33,30,39,31,49];
        D:=[1,317/10,1,0,2,358/15,1,121/5,1,0,4,791/30,1,57/2,1,0,6,62/3,1,0,7,
            181/12,1,1117/60,1,581/30,1,0,10,61/4,1,0,11,421/30,1,359/20,1,0,13,
            206/15,1,211/15,1,847/60,1,236/15,1,0,17,2897/210,1,0];
        I:=[539888,539912,539886,539904,539674,539686,539948,539877,539883,
            539687,539882,539899,539873,539878,539875,539897,539868];
    elif hash1 eq 111 and hash2 eq 2989 and hash3 eq 1341 then
        V:=[20,1,21,11,22,17,23,25,26,37,27,41,28,47];
        D:=[1,25,1,80/3,1,29,1,39,1,0,5,151/6,2,55/2,2,0,9,71/3,1,83/3,1,107/3,
            1,0,12,217/10,1,131/6,1,155/6,1,167/6,1,169/6,1,0,17,46/3,1,0,18,
            461/30,1,511/30,1,0,20,226/15,1,907/60,1,77/5,1,187/12,1,0];
        I:=[420434,420435,420278,420437,420081,420097,420300,420354,419421,
            420307,419598,419433,420027,419584,419535,419767,420310,419580,
            419162,419039,419047,419016,419056];
    elif hash1 eq 2884 and hash2 eq 1547 and hash3 eq 2782 then
        V:=[16,1,18,9,19,13,20,23,21,29,22,35,23,39];
        D:=[1,22,1,24,2,26,2,0,6,27,1,0,7,39/2,1,43/2,1,139/6,1,71/2,1,0,11,
            53/3,1,77/4,1,0,13,119/6,1,257/12,2,0,16,81/5,1,0,17,103/6,1,391/20,
            1,0];
        I:=[542497,542498,542502,542501,542503,542399,542057,542393,542490,
            542414,542387,542398,542048,542409,542441,542234,542066,542302];
    elif hash1 eq 581 and hash2 eq 2607 and hash3 eq 3145 then
        V:=[21,1,22,7,23,15,26,25,27,29];
        D:=[1,163/6,1,199/6,2,0,4,76/3,1,83/3,3,95/3,1,0,9,131/6,1,133/6,1,
            237/10,1,145/6,1,0,13,253/15,1,0,14,59/4,1,181/12,1,31/2,1,0];
        I:=[424317,424331,424539,424309,422893,423749,424289,424274,424273,
            424488,423060,424304,423690,423414,422978,423047];
    elif hash1 eq 1521 and hash2 eq 1966 and hash3 eq 1697 then
        V:=[30,1,31,5,32,13,34,19];
        D:=[1,65/3,5,0,6,107/6,1,109/6,2,119/6,1,0,10,49/3,4,50/3,1,0,15,40/3,6,
            0];
        I:=[57714,57752,58434,58438,58490,58416,57774,58462,57699,57776,58441,
            58470,58527,58545,57680,57702,58431,58440,58452,58562];
    elif hash1 eq 3213 and hash2 eq 56 and hash3 eq 3010 then
        V:=[25,1,26,5,27,17,28,23,29,27,30,31];
        D:=[1,151/6,1,0,2,59/3,1,85/4,1,64/3,1,65/3,5,71/3,1,0,11,79/4,1,119/6,
            1,0,13,49/3,1,0,14,167/10,1,0,15,15,1,46/3,1,0];
        I:=[670267,670081,669975,669996,669970,669971,669974,670266,670338,
            670008,670144,670015,670155,669901,669889,669890];
    elif hash1 eq 2226 and hash2 eq 2649 and hash3 eq 803 then
        V:=[23,1,25,5,26,15,27,21,28,31,29,39];
        D:=[1,51/2,1,0,2,257/12,1,87/4,1,143/6,1,193/6,1,0,6,61/3,1,67/3,1,0,8,
            599/30,2,207/10,1,127/6,1,22,1,0,13,56/3,1,227/12,1,313/15,1,0,16,
            781/42,1,75/4,1,201/10,1,0];
        I:=[502033,501727,501734,502007,501703,501707,501433,501742,501755,
            501667,501669,501658,501869,501705,501863,501757,501751,501802];
    elif hash1 eq 3072 and hash2 eq 1301 and hash3 eq 594 then
        V:=[21,1,22,7,23,15,24,31,25,37,26,43,27,47,28,51,29,55];
        D:=[1,139/6,1,259/6,1,0,3,83/3,1,95/3,1,194/5,1,0,6,167/6,1,179/6,1,
            181/6,1,317/10,1,127/4,1,193/6,1,203/6,1,0,13,22,1,194/7,1,0,15,
            359/20,1,391/20,1,0,17,223/12,1,0,18,461/28,1,0,19,203/15,1,0,20,
            55/4,1,0];
        I:=[512538,512552,512677,512509,511806,511810,511483,511251,511867,
            512478,512771,511801,511811,511250,511866,511805,511542,511267,
            511861,512482];
    elif hash1 eq 440 and hash2 eq 2616 and hash3 eq 265 then
        V:=[26,1,27,5,29,13,30,17,31,21,32,29,33,35];
        D:=[1,79/3,1,0,2,269/12,1,149/6,1,519/20,1,0,5,571/30,1,0,6,1123/60,1,0,
            7,35/3,1,421/30,2,85/6,1,0,11,707/60,3,71/6,2,0,16,283/24,1,
            2477/210,1,167/14,1,0];
        I:=[502056,503336,502063,503055,503332,501952,503039,502155,503022,
            502061,502236,503011,503261,502140,502243,501278,501276,501289];
    elif hash1 eq 2978 and hash2 eq 3011 and hash3 eq 1959 then
        V:=[23,1,24,5,25,15,26,23,27,35];
        D:=[1,53/2,3,0,4,21,1,68/3,1,23,1,25,1,0,8,103/6,2,115/6,2,39/2,1,0,13,
            69/4,1,53/3,1,18,1,59/3,1,20,1,0,18,95/6,2,107/6,1,0];
        I:=[245282,245542,245659,245296,242898,244389,244513,244978,245308,
            243004,245563,245456,244299,243535,243118,244132,244386,243567,
            245102,244131];
    elif hash1 eq 1192 and hash2 eq 3452 and hash3 eq 505 then
        V:=[19,1,20,9,21,15,22,29,23,35,24,45,25,51,26,57];
        D:=[1,53/2,1,61/2,1,69/2,1,0,4,23,1,27,1,0,6,39/2,1,139/6,1,47/2,1,
            151/6,1,51/2,3,163/6,1,0,14,65/3,3,22,1,0,18,71/4,1,107/6,2,233/12,
            1,119/6,1,0,23,39/2,1,20,1,0,25,101/6,1,599/30,1,0,27,218/15,1,
            175/12,1,223/15,1,179/12,1,15,1,46/3,1,0];
        I:=[253438,253426,253257,253429,253430,253212,253233,253246,252943,
            252922,253210,253211,253129,253167,253241,253394,253163,252859,
            252861,253248,253125,253095,252692,252607,253113,252513,252690,
            252586,252688,252672,252585,252820];
    elif hash1 eq 675 and hash2 eq 1963 and hash3 eq 2767 then
        V:=[25,1,26,5,27,13,28,27,30,31];
        D:=[1,139/6,1,0,2,65/3,1,93/4,1,71/3,1,0,5,197/10,2,79/4,1,217/10,1,
            87/4,1,133/6,1,143/6,1,0,12,124/7,1,0,13,302/21,1,321/20,2,0];
        I:=[235903,235299,236119,236032,235293,235398,236124,235524,236045,
            236098,235329,235308,235306,235357,235518];
    elif hash1 eq 1521 and hash2 eq 516 and hash3 eq 1209 then
        V:=[21,1,22,5,23,27,24,39,25,57,26,73,27,91,28,103,29,111];
        D:=[1,175/6,1,0,2,20,1,93/4,1,77/3,1,146/5,1,89/3,1,30,1,94/3,2,166/5,1,
            101/3,1,34,1,0,13,121/6,1,257/12,1,145/6,1,103/4,1,111/4,1,0,18,
            50/3,1,73/4,1,56/3,1,61/3,2,62/3,1,89/4,1,287/12,1,73/3,1,0,27,
            193/12,1,33/2,1,551/30,3,37/2,1,283/14,1,245/12,1,125/6,1,0,36,
            289/20,1,46/3,1,203/12,1,17,2,69/4,1,94/5,1,283/15,1,227/12,1,0,45,
            27/2,1,451/30,1,461/30,1,697/42,1,67/4,1,0,50,269/20,1,27/2,2,
            203/15,1,0,54,171/14,1,371/30,1,325/24,2,407/30,1,0];
        I:=[510612,510662,510601,510481,509569,510477,510604,510485,510543,
            509905,509644,510245,510274,510660,510360,510632,509719,509802,
            509346,509543,509222,510273,508396,509315,509440,508886,509655,
            510617,509582,509648,509805,509281,508813,508743,510438,509717,
            509676,510501,509180,509366,509244,509247,509628,509196,509187,
            509635,509716,508746,509254,509216,509183,509197,510122,508871,
            508870,508363,508372,508438];
    elif hash1 eq 1850 and hash2 eq 1536 and hash3 eq 244 then
        V:=[26,1,27,7,28,11,29,19,30,29,31,37];
        D:=[1,70/3,1,76/3,1,0,3,107/6,1,0,4,215/12,1,101/5,1,62/3,1,0,7,147/10,
            1,59/4,1,89/6,1,245/12,1,0,11,44/3,1,76/5,1,329/20,1,0,14,259/20,2,
            389/30,1,157/12,2,439/30,2,0];
        I:=[405665,406483,408364,406784,407664,408223,407167,406781,407168,
            406531,405678,404421,406518,406742,406794,404526,406365,406367,
            404506,404522];
    elif hash1 eq 299 and hash2 eq 1362 and hash3 eq 840 then
        V:=[20,1,21,11,22,23,23,33,24,41,25,47,26,55,27,59,28,67,29,77];
        D:=[1,80/3,1,31,1,33,1,104/3,1,0,5,127/6,1,43/2,1,63/2,2,235/6,1,163/4,
            1,0,11,77/3,2,32,1,100/3,1,101/3,1,0,16,109/6,2,87/4,1,257/10,1,0,
            20,55/3,1,263/12,1,0,22,167/10,1,551/30,1,113/6,1,0,25,72/5,1,0,26,
            391/30,1,137/10,1,115/6,1,0,29,452/35,1,363/28,1,203/15,2,14,1,0,34,
            485/42,1,0];
        I:=[512059,511899,511868,512214,511984,512217,512016,512133,511448,
            512399,512006,512415,512144,512007,512351,512298,512299,512002,
            511563,511967,512003,512058,511335,511929,512424,511137,512281,
            511172,511319,511102,511020,512283,512335,511295];
    elif hash1 eq 2790 and hash2 eq 1998 and hash3 eq 3030 then
        V:=[19,1,20,7,21,21,22,27,23,37,24,45,26,51];
        D:=[1,61/2,1,65/2,3,0,5,27,1,29,1,92/3,1,31,1,33,1,35,1,0,11,43/2,1,
            163/6,4,0,16,77/4,1,70/3,1,71/3,1,77/3,1,0,20,131/6,1,133/6,3,155/6,
            1,0,25,67/3,1,68/3,1,0,27,175/12,2,179/12,1,15,1,0];
        I:=[423856,423862,423917,423948,424035,424120,424017,424336,423864,
            423935,423833,422937,423846,423898,423920,424330,424151,423960,
            422928,423019,422926,424164,424327,422683,422610,422679,422946,
            423840,422943,422944];
    elif hash1 eq 393 and hash2 eq 431 and hash3 eq 3302 then
        V:=[25,1,26,5,27,13,28,25,29,43,30,51,31,59];
        D:=[1,127/6,2,0,3,77/4,1,59/3,5,85/4,1,0,10,95/6,1,209/12,1,177/10,1,
            107/6,1,197/10,1,0,15,44/3,1,191/12,1,81/5,1,65/4,1,49/3,1,50/3,1,
            268/15,1,55/3,1,0,23,169/12,1,29/2,1,147/10,1,0,26,155/12,1,13,1,
            40/3,2,0,30,23/2,1,0];
        I:=[240683,240822,240611,239786,240827,240890,240905,241194,240103,
            241400,241230,239606,240813,239746,240476,241271,240350,240914,
            239903,240893,239922,238072,239635,239725,240179,239678,239671,
            238877,239403,240508];
    elif hash1 eq 2931 and hash2 eq 131 and hash3 eq 2876 then
        V:=[20,1,21,5,22,13,23,19,24,29,25,35,26,41,27,49];
        D:=[1,92/3,2,0,3,39/2,1,51/2,1,55/2,1,0,6,59/3,1,22,1,0,8,97/6,2,109/6,
            1,119/6,2,121/6,1,0,14,18,1,73/4,1,0,16,431/30,1,89/6,1,0,18,15,1,
            46/3,1,52/3,1,0,21,181/12,1,31/2,1,0];
        I:=[509570,510542,510256,510338,509915,510254,510340,510329,510398,
            510066,509542,510541,508608,510209,509581,509587,510618,510382,
            510060,508636,508774,508773];
    elif hash1 eq 769 and hash2 eq 201 and hash3 eq 2629 then
        V:=[22,1,23,19,24,29,25,39,26,51,27,57,28,67,29,73];
        D:=[1,25,2,80/3,1,86/3,1,92/3,1,31,1,35,1,110/3,1,37,1,0,10,43/2,1,
            151/6,1,51/2,1,199/6,1,0,14,59/3,1,65/3,1,93/4,1,83/3,1,0,18,197/10,
            1,119/6,2,131/6,1,95/4,1,143/6,1,0,24,73/4,1,298/15,1,0,26,89/6,3,
            491/30,1,101/6,2,18,1,0,33,248/15,2,50/3,1,0,36,299/20,1,461/30,1,
            0];
        I:=[411726,412080,411722,411599,412081,411739,410421,411971,411609,
            412058,411559,411646,411676,412133,411638,411378,411405,411145,
            410853,411613,411669,410104,409483,410588,411458,411590,411637,
            411670,411512,409513,410146,409522,409521,410117,409518,409409,
            409369];
    elif hash1 eq 2837 and hash2 eq 477 and hash3 eq 3356 then
        V:=[27,1,28,11,29,15,30,21,31,29];
        D:=[1,269/12,1,91/4,1,731/30,1,49/2,2,0,6,308/15,1,0,7,511/30,1,205/12,
            1,0,9,61/4,1,187/12,2,17,1,0,13,133/10,1,161/12,1,937/60,1,0];
        I:=[413977,413861,414094,413908,413978,414179,414150,413804,414233,
            413853,413979,413576,414110,414213,413450];
    elif hash1 eq 534 and hash2 eq 2680 and hash3 eq 1635 then
        V:=[19,1,20,7,21,15,22,19,23,31,24,35];
        D:=[1,49/2,1,57/2,2,0,4,21,2,74/3,1,25,1,0,8,127/6,1,0,9,52/3,1,53/3,1,
            59/3,1,22,1,77/3,1,0,14,109/6,2,0,16,49/3,2,0];
        I:=[422133,421982,422119,421967,422136,421985,422131,422139,422064,
            422022,421761,421983,421267,421977,421998,421550,422151];
    elif hash1 eq 1380 and hash2 eq 2024 and hash3 eq 1814 then
        V:=[22,1,23,7,24,19,25,27,26,33,27,41,28,45];
        D:=[1,17,1,19,1,0,3,103/6,1,115/6,1,127/6,1,43/2,1,139/6,1,0,8,69/4,1,
            52/3,1,77/4,2,0,12,97/6,1,209/12,1,0,14,238/15,2,18,1,91/5,1,0,18,
            351/20,1,0,19,179/12,1,0];
        I:=[412335,412351,412331,412169,410871,411799,411907,410526,412313,
            410805,411959,410512,411932,410225,410946,409159,409964,409979,
            409239];
    elif hash1 eq 1192 and hash2 eq 2617 and hash3 eq 1505 then
        V:=[16,1,17,5,18,9,19,15,20,25,21,31,22,41,23,53,25,59];
        D:=[1,35,1,0,2,59/2,1,0,3,95/3,2,101/3,1,0,6,103/4,1,155/6,3,157/6,1,
            53/2,1,0,12,68/3,1,311/12,1,0,14,22,1,265/12,1,269/12,1,45/2,1,0,18,
            19,2,77/4,1,58/3,4,59/3,1,332/15,1,0,27,581/30,4,451/20,1,0,32,
            937/60,1,0];
        I:=[254973,254967,254996,254998,254971,254950,254969,254987,254988,
            254979,254986,254977,254943,254955,254952,254935,254932,254924,
            254980,254948,254970,254984,254991,254992,254994,254942,254923,
            254925,254929,254931,254915,254918];
    elif hash1 eq 1380 and hash2 eq 2855 and hash3 eq 137 then
        V:=[22,1,23,5,24,9,25,25,26,41,27,45,28,51,29,57,30,65];
        D:=[1,23,1,0,2,211/6,1,0,3,71/3,1,88/3,1,30,1,154/5,1,94/3,2,95/3,1,
            100/3,2,0,12,121/6,1,133/6,1,237/10,1,95/4,1,145/6,1,277/10,1,167/6,
            1,0,19,239/12,1,0,20,491/30,1,37/2,1,0,22,86/5,1,88/5,1,0,24,59/4,1,
            103/6,1,229/12,1,0,27,66/5,1,41/3,1,0];
        I:=[411222,411274,411221,409970,411136,409536,411156,411425,411390,
            409508,411137,411201,409333,409567,411167,411437,409684,411140,
            411248,411202,410018,410098,409418,411030,409446,409379,411197,
            411518];
    elif hash1 eq 1192 and hash2 eq 1477 and hash3 eq 1466 then
        V:=[24,1,25,5,26,23,27,35,28,39,29,47,31,59];
        D:=[1,86/3,1,0,2,139/6,1,47/2,1,175/6,1,59/2,2,187/6,2,63/2,1,199/6,1,
            67/2,1,0,12,52/3,1,53/3,1,71/3,2,24,1,83/3,1,0,18,133/6,1,0,19,55/3,
            1,101/5,1,61/3,1,0,22,311/20,1,197/12,1,33/2,1,101/6,1,183/10,1,0,
            27,391/30,2,27/2,2,0];
        I:=[235886,236077,235951,236046,235994,236014,236039,236041,236073,
            236043,236186,236089,236088,235944,236040,236044,236078,235523,
            236022,235525,236079,235522,235968,235970,235327,235352,236010,
            236042,236011,236012];
    elif hash1 eq 1756 and hash2 eq 2635 and hash3 eq 460 then
        V:=[23,1,24,5,26,17,27,25,28,29,29,37,30,45];
        D:=[1,163/6,2,0,3,59/3,1,93/4,1,126/5,1,88/3,1,95/3,1,0,8,89/5,1,91/5,1,
            68/3,2,0,12,207/10,1,0,13,179/12,1,248/15,1,409/20,1,0,16,613/42,1,
            149/10,1,75/4,1,0,19,178/15,1,261/20,1,0];
        I:=[505892,505946,504296,504426,506344,506171,506477,505051,506111,
            505959,506198,506222,504169,504417,505717,504197,505301,504910,
            506365,505614];
    elif hash1 eq 1286 and hash2 eq 885 and hash3 eq 2948 then
        V:=[12,1,13,5,14,9,15,13,16,19,17,25,18,31,19,35,20,39,21,43,22,49,23,
            53];
        D:=[1,46,1,0,2,81/2,1,0,3,122/3,1,0,4,199/6,1,259/6,1,0,6,133/4,1,101/3,
            1,0,8,135/4,1,215/6,1,0,10,407/12,1,0,11,45/2,1,0,12,247/12,1,0,13,
            619/30,1,691/30,1,0,15,1243/60,1,0,16,63/4,1,0];
        I:=[430072,430070,430031,430030,429996,430003,430026,429988,429993,
            429970,429995,429997,429969,429963,429964,429971];
    elif hash1 eq 3354 and hash2 eq 1481 and hash3 eq 2179 then
        V:=[26,1,27,7,28,13,29,17,30,23,31,31];
        D:=[1,65/3,2,71/3,1,0,4,119/6,1,121/6,1,0,6,91/5,1,0,7,197/12,2,101/6,1,
            0,10,179/12,2,15,2,46/3,2,0,16,181/12,1,0];
        I:=[670460,670877,670966,670959,670805,670162,670317,670318,670900,
            669928,670252,669952,670344,670439,670467,669882];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 18 mod 47.
function cached_ids_18(P,hash1,hash2,hash3)
    if hash1 eq 3402 and hash2 eq 294 and hash3 eq 743 then
        V:=[23,1,24,5,25,9,26,15,27,21,28,31,30,37];
        D:=[1,55/2,1,0,2,71/3,2,0,4,133/6,1,49/2,1,0,6,55/3,1,62/3,1,0,8,33/2,2,
            67/4,1,101/6,2,187/10,1,0,14,175/12,1,15,2,0,17,35/3,1,0];
        I:=[246967,246966,247056,246561,246938,246599,246849,247093,247180,
            246730,246755,247092,246539,246718,246979,247169,246739];
    elif hash1 eq 18 and hash2 eq 3227 and hash3 eq 1572 then
        V:=[28,1,29,11,30,19,31,23,32,27];
        D:=[1,70/3,1,71/3,2,77/3,1,83/3,1,0,6,217/10,1,133/6,1,143/6,1,0,9,91/5,
            1,0,10,101/6,1,0,11,218/15,1,223/15,1,15,3,46/3,1,0];
        I:=[666625,666298,666449,665368,666292,665478,665372,666591,665438,
            666573,665459,665476,665371,665467,666606,666296];
    elif hash1 eq 1005 and hash2 eq 3302 and hash3 eq 1709 then
        V:=[22,1,23,5,24,11,25,21,26,31,27,37,28,45,29,49,31,57];
        D:=[1,82/3,1,0,2,167/6,1,127/4,1,0,4,328/15,1,73/3,2,311/12,1,141/5,1,0,
            9,611/30,1,125/6,1,289/12,1,293/12,1,0,13,247/12,1,313/15,2,0,16,
            67/4,1,205/12,1,217/12,1,0,19,76/5,1,0,20,569/42,1,137/10,2,157/10,
            1,0,24,1019/84,1,0];
        I:=[423146,422750,422852,422724,422789,423195,423078,423259,423073,
            422745,422368,422594,422659,423161,423370,423165,423164,422422,
            422791,422385,422416,423225,422351,422335];
    elif hash1 eq 1052 and hash2 eq 1955 and hash3 eq 1660 then
        V:=[20,1,21,5,22,13,23,27,24,39,25,45,26,51,27,63];
        D:=[1,35,1,0,2,187/6,1,67/2,1,223/6,1,0,5,26,1,83/3,1,89/3,1,30,1,94/3,
            1,95/3,1,0,11,109/6,1,197/10,1,119/6,2,95/4,1,157/6,2,0,18,328/15,1,
            263/12,1,0,20,491/30,2,439/20,1,0,23,223/12,1,283/15,1,227/12,3,
            58/3,4,441/20,1,0,33,451/30,1,559/30,1,0];
        I:=[510345,509566,510359,510564,509994,510555,509354,509987,510326,
            509468,509624,509556,509620,509621,509378,509257,510514,509099,
            509379,509484,509552,508770,508661,509026,509030,509057,509245,
            509085,509362,509499,510061,508419,509766,508428];
    elif hash1 eq 629 and hash2 eq 2210 and hash3 eq 817 then
        V:=[24,1,25,9,26,21,27,31,28,55,29,65];
        D:=[1,70/3,1,24,1,101/4,1,0,4,107/6,1,119/6,1,121/6,1,41/2,2,217/10,1,0,
            10,50/3,2,56/3,1,61/3,1,263/12,1,0,15,89/6,1,91/6,1,33/2,2,67/4,1,
            101/6,2,103/6,1,35/2,1,221/12,1,37/2,2,187/10,1,75/4,1,0,29,44/3,1,
            15,2,253/15,1,17,1,0,34,401/30,1,161/12,1,27/2,1,83/6,3,181/12,1,
            185/12,1,0];
        I:=[414265,414040,414248,414251,414261,414435,414067,414312,414035,
            414066,414254,414106,414262,413843,414092,413993,414260,414263,
            414099,414299,414324,414024,414070,414373,413817,413818,413659,
            413840,413996,414107,414247,414316,414015,414286,413529,414079,
            414311,414314,414327,413503,413504];
    elif hash1 eq 2650 and hash2 eq 1522 and hash3 eq 865 then
        V:=[18,1,19,7,20,17,24,23,25,29];
        D:=[1,171/5,1,109/3,1,0,3,61/2,1,719/20,1,1091/30,1,437/12,1,0,7,569/20,
            1,428/15,1,0,9,947/60,1,268/15,1,0,11,3293/210,2,1327/84,1,949/60,1,
            95/6,1,0];
        I:=[519461,519538,519563,519421,519419,519418,519480,519472,519423,
            519380,519334,519357,519335,519355,519333];
    elif hash1 eq 2415 and hash2 eq 254 and hash3 eq 2837 then
        V:=[23,1,24,9,25,17,26,23,27,29,28,41];
        D:=[1,47/2,1,59/2,1,187/6,1,0,4,64/3,1,65/3,1,76/3,2,0,8,257/12,1,
            217/10,1,0,10,215/12,1,55/3,1,0,12,101/6,1,103/6,1,183/10,1,551/30,
            1,113/6,1,0,17,17,3,0];
        I:=[507991,507920,507974,507969,507979,507966,507968,507939,507982,
            508009,507994,507989,507923,507871,508004,507954,507924,507938,
            508018];
    elif hash1 eq 2603 and hash2 eq 1852 and hash3 eq 2996 then
        V:=[21,1,22,5,23,15,24,21,25,33,28,41];
        D:=[1,175/6,1,0,2,76/3,1,77/3,1,26,1,89/3,1,0,6,257/10,1,157/6,2,0,9,
            61/3,1,263/12,1,111/5,1,67/3,4,311/12,1,0,17,113/6,2,125/6,1,659/30,
            1,0,21,208/15,1,14,1,187/12,1,0];
        I:=[251904,251960,251963,251902,251881,251953,251901,251934,251945,
            251932,251928,251925,251926,251965,251966,252007,251920,251935,
            251950,251854,251943,251957,251833];
    elif hash1 eq 2838 and hash2 eq 420 and hash3 eq 1962 then
        V:=[18,1,19,5,20,9,21,15,22,25,23,37,24,41,25,45,26,49,28,55,29,61];
        D:=[1,134/3,1,0,2,247/6,1,0,3,112/3,1,119/3,1,0,5,297/10,1,337/10,1,
            393/10,1,253/6,1,0,9,180/7,1,478/15,1,32,1,34,1,171/5,1,0,14,779/30,
            1,0,15,521/20,1,0,16,793/42,1,0,17,274/15,1,287/15,1,0,19,208/15,1,
            31/2,1,0,21,82/7,1,2537/210,1,0];
        I:=[516253,516058,516158,516217,515657,516072,515842,516345,515660,
            516019,516332,516207,516045,515818,515769,515666,515767,515841,
            516218,515700,515699,515683];
    elif hash1 eq 1710 and hash2 eq 1492 and hash3 eq 362 then
        V:=[24,1,25,5,26,11,27,19,28,31,29,39,30,47,31,51];
        D:=[1,68/3,1,0,2,115/6,1,99/4,1,0,4,53/3,2,18,1,85/4,1,0,8,95/6,2,97/6,
            1,71/4,1,107/6,1,109/6,1,0,14,43/3,2,49/3,1,50/3,1,0,18,169/12,1,
            431/30,1,173/12,1,0,21,151/12,1,0,22,341/30,2,137/12,1,0];
        I:=[244923,245271,243980,243291,245371,245283,242175,243076,243975,
            245340,243224,242870,244995,242942,243155,242848,242341,242117,
            242868,243942,243117,243316,243933,243330];
    elif hash1 eq 2415 and hash2 eq 125 and hash3 eq 3056 then
        V:=[22,1,23,7,24,13,26,25,27,31,28,37];
        D:=[1,94/3,1,95/3,1,0,3,167/6,2,169/6,1,0,6,73/4,1,61/3,1,22,1,89/4,1,
            287/12,1,0,11,69/4,1,77/4,1,0,13,27/2,1,31/2,1,0,15,269/20,1,203/15,
            1,0];
        I:=[510609,510508,509189,509234,510470,509376,509233,509191,508905,
            509387,509322,510626,509324,509316,509199,510119];
    elif hash1 eq 2744 and hash2 eq 2529 and hash3 eq 1170 then
        V:=[25,1,26,11,27,19,28,25,29,31,31,37,32,45];
        D:=[1,167/6,2,119/4,1,191/6,1,205/6,1,0,6,131/5,1,79/3,1,28,1,0,9,
            731/30,2,49/2,1,0,12,102/5,1,106/5,1,0,14,205/12,1,35/2,2,0,17,
            137/10,2,913/60,2,157/10,2,0,23,707/60,1,0];
        I:=[418675,418756,417818,418736,416739,417537,416027,416048,417341,
            417837,416082,417787,416078,417918,414924,418697,416070,416213,
            416258,416269,416062,416283,415368];
    elif hash1 eq 2180 and hash2 eq 3006 and hash3 eq 2179 then
        V:=[21,1,23,5,24,9,26,25,27,31,28,35,29,45,30,53];
        D:=[1,187/6,1,0,2,181/6,1,0,3,311/12,1,105/4,1,418/15,1,335/12,1,141/5,
            1,85/3,1,86/3,1,0,10,223/12,1,21,1,0,12,559/30,1,0,13,309/20,1,
            238/15,2,16,1,223/12,1,0,18,817/60,1,55/4,1,46/3,1,0,21,61/5,1,0];
        I:=[421743,421737,421415,421341,421363,421445,421670,421715,421632,
            421607,421714,421256,421207,421668,421698,421625,421112,421762,
            421707,421211,421155];
    elif hash1 eq 488 and hash2 eq 2312 and hash3 eq 2397 then
        V:=[24,1,25,7,26,21,27,25,28,35];
        D:=[1,23,1,29,1,0,3,115/6,1,39/2,1,127/6,1,139/6,1,47/2,1,51/2,1,0,9,
            65/3,1,0,10,107/6,2,109/6,3,119/6,1,121/6,2,0,18,55/3,1,56/3,1,0];
        I:=[240378,240105,240085,239830,240123,239352,240352,240277,239238,
            239200,240346,240101,240279,240337,240489,239354,240326,240416,
            239823];
    elif hash1 eq 2462 and hash2 eq 334 and hash3 eq 1157 then
        V:=[16,1,17,7,18,11,19,17,20,23,21,31,23,37,25,41];
        D:=[1,38,1,42,1,0,3,85/2,2,0,5,74/3,1,35,1,0,7,163/6,1,211/6,1,0,9,65/3,
            1,82/3,1,83/3,1,0,12,79/4,1,167/6,1,0,14,137/6,1,0,15,181/12,1,0];
        I:=[516473,516501,516476,516525,516493,516432,516475,516512,516469,
            516532,516036,516506,515878,515807,516049];
    elif hash1 eq 3355 and hash2 eq 2986 and hash3 eq 1285 then
        V:=[22,1,23,5,24,11,25,21,26,29];
        D:=[1,68/3,2,0,3,115/6,1,39/2,1,0,5,52/3,1,53/3,6,18,1,20,2,0,15,95/6,1,
            97/6,1,33/2,1,0,18,44/3,1,0];
        I:=[246957,247366,247418,247488,247439,247234,247373,247398,247413,
            247414,247469,247468,246914,247158,247446,247445,247257,246887];
    elif hash1 eq 2462 and hash2 eq 1333 and hash3 eq 1116 then
        V:=[28,1,29,7,31,11,32,25,34,33];
        D:=[1,322/15,1,412/15,1,0,3,217/12,1,0,4,877/60,1,31/2,1,95/6,1,50/3,1,
            43/2,1,131/6,1,0,10,148/15,1,127/10,1,191/12,1,0,13,1006/105,1,68/7,
            1,146/15,1,148/15,1,0];
        I:=[534922,534977,534891,534883,534944,534945,534889,534994,534995,
            534902,534905,534980,534914,534896,534935,534989];
    elif hash1 eq 2697 and hash2 eq 838 and hash3 eq 1766 then
        V:=[22,1,23,5,24,17,25,23,26,31,27,35,28,43];
        D:=[1,17,1,0,2,35/2,1,115/6,1,39/2,1,47/2,1,51/2,1,0,7,16,1,85/4,1,0,9,
            85/6,1,95/6,1,177/10,1,0,12,43/3,1,0,13,431/30,1,173/12,1,89/6,1,0,
            16,193/15,1,53/4,1,0];
        I:=[503684,502809,503573,502513,502519,502346,503694,503651,503333,
            503570,502027,503305,503129,503165,502479,502048,502052];
    elif hash1 eq 958 and hash2 eq 805 and hash3 eq 948 then
        V:=[16,1,17,5,18,11,19,19,20,25,21,31,22,35,23,39,24,43,26,49,27,53];
        D:=[1,47,1,0,2,139/4,1,95/2,1,0,4,117/4,1,118/3,1,119/3,1,0,7,317/10,1,
            449/12,1,0,9,263/12,1,479/12,1,0,11,639/20,1,0,12,481/20,1,0,13,
            241/10,1,0,14,89/6,1,113/6,1,0,16,227/12,1,0,17,709/60,1,0];
        I:=[518625,518497,518542,518451,518593,518484,518390,518475,518453,
            518236,518298,518235,518217,518417,518295,518184,518426];
    elif hash1 eq 112 and hash2 eq 416 and hash3 eq 2514 then
        V:=[26,1,27,5,28,15,29,25,30,39];
        D:=[1,59/3,1,0,2,95/6,1,97/6,1,33/2,1,71/4,1,0,6,16,1,49/3,2,50/3,2,
            215/12,2,0,13,143/10,1,173/12,3,29/2,6,89/6,4,91/6,1,491/30,2,0,30,
            13,1,0];
        I:=[241226,241451,241707,241172,238981,240572,241453,241692,240797,
            241693,239098,239100,240164,238968,240600,241352,239093,241133,
            241268,241632,241658,241714,240573,241247,241266,241566,239013,
            238377,240006,239344];
    elif hash1 eq 1005 and hash2 eq 2728 and hash3 eq 3118 then
        V:=[14,1,15,7,16,11,17,17,18,21,19,27,20,33,21,37,22,41,23,45,25,49];
        D:=[1,28,1,40,1,0,3,69/2,2,0,5,92/3,1,104/3,1,0,7,235/6,1,0,8,100/3,1,
            141/4,1,0,10,107/6,1,377/12,1,0,12,388/15,1,0,13,33/2,1,0,14,248/15,
            1,0,15,781/42,1,0,16,1271/84,1,0];
        I:=[518725,518605,518712,518744,518723,518740,518597,518717,518669,
            518720,518606,518425,518741,518435,518310,518291];
    elif hash1 eq 3026 and hash2 eq 324 and hash3 eq 995 then
        V:=[20,1,21,9,22,21,23,31,24,37,26,43,27,53];
        D:=[1,101/3,1,107/3,1,36,1,0,4,167/6,1,169/6,1,297/10,1,179/6,1,181/6,2,
            0,10,131/5,2,79/3,1,418/15,1,85/3,2,0,16,45/2,1,731/30,1,0,18,
            308/15,1,313/15,1,0,20,309/20,1,187/12,1,47/3,1,263/15,4,0,27,
            433/28,1,937/60,2,157/10,1,0];
        I:=[426714,427330,427332,427152,426982,426856,427095,426713,427162,
            426732,426855,427150,426583,426605,426612,427106,426602,426865,
            426596,426639,426975,426977,426520,426521,426552,426570,426467,
            426445,426465,426455];
    elif hash1 eq 1616 and hash2 eq 340 and hash3 eq 840 then
        V:=[22,1,23,5,24,15,25,31,26,41,27,51,28,63,29,71,30,75];
        D:=[1,29,1,0,2,139/6,1,151/6,2,59/2,1,187/6,1,0,7,64/3,2,65/3,1,93/4,1,
            70/3,3,71/3,2,76/3,1,77/3,1,0,18,109/6,2,119/6,1,121/6,2,145/6,1,0,
            24,268/15,1,215/12,1,73/4,1,55/3,3,0,30,33/2,1,167/10,1,101/6,4,
            221/12,1,75/4,3,0,40,179/12,1,15,1,199/12,1,0,43,91/6,1,0,44,139/12,
            1,0];
        I:=[248815,249080,249053,249202,249259,248248,248192,249177,249465,
            247928,248546,249113,249178,248201,249267,247936,248238,248230,
            248563,248499,248233,249233,248171,248548,248212,247776,248114,
            248203,248505,248460,248419,248081,248220,248491,249028,248393,
            247692,248389,248392,248622,248975,247502,247503,248374];
    elif hash1 eq 3073 and hash2 eq 2266 and hash3 eq 932 then
        V:=[24,1,25,11,26,29,27,39,28,45];
        D:=[1,18,1,59/3,2,65/3,2,71/3,1,0,7,33/2,1,71/4,1,107/6,2,109/6,2,119/6,
            2,121/6,1,213/10,1,217/10,1,0,18,65/4,2,49/3,3,215/12,1,109/5,1,0,
            25,29/2,1,91/6,2,0,28,40/3,1,179/12,1,0];
        I:=[414822,414867,414872,414568,414606,413917,414617,414551,414184,
            414224,414831,414842,413707,414523,414850,413932,413897,414530,
            414559,414445,414848,414851,413709,413468,414823,414527,414691,
            413995,413489];
    elif hash1 eq 3073 and hash2 eq 1522 and hash3 eq 477 then
        V:=[24,1,25,5,26,17,27,23,28,31,29,35,30,45,31,51];
        D:=[1,89/3,1,0,2,305/12,1,257/10,1,167/6,1,169/6,1,181/6,1,0,7,111/5,1,
            67/3,1,0,9,197/12,1,113/6,1,207/10,1,0,12,21,1,0,13,299/20,1,461/30,
            1,205/12,2,217/12,1,0,18,787/60,1,163/12,1,0,20,361/30,1,0];
        I:=[417676,417505,417677,417786,417670,417595,416079,415481,417770,
            417356,417501,417794,415436,417559,416148,416627,415027,417709,
            417766,415374];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 19 mod 47.
function cached_ids_19(P,hash1,hash2,hash3)
    if hash1 eq 2980 and hash2 eq 958 and hash3 eq 1003 then
        V:=[18,1,19,5,20,15,21,31,22,49,23,67,24,79,25,87,26,95,27,103];
        D:=[1,98/3,1,0,2,55/2,1,175/6,1,199/6,1,139/4,1,0,6,28,1,146/5,1,117/4,
            1,89/3,1,94/3,1,107/3,1,149/4,1,0,13,257/10,1,111/4,1,353/12,1,
            179/6,1,191/6,1,333/10,1,203/6,1,205/6,1,0,21,68/3,1,121/5,1,388/15,
            1,26,1,105/4,1,412/15,1,442/15,1,208/7,1,0,29,265/12,1,269/12,1,
            227/10,1,511/20,1,161/6,1,0,34,428/21,1,247/12,1,21,1,0,37,521/30,1,
            379/20,1,265/12,1,0,40,86/5,1,401/21,1,77/4,1,0,43,421/30,3,85/6,1,
            917/60,1,0];
        I:=[514486,514493,514485,514582,514933,514728,514481,514929,514713,
            514984,514359,514368,514484,514905,514653,514864,514367,515007,
            514366,514120,514364,514038,514692,514829,514276,515005,514569,
            514122,514618,514807,514447,514549,514017,514035,514620,514275,
            514239,514007,514063,514278,514059,514060,514013,514446,514818,
            514365,514162];
    elif hash1 eq 2087 and hash2 eq 1422 and hash3 eq 2223 then
        V:=[22,1,24,7,25,13,26,19,27,29];
        D:=[1,21,1,86/3,1,0,3,53/3,2,59/3,2,0,7,107/6,2,37/2,1,0,10,191/12,2,16,
            2,49/3,2,99/5,1,0,17,431/30,1,173/12,1,91/6,1,0];
        I:=[249813,248935,248759,249536,248747,248931,248990,249149,249537,
            248911,249501,248898,249493,248767,249140,247586,247769,247798,
            247979];
    elif hash1 eq 2651 and hash2 eq 15 and hash3 eq 2117 then
        V:=[21,1,22,5,23,13,24,21,25,33,26,41,27,55,28,69];
        D:=[1,175/6,1,0,2,82/3,1,83/3,4,89/3,1,0,8,131/6,1,143/6,3,155/6,1,0,13,
            61/3,2,62/3,1,328/15,1,111/5,1,67/3,1,0,19,221/12,1,37/2,1,113/6,1,
            0,22,253/15,1,52/3,1,369/20,1,223/12,1,283/15,1,227/12,1,0,28,59/4,
            1,15,1,181/12,3,91/6,1,461/30,2,31/2,4,0,40,61/4,1,0];
        I:=[252450,252359,252197,252451,252452,252455,252279,252289,251984,
            252196,252290,252195,252170,252361,252292,252042,252038,251980,
            252035,252118,252165,252030,252028,251982,251848,251871,251892,
            252017,252062,252015,252063,252065,252115,251995,252001,252044,
            252281,252443,252454,251838];
    elif hash1 eq 2134 and hash2 eq 689 and hash3 eq 2448 then
        V:=[22,1,23,15,24,29,25,33,26,41,27,47,28,59];
        D:=[1,88/3,1,125/4,2,100/3,1,34,1,106/3,1,107/3,1,0,8,305/12,1,53/2,1,
            277/10,1,167/6,1,169/6,1,313/10,1,0,14,74/3,1,0,15,29/2,1,91/6,1,
            33/2,3,0,20,277/15,1,19,1,0,22,15,1,451/30,2,181/12,1,91/6,1,559/30,
            1,0,28,66/5,1,14,1,317/21,1,907/60,1,238/15,2,0];
        I:=[510707,509226,509708,510504,509407,509433,510205,509517,509186,
            510204,509241,509190,508517,509221,510765,509217,509406,510756,
            510764,508598,508312,509741,509240,509272,510272,510739,508446,
            510141,509311,508820,508827,509213,509314];
    elif hash1 eq 3121 and hash2 eq 3100 and hash3 eq 1272 then
        V:=[18,1,19,15,20,25,21,29];
        D:=[1,26,1,28,1,30,1,34,1,36,1,38,1,0,7,53/2,2,57/2,3,61/2,2,65/2,1,0,
            15,23,1,0,16,139/6,2,0];
        I:=[251657,251653,251654,251670,251655,251668,251652,251672,251660,
            251666,251671,251649,251651,251656,251659,251619,251661];
    elif hash1 eq 66 and hash2 eq 1078 and hash3 eq 1037 then
        V:=[26,1,27,5,28,19,29,23,30,31,31,39];
        D:=[1,71/3,1,0,2,97/6,1,71/4,1,107/6,1,109/6,1,197/10,1,41/2,1,0,8,50/3,
            2,0,10,29/2,2,147/10,1,89/6,2,0,15,151/12,1,13,2,76/5,1,0,19,71/6,1,
            27/2,1,0];
        I:=[403346,403708,403744,403704,403702,403367,403835,403530,403843,
            403550,403693,403371,403368,403457,403814,403463,403797,402929,
            402993,402755];
    elif hash1 eq 207 and hash2 eq 1023 and hash3 eq 2398 then
        V:=[20,1,21,7,22,15,23,21,24,27,25,37,26,41];
        D:=[1,27,2,31,1,0,4,151/6,2,51/2,1,55/2,1,0,8,65/3,1,71/3,1,0,10,121/6,
            2,41/2,1,0,13,55/3,2,56/3,1,62/3,1,263/12,1,0,18,33/2,1,0,19,76/5,1,
            0];
        I:=[252272,252432,252313,252268,252273,252324,252304,252226,252008,
            252071,252233,252405,252253,252306,252453,252005,251852,252224,
            251885];
    elif hash1 eq 160 and hash2 eq 955 and hash3 eq 2852 then
        V:=[23,1,24,7,25,17,26,25,27,39,28,47,30,53,31,59];
        D:=[1,55/2,1,199/6,1,0,3,77/3,2,89/3,2,95/3,1,101/3,1,0,9,143/6,3,155/6,
            1,157/6,1,0,14,101/5,1,61/3,1,263/12,1,67/3,1,73/3,1,382/15,1,0,20,
            255/14,1,551/30,1,20,1,0,23,253/15,1,277/15,1,0,25,41/3,1,46/3,1,0,
            27,499/42,1,0];
        I:=[250181,250034,250078,250631,249991,249995,249992,250003,250045,
            250050,250311,250077,250030,250692,250144,250025,250190,250024,
            249868,249835,250306,250388,250021,249952,250252,249859,249840];
    elif hash1 eq 583 and hash2 eq 2862 and hash3 eq 959 then
        V:=[19,1,20,7,21,15,22,25,23,33,24,39];
        D:=[1,49/2,1,53/2,1,0,3,68/3,1,23,3,74/3,1,0,8,127/6,4,43/2,1,47/2,2,
            99/4,1,0,16,59/3,2,20,2,85/4,1,0,21,109/6,1,197/10,1,0,23,50/3,2,0];
        I:=[252466,252477,252278,252457,252458,252465,252390,252295,252398,
            252407,252422,252460,252069,252090,252091,252259,252296,252055,
            252096,252191,252093,252016,252031,252221];
    elif hash1 eq 2087 and hash2 eq 3398 and hash3 eq 1133 then
        V:=[24,1,25,11,26,15,28,23,29,37,30,41];
        D:=[1,76/3,1,88/3,1,125/4,1,133/4,1,0,5,305/12,1,0,6,73/4,1,68/3,1,
            131/5,1,0,9,44/3,1,223/15,1,15,1,199/12,3,37/2,1,41/2,1,0,17,213/14,
            1,0,18,787/60,1,66/5,2,41/3,1,0];
        I:=[502691,502692,503689,503676,503348,503355,503683,503036,503632,
            503391,502673,502776,502800,502801,502773,502747,502039,503349,
            503346,503362,503673];
    elif hash1 eq 912 and hash2 eq 2016 and hash3 eq 55 then
        V:=[24,1,25,5,26,11,27,19,28,27];
        D:=[1,68/3,1,0,2,115/6,3,39/2,4,0,9,53/3,1,18,1,59/3,1,0,12,95/6,1,97/6,
            3,33/2,2,0,18,57/4,1,43/3,4,0];
        I:=[670604,670634,670685,670737,670612,670727,670852,670858,670790,
            670848,670706,670109,670067,670222,670617,670225,670684,670648,
            670100,670658,670679,670756];
    elif hash1 eq 1194 and hash2 eq 2357 and hash3 eq 9 then
        V:=[22,1,23,7,24,13,25,19];
        D:=[1,21,1,23,1,0,3,115/6,1,39/2,5,0,9,53/3,4,20,1,0,14,97/6,1,33/2,1,
            0];
        I:=[673210,673137,673141,673147,673172,673194,673209,673211,673076,
            673142,673182,673202,673022,673169,673074];
    elif hash1 eq 3356 and hash2 eq 2905 and hash3 eq 2482 then
        V:=[23,1,24,7,25,15,26,21,27,33];
        D:=[1,115/6,2,91/4,1,0,4,53/3,3,59/3,1,20,1,0,9,95/6,1,177/10,1,0,11,
            44/3,2,81/5,1,65/4,1,49/3,1,50/3,1,0,17,91/6,1,0];
        I:=[414828,414845,414603,414838,414849,414880,414352,414785,414443,
            414021,414218,414235,414240,414402,414349,413677,413477];
    elif hash1 eq 536 and hash2 eq 201 and hash3 eq 2757 then
        V:=[19,1,20,11,21,29,22,49,23,61,24,69,25,83];
        D:=[1,45/2,1,53/2,1,65/2,2,77/2,1,0,6,62/3,1,21,1,80/3,1,27,1,86/3,3,29,
            1,92/3,2,33,1,0,17,75/4,1,115/6,3,127/6,1,43/2,1,151/6,1,163/6,1,
            55/2,1,187/6,2,63/2,1,0,29,53/3,1,18,2,85/4,1,65/3,1,26,1,0,35,
            107/6,4,233/12,2,119/6,1,0,42,238/15,1,49/3,2,35/2,1,91/5,1,56/3,2,
            61/3,1,0,50,147/10,1,16,1,193/12,3,33/2,1,113/6,1,115/6,1,0];
        I:=[420254,420605,420519,420583,420373,420588,420535,420090,420602,
            420207,420404,420607,419245,419454,420616,420327,420094,419985,
            420567,420568,419529,420062,420109,420408,420411,419084,420413,
            420381,420359,420332,420350,419531,420345,419567,419083,420124,
            420177,420542,419833,420255,419744,419714,419536,419811,419677,
            419099,419490,420311,420223,419258,419753,419123,419761,420217,
            420383,419008,419673];
    elif hash1 eq 818 and hash2 eq 605 and hash3 eq 226 then
        V:=[20,1,21,7,22,19,23,27,24,35,25,41,26,51,27,61,28,81];
        D:=[1,35,2,110/3,1,0,4,123/4,1,199/6,1,67/2,1,211/6,1,71/2,1,0,9,28,1,
            89/3,1,125/4,1,0,12,107/6,1,257/10,1,155/6,2,0,16,268/15,1,20,1,0,
            18,193/12,1,197/12,1,18,1,611/30,1,0,22,289/20,1,278/15,1,283/15,2,
            227/12,1,0,27,291/20,1,613/42,1,439/30,2,44/3,1,59/4,1,149/10,2,
            451/30,2,31/2,1,95/6,1,0,39,59/4,1,0];
        I:=[420526,420606,420387,419724,420603,419855,419603,420314,419455,
            419857,419831,420213,419071,419252,419784,419565,420283,419827,
            419608,419699,419250,419292,419257,419122,419549,419007,419803,
            419009,419015,419448,420432,419459,419628,419679,419828,419870,
            419745,420364,418943];
    elif hash1 eq 1523 and hash2 eq 1069 and hash3 eq 1244 then
        V:=[23,1,24,11,25,17,26,23,27,31,29,39];
        D:=[1,167/6,1,297/10,1,203/6,1,449/12,1,0,5,194/7,1,206/7,1,0,7,551/30,
            1,47/2,1,0,9,259/12,1,173/7,1,373/15,1,0,12,397/24,1,257/12,1,
            649/30,1,0,15,917/60,1,0];
        I:=[511680,510906,511690,511268,510913,511006,510874,511205,510952,
            510812,510890,510877,510978,510881,510975];
    elif hash1 eq 1805 and hash2 eq 629 and hash3 eq 934 then
        V:=[21,1,22,5,23,17,24,27,25,39,26,49,27,65];
        D:=[1,35/2,1,0,2,53/3,1,59/3,1,109/4,1,28,1,88/3,1,0,7,95/6,1,97/6,1,
            145/6,1,103/4,1,0,11,81/5,1,49/3,1,215/12,2,62/3,1,21,1,0,17,29/2,1,
            491/30,2,33/2,1,279/14,1,0,22,44/3,1,15,1,46/3,1,199/12,1,117/7,1,
            203/12,1,302/15,1,0,29,161/12,1,0];
        I:=[513868,513745,513863,513905,513482,513475,513904,513869,513496,
            513584,513719,513819,513617,513638,513494,513489,513725,513444,
            513590,513582,513437,513853,513823,513479,513755,513433,513366,
            513312,513451];
    elif hash1 eq 207 and hash2 eq 2810 and hash3 eq 980 then
        V:=[34,1,35,11,36,21];
        D:=[1,167/12,1,14,2,71/5,1,43/3,4,0,9,149/12,1,25/2,1,51/4,1,77/6,3,0,
            15,34/3,1,0];
        I:=[54545,54962,55266,55669,54761,54808,55514,55568,55222,55802,55939,
            54579,54699,55969,54418];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 20 mod 47.
function cached_ids_20(P,hash1,hash2,hash3)
    if hash1 eq 3310 and hash2 eq 459 and hash3 eq 3279 then
        V:=[24,1,25,5,26,19,27,25,28,33];
        D:=[1,19,1,0,2,31/2,1,103/6,1,35/2,3,115/6,2,39/2,1,83/4,1,0,11,86/5,1,
            69/4,1,0,13,157/10,1,63/4,2,109/6,1,0,17,44/3,2,0];
        I:=[245717,245707,245706,245750,245792,245799,245230,245366,245165,
            243351,243846,242663,243869,243874,244929,242455,243152,243349];
    elif hash1 eq 2746 and hash2 eq 787 and hash3 eq 1548 then
        V:=[27,1,28,9,29,17,30,25,31,31,32,41];
        D:=[1,151/6,3,107/4,1,163/6,1,0,6,53/3,1,18,1,106/5,1,0,9,97/6,1,107/6,
            1,109/6,1,0,12,167/12,1,191/12,2,0,15,29/2,2,147/10,1,479/30,1,33/2,
            1,0,20,188/15,1,13,1,0];
        I:=[233986,234028,234120,233239,233104,231870,233881,233961,233005,
            233101,233518,233851,233266,234002,233270,233375,234066,230064,
            229195,233382,232125];
    elif hash1 eq 1477 and hash2 eq 750 and hash3 eq 367 then
        V:=[22,1,23,5,24,17,25,23,26,39,28,45,29,53];
        D:=[1,71/3,1,0,2,87/4,1,131/6,1,281/12,1,253/10,1,103/4,1,0,7,55/3,1,
            79/3,1,0,9,551/30,1,203/10,1,245/12,1,41/2,1,265/12,1,47/2,1,49/2,1,
            0,16,248/15,1,278/15,1,0,18,67/5,1,208/15,1,907/60,1,0,21,377/28,1,
            279/20,1,0];
        I:=[513543,513539,513542,513568,513555,513578,513537,513392,513521,
            513529,513522,513471,513411,513573,513324,513561,513316,513417,
            513414,513424,513303,513300];
    elif hash1 eq 114 and hash2 eq 1822 and hash3 eq 931 then
        V:=[25,1,26,13,27,17,28,27,30,37];
        D:=[1,39/2,1,127/6,2,43/2,2,139/6,1,47/2,1,0,8,20,1,0,9,107/6,1,109/6,2,
            119/6,2,121/6,1,0,15,43/3,1,49/3,2,50/3,1,55/3,1,0,20,193/15,2,66/5,
            1,40/3,1,0];
        I:=[240039,239435,240186,239649,240269,240063,240184,239471,240382,
            239780,240133,240056,240198,239621,240379,239251,240216,239467,
            240402,240365,240366,239776,239769];
    elif hash1 eq 1853 and hash2 eq 2026 and hash3 eq 2466 then
        V:=[22,1,23,11,24,17,25,31,26,41,27,51,28,67];
        D:=[1,25,1,86/3,1,29,2,33,1,0,6,151/6,4,51/2,3,0,13,85/4,1,65/3,4,93/4,
            1,70/3,1,71/3,2,77/3,1,0,23,197/10,2,119/6,1,217/10,1,131/6,2,0,29,
            215/12,2,73/4,1,55/3,3,56/3,3,0,38,491/30,1,33/2,1,551/30,2,221/12,
            2,37/2,2,187/10,1,113/6,1,0,48,212/15,1,85/6,1,29/2,1,0];
        I:=[674102,673917,673974,674012,674009,673789,673815,673889,674026,
            673975,674033,674117,673807,673845,673927,673990,674031,673809,
            673846,673869,674007,673857,673802,673803,673977,673759,673780,
            673860,673796,673818,673808,673842,673959,673972,673841,673971,
            673981,673798,673762,673755,673766,673752,673767,673756,673773,
            673760,673758,673772,673848,673819];
    elif hash1 eq 2464 and hash2 eq 2565 and hash3 eq 184 then
        V:=[29,1,30,5,31,15,33,23];
        D:=[1,143/6,1,0,2,239/12,2,20,1,62/3,1,22,1,0,7,33/2,1,101/6,1,103/6,1,
            0,10,401/30,1,27/2,2,83/6,1,291/20,2,0];
        I:=[227264,226250,227810,225188,226138,227487,226381,226158,225587,
            225840,225094,225585,225457,224422,224504];
    elif hash1 eq 1994 and hash2 eq 3390 and hash3 eq 1650 then
        V:=[25,1,28,5,29,11,30,17,31,35];
        D:=[1,119/6,1,0,2,175/12,1,44/3,1,0,4,451/30,1,181/12,1,0,6,77/6,1,67/5,
            1,203/15,2,411/28,1,74/5,1,317/21,1,907/60,1,91/6,1,0,15,361/30,2,
            0];
        I:=[417350,417438,418588,417436,417444,416563,416559,417439,417725,
            415380,415015,415429,415996,415428,416477,416758];
    elif hash1 eq 772 and hash2 eq 366 and hash3 eq 2756 then
        V:=[18,1,19,7,20,19,21,33,22,49,23,57,25,61,26,75,27,83,28,87];
        D:=[1,35,1,41,1,0,3,175/6,1,131/4,1,75/2,1,235/6,2,247/6,1,0,9,125/4,1,
            101/3,2,34,1,149/4,1,112/3,1,113/3,1,0,16,131/6,1,155/6,1,297/10,1,
            377/12,1,317/10,1,127/4,1,193/6,1,0,23,47/2,1,311/12,1,85/3,1,0,26,
            611/30,1,0,27,59/4,1,371/20,1,379/20,1,229/12,1,581/30,1,233/12,2,0,
            34,498/35,1,46/3,1,101/6,1,0,37,185/12,1,0,38,233/20,1,176/15,1,0];
        I:=[516596,516619,516109,515972,516160,515901,516632,516222,516163,
            516422,516629,516061,516020,516095,515945,516096,516536,516076,
            515865,516165,516308,516311,515868,516205,516211,515924,516111,
            515784,515711,515785,515906,515797,515905,515772,516638,515710,
            515703,516290,516037];
    elif hash1 eq 443 and hash2 eq 2838 and hash3 eq 1174 then
        V:=[14,1,15,5,16,9,17,17,19,23,20,31,22,41];
        D:=[1,26,1,0,2,65/2,1,0,3,25,1,29,1,37,2,0,7,187/6,1,63/2,2,0,10,107/6,
            1,143/6,1,145/6,1,0,13,56/3,1,61/3,1,328/15,1,67/3,1,0,17,50/3,1,
            278/15,1,56/3,1,19,2,0];
        I:=[428413,428402,428410,428398,428431,428447,428414,428405,428429,
            428430,428418,428401,427712,427731,427872,427898,428453,427710,
            427709,427713,427727];
    elif hash1 eq 866 and hash2 eq 3222 and hash3 eq 2243 then
        V:=[20,1,21,7,22,15,23,23,24,29,25,39,26,45];
        D:=[1,86/3,1,31,1,0,3,151/6,2,51/2,1,163/6,1,0,7,64/3,1,26,1,30,1,0,10,
            71/4,1,133/6,1,0,12,55/3,1,56/3,1,99/5,1,20,1,0,16,173/12,1,29/2,1,
            0,18,175/12,1,0];
        I:=[513843,513847,513838,513854,513797,513845,513862,513801,513315,
            513814,513840,513852,513834,513532,513570,513519,513620,513851];
    elif hash1 eq 2229 and hash2 eq 2453 and hash3 eq 2502 then
        V:=[25,1,26,7,27,15,28,19,29,27,30,31,31,37];
        D:=[1,51/2,2,59/2,1,0,4,64/3,2,76/3,1,26,1,0,8,41/2,2,0,10,91/5,1,
            298/15,1,61/3,1,0,13,551/30,1,0,14,76/5,1,248/15,1,0,16,391/30,1,0];
        I:=[243299,243501,244064,243094,243549,244096,244187,243464,243712,
            242297,242257,244578,242323,242324,242307,244169];
    elif hash1 eq 1148 and hash2 eq 1884 and hash3 eq 2109 then
        V:=[26,1,27,7,28,17,29,25,31,29];
        D:=[1,71/3,2,77/3,3,0,6,87/4,1,131/6,2,133/6,1,145/6,1,0,11,55/3,1,20,1,
            61/3,1,0,14,221/12,3,0,17,401/30,1,27/2,1,0];
        I:=[670372,670431,670415,670423,670556,669965,670376,670475,669960,
            670406,670726,670371,670001,669985,669994,670258,669990,670255];
    elif hash1 eq 2558 and hash2 eq 2180 and hash3 eq 2174 then
        V:=[27,1,28,7,29,19,30,27,31,33];
        D:=[1,115/6,2,39/2,2,0,5,47/3,1,52/3,1,53/3,1,18,2,96/5,1,0,11,95/6,2,
            97/6,2,33/2,1,0,16,43/3,1,44/3,1,0,18,77/6,1,0];
        I:=[667426,667486,667390,667393,667252,667264,667346,665573,666885,
            665505,665613,666027,665687,667414,666013,667213,665546,665114];
    elif hash1 eq 161 and hash2 eq 2337 and hash3 eq 2690 then
        V:=[26,1,27,11,28,19,29,25,30,29];
        D:=[1,62/3,1,21,1,23,3,74/3,1,0,7,115/6,1,39/2,2,127/6,1,0,11,46/3,1,16,
            1,0,13,109/6,1,0,14,57/4,1,43/3,1,0];
        I:=[669308,669279,668759,668800,668947,668815,668705,668874,669282,
            668868,668858,668854,668688,668801,669210];
    elif hash1 eq 2323 and hash2 eq 143 and hash3 eq 558 then
        V:=[26,1,27,5,28,13,29,21,30,29,31,33];
        D:=[1,83/3,1,0,2,217/10,1,131/6,1,133/6,2,0,6,268/15,1,55/3,1,19,1,0,9,
            551/30,1,221/12,1,115/6,1,0,12,76/5,2,0,14,449/30,1,181/12,1,0];
        I:=[404057,404294,404300,404272,404285,404263,404067,404293,404062,
            404273,404303,404068,404309,404279,404106];
    elif hash1 eq 2793 and hash2 eq 2781 and hash3 eq 612 then
        V:=[17,1,18,5,19,9,20,15,21,21,24,35,25,39,26,51];
        D:=[1,85/2,1,0,2,37,1,0,3,211/6,1,223/6,1,0,5,88/3,1,94/3,1,0,7,329/12,
            1,353/12,1,179/6,2,181/6,1,377/12,1,401/12,1,0,14,401/20,1,0,15,
            169/10,1,521/30,2,35/2,1,201/10,1,411/20,1,0,21,302/15,1,0];
        I:=[515150,515366,515153,515045,515338,515343,515038,515339,515348,
            515350,515362,515031,514962,514258,514722,514979,514999,515352,
            514155,514043,513980];
    elif hash1 eq 1665 and hash2 eq 3394 and hash3 eq 2155 then
        V:=[20,1,21,9,22,19,23,31,24,43,25,53,26,59];
        D:=[1,23,1,27,1,29,1,0,4,127/6,1,151/6,2,51/2,2,55/2,1,0,10,64/3,1,65/3,
            1,22,3,77/3,1,26,1,0,17,109/6,1,119/6,1,87/4,1,131/6,1,133/6,1,0,22,
            91/5,1,55/3,2,56/3,9,263/12,1,0,35,113/6,2,115/6,4,0,41,223/15,1,15,
            1,0];
        I:=[252366,252350,252351,252388,252287,252383,252182,252395,252152,
            252419,252148,252156,252391,252393,252060,252205,252357,252360,
            251999,252167,252179,251994,251959,251977,251946,251955,252006,
            252076,252160,252180,252184,252368,252369,251970,251856,252027,
            251853,251874,251882,251933,251998,252045];
    elif hash1 eq 1994 and hash2 eq 1431 and hash3 eq 2193 then
        V:=[19,1,20,13,21,29,22,35,23,49,24,57,26,67];
        D:=[1,45/2,1,49/2,1,53/2,2,57/2,1,65/2,1,0,7,21,2,68/3,1,23,1,74/3,2,25,
            1,27,1,86/3,1,0,16,139/6,4,55/2,1,0,21,18,1,58/3,1,59/3,1,85/4,1,
            65/3,2,126/5,1,0,28,97/6,1,257/12,1,87/4,1,0,31,49/3,1,268/15,1,18,
            1,55/3,1,0,35,47/3,1,0];
        I:=[424730,424707,424613,424735,424618,424622,424654,424747,424100,
            424708,424612,424616,424621,424583,423775,423930,424139,424162,
            424300,423042,422646,424655,422656,423786,424118,424754,422633,
            424041,422803,422799,424314,422652,422647,422666,422438];
    elif hash1 eq 1947 and hash2 eq 3265 and hash3 eq 1077 then
        V:=[22,1,23,5,24,17,25,21,26,29,27,37,28,45,29,61,30,69,31,75];
        D:=[1,109/4,1,0,2,109/6,1,121/6,1,281/12,1,257/10,1,353/12,1,0,7,151/5,
            1,0,8,167/10,1,551/30,2,113/6,1,0,12,76/5,1,283/15,1,96/5,1,0,15,
            793/42,1,127/6,1,641/30,1,0,18,203/15,2,14,1,137/9,1,323/21,1,
            309/20,1,582/35,1,394/21,1,0,26,119/10,1,239/20,2,431/30,1,0,30,
            122/9,1,41/3,1,0,32,1073/90,1,0];
        I:=[540205,540222,539824,539825,540178,540216,540239,540255,540167,
            540241,540210,540229,540192,540240,539982,540036,540188,540223,
            540249,540206,539696,539964,539639,539968,539980,540194,540196,
            540234,540195,539692,539695,539640];
    elif hash1 eq 3028 and hash2 eq 1281 and hash3 eq 1921 then
        V:=[21,1,23,5,24,13,25,29,26,37,27,45,28,55,29,77,30,85];
        D:=[1,139/6,1,0,2,257/12,1,119/4,1,353/10,1,0,5,298/15,1,73/3,1,129/5,1,
            30,1,121/4,1,129/4,1,34,1,0,12,391/20,1,293/12,1,55/2,1,0,15,278/15,
            1,58/3,1,299/12,1,0,18,401/30,2,461/28,1,529/30,1,265/14,1,0,23,
            313/24,1,274/21,1,196/15,1,67/5,1,582/35,1,1003/60,1,101/6,1,
            1027/60,1,263/15,1,563/28,1,0,33,263/20,1,137/10,1,603/40,1,0,36,
            1216/105,1,41/3,1,0];
        I:=[540065,540079,539811,540084,540080,540053,540157,540007,540003,
            540004,539848,540150,539917,540059,539931,540052,539788,539804,
            539940,539967,540160,539862,539801,539979,540122,540118,539941,
            540033,540104,539911,539765,539799,539932,539858,539816,539984,
            539689];
    elif hash1 eq 1759 and hash2 eq 1260 and hash3 eq 2442 then
        V:=[26,1,27,9,28,17,29,27,30,35,31,41];
        D:=[1,22,1,71/3,1,77/3,1,0,4,107/6,1,133/6,1,95/4,1,0,7,215/12,1,18,1,
            91/5,1,239/12,1,0,11,91/6,2,193/12,1,113/6,1,0,15,248/15,1,253/15,2,
            0,18,27/2,1,0];
        I:=[407393,407603,407627,407536,407770,407905,407900,407017,407551,
            407899,407372,407403,407485,407775,407548,407935,407968,407983];
    elif hash1 eq 3169 and hash2 eq 1771 and hash3 eq 3112 then
        V:=[21,1,22,7,23,19,24,25,25,31,26,37,27,47,28,59];
        D:=[1,199/6,1,211/6,2,0,4,109/4,1,82/3,1,83/3,1,125/4,1,95/3,1,0,9,
            237/10,1,305/12,1,0,11,298/15,1,239/12,1,0,13,241/12,1,41/2,1,0,15,
            41/3,1,179/12,1,361/20,1,369/20,2,0,20,401/30,1,44/3,1,451/30,3,
            91/6,1,1093/60,1,0,27,787/60,1,0];
        I:=[424824,422887,424487,423501,424503,423905,422606,423686,422888,
            423488,423590,422542,422496,423627,423632,423675,422501,423636,
            423643,423170,424209,422713,423689,423692,424779,422340,423197];
    elif hash1 eq 1712 and hash2 eq 3270 and hash3 eq 473 then
        V:=[27,1,28,7,29,21,30,31,31,39];
        D:=[1,133/6,1,145/6,1,0,3,50/3,1,55/3,1,56/3,1,61/3,2,111/5,1,67/3,1,0,
            10,197/12,1,167/10,1,101/6,2,103/6,1,0,15,179/12,1,15,2,33/2,1,0,19,
            83/6,3,0];
        I:=[237575,237202,237555,237673,237613,237560,237670,237542,237102,
            237590,237432,237563,237591,237548,237574,237107,237131,236972,
            237562,237897,237898];
    elif hash1 eq 678 and hash2 eq 944 and hash3 eq 3263 then
        V:=[20,1,21,11,22,23,23,27,24,41,25,51,26,63];
        D:=[1,29,1,92/3,2,31,2,35,1,0,7,43/2,1,107/4,1,163/6,1,55/2,1,175/6,1,0,
            12,70/3,1,0,13,71/4,1,107/6,1,109/6,2,79/4,1,119/6,2,121/6,1,0,21,
            50/3,1,55/3,1,99/5,1,61/3,1,0,25,16,1,197/12,1,33/2,2,101/6,1,
            551/30,1,0,31,179/12,1,0];
        I:=[424855,424173,424797,424857,424859,423903,424858,424501,424383,
            423991,424549,424463,424524,424774,424316,424523,424519,424382,
            424509,424295,423902,424283,423638,424284,424528,424339,424358,
            424391,424310,422538,423440];
    elif hash1 eq 1806 and hash2 eq 2040 and hash3 eq 2462 then
        V:=[19,1,20,9,21,15,22,23,23,27];
        D:=[1,57/2,1,69/2,1,77/2,1,0,4,92/3,1,110/3,1,0,6,115/4,1,175/6,5,187/6,
            1,0,13,126/5,1,0,14,143/6,2,0];
        I:=[251667,251648,251635,251629,251604,251606,251607,251608,251620,
            251637,251642,251610,251609,251640,251641];
    elif hash1 eq 443 and hash2 eq 147 and hash3 eq 347 then
        V:=[21,1,22,9,23,23,24,31,25,37,26,43,27,53,28,67];
        D:=[1,55/2,1,59/2,1,67/2,1,0,4,24,1,83/3,1,28,1,89/3,1,156/5,1,95/3,2,0,
            11,145/6,1,49/2,1,103/4,1,0,14,101/5,1,62/3,1,0,16,33/2,1,101/6,2,0,
            19,33/2,1,17,1,52/3,1,227/12,1,0,23,27/2,1,44/3,1,181/12,1,461/30,1,
            31/2,1,50/3,1,0,29,787/60,1,66/5,1,0];
        I:=[424473,424755,424412,424172,424834,424174,424544,423447,424773,
            424823,422704,424378,422630,423450,423551,424394,424448,424776,
            422574,424396,423691,422547,423652,424393,423433,423722,424510,
            422546,423611,423671];
    elif hash1 eq 1618 and hash2 eq 1072 and hash3 eq 1189 then
        V:=[20,1,21,7,22,11,23,19,24,27,25,33,26,41,27,49,28,63];
        D:=[1,29,1,33,1,0,3,187/6,1,0,4,109/4,1,83/3,2,30,1,0,8,121/6,1,45/2,1,
            143/6,1,0,11,263/12,1,68/3,2,0,14,37/2,1,113/6,2,245/12,1,0,18,46/3,
            1,17,1,52/3,1,0,21,451/30,1,181/12,1,91/6,2,461/30,1,511/30,1,103/6,
            2,0,29,15,2,232/15,1,0];
        I:=[421848,421846,421959,421847,421658,421952,421840,421797,421843,
            421306,421323,421795,421841,421926,421321,421820,421912,421842,
            421729,421233,421783,421334,421792,421828,421328,421225,421231,
            421232,421125,421218,421126];
    elif hash1 eq 114 and hash2 eq 1273 and hash3 eq 2354 then
        V:=[22,1,23,7,24,15,25,27,26,37,27,45];
        D:=[1,68/3,2,25,1,0,4,115/6,2,127/6,1,91/4,1,0,8,52/3,1,53/3,3,18,1,
            96/5,1,77/4,1,0,15,63/4,1,95/6,3,97/6,1,193/10,1,0,21,43/3,1,44/3,2,
            81/5,1,0,25,169/12,1,431/30,1,89/6,2,0];
        I:=[418071,418386,418041,418558,418652,417649,418420,418435,416333,
            418473,418619,418120,416323,418322,418359,417681,418311,418358,
            416805,416240,418574,416825,417448,416322,418577,416307,415025,
            418187];
    elif hash1 eq 2041 and hash2 eq 1066 and hash3 eq 1791 then
        V:=[30,1,31,5,32,13,33,21];
        D:=[1,53/3,3,0,4,63/4,1,95/6,3,97/6,1,0,9,43/3,2,44/3,1,191/12,2,0,14,
            77/6,1,79/6,1,0];
        I:=[658421,658464,658710,656512,657883,658412,659171,658662,658401,
            658656,657162,656198,656690,656713,659159];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 21 mod 47.
function cached_ids_21(P,hash1,hash2,hash3)
    if hash1 eq 1290 and hash2 eq 2431 and hash3 eq 1776 then
        V:=[19,1,20,5,21,13,22,27,23,41,24,55,25,63,26,71];
        D:=[1,49/2,1,0,2,68/3,1,23,1,86/3,1,0,5,127/6,3,43/2,2,151/6,5,51/2,3,
            55/2,1,59/2,1,0,20,18,1,77/4,1,58/3,3,64/3,1,65/3,3,116/5,1,0,30,
            177/10,1,107/6,5,197/10,1,79/4,1,119/6,1,41/2,2,0,41,50/3,2,18,1,
            55/3,1,0,45,89/6,2,167/10,1,551/30,1,0,49,44/3,1,223/15,1,15,1,46/3,
            1,17,1,0];
        I:=[424079,423944,424019,423791,422847,424007,424013,423783,423908,
            423020,423391,423915,423950,424026,423758,423764,423804,422736,
            422730,424242,423939,423893,424037,424063,423968,423789,424102,
            424267,422727,423034,423153,423982,424206,424210,424228,423252,
            424233,423330,423778,423799,423404,423808,422998,424012,423395,
            424285,423265,422999,424077,422349,422793,423831,422473];
    elif hash1 eq 2418 and hash2 eq 1552 and hash3 eq 276 then
        V:=[17,1,18,9,19,25,20,47,21,75,22,97,23,119,24,133,25,153,26,159,27,
            165];
        D:=[1,65/2,1,73/2,1,81/2,1,0,4,25,1,27,2,86/3,1,29,2,92/3,1,98/3,2,37,1,
            0,14,47/2,1,151/6,2,51/2,2,163/6,3,55/2,1,59/2,1,187/6,1,199/6,4,
            67/2,1,235/6,1,0,31,59/3,1,65/3,1,22,1,116/5,1,71/3,1,126/5,1,76/3,
            1,77/3,2,109/4,1,82/3,1,83/3,3,89/3,1,156/5,1,0,47,233/12,1,119/6,2,
            131/6,2,45/2,1,237/10,1,143/6,2,145/6,2,257/10,1,329/12,1,57/2,1,0,
            61,50/3,1,215/12,2,55/3,1,101/5,1,62/3,1,328/15,2,263/12,1,111/5,1,
            67/3,1,74/3,1,0,73,491/30,1,101/6,1,18,1,207/10,1,125/6,2,127/6,2,0,
            81,175/12,1,179/12,1,46/3,1,16,1,321/20,1,50/3,1,17,1,19,1,22,1,0,
            90,185/12,1,453/28,1,0,92,323/21,1,47/3,1,0,94,119/10,1,0];
        I:=[515088,515143,515446,515050,515012,515087,515144,515011,515112,
            515149,515014,515041,515053,515016,514708,515146,515094,515215,
            515018,515052,515148,515347,515349,515021,515042,515084,515131,
            515207,515381,514355,515111,514490,514489,514680,515152,514737,
            515066,514776,515044,514729,514522,514416,515299,515340,515151,
            514600,515036,515162,515170,514487,514507,515218,514662,514428,
            515233,514916,515445,514843,514551,514567,514597,515032,515085,
            515132,514801,514539,514651,514889,514571,514350,514753,514561,
            514891,514623,514281,514716,514706,514717,514574,515212,514859,
            515248,515055,514259,514328,514646,515137,514757,514056,514944,
            514032,514118,514070,514492];
    elif hash1 eq 2888 and hash2 eq 3253 and hash3 eq 1148 then
        V:=[22,1,23,5,24,11,25,17,26,27,28,33];
        D:=[1,21,1,0,2,103/6,1,115/6,1,0,4,53/3,2,65/3,1,0,7,95/6,1,97/6,1,
            209/12,1,41/2,1,0,11,49/3,3,262/15,1,0,15,66/5,1,0];
        I:=[418725,418809,418913,417897,418904,417516,418878,417872,417540,
            416343,416706,418876,418921,417799,416696];
    elif hash1 eq 2089 and hash2 eq 3347 and hash3 eq 1796 then
        V:=[26,1,27,7,28,13,29,19,30,27];
        D:=[1,21,1,23,7,0,9,115/6,1,127/6,3,0,13,16,1,53/3,1,0,15,95/6,1,97/6,2,
            33/2,2,0,20,43/3,1,44/3,1,0];
        I:=[60856,60557,60600,60643,60645,60659,60849,60860,60531,60484,60635,
            60698,60634,60568,60657,60596,60656,60486,60598,60404,60459];
    elif hash1 eq 3029 and hash2 eq 1592 and hash3 eq 2226 then
        V:=[26,1,27,5,28,11,29,17,30,27,31,33,32,41,33,47];
        D:=[1,59/3,1,0,2,233/12,1,217/10,1,0,4,50/3,1,268/15,1,0,6,431/30,2,
            89/6,1,197/12,1,207/10,1,0,11,40/3,1,46/3,1,0,13,117/10,1,73/6,1,
            85/6,1,0,16,178/15,2,269/20,1,0,19,301/30,1,0];
        I:=[494518,493128,494519,494476,494602,494421,494448,494590,494482,
            494613,494747,493213,494416,494520,494609,494505,494521,492624,
            493078];
    elif hash1 eq 2841 and hash2 eq 2308 and hash3 eq 2513 then
        V:=[25,1,26,13,27,23,28,33,29,39];
        D:=[1,35/2,1,115/6,1,39/2,2,127/6,1,43/2,2,0,8,52/3,1,53/3,4,18,1,58/3,
            2,0,16,63/4,3,95/6,6,97/6,6,109/6,2,0,33,43/3,3,44/3,2,0,38,79/6,1,
            0];
        I:=[671071,670933,670639,670976,670219,670098,670531,670522,670631,
            670686,670822,671043,671045,670113,670723,670451,670599,670608,
            670524,670592,670673,670709,670771,671044,670632,670705,670708,
            670724,670830,671072,669918,670336,670142,670154,670328,670299,
            670304,670308];
    elif hash1 eq 3405 and hash2 eq 2346 and hash3 eq 3039 then
        V:=[14,1,15,5,16,13,17,23,18,39,19,65,20,89,21,121,22,131,23,143,24,161,
            25,179,26,191];
        D:=[1,30,1,0,2,57/2,1,69/2,1,77/2,1,0,5,92/3,1,31,1,110/3,1,116/3,1,0,9,
            151/6,1,163/6,3,55/2,2,59/2,1,67/2,1,139/4,1,235/6,1,0,19,24,1,
            126/5,1,26,1,82/3,1,28,1,30,1,156/5,1,94/3,1,133/4,1,101/3,1,34,2,
            113/3,1,0,32,79/4,1,121/6,4,217/10,2,133/6,2,237/10,1,145/6,1,
            305/12,1,257/10,1,277/10,1,297/10,1,337/10,1,0,48,73/4,1,298/15,1,
            239/12,3,20,2,61/3,1,152/7,1,22,1,68/3,1,358/15,1,287/12,1,73/3,1,
            26,1,139/5,1,141/5,1,407/12,1,0,66,241/12,1,283/14,1,41/2,1,671/30,
            1,0,70,17,1,52/3,1,369/20,1,19,1,85/4,1,0,75,451/30,1,181/12,3,
            461/30,2,31/2,2,521/30,1,571/30,1,81/4,1,865/42,1,0,87,361/24,2,
            61/4,1,187/12,1,47/3,1,475/28,1,365/21,1,211/12,1,1243/60,1,0,96,
            377/28,1,817/60,2,1271/84,1,653/42,1,157/10,1,0,102,181/15,2,0];
        I:=[518129,518109,518103,518087,518091,518029,518026,518023,518095,
            517906,518099,518101,518032,518060,517934,518108,518118,518096,
            518021,517913,518022,518077,517997,517999,517887,517885,518104,
            518004,518011,518015,517823,518126,517982,518028,518030,518073,
            517870,517991,517942,517943,517718,517705,518046,517946,517899,
            517908,517955,518054,517979,518081,518083,518100,518066,518082,
            518089,517762,517944,518057,517897,518040,518123,517778,517803,
            517974,517788,517959,517716,518085,517871,517918,518010,517958,
            517707,518041,517957,517926,518049,518056,517661,517880,517653,
            518084,517961,517891,517665,517741,517655,517662,517656,517838,
            518009,517711,517713,517670,517646,517757,517760,517967,517749,
            517746,517747,517963,517964];
    elif hash1 eq 1525 and hash2 eq 3233 and hash3 eq 2752 then
        V:=[14,1,16,5,18,9,20,15,21,23,22,29,23,33,24,37,25,41];
        D:=[1,38,1,0,2,43,1,0,3,106/3,1,112/3,1,0,5,55/2,1,91/3,1,63/2,1,0,8,
            911/30,2,671/20,1,0,11,122/5,1,0,12,201/10,1,0,13,53/3,1,0,14,121/6,
            1,619/30,1,0];
        I:=[518120,517996,518005,518006,518019,518014,518018,517816,517839,
            517792,517925,517790,518012,517658,517657];
    elif hash1 eq 2982 and hash2 eq 3272 and hash3 eq 2105 then
        V:=[24,1,25,7,26,15,27,19,28,25,29,31];
        D:=[1,56/3,1,21,2,0,4,31/2,2,35/2,3,115/6,2,0,11,46/3,1,0,12,55/4,1,
            85/6,1,0,14,43/3,1,16,1,0,16,127/10,1,0];
        I:=[400049,399094,399678,400007,400526,399028,400459,400532,398898,
            400084,400140,399301,399513,400482,396623,398704];
    elif hash1 eq 2042 and hash2 eq 33 and hash3 eq 2497 then
        V:=[28,1,30,5,31,11,32,17,33,23,34,39];
        D:=[1,59/3,1,0,2,50/3,1,18,1,0,4,33/2,1,221/12,1,0,6,40/3,1,15,1,0,8,
            51/4,2,391/30,1,29/2,1,439/30,1,625/42,1,149/10,1,451/30,2,0,17,
            178/15,1,0];
        I:=[234072,232033,234115,233834,234139,230250,234135,233524,233920,
            230246,228823,230146,228973,228980,230248,230251,231925];
    elif hash1 eq 726 and hash2 eq 1604 and hash3 eq 2548 then
        V:=[22,1,24,5,25,11,26,19,27,23,28,29,30,35,31,45];
        D:=[1,31,1,0,2,117/4,1,30,1,0,4,237/10,1,253/10,1,103/4,1,0,7,239/12,1,
            0,8,551/30,1,125/6,1,0,10,277/15,1,94/5,1,0,12,269/20,1,203/15,1,
            163/12,1,61/4,1,0,16,119/10,1,0];
        I:=[247908,247803,247713,248236,247581,247966,248067,247504,247914,
            247612,247582,247818,247819,248071,247572,247535];
    elif hash1 eq 2230 and hash2 eq 3273 and hash3 eq 3056 then
        V:=[23,1,24,7,25,15,26,23,27,33,28,43,29,47];
        D:=[1,163/6,1,175/6,1,0,3,71/3,1,77/3,1,26,1,0,6,37/2,1,119/6,1,145/6,1,
            0,9,215/12,1,56/3,1,61/3,2,21,1,0,14,89/6,1,33/2,1,101/6,1,221/12,1,
            0,18,223/15,1,0,19,449/30,2,0];
        I:=[414679,414812,414805,414814,414802,414809,414290,414821,414434,
            414806,414659,414762,414820,414803,414671,413895,414813,414433,
            413649,413650];
    elif hash1 eq 2935 and hash2 eq 1104 and hash3 eq 3262 then
        V:=[28,1,29,5,30,13,31,23];
        D:=[1,53/3,1,0,2,95/6,2,97/6,3,107/6,1,0,8,167/12,1,14,1,44/3,1,15,1,0,
            12,51/4,3,77/6,1,79/6,1,0];
        I:=[228592,227971,228095,228041,228050,228511,226195,227406,228422,
            228288,227579,225676,225957,226955,227731,227136];
    elif hash1 eq 2183 and hash2 eq 1333 and hash3 eq 608 then
        V:=[14,1,16,5,18,13,20,27,22,43];
        D:=[1,26,1,0,2,23,1,25,3,39,1,0,7,85/4,1,22,1,93/4,2,70/3,2,24,3,77/3,1,
            0,17,91/5,1,19,2,61/3,1,62/3,1,21,1,43/2,1,73/3,1,0,25,223/12,1,0];
        I:=[518175,518176,518124,518130,518174,518078,518125,517988,518121,
            518122,518169,518173,518036,518048,518111,518092,517990,518042,
            518070,517808,518061,518044,518051,517806,517805];
    elif hash1 eq 2042 and hash2 eq 2487 and hash3 eq 1575 then
        V:=[26,1,27,9,28,21,29,31,30,35,31,45];
        D:=[1,69/4,1,96/5,1,106/5,1,0,4,83/6,1,185/12,1,157/10,1,95/6,3,209/12,
            1,0,11,43/3,1,44/3,1,232/15,1,91/5,1,0,15,77/6,1,0,16,193/15,1,
            155/12,1,76/5,1,46/3,1,0,20,219/20,1,341/30,1,0];
        I:=[494033,494224,493919,493886,494160,493992,493980,494193,494212,
            494209,493967,494230,494215,494259,494082,493966,494243,494032,
            494018,493985,494252];
    elif hash1 eq 538 and hash2 eq 2052 and hash3 eq 187 then
        V:=[20,1,21,9,22,19,23,39,24,49,26,61,27,65,28,69];
        D:=[1,21,1,25,1,29,1,0,4,39/2,1,127/6,3,91/4,1,47/2,1,0,10,77/4,1,59/3,
            2,64/3,2,65/3,1,22,1,93/4,2,70/3,1,71/3,1,26,1,0,22,33/2,1,209/12,1,
            109/6,2,37/2,1,0,27,191/12,1,49/3,2,35/2,1,55/3,1,39/2,1,0,33,
            199/12,2,0,35,181/12,1,0,36,178/15,2,0];
        I:=[420403,420083,419924,420339,420292,420326,420597,419956,420046,
            419456,419799,420569,419533,419578,420595,420333,419106,420368,
            420433,420298,419527,419798,420055,419730,420164,420405,419606,
            419643,420320,420291,420165,419550,419044,419160,419027,419652,
            419654];
    elif hash1 eq 3311 and hash2 eq 2123 and hash3 eq 1976 then
        V:=[19,1,20,5,21,13,22,21,23,27,24,39,25,45,26,49];
        D:=[1,65/2,1,0,2,27,4,29,1,33,1,0,8,139/6,2,151/6,3,51/2,1,0,14,64/3,3,
            71/3,2,0,19,197/10,1,121/6,2,87/4,1,131/6,2,305/12,1,0,26,268/15,1,
            61/3,1,0,28,491/30,2,0,30,15,1,0];
        I:=[252888,252997,252999,253025,253028,253026,253000,253032,253050,
            253024,253030,253049,253044,252982,253020,253033,253004,253072,
            252789,252779,253043,252577,252773,252811,252522,252801,252767,
            252797,252798,253018];
    elif hash1 eq 1525 and hash2 eq 3097 and hash3 eq 2509 then
        V:=[25,1,26,11,27,17,28,23,29,29,30,35];
        D:=[1,115/6,1,43/2,1,163/6,2,59/2,1,0,6,52/3,1,126/5,1,0,8,109/6,1,
            133/6,1,0,10,215/12,1,91/5,1,0,12,89/6,1,91/6,1,0,14,248/15,1,
            253/15,1,0];
        I:=[398717,399464,398991,399424,399577,399282,399363,398752,399315,
            398573,399894,398753,399340,395874,399354];
    elif hash1 eq 3217 and hash2 eq 1212 and hash3 eq 6 then
        V:=[25,1,26,5,27,9,28,21,29,25,30,39];
        D:=[1,139/6,1,0,2,85/4,2,0,4,71/4,1,107/6,2,119/6,1,257/12,1,217/10,1,0,
            10,136/7,1,0,11,491/30,1,33/2,1,67/4,1,101/6,1,245/12,1,43/2,1,0,17,
            218/15,1,0];
        I:=[397983,397357,397766,397383,397529,397732,397421,396580,397980,
            396182,395425,396409,397362,396423,395496,395291,397373];
    elif hash1 eq 585 and hash2 eq 3005 and hash3 eq 1615 then
        V:=[27,1,28,5,29,17,30,25,31,35,32,41];
        D:=[1,139/6,1,0,2,53/3,1,59/3,1,20,1,85/4,1,71/3,1,0,7,97/6,1,107/6,1,
            119/6,1,0,10,65/4,1,50/3,1,17,1,89/5,1,0,14,173/12,1,29/2,1,0,16,
            193/15,1,13,1,40/3,1,0];
        I:=[666408,666235,666290,666412,666832,665283,666709,665365,666351,
            665449,666847,666501,665435,666405,665379,666529,666313,666755];
    elif hash1 eq 2982 and hash2 eq 86 and hash3 eq 3228 then
        V:=[19,1,20,7,21,11,22,17,23,25,25,31,26,35,27,41,28,45];
        D:=[1,69/2,1,77/2,1,0,3,41,1,0,4,175/6,1,199/6,1,0,6,82/3,1,117/4,1,
            94/3,2,0,10,305/12,1,169/6,1,0,12,197/12,1,0,13,369/20,1,37/2,1,0,
            15,44/3,1,0,16,61/4,1,0];
        I:=[512649,512696,512380,512655,511884,512658,512571,512645,512688,
            512139,512138,512622,511444,511406,512290,511058];
    elif hash1 eq 2606 and hash2 eq 1036 and hash3 eq 3380 then
        V:=[19,1,20,7,21,15,22,25,23,39,24,55,25,69,27,75,28,85];
        D:=[1,163/6,1,115/4,1,0,3,71/3,1,76/3,1,109/4,1,0,6,237/10,2,155/6,1,
            273/10,1,179/6,1,0,11,89/4,1,287/12,1,382/15,1,105/4,1,80/3,1,103/3,
            1,0,17,551/30,1,187/10,1,611/30,1,335/14,1,289/12,1,51/2,1,207/8,1,
            0,24,386/21,1,307/15,1,21,1,464/21,1,283/12,1,25,1,0,30,445/24,1,
            293/14,1,0,32,939/70,1,569/42,1,421/30,2,303/20,1,0,37,604/45,1,
            485/36,1,847/60,1,0];
        I:=[542116,542170,542169,542167,542140,542120,542142,542104,542115,
            541840,542124,542148,542136,542103,541984,541981,542095,541967,
            541827,541979,541970,542082,541842,541915,542091,542184,541971,
            542090,541815,541838,541922,541927,541775,541788,541934,541926,
            541797,541785,541787];
    elif hash1 eq 2465 and hash2 eq 1572 and hash3 eq 2282 then
        V:=[22,1,23,7,24,13,25,23,26,43,27,51,28,57,29,69];
        D:=[1,22,1,70/3,1,0,3,95/6,1,277/10,1,0,5,328/15,1,111/5,1,67/3,1,80/3,
            1,0,9,173/12,2,29/2,1,89/6,1,91/6,1,551/30,1,37/2,1,125/6,1,127/6,1,
            659/30,1,0,19,218/15,1,86/5,1,77/4,1,0,22,451/30,1,461/30,1,0,24,
            787/60,4,79/6,2,53/4,1,14,1,907/60,1,0,33,1103/84,1,0];
        I:=[509004,508990,509844,509368,509276,508997,508996,508790,509827,
            509883,509907,509175,508987,509172,508993,508928,508920,509262,
            510108,509173,508880,509261,509371,509889,509898,510107,510109,
            509901,509963,509910,508777,508747,508694];
    elif hash1 eq 162 and hash2 eq 1768 and hash3 eq 2164 then
        V:=[28,1,29,13,30,29,31,33,32,39,33,45];
        D:=[1,59/3,3,85/4,1,64/3,1,65/3,1,93/4,1,0,8,97/6,1,209/12,1,177/10,1,
            107/6,2,109/6,1,197/10,1,79/4,2,0,17,81/5,1,0,18,173/12,1,479/30,1,
            0,20,13,1,40/3,4,0,25,117/10,1,0];
        I:=[665701,666173,666280,666652,666568,665568,666765,666522,666050,
            666172,665726,666524,664990,666790,666346,666392,665159,665728,
            665425,666523,666508,666656,666689,666844,664997];
    elif hash1 eq 3170 and hash2 eq 2144 and hash3 eq 2089 then
        V:=[29,1,30,9,31,17];
        D:=[1,91/6,1,31/2,1,103/6,2,0,5,46/3,2,47/3,2,16,1,0,10,137/10,3,55/4,1,
            83/6,2,85/6,4,157/10,1,0];
        I:=[219284,219364,219016,219018,219275,219276,218420,218450,217068,
            216928,217604,217805,217288,219366,219463,214797,215801,219047,
            219405,216479];
    elif hash1 eq 1243 and hash2 eq 660 and hash3 eq 2739 then
        V:=[25,1,26,9,27,17,28,25,29,31];
        D:=[1,35/2,2,39/2,1,127/6,1,0,5,53/3,2,18,2,85/4,1,0,10,29/2,1,63/4,1,
            95/6,2,0,14,43/3,2,191/12,1,0,17,91/6,1,0];
        I:=[241619,241751,241637,239763,239463,240972,241052,241744,239272,
            241062,241026,241054,241158,240687,241036,239774,238163];
    elif hash1 eq 2183 and hash2 eq 379 and hash3 eq 2941 then
        V:=[24,1,25,9,26,17,27,27,28,31,29,41];
        D:=[1,52/3,1,59/3,1,65/3,2,0,5,209/12,1,107/6,1,197/10,1,0,8,49/3,1,
            50/3,1,298/15,1,239/12,2,0,13,89/6,1,0,14,40/3,2,223/15,1,76/5,1,
            248/15,1,0,19,157/12,1,44/3,1,0];
        I:=[411576,412290,411031,412014,411092,411610,411511,412048,411970,
            411520,411318,411505,411147,411061,411223,411553,410922,410847,
            411253,408958];
    elif hash1 eq 1102 and hash2 eq 2283 and hash3 eq 679 then
        V:=[16,1,17,5,18,11,19,17,20,23,21,31,22,43,23,47,25,55];
        D:=[1,26,1,0,2,49/2,1,53/2,1,0,4,27,1,29,1,0,6,199/6,1,223/6,1,0,8,59/3,
            2,101/4,1,136/5,1,0,12,197/10,1,79/4,1,119/6,1,217/10,1,203/6,1,0,
            17,55/3,1,0,18,551/30,1,391/20,1,611/30,1,0,21,181/12,1,62/3,1,0];
        I:=[516463,516541,516483,516668,516261,516581,516081,516470,516613,
            516522,516155,516126,516515,516612,516105,515802,516227,515672,
            515805,515822,515639,515637];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 22 mod 47.
function cached_ids_22(P,hash1,hash2,hash3)
    if hash1 eq 2983 and hash2 eq 2259 and hash3 eq 1323 then
        V:=[28,1,29,11,31,19,32,31,33,37];
        D:=[1,65/3,1,71/3,4,26,1,117/4,1,0,8,217/10,1,131/6,1,145/6,1,0,11,29/2,
            1,103/6,1,37/2,1,187/10,1,113/6,1,0,16,175/12,1,253/15,1,0,18,
            401/30,1,0];
        I:=[229648,229356,229675,229738,229907,229093,229353,229985,229122,
            229096,229906,229735,229104,229987,229850,229228,229069,229235];
    elif hash1 eq 210 and hash2 eq 3011 and hash3 eq 2267 then
        V:=[22,1,23,5,24,9,25,19,26,23];
        D:=[1,35,1,0,2,151/6,2,0,4,59/3,1,71/3,1,76/3,1,26,2,0,9,133/6,1,0,10,
            18,1,55/3,1,56/3,4,89/4,1,0];
        I:=[247736,247886,247923,248254,248482,248024,247741,248070,247909,
            248475,248063,247767,247892,247940,247941,247628];
    elif hash1 eq 116 and hash2 eq 3307 and hash3 eq 2313 then
        V:=[29,1,31,5,32,11,33,19,34,27,35,31,36,39];
        D:=[1,671/30,1,0,2,17,1,751/30,1,0,4,317/21,1,233/15,1,256/15,1,0,7,
            913/60,1,157/10,1,319/20,1,0,10,313/20,1,0,11,995/84,1,73/6,1,40/3,
            1,0,14,133/15,1,9,1,0];
        I:=[535359,535355,535350,535370,535367,535360,535343,535347,535345,
            535346,535371,535342,535349,535344,535354];
    elif hash1 eq 868 and hash2 eq 2167 and hash3 eq 2489 then
        V:=[20,1,21,7,22,19,23,33,25,37,26,43,27,51,28,59,29,63];
        D:=[1,29,1,98/3,1,0,3,163/6,1,123/4,1,63/2,1,223/6,1,235/6,1,0,8,65/3,1,
            24,1,109/4,1,83/3,1,89/3,1,94/3,1,0,14,131/6,1,0,15,37/2,1,41/2,1,0,
            17,40/3,1,175/12,1,86/5,1,0,20,149/10,1,161/10,1,103/6,1,0,23,77/6,
            1,0,24,347/30,1,0];
        I:=[511943,512747,512092,511856,512629,511982,511944,512379,512617,
            512172,512638,512656,512171,511203,511213,511973,512825,512243,
            511974,512274,511068,511186,512824,512372];
    elif hash1 eq 1855 and hash2 eq 2920 and hash3 eq 3307 then
        V:=[20,1,21,5,22,9,23,15,24,23,25,31,26,35,27,39];
        D:=[1,35,1,0,2,187/6,1,0,3,94/3,1,141/4,1,0,5,329/12,1,179/6,1,181/6,1,
            0,8,358/15,1,412/15,1,55/2,1,0,11,471/20,1,0,12,118/5,1,0,13,339/20,
            1,521/30,1,241/12,1,0];
        I:=[509397,509019,509394,509023,509003,508984,509012,508956,508947,
            509383,509229,509350,508506,509008,508527];
    elif hash1 eq 2983 and hash2 eq 1258 and hash3 eq 1122 then
        V:=[26,1,27,7,28,13,29,23,30,31];
        D:=[1,85/4,1,65/3,2,0,4,33/2,2,121/6,2,0,8,49/3,2,50/3,1,18,1,56/3,1,0,
            13,29/2,1,89/6,2,167/10,1,0,17,40/3,2,179/12,1,0];
        I:=[237193,237156,237229,237158,237222,237226,237534,237172,237333,
            237419,237152,236950,237325,237109,237355,237405,237329,237339,
            236858];
    elif hash1 eq 3124 and hash2 eq 1081 and hash3 eq 2 then
        V:=[27,1,28,9,29,19,30,23,31,37,32,47,33,57];
        D:=[1,143/6,1,145/6,1,155/6,1,0,4,328/15,1,263/12,1,358/15,1,311/12,1,0,
            8,399/20,1,0,9,175/12,1,44/3,1,52/3,1,223/12,1,283/15,1,19,1,0,15,
            137/10,1,83/6,1,59/4,1,205/12,1,0,19,203/15,3,14,1,77/5,1,233/15,1,
            0,25,571/42,1,817/60,1,0];
        I:=[405053,404678,404553,404699,405349,404698,405056,404849,405073,
            404708,405099,405213,405096,404983,405257,404679,405330,405077,
            405014,405074,405217,405033,405271,404723,404728,405153];
    elif hash1 eq 2043 and hash2 eq 1473 and hash3 eq 3141 then
        V:=[24,1,25,11,26,17,27,23,28,29];
        D:=[1,80/3,2,27,1,29,2,92/3,1,0,7,127/6,2,139/6,1,0,10,59/3,1,64/3,1,0,
            12,157/10,2,109/6,1,0,15,49/3,2,91/5,1,55/3,1,56/3,1,0];
        I:=[243973,245001,245284,243196,245130,243334,243355,245229,243202,
            244999,242330,242983,243261,245135,242543,243180,242402,242417,
            242432];
    elif hash1 eq 2325 and hash2 eq 1922 and hash3 eq 7 then
        V:=[25,1,26,7,27,13,28,19,29,27,30,39];
        D:=[1,103/6,1,35/2,1,0,3,52/3,1,53/3,2,0,6,95/6,1,97/6,3,0,10,57/4,1,
            81/5,2,49/3,1,0,14,79/6,1,169/12,1,431/30,1,147/10,1,89/6,1,0,19,
            155/12,1,0];
        I:=[245825,245776,245519,243064,245795,245801,245244,245246,245376,
            243317,242405,244387,243326,242900,244214,243065,243343,244253,
            244290];
    elif hash1 eq 2184 and hash2 eq 3328 and hash3 eq 3184 then
        V:=[20,1,21,13,22,25,23,37,24,49,25,55,26,65,27,73,28,77];
        D:=[1,25,1,80/3,1,31,1,98/3,1,37,1,0,6,39/2,2,43/2,1,47/2,1,51/2,2,55/2,
            1,0,13,47/3,1,18,1,64/3,1,65/3,1,70/3,1,0,18,29/2,1,97/6,1,233/12,1,
            197/10,1,119/6,1,0,23,49/3,1,239/12,1,0,25,79/6,1,319/20,1,193/12,1,
            113/6,1,0,29,193/15,5,40/3,1,179/12,1,0,36,625/42,1,0,37,57/5,1,0];
        I:=[540154,540100,540082,540081,540107,540106,540163,540076,540103,
            540083,540105,540074,540098,540123,540027,540088,540132,540151,
            540075,540143,540039,540125,540097,540110,540026,540092,540093,
            540061,540024,540028,540096,540134,540135,540089,540042,539867,
            540041];
    elif hash1 eq 2090 and hash2 eq 2537 and hash3 eq 2270 then
        V:=[14,1,15,13,16,19,17,27,18,35,19,43,20,49,22,53,24,57];
        D:=[1,38,1,40,1,42,1,44,1,46,1,0,6,73/2,1,85/2,1,0,8,98/3,1,33,1,43,2,0,
            12,175/6,1,199/6,2,211/6,1,0,16,101/4,1,106/3,1,107/3,2,0,20,103/4,
            1,167/6,1,0,22,67/3,1,0,23,23,1,0,24,187/12,1,0];
        I:=[254859,254875,254860,254877,254864,254858,254873,254869,254874,
            254809,254822,254871,254818,254868,254816,254811,254788,254791,
            254819,254820,254824,254825,254777,254787];
    elif hash1 eq 2231 and hash2 eq 810 and hash3 eq 1611 then
        V:=[20,1,21,7,22,15,23,27,24,37,25,41,26,47];
        D:=[1,23,1,80/3,1,0,3,115/6,1,39/2,1,127/6,2,0,7,52/3,1,59/3,1,85/4,1,
            64/3,2,22,1,0,13,33/2,1,233/12,1,197/10,1,121/6,1,0,17,39/2,1,0,18,
            91/6,1,101/6,1,0,20,61/4,1,0];
        I:=[510758,510380,510519,510296,510280,510718,510517,509849,510243,
            510431,510575,510569,509588,510437,509337,510100,510103,509497,
            509408,508853];
    elif hash1 eq 2654 and hash2 eq 235 and hash3 eq 2793 then
        V:=[21,1,22,5,23,17,24,27,25,39,26,43,27,49];
        D:=[1,45/2,2,0,3,62/3,1,21,4,23,3,74/3,1,25,1,0,13,115/6,3,39/2,1,127/6,
            1,43/2,1,0,19,69/4,1,52/3,4,53/3,1,59/3,3,20,2,0,30,33/2,1,0,31,
            238/15,1,191/12,2,0,34,431/30,1,0];
        I:=[251047,251111,251071,250994,251073,251103,251112,250996,251173,
            251187,251181,250979,250502,250696,251067,251218,251133,250992,
            251196,251052,251062,251065,251199,251177,250513,251180,251241,
            250565,250984,250509,250615,249913,250418,250685];
    elif hash1 eq 1996 and hash2 eq 2766 and hash3 eq 2725 then
        V:=[8,1,9,13,10,23,11,31,12,45,13,61,14,71,15,81,16,101,17,117,18,121,
            19,129,20,137,21,141,22,151];
        D:=[1,40,1,46,1,48,1,52,1,56,2,0,7,81/2,2,85/2,1,93/2,1,113/2,1,0,12,
            104/3,1,122/3,1,146/3,1,0,15,175/6,1,123/4,1,187/6,1,199/6,1,247/6,
            1,83/2,1,0,21,117/4,1,101/3,1,36,1,112/3,1,119/3,1,124/3,2,44,1,0,
            29,169/6,1,181/6,1,437/10,1,265/6,1,0,33,311/12,1,478/15,1,431/12,1,
            109/3,1,0,37,45/2,1,227/10,1,137/6,1,337/12,1,851/30,1,57/2,1,
            719/20,1,1091/30,1,437/12,1,0,46,313/15,1,338/15,2,24,1,49/2,1,
            319/12,1,569/20,1,57/2,1,0,54,1237/60,1,0,55,263/15,1,1363/60,1,
            93/4,1,0,58,1591/70,1,977/42,1,233/10,1,0,61,967/60,1,0,62,2193/140,
            1,3377/210,1,1355/84,1,977/60,1,0,66,63/5,1,0];
        I:=[544339,544343,544342,544340,544341,544345,544338,544344,544336,
            544346,544333,544325,544328,544331,544330,544314,544304,544327,
            544334,544317,544319,544316,544307,544326,544308,544301,544313,
            544305,544311,544302,544294,544295,544318,544298,544306,544315,
            544329,544277,544280,544271,544299,544270,544284,544285,544282,
            544276,544278,544279,544310,544312,544273,544275,544272,544286,
            544300,544290,544274,544266,544267,544265,544281,544263,544262,
            544259,544260,544288];
    elif hash1 eq 1244 and hash2 eq 895 and hash3 eq 418 then
        V:=[19,1,20,5,21,11,22,15,23,21,24,27,25,31,26,37,28,43];
        D:=[1,49/2,1,0,2,29,1,116/3,1,0,4,51/2,1,0,5,65/3,1,101/4,1,0,7,217/10,
            1,103/4,1,0,9,239/12,1,0,10,91/6,1,399/20,1,0,12,248/15,1,253/15,2,
            0,15,76/5,1,0];
        I:=[512896,512844,512678,512682,512874,511956,511950,511979,511909,
            512499,512464,511886,511780,511819,511385];
    elif hash1 eq 2701 and hash2 eq 1243 and hash3 eq 1151 then
        V:=[22,1,23,5,24,11,25,25,26,29,27,39];
        D:=[1,21,1,0,2,103/6,1,175/6,1,0,4,47/3,1,53/3,1,18,1,58/3,1,59/3,1,20,
            1,0,10,209/12,1,0,11,44/3,1,81/5,1,49/3,1,263/12,1,0,15,173/12,1,
            29/2,1,0];
        I:=[418799,418769,415921,418787,418788,418778,418282,418116,418780,
            416948,418202,417149,418156,414994,417549,418169];
    elif hash1 eq 3406 and hash2 eq 434 and hash3 eq 2773 then
        V:=[28,1,29,5,30,17,31,21,32,31,33,37];
        D:=[1,58/3,2,0,3,97/6,2,33/2,2,209/12,2,71/4,1,109/6,4,0,14,49/3,1,0,15,
            173/12,1,91/6,1,193/12,1,197/12,1,0,19,13,1,40/3,1,0,21,71/6,2,0];
        I:=[233018,234309,233475,234613,234314,234621,232167,233905,231815,
            234273,234298,234299,234377,235137,233841,230574,233855,229298,
            233553,235131,232065,234034];
    elif hash1 eq 2325 and hash2 eq 1811 and hash3 eq 2072 then
        V:=[22,1,23,7,24,25,25,31,26,41,28,45];
        D:=[1,86/3,1,29,1,0,3,47/2,1,151/6,1,51/2,1,163/6,1,55/2,1,175/6,3,
            187/6,1,199/6,1,0,13,136/5,1,82/3,1,0,15,87/4,1,131/6,1,133/6,1,
            305/12,1,0,19,61/3,2,0,21,17,1,0];
        I:=[416335,416587,416352,417813,417819,416331,417420,417806,417815,
            417851,416573,417641,416567,417321,417838,417699,417479,417814,
            416541,416574,416300];
    elif hash1 eq 2513 and hash2 eq 1279 and hash3 eq 912 then
        V:=[23,1,24,7,25,19,26,25,28,29,29,33,30,43,31,47];
        D:=[1,199/6,2,67/2,1,0,4,136/5,1,83/3,2,28,1,117/4,1,89/3,2,0,11,95/4,1,
            143/6,2,0,14,61/3,1,0,15,17,1,0,16,291/20,1,451/30,1,91/6,1,31/2,1,
            0,20,883/60,1,0,21,677/60,1,0];
        I:=[250247,250707,251279,250221,250123,251136,251278,250321,251515,
            251526,250325,250348,251312,250346,250340,249858,250324,250925,
            251315,249822,250359];
    elif hash1 eq 2231 and hash2 eq 1124 and hash3 eq 318 then
        V:=[25,1,26,5,27,13,28,17];
        D:=[1,41/2,2,0,3,17,2,56/3,2,19,2,0,9,103/6,3,0,12,46/3,2,47/3,3,0];
        I:=[61273,61285,61120,61243,61139,61209,61168,61186,61086,61092,61164,
            61161,61184,60944,60949,61128];
    elif hash1 eq 2701 and hash2 eq 290 and hash3 eq 2851 then
        V:=[18,1,19,5,20,13,21,27,23,33,24,37,25,43,26,47];
        D:=[1,27,1,0,2,99/4,1,199/6,1,67/2,1,0,5,109/4,1,82/3,1,83/3,1,88/3,1,
            94/3,1,95/3,1,0,11,111/4,1,167/6,1,0,13,22,1,0,14,337/15,1,114/5,1,
            0,16,91/6,2,0,18,31/2,1,1123/60,1,0];
        I:=[517590,517465,517588,517591,517486,517564,517584,517573,517574,
            517157,517261,517547,517355,517049,517064,517586,517587,517037,
            516979];
    elif hash1 eq 2043 and hash2 eq 1461 and hash3 eq 3131 then
        V:=[28,1,29,7,30,17,31,21,32,25,33,29];
        D:=[1,71/3,1,77/3,2,0,4,119/6,2,121/6,1,217/10,1,143/6,1,0,9,61/3,1,0,
            10,491/30,2,0,12,175/12,1,0,13,27/2,2,439/30,1,44/3,1,0];
        I:=[667753,666759,667619,666873,667556,667666,666803,667687,665453,
            666426,666660,666443,665475,667671,665070,665049];
    elif hash1 eq 2795 and hash2 eq 964 and hash3 eq 214 then
        V:=[12,1,13,5,14,9,15,19,16,27,17,49,18,69,19,91,20,111,21,133,22,147,
            23,173,24,195,25,209];
        D:=[1,44,1,0,2,77/2,1,0,3,37,1,41,2,128/3,1,47,1,0,8,199/6,1,211/6,2,
            247/6,3,0,14,89/3,1,166/5,1,100/3,1,101/3,2,141/4,1,106/3,1,107/3,1,
            113/3,1,125/3,1,130/3,1,0,25,257/10,1,103/4,1,167/6,1,377/12,1,
            401/12,4,337/10,1,203/6,2,377/10,1,227/6,1,0,38,111/5,1,67/3,1,
            121/5,1,418/15,2,85/3,1,448/15,1,91/3,1,31,1,508/15,1,431/12,1,0,49,
            311/14,1,519/20,1,313/12,1,839/30,1,337/12,1,287/10,1,173/6,1,187/6,
            2,34,1,0,59,283/15,1,302/15,1,313/15,1,338/15,1,271/12,1,343/15,2,
            23,1,373/15,1,26,1,521/20,2,0,71,793/42,2,571/30,1,41/2,1,411/20,1,
            631/30,1,641/30,1,0,78,907/60,1,76/5,1,61/4,1,233/15,1,187/12,2,
            47/3,1,263/15,1,708/35,1,203/10,1,579/28,1,68/3,1,23,1,0,91,365/24,
            2,917/60,1,185/12,2,1439/84,1,489/28,1,737/42,1,1057/60,1,851/42,1,
            203/10,1,623/30,1,0,103,1145/84,1,273/20,1,1313/84,1,110/7,1,238/15,
            1,191/12,1,0,109,367/30,3,0];
        I:=[519198,519215,519221,519224,519232,519219,519241,519240,519103,
            519207,519208,519209,519233,519108,519082,519226,519210,519238,
            519162,519231,519234,519236,519144,519131,519032,519161,519235,
            519164,519211,519214,519225,519227,519119,519230,519237,519086,
            519118,519059,519147,518899,519033,519095,519242,519092,519136,
            518860,519097,519117,518891,519054,518953,518947,519102,519093,
            519125,518861,518864,518977,519139,518956,519067,519115,519116,
            519057,519135,519064,519056,518984,518976,518978,518869,518890,
            518927,518985,518975,519112,519071,519072,519077,519127,518856,
            519156,519158,519229,519120,518894,518853,518910,518925,518928,
            518857,518867,518851,519228,519243,518919,518923,518917,518902,
            518866,518859,518862,518898,518892,518904,518920,518908,518931,
            519078,519079,519080];
    elif hash1 eq 2842 and hash2 eq 3245 and hash3 eq 2209 then
        V:=[25,1,26,5,28,9,29,17,30,25,31,35];
        D:=[1,217/10,1,0,2,215/12,1,0,3,53/3,1,369/20,1,223/12,1,0,6,27/2,1,
            83/6,1,241/14,1,0,9,77/6,1,66/5,1,67/5,1,41/3,1,0,13,499/42,1,73/6,
            1,939/70,2,0];
        I:=[503012,503665,503446,502216,501557,503353,503476,501883,503496,
            503234,503026,503437,501916,501576,501893,502050];
    elif hash1 eq 2701 and hash2 eq 2500 and hash3 eq 1012 then
        V:=[22,1,23,7,24,13,25,19,26,27,27,35,30,45];
        D:=[1,29,1,31,1,0,3,151/6,2,175/6,2,0,7,93/4,1,71/3,1,0,9,107/6,1,119/6,
            1,121/6,1,0,12,73/4,1,55/3,1,56/3,1,0,15,197/12,1,33/2,1,113/6,2,
            115/6,1,0,20,173/15,1,0];
        I:=[248852,248851,247795,248903,247950,248781,247739,248937,248485,
            248555,248870,247760,247899,247919,247747,248550,247835,247929,
            247830,248463];
    elif hash1 eq 2184 and hash2 eq 264 and hash3 eq 2501 then
        V:=[27,1,28,5,29,9,30,15,31,19,32,29,33,35];
        D:=[1,121/6,1,0,2,287/12,1,0,3,241/12,1,207/10,2,0,6,37/2,1,0,7,79/6,1,
            27/2,1,181/12,1,709/42,1,0,11,40/3,1,14,1,0,13,677/60,1,485/42,1,
            47/4,1,71/6,1,145/12,1,0];
        I:=[407142,404479,404460,405287,407293,404906,407813,407298,404914,
            405109,407321,407920,407279,404892,404539,404940,404933];
    elif hash1 eq 1620 and hash2 eq 732 and hash3 eq 2981 then
        V:=[20,1,21,11,22,17,23,27,24,33,25,37,26,45];
        D:=[1,19,1,23,1,25,1,98/3,2,0,6,127/6,1,211/6,1,0,8,69/4,1,85/4,1,126/5,
            1,77/3,1,0,12,209/12,1,107/6,1,0,14,49/3,1,0,15,351/20,1,37/2,1,
            187/10,1,0,18,88/5,1,0];
        I:=[512878,512287,512650,512639,512671,512284,512765,512717,511928,
            511675,511965,512276,511425,511916,512166,511065,511391,511555];
    elif hash1 eq 2043 and hash2 eq 518 and hash3 eq 41 then
        V:=[23,1,24,7,25,19,27,31,29,39];
        D:=[1,59/2,1,211/6,1,0,3,76/3,1,77/3,1,82/3,1,88/3,1,89/3,1,0,8,133/6,1,
            169/6,1,297/10,1,181/6,1,61/2,1,0,13,221/12,1,37/2,2,127/6,1,0,17,
            521/30,1,35/2,2,107/6,2,0];
        I:=[415175,415172,415497,415182,415478,415399,415484,415480,415229,
            415495,415503,415488,415493,415135,415390,415231,415502,415393,
            415489,415394,415395];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 23 mod 47.
function cached_ids_23(P,hash1,hash2,hash3)
    if hash1 eq 822 and hash2 eq 1298 and hash3 eq 1849 then
        V:=[20,1,21,7,22,13,23,17,24,27,26,35,27,41];
        D:=[1,68/3,1,86/3,1,0,3,151/6,4,51/2,1,0,8,65/3,1,0,9,97/6,1,79/4,1,
            217/10,1,145/6,1,0,13,81/5,1,73/4,1,55/3,1,0,16,223/15,3,15,2,0,21,
            625/42,1,0];
        I:=[424757,424287,424290,424565,424587,424599,424558,424218,424756,
            424184,423413,424191,423168,424257,424211,423171,423172,423354,
            423405,424709,422411];
    elif hash1 eq 2937 and hash2 eq 2274 and hash3 eq 1460 then
        V:=[27,1,28,5,29,11,30,21,31,29];
        D:=[1,127/6,2,0,3,47/3,2,59/3,1,0,6,63/4,1,33/2,1,71/4,1,109/6,1,0,10,
            71/5,1,57/4,1,43/3,1,0,13,25/2,2,89/6,1,0];
        I:=[226345,228499,228091,228431,227329,228529,228497,225478,226628,
            227601,226890,228507,228068,228151,224699];
    elif hash1 eq 70 and hash2 eq 2688 and hash3 eq 1290 then
        V:=[21,1,22,5,23,11,24,17,25,29,26,39,27,47];
        D:=[1,47/2,1,0,2,71/3,1,24,1,0,4,37/2,1,119/6,1,0,6,73/4,1,55/3,1,56/3,
            1,61/3,1,109/5,1,0,11,33/2,2,101/6,4,103/6,1,551/30,1,0,19,15,2,
            253/15,2,203/12,1,0,24,27/2,1,185/12,1,0];
        I:=[421099,421102,421045,421105,421106,421053,421024,421055,421093,
            420955,421059,421061,421025,421082,421083,421095,421046,421001,
            420988,421005,420994,421000,420946,420997,420935];
    elif hash1 eq 2843 and hash2 eq 2160 and hash3 eq 915 then
        V:=[16,1,17,9,18,13,19,19,20,33,21,37,23,41,24,45,25,49];
        D:=[1,98/3,1,104/3,1,35,1,0,4,115/4,2,0,6,100/3,1,141/4,2,0,9,107/6,1,
            377/12,1,317/10,1,191/6,1,425/12,1,473/12,1,0,15,220/7,1,0,16,33/2,
            1,0,17,689/30,1,0,18,46/3,1,0,19,47/3,1,337/18,1,1133/60,1,0];
        I:=[518151,518149,518132,518025,518033,518142,517903,518007,518156,
            518027,517818,518161,517962,517725,517723,518141,517742,518137,
            517744,517613,517614];
    elif hash1 eq 2514 and hash2 eq 3443 and hash3 eq 1166 then
        V:=[28,1,29,11,30,23];
        D:=[1,277/15,1,223/12,2,307/15,1,45/2,1,0,6,509/30,1,781/42,1,19,1,
            619/30,1,83/4,1,0,11,61/4,1,50/3,2,352/21,1,84/5,1,103/6,1,0];
        I:=[507833,507829,507851,507818,507862,507855,507819,507838,507751,
            507877,507812,507857,507881,507899,507823,507754];
    elif hash1 eq 869 and hash2 eq 2933 and hash3 eq 1569 then
        V:=[20,1,22,7,23,17,24,23,25,29,26,33,27,37,28,41,29,45];
        D:=[1,92/3,1,31,1,0,3,24,1,82/3,1,117/4,1,88/3,1,0,7,233/10,1,143/6,1,0,
            9,298/15,1,239/12,1,0,11,101/6,1,0,12,76/5,1,0,13,625/42,1,0,14,
            53/4,1,0,15,817/60,1,0];
        I:=[512609,512567,512780,511670,511726,511718,511822,511736,511704,
            511395,511725,510894,510901,511688,510830];
    elif hash1 eq 728 and hash2 eq 2995 and hash3 eq 1556 then
        V:=[27,1,28,7,29,17,30,25,31,31];
        D:=[1,109/6,1,133/6,1,0,3,215/12,1,55/3,2,99/5,1,20,1,0,8,197/12,1,
            103/6,1,245/12,1,0,11,223/15,1,199/12,1,0,13,157/12,1,79/6,2,161/12,
            2,451/30,1,0];
        I:=[239963,240017,239225,238901,240307,239307,239198,239644,240377,
            239331,240477,238142,239385,239334,239998,238138,239310,238013];
    elif hash1 eq 2890 and hash2 eq 2888 and hash3 eq 3082 then
        V:=[24,1,25,7,26,11,27,17,28,23,29,29,30,35,31,41];
        D:=[1,82/3,1,100/3,1,0,3,169/6,1,0,4,263/12,1,73/3,1,0,6,207/10,1,125/6,
            2,0,9,223/15,1,179/12,2,0,12,59/4,1,91/6,1,0,14,13,1,227/15,1,0,16,
            817/60,1,0];
        I:=[415894,416199,415763,417380,415744,415785,416237,416243,416401,
            417399,417472,417389,415855,415040,415525,414999];
    elif hash1 eq 2843 and hash2 eq 2299 and hash3 eq 702 then
        V:=[22,1,23,5,24,11,25,15,26,21,27,29,28,33,29,39,30,49,31,53];
        D:=[1,94/3,1,0,2,145/6,1,377/12,1,0,4,358/15,2,0,6,791/30,1,161/6,1,0,8,
            313/15,1,106/5,1,373/15,1,0,11,115/6,1,0,12,66/5,1,310/21,1,0,14,
            361/30,1,145/12,1,569/42,1,137/10,1,0,18,707/60,1,0,19,217/18,1,0];
        I:=[509858,509880,509868,509703,509869,508688,508622,509854,509853,
            508689,508680,508415,508370,510106,509962,508492,509947,510112,
            508302];
    elif hash1 eq 1809 and hash2 eq 2791 and hash3 eq 610 then
        V:=[22,1,23,7,24,15,25,25,26,31,27,39,28,45,29,55,30,69];
        D:=[1,64/3,1,164/5,1,0,3,95/6,1,109/6,1,53/2,1,0,6,57/4,1,138/7,1,
            287/12,1,25,1,0,10,75/4,1,113/6,1,0,12,289/20,3,29/2,1,86/5,1,0,17,
            391/30,1,157/12,1,0,19,40/3,1,208/15,1,43/3,2,233/15,1,0,24,145/12,
            1,911/70,2,1103/84,1,793/60,2,797/60,1,40/3,2,0,33,407/35,1,254/21,
            1,0];
        I:=[540085,540254,540168,540159,540070,540248,539971,540244,540062,
            539915,540040,540169,540193,540243,540197,540164,539826,540246,
            540047,540153,540043,540060,540225,539954,539869,539986,539937,
            539830,539865,540023,540022,540046,539699,539703];
    elif hash1 eq 1480 and hash2 eq 3392 and hash3 eq 2350 then
        V:=[18,1,19,11,20,27,21,43,22,65,23,79,24,99,25,111,26,123,27,129];
        D:=[1,29,1,98/3,1,35,1,116/3,1,0,5,127/6,2,43/2,1,51/2,2,123/4,1,199/6,
            1,71/2,1,235/6,1,0,14,59/3,1,85/4,1,65/3,1,71/3,1,126/5,1,83/3,1,
            94/3,1,0,21,109/6,2,37/2,1,79/4,1,119/6,2,87/4,1,131/6,3,95/4,1,
            143/6,1,257/10,1,157/6,2,0,36,268/15,1,215/12,1,18,3,239/12,2,
            287/12,1,79/3,1,0,45,197/12,1,101/6,1,18,1,217/12,1,551/30,1,599/30,
            1,241/12,2,611/30,1,125/6,1,0,55,361/20,1,380/21,1,272/15,2,96/5,2,
            464/21,1,0,62,137/10,1,15,1,181/10,1,1093/60,2,73/4,2,0,69,203/15,1,
            75/4,1,0,71,239/20,1,0];
        I:=[515090,515545,515078,515063,515497,515550,515210,515092,515266,
            514682,515156,514719,514419,515492,515268,515030,515546,514778,
            515057,514267,515060,515142,515187,514381,515484,515554,514658,
            514284,514382,515121,514285,514879,514888,514420,514625,514570,
            515155,515324,515475,515508,514386,514393,514167,514121,514675,
            514369,514825,515226,514518,514607,514392,514685,514512,514370,
            514421,514110,514010,514687,514266,514300,513988,514824,514384,
            514247,514295,514299,513979,514293,514554,513952,514462];
    elif hash1 eq 2843 and hash2 eq 2419 and hash3 eq 2804 then
        V:=[26,1,27,9,28,19,29,29,30,33];
        D:=[1,52/3,1,53/3,1,59/3,1,0,4,209/12,1,107/6,1,233/12,1,79/4,1,0,8,
            43/3,1,44/3,1,49/3,1,50/3,1,0,12,29/2,1,0,13,66/5,1,40/3,2,41/3,3,
            0];
        I:=[239598,240153,239719,240090,240145,239637,240453,240364,239714,
            240159,240149,239751,239721,238261,239494,239728,240412,240483];
    elif hash1 eq 70 and hash2 eq 1468 and hash3 eq 2661 then
        V:=[22,1,25,5,26,13,27,17,28,31,29,41];
        D:=[1,76/3,1,0,2,245/12,1,28,1,389/12,1,0,5,271/12,1,0,6,559/30,1,
            569/30,1,619/30,1,62/3,1,21,1,679/30,1,0,12,17,1,56/3,1,113/5,1,
            93/4,1,0,16,641/42,1,67/4,1,52/3,1,0];
        I:=[539970,539924,539903,539879,539890,539966,539887,539957,539770,
            539894,539771,539900,539902,540356,540357,539973,539872,539772];
    elif hash1 eq 2514 and hash2 eq 1293 and hash3 eq 379 then
        V:=[18,1,19,7,20,15,21,27,22,45,23,55,24,67,25,71,27,95,28,99];
        D:=[1,104/3,1,43,1,0,3,67/2,1,223/6,1,235/6,1,0,6,101/4,1,156/5,1,94/3,
            1,34,1,112/3,1,0,11,87/4,1,257/10,1,157/6,1,329/12,1,167/6,1,353/12,
            1,181/6,1,193/6,1,0,19,311/12,1,131/5,1,79/3,1,80/3,1,0,23,221/12,1,
            187/10,1,719/30,1,519/20,1,317/12,1,0,28,441/20,1,0,29,44/3,1,59/4,
            1,205/12,1,559/30,1,56/3,1,75/4,1,19,1,581/30,1,233/12,1,39/2,2,
            221/10,1,0,41,75/4,1,0,42,233/20,1,707/60,2,0];
        I:=[515359,515125,514415,514531,514591,514800,514934,514536,515054,
            515221,514614,514928,514585,514832,515201,514802,514541,514749,
            514912,514904,515128,514922,514833,514482,514031,514044,514096,
            514131,514821,514804,514613,514136,514045,514104,514269,514915,
            514480,514694,514752,513992,513955,514747,514552,514784];
    elif hash1 eq 117 and hash2 eq 3276 and hash3 eq 167 then
        V:=[16,1,17,5,18,9,19,15,20,19,21,23,22,33,23,39,24,45,25,49,26,53,27,
            57];
        D:=[1,32,1,0,2,61/2,1,0,3,27,1,86/3,1,0,5,175/6,1,0,6,109/4,1,0,7,155/6,
            1,329/12,1,169/6,1,205/6,1,0,11,382/15,1,131/5,1,0,13,127/6,1,
            511/20,1,0,15,117/7,1,0,16,43/2,1,0,17,401/21,1,0,18,937/60,1,0];
        I:=[542986,543091,542903,542939,543093,542910,542950,542926,542897,
            542958,542900,542968,542699,542971,542834,542742,542780,542733];
    elif hash1 eq 3219 and hash2 eq 2477 and hash3 eq 2031 then
        V:=[22,1,23,9,24,13,25,19,26,27,27,33,28,43,29,51];
        D:=[1,86/3,1,29,1,31,1,0,4,187/6,1,0,5,58/3,1,83/3,2,0,8,33/2,1,237/10,
            1,143/6,2,0,12,191/12,1,49/3,1,0,14,89/6,2,167/10,2,187/10,1,113/6,
            1,0,20,218/15,1,175/12,1,44/3,1,0,23,51/4,1,0];
        I:=[411741,411740,411834,411623,410493,410319,410925,410967,410100,
            410108,410109,410400,411734,410008,410065,409973,409974,411326,
            409792,410007,410006,411966,409778];
    elif hash1 eq 1762 and hash2 eq 2160 and hash3 eq 1885 then
        V:=[23,1,24,5,25,13,26,19,27,29,29,35];
        D:=[1,99/4,1,0,2,65/3,3,70/3,1,101/4,2,0,8,87/4,1,131/6,1,0,10,91/5,1,
            20,1,81/4,2,352/15,1,0,15,491/30,3,125/6,1,0,19,401/30,1,0];
        I:=[415883,415711,415827,415853,415806,415718,415822,415897,415760,
            415887,415673,415817,415892,415774,415204,415308,415866,415299,
            415212];
    elif hash1 eq 1574 and hash2 eq 2164 and hash3 eq 962 then
        V:=[21,1,22,13,23,21,24,41,25,57,26,79,28,91];
        D:=[1,47/2,1,151/6,1,163/6,1,187/6,1,211/6,1,0,6,76/3,1,26,1,83/3,1,0,9,
            121/6,1,131/6,2,133/6,1,145/6,2,49/2,1,305/12,1,155/6,2,157/6,2,
            53/2,1,0,22,62/3,1,328/15,1,22,1,111/5,1,67/3,1,358/15,1,74/3,1,0,
            29,101/6,1,551/30,1,37/2,2,113/6,3,20,1,245/12,2,125/6,2,127/6,1,
            223/10,1,227/10,1,0,44,17,1,52/3,1,56/3,1,227/12,2,19,1,0,50,269/20,
            1,203/15,2,208/15,2,14,1,187/12,2,0];
        I:=[421332,421721,421666,421669,421428,421413,421408,421750,421651,
            421701,421741,421727,421661,421677,421650,421645,421685,421687,
            421330,421401,421485,421704,421433,421748,421414,421728,421681,
            421369,421699,421429,421671,421674,421361,421375,421653,421477,
            421557,421752,421486,421723,421478,421252,421230,421744,421731,
            421673,421594,421652,421675,421170,421695,421700,421405,421664,
            421402,421158,421183];
    elif hash1 eq 3125 and hash2 eq 2033 and hash3 eq 1589 then
        V:=[17,1,18,5,19,17,21,23,23,31,24,37];
        D:=[1,69/2,1,0,2,62/3,1,68/3,2,74/3,1,86/3,1,39,1,0,8,127/6,1,43/2,1,0,
            10,109/6,2,237/10,1,191/6,1,0,14,197/12,1,289/12,1,0,16,223/15,1,
            56/3,1,0];
        I:=[516752,516724,516497,516703,515996,516555,516071,516757,516753,
            516513,516641,515810,515724,515961,515636,516154,515664];
    elif hash1 eq 3407 and hash2 eq 334 and hash3 eq 2932 then
        V:=[25,1,26,7,27,13,28,19,29,29,30,45,31,57,32,61];
        D:=[1,107/6,1,181/6,1,0,3,268/15,1,328/15,1,0,5,101/6,1,115/6,1,0,7,
            223/15,1,253/15,1,283/15,1,21,1,0,11,79/6,1,83/6,1,613/42,1,299/20,
            1,781/42,1,75/4,1,569/30,1,0,18,178/15,1,143/12,2,66/5,1,41/3,2,
            208/15,1,0,25,697/60,2,0,27,107/9,1,0];
        I:=[502993,503172,503358,502891,502878,502996,503053,502904,501674,
            502396,501316,502867,501266,502509,501648,502004,501948,503137,
            502970,502991,502984,501687,503035,502918,503003,503013,501068];
    elif hash1 eq 446 and hash2 eq 2102 and hash3 eq 2980 then
        V:=[24,1,25,5,26,9,27,17,28,25,29,29,30,37,31,47];
        D:=[1,94/3,1,0,2,277/10,1,0,3,55/3,1,328/15,1,89/4,1,0,6,197/12,1,33/2,
            1,207/10,1,0,9,223/15,3,0,12,451/30,1,91/6,1,997/60,1,0,15,787/60,2,
            66/5,1,53/4,1,203/15,1,0,20,347/30,1,0];
        I:=[418737,416267,418303,416644,416708,417833,418690,416266,415373,
            416678,417507,416198,416570,415360,415538,417796,417526,417832,
            416571,415381];
    elif hash1 eq 493 and hash2 eq 1087 and hash3 eq 1074 then
        V:=[24,1,25,9,26,17,27,25,28,33,29,49];
        D:=[1,18,1,93/4,1,109/4,1,0,4,37/2,1,45/2,1,237/10,1,0,7,49/3,1,73/4,1,
            56/3,1,0,10,89/6,1,491/30,2,167/10,1,0,14,155/12,1,40/3,2,44/3,1,
            223/15,1,179/12,1,76/5,1,199/12,3,0,24,391/30,1,157/12,1,213/14,1,
            0];
        I:=[508257,508184,508205,508224,508209,508146,508246,508258,508212,
            508053,508125,508145,508115,508182,508219,508220,508244,508118,
            507722,508178,507934,507950,508013,508056,508117,507822];
    elif hash1 eq 1104 and hash2 eq 3020 and hash3 eq 48 then
        V:=[22,1,23,5,24,17,25,25,26,31,27,37];
        D:=[1,21,1,0,2,115/6,1,127/6,1,43/2,1,91/4,1,47/2,1,0,7,53/3,3,59/3,1,
            20,1,0,12,33/2,1,107/6,2,0,15,44/3,1,50/3,1,0,17,173/12,1,29/2,2,
            89/6,1,0];
        I:=[248948,249348,249405,249201,249119,248651,248523,249002,249471,
            248161,249169,249172,248339,249360,248190,248628,248947,249306,
            249435,248624];
    elif hash1 eq 3407 and hash2 eq 1359 and hash3 eq 2549 then
        V:=[19,1,20,5,21,9,22,27,23,37,24,49,25,53,26,59,28,67];
        D:=[1,49/2,1,0,2,80/3,2,0,4,139/6,1,47/2,1,151/6,1,51/2,1,175/6,1,187/6,
            1,63/2,1,223/6,1,0,12,116/5,2,71/3,2,77/3,1,32,1,0,18,131/6,2,
            237/10,1,143/6,2,145/6,2,155/6,1,0,26,74/3,1,0,27,491/30,1,245/12,2,
            0,30,248/15,1,361/20,1,332/15,1,0,33,203/15,2,0];
        I:=[419585,419427,419823,419277,419204,419416,419190,419707,419746,
            419326,419586,419208,419460,419205,419506,419270,419587,419265,
            419486,419185,419191,419386,419216,419469,419068,419155,419499,
            419262,419308,419288,419103,419022,418971,419218];
    elif hash1 eq 1997 and hash2 eq 3420 and hash3 eq 291 then
        V:=[24,1,25,5,26,13,27,25,28,41];
        D:=[1,119/5,1,0,2,241/12,1,45/2,1,149/6,1,0,5,247/12,1,251/12,1,22,1,
            275/12,1,25,1,0,10,559/30,1,56/3,1,379/20,1,19,1,233/12,1,865/42,1,
            1237/60,1,0,17,563/28,1,0];
        I:=[513401,513389,513364,513347,513428,513339,513337,513381,513320,
            513362,513358,513344,513361,513338,513407,513346,513356];
    elif hash1 eq 2185 and hash2 eq 2922 and hash3 eq 372 then
        V:=[21,1,22,9,23,17,24,29,25,33,26,39,27,43];
        D:=[1,115/6,1,199/6,1,67/2,1,0,4,59/3,1,126/5,1,82/3,1,0,7,95/6,1,109/6,
            1,193/10,1,233/12,2,281/12,1,0,13,91/5,1,0,14,89/6,1,197/12,1,0,16,
            179/12,1,0,17,185/12,1,0];
        I:=[510757,510170,510465,510694,510169,509302,510690,510717,510167,
            509723,510292,508836,509941,510478,510410,509538,508863];
    elif hash1 eq 258 and hash2 eq 2274 and hash3 eq 3315 then
        V:=[18,1,19,7,20,15,21,27,22,43,23,57,24,75,25,87];
        D:=[1,86/3,1,29,1,0,3,51/2,1,163/6,1,187/6,2,0,7,70/3,1,76/3,1,77/3,1,
            26,1,117/4,1,0,12,119/6,1,121/6,1,217/10,1,131/6,1,133/6,1,145/6,3,
            157/6,1,0,21,56/3,1,19,1,239/12,2,61/3,1,328/15,1,263/12,1,0,28,
            67/4,1,18,1,217/12,1,221/12,1,37/2,1,113/6,2,115/6,2,20,1,0,38,
            248/15,1,199/12,1,109/6,3,223/12,1,283/15,2,0,46,15,1,1093/60,1,0];
        I:=[427513,427077,427529,427531,427299,427479,427530,427212,427434,
            426624,426787,427101,427154,426785,427144,427295,426879,427277,
            427297,426841,427076,427301,426577,427124,427068,426491,426784,
            426643,427204,427057,427298,426565,426762,427069,426895,427288,
            426811,426797,426760,426466,426611,426761,426449,426556,426653,
            426613,426501];
    elif hash1 eq 1151 and hash2 eq 411 and hash3 eq 3024 then
        V:=[23,1,24,13,25,19,26,23,27,29,28,41];
        D:=[1,127/6,1,163/6,1,175/6,1,59/2,1,187/6,2,0,7,71/3,2,26,2,0,11,109/6,
            1,0,12,238/15,1,328/15,1,0,14,551/30,1,221/12,1,113/6,3,115/6,3,22,
            1,0,23,223/12,1,0];
        I:=[418006,417888,418610,415977,416774,418140,416554,418603,416235,
            418015,416965,417323,415391,415818,416009,415676,415695,416789,
            415871,416233,416969,415026,414998];
    elif hash1 eq 1762 and hash2 eq 2549 and hash3 eq 2245 then
        V:=[29,1,30,5,31,9,32,17,33,29];
        D:=[1,103/6,1,0,2,46/3,1,0,3,55/4,2,85/6,3,157/10,1,0,9,37/3,2,208/15,1,
            57/4,1,16,1,81/5,1,0,15,25/2,1,0];
        I:=[366989,366994,366939,366956,366326,367038,367072,366663,366472,
            366473,366671,364753,364580,363784,363300];
    elif hash1 eq 1433 and hash2 eq 1411 and hash3 eq 158 then
        V:=[30,1,31,11,32,21,34,29,35,37];
        D:=[1,77/4,1,58/3,1,85/4,1,64/3,1,0,5,71/4,1,107/6,1,109/6,1,197/10,1,0,
            9,81/5,1,49/3,2,50/3,1,0,13,151/12,1,155/12,1,13,1,0,16,341/30,1,0];
        I:=[58871,59019,58680,58976,58856,58974,58750,58714,59007,58701,58965,
            59010,59128,58809,58823,58650];
    elif hash1 eq 634 and hash2 eq 954 and hash3 eq 313 then
        V:=[14,1,15,7,16,13,17,17,18,21,19,31,20,35,21,41,22,45,24,51,26,57];
        D:=[1,38,1,42,1,0,3,81/2,1,105/2,1,0,5,128/3,1,0,6,235/6,1,0,7,83/3,1,
            36,2,118/3,1,165/4,1,0,12,377/12,1,0,13,181/5,1,145/4,1,0,15,519/20,
            1,0,16,409/20,1,481/20,1,0,18,47/3,1,53/3,1,0,20,1594/105,2,110/7,1,
            79/5,1,0];
        I:=[518706,518702,518683,518693,518652,518698,518703,518694,518695,
            518687,518637,518700,518538,518528,518537,518569,518366,518696,
            518364,518225,518234,518233,518238];
    elif hash1 eq 3031 and hash2 eq 3115 and hash3 eq 2872 then
        V:=[21,1,22,7,23,19,24,29,25,37];
        D:=[1,45/2,1,49/2,1,0,3,62/3,1,21,1,68/3,2,23,4,25,2,0,13,39/2,1,127/6,
            5,43/2,2,47/2,1,0,22,53/3,1,59/3,2,64/3,1,0,26,177/10,2,107/6,4,
            109/6,4,37/2,2,0];
        I:=[674131,674129,674130,674121,674036,674120,673998,674016,674101,
            674111,673890,674056,673895,673782,673935,673956,674004,674127,
            673836,673877,673822,674122,673921,674018,673829,673784,673835,
            673827,673901,673903,673939,673811,673823,673824,673922,673844,
            673908];
    elif hash1 eq 258 and hash2 eq 808 and hash3 eq 654 then
        V:=[20,1,21,15,22,19,23,23,24,27];
        D:=[1,24,1,28,1,30,2,32,1,34,1,36,1,0,8,45/2,1,0,9,21,1,0,10,115/6,2,0,
            12,69/4,1,52/3,2,58/3,1,0];
        I:=[506661,506627,506628,506708,506601,506666,506663,506680,506564,
            506676,506799,506624,506696,506762,506113];
    elif hash1 eq 1292 and hash2 eq 2126 and hash3 eq 1391 then
        V:=[21,1,22,7,23,15,24,27,25,39];
        D:=[1,45/2,1,53/2,2,0,4,19,7,21,3,23,1,0,15,103/6,1,115/6,1,39/2,1,
            127/6,1,43/2,2,0,21,69/4,2,52/3,2,53/3,4,18,1,59/3,1,0,31,97/6,1,0];
        I:=[249814,248994,249768,249572,249611,249617,249629,249645,249810,
            249811,248862,249585,249756,248885,249525,249741,249646,249429,
            247750,249032,248720,249001,249096,249740,249007,249116,249287,
            249635,247975,247587,248783];
    elif hash1 eq 3078 and hash2 eq 2017 and hash3 eq 1625 then
        V:=[23,1,24,5,25,9,26,17,27,27,28,41,29,51,30,61];
        D:=[1,187/6,1,0,2,28,1,0,3,95/4,1,145/6,1,305/12,1,0,6,215/12,1,111/5,1,
            68/3,1,76/3,1,0,10,33/2,1,101/6,1,539/30,1,217/12,1,221/12,2,41/2,2,
            0,18,17,1,69/4,2,53/3,1,251/12,1,0,23,439/30,1,451/30,2,331/20,1,
            697/42,1,0,28,66/5,2,163/12,2,361/24,1,226/15,1,0];
        I:=[416309,417482,415890,416312,416686,416799,416665,416666,415344,
            416826,415683,416718,417520,416359,417018,414978,417470,416030,
            416033,416952,417805,416635,416897,416914,416956,417338,415354,
            416672,416827,415562,417798,414981,414979];
    elif hash1 eq 3313 and hash2 eq 3253 and hash3 eq 1850 then
        V:=[29,1,31,7];
        D:=[1,31/2,1,39/2,1,0,3,83/6,4,85/6,7,29/2,1,63/4,1,0];
        I:=[210393,210087,210413,210683,210692,210791,209488,210374,210406,
            210532,210535,210685,210701,210372,208709];
    elif hash1 eq 3125 and hash2 eq 2794 and hash3 eq 211 then
        V:=[17,1,18,7,19,17,20,23,21,33,23,41];
        D:=[1,49/2,1,73/2,1,0,3,25,1,29,1,92/3,1,31,1,0,7,163/6,2,55/2,1,0,10,
            53/3,1,71/3,1,24,3,77/3,1,0,16,121/6,1,217/10,1,133/6,1,0,19,33/2,1,
            551/30,1,113/6,2,0];
        I:=[427587,427584,427560,427535,427555,427546,427232,427574,427567,
            427573,427551,427533,427547,427548,427260,426560,426927,426958,
            427563,426561,426554,426562];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 24 mod 47.
function cached_ids_24(P,hash1,hash2,hash3)
    if hash1 eq 2703 and hash2 eq 2256 and hash3 eq 380 then
        V:=[21,1,22,9,23,17,24,21,25,29,26,35,27,39];
        D:=[1,53/2,1,57/2,1,65/2,1,0,4,25,1,80/3,1,29,1,0,7,151/6,1,0,8,71/3,1,
            76/3,1,26,1,0,11,257/12,1,237/10,1,0,13,101/5,1,0,14,103/6,2,431/20,
            1,0];
        I:=[410719,410744,411416,410749,410761,411539,410498,411411,410407,
            411310,411150,410893,410887,409525,409526,409438];
    elif hash1 eq 3032 and hash2 eq 2542 and hash3 eq 2253 then
        V:=[22,1,23,5,24,13,25,19,26,29,28,33,29,37];
        D:=[1,86/3,1,0,2,51/2,2,107/4,1,59/2,1,0,6,76/3,1,77/3,2,0,9,217/10,1,
            131/6,2,133/6,2,237/10,1,0,15,55/3,1,0,16,253/15,1,0,17,181/12,1,
            31/2,1,0];
        I:=[248066,247758,247860,248540,247768,248080,247762,247976,247849,
            248069,248102,247756,247861,248002,247946,247578,247555,247554];
    elif hash1 eq 729 and hash2 eq 1206 and hash3 eq 1416 then
        V:=[24,1,25,11,26,17,27,27,28,37,29,49];
        D:=[1,65/3,1,22,1,70/3,1,71/3,1,0,5,119/6,2,41/2,2,0,9,55/3,2,298/15,1,
            101/5,1,61/3,1,0,14,167/10,1,101/6,3,103/6,1,221/12,3,0,22,175/12,1,
            179/12,3,15,4,46/3,2,199/12,1,0,33,79/6,1,181/12,1,0];
        I:=[247404,247023,247478,247305,247364,247410,247019,247038,247142,
            247360,246809,246584,246587,246864,246737,246946,247146,247403,
            246494,246495,246902,246913,246779,246973,247191,247314,247371,
            247406,247407,246999,247232,246533,246918,246469];
    elif hash1 eq 588 and hash2 eq 2739 and hash3 eq 302 then
        V:=[22,1,23,5,24,11,25,23,26,27,28,33];
        D:=[1,92/3,1,0,2,163/6,1,175/6,1,0,4,85/4,1,64/3,1,65/3,3,126/5,1,77/3,
            1,0,11,197/10,1,0,12,191/12,2,215/12,1,0,15,66/5,1,0];
        I:=[251042,250307,249905,250297,250875,251005,251413,251477,249904,
            249910,250602,250175,250240,250293,250463];
    elif hash1 eq 2844 and hash2 eq 2613 and hash3 eq 1368 then
        V:=[8,1,9,5,10,9,11,17,14,25,16,35,17,47,18,53,19,57,20,61,22,69,23,79];
        D:=[1,58,1,0,2,117/2,1,0,3,152/3,1,158/3,1,182/3,1,0,6,163/4,1,307/6,1,
            211/4,1,0,9,191/5,1,598/15,1,479/12,1,527/12,1,0,13,193/6,1,493/15,
            1,217/6,1,842/21,1,241/6,1,0,18,1597/60,1,901/28,1,0,20,747/28,1,0,
            21,143/6,1,0,22,1984/105,1,301/15,1,349/15,1,0,25,667/42,1,170/9,1,
            856/45,1,381/20,1,0,29,247/20,1,0];
        I:=[544337,544323,544324,544335,544322,544321,544332,544303,544292,
            544296,544309,544297,544320,544291,544287,544269,544257,544283,
            544258,544256,544254,544264,544293,544268,544261,544253,544255,
            544252,544289];
    elif hash1 eq 1669 and hash2 eq 726 and hash3 eq 2104 then
        V:=[18,1,19,7,20,19,21,37,22,53,23,63,24,71,26,77];
        D:=[1,28,1,30,1,0,3,49/2,1,57/2,1,65/2,1,69/2,3,77/2,1,0,10,25,2,80/3,1,
            27,1,86/3,2,29,1,31,6,104/3,1,35,1,0,25,115/6,1,139/6,3,47/2,1,
            151/6,2,51/2,2,163/6,4,175/6,1,0,39,85/4,1,64/3,3,65/3,1,70/3,2,0,
            46,209/12,1,119/6,2,121/6,1,0,50,215/12,1,55/3,2,0,53,289/20,1,46/3,
            1,0];
        I:=[253879,253877,253878,253932,253955,253683,253697,253894,253700,
            253753,253933,253869,253687,253695,253867,253765,253694,253759,
            253775,253929,253941,253975,253584,253735,253764,253693,253696,
            253876,253692,253701,253928,253685,253716,253586,253666,253690,
            253731,253585,253734,253671,253957,253964,253592,253583,253732,
            253736,253568,253783,253758,253720,253714,253870,253524,253506];
    elif hash1 eq 3314 and hash2 eq 2344 and hash3 eq 438 then
        V:=[28,1,29,7,30,15,31,23];
        D:=[1,47/3,1,85/4,1,0,3,29/2,2,33/2,1,119/6,1,0,7,14,1,44/3,2,81/5,1,0,
            11,127/10,1,77/6,1,79/6,3,29/2,1,0];
        I:=[394841,392878,394598,394779,393646,392956,394844,394701,394748,
            393110,394194,393856,392043,393937,394781,391586];
    elif hash1 eq 1293 and hash2 eq 2784 and hash3 eq 1501 then
        V:=[22,1,23,11,24,15,25,23,26,27,27,35,29,41];
        D:=[1,21,1,25,1,80/3,1,29,1,0,5,139/6,1,0,6,58/3,1,65/3,2,71/3,1,0,10,
            37/2,1,0,11,81/5,1,50/3,1,20,1,0,14,431/30,1,491/30,1,0,16,129/10,2,
            0];
        I:=[410685,410634,410246,410875,410266,410604,410236,410609,410233,
            410177,410169,410182,410602,410466,410976,410468,411056];
    elif hash1 eq 353 and hash2 eq 851 and hash3 eq 165 then
        V:=[19,1,20,5,21,15,22,27,23,41,24,49];
        D:=[1,53/2,5,0,6,21,2,23,2,74/3,1,25,1,0,12,115/6,1,127/6,1,43/2,2,47/2,
            1,151/6,1,0,18,18,1,59/3,3,20,4,65/3,1,22,1,77/3,1,0,29,107/6,1,
            109/6,1,131/6,1,0,32,49/3,3,0];
        I:=[252416,252461,252478,252479,252482,252433,252483,252411,252440,
            252269,252265,252420,252066,252288,252399,252139,252081,252430,
            252334,252340,252434,252109,252404,252480,252481,252218,252000,
            251868,252248,252135,251900,252402,252414,252427];
    elif hash1 eq 1246 and hash2 eq 3215 and hash3 eq 1836 then
        V:=[12,1,13,11,14,19,15,31,16,43,17,53,18,61,19,69,20,77,21,83,23,89];
        D:=[1,36,2,40,1,42,1,48,1,0,6,53/2,2,73/2,2,77/2,1,0,11,74/3,2,25,1,29,
            1,110/3,1,122/3,1,0,17,99/4,1,151/6,1,163/6,1,55/2,1,87/2,1,0,22,
            65/3,1,101/4,1,76/3,1,107/3,1,0,26,121/6,1,143/4,1,217/6,1,0,29,
            239/12,1,67/3,1,141/5,1,0,32,31/2,1,399/20,1,241/12,1,0,35,253/15,2,
            271/12,1,0,38,679/30,1,139/6,1,0,40,193/12,1,0];
        I:=[519654,519655,519610,519650,519618,519626,519652,519643,519648,
            519625,519615,519621,519653,519544,519613,519562,519577,519622,
            519565,519573,519498,519614,519578,519552,519555,519554,519443,
            519438,519579,519435,519388,519580,519442,519383,519475,519491,
            519384,519349,519351,519345];
    elif hash1 eq 259 and hash2 eq 2166 and hash3 eq 3203 then
        V:=[28,1,29,9,30,13,31,19,32,29,33,35];
        D:=[1,53/3,1,96/5,1,20,1,0,4,95/6,2,0,6,167/12,1,44/3,1,0,8,77/6,1,
            419/30,1,431/30,1,173/12,1,0,12,151/12,1,13,1,0,14,331/30,2,0];
        I:=[232551,232808,231874,231934,234635,232627,232520,232570,229674,
            233264,232563,232807,230717,232228,232587];
    elif hash1 eq 2891 and hash2 eq 1387 and hash3 eq 936 then
        V:=[12,1,14,7,16,17,18,25,20,41,22,45];
        D:=[1,32,1,40,1,0,3,98/3,1,37,1,39,1,43,1,0,7,136/5,1,28,1,196/5,1,0,10,
            101/5,1,81/4,2,61/3,1,67/3,1,166/7,1,73/3,1,34,1,0,18,428/21,1,0,19,
            123/8,1,187/12,1,83/4,1,0];
        I:=[543998,543995,544018,544004,544003,543976,543989,544009,543973,
            543969,543999,544008,544002,543980,543947,543936,543956,543939,
            543895,543903,543900];
    elif hash1 eq 1293 and hash2 eq 3129 and hash3 eq 1809 then
        V:=[18,1,19,11,20,21,21,29,22,39,23,55,24,59,25,71,26,77,27,81];
        D:=[1,92/3,1,31,1,33,3,37,1,0,7,163/6,3,55/2,1,175/6,3,199/6,1,0,15,
            83/3,1,28,1,88/3,2,0,19,257/12,1,237/10,1,143/6,1,103/4,1,0,23,55/3,
            1,61/3,1,62/3,2,328/15,1,89/4,2,67/3,3,68/3,1,0,34,125/6,1,0,35,
            61/4,1,46/3,1,253/15,1,17,1,337/15,1,0,40,181/12,3,91/6,1,0,44,
            323/21,1,0,45,361/30,1,0];
        I:=[426158,426121,426072,426127,426157,426166,426115,426137,426170,
            426138,426156,426193,426196,426122,426149,426146,425976,425992,
            426120,425931,425887,425877,425904,425891,426147,426150,425871,
            425731,425912,425880,425882,425948,425972,425973,425995,426142,
            425886,425907,425640,426129,426130,426141,426144,425634,425964];
    elif hash1 eq 118 and hash2 eq 1291 and hash3 eq 3362 then
        V:=[26,1,27,7,28,13,29,25,30,29,31,33,32,37];
        D:=[1,65/3,1,24,1,0,3,107/6,1,133/6,1,0,5,50/3,2,18,1,55/3,2,56/3,2,19,
            1,0,13,113/6,1,0,14,199/12,1,0,15,71/6,1,0,16,229/20,1,173/15,3,0];
        I:=[405431,406578,405704,406571,405427,405453,405528,405829,406573,
            406594,406618,406620,406622,406607,406581,406570,406565,406593,
            406595];
    elif hash1 eq 2139 and hash2 eq 640 and hash3 eq 345 then
        V:=[24,1,25,5,26,13,27,17];
        D:=[1,19,2,0,3,103/6,6,35/2,1,115/6,1,0,11,69/4,2,0,13,29/2,1,185/12,1,
            63/4,1,95/6,2,97/6,1,209/12,1,0];
        I:=[245772,245833,244960,245755,245770,245773,245796,245815,245100,
            244677,242977,243051,242865,243142,243174,242550,245363,245145,
            243175];
    elif hash1 eq 1575 and hash2 eq 2869 and hash3 eq 2613 then
        V:=[21,1,22,17,23,33,24,41,25,49,26,61,27,77,28,87,29,93];
        D:=[1,175/6,1,187/6,1,63/2,2,199/6,1,71/2,1,223/6,1,75/2,1,0,9,85/4,2,
            65/3,1,136/5,1,156/5,1,94/3,1,32,1,101/3,1,0,17,41/2,1,131/6,1,
            133/6,1,0,20,61/3,1,62/3,1,119/5,1,0,23,33/2,1,167/10,1,37/2,1,
            611/30,2,41/2,1,0,29,289/20,1,175/12,1,223/15,1,179/12,1,18,1,
            380/21,1,223/12,1,0,36,149/10,1,461/30,1,499/30,1,103/6,1,0,40,
            787/60,1,79/6,1,0,42,697/60,1,0];
        I:=[510511,509704,510389,510411,510387,510600,509550,509781,509598,
            509599,509495,510057,510073,510242,509779,510435,510642,509349,
            510452,510653,509641,508861,509507,509458,509677,509449,509909,
            509510,509724,509589,509745,509576,508768,508664,508851,509787,
            509445,508626,509364,509690,509482,509525];
    elif hash1 eq 3173 and hash2 eq 287 and hash3 eq 2138 then
        V:=[19,1,20,7,21,15,22,23,23,29,24,35,25,39];
        D:=[1,45/2,1,53/2,1,0,3,23,1,25,1,33,2,0,7,115/6,1,99/4,1,163/6,1,0,10,
            59/3,4,106/5,1,0,15,177/10,1,121/6,1,0,17,49/3,1,0,18,491/30,1,0];
        I:=[420040,420050,419904,419189,419918,420019,420037,420118,419197,
            419481,420024,420052,420057,419298,419317,419296,419919,419381];
    elif hash1 eq 2703 and hash2 eq 2639 and hash3 eq 574 then
        V:=[15,1,17,5,18,13,19,23,20,33,21,41,22,55,23,65,24,73,26,77];
        D:=[1,81/2,1,0,2,199/6,1,235/6,1,91/2,1,0,5,101/3,1,112/3,1,113/3,1,
            119/3,1,0,9,179/6,1,191/6,1,193/6,1,227/6,1,0,13,359/12,2,30,1,97/3,
            1,0,17,731/30,3,313/12,1,267/10,1,851/30,1,57/2,1,30,1,0,25,102/5,1,
            338/15,1,368/15,1,157/6,1,0,29,571/30,1,451/20,1,679/30,1,0,32,
            113/5,1,0,33,211/15,2,71/5,2,947/60,1,95/6,1,0];
        I:=[428044,427871,427848,428149,428064,427873,427823,428082,427860,
            427829,427868,428147,427824,427877,427913,427832,427821,427834,
            427835,427984,427830,427819,427870,428126,428111,428086,427974,
            428123,428062,427711,427797,427695,427987,428105,427833,427839,
            427679,427630];
    elif hash1 eq 2938 and hash2 eq 2922 and hash3 eq 1073 then
        V:=[24,1,25,5,26,17,27,23,28,33,29,37];
        D:=[1,82/3,1,0,2,131/6,1,133/6,1,143/6,1,145/6,1,111/4,1,0,7,62/3,1,
            111/5,1,0,9,221/12,2,37/2,1,41/2,1,125/6,1,0,14,53/3,1,0,15,511/30,
            1,521/30,3,177/10,1,0];
        I:=[413619,413565,413574,413563,413620,413612,413610,413530,413545,
            413546,413584,413559,413554,413606,413551,413523,413550,413552,
            413537];
    elif hash1 eq 823 and hash2 eq 1328 and hash3 eq 3327 then
        V:=[22,1,23,5,24,13,25,21,26,31,27,45,28,51];
        D:=[1,23,2,0,3,151/6,1,55/2,2,175/6,2,0,8,64/3,1,71/3,2,28,2,0,13,37/2,
            1,119/6,1,121/6,1,145/6,1,0,17,91/5,1,73/4,1,55/3,1,56/3,1,101/5,1,
            62/3,1,0,23,101/6,1,221/12,1,0,25,44/3,2,50/3,1,0];
        I:=[418364,418463,418475,417611,418329,418469,418689,418317,417754,
            418595,418671,418695,418299,417759,418687,417840,416403,417803,
            418685,417758,417760,417856,417522,417797,418376,418686,416289];
    elif hash1 eq 2891 and hash2 eq 2655 and hash3 eq 2875 then
        V:=[25,1,26,9,27,17,28,29];
        D:=[1,35/2,1,115/6,1,39/2,4,0,7,53/3,1,18,1,77/4,1,0,10,63/4,1,95/6,2,
            97/6,10,33/2,1,109/6,1,0,25,44/3,2,0];
        I:=[670995,671078,670590,670596,670939,671086,670951,670550,670192,
            670758,670463,670546,670334,670440,670441,670519,670542,670637,
            670652,670734,670800,671026,670765,670362,670172,670320];
    elif hash1 eq 3314 and hash2 eq 408 and hash3 eq 1538 then
        V:=[22,1,23,17,24,33,25,49,26,55,27,65,28,81];
        D:=[1,86/3,1,29,2,92/3,1,31,1,33,1,104/3,1,35,1,0,9,99/4,1,151/6,2,
            163/6,3,55/2,2,175/6,1,59/2,2,187/6,1,0,21,53/3,1,65/3,1,71/3,1,24,
            2,126/5,1,76/3,3,77/3,3,0,33,107/6,1,119/6,1,0,35,191/12,1,49/3,2,
            215/12,1,61/3,1,0,40,319/20,1,16,1,101/6,1,37/2,1,187/10,1,75/4,1,
            113/6,2,0,48,281/20,2,0];
        I:=[251325,251463,251511,250756,251358,251524,250027,250924,250257,
            250883,251592,250242,250866,251559,250358,250729,249894,250320,
            250904,250841,251151,250882,250361,250674,251346,249895,250220,
            250573,250888,249893,250047,250059,250363,250865,250336,250179,
            250868,250060,250354,249942,249937,251364,250013,249919,249828,
            249917,250382,250217,250352];
    elif hash1 eq 118 and hash2 eq 1251 and hash3 eq 1374 then
        V:=[20,1,21,9,22,15,23,23,24,29,25,37,26,45,27,51,28,59,29,71,30,77];
        D:=[1,94/3,1,112/3,1,113/3,1,0,4,203/6,1,417/10,1,0,6,388/15,1,418/15,1,
            383/12,1,0,9,731/30,1,49/2,1,0,11,506/21,1,295/12,1,26,1,0,14,
            631/30,1,253/12,1,641/30,1,0,17,69/4,1,272/15,1,0,19,157/10,1,
            1199/70,1,1037/60,1,0,22,96/7,1,208/15,1,623/40,1,377/24,1,236/15,1,
            0,27,1019/84,1,61/5,1,0,29,437/36,1,0];
        I:=[514873,514700,514598,514981,514780,514972,514961,514689,514697,
            514906,514148,514147,514076,513973,514308,514513,514867,514055,
            514937,514054,514051,514091,514786,513962,513967,513960,514065,
            514083,513956];
    elif hash1 eq 1152 and hash2 eq 784 and hash3 eq 2045 then
        V:=[25,1,26,11,27,19,28,29,29,41,30,47,31,53];
        D:=[1,151/6,1,163/6,1,187/6,1,199/6,1,0,5,116/5,1,101/4,1,28,1,0,8,
            177/10,1,107/6,1,79/4,1,45/2,1,0,12,49/3,1,268/15,1,215/12,1,55/3,1,
            62/3,1,0,17,59/4,1,89/6,2,0,20,218/15,1,175/12,1,0,22,259/20,1,
            391/30,1,59/4,1,0];
        I:=[397914,398308,400316,400091,398307,399828,399492,398783,400056,
            398610,398460,400318,399875,399489,400092,398969,398472,397844,
            399703,397837,398568,395745,398041,395800];
    elif hash1 eq 1716 and hash2 eq 2137 and hash3 eq 630 then
        V:=[18,1,19,9,20,19,21,37,22,49,23,65,24,75];
        D:=[1,24,1,26,1,28,1,0,4,49/2,2,53/2,4,57/2,5,65/2,1,0,16,62/3,1,21,4,
            68/3,3,23,5,74/3,3,25,1,27,1,29,1,0,35,115/6,4,39/2,1,127/6,5,43/2,
            2,139/6,1,0,48,53/3,2,58/3,1,59/3,5,20,4,64/3,1,65/3,4,22,2,0,67,
            209/12,1,107/6,1,109/6,1,119/6,2,0,72,44/3,2,81/5,2,49/3,3,50/3,2,
            0];
        I:=[253431,253446,253449,253232,253420,253258,253321,253445,253447,
            253181,253229,253262,253319,253381,252886,253281,253267,253410,
            253444,253451,253291,253323,253386,253134,253268,253349,253395,
            253448,252947,253053,253102,253083,253139,252736,253273,253375,
            253382,253416,253376,253097,253116,253128,253387,253423,253068,
            253351,253126,253180,253261,253127,253081,253109,253155,253358,
            253408,252731,253103,253340,253450,252957,252574,253038,253112,
            253122,252742,253107,253141,253058,253098,252514,252573,252852,
            253085,252719,252741,252752,253192,253417,252659,252720];
    elif hash1 eq 1434 and hash2 eq 2351 and hash3 eq 2121 then
        V:=[21,1,22,7,23,11,24,17,25,27,26,35,27,45,29,53];
        D:=[1,151/6,1,175/6,1,0,3,77/3,2,0,5,131/6,2,143/6,1,0,8,20,1,61/3,1,
            62/3,2,22,2,0,14,217/12,1,221/12,1,37/2,2,0,18,37/2,1,278/15,1,
            223/12,1,283/15,2,0,23,44/3,1,181/12,2,31/2,2,0,28,47/4,1,0];
        I:=[253276,253263,253035,253285,252981,253235,252980,252916,253331,
            252625,253247,252841,252917,252651,252515,252685,253148,252500,
            252649,252826,252504,252830,252681,252646,252898,252762,253284,
            252716];
    elif hash1 eq 2891 and hash2 eq 1215 and hash3 eq 1284 then
        V:=[23,1,24,5,25,13,26,23,27,31,29,37];
        D:=[1,43/2,1,0,2,58/3,1,26,1,83/3,1,0,5,119/6,1,121/6,1,41/2,1,131/6,1,
            0,9,50/3,2,268/15,1,298/15,1,0,13,147/10,2,103/6,1,0,16,129/10,1,0];
        I:=[503168,502839,503181,503249,502821,503244,503252,502838,502840,
            503192,503167,502822,502823,503170,503158,502854];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 25 mod 47.
function cached_ids_25(P,hash1,hash2,hash3)
    if hash1 eq 2657 and hash2 eq 2190 and hash3 eq 180 then
        V:=[23,1,24,7,25,13,26,23,27,29];
        D:=[1,115/6,1,99/4,1,0,3,53/3,3,77/4,1,0,7,177/10,1,71/4,2,109/6,2,
            133/6,1,0,13,79/5,1,191/12,4,0,18,431/30,1,29/2,1,91/6,1,0];
        I:=[418857,418613,418533,418593,418841,418624,416598,418576,418657,
            418838,418865,417387,417591,417567,418082,418519,418662,416590,
            418446,418313];
    elif hash1 eq 2939 and hash2 eq 234 and hash3 eq 365 then
        V:=[26,1,27,9,28,15,29,21,30,25];
        D:=[1,62/3,1,21,1,68/3,1,0,4,115/6,2,39/2,4,0,10,47/3,1,53/3,3,0,14,
            97/6,1,0,15,43/3,1,44/3,1,0];
        I:=[61210,61205,61200,61108,61147,60942,60968,61107,61270,61193,61115,
            61122,61153,61113,60999,61136];
    elif hash1 eq 495 and hash2 eq 1766 and hash3 eq 1342 then
        V:=[23,1,24,5,25,19,26,31,27,45,29,57];
        D:=[1,47/2,1,0,2,20,1,65/3,1,22,1,71/3,1,101/4,1,77/3,1,0,8,41/2,1,
            257/12,1,217/10,1,133/6,1,143/6,1,0,13,18,1,55/3,4,239/12,1,20,1,
            61/3,1,109/5,1,0,22,101/6,1,103/6,1,221/12,1,37/2,2,113/6,1,0,28,
            401/30,2,137/10,1,83/6,3,185/12,1,0];
        I:=[246541,246608,246760,246679,246623,246552,246663,246799,246642,
            246689,246678,246611,246793,246569,246643,246681,246814,246785,
            246783,246765,246496,246795,246839,246645,246614,246787,246786,
            246823,246824,246674,246557,246655,246668,246464];
    elif hash1 eq 1012 and hash2 eq 943 and hash3 eq 2088 then
        V:=[12,1,13,19,14,33,15,49,16,73,17,103,18,125,19,143,20,163,21,185,22,
            193,23,201,24,213];
        D:=[1,28,1,32,1,34,1,36,2,38,1,40,1,42,1,46,1,0,10,53/2,1,57/2,1,61/2,2,
            65/2,1,73/2,1,89/2,1,0,17,25,1,92/3,1,98/3,1,33,1,35,1,110/3,2,
            122/3,1,0,25,151/6,3,51/2,1,107/4,1,175/6,1,59/2,2,187/6,1,67/2,1,
            71/2,1,223/6,2,75/2,1,79/2,1,0,40,22,1,93/4,1,71/3,1,24,1,101/4,1,
            76/3,1,77/3,2,26,1,89/3,1,156/5,1,95/3,1,100/3,1,141/4,1,107/3,2,0,
            56,131/6,2,133/6,2,45/2,1,281/12,1,145/6,1,277/10,1,167/6,1,169/6,1,
            353/12,1,357/10,1,0,68,101/5,1,328/15,1,263/12,1,67/3,2,358/15,1,
            97/4,2,79/3,1,113/4,1,0,78,37/2,1,187/10,1,113/6,1,611/30,1,245/12,
            1,41/2,1,125/6,1,289/12,1,49/2,1,0,87,46/3,4,52/3,1,37/2,1,278/15,2,
            409/20,1,41/2,1,441/20,1,337/15,1,271/12,1,23,1,0,101,35/2,1,75/4,1,
            949/42,1,0,104,309/20,1,238/15,1,191/12,1,0,107,1079/70,1,433/28,1,
            937/60,1,667/42,1,481/30,1,0,112,62/5,1,0];
        I:=[519315,519308,519309,519311,519314,519310,519319,519320,519318,
            519305,519253,519151,519307,519251,519316,519313,519312,519256,
            519252,519204,519317,519249,519264,519197,519192,519265,519266,
            519194,519199,519267,519193,519201,519268,519202,519191,519255,
            519257,519186,519175,519303,519183,519184,519306,519167,519185,
            519152,519254,519304,519180,519027,519034,519283,519222,519165,
            519171,519250,519262,519019,519190,519007,519166,519189,519046,
            518949,519181,519182,518990,519021,519011,519200,519014,519280,
            519052,519157,519172,518937,518951,519261,519008,519017,519031,
            518995,518946,519187,518948,518945,519163,519174,519178,519206,
            519047,519173,519013,519015,518941,518952,519001,519098,519000,
            518940,519188,518939,518885,518998,518997,518994,518881,518877,
            518875,518876,518874,519023];
    elif hash1 eq 3174 and hash2 eq 2387 and hash3 eq 809 then
        V:=[24,1,25,15,26,23,27,31];
        D:=[1,19,1,21,1,68/3,2,23,1,74/3,1,25,1,0,8,115/6,1,39/2,1,83/4,1,0,11,
            53/3,2,59/3,1,20,2,0,16,95/6,1,107/6,1,0];
        I:=[406019,406092,406121,406199,406041,405943,406020,406126,406327,
            405890,405459,406002,406326,406094,406184,405964,406087];
    elif hash1 eq 965 and hash2 eq 67 and hash3 eq 137 then
        V:=[14,1,16,5,18,11,20,15,22,25,26,37];
        D:=[1,34,1,0,2,29,1,92/3,1,0,4,117/4,1,0,5,79/3,1,412/15,1,91/3,1,104/3,
            1,0,9,131/7,1,163/8,1,506/21,1,307/12,1,211/8,1,0,14,122/9,1,71/5,1,
            0];
        I:=[543733,543699,543706,543746,543720,543705,543611,543644,543668,
            543620,543646,543694,543604,543588,543583];
    elif hash1 eq 2704 and hash2 eq 2018 and hash3 eq 1992 then
        V:=[22,1,23,9,24,19,26,31,27,39,28,47,30,51];
        D:=[1,74/3,1,86/3,1,29,1,0,4,151/6,1,163/6,1,187/6,1,211/6,1,0,8,93/4,1,
            71/3,1,76/3,1,136/5,2,83/3,1,0,14,81/5,1,49/3,1,55/3,1,0,17,25/2,1,
            551/30,1,187/10,1,0,20,223/15,1,0,21,667/60,1,0];
        I:=[502313,502303,502315,502416,502319,502311,502417,502302,502360,
            502312,502351,502354,502350,502406,502359,502401,502287,502421,
            502352,502361,502418];
    elif hash1 eq 1717 and hash2 eq 1955 and hash3 eq 3452 then
        V:=[16,1,17,11,18,21,19,31,20,43,21,49,22,53];
        D:=[1,34,1,36,2,38,1,42,1,0,6,57/2,1,65/2,3,73/2,1,85/2,1,0,12,80/3,1,
            29,1,98/3,2,33,1,0,17,99/4,1,151/6,2,175/6,1,59/2,1,211/6,1,0,23,
            77/3,2,83/3,1,0,26,131/6,1,0,27,67/3,1,0];
        I:=[254728,254716,254720,254682,254721,254726,254715,254717,254731,
            254725,254678,254730,254729,254681,254683,254672,254679,254670,
            254718,254675,254656,254650,254654,254735,254677,254676,254645];
    elif hash1 eq 965 and hash2 eq 331 and hash3 eq 2616 then
        V:=[14,1,15,13,16,23,17,29,18,37,19,41,20,47];
        D:=[1,30,1,32,3,34,2,36,2,38,1,0,10,57/2,2,61/2,1,65/2,1,69/2,3,0,17,29,
            3,35,1,0,21,151/6,1,163/6,2,55/2,1,0,25,83/3,1,0,26,131/6,1,133/6,1,
            0,28,67/3,1,0];
        I:=[62083,62070,62074,62084,62073,62079,62075,62077,62076,62071,62080,
            62082,62064,62063,62072,62081,62062,62068,62078,62065,62061,62059,
            62060,62069,62058,62067,62066,62057];
    elif hash1 eq 354 and hash2 eq 2061 and hash3 eq 2160 then
        V:=[15,1,16,5,17,13,18,21,19,25,20,31,21,37,22,45];
        D:=[1,69/2,1,0,2,86/3,1,110/3,1,37,1,0,5,151/6,1,187/6,2,63/2,1,0,9,
            71/3,2,0,11,143/6,2,257/10,1,0,14,287/12,1,32,1,0,16,20,1,24,1,
            289/12,1,0,19,145/6,1,0];
        I:=[428976,428977,428860,428847,428965,428865,428960,428831,428863,
            428974,428787,428850,428669,428912,428604,428786,428780,428690,
            428581];
    elif hash1 eq 1764 and hash2 eq 634 and hash3 eq 488 then
        V:=[30,1,31,5,32,13,33,23];
        D:=[1,81/5,1,0,2,79/6,2,29/2,2,167/10,1,0,7,151/12,1,193/15,4,13,2,40/3,
            4,0,18,67/6,1,0];
        I:=[226633,226505,226539,228272,228386,224598,228404,226678,226744,
            226836,226963,227161,228332,227172,228266,228369,228463,226973];
    elif hash1 eq 2234 and hash2 eq 1113 and hash3 eq 587 then
        V:=[18,1,19,9,20,31,21,51,22,75,23,105,24,129,25,153,26,179,27,191];
        D:=[1,25,1,29,2,98/3,1,0,5,139/6,2,151/6,1,51/2,1,163/6,1,55/2,1,59/2,1,
            123/4,1,187/6,1,199/6,1,67/2,2,0,17,65/3,2,70/3,1,71/3,2,76/3,1,
            77/3,2,136/5,1,83/3,1,28,1,89/3,2,0,30,233/12,1,119/6,2,121/6,1,
            41/2,1,131/6,1,281/12,1,237/10,2,95/4,1,143/6,3,49/2,1,257/10,1,0,
            45,215/12,1,18,1,55/3,2,56/3,1,298/15,1,20,1,101/5,1,61/3,1,21,1,
            328/15,2,67/3,1,121/5,1,129/5,1,311/12,1,0,61,101/6,1,103/6,2,
            217/12,1,551/30,1,37/2,1,113/6,1,611/30,1,245/12,1,41/2,1,207/10,1,
            83/4,2,0,74,179/12,2,15,2,76/5,1,46/3,2,329/20,1,50/3,1,253/15,2,17,
            1,52/3,2,283/15,2,422/21,1,0,91,59/4,3,451/30,2,181/12,1,91/6,2,
            185/12,1,31/2,3,709/42,1,339/20,1,205/12,1,241/14,1,209/12,1,
            1213/60,1,0,109,203/15,1,163/12,2,316/21,1,323/21,1,47/3,1,0,115,
            361/30,4,0];
        I:=[517475,517439,517441,517437,517386,517434,517435,517444,517467,
            517497,517494,517432,517398,517421,517485,517503,517267,517325,
            517407,517424,517450,517415,517443,517487,517356,517161,517517,
            517287,517483,517404,517374,517410,517423,517468,517357,517430,
            517227,517258,517449,517239,517307,517426,517511,517232,517401,
            517412,517390,517429,517236,517286,517388,517354,517391,517520,
            517306,517315,517480,517305,517095,517108,517381,517500,517513,
            517294,517234,517393,517067,517314,517469,517342,517301,517459,
            517460,517330,517445,517247,517382,517344,517268,517411,517327,
            517252,517293,517312,517498,517191,517280,517254,517295,517062,
            517375,517399,517400,516986,517257,517146,516987,517403,517201,
            517504,517505,517506,517046,517060,517014,517038,517018,516993,
            517337,517068,517222,517050,517056,517075,517335,517338,517360,
            517361];
    elif hash1 eq 1811 and hash2 eq 2492 and hash3 eq 1899 then
        V:=[20,1,21,23,22,29,23,37,24,51,25,55,26,65];
        D:=[1,52/3,1,59/3,1,104/5,1,85/4,1,65/3,2,88/3,1,94/3,1,166/5,1,101/3,2,
            106/3,2,0,14,119/6,1,87/4,1,0,16,215/12,1,164/7,1,121/5,1,0,19,
            197/12,1,101/6,1,551/30,1,599/30,2,22,1,719/30,1,0,26,15,1,0,27,
            697/42,1,217/12,1,1093/60,1,56/3,1,0,31,203/15,1,164/9,1,0];
        I:=[542493,542496,542240,542391,542499,542504,542046,542047,542340,
            542451,542455,542049,542491,542500,542265,542435,541886,542281,
            542269,542452,542299,542237,542248,541864,541865,542482,541879,
            542137,542051,541863,542333,541778];
    elif hash1 eq 2939 and hash2 eq 3080 and hash3 eq 45 then
        V:=[18,1,19,5,20,15,21,25,22,31,24,35,25,45,26,57];
        D:=[1,113/3,1,0,2,205/6,1,377/10,1,227/6,1,159/4,1,0,6,27,1,85/3,1,
            359/12,1,32,1,0,10,731/30,2,791/30,1,0,13,313/15,1,0,14,263/15,1,
            1147/60,1,293/15,1,235/12,1,0,18,467/30,1,157/10,1,163/10,1,327/20,
            1,1311/70,1,0,23,229/14,1,0];
        I:=[428529,427931,427933,428055,427744,428281,427972,427736,428327,
            427948,427980,427780,427971,428053,427789,427785,427738,428108,
            427991,427742,427743,427614,427707];
    elif hash1 eq 2751 and hash2 eq 2759 and hash3 eq 3227 then
        V:=[28,1,29,5,30,11,31,15,32,23,33,27,34,33,35,39];
        D:=[1,129/5,1,0,2,265/12,1,671/30,1,0,4,308/15,1,0,5,499/30,1,793/42,1,
            1237/60,1,0,8,227/15,1,0,9,817/60,1,83/6,1,0,11,707/60,1,487/35,1,0,
            13,617/60,2,71/6,1,0];
        I:=[406382,404603,406837,406881,406380,404591,407020,406379,408103,
            406357,404601,404523,407751,407791,404099];
    elif hash1 eq 2939 and hash2 eq 20 and hash3 eq 1288 then
        V:=[22,1,23,11,24,19,25,29,26,39,27,43,28,61,29,67];
        D:=[1,21,1,23,1,25,1,27,1,0,5,115/6,1,127/6,1,47/2,1,0,8,53/3,1,18,3,
            77/4,1,20,1,0,14,85/6,1,29/2,1,177/10,1,233/12,1,0,18,238/15,1,0,19,
            127/10,1,77/6,4,79/6,1,279/20,1,169/12,1,431/30,1,29/2,1,59/4,1,0,
            30,188/15,1,302/21,1,0,32,341/30,1,0];
        I:=[502434,502394,502444,502475,502430,502332,502450,502426,502496,
            502517,502548,502252,502432,502576,502203,502153,502553,502410,
            502374,502122,502187,502400,502600,502395,502399,502392,502602,
            502205,502366,502409,501610,502368];
    elif hash1 eq 871 and hash2 eq 2680 and hash3 eq 3229 then
        V:=[25,1,26,7,27,17,29,23,30,29];
        D:=[1,127/6,1,151/6,1,0,3,53/3,2,24,1,77/3,1,89/3,1,0,8,109/6,1,133/6,1,
            0,10,89/6,3,101/6,1,0,14,199/12,1,253/15,1,52/3,1,0];
        I:=[395993,396421,396044,396377,396061,396347,395673,396350,396819,
            396628,396783,396789,396281,395623,396784,396695];
    elif hash1 eq 2093 and hash2 eq 222 and hash3 eq 1150 then
        V:=[18,1,19,13,20,27,21,43,22,59,23,77,24,87,25,93];
        D:=[1,23,1,27,1,31,1,98/3,1,33,1,0,6,127/6,2,43/2,1,91/4,1,139/6,2,
            115/4,1,59/2,1,0,14,20,1,85/4,1,64/3,2,65/3,1,22,3,116/5,1,71/3,1,0,
            24,107/6,1,109/6,1,233/12,1,121/6,2,41/2,2,257/12,1,143/6,1,0,33,
            268/15,1,73/4,1,55/3,2,56/3,1,39/2,1,81/4,1,62/3,1,109/5,1,0,42,
            33/2,1,101/6,1,103/6,1,37/2,2,0,47,76/5,1,61/4,1,0,49,213/14,1,
            461/30,1,0];
        I:=[427605,517522,517528,517544,427602,517599,517600,517521,517527,
            517532,517567,517476,517479,427604,517526,517540,517595,517598,
            427603,517413,517452,517130,517422,517596,517124,517523,517529,
            517530,517447,517461,517454,517079,517134,517524,517560,517594,
            517406,517457,517081,517453,517266,517593,517531,517463,517077,
            517082,517133,517128,517005,517003];
    elif hash1 eq 1012 and hash2 eq 1515 and hash3 eq 786 then
        V:=[19,1,20,5,21,11,23,19,24,25,25,33];
        D:=[1,69/2,1,0,2,74/3,1,31,2,0,5,47/2,1,151/6,2,51/2,2,0,10,109/6,1,
            133/6,1,0,12,215/12,1,55/3,2,56/3,2,0,17,217/12,1,0];
        I:=[253176,253164,253274,253320,252941,253223,253224,253051,253290,
            253308,252818,252517,253054,253151,252988,253377,252499];
    elif hash1 eq 2845 and hash2 eq 834 and hash3 eq 1651 then
        V:=[19,1,20,5,21,13,22,19,23,23,24,31,25,39];
        D:=[1,57/2,1,0,2,80/3,1,27,1,92/3,1,0,5,139/6,1,47/2,1,0,7,65/3,2,0,9,
            109/6,2,233/12,1,119/6,3,0,15,18,1,55/3,1,292/15,1,0,18,33/2,1,
            551/30,1,0];
        I:=[253428,253280,253427,252948,252934,253433,253197,253277,252894,
            253174,252876,252863,252878,252976,253145,252896,252605,252881,
            252560];
    elif hash1 eq 1482 and hash2 eq 2186 and hash3 eq 2466 then
        V:=[23,1,24,5,25,15,26,23,27,29];
        D:=[1,49/2,2,0,3,56/3,1,19,3,62/3,2,68/3,1,0,10,35/2,1,115/6,1,39/2,1,0,
            13,52/3,2,18,2,0,17,185/12,2,95/6,1,0];
        I:=[672746,672894,672893,672644,672770,672890,672777,672836,672563,
            672761,672475,672651,672503,672767,672560,672642,672537,672784,
            672501];
    elif hash1 eq 1811 and hash2 eq 1855 and hash3 eq 1015 then
        V:=[29,1,30,5,31,15,32,21,33,25];
        D:=[1,127/6,2,0,3,16,1,18,1,58/3,2,59/3,1,0,8,97/6,3,33/2,3,0,14,43/3,1,
            0,15,79/6,1,0];
        I:=[58136,58437,58159,58556,57892,58534,58465,57833,58387,58525,58395,
            58407,58561,58526,57843];
    elif hash1 eq 2234 and hash2 eq 3089 and hash3 eq 904 then
        V:=[26,1,27,11,28,19];
        D:=[1,239/12,1,109/5,1,67/3,1,74/3,1,0,5,125/6,1,671/30,1,91/4,1,0,8,
            277/15,1,223/12,2,227/12,1,19,1,20,1,308/15,1,21,1,0];
        I:=[413557,413558,413621,413519,413562,413536,413605,413567,413556,
            413571,413526,413458,413538,413547,413541];
    elif hash1 eq 25 and hash2 eq 3138 and hash3 eq 3362 then
        V:=[22,1,23,7,24,11,25,19];
        D:=[1,21,1,68/3,1,0,3,39/2,1,0,4,53/3,6,18,4,59/3,3,0,17,97/6,1,33/2,2,
            0];
        I:=[247491,247431,247486,247213,247252,247319,247326,247419,247442,
            247248,247281,247401,247415,246890,246922,247171,247238,247101,
            247216];
    elif hash1 eq 2422 and hash2 eq 2791 and hash3 eq 2634 then
        V:=[18,1,20,5,21,9,22,17,23,27,24,35,25,41,26,45,27,51,28,55];
        D:=[1,35,1,0,2,94/3,1,0,3,257/10,1,157/6,1,329/12,2,0,7,388/15,1,79/3,1,
            80/3,1,161/5,1,0,11,125/6,1,269/12,2,293/12,1,0,15,283/15,2,313/15,
            1,0,18,793/42,1,0,19,316/21,1,244/15,1,0,21,46/3,1,0,22,707/60,2,
            827/60,1,0];
        I:=[515191,515372,514947,515378,514389,514987,514988,515334,513964,
            514097,514991,515176,515290,514899,514543,514564,514100,514102,
            513994,513991,514152,514745,514898,513986];
    elif hash1 eq 730 and hash2 eq 3019 and hash3 eq 1004 then
        V:=[28,1,29,5,30,13,31,19,32,29];
        D:=[1,47/3,2,0,3,63/4,1,95/6,1,97/6,2,0,7,71/5,1,43/3,1,0,9,149/12,2,
            25/2,2,127/10,1,77/6,1,0,15,34/3,1,0];
        I:=[228243,228244,226585,226594,226588,228333,226752,226783,228327,
            228337,227453,228209,227470,226469,227549];
    elif hash1 eq 1059 and hash2 eq 2250 and hash3 eq 2053 then
        V:=[12,1,13,9,14,19,15,31,16,43,17,59,18,71,19,81,21,93];
        D:=[1,36,2,38,1,40,1,0,5,65/2,2,69/2,1,73/2,2,77/2,2,0,12,31,2,33,1,
            104/3,1,35,2,39,1,0,19,175/6,2,59/2,3,187/6,1,211/6,1,71/2,2,0,28,
            77/3,1,26,1,83/3,3,88/3,1,89/3,1,30,1,107/3,1,0,37,143/6,1,157/6,1,
            111/4,1,179/6,1,181/6,1,0,42,67/3,1,68/3,3,23,2,79/3,1,0,49,671/30,
            2,45/2,2,227/10,2,137/6,2,139/6,1,0,58,35/2,2,0];
        I:=[255835,255837,255836,255838,255822,255832,255823,255815,255831,
            255828,255833,255826,255827,255807,255805,255829,255834,255795,
            255800,255811,255810,255816,255824,255796,255808,255818,255819,
            255801,255825,255794,255804,255821,255806,255802,255813,255785,
            255817,255814,255792,255790,255784,255803,255797,255799,255812,
            255809,255830,255793,255781,255786,255789,255791,255783,255788,
            255780,255782,255787,255798,255820];
    elif hash1 eq 589 and hash2 eq 390 and hash3 eq 2942 then
        V:=[23,1,24,5,25,11,26,25,27,31,28,35,29,41,30,49];
        D:=[1,119/4,1,0,2,111/5,1,129/5,1,0,4,167/10,1,45/2,1,731/30,1,367/14,1,
            317/12,1,53/2,1,0,10,56/3,1,19,1,0,12,33/2,1,0,13,419/28,1,69/4,1,0,
            15,797/60,1,137/10,1,653/42,2,0,19,417/35,1,0];
        I:=[509529,509328,509266,508882,509423,508999,508744,509434,508741,
            508883,509422,508793,508314,508487,508884,508855,508739,508740,
            508766];
    elif hash1 eq 354 and hash2 eq 2118 and hash3 eq 2992 then
        V:=[22,1,23,7,24,11,25,15,26,21,28,27,30,33];
        D:=[1,74/3,1,86/3,1,0,3,151/6,1,0,4,65/3,2,0,6,119/6,1,121/6,2,0,9,
            215/12,1,55/3,1,0,11,29/2,2,175/12,2,0,15,35/3,1,0];
        I:=[674034,674029,673884,673898,674055,673885,673847,674015,673810,
            673902,673830,673933,673833,673992,673838];
    elif hash1 eq 2939 and hash2 eq 237 and hash3 eq 788 then
        V:=[27,1,28,9,29,19,30,29,31,37];
        D:=[1,115/6,5,39/2,3,127/6,1,0,10,16,1,52/3,1,53/3,2,18,1,0,15,95/6,1,
            97/6,4,33/2,2,107/6,3,0,25,57/4,2,43/3,1,44/3,1,0,29,77/6,4,0];
        I:=[60530,60542,60559,60834,60854,60772,60835,60892,60682,60800,60515,
            60408,60549,60592,60775,60407,60485,60578,60597,60360,60586,60314,
            60315,60400,60401,60489,60396,60879,60397,60437,60438,60441];
    elif hash1 eq 1811 and hash2 eq 2183 and hash3 eq 219 then
        V:=[27,1,29,13,30,23,32,31];
        D:=[1,39/2,1,127/6,1,139/6,1,47/2,1,151/6,1,0,6,95/6,2,177/10,1,109/6,1,
            37/2,2,0,12,16,1,49/3,1,215/12,2,0,16,218/15,1,0];
        I:=[669104,669029,669202,669027,669028,668443,668945,668704,668537,
            669042,669265,669030,668483,668086,668134,667953];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 26 mod 47.
function cached_ids_26(P,hash1,hash2,hash3)
    if hash1 eq 2705 and hash2 eq 2473 and hash3 eq 376 then
        V:=[17,1,18,9,19,29,20,47,21,69,22,81,23,99,24,127,25,147,26,151,27,
            157];
        D:=[1,57/2,1,65/2,1,77/2,3,0,6,80/3,1,27,1,86/3,1,29,3,31,4,33,1,104/3,
            1,35,2,116/3,1,0,21,151/6,2,163/6,1,175/6,2,59/2,2,123/4,1,187/6,8,
            63/2,2,199/6,3,0,42,65/3,1,71/3,1,126/5,1,76/3,1,77/3,5,26,1,109/4,
            1,83/3,1,32,2,107/3,1,0,57,131/6,1,95/4,1,143/6,1,305/12,1,155/6,1,
            0,62,50/3,1,215/12,1,55/3,1,56/3,1,239/12,1,61/3,2,89/4,1,67/3,1,0,
            71,491/30,1,33/2,1,359/20,1,18,1,217/12,1,221/12,3,37/2,1,187/10,2,
            75/4,2,113/6,4,115/6,2,245/12,1,41/2,1,0,92,97/6,1,199/12,1,17,1,
            380/21,1,223/12,1,19,1,58/3,1,271/12,1,275/12,1,0,101,451/30,1,0,
            102,187/12,1,47/3,1,0,104,361/30,2,0];
        I:=[427580,427576,427348,427424,427470,427577,427355,427334,427167,
            427174,427590,427226,427353,427423,427585,427337,427414,427410,
            427422,427200,427253,427397,427565,426930,427136,427131,427336,
            427121,427122,427161,427411,427416,427499,427518,427520,427579,
            427227,427351,427363,427402,427570,427471,427175,426929,427377,
            426706,427001,427218,427247,427448,427002,426668,426814,426829,
            426983,426846,426715,426525,426837,427040,426477,426618,427016,
            427571,427566,426790,427024,427169,426736,426773,426802,426876,
            426490,426526,426848,426495,427109,427170,427114,426663,426825,
            426585,427249,426801,427125,427251,427446,427079,427496,426817,
            426798,427130,426666,427112,426468,426462,426547,426478,426456,
            426452,426822,426469,426470,426688,426745];
    elif hash1 eq 2564 and hash2 eq 62 and hash3 eq 1160 then
        V:=[22,1,23,11,24,17,25,21,26,27,27,31];
        D:=[1,23,3,74/3,2,25,1,27,1,0,8,127/6,1,139/6,1,0,10,59/3,2,0,12,121/6,
            2,41/2,1,0,15,44/3,1,0,16,431/30,1,173/12,1,29/2,1,0];
        I:=[249081,249220,249402,249234,249374,249085,249361,249407,249270,
            248967,249362,248321,248582,248448,248950,248338,249105,249470];
    elif hash1 eq 2893 and hash2 eq 1576 and hash3 eq 97 then
        V:=[24,1,25,9,26,21,27,33,28,41,29,47,30,51];
        D:=[1,59/3,1,71/3,1,77/3,1,0,4,107/6,1,109/6,1,233/12,1,121/6,1,257/12,
            1,0,9,16,1,49/3,1,18,1,55/3,1,298/15,1,0,14,491/30,1,103/6,2,551/30,
            1,0,18,223/15,1,247/15,1,0,20,401/30,1,0,21,139/12,1,41/3,1,0];
        I:=[414557,414575,414138,414467,414571,414780,414232,414123,414468,
            414622,414075,414625,413990,414143,414589,414599,414060,414168,
            413713,413927,414023,413439];
    elif hash1 eq 2611 and hash2 eq 751 and hash3 eq 1633 then
        V:=[28,1,29,11,30,21,31,27,32,33];
        D:=[1,64/3,1,65/3,1,22,1,71/3,1,0,5,109/6,2,79/4,1,119/6,1,121/6,1,0,10,
            65/4,1,50/3,2,0,13,89/6,2,197/12,1,0,16,155/12,1,0];
        I:=[664664,664562,664796,664258,664574,664638,663990,664688,664257,
            664002,664606,664639,664154,664655,663247,663700];
    elif hash1 eq 2235 and hash2 eq 1257 and hash3 eq 546 then
        V:=[19,1,20,11,21,21,22,33,23,45,24,49,25,55,27,59];
        D:=[1,57/2,1,61/2,1,65/2,1,77/2,1,0,5,29,1,92/3,1,104/3,1,35,1,0,9,
            151/6,1,59/2,1,187/6,1,211/6,1,71/2,3,0,16,76/3,1,83/3,2,88/3,1,
            89/3,1,30,1,0,22,119/6,1,0,23,111/5,1,68/3,2,0,26,551/30,1,0,27,
            877/60,2,91/6,2,511/30,1,103/6,2,0];
        I:=[420063,420379,420045,420436,420061,420377,419999,420378,420209,
            420280,420438,420367,420281,420312,420323,420275,420189,420418,
            420287,420430,419824,420071,419566,419537,419575,419556,419285,
            419530,420303,420304,419148,419145,419170];
    elif hash1 eq 2047 and hash2 eq 974 and hash3 eq 2224 then
        V:=[17,1,18,7,19,13,21,23,22,33,25,39,26,43,27,49];
        D:=[1,61/2,1,69/2,1,0,3,116/3,1,39,1,0,5,163/6,1,175/6,1,71/2,1,235/6,1,
            0,9,155/6,1,111/4,1,357/10,1,217/6,2,0,14,388/15,1,418/15,1,0,16,
            31/2,1,0,17,907/60,1,86/5,1,0,19,1271/84,1,303/20,1,653/42,1,467/30,
            1,157/10,1,63/4,1,0];
        I:=[516433,516366,516453,516456,516450,516658,516458,516455,516430,
            516482,516335,516329,516330,515909,515953,516457,516343,515847,
            515770,515765,515762,515750,515763,515747];
    elif hash1 eq 1671 and hash2 eq 2546 and hash3 eq 3277 then
        V:=[19,1,20,7,21,13,22,19,25,23];
        D:=[1,215/6,1,217/6,2,0,4,538/15,1,431/12,1,0,6,559/20,1,341/12,1,0,8,
            70/3,1,0,9,937/60,1,157/10,1,63/4,1,95/6,2,107/6,2,0];
        I:=[428527,428463,428521,427943,428047,427975,427967,428526,428046,
            427944,427976,428466,428467,427747,427748];
    elif hash1 eq 1154 and hash2 eq 428 and hash3 eq 2805 then
        V:=[25,1,26,5,27,15,28,23,29,29];
        D:=[1,115/4,1,0,2,65/3,1,71/3,3,101/4,1,77/3,1,0,8,197/10,1,87/4,1,
            133/6,1,0,11,91/5,2,239/12,2,0,15,539/30,1,0];
        I:=[671499,671463,671196,671243,671263,671174,671242,671450,671276,
            671568,671105,671453,671277,671638,671159];
    elif hash1 eq 1906 and hash2 eq 2184 and hash3 eq 1228 then
        V:=[17,1,18,7,19,15,20,23,21,27,22,37,23,43];
        D:=[1,57/2,1,69/2,1,0,3,25,2,27,1,29,1,0,7,127/6,1,151/6,1,51/2,1,0,10,
            65/3,1,0,11,107/6,2,197/10,1,45/2,1,103/4,1,0,16,50/3,1,55/3,1,0,18,
            491/30,1,33/2,1,0];
        I:=[517436,517438,517433,517440,517477,517478,517389,517451,517428,
            517397,517371,517384,517329,517458,517101,517472,517378,517333,
            517409];
    elif hash1 eq 2752 and hash2 eq 2623 and hash3 eq 1833 then
        V:=[25,1,27,9,28,13,29,21,30,29,31,33,32,37,33,45];
        D:=[1,317/12,1,839/30,1,409/12,1,0,4,1033/42,1,0,5,662/35,1,407/21,1,21,
            1,0,8,1219/70,1,517/24,1,175/8,1,0,11,827/60,1,0,12,331/24,1,0,13,
            71/7,1,254/21,1,73/6,1,0,16,917/90,1,0];
        I:=[539852,540001,539856,539736,539958,539916,539955,539956,539744,
            539745,539853,539661,539700,539675,539672,540332];
    elif hash1 eq 2940 and hash2 eq 1401 and hash3 eq 3319 then
        V:=[23,1,24,5,26,11,27,19,28,29];
        D:=[1,133/6,1,0,2,263/12,1,352/15,1,0,4,50/3,1,422/21,1,22,1,0,7,511/30,
            1,19,1,39/2,1,127/6,1,0,11,309/20,1,113/6,2,287/15,1,424/21,1,0];
        I:=[539952,539925,539965,539921,539913,539876,539769,539778,539884,
            539880,539896,539754,539785,539784,539871];
    elif hash1 eq 2517 and hash2 eq 2862 and hash3 eq 902 then
        V:=[20,1,21,7,22,11,23,17,24,25,25,29,26,35,27,43];
        D:=[1,77/3,1,89/3,1,0,3,155/6,1,0,4,22,1,24,1,0,6,125/6,1,265/12,1,
            671/30,1,0,9,223/12,1,0,10,371/20,1,379/20,1,0,12,233/15,2,187/12,1,
            47/3,4,0,19,63/4,1,0];
        I:=[426322,426325,426326,426287,426286,426388,425866,425879,425748,
            425650,425876,425763,425814,426053,425861,426001,426273,426377,
            425657];
    elif hash1 eq 1013 and hash2 eq 2630 and hash3 eq 3437 then
        V:=[14,1,15,15,16,27,17,39,18,45,19,53];
        D:=[1,30,3,32,1,34,2,36,1,38,2,42,1,0,11,53/2,1,57/2,1,61/2,4,65/2,1,
            69/2,2,0,20,25,2,80/3,1,27,3,33,1,35,1,0,28,163/6,3,55/2,2,0,33,20,
            1,82/3,1,83/3,1,0,36,119/6,1,0];
        I:=[254885,254887,254888,254883,254886,254896,254890,254895,254899,
            254829,254894,254892,254840,254884,254889,254893,254855,254837,
            254853,254854,254891,254831,254803,254835,254836,254795,254794,
            254793,254833,254838,254799,254848,254808,254801,254797,254796];
    elif hash1 eq 167 and hash2 eq 1245 and hash3 eq 689 then
        V:=[18,1,19,7,20,13,21,21,22,29,23,33,24,43,25,49];
        D:=[1,116/3,1,39,1,0,3,139/6,1,139/4,1,0,5,77/3,1,109/4,1,95/3,1,0,8,
            79/4,1,155/6,1,273/10,1,0,11,352/15,1,0,12,491/30,1,33/2,1,221/12,1,
            245/12,1,0,16,20,1,121/6,1,0,18,451/30,1,0];
        I:=[515504,515486,515481,515254,515531,515281,515278,515217,514845,
            514659,514606,514511,514501,515239,514902,514290,514322,514515];
    elif hash1 eq 402 and hash2 eq 2525 and hash3 eq 2805 then
        V:=[26,1,27,9,28,13,29,19,30,27,31,35];
        D:=[1,70/3,1,71/3,1,83/3,1,0,4,121/6,1,0,5,55/3,1,62/3,1,0,7,197/12,3,
            101/6,2,539/30,1,0,13,175/12,1,44/3,1,223/15,1,0,16,79/6,1,161/12,1,
            83/6,1,0];
        I:=[241557,241454,240313,241517,238394,238936,238400,241481,241535,
            241427,241434,238402,240534,238390,240114,239179,239114,241600];
    elif hash1 eq 1671 and hash2 eq 2063 and hash3 eq 499 then
        V:=[25,1,26,11,27,25,28,29,29,35];
        D:=[1,109/6,1,119/6,1,143/6,1,253/10,1,0,5,91/5,1,55/3,1,56/3,1,61/3,3,
            22,1,89/4,1,0,13,221/12,1,0,14,44/3,1,329/20,1,0,16,27/2,1,0];
        I:=[414245,414243,413820,414276,414244,414173,414241,413493,414139,
            414267,413652,413658,414284,414348,413495,414291];
    elif hash1 eq 26 and hash2 eq 3087 and hash3 eq 2687 then
        V:=[18,1,19,5,20,15,21,21,22,27,23,33,24,47];
        D:=[1,29,1,0,2,151/6,1,51/2,1,163/6,1,63/2,2,0,7,76/3,1,77/3,4,0,12,
            131/6,3,155/6,1,0,16,263/12,2,22,1,0,19,37/2,2,187/10,1,113/6,3,
            115/6,3,39/2,1,265/12,1,0,30,19,1,0];
        I:=[254285,254310,254262,254278,254269,254315,254279,254136,254219,
            254246,254306,254250,254253,254276,254186,254128,254141,254193,
            254140,254325,254137,254153,254199,254298,254130,254191,254292,
            254266,254121,254148];
    elif hash1 eq 120 and hash2 eq 478 and hash3 eq 2855 then
        V:=[26,1,27,7,28,13,30,21];
        D:=[1,17,2,19,3,0,6,103/6,2,35/2,2,0,10,61/4,1,47/3,6,16,3,0,20,13,1,0];
        I:=[667831,667863,667651,667827,667909,667248,667312,667830,667875,
            667234,666851,666854,667033,667536,667714,667839,666922,667047,
            667169,665936];
    elif hash1 eq 966 and hash2 eq 668 and hash3 eq 266 then
        V:=[21,1,22,7,23,13,24,25,25,33,26,39,28,45];
        D:=[1,55/2,1,211/6,1,0,3,95/3,1,101/3,2,0,6,87/4,1,133/6,1,95/4,1,143/6,
            1,257/10,1,0,11,111/5,1,89/4,1,80/3,1,0,14,551/30,1,599/30,1,0,16,
            386/21,1,283/15,1,0,18,281/21,1,96/7,1,208/15,2,0];
        I:=[419505,419743,419863,419723,419839,419203,419742,419477,419509,
            419466,419601,419378,419050,419226,419199,419006,419048,418956,
            418976,419049,419709];
    elif hash1 eq 2987 and hash2 eq 901 and hash3 eq 2725 then
        V:=[21,1,22,5,23,13,24,19,25,27,26,31,27,35,28,41];
        D:=[1,115/4,1,0,2,93/4,1,24,1,89/3,2,0,6,119/6,1,257/10,1,0,8,268/15,1,
            298/15,1,382/15,1,0,11,187/10,1,0,12,40/3,1,0,13,157/12,1,461/30,1,
            0,15,77/6,1,274/21,1,0];
        I:=[510288,510285,510474,510301,510302,509691,510163,509685,510191,
            509206,510197,510298,510299,510198,510631,508852];
    elif hash1 eq 261 and hash2 eq 1233 and hash3 eq 2036 then
        V:=[19,1,20,9,21,15,22,29,23,39,24,49,25,67,26,71];
        D:=[1,175/6,1,59/2,1,187/6,2,0,5,77/3,1,28,2,0,8,87/4,1,45/2,3,143/6,2,
            145/6,2,103/4,1,157/6,1,0,18,61/3,3,62/3,1,263/12,1,67/3,3,0,26,
            37/2,1,75/4,1,113/6,6,659/30,1,0,35,199/12,1,203/12,1,17,1,283/15,3,
            19,1,96/5,1,77/4,1,337/15,1,0,45,91/6,1,0,46,309/20,1,187/12,1,0];
        I:=[425608,425615,425584,425586,425600,425566,425574,425564,425565,
            425568,425573,425569,425613,425571,425593,425535,425572,425547,
            425579,425610,425590,425570,425489,425515,425560,425484,425552,
            425524,425546,425549,425554,425589,425601,425486,425490,425531,
            425533,425483,425487,425499,425479,425497,425477,425471,425494,
            425472,425473];
    elif hash1 eq 2188 and hash2 eq 1386 and hash3 eq 355 then
        V:=[25,1,26,17,27,27,29,33,30,45,31,49];
        D:=[1,163/6,1,55/2,1,115/4,1,175/6,1,187/6,1,63/2,1,67/2,1,0,8,22,1,
            116/5,1,71/3,2,83/3,1,0,13,107/6,1,79/4,1,0,15,16,1,491/30,1,33/2,1,
            221/12,1,37/2,1,0,20,223/15,1,0,21,161/12,1,0];
        I:=[245527,244685,244012,244633,244634,244655,244666,244594,244179,
            244637,244683,244671,243439,244571,242194,242255,244648,244547,
            242789,244546,244569];
    elif hash1 eq 2658 and hash2 eq 196 and hash3 eq 3123 then
        V:=[19,1,20,9,21,25,22,37,23,47,24,57,25,67,26,85,27,111,28,125];
        D:=[1,211/6,1,71/2,1,259/6,1,0,4,28,1,88/3,1,100/3,1,176/5,1,141/4,1,
            107/3,2,157/4,1,0,12,131/6,1,277/10,2,297/10,1,317/10,1,191/6,1,0,
            18,73/3,1,311/12,1,418/15,1,159/5,1,0,22,18,1,551/30,1,399/20,1,
            265/12,1,0,26,199/12,1,422/21,1,308/15,1,247/12,1,0,30,83/6,1,
            299/20,1,181/12,1,169/10,1,511/30,1,181/10,1,379/20,1,569/30,1,0,38,
            203/15,2,74/5,1,301/20,1,91/6,2,61/4,1,46/3,2,77/5,1,233/15,2,16,1,
            1003/60,1,352/21,1,94/5,1,0,54,797/60,1,179/12,1,1059/70,1,913/60,1,
            917/60,1,183/10,1,0,60,117/10,1,0];
        I:=[515472,515253,514856,515297,515482,515509,514849,514456,515296,
            515436,514317,514754,514457,514848,514681,514828,514426,515127,
            514439,514855,514166,514535,514887,514257,514064,514378,514030,
            514819,514461,514394,514748,514380,514455,514440,514028,514431,
            514126,514401,514868,514005,514865,514669,515321,514319,515469,
            515551,514896,514371,514464,515405,514176,514173,513990,514894,
            513993,514135,513995,514027,513959,514391];
    elif hash1 eq 2047 and hash2 eq 1222 and hash3 eq 3209 then
        V:=[25,1,26,5,27,9,28,23,30,29];
        D:=[1,91/4,1,0,2,85/4,1,0,3,71/4,1,109/6,2,193/10,1,79/4,1,119/6,3,
            233/10,1,0,12,16,1,49/3,1,0,14,193/15,1,223/15,1,0];
        I:=[405694,405905,406264,405749,405868,405645,405605,405602,405935,
            405937,405819,406068,404736,405360,404729];
    elif hash1 eq 3081 and hash2 eq 3080 and hash3 eq 2920 then
        V:=[23,1,24,9,25,13,26,23,27,35,28,53,29,67,30,71];
        D:=[1,139/6,1,59/2,3,63/2,1,0,6,24,1,0,7,37/2,2,119/6,1,41/2,1,257/12,1,
            0,12,49/3,1,215/12,1,55/3,1,298/15,1,81/4,1,0,17,431/30,1,173/12,3,
            491/30,1,33/2,1,551/30,1,37/2,1,75/4,1,659/30,1,0,27,72/5,1,218/15,
            1,175/12,1,223/15,1,33/2,1,199/12,1,0,33,391/30,1,0,34,229/20,2,
            139/12,1,0];
        I:=[418112,416533,417986,418023,416549,416745,417988,417998,418559,
            416413,417261,417650,417179,418350,416907,417276,417363,416432,
            417391,417500,417182,417327,416954,416721,416017,415016,417462,
            417221,416785,416934,416244,415314,416405,416451,416899,416780];
    elif hash1 eq 2940 and hash2 eq 2940 and hash3 eq 1219 then
        V:=[21,1,22,9,23,15,24,31,25,43,26,47,27,61];
        D:=[1,35/2,1,55/2,1,67/2,1,0,4,71/3,1,89/3,1,0,6,95/6,2,107/6,1,119/6,1,
            217/10,1,131/6,1,49/2,1,257/10,1,0,14,49/3,1,91/5,1,56/3,1,81/4,1,
            21,1,0,19,431/30,1,0,20,289/20,2,218/15,1,175/12,2,253/15,1,52/3,1,
            422/21,1,0,28,129/10,1,0];
        I:=[510264,508981,508938,509308,508970,510328,510399,509330,509299,
            508948,508972,509018,509398,510405,508775,508840,509293,508910,
            509768,509888,510093,510019,509924,509932,508786,508496,508429,
            508872];
    elif hash1 eq 2705 and hash2 eq 781 and hash3 eq 1151 then
        V:=[21,1,22,5,23,11,24,17,25,29,27,39];
        D:=[1,41/2,1,0,2,19,1,21,1,0,4,35/2,1,115/6,1,0,6,52/3,3,53/3,1,18,1,
            96/5,1,77/4,1,0,13,63/4,1,33/2,1,109/6,1,233/12,1,0,17,27/2,1,33/2,
            1,0];
        I:=[412399,412365,412343,411579,412409,412349,412391,412393,411224,
            412284,411327,411231,411199,412211,412010,411218,410849,409429];
    elif hash1 eq 26 and hash2 eq 2474 and hash3 eq 3011 then
        V:=[23,1,24,7,25,17,26,23,27,31,28,35];
        D:=[1,45/2,1,49/2,2,0,4,62/3,2,21,3,68/3,1,23,3,0,13,115/6,5,39/2,3,0,
            21,69/4,1,53/3,2,18,1,0,25,97/6,1,0,26,43/3,1,44/3,1,0];
        I:=[61720,61656,61710,61712,61718,61668,61669,61681,61626,61556,61599,
            61616,61564,61588,61593,61618,61714,61570,61614,61677,61584,61624,
            61717,61667,61545,61539,61547];
    elif hash1 eq 2987 and hash2 eq 746 and hash3 eq 2346 then
        V:=[18,1,19,17,20,33,21,57,22,75,23,85,24,105,25,129,26,137];
        D:=[1,23,1,80/3,1,31,1,98/3,1,33,2,37,3,116/3,1,0,11,127/6,1,107/4,1,
            163/6,1,59/2,1,63/2,1,199/6,1,211/6,1,0,18,85/4,1,64/3,2,65/3,1,22,
            1,71/3,1,26,1,88/3,1,156/5,1,125/4,2,95/3,2,141/4,1,0,32,193/10,1,
            119/6,1,41/2,1,257/12,1,131/6,1,237/10,1,155/6,1,157/6,2,0,41,17,1,
            55/3,1,56/3,1,61/3,5,0,49,197/12,1,101/6,1,103/6,1,351/20,1,217/12,
            1,113/6,1,391/20,1,611/30,3,207/10,2,0,61,76/5,1,247/15,1,199/12,2,
            88/5,1,92/5,1,129/7,1,369/20,1,283/15,2,19,1,96/5,1,337/15,1,0,74,
            83/6,1,697/42,1,67/4,1,0,77,269/20,1,203/15,2,0];
        I:=[515412,515596,515091,515498,515109,515556,515251,515298,515471,
            515431,515608,515327,515597,515118,515157,515402,515604,515434,
            515220,515460,515603,515024,515080,515458,514274,514404,514327,
            515435,515428,515429,514467,514586,515607,515079,514375,515439,
            514418,514326,514770,514854,514417,515202,515312,514756,515120,
            515305,515316,515595,514433,515388,515414,514809,514435,514296,
            514432,514351,514353,514463,514373,514566,514790,514698,514249,
            514395,514438,514437,514127,514245,514573,514862,514248,514407,
            514004,514427,514019,514250,514458,514358,514478];
    elif hash1 eq 261 and hash2 eq 1403 and hash3 eq 534 then
        V:=[15,1,16,7,17,15,18,21,19,31,20,39,21,47,22,55,23,67,24,71];
        D:=[1,73/2,1,77/2,1,0,3,33,2,37,1,128/3,1,0,7,199/6,3,67/2,1,0,11,77/3,
            2,133/4,1,100/3,1,107/3,1,0,16,305/12,1,103/4,1,167/6,3,0,21,22,1,
            67/3,1,311/12,1,0,24,241/12,1,41/2,1,137/6,1,0,27,302/15,1,307/15,1,
            41/2,1,338/15,1,343/15,1,0,32,1213/60,1,0,33,907/60,1,76/5,1,61/4,1,
            187/12,1,47/3,1,0];
        I:=[255729,255728,255682,255727,255662,255545,255515,255525,255672,
            255685,255671,255684,255521,255509,255530,255522,255531,255519,
            255529,255538,255512,255520,255423,255527,255540,255516,255436,
            255437,255435,255360,255359,255361,255424,255425,255433,255432,
            255418];
    elif hash1 eq 1107 and hash2 eq 203 and hash3 eq 360 then
        V:=[26,1,27,7,28,13,29,21,30,29,31,33];
        D:=[1,23,2,74/3,1,0,4,35/2,1,115/6,1,0,6,47/3,3,53/3,1,18,1,0,11,97/6,1,
            177/10,1,109/6,1,0,14,38/3,1,0,15,371/30,3,149/12,1,0];
        I:=[234901,235025,234711,234990,234732,232229,232535,233088,234472,
            234404,233663,231467,232029,232249,232108,232244,232246,232308];
    elif hash1 eq 3316 and hash2 eq 1126 and hash3 eq 979 then
        V:=[27,1,28,11,29,15,30,21,32,27];
        D:=[1,143/6,2,155/6,2,157/6,1,167/6,2,0,8,73/3,1,0,9,125/6,1,127/6,1,0,
            11,17,1,52/3,1,0,13,41/3,4,227/15,1,91/6,1,233/15,2,47/3,2,0];
        I:=[241531,241609,241378,241501,241528,238442,241380,238512,238357,
            238513,241122,241523,238526,238618,241538,241573,238331,238371,
            238372,238630,238399,238510];
    elif hash1 eq 496 and hash2 eq 2268 and hash3 eq 2081 then
        V:=[19,1,20,9,21,27,22,39,23,49,24,57,26,65];
        D:=[1,57/2,1,61/2,3,65/2,2,0,7,74/3,1,25,1,80/3,1,27,1,86/3,3,29,1,92/3,
            1,31,1,0,17,139/6,1,151/6,1,51/2,5,55/2,2,175/6,1,0,27,85/4,1,64/3,
            1,65/3,3,71/3,2,0,34,119/6,1,121/6,1,133/6,4,0,40,215/12,1,73/4,1,
            55/3,1,0,43,223/15,1,179/12,1,15,1,0];
        I:=[674525,674513,674517,674528,674511,674520,674526,674522,674524,
            674527,674431,674435,674510,674521,674514,674512,674515,674445,
            674444,674449,674508,674519,674523,674441,674454,674397,674434,
            674429,674428,674446,674516,674426,674427,674453,674437,674393,
            674430,674438,674447,674433,674432,674448,674395,674398,674394];
    elif hash1 eq 872 and hash2 eq 661 and hash3 eq 964 then
        V:=[25,1,26,13,27,19,28,35,29,41];
        D:=[1,103/6,1,43/2,1,47/2,1,151/6,1,115/4,1,0,6,59/3,1,24,1,0,8,161/12,
            1,83/6,1,95/6,2,97/6,1,209/12,1,177/10,1,121/6,1,0,16,14,1,73/4,1,0,
            18,195/14,1,29/2,1,0];
        I:=[398844,397730,397568,397789,397834,397995,395727,400063,400474,
            397994,400041,400542,399349,397715,395863,398845,398059,395836,
            400272];
    elif hash1 eq 2611 and hash2 eq 644 and hash3 eq 1926 then
        V:=[25,1,26,7,27,15,28,27,29,33,30,37];
        D:=[1,39/2,1,99/4,1,0,3,59/3,1,106/5,1,101/4,1,0,6,177/10,1,107/6,1,
            233/12,1,121/6,1,133/6,1,0,11,49/3,2,18,1,0,14,29/2,1,0,15,40/3,1,
            41/3,1,46/3,1,0];
        I:=[239171,239090,240160,239259,239034,240000,240070,240526,240151,
            238410,239199,240155,240525,238969,238988,238966,238045];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 27 mod 47.
function cached_ids_27(P,hash1,hash2,hash3)
    if hash1 eq 1954 and hash2 eq 1126 and hash3 eq 2331 then
        V:=[16,1,17,9,18,15,19,29,20,37,21,45,24,53];
        D:=[1,28,1,34,1,40,1,0,4,45/2,1,65/2,1,0,6,56/3,1,21,1,104/3,1,35,1,
            110/3,1,37,2,0,13,47/2,1,175/6,1,63/2,1,0,16,93/4,1,71/3,1,83/3,1,0,
            19,107/6,2,119/6,1,217/10,1,0,23,223/15,1,52/3,1,0];
        I:=[516764,516777,516766,516717,516767,516711,516763,516524,516509,
            516478,516589,516655,516518,516510,516197,516538,516519,515829,
            516527,516654,515803,516031,516027,515813];
    elif hash1 eq 1531 and hash2 eq 3127 and hash3 eq 540 then
        V:=[20,1,21,5,22,11,23,19,24,25,25,43,26,53,27,59,28,69,29,85,30,97];
        D:=[1,166/5,1,0,2,193/6,1,425/12,1,0,4,101/5,1,287/12,1,478/15,4,0,10,
            779/30,1,347/10,1,0,12,58/3,1,338/15,1,126/5,1,101/4,1,554/21,1,
            529/20,1,403/15,1,27,1,0,20,75/4,1,571/30,1,79/4,1,1045/42,1,0,24,
            323/21,1,77/4,1,0,26,277/18,1,433/28,1,667/42,2,52/3,1,0,31,181/15,
            1,727/60,3,401/30,1,1145/84,1,96/7,1,208/15,3,110/7,1,0,42,709/60,3,
            2537/210,1,1217/90,1,491/36,1,2873/210,1,0,49,544/45,1,0];
        I:=[542246,542377,542376,541875,541892,542280,542339,542374,542375,
            542035,542382,541849,542097,541847,541820,541929,542191,542378,
            542011,541888,542099,541817,541911,541880,541885,541779,541780,
            541956,541961,542002,542383,542288,542295,542369,541889,541887,
            541986,541881,542328,542355,541985,542291,542292,542380,541871,
            541781,541772,541771,541782];
    elif hash1 eq 3035 and hash2 eq 472 and hash3 eq 2631 then
        V:=[21,1,22,7,23,15,24,25,25,37,26,47,27,53];
        D:=[1,49/2,1,61/2,1,0,3,21,1,23,1,25,1,0,6,39/2,1,127/6,1,139/6,2,47/2,
            3,0,13,53/3,2,18,1,58/3,1,59/3,3,20,1,0,21,95/6,1,107/6,1,109/6,1,
            121/6,1,0,25,43/3,1,49/3,4,0,30,89/6,2,0];
        I:=[249228,248805,249276,249439,249072,249056,249046,248118,249076,
            248501,249031,249204,248671,249033,249044,249006,248568,249005,
            249066,249073,248500,248356,248232,248041,248496,248119,248603,
            249035,249269,248979,249091];
    elif hash1 eq 1719 and hash2 eq 1293 and hash3 eq 2022 then
        V:=[16,1,17,13,18,29,19,47,20,61,21,79,22,97,23,107];
        D:=[1,28,2,30,1,32,1,34,1,36,1,0,7,45/2,1,49/2,3,53/2,1,61/2,2,65/2,4,
            69/2,2,73/2,1,0,21,21,3,25,1,80/3,6,27,3,86/3,1,29,3,33,1,35,2,0,41,
            127/6,1,43/2,3,139/6,1,47/2,3,151/6,1,55/2,1,0,51,77/4,1,58/3,1,
            59/3,2,64/3,1,65/3,2,22,1,71/3,2,83/3,1,0,62,107/6,2,109/6,1,119/6,
            1,121/6,2,131/6,1,133/6,2,45/2,2,157/6,1,0,74,15,1,35/2,1,239/12,1,
            67/3,1,0,78,491/30,1,33/2,2,101/6,1,0];
        I:=[254751,254772,254752,254750,254743,254744,254768,254765,427458,
            427469,427097,254759,427583,254745,254748,427369,427477,427352,
            427578,427455,254758,254771,427457,427459,427096,427372,427443,
            427493,427515,427581,254749,427454,427519,427368,427080,427085,
            427341,427015,426709,426985,427467,254746,254747,427339,427160,
            427098,427456,427497,427014,427064,426984,427120,427017,427500,
            427025,426991,427082,426844,427157,427172,426545,427398,427503,
            427396,427090,427092,427589,427093,426628,427250,426675,426771,
            426499,427208,427013,426543,426767,426659,427382,427504,426781];
    elif hash1 eq 2330 and hash2 eq 1974 and hash3 eq 1024 then
        V:=[24,1,25,9,26,15,27,23,28,27,30,35,31,39];
        D:=[1,89/3,1,95/3,1,118/3,1,0,4,45/2,1,157/6,1,0,6,121/5,1,388/15,1,
            79/3,1,0,9,115/6,1,0,10,253/15,1,58/3,2,313/15,1,0,14,1027/60,2,0,
            16,407/30,1,137/10,2,0];
        I:=[501165,501365,501178,501336,501183,501260,501259,501325,501327,
            501338,501117,501122,501371,501257,501357,501256,501251,501374];
    elif hash1 eq 2518 and hash2 eq 3253 and hash3 eq 2410 then
        V:=[23,1,24,5,25,11,26,21,27,25,28,31];
        D:=[1,61/2,1,0,2,74/3,1,29,3,0,6,91/4,1,139/6,2,47/2,1,151/6,2,0,12,
            53/3,1,0,13,109/6,1,121/6,1,0,15,191/12,1,49/3,1,50/3,1,55/3,1,0];
        I:=[245294,245039,244195,245050,245051,244049,243097,244000,244941,
            242860,242928,243947,242890,242929,244381,243234,245061,242894];
    elif hash1 eq 262 and hash2 eq 3080 and hash3 eq 861 then
        V:=[15,1,16,5,17,15,18,21,19,39,20,61,21,87,22,101,23,119,24,143,25,151,
            26,161];
        D:=[1,69/2,1,0,2,29,1,39,1,122/3,1,128/3,1,0,6,63/2,1,171/4,1,0,8,77/3,
            1,26,1,117/4,1,88/3,1,30,2,156/5,1,94/3,1,119/3,1,0,17,133/6,1,95/4,
            1,143/6,1,49/2,1,155/6,1,157/6,1,119/4,1,317/10,1,191/6,1,193/6,2,0,
            28,239/12,1,81/4,1,328/15,1,263/12,2,22,1,67/3,1,287/12,2,73/3,2,
            131/5,1,79/3,2,335/12,1,383/12,1,0,44,115/6,1,39/2,1,20,2,125/6,1,
            719/30,1,53/2,1,0,51,17,1,109/6,1,223/12,1,19,1,401/20,1,121/6,1,
            308/15,1,21,1,0,59,461/30,1,997/60,1,509/28,1,1093/60,3,75/4,1,
            379/20,1,19,1,115/6,1,581/30,2,197/10,2,621/28,1,0,74,31/2,1,274/15,
            1,183/10,1,0,77,1271/84,1,76/5,1,367/20,1,94/5,1,0,81,727/60,2,0];
        I:=[428450,428451,428291,428436,428336,428166,428098,428390,428168,
            428162,428389,428158,428238,427914,428448,428066,428284,428237,
            428322,428157,428433,428534,428003,427769,427815,427881,428177,
            428330,427936,427946,427986,428049,427763,428388,427775,428163,
            428236,428362,427916,427985,428366,428073,427651,428279,428164,
            427952,428273,428365,427942,427669,428070,428277,427729,427777,
            428028,427953,427817,428068,428051,428135,427673,427699,428012,
            428114,427745,427720,427770,427715,427734,427989,427718,427958,
            427616,428023,427776,427786,427690,427687,427607,427606,428116,
            428128];
    elif hash1 eq 2659 and hash2 eq 1232 and hash3 eq 1601 then
        V:=[23,1,24,7,25,13,26,19,27,23,28,31,29,35];
        D:=[1,49/2,1,57/2,1,0,3,68/3,1,23,2,0,6,127/6,2,139/6,2,0,10,59/3,1,0,
            11,95/6,1,109/6,1,121/6,1,0,14,268/15,1,0,15,14,1,0];
        I:=[244915,244904,244916,244865,244898,244905,244912,244867,245332,
            243059,244802,245350,243873,242569,243264];
    elif hash1 eq 2565 and hash2 eq 3105 and hash3 eq 761 then
        V:=[18,1,19,5,20,9,21,19,22,25,23,33,24,37,25,45,26,55];
        D:=[1,101/3,1,0,2,203/6,1,0,3,26,1,28,1,85/3,1,30,1,0,7,313/12,1,791/30,
            1,0,9,64/3,1,343/15,1,397/15,1,0,12,56/3,1,0,13,93/5,1,1147/60,1,
            389/20,1,0,16,47/3,1,157/10,1,63/4,1,95/6,3,0,22,331/21,1,947/60,1,
            94/5,1,0];
        I:=[429513,429481,429523,429557,429504,429530,429305,429210,429472,
            429206,429128,429279,429130,429328,429228,429458,429323,429441,
            429484,429485,429531,429138,429135,429087];
    elif hash1 eq 1296 and hash2 eq 1471 and hash3 eq 751 then
        V:=[25,1,26,5,27,13,28,21,29,29];
        D:=[1,139/6,1,0,2,59/3,1,20,2,64/3,1,0,6,107/6,1,109/6,1,197/10,1,0,9,
            49/3,1,50/3,1,215/12,2,0,13,173/12,1,29/2,1,89/6,1,0];
        I:=[671011,671075,670802,670804,670450,670489,670377,670200,670190,
            670445,669917,670355,670742,669959,670740];
    elif hash1 eq 1531 and hash2 eq 2023 and hash3 eq 151 then
        V:=[18,1,19,5,20,13,22,19,23,29,24,33,25,37,26,41];
        D:=[1,39,1,0,2,151/6,1,235/6,1,79/2,1,0,5,116/5,1,119/3,2,0,8,91/5,1,
            298/15,2,358/15,1,287/12,1,0,13,479/20,1,0,14,481/20,1,0,15,75/4,1,
            0,16,94/5,1,0];
        I:=[543355,543149,543382,543259,543111,543039,543068,543174,543138,
            543139,542927,543060,542795,542656,542749,542577];
    elif hash1 eq 262 and hash2 eq 2326 and hash3 eq 448 then
        V:=[20,1,21,7,22,13,23,19,24,29,25,39,27,43,28,47];
        D:=[1,98/3,1,33,1,0,3,51/2,1,259/6,1,0,5,83/3,1,107/3,1,0,7,133/6,1,
            155/6,1,277/10,1,169/6,1,0,11,298/15,1,239/12,1,20,1,91/3,1,0,15,
            671/30,1,0,16,451/30,2,0,18,256/15,1,0];
        I:=[511024,511013,511022,511291,511178,511287,511025,511019,510944,
            510992,511021,511047,511045,511182,510942,511014,511015,511292];
    elif hash1 eq 685 and hash2 eq 1899 and hash3 eq 1813 then
        V:=[17,1,19,5,20,9,21,13,22,17,23,21,24,27,25,31,26,37,27,49];
        D:=[1,61/2,1,0,2,163/6,1,0,3,76/3,1,0,4,155/6,1,0,5,287/12,1,0,6,289/12,
            1,327/10,1,0,8,332/15,1,0,9,641/30,1,1333/60,1,0,11,394/21,1,155/8,
            1,293/15,1,81/4,1,65/3,1,0,16,603/40,1,203/10,1,0];
        I:=[542840,542842,542708,542836,542694,542766,542680,542709,542683,
            542750,542690,542592,542641,542757,542744,542564,542741];
    elif hash1 eq 3176 and hash2 eq 3165 and hash3 eq 132 then
        V:=[15,1,16,5,17,15,18,23,19,39,20,57,21,65,22,79,23,97,24,103,25,113,
            26,117];
        D:=[1,77/2,1,0,2,104/3,1,35,1,37,1,116/3,1,0,6,187/6,1,199/6,3,67/2,2,0,
            12,83/3,1,88/3,1,89/3,3,95/3,2,32,1,101/3,1,107/3,1,0,22,305/12,1,
            103/4,1,155/6,3,157/6,1,167/6,1,119/4,1,317/10,1,191/6,1,0,32,89/4,
            1,67/3,2,73/3,1,0,36,41/2,1,671/30,1,269/12,1,45/2,1,91/4,1,137/6,1,
            0,42,37/2,1,223/12,1,227/12,2,19,2,58/3,2,45/2,1,271/12,1,275/12,1,
            0,53,75/4,1,451/20,1,0,55,61/4,1,233/15,1,187/12,1,47/3,1,0,59,
            653/42,1,0,60,61/5,1,0];
        I:=[255306,255315,255317,255199,255249,255308,255187,255253,255316,
            255203,255213,255320,255189,255194,255303,255318,255208,255312,
            255241,255197,255172,255132,255136,255195,255305,255319,255144,
            255147,255070,255061,255048,255138,255047,255143,255064,255190,
            255022,255024,255105,255078,255145,255146,255127,255126,255177,
            255129,255198,255310,255314,255015,255020,255016,255011,255013,
            255063,255026,255068,255067,255014,255085];
    elif hash1 eq 1766 and hash2 eq 2042 and hash3 eq 2053 then
        V:=[25,1,26,7,27,19,28,27,29,33,30,37,32,43];
        D:=[1,179/6,1,205/6,1,0,3,358/15,1,131/5,1,335/12,1,359/12,1,103/3,1,0,
            8,45/2,1,731/30,1,149/6,1,0,11,247/12,1,313/15,2,0,14,521/30,1,0,15,
            86/5,1,263/15,1,0,17,827/60,1,83/6,1,0];
        I:=[415103,415387,415101,415195,415379,415253,415257,415136,415193,
            415180,415260,415189,415276,415105,415138,415137,415124,415126];
    elif hash1 eq 2612 and hash2 eq 3311 and hash3 eq 2046 then
        V:=[24,1,25,9,26,19,27,31,28,41];
        D:=[1,64/3,1,70/3,1,71/3,1,0,4,95/6,1,97/6,2,33/2,1,79/4,1,0,9,43/3,1,
            44/3,1,191/12,2,49/3,1,56/3,1,0,15,431/30,1,29/2,1,89/6,2,193/12,1,
            0,20,66/5,1,289/20,1,0];
        I:=[507910,507937,508195,508160,508232,508245,508252,508238,508201,
            508194,508148,508152,507779,508003,508057,508251,508236,508254,
            508156,508047,507894];
    elif hash1 eq 2330 and hash2 eq 2269 and hash3 eq 3077 then
        V:=[12,1,13,5,14,11,15,21,16,29,17,35,18,45,19,61,21,67,23,73,24,77];
        D:=[1,36,2,0,3,69/2,1,73/2,1,0,5,98/3,1,33,1,35,1,116/3,1,0,9,175/6,1,
            59/2,1,63/2,1,0,12,26,1,89/3,3,0,16,155/6,2,157/6,1,297/10,1,181/6,
            1,0,21,67/3,1,68/3,1,23,2,311/12,1,26,1,105/4,1,85/3,1,0,29,269/12,
            1,137/6,1,0,31,35/2,2,68/3,1,0,34,481/30,1,0,35,38/3,1,0];
        I:=[255778,255779,255772,255775,255770,255773,255771,255758,255759,
            255774,255756,255757,255766,255767,255769,255763,255764,255761,
            255750,255748,255762,255760,255776,255777,255753,255755,255754,
            255747,255749,255751,255765,255768,255746,255745,255752];
    elif hash1 eq 2236 and hash2 eq 3039 and hash3 eq 2275 then
        V:=[25,1,26,7,27,11,28,19,29,31,30,43,31,49,32,61];
        D:=[1,79/4,1,257/12,1,0,3,55/3,2,0,5,147/10,1,491/30,3,41/2,1,0,10,
            103/7,1,199/12,1,86/5,1,386/21,1,277/15,1,0,15,401/30,2,213/14,1,
            461/30,1,997/60,1,67/4,1,0,21,274/21,1,787/60,1,0,23,347/30,1,
            697/60,1,241/18,1,569/42,1,817/60,1,0,28,417/35,1,0];
        I:=[502893,502898,502802,503075,502982,502896,502933,503076,501553,
            501871,502922,503077,501777,502105,503033,503103,501056,501083,
            501420,501665,501569,501957,502940,503045,501054,501052,501057,
            501078];
    elif hash1 eq 1155 and hash2 eq 247 and hash3 eq 2157 then
        V:=[28,1,29,7,30,11,31,21,32,33];
        D:=[1,52/3,1,53/3,1,0,3,97/6,3,0,6,43/3,2,44/3,1,238/15,1,191/12,1,0,11,
            79/6,1,169/12,1,431/30,2,29/2,1,319/20,2,0,18,188/15,1,155/12,1,0];
        I:=[233326,232322,232090,232591,232879,232149,232972,232649,232682,
            232312,233761,232163,232006,232145,231039,231633,232028,232180,
            232111];
    elif hash1 eq 544 and hash2 eq 2065 and hash3 eq 3020 then
        V:=[23,1,24,11,25,17,26,25,27,37];
        D:=[1,39/2,1,127/6,2,139/6,1,47/2,2,0,7,18,1,65/3,1,0,9,107/6,1,121/6,1,
            41/2,1,0,12,44/3,1,49/3,2,50/3,1,91/5,1,55/3,2,0,19,431/30,1,59/4,1,
            89/6,1,0];
        I:=[414672,414646,414655,414658,414541,414660,414630,414289,414669,
            414756,414767,414268,414496,414653,414554,414285,414306,414323,
            414264,413808,414580];
    elif hash1 eq 1672 and hash2 eq 121 and hash3 eq 3412 then
        V:=[17,1,19,5,21,9,22,13,23,19,24,23,25,27,26,33,27,37];
        D:=[1,77/2,1,0,2,199/6,1,0,3,357/10,2,0,5,29,1,502/15,1,0,7,611/30,1,0,
            8,248/15,1,0,9,757/60,1,51/4,1,0,11,308/15,1,0,12,507/40,1,891/70,1,
            51/4,1,481/30,1,0];
        I:=[547142,547143,547141,547168,547057,547144,547140,547174,547146,
            547007,547092,547017,547041,547019,547145];
    elif hash1 eq 3223 and hash2 eq 2188 and hash3 eq 796 then
        V:=[24,1,25,5,26,15,27,25,28,37,29,41,30,51];
        D:=[1,31,1,0,2,127/6,1,139/6,1,47/2,1,151/6,1,0,6,59/3,1,20,1,64/3,1,
            71/3,1,0,10,95/6,2,97/6,2,71/4,1,109/6,1,121/6,2,0,18,73/4,1,0,19,
            77/6,2,431/30,1,173/12,2,101/6,1,0,25,15,1,0];
        I:=[408756,408401,408028,408757,408758,407050,408748,408012,408081,
            408335,408702,407068,408704,408677,408064,406695,407410,405490,
            408277,408308,408124,407943,408760,404668,405263];
    elif hash1 eq 2612 and hash2 eq 1157 and hash3 eq 3368 then
        V:=[22,1,23,11,24,21,25,27,26,45,27,69,28,85,29,91,30,97];
        D:=[1,65/3,1,76/3,1,109/4,1,28,1,0,5,109/6,1,237/10,1,49/2,1,157/6,1,0,
            9,62/3,1,73/3,2,0,12,91/6,1,31/2,1,101/6,3,221/12,1,75/4,1,113/6,2,
            41/2,1,125/6,2,0,24,223/15,1,15,1,46/3,1,199/12,1,203/12,1,86/5,1,
            53/3,1,277/15,1,223/12,1,56/3,1,227/12,1,0,35,83/6,1,299/20,1,
            181/12,1,709/42,1,511/30,1,205/12,1,265/14,1,0,42,203/15,4,907/60,1,
            0,47,361/30,1,73/6,1,0,49,727/60,1,0];
        I:=[513747,513746,513672,513691,513694,513703,513695,513657,513680,
            513560,513564,513662,513651,513670,513683,513686,513678,513665,
            513533,513658,513613,513449,513656,513741,513677,513652,513653,
            513395,513629,513669,513772,513769,513524,513528,513698,513450,
            513650,513398,513642,513453,513386,513623,513679,513684,513731,
            513440,513416,513393,513296];
    elif hash1 eq 920 and hash2 eq 2016 and hash3 eq 3109 then
        V:=[22,1,23,7,24,15,25,27,26,39,27,51,28,55,29,61];
        D:=[1,25,1,80/3,1,0,3,107/4,1,163/6,3,175/6,2,0,9,22,1,71/3,1,77/3,1,26,
            2,109/4,1,0,15,217/10,1,131/6,1,133/6,1,49/2,1,305/12,1,0,20,20,1,
            81/4,1,61/3,1,21,1,322/15,1,0,25,103/6,1,0,26,253/15,1,17,1,0,28,
            31/2,1,0];
        I:=[247945,248404,248134,248267,248386,248405,247827,247843,248053,
            248310,247965,247836,247882,248223,247839,248007,247845,247894,
            247797,248100,248170,247999,247964,247611,247607,247608,247625,
            247550];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 28 mod 47.
function cached_ids_28(P,hash1,hash2,hash3)
    if hash1 eq 28 and hash2 eq 1830 and hash3 eq 1231 then
        V:=[20,1,21,9,22,19,23,27,24,39,25,55,26,77,27,95,28,105];
        D:=[1,23,1,25,1,92/3,1,0,4,139/6,1,47/2,1,163/6,1,55/2,3,0,10,71/3,3,
            126/5,1,77/3,1,0,15,107/6,1,109/6,2,119/6,2,121/6,2,41/2,1,0,23,
            50/3,3,91/5,1,55/3,2,56/3,1,239/12,2,81/4,2,61/3,1,0,35,59/4,1,89/6,
            2,91/6,1,197/12,1,33/2,3,167/10,1,101/6,2,103/6,1,551/30,1,241/12,1,
            0,49,175/12,2,44/3,2,223/15,1,15,3,46/3,3,117/7,1,86/5,1,52/3,1,0,
            63,401/30,1,451/30,1,91/6,1,461/30,1,0,67,12,1,0];
        I:=[422138,421923,421947,422088,421888,421757,421871,421877,421879,
            421740,421771,421864,421430,421720,422078,421940,422054,421521,
            421948,421524,421930,421875,421874,421897,421898,421616,421482,
            421766,421203,421277,422017,422123,422128,421712,421531,421574,
            421724,422102,421480,421520,421901,421903,421734,421495,421561,
            421365,421713,421163,422023,422055,422048,422056,421469,421473,
            421945,422053,421410,421870,421895,421197,421169,421144,421186,
            421208,421209,421214,421483];
    elif hash1 eq 2848 and hash2 eq 2185 and hash3 eq 1633 then
        V:=[27,1,28,15,29,23,30,29,31,37,32,49];
        D:=[1,127/6,1,139/6,2,47/2,1,151/6,2,51/2,3,55/2,1,0,11,52/3,1,22,1,
            71/3,1,0,14,107/6,2,119/6,2,0,18,43/3,2,49/3,1,55/3,1,0,22,89/6,3,
            91/6,1,491/30,2,197/12,1,101/6,1,0,30,188/15,1,155/12,1,0];
        I:=[234027,232916,233904,232614,232137,233217,232752,233242,233618,
            233856,233573,233394,233212,232069,233495,233019,233206,232147,
            233777,232768,233556,232136,233227,233617,233196,229269,230438,
            229894,232464,232153,232498];
    elif hash1 eq 733 and hash2 eq 1048 and hash3 eq 1941 then
        V:=[23,1,24,5,25,13,26,29,27,41,28,49,29,61,30,71,31,79];
        D:=[1,187/6,1,0,2,71/3,1,30,1,94/3,2,0,6,237/10,1,145/6,3,157/6,1,
            277/10,1,111/4,1,167/6,1,169/6,1,0,15,215/12,1,55/3,1,263/12,1,67/3,
            2,68/3,1,0,21,221/12,2,113/6,1,207/10,1,0,25,329/20,1,199/12,2,17,1,
            277/15,2,223/12,1,0,32,59/4,1,451/30,2,181/12,1,50/3,1,0,37,203/15,
            1,163/12,1,14,2,0,41,361/30,1,0];
        I:=[418698,417883,418762,418664,418667,416688,417886,418597,418666,
            418761,417491,416305,418582,417614,417733,418602,416260,415363,
            416251,417360,416181,418304,418583,417811,415061,417917,418556,
            418678,416268,416286,416276,417916,416729,417524,418758,416189,
            417788,417780,418672,418740,415357];
    elif hash1 eq 3224 and hash2 eq 1456 and hash3 eq 1672 then
        V:=[19,1,20,7,21,27,22,39,23,45,24,53,25,67,26,75,28,87,29,93];
        D:=[1,51/2,1,139/4,1,0,3,65/3,2,70/3,1,71/3,1,136/5,1,95/3,1,112/3,1,38,
            1,157/4,1,204/5,1,0,13,145/6,1,155/6,1,191/6,1,65/2,1,205/6,1,0,18,
            56/3,1,166/7,1,0,20,187/10,3,127/6,1,45/2,1,0,25,46/3,1,47/3,1,
            203/12,1,17,1,227/12,2,19,1,0,32,149/10,1,451/30,1,521/30,1,0,35,
            208/15,1,522/35,1,419/28,1,652/35,1,401/21,1,0,40,253/21,1,49/4,1,0,
            42,217/18,1,849/70,1,0];
        I:=[542342,542361,542254,542336,542257,542363,542341,541874,542145,
            542335,542165,542320,542329,542238,542130,542338,542315,542354,
            541974,542253,542321,542337,542364,542118,542314,542346,542345,
            541969,541972,542352,542110,542252,541869,542326,542242,541870,
            541868,541921,541891,541925,541935,541776,541798];
    elif hash1 eq 2472 and hash2 eq 2609 and hash3 eq 3018 then
        V:=[23,1,24,15,25,21,26,25,27,29,28,33];
        D:=[1,43/2,1,107/4,1,163/6,1,175/6,1,187/6,1,211/6,1,0,7,76/3,1,83/3,1,
            0,9,217/10,1,0,10,49/3,2,0,12,221/12,1,0,13,101/7,1,217/15,1,175/12,
            1,15,1,0];
        I:=[409543,410824,412346,411863,410652,409575,411008,409652,409892,
            410059,411789,409033,409262,410224,410527,412237];
    elif hash1 eq 28 and hash2 eq 308 and hash3 eq 706 then
        V:=[26,1,29,7,30,17,31,29];
        D:=[1,101/5,1,74/3,1,0,3,27/2,1,83/6,3,181/12,1,91/6,1,0,9,66/5,1,53/4,
            2,203/15,1,41/3,2,208/15,1,0,16,677/60,1,817/60,1,583/42,2,0];
        I:=[508128,508208,508159,508221,508222,508223,508216,508217,508107,
            508176,508229,508096,508211,508215,508180,508045,507760,507886,
            507907];
    elif hash1 eq 2143 and hash2 eq 2259 and hash3 eq 3397 then
        V:=[21,1,22,7,23,15,24,19,27,27];
        D:=[1,211/6,1,71/2,1,0,3,83/3,1,89/3,1,30,1,0,6,157/6,1,0,7,328/15,1,
            67/3,2,68/3,2,0,12,451/30,1,181/12,2,91/6,1,35/2,4,0];
        I:=[252844,252843,252873,253150,252958,252892,252594,252610,252908,
            252890,252970,252839,252601,252837,252604,252526,252528,252537,
            252564];
    elif hash1 eq 3271 and hash2 eq 454 and hash3 eq 530 then
        V:=[23,1,24,7,25,17,26,29,27,45,28,63,29,69,30,81];
        D:=[1,139/6,1,151/6,2,0,4,93/4,1,71/3,1,76/3,1,77/3,1,0,8,71/4,1,109/6,
            1,121/6,1,213/10,1,145/6,1,0,13,16,1,215/12,1,91/5,1,55/3,3,56/3,1,
            101/5,1,61/3,1,0,22,431/30,1,173/12,1,491/30,2,33/2,1,67/4,1,101/6,
            1,255/14,1,551/30,1,0,31,223/15,1,15,1,0,33,391/30,1,157/12,1,79/6,
            1,401/30,1,161/12,2,0,39,173/15,1,203/15,1,0];
        I:=[418111,417939,418073,417606,418205,418502,416748,418305,418223,
            418107,415971,415490,418483,415973,416734,416517,418076,418351,
            418575,415458,415445,417453,417596,417125,417616,415880,415746,
            415808,415197,417637,415009,417563,416374,417128,415773,416996,
            415856,417092,416462,414945];
    elif hash1 eq 3083 and hash2 eq 806 and hash3 eq 2856 then
        V:=[22,1,23,17,24,31,25,47,26,67,27,79,28,91,29,97];
        D:=[1,21,1,68/3,1,23,2,25,1,80/3,3,27,2,92/3,1,0,12,115/6,5,127/6,2,
            91/4,1,139/6,3,47/2,1,51/2,1,0,25,52/3,3,53/3,1,58/3,2,59/3,1,106/5,
            3,65/3,1,24,1,0,37,185/12,1,97/6,4,33/2,1,209/12,1,177/10,1,109/6,1,
            121/6,1,41/2,1,133/6,2,0,50,57/4,1,191/12,1,16,1,49/3,1,50/3,1,0,55,
            431/30,1,29/2,3,59/4,1,89/6,3,193/12,1,0,64,155/12,1,15,1,0,66,
            117/10,1,0];
        I:=[417863,418405,418131,418135,418188,418058,418352,418459,417987,
            418454,418600,417869,417954,418326,418416,418569,417656,417862,
            417671,416674,416676,418510,417585,418043,417867,418395,418470,
            418374,416716,418485,417868,416928,416931,417636,418044,416535,
            418480,416910,416930,418016,418274,418164,416577,416802,416761,
            418561,416581,415776,416134,416710,416067,417581,418438,416783,
            416364,415719,415724,415873,417202,415710,418017,418168,415058,
            417528,415341,416777];
    elif hash1 eq 2848 and hash2 eq 1186 and hash3 eq 2544 then
        V:=[16,1,17,11,18,17,19,25,20,31,21,41,22,45];
        D:=[1,34,1,36,1,40,1,46,1,0,5,57/2,1,73/2,1,0,7,74/3,1,98/3,1,43,1,0,10,
            175/6,1,199/6,2,0,13,106/5,1,65/3,3,101/4,1,76/3,1,0,19,167/6,1,0,
            20,67/3,1,0];
        I:=[425220,425133,425236,425197,425218,425221,425215,425226,424955,
            425219,425214,425222,424925,424929,425234,425235,425230,425223,
            424933,424939];
    elif hash1 eq 263 and hash2 eq 2487 and hash3 eq 1753 then
        V:=[22,1,23,7,25,13,26,19,27,25,28,31,29,35];
        D:=[1,32,1,206/5,1,0,3,317/10,2,203/6,1,0,6,221/12,1,267/10,1,0,8,
            329/20,1,313/15,1,0,10,401/30,1,461/28,1,0,12,313/24,1,0,13,797/60,
            1,569/42,1,137/10,1,0];
        I:=[512445,511782,512332,512383,511130,512493,511220,512378,511475,
            510873,511224,510872,512460,511259,512459];
    elif hash1 eq 1391 and hash2 eq 2101 and hash3 eq 2325 then
        V:=[22,1,23,9,24,15,25,21,26,25,27,33,28,37];
        D:=[1,62/3,2,21,1,25,1,0,5,127/6,1,51/2,1,0,7,69/4,1,53/3,1,0,9,209/12,
            1,0,10,43/3,1,79/5,1,55/3,1,0,13,79/6,1,0,14,193/15,2,0];
        I:=[503541,503628,503552,503608,503614,503613,503553,503575,503200,
            503144,502192,502281,502808,503257,503302];
    elif hash1 eq 3083 and hash2 eq 2254 and hash3 eq 902 then
        V:=[22,1,23,7,24,17,25,27,26,33,27,39,28,47,29,53,30,61];
        D:=[1,95/3,1,106/3,1,0,3,95/4,1,143/6,1,317/10,1,193/6,1,0,7,79/3,1,28,
            1,141/5,1,91/3,1,0,11,221/12,1,227/10,1,0,13,223/12,1,19,1,0,15,
            339/20,1,511/30,1,241/12,1,0,18,61/4,1,103/6,1,0,20,797/60,1,137/10,
            1,85/6,1,0,23,61/5,1,0];
        I:=[510589,510681,509439,510672,509615,510669,510649,510646,509558,
            508748,510235,509742,510638,510647,508738,509753,508456,510567,
            509360,509485,510225,510680,508728];
    elif hash1 eq 2096 and hash2 eq 873 and hash3 eq 865 then
        V:=[17,1,18,11,19,31,20,47,21,61,22,83,23,103,24,117,25,123];
        D:=[1,57/2,1,61/2,2,65/2,3,69/2,1,0,8,23,1,74/3,1,25,1,80/3,1,27,6,86/3,
            3,29,3,31,1,98/3,1,0,26,127/6,2,43/2,3,139/6,1,47/2,2,151/6,3,163/6,
            2,55/2,2,0,41,59/3,6,64/3,1,65/3,5,71/3,3,76/3,1,28,1,0,58,109/6,3,
            79/4,1,119/6,1,121/6,4,41/2,1,217/10,1,87/4,5,131/6,1,133/6,2,
            257/10,1,0,78,50/3,1,215/12,2,18,1,91/5,1,55/3,2,239/12,1,62/3,1,
            263/12,2,22,2,0,91,89/6,2,491/30,1,33/2,3,101/6,1,103/6,2,45/2,1,0,
            101,179/12,1,46/3,1,0,103,95/6,1,0];
        I:=[426402,426303,426395,426396,426400,426406,426416,426403,426252,
            426295,426355,426238,426270,426305,426397,426398,426405,426263,
            426299,426353,426274,426276,426364,426098,426087,426282,426292,
            426092,426401,426404,426226,426262,426271,426033,426080,426264,
            425845,426334,425991,426045,426210,426212,426223,426266,426302,
            426306,426082,425988,426095,426218,426339,426375,425780,425993,
            426088,426041,425739,426040,426243,426251,426036,426343,426084,
            426100,426205,426370,426399,425812,425864,425986,426007,426160,
            426162,426128,425844,426049,425686,425737,426027,426047,426032,
            425837,425840,426065,425693,425764,425801,425930,425669,425823,
            426012,426099,425777,425733,426213,426289,426067,426038,426042,
            425649,425735,425842,425663];
    elif hash1 eq 1485 and hash2 eq 2838 and hash3 eq 2148 then
        V:=[25,1,26,9,27,21,28,27,29,37,30,45,31,49];
        D:=[1,127/6,1,43/2,1,139/6,2,0,5,58/3,1,59/3,2,20,2,85/4,1,65/3,1,0,12,
            177/10,1,107/6,1,0,14,44/3,1,81/5,1,49/3,3,268/15,1,0,20,29/2,1,
            89/6,2,491/30,1,0,24,155/12,1,0,25,23/2,1,0];
        I:=[671809,671879,671467,672092,671729,672174,672268,671877,671882,
            671131,671468,671222,671875,671188,671510,671141,671223,671790,
            671189,671534,671783,671946,671115,671215,671192];
    elif hash1 eq 1955 and hash2 eq 2358 and hash3 eq 2808 then
        V:=[27,1,28,9,29,17,30,23,31,29,32,41];
        D:=[1,121/6,1,87/4,1,95/4,1,0,4,49/3,1,18,1,101/5,1,0,7,89/6,1,207/10,1,
            0,9,223/15,1,199/12,1,0,11,391/30,2,79/6,1,27/2,1,85/6,1,451/30,1,0,
            17,173/15,1,0];
        I:=[403310,403597,403587,403303,403656,403448,403327,403613,403618,
            402944,403351,403678,403349,403539,403498,402996,403649];
    elif hash1 eq 3224 and hash2 eq 677 and hash3 eq 1297 then
        V:=[21,1,22,19,23,25,24,37,25,45,27,55];
        D:=[1,103/6,1,39/2,1,43/2,1,47/2,1,51/2,1,55/2,1,59/2,1,139/4,1,0,9,
            52/3,1,18,1,0,11,97/6,1,109/6,1,197/10,1,79/4,1,145/6,1,0,16,238/15,
            2,49/3,2,124/7,1,0,21,89/6,1,91/6,1,319/20,2,227/14,1,0,26,405/28,1,
            0];
        I:=[540281,540286,539987,540077,539991,540066,540095,540270,540309,
            540312,540277,539823,540263,540235,540282,540175,540250,540311,
            540313,539821,539710,539992,540233,540260,539822,539927];
    elif hash1 eq 2895 and hash2 eq 880 and hash3 eq 672 then
        V:=[24,1,25,5,26,15,27,23];
        D:=[1,68/3,1,0,2,127/6,1,43/2,1,91/4,1,51/2,2,0,7,106/5,1,65/3,2,71/3,1,
            0,11,107/6,1,197/10,1,79/4,2,121/6,2,0];
        I:=[671452,671454,671574,671364,671293,671650,671229,671197,671579,
            671117,671528,671262,671334,671382,671194,671527];
    elif hash1 eq 2284 and hash2 eq 1821 and hash3 eq 2117 then
        V:=[25,1,26,7,27,17,28,23,29,27,31,31,32,43];
        D:=[1,337/10,2,205/6,1,0,4,47/2,1,141/5,1,85/3,1,359/12,1,0,8,395/14,1,
            851/30,1,0,10,428/21,1,0,11,103/6,1,0,12,40/3,1,161/12,1,913/60,1,
            153/10,1,157/10,2,0,18,977/84,1,0];
        I:=[409685,410125,410082,411402,409691,409673,410708,408956,408975,
            408944,408903,411307,411419,409440,409515,409472,409540,409299];
    elif hash1 eq 2378 and hash2 eq 276 and hash3 eq 1777 then
        V:=[26,1,27,5,28,13,29,21,30,29];
        D:=[1,55/3,1,0,2,551/30,1,113/6,1,599/30,1,0,5,199/12,1,17,1,52/3,1,0,8,
            451/30,2,31/2,1,63/4,1,0,12,163/12,4,89/6,1,0];
        I:=[414118,414155,414171,414111,414152,413879,414201,414193,414221,
            413862,413857,413581,413971,413972,414013,413643];
    elif hash1 eq 1109 and hash2 eq 2782 and hash3 eq 342 then
        V:=[21,1,22,5,23,17,24,25,25,35,26,39,28,45,30,53];
        D:=[1,73/2,1,0,2,80/3,1,27,1,86/3,1,31,1,33,1,0,7,43/2,1,139/6,1,51/2,1,
            0,10,64/3,1,65/3,2,70/3,1,77/3,1,0,15,97/6,1,0,16,16,3,65/4,1,0,20,
            66/5,1,85/6,2,29/2,1,0,24,173/15,1,0];
        I:=[246061,246118,246235,246124,246034,245948,246163,246138,245943,
            246072,246065,246137,246054,245928,245976,245902,245965,246075,
            245942,245884,245896,246039,245944,245952];
    elif hash1 eq 733 and hash2 eq 2670 and hash3 eq 2724 then
        V:=[21,1,22,11,23,21,25,25];
        D:=[1,49/2,1,53/2,2,57/2,3,61/2,1,0,8,21,1,23,1,74/3,1,25,2,0,13,127/6,
            3,0,16,109/6,1,0];
        I:=[674074,674051,674072,673896,674046,674097,674078,674086,673900,
            673960,673964,674099,673924,674069,674070,674071];
    elif hash1 eq 3412 and hash2 eq 2284 and hash3 eq 782 then
        V:=[25,1,26,5,27,9,28,25,30,35];
        D:=[1,127/6,1,0,2,59/3,1,0,3,97/6,3,33/2,2,177/10,1,107/6,1,109/6,1,
            119/6,1,121/6,2,0,14,44/3,2,238/15,1,268/15,1,215/12,1,0,19,193/15,
            3,40/3,1,0];
        I:=[407605,408095,407540,407649,407992,407204,407500,407954,407955,
            407408,408111,407516,407708,407604,407606,407615,408077,408075,
            407582,407825,408099,407762];
    elif hash1 eq 3412 and hash2 eq 2753 and hash3 eq 241 then
        V:=[29,1,31,5,33,11,34,23,35,33];
        D:=[1,39/2,1,0,2,97/6,3,107/6,1,0,6,79/6,1,431/30,1,173/12,1,29/2,1,
            89/6,1,0,11,151/12,1,38/3,2,193/15,1,13,2,0,17,117/10,1,0];
        I:=[660580,660648,660841,661322,660729,660501,660196,661149,661031,
            660898,660659,661222,661337,659790,659386,659387,661367];
    elif hash1 eq 3412 and hash2 eq 2538 and hash3 eq 855 then
        V:=[21,1,22,7,23,17,24,21,26,29,27,37,28,45,29,49];
        D:=[1,191/6,1,449/12,1,0,3,149/5,1,359/12,1,32,1,33,1,0,7,151/6,1,0,8,
            58/3,1,337/15,1,295/12,1,0,11,1003/60,1,69/4,1,287/15,1,0,14,
            1199/70,1,725/42,1,1037/60,1,0,17,208/15,1,0,18,2537/210,1,0];
        I:=[514650,514594,514537,514627,514647,514823,514793,514736,514763,
            514113,514504,514639,514517,514058,514061,514143,514631,514068];
    elif hash1 eq 263 and hash2 eq 1681 and hash3 eq 2916 then
        V:=[16,1,17,7,18,23,19,35,20,47,21,59,22,79,23,93,24,105,25,113,27,117];
        D:=[1,28,1,34,1,0,3,57/2,1,61/2,2,65/2,1,69/2,1,73/2,1,77/2,1,81/2,1,0,
            11,74/3,1,31,1,104/3,3,110/3,1,39,1,0,18,51/2,2,55/2,1,187/6,1,
            199/6,3,67/2,1,0,26,106/5,1,65/3,1,77/3,1,28,1,106/3,1,0,31,109/6,1,
            233/12,1,119/6,1,121/6,1,131/6,1,281/12,1,237/10,1,257/10,1,111/4,1,
            0,40,81/5,1,268/15,1,18,2,239/12,1,20,1,61/3,1,0,47,491/30,1,539/30,
            1,551/30,1,20,1,611/30,1,0,52,289/20,1,218/15,1,449/20,1,0,55,
            973/60,1,0,56,119/10,2,0];
        I:=[516492,516500,516460,516488,516588,516587,516505,516485,516462,
            516427,516489,516666,516359,516415,516508,516467,516424,516477,
            516569,516268,515926,516370,516466,516520,516196,516069,516388,
            516082,516107,516032,516523,516552,516652,516048,516213,515925,
            516097,516062,515875,515942,516232,516472,516504,516139,516078,
            516583,516068,515801,515935,515796,516136,515976,515908,515686,
            515659,516299,516302];
    elif hash1 eq 1250 and hash2 eq 784 and hash3 eq 2756 then
        V:=[20,1,21,5,22,13,23,19,24,29];
        D:=[1,80/3,1,0,2,43/2,1,139/6,1,47/2,1,0,5,59/3,1,20,2,0,8,109/6,4,37/2,
            1,119/6,1,121/6,2,0,16,49/3,1,0];
        I:=[251779,251828,251803,251737,251823,251829,251830,251742,251786,
            251789,251795,251774,251731,251730,251733,251796];
    elif hash1 eq 1485 and hash2 eq 3352 and hash3 eq 103 then
        V:=[19,1,20,5,21,11,22,17,23,25,24,39,25,45,26,53,27,57,28,71,29,81];
        D:=[1,259/6,1,0,2,113/3,1,157/4,1,0,4,203/6,1,251/6,1,0,6,30,1,383/12,1,
            508/15,1,0,9,479/20,1,313/12,1,337/12,1,851/30,1,287/10,1,173/6,1,0,
            15,159/7,1,343/15,1,0,17,115/6,1,641/30,1,43/2,1,0,20,1123/60,1,0,
            21,433/28,1,653/42,1,937/60,1,481/30,1,559/30,1,75/4,1,0,27,827/60,
            2,167/12,1,185/12,1,56/3,1,0,32,171/14,1,367/30,1,0];
        I:=[514805,514615,514424,514921,514448,514803,514357,514995,514473,
            514846,514483,515004,514602,514542,514133,514128,513950,515009,
            514529,514272,514130,514115,514149,514699,513989,513996,514557,
            514762,514526,514042,513958,514125,514062];
    elif hash1 eq 545 and hash2 eq 1228 and hash3 eq 2321 then
        V:=[16,1,17,17,18,33,19,45,20,63,21,67,22,71,23,75];
        D:=[1,24,1,26,1,28,4,30,2,32,2,34,1,40,1,0,13,45/2,1,49/2,1,53/2,1,57/2,
            2,65/2,1,69/2,4,77/2,1,0,24,21,3,23,1,74/3,1,25,2,86/3,1,0,32,39/2,
            1,127/6,1,43/2,1,139/6,5,47/2,1,151/6,1,163/6,1,59/2,1,0,44,58/3,2,
            0,46,121/6,1,0,47,15,1,0,48,491/30,1,0];
        I:=[425397,425396,425293,425383,425403,425407,425267,425353,425265,
            425384,425388,425381,425393,425308,425369,425320,425374,425114,
            425286,425287,425318,425376,425080,425302,425319,425398,425390,
            425121,425113,425258,425087,425142,425268,425141,425082,425095,
            425283,425289,425300,425299,425116,425136,424947,425139,425291,
            424953,425187,424951];
    elif hash1 eq 921 and hash2 eq 1393 and hash3 eq 3452 then
        V:=[22,1,23,7,24,21,25,37,26,51,27,61,28,67];
        D:=[1,86/3,1,31,1,0,3,139/6,1,47/2,1,151/6,1,51/2,1,163/6,1,59/2,1,0,9,
            65/3,1,93/4,1,71/3,3,24,1,77/3,1,26,2,83/3,1,0,19,119/6,1,87/4,1,
            131/6,2,133/6,3,237/10,1,145/6,1,0,28,215/12,2,55/3,3,61/3,5,62/3,1,
            0,39,103/6,1,37/2,1,0,41,223/15,1,179/12,1,15,2,253/15,1,17,3,0];
        I:=[249438,249298,249384,249433,249495,248731,248912,248705,249392,
            247785,248844,249437,249464,248723,248709,247751,248711,247971,
            249388,248044,248729,248871,248700,248701,249489,247995,248040,
            248033,248129,248150,248831,249391,247773,248085,248195,249457,
            249458,247984,247585,247841,248082,248047,248713,248809,247614,
            247594,247621,247627];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 29 mod 47.
function cached_ids_29(P,hash1,hash2,hash3)
    if hash1 eq 1345 and hash2 eq 819 and hash3 eq 2577 then
        V:=[20,1,21,5,22,17,23,21,24,33,25,49,26,55,27,59,28,69];
        D:=[1,92/3,1,0,2,151/6,1,51/2,1,175/6,1,123/4,1,211/6,2,0,8,77/3,1,0,9,
            131/6,1,95/4,1,145/6,1,53/2,1,111/4,1,0,14,239/12,1,20,1,61/3,1,
            263/12,1,111/5,1,68/3,1,352/15,1,0,21,37/2,1,245/12,1,0,23,46/3,2,0,
            25,299/20,1,91/6,1,185/12,2,511/30,2,0,31,419/28,1,907/60,1,0];
        I:=[424097,424453,424334,424341,424252,423183,424530,424518,424053,
            424386,424347,424430,423489,423423,424058,423421,424346,423456,
            424338,422533,424388,424248,424417,424507,423422,424416,423499,
            423500,422557,422571,422356,422507];
    elif hash1 eq 1486 and hash2 eq 2539 and hash3 eq 1158 then
        V:=[29,1,30,9,31,15,32,23];
        D:=[1,103/6,1,35/2,1,115/6,2,0,5,16,2,69/4,1,0,8,153/10,1,95/6,2,97/6,2,
            0,13,167/12,1,14,2,43/3,1,0];
        I:=[661197,660802,661292,661400,660665,660903,660604,659899,661147,
            661304,660183,661303,660884,659807,661058,660239];
    elif hash1 eq 2755 and hash2 eq 2844 and hash3 eq 2315 then
        V:=[12,1,13,5,14,9,15,17,16,23,17,41,18,51,19,59,21,69,22,83,23,93];
        D:=[1,32,1,0,2,77/2,1,0,3,21,1,31,1,98/3,1,0,6,175/6,1,75/2,3,0,10,59/3,
            1,77/3,1,26,1,95/3,2,176/5,1,107/3,1,112/3,1,113/3,2,0,20,41/2,1,
            179/6,1,317/10,1,215/6,1,0,24,55/3,2,418/15,1,141/5,1,0,28,37/2,2,
            115/6,2,611/30,1,41/2,1,0,34,103/6,1,571/30,2,115/6,1,949/42,1,
            321/14,1,277/12,1,0,41,1027/60,2,86/5,1,69/4,1,18,2,0,47,1079/70,1,
            467/30,1,319/20,1,193/12,1,0];
        I:=[519273,519277,519297,519274,519298,519294,519247,519270,519271,
            519291,519292,519248,519286,519300,519132,519145,519299,519290,
            519296,518935,519289,519062,519128,519245,519269,519104,519137,
            519246,519281,518963,518973,518982,518936,519287,518970,518980,
            518971,518918,518893,518895,519065,519089,519061,519143,518965,
            518966,518900,518907,518903,518897];
    elif hash1 eq 2191 and hash2 eq 2341 and hash3 eq 3092 then
        V:=[20,1,21,9,22,17,23,29,24,43,25,57,26,67,28,75];
        D:=[1,21,1,27,1,31,1,0,4,39/2,1,127/6,1,47/2,2,0,8,77/4,1,59/3,3,22,1,
            93/4,1,76/3,1,0,15,107/6,3,109/6,4,233/12,1,79/4,3,119/6,1,121/6,1,
            0,28,49/3,2,215/12,2,18,1,55/3,1,298/15,1,239/12,1,0,36,491/30,1,
            217/12,3,187/10,2,113/6,1,0,43,41/3,1,361/20,1,272/15,1,0,46,143/12,
            1,0];
        I:=[424664,424131,424010,424560,424227,423961,424153,423914,424081,
            424369,424734,424165,423128,422667,423406,424585,424749,423396,
            423943,424222,424713,423509,423015,423452,424114,424000,424043,
            422896,423940,423278,424033,424244,422649,423345,423427,423419,
            422404,423035,423127,422670,422687,423580,423196,422549,422681,
            423473];
    elif hash1 eq 2426 and hash2 eq 3185 and hash3 eq 1278 then
        V:=[26,1,27,17,28,27,30,31,31,39,32,45];
        D:=[1,101/4,1,76/3,1,77/3,1,109/4,1,82/3,2,83/3,1,88/3,1,0,9,237/10,1,
            143/6,1,145/6,1,167/6,1,0,13,287/12,2,0,15,199/12,1,253/15,1,17,2,0,
            19,91/6,1,31/2,1,0,21,233/15,1,0];
        I:=[242261,243437,242378,242972,243099,244606,244283,242884,243068,
            243671,242930,243551,241976,243477,241982,242294,242292,243828,
            243443,243677,241796];
    elif hash1 eq 3225 and hash2 eq 1135 and hash3 eq 3160 then
        V:=[21,1,22,5,23,9,24,13,25,19,26,25,27,29,28,33];
        D:=[1,41/2,2,0,3,19,1,0,4,43/2,1,0,5,52/3,2,59/3,3,0,10,33/2,2,121/6,2,
            0,14,31/2,1,0,15,27/2,1,0,16,40/3,1,0];
        I:=[412408,412411,412412,411601,412397,412413,411605,411957,411967,
            411650,411666,411289,411495,411982,411356,411507];
    elif hash1 eq 593 and hash2 eq 921 and hash3 eq 201 then
        V:=[25,1,26,5,27,11,28,17,29,21,30,31];
        D:=[1,139/6,1,0,2,64/3,1,24,1,0,4,97/6,1,107/6,1,0,6,191/12,1,0,7,
            143/10,1,173/12,2,29/2,2,89/6,3,0,15,193/15,1,0];
        I:=[241629,238553,238541,241069,241078,238715,239478,240019,241082,
            240548,241640,238516,240953,241134,239748];
    elif hash1 eq 1298 and hash2 eq 1484 and hash3 eq 423 then
        V:=[14,1,15,5,16,11,17,15,18,21,19,25,20,33,21,37,24,43];
        D:=[1,38,1,0,2,65/2,1,81/2,1,0,4,122/3,1,0,5,175/6,1,199/6,2,0,8,133/4,
            1,0,9,131/6,1,305/12,1,277/10,1,0,12,407/12,1,0,13,245/12,1,45/2,1,
            0,15,907/60,1,1243/60,1,0];
        I:=[428832,428862,428829,428844,428672,428834,428840,428839,428843,
            428853,428666,428621,428667,428684,428585,428584];
    elif hash1 eq 3084 and hash2 eq 1072 and hash3 eq 2753 then
        V:=[22,1,23,5,24,9,25,25,26,41,27,53,28,57];
        D:=[1,71/3,1,0,2,133/6,1,0,3,61/3,1,62/3,1,263/12,1,111/5,1,67/3,1,73/3,
            1,129/5,1,0,10,37/2,1,113/6,1,241/12,1,611/30,1,245/12,2,45/2,1,
            91/4,1,0,18,253/15,1,17,2,69/4,1,223/12,1,227/12,1,0,24,339/20,1,0,
            25,203/15,1,41/3,1,14,3,61/4,1,187/12,1,0];
        I:=[421019,420966,421021,421018,420990,420980,420964,420971,421002,
            420972,420959,420970,420996,420967,420969,420950,420947,421009,
            421010,421011,420982,421006,421008,420936,420984,421013,420961,
            420963,420968,420940,420939];
    elif hash1 eq 2191 and hash2 eq 11 and hash3 eq 1744 then
        V:=[25,1,26,5,27,15,28,21,29,27];
        D:=[1,151/6,2,0,3,20,1,65/3,2,22,1,24,2,0,9,109/6,2,121/6,2,0,13,49/3,1,
            55/3,1,0,15,29/2,1,33/2,1,0];
        I:=[237568,237616,237615,237576,237657,237899,237558,237901,237644,
            237652,237708,237738,237638,237559,237648,237079];
    elif hash1 eq 1674 and hash2 eq 3217 and hash3 eq 2339 then
        V:=[24,1,25,9,26,19,27,33,28,41,29,45];
        D:=[1,18,1,59/3,1,85/4,1,0,4,109/6,2,197/10,1,119/6,2,143/6,1,0,10,49/3,
            1,50/3,1,268/15,1,292/15,1,20,1,101/5,1,0,16,197/12,2,33/2,1,37/2,1,
            0,20,223/15,1,0,21,401/30,6,83/6,2,0];
        I:=[417303,416886,416858,416876,417628,417277,416856,417283,415807,
            417210,416884,417050,417012,415842,415151,417107,417209,415852,
            415945,415375,415528,416595,416859,417002,417208,417394,415543,
            417169];
    elif hash1 eq 123 and hash2 eq 1330 and hash3 eq 2463 then
        V:=[22,1,23,5,24,15,25,29,26,35,27,39,28,55,29,73];
        D:=[1,31,1,0,2,163/6,6,55/2,1,175/6,3,59/2,2,0,14,65/3,1,70/3,1,71/3,1,
            24,1,101/4,2,83/3,1,0,21,197/10,1,121/6,2,0,24,49/3,2,0,26,89/6,1,
            197/12,1,33/2,1,103/6,3,551/30,1,221/12,2,75/4,1,0,36,40/3,1,85/6,1,
            175/12,1,44/3,1,179/12,1,15,2,46/3,2,199/12,2,0,47,391/30,1,0];
        I:=[249604,248735,248827,248828,249499,249514,249531,249376,248412,
            249475,249530,249573,249620,248775,248745,249381,249423,247676,
            248042,248030,248093,247771,247932,248825,249255,248032,247599,
            248760,248414,248897,249633,248076,247832,247834,247646,247879,
            248396,247931,247808,247800,247831,249598,248874,249323,247569,
            247645,248015];
    elif hash1 eq 687 and hash2 eq 1047 and hash3 eq 1980 then
        V:=[25,1,26,9,29,15,31,19];
        D:=[1,179/6,1,337/10,2,205/6,3,0,7,141/5,1,85/3,1,0,9,169/10,1,0,10,
            83/6,2,913/60,1,61/4,1,46/3,1,157/10,2,95/6,1,0];
        I:=[411635,409937,411184,409696,410079,410106,409942,410101,411364,
            409992,411621,409370,409442,409524,409938,410089,409539];
    elif hash1 eq 2755 and hash2 eq 1004 and hash3 eq 824 then
        V:=[20,1,21,7,22,11,23,15,24,21,26,25,27,29,28,33,29,39];
        D:=[1,86/3,1,110/3,1,0,3,175/6,2,0,5,109/4,1,0,6,257/10,2,155/6,1,0,9,
            215/12,1,0,10,248/15,1,0,11,625/42,1,0,12,274/21,1,787/60,2,0,15,
            117/10,1,0];
        I:=[419901,420038,419359,419470,420220,419202,419274,419282,420169,
            419475,418939,419005,418995,419347,419368];
    elif hash1 eq 2144 and hash2 eq 1011 and hash3 eq 1706 then
        V:=[21,1,22,5,23,11,24,21,25,29];
        D:=[1,41/2,3,0,4,19,1,62/3,2,0,7,75/4,1,115/6,1,39/2,1,127/6,1,0,11,
            69/4,1,52/3,1,53/3,1,0,14,33/2,1,107/6,1,0];
        I:=[418925,418926,418927,418924,418812,418837,418723,418729,418720,
            418547,418429,418891,417728,418655,416279];
    elif hash1 eq 1815 and hash2 eq 2208 and hash3 eq 3159 then
        V:=[21,1,22,5,23,9,24,13,26,19,28,23,29,27,30,33];
        D:=[1,259/6,1,0,2,95/3,3,0,5,179/6,1,0,6,24,1,73/3,3,0,10,21,1,0,11,
            69/4,3,0,14,361/30,1,145/12,1,0,16,176/15,1,707/60,4,0];
        I:=[509369,509395,509417,509430,509401,509396,509381,509399,509403,
            509295,509258,509260,509393,509385,509377,509347,509380,509382,
            509384,509402];
    elif hash1 eq 2050 and hash2 eq 917 and hash3 eq 1875 then
        V:=[21,1,22,7,23,17,24,25,25,33,26,39];
        D:=[1,45/2,5,49/2,3,0,9,19,2,62/3,2,21,5,74/3,1,0,19,115/6,3,39/2,6,
            127/6,4,0,32,52/3,1,53/3,2,18,1,0,36,97/6,5,33/2,1,0,42,43/3,4,0];
        I:=[673738,673739,673740,673741,673749,673585,673643,673684,673619,
            673629,673586,673660,673590,673614,673641,673691,673730,673453,
            673412,673582,673597,673419,673452,673601,673604,673742,673743,
            673307,673387,673411,673539,673443,673434,673444,673409,673265,
            673302,673644,673652,673718,673278,673360,673372,673438,673449];
    elif hash1 eq 546 and hash2 eq 555 and hash3 eq 1894 then
        V:=[23,1,24,11,25,21,26,31,27,47,28,65,29,73,30,85];
        D:=[1,127/6,2,139/6,2,151/6,2,163/6,1,0,8,53/3,1,58/3,1,71/3,2,24,1,0,
            13,107/6,1,109/6,2,79/4,1,121/6,2,0,19,50/3,2,268/15,1,55/3,1,56/3,
            1,239/12,1,61/3,1,62/3,2,0,28,431/30,1,89/6,1,91/6,1,491/30,1,
            539/30,1,221/12,1,37/2,2,113/6,1,0,37,41/3,1,253/15,1,19,1,0,40,
            259/20,1,401/30,5,27/2,1,83/6,3,299/20,1,0,51,173/15,1,0];
        I:=[416337,417710,417152,417280,416932,417721,417707,417288,417655,
            416967,417211,417260,416385,416889,417259,417234,417191,417722,
            416315,417279,416717,416319,417511,417222,417601,416579,416971,
            417623,417114,417110,417593,416063,417367,415681,416415,416108,
            417178,416467,415049,415371,415329,416809,416960,417203,417206,
            416852,416505,416962,417352,415492,415052];
    elif hash1 eq 1298 and hash2 eq 2842 and hash3 eq 1290 then
        V:=[26,1,27,7,28,15,29,27,30,31,31,35];
        D:=[1,85/4,1,101/4,1,0,3,71/4,2,79/4,1,119/6,2,0,8,49/3,1,89/5,1,268/15,
            1,239/12,1,61/3,1,0,13,479/30,1,0,14,218/15,1,0,15,259/20,1,27/2,1,
            0];
        I:=[243851,243173,243734,243909,242686,242639,243780,242642,243319,
            243844,241894,241809,243815,243876,241991,243342];
    elif hash1 eq 969 and hash2 eq 2506 and hash3 eq 3087 then
        V:=[29,1,30,15,31,19,32,25];
        D:=[1,95/6,1,97/6,3,209/12,1,109/6,1,79/4,1,119/6,1,0,9,50/3,1,0,10,
            143/10,1,173/12,1,0,12,193/15,2,40/3,2,0];
        I:=[665608,665611,665615,665625,665685,666205,665164,666079,666207,
            665727,665859,665609,665610,664987,665598];
    elif hash1 eq 1815 and hash2 eq 1405 and hash3 eq 2576 then
        V:=[25,1,26,5,27,13,28,21,29,33];
        D:=[1,43/2,1,0,2,64/3,1,65/3,1,71/3,1,0,5,33/2,1,109/6,1,121/6,1,0,8,
            65/4,1,49/3,1,50/3,1,17,1,55/3,1,0,13,173/12,1,29/2,2,0];
        I:=[237622,237666,237612,237197,237661,237651,237815,237189,237135,
            237754,237879,237425,237420,237198,237463];
    elif hash1 eq 2990 and hash2 eq 2388 and hash3 eq 2479 then
        V:=[19,1,20,7,21,15,22,31,23,47,24,63,25,73,27,95,28,105];
        D:=[1,187/6,1,223/6,1,0,3,88/3,1,95/3,1,113/3,1,0,6,155/6,1,167/6,1,
            353/12,1,297/10,2,181/6,1,317/10,1,193/6,1,0,14,111/5,1,287/12,1,
            73/3,1,388/15,1,105/4,1,80/3,1,113/4,1,0,21,245/12,1,207/10,1,125/6,
            1,265/12,1,227/10,1,731/30,1,149/6,1,0,28,223/12,1,19,1,251/12,1,21,
            1,0,32,511/30,1,205/12,2,521/30,2,37/2,1,781/42,1,379/20,1,569/30,1,
            571/30,3,581/30,1,265/12,1,0,46,817/60,1,137/10,1,55/4,1,185/12,1,0,
            50,707/60,2,188/15,1,329/24,2,206/15,1,0];
        I:=[517508,517481,517491,517482,517243,517237,517507,517455,517179,
            517288,517490,517255,517347,517366,517466,517276,517159,517488,
            516995,517160,517346,517263,517352,517211,517187,517145,517518,
            517184,517326,517364,516985,517173,517380,517392,517303,517365,
            517044,517030,517041,517123,517147,517177,517317,517348,516994,
            517311,517218,517310,517073,517246,517336,517024,516988,516996,
            516990];
    elif hash1 eq 2708 and hash2 eq 2288 and hash3 eq 2974 then
        V:=[17,1,18,5,20,9,21,15,22,27,23,33,24,41,25,47,27,51];
        D:=[1,73/2,1,0,2,134/3,1,0,3,71/3,1,95/3,1,0,5,157/6,1,317/10,1,337/10,
            1,203/6,1,205/6,1,0,10,89/4,1,151/5,1,0,12,551/30,1,269/12,1,367/14,
            1,0,15,283/15,1,307/15,1,0,17,793/42,1,0,18,939/70,1,583/42,1,
            421/30,1,157/10,1,0];
        I:=[516131,516017,516067,516014,516024,516015,516324,516025,515753,
            516260,515773,516073,516226,515757,516339,516283,515764,515695,
            515760,516327,515759];
    elif hash1 eq 2755 and hash2 eq 1466 and hash3 eq 1368 then
        V:=[23,1,24,11,25,19,26,23,27,29,28,35,29,43];
        D:=[1,151/6,1,163/6,2,55/2,2,187/6,1,0,7,20,1,65/3,1,71/3,1,0,10,121/6,
            1,0,11,49/3,2,56/3,2,0,15,33/2,2,103/6,1,0,18,40/3,1,175/12,3,
            223/15,1,0,23,391/30,2,0];
        I:=[249562,248982,249494,248989,249520,248083,249273,248940,248896,
            247766,248985,249568,249561,249569,248678,248820,247858,247765,
            247799,247865,248758,247825,248075,248078];
    elif hash1 eq 2520 and hash2 eq 2693 and hash3 eq 1898 then
        V:=[24,1,25,5,27,13,28,19,29,29,30,35,31,43];
        D:=[1,116/5,1,0,2,233/12,1,297/10,1,191/6,1,0,5,311/20,1,41/2,1,0,7,
            40/3,1,76/5,1,248/15,1,369/20,1,0,11,79/6,1,401/30,1,0,13,261/20,1,
            66/5,1,53/4,2,0,17,377/28,1,0];
        I:=[506362,506096,504922,504892,505337,504899,506257,506275,506341,
            505077,505672,506097,505797,505677,505432,505674,503978];
    elif hash1 eq 2332 and hash2 eq 2243 and hash3 eq 3354 then
        V:=[21,1,22,5,23,9,24,15,25,23,26,31,27,41,28,45,29,49];
        D:=[1,57/2,1,0,2,80/3,1,0,3,139/6,3,151/6,2,0,8,65/3,1,71/3,1,77/3,1,0,
            11,233/12,1,119/6,2,131/6,1,0,15,191/12,1,292/15,1,298/15,1,61/3,1,
            0,19,33/2,1,0,20,25/2,1,0,21,91/6,1,0];
        I:=[418128,418119,418105,418132,418220,418126,418243,418213,418176,
            418241,417233,418175,418212,418121,418137,415839,415775,417238,
            418198,415789,415287];
    elif hash1 eq 264 and hash2 eq 1236 and hash3 eq 1601 then
        V:=[22,1,23,7,24,19,25,27,26,37];
        D:=[1,19,1,62/3,1,0,3,103/6,1,35/2,2,115/6,3,127/6,1,51/2,1,0,11,52/3,2,
            53/3,1,77/4,1,0,15,157/10,1,97/6,1,109/6,1,233/12,1,0,19,191/12,2,
            55/3,1,19,1,0];
        I:=[411761,411756,412382,411949,412159,410651,412407,412410,411217,
            410379,412167,412385,412416,411606,410690,410433,410740,411528,
            410644,412174,409337,411521];
    elif hash1 eq 1909 and hash2 eq 1649 and hash3 eq 2464 then
        V:=[19,1,20,5,21,11,22,15,23,23,24,31,25,35,26,41,27,45,29,49];
        D:=[1,61/2,1,0,2,80/3,1,98/3,3,0,6,199/6,2,0,8,126/5,1,76/3,1,77/3,1,0,
            11,131/6,1,95/4,1,143/6,2,0,15,101/5,1,0,16,18,1,113/6,2,0,19,
            278/15,2,0,21,44/3,4,0,25,117/10,1,0];
        I:=[253931,253958,253611,253728,253873,253533,253972,253573,253843,
            253520,253927,253519,253531,253755,253578,253498,253532,253729,
            253465,253534,253457,253517,253565,253704,253501];
    elif hash1 eq 2144 and hash2 eq 810 and hash3 eq 877 then
        V:=[21,1,22,7,23,17,25,21];
        D:=[1,41/2,2,49/2,1,0,4,56/3,2,19,3,62/3,2,21,2,0,13,103/6,1,0,14,
            157/10,1,97/6,2,0];
        I:=[251457,251501,251411,251519,251590,251520,251522,251530,250808,
            251342,250837,251023,251560,250196,250216,250401];
    elif hash1 eq 593 and hash2 eq 2970 and hash3 eq 252 then
        V:=[18,1,19,5,20,13,21,25,22,43,23,55,24,61,25,69,26,87,28,99];
        D:=[1,110/3,1,0,2,199/6,1,67/2,1,211/6,1,0,5,89/3,1,166/5,1,101/3,2,34,
            1,113/3,1,0,11,305/12,1,155/6,2,157/6,1,167/6,2,297/10,1,119/4,1,
            179/6,1,127/4,1,0,21,121/5,1,73/3,2,311/12,1,26,1,79/3,1,0,27,
            671/30,1,269/12,1,0,29,37/2,1,227/12,1,19,2,0,33,50/3,1,521/30,1,
            56/3,1,569/30,1,571/30,2,229/12,2,581/30,1,233/12,1,0,43,73/5,1,
            883/60,2,59/4,1,61/4,1,47/3,1,0,49,707/60,1,0];
        I:=[427415,426992,427394,426986,427195,426772,426826,427361,427378,
            426854,426671,426845,427517,426838,427173,427400,426853,426485,
            427404,426877,426816,426851,426858,426843,427134,426796,426528,
            426527,426667,426665,426661,427211,426531,426818,426457,426464,
            426519,426870,426871,426872,426869,426875,426805,426582,426617,
            426587,426813,427438,426689];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 30 mod 47.
function cached_ids_30(P,hash1,hash2,hash3)
    if hash1 eq 1816 and hash2 eq 3199 and hash3 eq 419 then
        V:=[19,1,20,5,21,13,23,19,24,25,25,29,26,35,27,51];
        D:=[1,71/2,1,0,2,88/3,1,34,1,113/3,1,0,5,49/2,1,169/6,1,0,7,101/6,1,
            217/12,1,0,9,199/12,1,0,10,95/6,1,511/30,1,0,12,203/15,1,74/5,1,
            77/5,1,233/15,1,83/5,1,467/28,1,101/6,1,0,19,179/12,1,0];
        I:=[516602,516399,516320,515891,516411,516180,515952,516209,516304,
            516447,516297,515947,516176,515899,516006,515756,515778,515708,
            515653];
    elif hash1 eq 453 and hash2 eq 2497 and hash3 eq 1305 then
        V:=[20,1,21,11,22,17,23,25,25,31];
        D:=[1,22,1,24,2,26,1,28,2,0,7,41/2,1,45/2,5,0,13,56/3,2,19,1,21,1,0,17,
            103/6,1,43/2,1,0,19,95/6,1,0];
        I:=[251596,251470,251485,251469,251473,251574,251541,251234,251459,
            251467,251486,251487,251480,251586,251492,251479,251591,250379,
            251523];
    elif hash1 eq 2333 and hash2 eq 2062 and hash3 eq 1656 then
        V:=[21,1,22,7,23,13,24,25,25,29];
        D:=[1,53/2,1,65/2,3,0,5,74/3,1,80/3,1,0,7,115/6,1,127/6,2,43/2,1,139/6,
            1,107/4,1,0,13,58/3,1,0,14,79/4,1,41/2,1,0];
        I:=[505423,505143,505415,505630,505057,505412,505507,505465,505482,
            505136,505148,505493,505348,505356,505156];
    elif hash1 eq 2615 and hash2 eq 2194 and hash3 eq 120 then
        V:=[19,1,20,5,21,11,22,21,23,33,24,43,25,47,27,53,28,57];
        D:=[1,65/2,1,0,2,86/3,1,110/3,1,0,4,47/2,1,187/6,1,63/2,1,71/2,1,0,8,26,
            1,109/4,1,89/3,1,30,1,101/3,1,0,13,133/6,1,155/6,1,157/6,1,277/10,1,
            0,17,67/3,1,0,18,551/30,1,245/12,1,0,20,511/30,1,0,21,907/60,1,
            233/15,1,0];
        I:=[419866,419426,419867,419822,419861,419864,419738,419741,419263,
            419821,419835,419737,419465,419816,419163,419539,419802,419686,
            419291,419173,419035,419037];
    elif hash1 eq 2098 and hash2 eq 2734 and hash3 eq 1700 then
        V:=[20,1,21,7,22,17,23,25,24,33,25,43];
        D:=[1,25,1,27,1,0,3,127/6,1,139/6,1,151/6,3,51/2,1,0,9,59/3,1,64/3,1,
            65/3,1,0,12,107/6,1,119/6,1,131/6,1,0,15,215/12,1,91/5,1,55/3,2,19,
            2,0,21,113/6,2,0];
        I:=[253314,253335,253339,253329,252795,253031,253299,252756,253034,
            252979,252759,253337,253333,252520,252696,252804,252810,253047,
            252809,252995,252502,252748];
    elif hash1 eq 2098 and hash2 eq 3308 and hash3 eq 2495 then
        V:=[20,1,22,5,23,9,24,17,25,23,26,27,27,33,29,39,30,45];
        D:=[1,104/3,1,0,2,83/3,1,0,3,277/10,1,119/4,1,179/6,1,0,6,287/12,1,73/3,
            1,0,8,611/30,1,0,9,401/20,1,428/21,1,0,11,511/30,1,521/30,1,0,13,
            377/28,1,817/60,1,0,15,485/36,1,0];
        I:=[509028,508991,509056,508985,508925,509015,508995,509000,509146,
            508668,508988,508912,508605,508464,508316];
    elif hash1 eq 1628 and hash2 eq 1698 and hash3 eq 3256 then
        V:=[22,1,23,5,24,11,25,17,26,23,27,27,28,35,29,39,30,43];
        D:=[1,27,1,0,2,139/6,1,59/2,1,0,4,65/3,1,77/3,1,0,6,133/6,1,103/4,1,0,8,
            167/12,1,0,9,75/4,1,113/6,1,659/30,1,0,12,155/12,1,0,13,51/4,2,0,15,
            667/60,1,0];
        I:=[502385,502175,502299,502349,502384,502358,501603,502172,502512,
            502423,501102,502127,502348,502627,502371];
    elif hash1 eq 3085 and hash2 eq 1751 and hash3 eq 1061 then
        V:=[27,1,28,9,29,19,30,27,32,35];
        D:=[1,35/2,1,115/6,3,39/2,2,0,7,16,1,69/4,1,52/3,1,53/3,3,0,13,157/10,1,
            95/6,3,97/6,1,0,18,167/12,1,71/5,1,43/3,1,0,21,34/3,2,0];
        I:=[667124,666949,667144,667356,665577,667374,667160,665555,667105,
            665829,667127,667154,665786,665778,667077,667318,665754,665815,
            665518,665801,665592,665909];
    elif hash1 eq 1628 and hash2 eq 1714 and hash3 eq 910 then
        V:=[17,1,18,7,19,13,20,21,21,35,22,39,23,47,24,51,25,57,26,61];
        D:=[1,57/2,1,77/2,1,0,3,33,1,110/3,1,0,5,175/6,1,187/6,1,211/6,1,0,8,22,
            1,101/4,1,83/3,1,89/3,1,30,1,107/3,1,0,14,155/6,1,0,15,268/15,1,
            263/12,2,68/3,2,0,20,221/12,1,0,21,271/12,1,275/12,1,0,23,181/12,1,
            0,24,309/20,1,233/15,1,0];
        I:=[427246,427265,427244,427266,427323,427257,426956,427268,427242,
            427270,426957,426954,426916,426934,426949,426932,426933,426766,
            426959,426931,426441,426440,427252,426443,426430];
    elif hash1 eq 1017 and hash2 eq 992 and hash3 eq 2342 then
        V:=[18,1,19,5,20,15,21,23,22,29,23,37,24,47,25,53,26,57];
        D:=[1,26,1,0,2,49/2,1,53/2,1,57/2,2,73/2,1,0,7,80/3,1,86/3,1,33,1,0,10,
            39/2,1,47/2,1,0,12,59/3,1,22,1,71/3,1,0,15,63/4,2,97/6,1,133/6,1,
            145/6,1,0,20,191/12,1,239/12,1,0,22,173/12,2,0,24,289/20,1,0];
        I:=[512673,512831,512543,512828,512834,512694,512806,512697,512699,
            512840,512709,512681,512779,512544,512539,512788,512829,512398,
            511334,512785,511388,512546,512710,511963];
    elif hash1 eq 2098 and hash2 eq 3243 and hash3 eq 2426 then
        V:=[16,1,17,7,18,15,19,21,20,29,21,35,23,41];
        D:=[1,32,1,34,1,0,3,61/2,2,65/2,1,69/2,1,0,7,25,1,29,1,0,9,127/6,2,43/2,
            1,55/2,1,0,13,58/3,1,26,1,0,15,133/6,1,45/2,1,0,17,33/2,1,0];
        I:=[427481,254757,427379,427484,427386,427480,427487,427507,427431,
            427502,427444,426598,427492,426610,426581,426603,427381];
    elif hash1 eq 1863 and hash2 eq 28 and hash3 eq 3298 then
        V:=[21,1,22,5,23,13,24,27,25,43,26,57,27,65,28,73,29,93,30,99];
        D:=[1,139/6,1,0,2,101/3,1,106/3,1,107/3,1,0,5,157/6,1,329/12,1,277/10,1,
            169/6,1,179/6,1,181/6,1,0,11,73/3,1,388/15,1,311/12,1,141/5,1,86/3,
            1,151/5,1,91/3,2,0,19,359/20,1,221/12,1,241/12,1,311/14,1,671/30,2,
            137/6,1,0,26,329/20,1,283/15,1,313/15,1,0,29,31/2,2,521/30,1,631/30,
            1,0,33,76/5,1,61/4,1,323/21,1,187/12,1,238/15,1,16,1,1027/60,2,
            263/15,2,56/3,1,0,44,797/60,2,41/3,1,0,47,253/21,1,61/5,1,0];
        I:=[510227,510047,509688,509605,509931,509584,509979,510076,509754,
            509912,509526,510095,509701,509341,509610,509942,509617,509927,
            510233,510217,509502,508566,509339,509928,508742,509512,509469,
            509900,509892,509955,509365,510099,509939,509639,508593,508857,
            509263,509687,509131,509500,509511,509953,508455,509503,509513,
            509729,508725,508737];
    elif hash1 eq 3367 and hash2 eq 872 and hash3 eq 995 then
        V:=[20,1,21,5,22,15,23,23,24,33,25,37,26,47,27,51];
        D:=[1,32,1,0,2,49/2,1,53/2,1,57/2,1,65/2,1,0,6,21,1,68/3,1,25,1,0,9,
            115/6,2,39/2,1,43/2,1,139/6,1,0,14,47/3,1,0,15,185/12,1,63/4,1,
            233/12,1,119/6,1,0,19,43/3,1,0,20,279/20,1,169/12,1,173/12,1,0];
        I:=[411745,411715,412292,411765,412225,412049,412095,411951,412085,
            412154,412249,412218,412204,411744,412023,412129,410044,409659,
            412286,411099,412164,410656];
    elif hash1 eq 1111 and hash2 eq 1750 and hash3 eq 47 then
        V:=[23,1,24,7,25,15,27,21,28,25];
        D:=[1,43/2,1,139/6,2,0,4,59/3,1,20,1,65/3,1,0,7,121/6,1,41/2,3,0,11,
            29/2,1,0,12,193/15,1,13,3,0];
        I:=[413792,413779,413795,413787,413836,413849,413758,413733,413746,
            413791,413878,413719,413763,413777,413881];
    elif hash1 eq 1205 and hash2 eq 2024 and hash3 eq 75 then
        V:=[21,1,22,5,23,15,24,23,25,35,26,41];
        D:=[1,53/2,1,0,2,19,4,62/3,3,21,4,74/3,1,0,14,35/2,2,115/6,4,39/2,2,0,
            22,69/4,1,53/3,6,18,2,20,1,85/4,1,0,33,95/6,4,97/6,4,0,41,15,1,0];
        I:=[249694,249540,249586,249623,249815,248694,248936,249673,248782,
            249609,249622,249653,249372,248858,249548,249275,249277,249648,
            249649,248777,248814,249353,247790,248738,248769,248846,249020,
            249559,248006,249045,249257,248288,247856,247942,249529,249625,
            247952,248021,248849,248857,247778];
    elif hash1 eq 2803 and hash2 eq 320 and hash3 eq 292 then
        V:=[14,1,15,17,16,27,17,39,18,55,19,71,20,81,21,95,22,109,23,115,24,
            127];
        D:=[1,26,1,28,1,30,1,32,2,34,1,36,1,38,1,0,9,41/2,1,57/2,1,69/2,1,81/2,
            1,0,13,23,1,86/3,1,92/3,1,110/3,1,37,1,0,18,115/6,1,99/4,1,151/6,1,
            59/2,1,199/6,1,211/6,1,223/6,2,0,26,20,1,65/3,1,93/4,1,82/3,2,117/4,
            1,88/3,1,32,1,0,34,109/6,1,131/6,1,95/4,2,111/4,1,0,39,18,2,56/3,1,
            298/15,2,20,1,121/5,1,74/3,1,0,47,167/10,1,101/6,3,611/30,1,659/30,
            1,269/12,1,45/2,1,0,55,53/3,2,157/7,1,0,58,461/30,1,997/60,2,511/30,
            1,205/12,1,35/2,2,0,65,522/35,1,309/20,1,233/15,1,0];
        I:=[518148,518146,518147,518136,518150,518145,518133,518139,518159,
            518153,518144,518134,518154,518170,518152,518135,518112,518157,
            518059,518024,518062,518140,518166,518052,518105,517782,518143,
            518055,517796,518138,518001,518165,518074,518094,518155,518034,
            518058,517794,518064,518158,517775,517772,517928,517773,517882,
            518075,517977,517983,518050,518115,517814,517853,517953,517800,
            517798,517799,517738,517877,517848,517927,517881,517873,517771,
            517797,517750,517751,517758];
    elif hash1 eq 1346 and hash2 eq 1114 and hash3 eq 2717 then
        V:=[23,1,24,5,25,17,26,29,27,45,28,49];
        D:=[1,115/6,4,0,5,69/4,1,52/3,1,53/3,1,77/4,1,59/3,6,0,15,97/6,2,33/2,2,
            173/10,1,209/12,3,121/6,1,0,24,15,1,79/5,1,191/12,1,49/3,2,50/3,1,
            262/15,1,91/5,2,0,33,89/6,1,0,34,40/3,1,0];
        I:=[418860,418869,418905,418914,417880,418911,418842,418727,416779,
            418318,418566,418873,418882,418909,418717,418718,417926,417956,
            416361,417731,418508,418625,417808,416693,417568,418397,415957,
            418555,418567,417562,416280,416582,418683,416506];
    elif hash1 eq 2991 and hash2 eq 1405 and hash3 eq 1193 then
        V:=[22,1,23,5,24,11,25,15,26,19,27,29,28,35,29,43,30,53];
        D:=[1,83/3,1,0,2,155/6,1,179/6,1,0,4,105/4,1,0,5,41/2,1,0,6,283/15,1,
            227/12,1,251/12,1,271/12,1,0,10,205/12,1,521/30,1,0,12,233/15,2,
            47/3,1,1027/60,1,0,16,137/10,1,83/6,1,373/24,1,467/30,1,0,20,727/60,
            1,0];
        I:=[421585,421438,421627,421629,421378,421426,421628,421236,421274,
            421385,421527,421376,421409,421352,421199,421503,421384,421116,
            421114,421185];
    elif hash1 eq 2051 and hash2 eq 1356 and hash3 eq 956 then
        V:=[23,1,24,7,26,11,27,15,28,21,29,27,31,31,32,37];
        D:=[1,377/12,1,373/10,1,0,3,137/4,1,0,4,159/7,1,0,5,67/4,1,56/3,1,0,7,
            77/5,1,87/5,1,0,9,751/42,1,0,10,107/10,1,12,1,0,12,218/21,1,
            1111/105,1,53/5,2,893/84,1,0];
        I:=[539977,540224,540203,539885,539891,539892,540268,540269,539893,
            539827,539901,539983,539978,540171,540190,539928];
    elif hash1 eq 2850 and hash2 eq 3208 and hash3 eq 2029 then
        V:=[21,1,22,11,23,19,24,31,25,37,26,43,27,53,28,59,29,65];
        D:=[1,57/2,1,61/2,1,65/2,1,73/2,1,0,5,86/3,2,29,1,31,1,0,9,139/6,1,51/2,
            1,163/6,3,175/6,1,187/6,1,0,16,71/3,1,28,1,0,18,121/6,2,257/12,1,0,
            21,191/12,1,215/12,1,55/3,2,62/3,1,0,26,101/6,1,221/12,1,0,28,85/6,
            1,175/12,1,0,30,91/6,1,0];
        I:=[417898,418003,417977,418008,417999,418002,417899,417981,417990,
            418013,416768,418001,418009,417992,418012,417984,417150,416602,
            417329,416941,418054,416978,417900,418011,416936,418026,415688,
            416803,416955,415156];
    elif hash1 eq 2850 and hash2 eq 365 and hash3 eq 1119 then
        V:=[20,1,21,9,22,15,23,21,24,25,25,35,26,41,27,53];
        D:=[1,31,1,104/3,1,37,1,0,4,99/4,1,199/6,1,0,6,59/3,1,95/3,1,0,8,233/12,
            1,0,9,50/3,1,215/12,1,111/5,1,358/15,1,0,13,37/2,1,245/12,1,0,15,14,
            1,281/20,1,223/15,1,179/12,1,86/5,1,0,20,157/12,1,33/2,1,67/4,1,0];
        I:=[512818,512815,512582,512224,512533,511913,512038,512339,512278,
            511877,511505,511480,511525,512288,511535,511453,511411,511523,
            511445,511669,511243,511092];
    elif hash1 eq 124 and hash2 eq 2843 and hash3 eq 3357 then
        V:=[20,1,21,9,22,17,24,25,26,37,27,49];
        D:=[1,74/3,1,31,2,33,3,0,7,151/6,2,51/2,1,163/6,1,0,11,22,1,70/3,1,71/3,
            1,0,14,43/3,3,44/3,1,268/15,2,62/3,1,21,1,0,22,66/5,1,40/3,1,289/20,
            1,218/15,1,44/3,1,0,27,391/30,1,0];
        I:=[508974,508941,508986,508967,509009,509214,509069,509203,509209,
            508965,509178,508977,508971,508918,508961,509098,509309,508937,
            509147,508994,509042,509184,509181,508927,508960,509225,508915];
    elif hash1 eq 2380 and hash2 eq 1082 and hash3 eq 2786 then
        V:=[20,1,21,5,22,11,23,17,24,23,25,27,26,35,29,43];
        D:=[1,29,1,0,2,247/6,1,83/2,1,0,4,64/3,1,133/4,1,0,6,167/6,1,169/6,1,0,
            8,311/12,1,0,9,241/12,1,41/2,3,125/6,1,0,14,40/3,1,332/15,1,337/15,
            1,0,17,1103/84,1,797/60,2,0];
        I:=[539730,539738,539720,539718,539735,539722,539726,539746,539741,
            539727,539728,539749,539715,539809,539743,539732,539752,539740,
            539747];
    elif hash1 eq 1816 and hash2 eq 1759 and hash3 eq 2214 then
        V:=[17,1,18,5,19,11,20,25,21,45,22,57,23,73,24,83,25,93];
        D:=[1,45/2,1,0,2,29,1,116/3,1,0,4,127/6,1,43/2,1,139/6,3,51/2,1,59/2,1,
            79/2,1,0,12,77/4,1,59/3,1,104/5,1,64/3,2,65/3,1,76/3,1,77/3,1,156/5,
            1,95/3,2,0,23,109/6,2,233/12,1,213/10,1,257/12,1,237/10,1,0,29,18,2,
            55/3,2,19,1,39/2,1,298/15,1,61/3,1,121/5,1,0,38,319/20,1,113/6,1,
            391/20,1,125/6,1,0,42,321/20,1,97/6,1,19,1,96/5,1,0,46,137/10,1,
            499/30,1,0];
        I:=[516692,516601,516627,516691,516377,516010,516533,516653,516603,
            516412,516167,516656,516689,515971,515998,516398,515954,516435,
            516104,515811,516199,516391,516374,516410,516055,516340,516305,
            515903,516651,516734,515809,515959,516380,516404,515969,516639,
            515852,515994,516039,516341,516026,515978,516633,515658,515888,
            515962,515777];
    elif hash1 eq 2474 and hash2 eq 2216 and hash3 eq 2471 then
        V:=[22,1,23,9,24,13,25,23,27,27,28,31,30,37];
        D:=[1,27,1,92/3,1,110/3,1,0,4,187/6,1,0,5,24,1,146/5,1,156/5,1,94/3,1,0,
            9,143/6,1,0,10,101/6,1,0,11,199/12,1,117/7,1,0,13,274/21,1,66/5,2,
            0];
        I:=[505548,505165,505554,505650,505560,505547,505552,505651,505508,
            505537,505501,504589,504593,505499,505625];
    elif hash1 eq 453 and hash2 eq 1525 and hash3 eq 1503 then
        V:=[22,1,23,11,24,19,25,27,26,33,27,39,28,43];
        D:=[1,86/3,1,31,1,98/3,1,116/3,1,0,5,127/6,1,47/2,1,151/6,1,0,8,96/5,1,
            70/3,1,83/3,1,0,11,119/6,1,121/6,1,0,13,81/5,1,61/3,1,0,15,79/6,1,0,
            16,12,1,241/20,1,16,1,17,1,0];
        I:=[506519,506644,506648,505909,506104,506647,506182,505943,506192,
            506059,504325,506202,506068,506454,506769,505108,505576,504985,
            504323];
    elif hash1 eq 735 and hash2 eq 1388 and hash3 eq 1408 then
        V:=[22,1,23,5,24,11,25,17,26,27,27,35];
        D:=[1,25,4,0,5,43/2,2,139/6,2,0,9,59/3,2,22,1,0,12,109/6,7,37/2,2,79/4,
            1,119/6,1,0,23,49/3,1,50/3,1,56/3,1,0,26,89/6,1,0];
        I:=[673130,673131,673156,673181,673139,673148,673112,673122,673058,
            673161,673105,673075,673098,673124,673128,673150,673157,673176,
            673057,673175,673004,673020,673056,673072,672972,673007];
    elif hash1 eq 3367 and hash2 eq 2400 and hash3 eq 359 then
        V:=[12,1,13,13,14,25,15,35,16,49,17,55,18,63,19,71];
        D:=[1,32,1,34,2,38,3,42,1,48,1,0,9,61/2,1,65/2,1,69/2,1,77/2,1,81/2,1,0,
            14,27,1,33,1,35,1,41,1,0,18,151/6,1,51/2,1,163/6,1,55/2,2,199/6,1,
            211/6,1,0,25,83/3,1,36,1,0,27,121/6,1,167/6,1,169/6,1,0,30,61/3,1,
            67/3,1,85/3,1,0,33,113/6,1,0];
        I:=[430077,430088,430090,430087,430089,430092,430091,430057,430093,
            430055,430086,430074,430067,430094,430064,430053,430011,430068,
            430075,430056,430012,430052,430023,430058,430051,429987,430000,
            430013,430017,430014,429975,429966,430020];
    elif hash1 eq 2051 and hash2 eq 1938 and hash3 eq 1739 then
        V:=[26,1,27,13,29,25,30,29,31,35,32,41];
        D:=[1,83/3,1,117/4,1,88/3,1,89/3,1,94/3,1,0,6,143/6,1,155/6,1,157/6,1,
            277/10,1,167/6,2,0,12,611/30,1,0,13,203/12,1,52/3,1,0,15,291/20,1,
            205/12,1,0,17,53/4,2,175/12,1,0];
        I:=[244552,242241,244613,244584,244309,244619,244585,242221,242277,
            242274,244631,241798,243135,244553,242165,241849,244123,244424,
            241792];
    elif hash1 eq 2333 and hash2 eq 2508 and hash3 eq 3399 then
        V:=[26,1,27,7,28,13,29,21,30,25];
        D:=[1,19,1,68/3,1,0,3,103/6,1,115/6,1,0,5,47/3,5,52/3,1,53/3,2,0,13,
            97/6,1,0,14,38/3,1,167/12,1,0];
        I:=[235021,233367,234996,231993,233330,234639,234908,235009,235024,
            232583,232203,234950,232593,234214,232726];
    elif hash1 eq 1205 and hash2 eq 949 and hash3 eq 3080 then
        V:=[23,1,24,5,25,11,26,17,28,21,30,29];
        D:=[1,163/6,1,0,2,71/3,1,95/3,1,0,4,143/6,1,145/6,2,0,7,101/5,1,0,8,
            199/12,2,203/12,2,18,1,0,13,281/21,1,203/15,3,41/3,3,0];
        I:=[250671,250436,249993,250476,249994,249999,250414,249848,250618,
            249850,250619,249837,249843,249877,250000,250001,249996,249997,
            249998];
    elif hash1 eq 1111 and hash2 eq 1275 and hash3 eq 1031 then
        V:=[25,1,26,9,27,13,28,31,29,41,30,49];
        D:=[1,35/2,1,139/6,1,163/6,1,0,4,77/4,1,0,5,95/6,2,97/6,1,33/2,1,107/6,
            2,109/6,1,197/10,1,121/6,1,41/2,1,0,15,44/3,2,49/3,1,50/3,1,268/15,
            1,0,20,147/10,2,479/30,1,33/2,1,0,24,151/12,1,155/12,1,13,2,0];
        I:=[408681,408280,408339,407035,407956,408689,408743,407197,407950,
            408395,407936,407731,407037,408808,407121,407894,408392,408735,
            406904,407887,407892,405061,404620,408320,407439,407402,408762];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 31 mod 47.
function cached_ids_31(P,hash1,hash2,hash3)
    if hash1 eq 1629 and hash2 eq 728 and hash3 eq 2078 then
        V:=[28,1,29,9,30,13,31,21,33,25,34,35];
        D:=[1,116/5,1,26,1,83/3,1,0,4,121/6,1,0,5,56/3,1,298/15,1,61/3,2,0,9,
            551/30,1,0,10,137/12,1,391/30,1,451/30,1,91/6,2,0,15,667/60,3,0];
        I:=[382281,382188,382190,381530,381568,381571,381436,381567,381435,
            381500,382124,382014,380762,380763,381345,381346,381478];
    elif hash1 eq 1347 and hash2 eq 2987 and hash3 eq 971 then
        V:=[20,1,21,5,22,9,23,17,24,31,25,41,26,45,27,51,28,61];
        D:=[1,98/3,1,0,2,163/6,1,0,3,71/3,1,101/4,1,30,1,0,6,131/6,1,133/6,1,
            143/6,2,155/6,2,111/4,1,167/6,1,0,14,239/12,1,328/15,1,68/3,1,
            358/15,1,0,18,245/12,1,0,19,86/5,1,369/20,1,0,21,299/20,1,461/30,1,
            511/30,1,103/6,1,0,25,233/15,1,0];
        I:=[424722,424562,424723,423403,424403,424725,424431,424717,424718,
            423322,424429,423504,423508,423364,423323,424375,423506,422707,
            422530,423468,423251,423462,422588,422585,422569];
    elif hash1 eq 2569 and hash2 eq 859 and hash3 eq 2749 then
        V:=[29,1,30,7,31,11,32,17];
        D:=[1,121/6,1,217/10,1,0,3,43/3,1,0,4,431/30,1,89/6,1,0,6,151/12,1,
            193/15,4,13,3,40/3,3,0];
        I:=[228653,227293,228652,226503,227430,226374,226231,226356,227301,
            227319,226190,227353,227564,227565,227870,227893];
    elif hash1 eq 1253 and hash2 eq 2020 and hash3 eq 3134 then
        V:=[22,1,23,7,24,13,25,19];
        D:=[1,23,1,25,1,0,3,127/6,1,43/2,3,0,7,59/3,1,64/3,1,0,9,107/6,1,109/6,
            5,37/2,1,0];
        I:=[61872,61874,61862,61876,61878,61879,61848,61841,61861,61839,61847,
            61867,61869,61875,61877];
    elif hash1 eq 1911 and hash2 eq 2240 and hash3 eq 683 then
        V:=[23,1,24,7,25,15,26,27,27,45,28,55,29,69,30,75];
        D:=[1,115/6,1,175/6,1,0,3,58/3,1,59/3,1,136/5,1,0,6,97/6,1,33/2,1,
            209/12,1,177/10,1,45/2,1,0,11,50/3,2,17,1,89/5,1,268/15,1,18,1,91/5,
            3,56/3,1,21,1,0,22,169/12,1,91/6,1,101/6,1,41/2,1,0,26,155/12,2,
            40/3,5,103/7,1,179/12,2,248/15,2,199/12,1,0,39,757/60,1,461/30,1,0,
            41,82/7,1,178/15,1,203/15,1,0];
        I:=[503431,503174,503180,503276,503247,503292,503472,503430,503282,
            503464,502811,503559,503469,502160,503178,503566,503082,503083,
            503150,503584,503460,503347,502963,503143,502355,503611,503622,
            503080,503084,503494,503596,503597,502042,501340,501724,502106,
            502154,501962,503394,501819,502012,501945,501394];
    elif hash1 eq 2522 and hash2 eq 2241 and hash3 eq 1010 then
        V:=[10,1,11,15,12,27,13,41,14,51,15,59,16,69,17,79,18,83,20,93];
        D:=[1,38,2,40,1,42,3,44,1,46,1,48,1,0,10,73/2,1,77/2,1,81/2,1,85/2,1,
            97/2,1,0,15,33,1,35,1,37,1,122/3,1,128/3,1,43,2,0,22,199/6,2,67/2,1,
            211/6,1,71/2,2,0,28,106/3,1,107/3,3,36,1,0,33,167/6,4,169/6,1,215/6,
            1,217/6,1,0,40,335/12,1,28,1,113/4,1,91/3,1,0,44,137/6,1,0,45,
            338/15,2,343/15,4,23,2,70/3,1,0,54,53/3,1,0];
        I:=[430424,430426,430417,430420,430423,430425,430422,430418,430411,
            430421,430419,430412,430415,430402,430413,430414,430403,430410,
            430408,430398,430400,430394,430405,430409,430392,430399,430416,
            430395,430388,430401,430404,430397,430389,430390,430393,430396,
            430407,430383,430373,430387,430385,430386,430381,430406,430375,
            430377,430376,430378,430379,430380,430374,430384,430382,430391];
    elif hash1 eq 1958 and hash2 eq 3356 and hash3 eq 1209 then
        V:=[16,1,17,5,19,13,20,23,21,29,22,33,23,39,26,45];
        D:=[1,86/3,1,0,2,59/2,1,223/6,1,79/2,1,0,5,143/6,1,145/6,1,193/6,1,
            241/6,1,0,9,121/5,1,161/5,2,0,12,269/12,1,0,13,278/15,1,283/15,2,0,
            16,793/42,1,619/30,1,0,18,1426/105,1,0];
        I:=[518406,518377,518379,518381,518404,518389,518396,518257,518380,
            518314,518513,518392,518407,518394,518437,518268,518374,518312];
    elif hash1 eq 924 and hash2 eq 3285 and hash3 eq 2643 then
        V:=[22,1,23,9,24,27,25,41,26,61,27,75,28,85];
        D:=[1,76/3,1,83/3,1,117/4,1,0,4,119/6,1,121/6,1,87/4,1,131/6,2,95/4,1,
            143/6,1,145/6,1,155/6,1,0,13,55/3,2,101/5,2,81/4,1,61/3,1,62/3,1,
            287/12,1,0,21,551/30,1,221/12,2,37/2,1,599/30,1,245/12,1,41/2,1,
            207/10,1,125/6,1,45/2,1,0,31,253/15,2,203/12,1,277/15,1,223/12,1,
            227/12,1,22,1,0,38,85/6,1,709/42,1,511/30,1,229/12,1,0,42,203/15,1,
            208/15,3,14,7,0];
        I:=[421462,421452,421591,421622,421631,421399,421450,421499,421355,
            421461,421393,421353,421345,421604,421583,421676,421346,421487,
            421357,421238,421454,421394,421397,421460,421448,421347,421237,
            421167,421539,421245,421630,421643,421343,421532,421605,421201,
            421152,421350,421165,421166,421191,421589,421390,421398,421541,
            421194,421372,421392,421395,421412,421449,421459];
    elif hash1 eq 501 and hash2 eq 907 and hash3 eq 1501 then
        V:=[24,1,25,5,26,15,27,29,28,45,29,53,30,57];
        D:=[1,71/3,5,0,6,107/6,1,79/4,1,119/6,1,237/10,1,0,10,50/3,1,17,1,
            215/12,1,55/3,2,56/3,1,101/5,1,0,17,197/12,1,33/2,3,167/10,1,101/6,
            2,103/6,6,551/30,2,599/30,1,0,33,223/15,1,46/3,1,199/12,1,0,36,51/4,
            2,0,38,173/15,1,178/15,1,41/3,1,0];
        I:=[414459,414460,414572,414573,414613,414409,414223,414124,413992,
            414405,414508,414540,414041,414562,413987,414250,414620,414127,
            414544,414588,414034,414134,414598,414454,414456,414602,414607,
            414612,414614,413997,414002,413513,413933,413986,413479,414522,
            414753,414215,413476,413436];
    elif hash1 eq 2710 and hash2 eq 1799 and hash3 eq 3202 then
        V:=[17,1,18,9,19,19,20,25,21,41,22,53,23,69,24,81,25,89,26,99,27,107];
        D:=[1,65/2,1,69/2,1,89/2,1,0,4,80/3,1,27,1,33,3,110/3,1,0,10,187/6,1,
            199/6,3,0,14,71/3,1,76/3,2,77/3,3,82/3,1,89/3,1,94/3,1,141/4,1,0,24,
            95/6,2,109/6,1,133/6,2,45/2,1,329/12,1,0,31,44/3,1,298/15,1,239/12,
            1,20,1,263/12,1,67/3,1,311/12,1,0,38,431/30,3,59/4,1,359/20,1,
            217/12,1,127/6,2,0,46,223/12,2,337/15,1,114/5,1,0,50,401/30,2,27/2,
            1,877/60,2,59/4,1,0,56,66/5,2,419/28,1,316/21,1,0,60,347/30,2,0];
        I:=[542186,542215,542188,542151,542107,542042,542210,542222,542192,
            542112,542134,542153,542205,542044,542127,542216,542141,542149,
            542204,542217,542213,542219,542181,542162,542214,542039,542036,
            542227,542133,542079,542111,542179,542031,542157,542183,542175,
            542202,542052,542055,542173,542032,542084,542171,542129,542201,
            542138,542199,542024,542026,542114,542126,542106,542121,542182,
            542168,542123,542176,542013,542017,542177,542178];
    elif hash1 eq 454 and hash2 eq 2820 and hash3 eq 402 then
        V:=[26,1,27,11,29,21,30,31,31,43];
        D:=[1,93/4,1,76/3,1,77/3,2,82/3,1,0,6,107/6,1,37/2,1,121/6,1,217/10,2,0,
            11,89/6,1,103/6,1,221/12,1,207/10,1,0,15,44/3,1,223/15,2,15,1,
            247/15,1,86/5,1,0,21,157/12,1,161/12,1,0];
        I:=[408101,407066,407472,407505,407910,407113,407471,407457,406958,
            407579,407445,407409,406992,407525,407178,407109,407644,407455,
            404473,407844,407520,407529];
    elif hash1 eq 360 and hash2 eq 2200 and hash3 eq 2737 then
        V:=[16,1,17,7,18,19,20,29,21,35,23,43];
        D:=[1,24,1,26,1,0,3,45/2,2,49/2,2,53/2,1,57/2,1,69/2,1,0,10,68/3,2,74/3,
            1,25,1,80/3,1,0,15,65/3,1,22,2,0,18,109/6,1,233/12,1,133/6,1,0,21,
            89/6,1,0];
        I:=[425411,425271,425295,425389,425361,425401,425386,425276,425272,
            425184,425380,425344,425385,425346,425010,425088,425127,425120,
            425180,425040,425178];
    elif hash1 eq 1394 and hash2 eq 2348 and hash3 eq 56 then
        V:=[17,1,18,5,19,11,20,15,21,19,22,25,24,29,26,41];
        D:=[1,77/2,1,0,2,21,1,128/3,1,0,4,199/6,1,0,5,107/3,1,0,6,103/4,1,167/6,
            1,0,8,67/3,1,0,9,289/20,1,218/15,1,302/15,1,338/15,1,343/15,1,0,14,
            187/12,1,47/3,1,0];
        I:=[515123,515025,514779,515159,514601,514792,514702,514514,514826,
            514810,514052,514138,514086,514161,514158];
    elif hash1 eq 125 and hash2 eq 1552 and hash3 eq 2373 then
        V:=[19,1,20,7,21,15,22,23,23,35,24,39,25,43,26,55];
        D:=[1,211/6,1,223/6,1,0,3,77/3,1,89/3,3,113/3,1,0,8,145/6,1,53/2,1,
            191/6,1,0,11,24,1,121/5,1,73/3,1,26,1,131/5,1,0,16,611/30,1,0,17,
            369/20,1,0,18,65/4,1,1117/60,1,571/30,1,229/12,1,581/30,1,0,23,77/5,
            1,233/15,1,65/4,1,244/15,1,523/28,1,0];
        I:=[427127,427433,427194,427118,427463,427483,426606,427486,427220,
            427045,427146,426757,426828,427067,426482,426579,426539,426737,
            426493,426515,426516,426487,426674,426651,426540,426513,426425];
    elif hash1 eq 830 and hash2 eq 898 and hash3 eq 1510 then
        V:=[24,1,25,5,26,13,27,19,28,27,29,35,31,45];
        D:=[1,27,1,0,2,151/6,1,175/6,2,59/2,1,0,6,71/3,2,24,1,0,9,119/6,2,
            213/10,1,145/6,1,0,13,91/5,1,61/3,1,62/3,1,0,16,33/2,1,167/10,1,
            101/6,2,183/10,1,0,21,27/2,1,0];
        I:=[671575,671182,671233,671656,671657,671211,671658,671662,671207,
            671284,671155,671210,671208,671113,671114,671178,671195,671217,
            671278,671167,671661];
    elif hash1 eq 2616 and hash2 eq 371 and hash3 eq 3429 then
        V:=[17,1,18,7,19,19,20,33,21,41,22,57,23,65,24,69];
        D:=[1,53/2,2,57/2,1,0,4,74/3,1,25,3,80/3,2,27,2,86/3,1,0,13,43/2,1,
            139/6,2,47/2,3,151/6,2,51/2,1,175/6,1,0,23,65/3,2,22,2,71/3,1,0,28,
            37/2,1,79/4,1,121/6,4,41/2,1,257/12,2,131/6,4,133/6,6,0,47,55/3,3,
            61/3,1,263/12,1,0,52,101/6,4,0,56,46/3,1,0];
        I:=[254341,254356,254344,254326,254342,254345,254362,254280,254309,
            254316,254320,254302,254330,254304,254308,254296,254355,254359,
            254227,254228,254217,254134,254297,254331,254192,254221,254135,
            254224,254185,254180,254213,254214,254294,254231,254223,254226,
            254125,254133,254205,254232,254127,254142,254145,254175,254176,
            254202,254143,254201,254225,254154,254155,254126,254290,254312,
            254317,254144];
    elif hash1 eq 1676 and hash2 eq 2160 and hash3 eq 233 then
        V:=[23,1,24,5,25,15,26,27,27,33,28,43,29,55,30,59];
        D:=[1,167/6,1,0,2,22,1,74/3,1,382/15,1,80/3,1,0,6,173/12,1,91/6,1,31/2,
            1,207/10,1,311/14,1,0,11,175/12,2,67/3,1,0,14,451/30,1,91/6,1,
            461/30,1,31/2,1,0,18,787/60,1,79/6,2,53/4,1,40/3,1,907/60,1,0,24,
            79/6,2,0,26,49/4,1,0];
        I:=[513486,513502,513472,513774,513462,513737,513580,513497,513576,
            513376,513596,513730,513439,513579,513506,513484,513466,513714,
            513601,513716,513608,513609,513432,513349,513435,513298];
    elif hash1 eq 689 and hash2 eq 170 and hash3 eq 534 then
        V:=[23,1,24,15,25,25,26,31,27,39,29,43];
        D:=[1,39/2,4,127/6,3,47/2,1,99/4,1,151/6,1,51/2,1,0,12,58/3,1,24,1,77/3,
            1,109/4,1,0,16,109/6,1,233/12,1,0,18,81/5,2,49/3,1,91/5,1,0,22,
            197/12,1,0,23,129/10,2,0];
        I:=[410217,410255,410327,410419,410397,410545,410572,411085,410617,
            410556,411286,410360,410245,411087,410826,410997,410750,410330,
            410371,410716,410928,410640,409189,411204];
    elif hash1 eq 1864 and hash2 eq 1185 and hash3 eq 821 then
        V:=[20,1,21,5,22,19,23,27,24,33,25,37,26,41,27,45,29,51];
        D:=[1,37,1,0,2,55/2,1,175/6,1,59/2,1,199/6,3,67/2,2,223/6,1,0,11,77/3,1,
            109/4,1,95/3,1,0,14,257/12,1,237/10,1,0,16,298/15,1,0,17,18,1,0,18,
            227/12,1,0,19,181/12,1,91/6,1,0,21,347/30,1,0];
        I:=[420365,420151,419750,419520,419355,420146,420309,420150,420301,
            419764,420114,419213,419749,419664,419215,419813,419070,419069,
            419401,420147,419702];
    elif hash1 eq 1065 and hash2 eq 1714 and hash3 eq 2169 then
        V:=[24,1,25,7,26,13,28,25];
        D:=[1,19,1,80/3,1,0,3,91/6,3,103/6,3,0,9,46/3,1,47/3,1,16,1,53/3,1,77/4,
            1,0,14,208/15,2,43/3,2,0];
        I:=[400380,400243,400068,400470,400530,397268,400042,400521,400513,
            400486,399012,397993,397194,398994,399358,399071,399347];
    elif hash1 eq 3274 and hash2 eq 582 and hash3 eq 2776 then
        V:=[21,1,22,9,23,25,24,35,25,47,26,59,27,71];
        D:=[1,57/2,1,61/2,5,65/2,2,0,9,19,2,21,1,74/3,3,25,1,80/3,3,27,2,29,1,0,
            22,115/6,1,127/6,1,43/2,1,139/6,1,0,26,52/3,1,53/3,6,59/3,1,20,1,
            65/3,1,0,36,209/12,1,107/6,1,109/6,2,37/2,3,121/6,1,0,44,31/2,2,
            191/12,2,16,2,49/3,1,50/3,1,0,52,91/6,1,0];
        I:=[251557,250908,251137,251498,251504,251558,250755,251533,251293,
            251509,251581,251260,251264,251303,251510,250675,250776,251336,
            250801,251466,250799,250826,251092,250710,250874,250742,250353,
            250750,250897,251334,251361,251465,250926,251152,250188,250351,
            250091,250884,251267,250672,250749,251588,251268,250249,250823,
            250114,251337,250093,250248,250846,251270,250364];
    elif hash1 eq 3039 and hash2 eq 1478 and hash3 eq 2429 then
        V:=[25,1,26,11,27,27,28,37,29,47,30,55];
        D:=[1,127/6,2,43/2,1,139/6,1,47/2,2,0,7,69/4,1,53/3,1,18,1,58/3,1,106/5,
            1,85/4,1,65/3,3,0,16,97/6,1,177/10,1,107/6,1,37/2,1,0,20,44/3,1,
            49/3,1,50/3,1,73/4,1,0,24,431/30,1,173/12,1,147/10,1,0,27,38/3,1,
            44/3,1,0];
        I:=[245266,245396,245197,242846,242968,243692,243394,242962,244918,
            245189,243977,243985,242953,244744,245207,243704,243912,245136,
            243369,243332,244808,243146,242456,242656,243887,243223,242608,
            241980];
    elif hash1 eq 172 and hash2 eq 2906 and hash3 eq 313 then
        V:=[19,1,20,7,21,15,22,23,23,31,24,39,25,43,26,51];
        D:=[1,41/2,2,45/2,3,0,6,68/3,2,23,1,80/3,1,0,10,35/2,1,75/4,1,127/6,1,0,
            13,52/3,1,96/5,1,70/3,1,0,16,97/6,1,71/4,2,107/6,1,0,20,17,1,0,21,
            147/10,2,197/12,1,101/6,1,0,25,103/7,1,76/5,1,0];
        I:=[510715,510753,510696,510697,510702,510293,510750,510521,510361,
            510275,510633,510294,510278,509248,509294,510630,510357,510522,
            510692,509345,509831,509833,509797,509256,508825,508868];
    elif hash1 eq 407 and hash2 eq 3276 and hash3 eq 629 then
        V:=[26,1,27,7,28,11,29,23,30,31,31,37];
        D:=[1,18,1,58/3,1,0,3,97/6,1,0,4,44/3,1,191/12,1,65/4,1,49/3,1,268/15,1,
            0,9,431/30,1,147/10,1,193/12,1,0,12,13,1,66/5,1,0,14,73/6,1,401/30,
            1,0];
        I:=[408232,408718,408353,406813,406572,408148,408595,407628,406838,
            406859,406291,405847,405711,404517,404157];
    elif hash1 eq 78 and hash2 eq 1287 and hash3 eq 343 then
        V:=[20,1,21,15,22,25,23,41,24,51,25,61,26,73,27,83];
        D:=[1,74/3,1,27,1,86/3,1,29,1,33,1,104/3,1,0,7,39/2,1,43/2,1,59/2,1,
            67/2,1,0,11,58/3,3,59/3,1,106/5,1,64/3,1,116/5,2,71/3,1,24,1,0,21,
            33/2,1,119/6,1,121/6,1,133/6,1,0,25,50/3,1,215/12,1,101/5,1,111/5,2,
            0,30,173/12,1,147/10,1,89/6,3,103/6,1,551/30,1,0,37,223/15,1,179/12,
            1,15,1,338/21,1,0,41,157/12,1,31/2,1,0];
        I:=[510538,510325,510536,510366,510316,510531,510590,510518,509608,
            510416,510516,510544,510577,510545,509578,509508,509446,509455,
            509547,510573,509456,510533,509561,510426,510333,509436,509567,
            509148,509267,509835,509574,509105,509159,509307,509571,509563,
            509127,509043,510383,508893,509168,508800];
    elif hash1 eq 3133 and hash2 eq 428 and hash3 eq 2117 then
        V:=[19,1,20,7,21,15,22,25,23,39,24,43,25,49,27,55];
        D:=[1,163/6,1,55/2,1,0,3,71/3,1,77/3,1,117/4,1,0,6,131/6,1,143/6,2,
            145/6,2,155/6,1,0,12,56/3,1,61/3,1,62/3,1,263/12,1,111/5,1,74/3,1,0,
            18,127/6,1,0,19,86/5,1,52/3,1,0,21,181/12,1,339/20,1,0,23,73/6,1,0];
        I:=[425553,425559,425556,425567,425575,425555,425518,425562,425558,
            425561,425528,425516,425557,425551,425520,425502,425498,425500,
            425510,425506,425563,425469,425521];
    elif hash1 eq 1958 and hash2 eq 2686 and hash3 eq 1088 then
        V:=[27,1,28,5,29,11,30,23,31,31,33,37];
        D:=[1,39/2,1,0,2,53/3,1,106/5,1,0,4,97/6,2,33/2,1,209/12,1,177/10,1,
            71/4,1,0,10,208/15,1,16,1,50/3,1,0,13,79/6,1,29/2,1,0,15,341/30,1,
            0];
        I:=[384307,383649,383732,383650,383730,384312,384119,384314,384342,
            384056,384246,384212,384289,378914,384024];
    elif hash1 eq 2428 and hash2 eq 289 and hash3 eq 2416 then
        V:=[20,1,21,5,22,11,23,17,24,29,25,45,27,49,28,55,29,59];
        D:=[1,128/3,1,0,2,199/6,1,155/4,1,0,4,117/4,1,101/3,2,0,7,167/6,1,
            353/12,1,181/6,2,203/6,1,425/12,1,0,13,358/15,1,287/12,1,26,1,105/4,
            1,79/3,1,335/12,1,28,1,0,20,399/20,1,0,21,56/3,1,229/12,3,0,25,61/4,
            2,0,27,817/60,1,917/60,1,0];
        I:=[420451,419604,419850,419852,419878,420452,420376,419176,420439,
            420446,420427,419859,419770,419777,420230,419826,419638,419637,
            419540,419782,419130,419174,419632,419858,419657,419698,419755,
            419045];
    elif hash1 eq 2287 and hash2 eq 928 and hash3 eq 1018 then
        V:=[15,1,16,11,17,33,18,53,19,81,20,105,21,129,22,147,23,157,24,163];
        D:=[1,57/2,1,61/2,3,65/2,3,73/2,2,0,10,25,1,27,2,86/3,1,29,2,92/3,1,31,
            2,98/3,2,33,3,104/3,1,35,1,0,26,47/2,1,151/6,3,51/2,3,163/6,6,55/2,
            2,175/6,4,59/2,1,187/6,2,211/6,1,0,49,65/3,1,93/4,1,71/3,5,24,3,
            76/3,1,77/3,4,26,2,109/4,1,83/3,2,28,1,146/5,1,89/3,1,30,1,0,73,
            121/6,1,131/6,3,133/6,4,45/2,1,281/12,1,237/10,1,143/6,1,145/6,1,
            305/12,1,167/6,1,169/6,1,0,89,56/3,2,61/3,6,62/3,2,43/2,1,328/15,1,
            263/12,5,111/5,2,89/4,1,67/3,4,68/3,5,388/15,1,0,119,103/6,1,221/12,
            1,37/2,1,113/6,2,611/30,1,125/6,1,22,1,265/12,1,0,128,253/15,2,17,5,
            133/6,1,271/12,1,0,137,185/12,1,31/2,1,0,139,238/15,1,0];
        I:=[255345,255328,255336,255344,255321,255342,255349,255326,255341,
            255218,255307,255323,255278,255335,255343,255247,255267,255347,
            255206,255252,255114,255215,255219,255242,255240,255264,255225,
            255233,255258,255176,255322,255338,255196,255207,255210,255226,
            255250,255283,255222,255337,255209,255211,255274,255275,255110,
            255108,255174,255055,255231,255188,255238,255266,255276,255280,
            255295,255216,255340,255348,255175,255121,255148,255173,255291,
            255117,255151,255180,255040,255169,255159,255046,255087,255075,
            255243,255163,255239,255290,255140,255179,255235,255237,255171,
            255153,255042,255142,255112,255152,255084,255091,255158,255162,
            255088,255103,255118,255135,255236,255270,255111,255113,255170,
            255072,255099,255149,255155,255160,255178,255043,255101,255107,
            255057,255059,255096,255168,255037,255053,255123,255130,255154,
            255018,255082,255076,255051,255156,255165,255102,255054,255039,
            255104,255071,255100,255058,255106,255228,255256,255300,255010,
            255008,255097,255030,255007];
    elif hash1 eq 2381 and hash2 eq 2357 and hash3 eq 3337 then
        V:=[24,1,25,7,26,11,27,23,28,33,29,41,30,57];
        D:=[1,47/3,1,58/3,1,0,3,41/2,1,0,4,167/12,1,43/3,1,268/15,1,62/3,1,
            263/12,1,0,9,431/30,1,173/12,1,101/6,1,35/2,1,0,13,155/12,1,76/5,1,
            203/12,1,0,16,51/4,1,259/20,3,13,2,401/30,1,161/12,1,83/6,2,31/2,1,
            0,27,173/15,1,139/12,1,363/28,1,66/5,1,0];
        I:=[502810,502453,502483,503269,503255,502708,501950,503289,503136,
            503390,502337,502460,503188,502092,502015,503248,503135,503231,
            503256,503140,503275,502091,501419,502380,502482,503106,501977,
            501781,501836,502076];
    elif hash1 eq 1770 and hash2 eq 197 and hash3 eq 1478 then
        V:=[20,1,21,5,22,17,23,27,24,39,25,47,26,53];
        D:=[1,68/3,1,0,2,83/4,1,151/6,3,175/6,1,123/4,1,187/6,1,0,9,85/4,1,64/3,
            1,65/3,1,109/4,1,0,13,97/6,1,79/4,1,119/6,1,121/6,1,143/6,1,0,18,
            49/3,1,55/3,1,89/4,1,0,21,217/12,1,221/12,1,0,23,223/15,1,15,2,0];
        I:=[424731,424180,424609,424650,424739,422718,423157,424575,423777,
            424758,424615,423179,424649,423773,424573,424741,424195,423166,
            424293,422629,422427,422415,423295,423340,424711];
    elif hash1 eq 3133 and hash2 eq 808 and hash3 eq 529 then
        V:=[24,1,25,13,26,19,27,25,28,29,29,37,30,45];
        D:=[1,64/3,1,77/3,1,26,1,82/3,1,83/3,1,0,6,297/10,1,181/6,1,0,8,20,1,
            328/15,1,0,10,187/10,1,0,11,46/3,2,277/15,1,313/15,1,0,15,451/30,1,
            511/30,1,521/30,1,0,18,163/12,1,0];
        I:=[502848,503227,503250,503204,503226,503132,503246,502831,503159,
            503199,503160,503254,502014,503218,502833,503193,503164,503215];
    elif hash1 eq 2757 and hash2 eq 2246 and hash3 eq 915 then
        V:=[22,1,23,15,24,33,25,53,26,67,27,87,28,99,29,105];
        D:=[1,68/3,1,23,1,74/3,1,25,1,27,1,29,1,0,7,115/6,3,39/2,2,83/4,1,127/6,
            2,43/2,3,139/6,3,47/2,2,151/6,1,0,24,53/3,2,18,2,58/3,2,59/3,1,20,3,
            106/5,1,85/4,1,64/3,1,65/3,1,0,38,33/2,2,71/4,1,107/6,3,109/6,4,
            197/10,1,119/6,1,0,50,191/12,2,16,1,81/5,1,65/4,3,49/3,4,50/3,1,
            91/5,2,55/3,2,56/3,1,0,67,173/12,1,147/10,1,89/6,1,91/6,1,33/2,1,0,
            72,40/3,2,15,1,0,75,71/6,2,0];
        I:=[249704,249675,249707,249710,249370,249411,249518,249599,249638,
            249541,249720,249442,249618,249706,248930,249280,249668,248955,
            249713,249734,249302,249689,248268,249712,249719,249226,249350,
            248793,248829,248763,248847,249434,249526,248266,247755,249366,
            248003,248889,249371,247875,247982,248830,249627,247930,248821,
            248941,249318,248090,248594,249265,249431,249239,247791,247792,
            248228,248995,247669,248764,248971,249385,249329,247696,247812,
            247593,247651,247782,248009,248168,247973,247969,247591,247877,
            247987,247574,247784,248061];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 32 mod 47.
function cached_ids_32(P,hash1,hash2,hash3)
    if hash1 eq 831 and hash2 eq 1055 and hash3 eq 1159 then
        V:=[24,1,25,5,26,11,27,23,28,37,29,51];
        D:=[1,71/3,1,0,2,107/6,1,41/2,1,0,4,50/3,3,215/12,1,99/5,1,61/3,1,62/3,
            1,0,11,89/6,1,193/12,1,491/30,1,33/2,1,67/4,1,103/6,1,0,17,41/3,1,
            179/12,1,117/7,1,253/15,1,17,1,227/12,1,0,23,391/30,1,401/30,1,27/2,
            3,0];
        I:=[414036,414115,414125,414102,414145,414158,414144,414181,414212,
            413642,414412,414205,414222,414167,414081,414137,414141,413994,
            413524,413518,413646,413533,414214,414147,413637,414142,414154];
    elif hash1 eq 1113 and hash2 eq 3086 and hash3 eq 861 then
        V:=[25,1,26,9,28,19,30,25,31,29,32,41];
        D:=[1,599/20,1,65/2,1,69/2,1,0,4,368/15,2,529/20,1,398/15,1,80/3,1,0,9,
            442/21,1,449/21,1,0,11,923/60,2,0,13,82/7,1,47/4,2,178/15,1,239/20,
            1,12,1,0,19,82/7,1,287/24,1,1256/105,1,0];
        I:=[509734,509086,509370,509106,509107,509758,509759,509264,508687,
            508691,508754,508762,508881,508796,508824,509740,509760,509680,
            508297,508437,508442];
    elif hash1 eq 737 and hash2 eq 1462 and hash3 eq 1405 then
        V:=[28,1,29,5,30,9,31,15,32,23,33,35];
        D:=[1,70/3,1,0,2,121/6,2,0,4,91/5,1,55/3,1,0,6,33/2,1,167/10,1,101/6,1,
            0,9,175/12,2,44/3,1,223/15,1,15,1,46/3,1,0,15,38/3,1,0];
        I:=[666052,665605,665747,665157,665838,665834,665780,666142,665153,
            665233,666164,665254,665294,666191,666023];
    elif hash1 eq 549 and hash2 eq 2709 and hash3 eq 2834 then
        V:=[27,1,28,7,29,13,30,23,31,31];
        D:=[1,115/6,1,127/6,2,0,4,53/3,1,20,1,0,6,95/6,1,33/2,4,71/4,1,109/6,2,
            0,14,71/5,1,44/3,1,49/3,2,0,18,25/2,1,0];
        I:=[228239,227972,228236,228000,228100,228528,228334,228414,228469,
            228477,225900,226885,227631,227739,226913,225872,226090,226192];
    elif hash1 eq 596 and hash2 eq 2690 and hash3 eq 1135 then
        V:=[23,1,24,9,25,15,26,25,27,33,28,41];
        D:=[1,47/2,1,163/6,1,55/2,1,0,4,65/3,2,71/3,2,0,8,119/6,3,121/6,1,41/2,
            1,217/10,1,0,14,18,1,73/4,1,55/3,2,0,18,197/12,1,167/10,1,551/30,2,
            0,22,175/12,2,44/3,2,15,1,46/3,2,0];
        I:=[673704,673669,673712,673440,673658,673291,673486,673423,673598,
            673715,673496,673697,673275,673435,673313,673299,673471,673227,
            673298,673232,673294,673295,673382,673261,673424,673259,673609,
            673709];
    elif hash1 eq 1818 and hash2 eq 2027 and hash3 eq 1924 then
        V:=[24,1,25,5,26,17,27,25,29,33,31,41];
        D:=[1,35,1,0,2,91/6,1,103/6,1,139/6,1,55/2,1,175/6,1,0,7,14,1,52/3,1,
            59/3,1,0,10,95/6,1,97/6,1,107/6,1,0,13,149/12,3,77/6,1,33/2,1,0,18,
            109/10,1,0];
        I:=[497489,497582,497302,497715,497466,498088,498059,498123,497306,
            497478,497450,496191,497146,497196,497581,498072,496668,497637];
    elif hash1 eq 2852 and hash2 eq 1065 and hash3 eq 3005 then
        V:=[17,1,18,7,19,19,20,35,21,41,22,47,24,53];
        D:=[1,49/2,1,53/2,1,0,3,23,2,74/3,1,27,1,86/3,2,92/3,1,0,10,127/6,2,
            43/2,1,139/6,5,151/6,1,51/2,1,107/4,1,55/2,1,0,22,20,1,116/5,1,0,24,
            119/6,1,121/6,2,0,27,17,1,55/3,2,0,30,44/3,1,46/3,1,0];
        I:=[427435,254761,427375,427387,427474,427073,427048,427462,427225,
            427432,427464,427177,427104,427178,427182,427485,427510,427532,
            427230,427302,427229,427003,426673,427051,426794,427183,426614,
            426769,427062,427224,426751];
    elif hash1 eq 1395 and hash2 eq 11 and hash3 eq 96 then
        V:=[19,1,20,5,21,13,22,29,23,45,24,61,25,69,26,79,27,89,28,93];
        D:=[1,53/2,1,0,2,74/3,1,25,3,80/3,2,0,8,43/2,1,139/6,3,47/2,1,151/6,2,
            51/2,1,163/6,4,115/4,1,0,21,20,1,85/4,1,64/3,1,65/3,2,22,2,71/3,1,
            77/3,2,0,31,109/6,1,197/10,1,119/6,2,121/6,3,41/2,1,257/12,1,87/4,2,
            0,42,215/12,2,55/3,1,56/3,2,0,47,491/30,1,197/12,5,33/2,3,221/12,1,
            0,57,223/15,1,179/12,1,15,2,46/3,1,0,62,451/30,1,0,63,12,1,0];
        I:=[253312,253297,253240,253336,253373,253338,253397,253330,252923,
            253270,253302,253154,252887,253189,252971,252796,252854,252942,
            252960,252735,253056,253029,253303,252939,253315,252624,253057,
            252652,252589,252613,253326,252679,252815,253115,252909,253082,
            253187,252889,252727,252665,252740,253027,253046,253120,252959,
            253286,252623,252725,252750,252829,252912,253078,252510,252650,
            252880,252497,252708,252698,252647,252702,252832,252485,252717];
    elif hash1 eq 1583 and hash2 eq 179 and hash3 eq 1182 then
        V:=[20,1,21,9,22,25,23,41,24,53,25,65,26,73,27,79,28,85];
        D:=[1,29,1,92/3,1,33,1,0,4,51/2,1,175/6,2,59/2,1,187/6,1,131/4,1,67/2,1,
            211/6,1,0,12,136/5,1,109/4,1,83/3,2,117/4,1,88/3,2,89/3,1,101/3,1,0,
            21,133/6,1,237/10,1,305/12,1,155/6,1,329/12,1,0,26,328/15,1,22,1,
            111/5,1,67/3,2,51/2,1,0,32,83/4,1,127/6,1,659/30,1,0,35,253/15,1,
            52/3,1,0,37,511/30,1,103/6,1,0,39,187/12,1,0];
        I:=[423727,423209,423573,423420,422706,423149,423554,423185,423311,
            423564,423490,423200,423314,423091,423545,423620,423680,423732,
            423565,423529,423279,423202,423528,423613,423382,423534,423681,
            423556,423572,423607,423703,423321,423712,422581,423327,422528,
            422589,422568,422524];
    elif hash1 eq 3087 and hash2 eq 2584 and hash3 eq 3039 then
        V:=[25,1,26,9,27,23,28,29,30,41];
        D:=[1,139/6,1,47/2,1,99/4,1,0,4,53/3,1,77/4,1,58/3,1,59/3,2,20,1,116/5,
            1,0,11,97/6,3,109/6,2,0,16,191/12,1,16,1,49/3,1,50/3,3,17,1,0,23,
            151/12,5,193/15,1,0];
        I:=[239507,241058,239692,240246,240839,239516,239577,240941,240947,
            239752,240191,240234,240242,239572,239685,239573,240876,239497,
            239683,239689,240948,240944,240594,240616,240935,240942,241326,
            240236];
    elif hash1 eq 3087 and hash2 eq 3079 and hash3 eq 497 then
        V:=[27,1,28,11,30,17,31,31];
        D:=[1,139/6,3,47/2,1,151/6,1,51/2,1,0,7,59/3,1,71/3,1,0,9,16,1,81/5,1,
            49/3,1,50/3,1,215/12,2,55/3,1,0,16,89/6,1,0];
        I:=[666972,667610,667793,666981,667807,667787,667759,667505,666162,
            666288,667506,667507,666409,666736,665322,666375];
    elif hash1 eq 2805 and hash2 eq 3051 and hash3 eq 2862 then
        V:=[18,1,20,13,22,31,24,53,26,59];
        D:=[1,25,1,29,1,31,1,35,1,110/3,1,0,6,59/3,1,20,2,65/3,1,22,1,116/5,1,
            71/3,1,26,1,101/3,1,0,15,65/4,1,55/3,3,138/7,1,101/5,1,81/4,1,89/4,
            1,67/3,2,68/3,1,24,2,97/4,2,0,30,46/3,1,19,1,0,32,163/12,1,208/15,1,
            0];
        I:=[542401,542464,542458,542474,542422,542389,542232,542402,542243,
            542408,542274,541893,542410,542460,542411,542041,542085,542143,
            541903,542150,542119,542443,542089,542147,541828,542069,542131,
            542064,542075,541854,542088,542087,542304];
    elif hash1 eq 126 and hash2 eq 2323 and hash3 eq 1327 then
        V:=[22,1,23,9,24,27,25,41,27,49,28,65,29,77,30,81];
        D:=[1,23,1,27,1,86/3,1,0,4,39/2,1,83/4,1,127/6,2,139/6,1,175/6,1,59/2,1,
            123/4,1,187/6,1,0,13,58/3,1,64/3,1,22,1,93/4,1,70/3,1,77/3,1,0,19,
            109/6,1,121/6,1,133/6,1,0,22,419/30,1,14,1,29/2,1,167/10,2,101/6,2,
            103/6,1,75/4,1,0,31,296/21,1,218/15,1,175/12,1,338/21,1,329/20,1,0,
            36,51/4,1,0,37,173/15,1,0];
        I:=[412294,411780,411810,412176,410424,410633,410635,412276,410541,
            410679,410695,412045,411066,412206,412217,410655,410622,412024,
            412099,412046,409610,410017,409603,410605,409960,410422,410423,
            410504,409962,409644,409328,409760,409605,409186,409129,409838,
            411115];
    elif hash1 eq 173 and hash2 eq 3224 and hash3 eq 3198 then
        V:=[21,1,22,5,23,13,24,31,26,47,27,61,28,67,29,73];
        D:=[1,199/6,1,0,2,76/3,1,83/3,1,89/3,1,0,5,131/6,1,133/6,1,281/12,1,
            237/10,1,95/4,1,143/6,1,145/6,2,155/6,1,0,14,101/5,1,81/4,1,61/3,1,
            328/15,1,121/5,1,73/3,1,74/3,1,0,21,199/12,2,253/15,1,203/12,1,86/5,
            1,369/20,1,94/5,1,0,28,59/4,1,181/12,1,0,30,269/20,1,203/15,1,0,32,
            117/10,1,371/30,1,137/10,1,0];
        I:=[421852,421884,421564,422109,421497,421644,422033,421614,422031,
            421603,421640,421862,421690,421619,421528,421587,421746,421368,
            421726,421988,422030,422068,421498,422100,421526,421153,421282,
            421374,421387,421512,421511,421530,421215,421110];
    elif hash1 eq 2852 and hash2 eq 2423 and hash3 eq 877 then
        V:=[15,1,16,7,17,15,18,23,19,37,20,41,21,53,22,67,23,87,24,97,25,115];
        D:=[1,49/2,1,53/2,1,0,3,104/3,1,35,1,39,1,0,6,115/4,1,67/2,1,223/6,1,0,
            9,65/3,2,114/5,1,70/3,1,89/3,1,196/5,1,118/3,2,0,17,133/6,1,0,18,
            56/3,1,20,2,109/5,1,263/12,1,358/15,1,0,24,103/6,1,359/20,1,221/12,
            2,241/12,1,611/30,1,335/14,1,0,31,361/20,1,109/6,1,58/3,2,302/15,2,
            102/5,1,313/15,1,133/6,1,68/3,1,362/15,1,0,42,85/6,1,461/30,1,
            205/12,1,81/4,1,0,46,208/15,3,352/21,1,84/5,1,256/15,1,272/15,1,
            73/4,1,183/10,1,113/6,1,0,56,817/60,1,659/36,1,0];
        I:=[518171,518127,518172,518160,518110,518119,518097,518162,518037,
            518164,517893,518065,518088,517989,517779,517780,518080,518069,
            518163,518168,517894,517993,517883,517994,517973,517892,517968,
            517804,517884,517715,517865,518053,517774,517781,517895,517995,
            517975,517992,517650,517612,517648,517984,517900,517807,517660,
            517879,517978,517980,517740,517737,517868,517802,517786,517784,
            517651,517966,517626];
    elif hash1 eq 32 and hash2 eq 42 and hash3 eq 2358 then
        V:=[22,1,23,5,24,11,25,21,26,27,27,37];
        D:=[1,27,1,0,2,127/6,1,47/2,2,0,5,53/3,1,59/3,1,20,2,64/3,1,0,10,71/4,1,
            109/6,1,0,12,191/12,1,49/3,1,50/3,1,55/3,2,0,17,147/10,1,0];
        I:=[249544,249624,249288,249326,248878,248922,248970,249621,248691,
            248918,248976,248848,248910,248902,247643,247781,247801];
    elif hash1 eq 220 and hash2 eq 1459 and hash3 eq 3083 then
        V:=[23,1,24,15,25,21,26,29,27,41,28,45,29,59];
        D:=[1,139/6,1,151/6,1,163/6,1,55/2,2,175/6,1,63/2,1,0,8,65/3,1,70/3,1,0,
            10,121/6,1,41/2,1,53/2,1,0,13,65/4,1,215/12,1,18,1,73/4,1,81/4,1,0,
            18,101/6,1,0,19,151/12,1,155/12,3,29/2,1,46/3,1,329/20,1,33/2,1,0,
            27,91/6,1,0];
        I:=[503589,503591,502844,503612,503616,503222,503607,503605,502853,
            503617,503000,501250,503600,503220,503593,503601,502054,503224,
            502145,503639,503654,503670,503599,503122,502079,502139,501877];
    elif hash1 eq 1912 and hash2 eq 3460 and hash3 eq 1558 then
        V:=[23,1,24,5,25,15,27,23,28,29];
        D:=[1,127/6,2,0,3,77/4,1,58/3,2,65/3,1,71/3,2,0,9,97/6,1,71/4,1,107/6,1,
            0,12,147/10,1,89/6,3,0,16,223/15,1,0];
        I:=[249664,249688,249328,249292,249715,248345,249319,249320,248360,
            249229,249293,248279,248222,248363,249711,247525];
    elif hash1 eq 2382 and hash2 eq 1801 and hash3 eq 1910 then
        V:=[24,1,25,9,26,15,28,23,29,33];
        D:=[1,83/3,1,117/4,1,100/3,1,0,4,145/6,1,49/2,1,0,6,61/3,1,67/3,2,121/5,
            1,0,10,203/12,1,52/3,1,223/12,1,19,1,0,14,91/6,1,511/30,1,0];
        I:=[417822,417691,416737,417082,416730,417065,415833,415914,416068,
            415952,417080,415965,416398,415951,415545];
    elif hash1 eq 1489 and hash2 eq 2470 and hash3 eq 35 then
        V:=[21,1,22,13,23,23,24,37,25,47,26,53,27,61,28,69];
        D:=[1,139/6,1,151/6,2,51/2,2,175/6,1,187/6,1,0,8,65/3,1,22,1,70/3,1,
            71/3,2,0,13,119/6,2,121/6,1,87/4,1,131/6,1,133/6,2,305/12,1,0,21,
            43/3,1,18,1,62/3,3,21,1,0,27,221/12,2,223/10,1,0,30,40/3,1,289/20,1,
            175/12,1,0,33,391/30,3,79/6,2,625/42,1,0,39,173/15,1,0];
        I:=[513557,513465,513478,513492,513516,513499,513481,513487,513498,
            513503,513488,513527,513480,513526,513551,513467,513461,513505,
            513509,513476,513525,513544,513474,513483,513523,513477,513473,
            513550,513357,513510,513574,513540,513460,513534,513536,513468,
            513511,513350,513463];
    elif hash1 eq 1536 and hash2 eq 3162 and hash3 eq 166 then
        V:=[17,1,18,5,19,25,20,45,21,71,22,85,23,103,24,121,25,137,26,157,27,
            171];
        D:=[1,61/2,1,0,2,23,1,25,1,80/3,1,27,2,86/3,1,29,2,104/3,1,110/3,1,
            116/3,1,0,13,127/6,1,51/2,1,163/6,1,175/6,1,187/6,2,131/4,1,199/6,2,
            67/2,3,75/2,1,0,26,58/3,1,59/3,1,106/5,1,65/3,1,22,1,24,1,126/5,1,
            77/3,1,83/3,1,28,1,146/5,1,101/3,1,0,38,109/6,3,79/4,1,119/6,4,
            121/6,1,133/6,2,237/10,1,0,50,215/12,1,18,2,91/5,1,239/12,1,20,3,
            101/5,1,61/3,1,388/15,1,0,61,101/6,3,255/14,1,551/30,1,221/12,1,
            113/6,1,599/30,1,245/12,1,83/4,1,0,71,223/15,2,179/12,1,15,1,46/3,1,
            253/15,2,278/15,1,247/12,1,0,80,59/4,1,119/8,1,451/30,1,91/6,2,31/2,
            2,709/42,1,521/30,1,35/2,1,565/28,1,0,91,281/21,1,67/5,1,907/60,1,
            76/5,1,61/4,1,233/15,1,0,97,361/30,2,0];
        I:=[515073,515062,515272,515072,515214,515283,515077,515337,515344,
            515371,515374,515059,515098,515267,515089,514679,514538,514806,
            515228,515040,515256,515342,515355,515380,514923,515035,515252,
            514718,515122,515365,515363,514709,514733,514796,515368,514812,
            514726,515017,515209,515246,515056,514596,515064,515069,515284,
            514496,514178,515229,514917,515173,515067,515160,514556,515200,
            514324,514495,514595,514103,515141,514305,514498,515356,515379,
            514165,514604,514725,514041,514270,514323,515236,514565,514882,
            514494,515164,515303,514621,514735,514642,514108,515238,513974,
            514599,513975,515028,514939,515353,514170,514090,514187,513977,
            514048,514085,514073,514088,514049,514084,514545,514546];
    elif hash1 eq 126 and hash2 eq 1963 and hash3 eq 2619 then
        V:=[28,1,29,7,30,13,31,23,32,29];
        D:=[1,16,1,18,1,0,3,63/4,1,95/6,1,0,5,14,1,71/5,2,43/3,3,44/3,4,0,15,
            25/2,1,77/6,3,0,19,34/3,1,0];
        I:=[228394,228457,228235,228401,226824,226511,226766,226566,226905,
            228392,226758,226840,227212,227243,228210,228263,228339,228345,
            226271];
    elif hash1 eq 3416 and hash2 eq 3424 and hash3 eq 1222 then
        V:=[26,1,27,5,28,13,29,21,30,33];
        D:=[1,53/3,2,0,3,95/6,1,97/6,1,107/6,2,0,7,44/3,2,81/5,1,49/3,1,0,11,
            169/12,1,29/2,2,59/4,1,89/6,5,91/6,1,0,21,155/12,1,15,2,0];
        I:=[241624,241711,241670,241369,239095,241220,241112,241292,240338,
            240046,240357,241672,241690,240163,238489,239814,240444,240798,
            241351,241208,239250,238135,238635];
    elif hash1 eq 3040 and hash2 eq 1592 and hash3 eq 2543 then
        V:=[24,1,25,5,26,13,27,23,28,29];
        D:=[1,23,1,0,2,127/6,2,139/6,2,51/2,1,0,7,53/3,1,77/4,1,58/3,1,59/3,1,0,
            11,71/4,1,109/6,1,0,13,191/12,1,49/3,1,55/3,1,56/3,1,0];
        I:=[671789,671734,671737,671218,671738,671183,671390,671649,671367,
            671787,671559,671611,671642,671362,671185,671608];
    elif hash1 eq 455 and hash2 eq 2748 and hash3 eq 1382 then
        V:=[29,1,31,7,32,19,33,31];
        D:=[1,207/10,1,125/6,2,0,4,339/20,1,511/30,1,205/12,1,521/30,1,209/12,2,
            0,10,203/15,1,41/3,2,14,1,233/15,1,47/3,1,0,16,55/4,1,0];
        I:=[402807,402766,402911,402858,402809,402765,402803,402792,402801,
            402811,402838,402914,402799,402794,402789,402931];
    elif hash1 eq 3181 and hash2 eq 1245 and hash3 eq 1755 then
        V:=[19,1,20,5,21,13,22,23,23,29,24,33,25,37,26,43];
        D:=[1,77/2,1,0,2,74/3,2,29,1,33,1,0,6,83/4,1,127/6,1,99/4,1,151/6,1,0,
            10,59/3,1,65/3,1,0,12,103/4,1,0,13,44/3,1,0,14,431/30,3,173/12,1,0,
            18,218/15,1,175/12,1,44/3,1,302/15,1,0];
        I:=[510358,510353,510412,510488,510464,510421,510414,510471,510378,
            510462,510406,509065,510082,510070,510085,510087,509045,508452,
            510479,510394,508379];
    elif hash1 eq 1442 and hash2 eq 1432 and hash3 eq 2997 then
        V:=[21,1,22,15,23,25,24,45,25,61,26,79,27,89,28,95];
        D:=[1,127/6,1,43/2,1,139/6,1,151/6,2,175/6,1,59/2,1,0,8,59/3,3,20,1,
            71/3,1,77/3,1,0,14,71/4,1,107/6,4,109/6,2,119/6,2,121/6,1,217/10,1,
            87/4,1,131/6,1,133/6,2,0,29,50/3,3,215/12,1,91/5,1,55/3,2,239/12,1,
            62/3,1,263/12,1,0,39,193/12,1,197/12,1,33/2,2,101/6,3,539/30,3,18,1,
            217/12,1,113/6,1,0,52,223/15,1,179/12,2,15,1,203/12,1,0,57,401/30,1,
            27/2,1,0,59,143/12,1,0];
        I:=[422244,421827,422307,422234,422280,421567,422127,422224,422274,
            422313,422219,421922,421579,421969,422035,422236,422242,422283,
            422063,422221,421833,421965,422223,421580,421253,421494,421314,
            421639,421997,422052,422087,422002,421678,421918,422229,421692,
            421638,421213,421667,422083,422285,422287,421949,422273,422301,
            421294,421649,421697,421738,421662,421270,421517,421635,421777,
            421309,421260,421543,421722,421336];
    elif hash1 eq 2617 and hash2 eq 457 and hash3 eq 1895 then
        V:=[19,1,20,9,21,19,22,29,23,39,24,47,25,51,26,57];
        D:=[1,53/2,1,57/2,1,65/2,1,0,4,23,1,80/3,1,29,2,92/3,1,0,9,91/4,1,139/6,
            1,47/2,1,51/2,1,0,13,85/4,1,64/3,1,65/3,2,82/3,1,0,18,71/4,2,119/6,
            1,131/6,1,0,22,35/2,1,0,23,193/12,2,33/2,1,0,26,179/12,1,0];
        I:=[253896,253883,253924,253910,253686,253724,253875,253691,253706,
            253919,253723,253905,253679,253721,253730,253871,253502,253509,
            253717,253490,253503,253722,253511,253788,253508,253499];
    elif hash1 eq 361 and hash2 eq 505 and hash3 eq 3400 then
        V:=[22,1,23,13,24,19,25,29,26,43,27,57,28,69,29,77];
        D:=[1,24,1,82/3,1,89/3,1,125/4,1,95/3,1,0,6,119/6,1,145/6,1,0,8,55/3,1,
            62/3,1,111/5,1,121/5,1,0,12,101/6,1,35/2,1,551/30,1,221/12,1,37/2,1,
            187/10,1,0,18,175/12,1,33/2,1,50/3,1,17,1,52/3,1,53/3,1,0,24,157/12,
            1,451/30,2,181/12,1,461/30,1,241/12,1,0,30,77/6,3,787/60,1,66/5,1,0,
            35,117/10,2,0];
        I:=[513796,513611,513780,513790,513779,513763,513811,513722,513856,
            513633,513634,513723,513809,513692,513810,513825,513659,513717,
            513807,513818,513614,513860,513776,513870,513720,513761,513310,
            513660,513305,513892,513910,513913,513727,513721,513713,513739];
    elif hash1 eq 1207 and hash2 eq 121 and hash3 eq 3229 then
        V:=[26,1,27,9,28,19,29,23,30,27,31,35,32,41,33,49];
        D:=[1,70/3,1,82/3,1,125/4,1,0,4,237/10,1,157/6,1,169/6,1,181/6,1,0,8,
            121/5,3,0,11,187/10,1,0,12,175/12,1,15,1,369/20,1,0,15,391/30,1,
            27/2,1,0,17,66/5,1,203/15,2,14,1,0,21,677/60,1,485/42,1,361/30,1,0];
        I:=[398001,397027,399691,397377,400203,399948,400258,395915,398685,
            398699,398379,398731,399994,395637,398700,400177,397378,398886,
            399326,399327,399951,395930,396449];
    elif hash1 eq 925 and hash2 eq 3277 and hash3 eq 2642 then
        V:=[23,1,24,9,25,17,26,25,27,35];
        D:=[1,115/6,1,151/6,1,51/2,2,0,5,53/3,3,106/5,1,85/4,1,0,10,177/10,1,
            79/4,1,119/6,1,0,13,44/3,1,79/5,1,191/12,2,215/12,1,0,18,431/30,1,
            29/2,1,0];
        I:=[418791,417262,418179,418300,418797,418840,418862,417294,415936,
            417153,417010,418152,418207,417151,418161,418518,415071,416504,
            418357];
    elif hash1 eq 1113 and hash2 eq 2522 and hash3 eq 2586 then
        V:=[26,1,27,5,28,19,29,25,30,33,31,39,32,55];
        D:=[1,71/3,1,0,2,97/6,1,119/6,1,217/10,1,131/6,1,133/6,2,143/6,1,0,9,
            56/3,1,101/5,1,0,11,89/6,1,33/2,1,101/6,1,0,14,175/12,1,33/2,1,0,16,
            71/6,1,541/42,1,389/30,1,391/30,1,157/12,1,27/2,2,83/6,1,0,24,
            173/15,2,139/12,1,0];
        I:=[408062,407249,408065,407817,407677,406930,407014,407095,407773,
            407256,404402,407038,406954,404131,404557,407426,405282,405822,
            407227,405188,407738,407981,407879,407425,408057,407421];
    elif hash1 eq 2899 and hash2 eq 1225 and hash3 eq 2333 then
        V:=[20,1,21,9,22,19,23,33,24,47,25,57,26,63];
        D:=[1,21,1,74/3,1,80/3,1,0,4,115/6,1,127/6,1,151/6,1,175/6,1,0,8,59/3,2,
            20,1,65/3,1,116/5,1,71/3,1,24,1,0,15,71/4,1,107/6,1,109/6,2,233/12,
            1,197/10,1,155/6,1,0,22,268/15,1,91/5,1,292/15,1,298/15,1,0,26,
            147/10,1,101/6,1,0,28,175/12,1,223/15,1,15,3,253/15,1,121/6,1,0];
        I:=[424050,424126,423759,424062,424122,424054,422961,423958,424006,
            423825,423198,423290,422741,423809,424071,424070,422756,423318,
            424108,422771,422618,423029,422447,422637,422675,423389,423393,
            424076,423018,422347,424075,424204,422373,422346];
    elif hash1 eq 549 and hash2 eq 1420 and hash3 eq 253 then
        V:=[21,1,22,7,23,17,24,25,26,29,27,37];
        D:=[1,169/6,1,179/6,1,0,3,26,1,79/3,1,80/3,1,335/12,1,0,7,45/2,1,137/6,
            2,49/2,1,0,11,21,1,0,12,233/15,1,187/12,1,263/15,2,0,16,31/2,1,0];
        I:=[425980,426037,426061,425983,425714,425716,426031,425981,425982,
            425836,425724,425717,425718,425680,425683,425661];
    elif hash1 eq 831 and hash2 eq 368 and hash3 eq 2709 then
        V:=[18,1,19,5,20,15,21,25,22,33,23,41];
        D:=[1,30,1,0,2,45/2,1,49/2,4,53/2,3,57/2,1,0,11,21,1,68/3,1,23,7,25,2,0,
            22,39/2,1,127/6,3,43/2,1,0,27,18,1,58/3,1,22,2,0,31,107/6,1,0];
        I:=[674536,674537,674469,674531,674532,674540,674467,674489,674533,
            674499,674464,674497,674486,674488,674491,674492,674493,674538,
            674539,674403,674421,674487,674409,674420,674506,674485,674401,
            674507,674381,674419,674400];
    elif hash1 eq 1865 and hash2 eq 1122 and hash3 eq 207 then
        V:=[22,1,23,11,24,19,25,25,27,33];
        D:=[1,21,1,80/3,1,27,2,86/3,1,0,6,115/6,1,127/6,1,151/6,1,0,9,47/3,1,
            53/3,1,0,11,95/6,1,109/6,2,257/12,1,0,15,29/2,2,0];
        I:=[418007,418871,418549,418828,418835,418900,417950,417938,418774,
            418773,416917,418000,418461,417465,418770,418883];
    elif hash1 eq 3040 and hash2 eq 687 and hash3 eq 178 then
        V:=[24,1,25,7,26,13,27,19,28,23,29,31,30,39];
        D:=[1,21,1,23,2,0,4,35/2,1,115/6,1,0,6,53/3,1,58/3,1,0,8,95/6,1,0,9,
            167/12,1,43/3,1,44/3,1,0,12,371/30,3,25/2,3,431/30,1,0,19,34/3,1,0];
        I:=[406059,406027,406204,406028,406182,405506,405903,406174,406207,
            405946,405888,405607,405892,405967,405435,405895,405945,404839,
            405880];
    elif hash1 eq 1865 and hash2 eq 2617 and hash3 eq 1690 then
        V:=[15,1,16,9,17,13,18,17,19,35,20,43,21,51,22,57,23,63,25,69,26,75,27,
            83];
        D:=[1,69/2,1,73/2,1,89/2,1,0,4,128/3,1,0,5,259/6,1,0,6,83/3,1,95/3,1,32,
            1,186/5,1,113/3,1,118/3,1,40,1,165/4,1,0,14,191/6,1,205/6,1,497/12,
            1,0,17,161/5,2,97/3,1,103/3,1,0,21,137/6,1,791/30,1,0,23,92/5,1,
            58/3,1,0,25,169/10,1,571/30,1,0,27,569/42,1,137/10,1,0,29,1426/105,
            1,1145/84,1,95/6,1,0,32,991/84,1,0];
        I:=[518550,518480,518571,518506,518489,518570,518383,518397,518509,
            518510,518534,518464,518495,518527,518502,518469,518385,518511,
            518536,518431,518496,518346,518548,518393,518568,518436,518239,
            518260,518227,518246,518299,518240];
    elif hash1 eq 1489 and hash2 eq 1937 and hash3 eq 2250 then
        V:=[25,1,26,9,27,17,28,21,29,27];
        D:=[1,127/6,1,139/6,4,151/6,1,0,7,77/4,1,59/3,2,20,1,0,11,177/10,1,0,12,
            81/5,1,49/3,2,0,15,89/6,1,0];
        I:=[671973,671273,671456,671460,671982,671493,671426,671834,671836,
            671830,671458,671465,671272,671325,671899];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 33 mod 47.
function cached_ids_33(P,hash1,hash2,hash3)
    if hash1 eq 691 and hash2 eq 447 and hash3 eq 2528 then
        V:=[20,1,22,5,23,11,24,21,25,31,28,41];
        D:=[1,98/3,1,0,2,77/3,2,101/3,1,0,5,87/4,1,257/10,1,155/6,2,297/10,1,0,
            10,55/3,1,263/12,2,68/3,1,180/7,1,0,15,113/6,1,203/10,1,245/12,1,
            207/10,1,0,19,67/5,1,208/15,1,14,1,233/15,1,47/3,1,0];
        I:=[423148,423144,423201,422405,422775,423076,422905,423282,422393,
            422938,423111,423328,423713,422398,423714,422764,423326,423710,
            422409,423320,423717,422494,422525];
    elif hash1 eq 362 and hash2 eq 3109 and hash3 eq 2100 then
        V:=[22,1,23,7,24,11,25,19,26,23,27,27,29,31,31,39];
        D:=[1,89/3,1,125/3,1,0,3,179/6,1,0,4,328/15,1,67/3,1,141/5,2,0,8,671/30,
            1,0,9,313/15,1,0,10,571/30,1,0,11,311/30,1,76/5,1,46/3,1,0,14,
            2117/210,2,101/10,1,143/14,1,307/30,1,0];
        I:=[539670,539659,539662,539663,539668,539656,539665,539648,539647,
            539642,539702,539649,539658,539641,539650,539651,539669,539667];
    elif hash1 eq 2759 and hash2 eq 1544 and hash3 eq 2220 then
        V:=[22,1,23,9,24,17,25,25,27,31,29,35];
        D:=[1,27,2,31,2,98/3,1,0,6,127/6,1,51/2,1,199/6,1,0,9,65/3,1,71/3,1,
            83/3,1,0,12,131/6,1,143/6,1,0,14,113/6,1,0,15,15,1,0];
        I:=[410442,410673,410375,410885,410746,410576,411265,411236,411540,
            410747,411533,411546,410113,409981,409412];
    elif hash1 eq 2759 and hash2 eq 538 and hash3 eq 3359 then
        V:=[16,1,17,9,18,17,19,23,23,29];
        D:=[1,30,1,32,2,36,2,0,6,45/2,2,49/2,1,73/2,1,0,10,21,1,25,2,0,13,127/6,
            2,55/2,1,0,16,101/6,1,0];
        I:=[516794,516725,516792,516712,516713,516787,516790,516756,516681,
            516793,516446,516487,516679,516690,515955,516419];
    elif hash1 eq 174 and hash2 eq 1720 and hash3 eq 419 then
        V:=[16,1,17,7,18,15,19,23,20,29,21,35,22,45,24,51];
        D:=[1,36,1,44,1,0,3,61/2,1,65/2,1,69/2,1,0,6,29,1,98/3,1,104/3,1,0,9,
            151/6,2,163/6,1,0,12,59/3,1,77/3,1,0,14,71/4,1,119/6,1,131/6,1,
            329/12,1,0,18,215/12,1,68/3,1,0,20,179/12,1,0];
        I:=[516542,516556,516461,516499,516438,516606,516503,516580,516577,
            516608,516442,516572,516648,516439,516471,516573,515876,516521,
            516191,516443];
    elif hash1 eq 2242 and hash2 eq 846 and hash3 eq 1778 then
        V:=[24,1,26,5,27,11,28,17,29,21,30,29];
        D:=[1,47/3,1,0,2,43/3,1,268/15,1,0,4,431/30,2,173/12,1,0,7,13,1,0,8,
            259/20,2,83/6,1,85/6,1,0,12,173/15,1,363/28,1,274/21,1,787/60,1,
            79/6,1,53/4,1,0];
        I:=[502814,503327,502229,502817,503211,503229,503187,502939,503110,
            502546,502555,501828,501800,501758,501760,501987,502582];
    elif hash1 eq 2383 and hash2 eq 1979 and hash3 eq 3361 then
        V:=[21,1,22,9,23,13,24,19,27,29,29,37];
        D:=[1,187/6,1,223/6,1,75/2,1,0,4,95/3,2,0,6,143/6,1,191/6,1,0,8,358/15,
            1,24,1,121/5,1,73/3,1,0,12,237/14,1,205/12,2,37/2,1,0,16,569/42,1,
            137/10,4,0];
        I:=[419336,419881,419896,419338,419899,419346,419883,419357,419882,
            419890,419884,418980,418986,419344,418964,419063,419880,419892,
            419893,419895];
    elif hash1 eq 2289 and hash2 eq 1829 and hash3 eq 2599 then
        V:=[22,1,23,9,24,29,25,41,26,55,27,67,28,81,29,97,30,109,31,115];
        D:=[1,64/3,1,149/4,1,113/3,1,0,4,87/4,1,155/6,1,329/12,1,111/4,1,353/12,
            1,297/10,1,179/6,1,193/6,1,401/12,1,0,13,328/15,1,74/3,1,311/12,1,
            28,1,448/15,1,0,18,183/10,1,75/4,1,125/6,1,671/30,1,137/6,2,155/6,1,
            0,25,17,1,369/20,1,422/21,1,313/15,1,21,1,0,30,449/30,1,181/12,1,
            31/2,1,103/6,1,209/12,1,419/20,1,0,36,74/5,1,76/5,1,61/4,1,582/35,1,
            394/21,1,113/6,1,287/15,1,0,43,793/60,1,271/20,1,55/4,1,373/24,1,
            157/10,1,0,48,335/28,1,73/6,1,0,50,431/36,1,0];
        I:=[509843,509220,510040,509971,509140,510202,509236,509659,509783,
            509640,510129,509720,510023,510137,510147,510175,510150,509638,
            510012,510013,509929,509673,509819,508887,509093,510151,508514,
            509902,508397,510092,509966,509068,509044,510055,510200,510089,
            510077,510136,508865,508675,508875,508876,510113,509823,509983,
            508411,508412,508867,508525,508307];
    elif hash1 eq 1255 and hash2 eq 917 and hash3 eq 2040 then
        V:=[26,1,27,7,28,21,29,33,30,45,31,61,32,71];
        D:=[1,70/3,1,71/3,1,0,3,107/6,1,121/6,2,257/12,1,131/6,2,133/6,1,281/12,
            1,0,11,215/12,1,91/5,1,56/3,1,298/15,2,239/12,2,0,18,101/6,2,103/6,
            1,539/30,1,551/30,1,611/30,1,0,24,175/12,2,223/15,1,179/12,1,15,3,
            76/5,1,61/4,1,329/20,1,0,34,391/30,1,161/12,1,27/2,4,59/4,1,0,41,
            139/12,1,178/15,1,0];
        I:=[240259,239434,240171,239399,239470,239925,239218,239673,240228,
            239929,239197,239583,239633,239744,239745,238794,239828,239653,
            239661,239396,238567,238445,239326,240021,240170,239582,238295,
            238832,239676,239891,238613,238905,238561,240502,239958,239306,
            239340,239948,240503,238290,238210,238305];
    elif hash1 eq 1960 and hash2 eq 2857 and hash3 eq 2461 then
        V:=[27,1,28,7,29,15,30,23,31,31];
        D:=[1,143/6,1,145/6,2,0,4,55/3,3,20,1,61/3,5,0,13,103/6,1,221/12,1,37/2,
            1,0,16,44/3,2,179/12,1,227/12,1,0,20,13,1,27/2,1,0];
        I:=[239608,238480,238514,238842,238855,239651,239664,239220,239838,
            240172,240334,240471,239654,238600,238494,238891,239650,238535,
            238319,240485,240512];
    elif hash1 eq 1161 and hash2 eq 2735 and hash3 eq 1687 then
        V:=[24,1,25,9,27,17,28,23,29,29,30,35,31,41];
        D:=[1,71/3,1,136/5,1,94/3,1,0,4,121/6,1,237/10,1,143/6,1,0,7,67/4,1,
            283/14,1,0,9,199/12,1,253/15,1,0,11,401/30,2,27/2,1,0,14,66/5,2,
            281/21,1,0,17,241/18,1,0];
        I:=[502466,502500,502536,502465,502510,502530,502439,501451,502436,
            502461,502335,502438,502471,502407,502472,501455,500975];
    elif hash1 eq 1584 and hash2 eq 766 and hash3 eq 2755 then
        V:=[24,1,25,13,26,19,27,23,28,27,29,35];
        D:=[1,83/3,1,117/4,1,89/3,1,95/3,1,101/3,1,0,6,49/2,1,155/6,2,0,9,55/3,
            1,0,10,67/4,1,0,11,203/12,1,52/3,1,227/12,1,0,14,27/2,1,91/6,1,0];
        I:=[418757,418755,418681,418705,418735,418682,416273,418700,418701,
            418743,416262,418750,418733,417765,416274];
    elif hash1 eq 2759 and hash2 eq 1266 and hash3 eq 1103 then
        V:=[27,1,28,5,29,15,30,21,31,29,32,37];
        D:=[1,139/6,1,0,2,18,1,65/3,1,22,1,70/3,1,0,6,97/6,2,109/6,2,0,10,
            167/12,1,71/5,2,239/12,1,0,14,431/30,1,89/6,1,197/12,1,0,17,249/20,
            1,155/12,1,40/3,1,0];
        I:=[233164,232178,230674,232183,233763,231942,233257,232189,233765,
            233458,229911,232293,230153,231627,230480,229860,229341,229259,
            233567];
    elif hash1 eq 268 and hash2 eq 34 and hash3 eq 398 then
        V:=[24,1,25,13,26,25,27,35,28,43,29,47,30,51];
        D:=[1,80/3,1,29,1,31,1,104/3,1,35,1,0,6,115/6,1,163/6,1,55/2,1,175/6,1,
            187/6,1,0,11,71/3,2,101/4,1,76/3,2,82/3,2,0,18,209/12,1,197/10,1,
            119/6,1,0,21,238/15,1,0,22,221/12,1,0,23,218/15,1,0];
        I:=[236496,236321,236506,235913,236532,236380,236525,236019,235941,
            236480,235486,236031,235468,235945,236485,235963,235977,236167,
            235469,235916,236105,235991,236023];
    elif hash1 eq 2430 and hash2 eq 1255 and hash3 eq 3020 then
        V:=[19,1,20,7,21,17,22,25,23,41,24,59,25,71,26,77,27,93,29,99];
        D:=[1,73/2,1,81/2,1,0,3,98/3,1,33,1,35,1,110/3,1,0,7,151/6,1,51/2,1,
            59/2,2,0,11,71/3,1,101/4,1,77/3,4,26,1,89/3,1,94/3,1,95/3,3,0,23,
            257/12,1,131/6,1,95/4,1,143/6,2,145/6,1,257/10,1,155/6,3,157/6,2,0,
            35,298/15,1,101/5,1,61/3,1,328/15,1,263/12,2,0,41,221/12,1,37/2,1,0,
            43,253/15,1,369/20,1,37/2,2,223/12,1,283/15,2,227/12,2,19,3,0,55,
            853/60,1,15,1,0,57,117/10,1,0];
        I:=[253980,253799,253763,253983,253795,253806,253785,253733,253797,
            253798,253698,253530,253545,253582,253969,253970,253976,253469,
            253535,253602,253802,253811,253527,253782,253528,253665,253813,
            253603,253473,253606,253794,253804,253471,253587,253543,253541,
            253668,253460,253459,253599,253529,253489,253538,253483,253458,
            253485,253484,253463,253481,253455,253464,253462,253480,253601,
            253593,253521,253522];
    elif hash1 eq 174 and hash2 eq 2504 and hash3 eq 2560 then
        V:=[24,1,25,5,26,9,27,17,28,21,29,39];
        D:=[1,93/4,1,0,2,95/4,1,0,3,20,1,81/4,1,61/3,1,0,6,659/30,1,0,7,129/7,1,
            277/15,1,223/12,4,283/15,2,58/3,1,422/21,1,247/12,1,21,1,0,19,
            451/30,1,241/12,1,0];
        I:=[415323,415514,415261,415190,415322,415226,415174,415160,415153,
            415157,415162,415185,414982,415274,415268,415244,415324,415316,
            414951,415121];
    elif hash1 eq 1725 and hash2 eq 3417 and hash3 eq 2022 then
        V:=[29,1,30,7,31,13,32,25];
        D:=[1,97/6,1,109/6,3,0,5,49/3,2,50/3,3,0,10,173/12,1,29/2,1,147/10,1,
            89/6,5,91/6,2,0,20,13,1,0];
        I:=[663867,663522,663536,663572,663557,663561,663505,663521,663974,
            663602,663854,663532,663443,663506,663528,663671,663746,663226,
            663585,663882];
    elif hash1 eq 127 and hash2 eq 3104 and hash3 eq 3011 then
        V:=[21,1,22,5,23,11,24,19,25,31,26,39,28,45,29,51];
        D:=[1,175/6,1,0,2,89/3,1,95/3,1,0,4,131/6,1,145/6,1,305/12,1,0,7,62/3,1,
            328/15,1,111/5,1,89/4,2,97/4,1,0,13,217/12,1,113/6,2,245/12,1,0,17,
            253/15,2,278/15,1,0,20,203/15,1,61/4,1,0,22,47/4,2,137/10,1,0];
        I:=[421867,421535,421496,421855,421854,421441,421529,421439,421354,
            421850,421857,421446,421344,421447,421755,421993,421391,421396,
            421356,421522,421160,421534,421569,421111];
    elif hash1 eq 2101 and hash2 eq 2377 and hash3 eq 1733 then
        V:=[16,1,17,7,19,17,20,27,22,31,24,35,25,39];
        D:=[1,68/3,1,80/3,1,0,3,91/4,1,139/6,1,47/2,1,99/4,1,0,7,87/4,1,143/6,1,
            257/10,1,239/6,1,0,11,55/3,2,0,13,506/21,1,0,14,183/10,1,0,15,55/3,
            1,227/12,1,0];
        I:=[518739,518572,518656,518751,518614,518503,518546,518352,518471,
            518247,518616,518661,518210,518205,518188,518182];
    elif hash1 eq 1349 and hash2 eq 380 and hash3 eq 2317 then
        V:=[22,1,23,9,24,17,25,21,27,25,28,31];
        D:=[1,83/3,2,28,1,89/3,2,0,6,257/10,1,157/6,2,167/6,1,0,10,67/3,1,0,11,
            611/30,1,0,12,461/30,1,521/30,1,0,14,309/20,1,233/15,1,0];
        I:=[252874,252907,252875,252636,252914,252576,252635,252910,252596,
            252913,252693,252595,252536,252489,252486];
    elif hash1 eq 1866 and hash2 eq 2296 and hash3 eq 1684 then
        V:=[19,1,20,5,21,11,22,15,23,25,24,29,26,37,27,43,28,47];
        D:=[1,49/2,2,0,3,27,1,98/3,1,0,5,123/4,1,0,6,20,1,65/3,1,77/3,1,100/3,1,
            0,10,87/4,1,0,11,239/12,1,81/4,1,263/12,1,0,14,17,1,361/20,1,0,16,
            137/10,1,0,17,203/15,1,0];
        I:=[512516,512545,512541,512670,512524,512519,512537,512501,512504,
            512508,512522,512485,512505,511826,511796,512468,511903];
    elif hash1 eq 503 and hash2 eq 401 and hash3 eq 1005 then
        V:=[24,1,25,7,26,17,27,25,28,33,29,41,30,47];
        D:=[1,25,1,27,1,0,3,127/6,1,139/6,2,47/2,1,151/6,2,0,9,77/4,1,59/3,3,24,
            1,0,14,209/12,1,107/6,1,109/6,1,0,17,49/3,1,50/3,1,215/12,1,0,20,
            169/12,1,173/12,1,0,22,15,1,0];
        I:=[240968,240974,240970,240980,240982,241008,239479,239886,240994,
            240992,241004,241032,239623,239469,240966,240965,241053,241011,
            238902,239447,239480,238171];
    elif hash1 eq 268 and hash2 eq 369 and hash3 eq 2999 then
        V:=[22,1,23,7,24,15,25,27,26,35,27,45,28,49,29,53];
        D:=[1,80/3,1,31,1,0,3,151/6,1,51/2,1,175/6,1,0,6,65/3,3,116/5,1,70/3,1,
            24,1,76/3,1,0,13,119/6,1,121/6,1,217/10,1,0,16,16,1,49/3,1,50/3,1,
            18,1,0,20,551/30,1,0,21,15,1,0,22,391/30,2,0];
        I:=[250780,251540,250747,251229,250243,250779,251362,251568,250246,
            250880,250398,250778,250792,250187,249900,250785,250829,251584,
            251373,249890,250330,250176,250238];
    elif hash1 eq 644 and hash2 eq 3454 and hash3 eq 504 then
        V:=[22,1,23,7,24,17,25,23,26,29,27,33];
        D:=[1,31,1,33,1,0,3,127/6,1,151/6,1,59/2,2,187/6,2,0,9,53/3,1,26,1,0,11,
            107/6,1,109/6,1,0,13,238/15,1,0,14,359/20,1,551/30,2,221/12,1,113/6,
            1,115/6,1,0];
        I:=[412007,412000,412253,411968,409596,410662,410047,412207,410430,
            409555,412239,412242,411113,409257,409141,409595,409597,409918,
            409893];
    elif hash1 eq 268 and hash2 eq 1538 and hash3 eq 1790 then
        V:=[25,1,26,11,27,17,28,21,29,27,30,33];
        D:=[1,127/6,1,107/4,1,163/6,1,199/6,1,0,5,126/5,1,77/3,1,0,7,119/6,1,0,
            8,65/4,1,49/3,1,0,10,431/30,1,351/20,1,0,12,44/3,1,223/15,2,199/12,
            1,0];
        I:=[398564,398430,398443,398381,398390,398407,398343,398352,399484,
            398588,395641,399575,398431,399898,395952];
    elif hash1 eq 2195 and hash2 eq 1691 and hash3 eq 3235 then
        V:=[19,1,20,5,21,11,22,17,23,25,24,37,25,41,26,45,27,51];
        D:=[1,61/2,1,0,2,27,1,37,1,0,4,175/6,2,59/2,1,0,7,70/3,2,30,1,94/3,1,0,
            11,131/6,1,133/6,1,281/12,1,143/6,1,305/12,1,0,16,39/2,1,0,17,
            391/20,1,0,18,253/15,2,52/3,2,0,22,201/10,1,0];
        I:=[424055,423779,423885,423780,423781,423787,423761,423771,423802,
            423849,423798,423800,423769,423784,422719,423937,422671,422738,
            422805,422774,423803,422348];
    elif hash1 eq 2571 and hash2 eq 2246 and hash3 eq 2188 then
        V:=[23,1,24,9,25,17,26,27,27,43];
        D:=[1,43/2,1,139/6,1,47/2,3,0,6,18,1,59/3,2,65/3,2,0,11,107/6,1,109/6,2,
            119/6,1,121/6,1,0,16,65/4,1,49/3,4,50/3,4,17,1,91/5,1,73/4,2,55/3,2,
            0,31,89/6,4,0];
        I:=[247301,247304,246958,246960,247316,247470,247250,247482,247223,
            247237,247285,246896,247457,247008,247321,247222,247236,247247,
            247259,247308,247017,247043,247136,247402,246951,246928,246507,
            246528,246891,246924,246784,246832,246919,246927];
    elif hash1 eq 1255 and hash2 eq 1518 and hash3 eq 1468 then
        V:=[26,1,27,7,28,13,29,19];
        D:=[1,59/3,2,20,1,0,4,107/6,1,109/6,1,0,6,49/3,1,50/3,2,0,9,59/4,2,89/6,
            6,0];
        I:=[669819,669870,669752,669872,669836,669820,669704,669825,669697,
            669726,669713,669823,669827,669850,669853,669856];
    elif hash1 eq 1537 and hash2 eq 1199 and hash3 eq 3444 then
        V:=[26,1,27,11,28,23,30,45];
        D:=[1,53/3,1,58/3,1,59/3,1,20,3,0,7,33/2,1,177/10,1,107/6,1,109/6,1,
            119/6,1,0,12,44/3,1,15,1,238/15,1,16,1,65/4,1,49/3,2,17,1,89/5,1,
            73/4,1,55/3,2,0,24,193/15,1,66/5,2,40/3,3,41/3,2,0];
        I:=[239203,240328,239553,239073,239579,240423,240369,239593,239777,
            240166,240180,239680,239679,239595,240121,239529,239183,240157,
            239620,238626,240217,238907,240248,240367,239548,240380,239483,
            239519,239736,239465,240495];
    elif hash1 eq 1631 and hash2 eq 702 and hash3 eq 3269 then
        V:=[25,1,26,11,27,19,28,29,29,35,30,43];
        D:=[1,35/2,1,115/6,1,139/6,1,47/2,1,0,5,52/3,1,59/3,1,20,1,0,8,63/4,2,
            107/6,4,109/6,1,197/10,1,0,16,16,1,49/3,2,0,19,149/12,1,51/4,1,14,1,
            0,22,151/12,2,44/3,1,0];
        I:=[408597,408707,407048,408502,408653,408507,408539,408637,408638,
            406940,407948,408573,408746,408525,405746,408599,406948,407635,
            405853,408415,408536,408282,408283,404715];
    elif hash1 eq 2853 and hash2 eq 2162 and hash3 eq 1820 then
        V:=[25,1,26,7,27,15,28,23,29,31,30,39];
        D:=[1,139/6,1,151/6,1,0,3,52/3,1,20,2,71/3,1,0,7,107/6,1,121/6,1,131/6,
            1,0,10,49/3,1,55/3,1,56/3,1,0,13,29/2,1,89/6,1,101/6,1,0,16,179/12,
            1,0];
        I:=[240287,240161,239944,239997,240432,238944,239076,239268,238360,
            239854,240211,238621,238982,238967,239075,238316];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 34 mod 47.
function cached_ids_34(P,hash1,hash2,hash3)
    if hash1 eq 1585 and hash2 eq 915 and hash3 eq 2206 then
        V:=[15,1,16,5,17,19,18,29,19,39,20,53,21,73,22,91,23,97,24,101,25,109];
        D:=[1,73/2,1,0,2,27,1,86/3,1,29,2,31,3,98/3,1,33,1,0,11,51/2,1,163/6,2,
            55/2,2,175/6,2,0,18,22,1,71/3,2,76/3,2,28,1,0,24,133/6,1,145/6,2,
            257/10,1,103/4,1,155/6,1,169/6,1,0,31,19,1,81/4,1,61/3,1,62/3,2,
            263/12,2,22,1,89/4,1,68/3,2,311/12,1,0,43,37/2,2,113/6,2,611/30,1,
            41/2,1,207/10,1,125/6,1,659/30,1,45/2,1,0,53,17,3,271/12,1,0,57,
            95/6,1,0,58,233/15,1,238/15,1,16,1,0,61,25/2,1,0];
        I:=[255005,255002,254990,255003,255006,254968,255000,255004,254982,
            254999,254978,254974,254989,254995,255001,254976,254997,254985,
            254966,254983,254954,254975,254945,254981,254928,254946,254934,
            254956,254957,254926,254947,254944,254941,254964,254965,254960,
            254963,254958,254933,254961,254962,254920,254953,254959,254949,
            254951,254937,254927,254939,254938,254940,254921,254930,254972,
            254993,254914,254922,254916,254917,254919,254936];
    elif hash1 eq 1726 and hash2 eq 2569 and hash3 eq 1418 then
        V:=[24,1,25,7,26,13,27,21,28,27];
        D:=[1,21,3,23,2,0,6,115/6,1,39/2,1,0,8,53/3,3,18,1,59/3,1,0,13,97/6,2,
            33/2,1,0,16,44/3,1,0];
        I:=[61466,61467,61510,61441,61477,61509,61491,61446,61488,61490,61445,
            61400,61440,61444,61434,61403];
    elif hash1 eq 2102 and hash2 eq 3107 and hash3 eq 18 then
        V:=[14,1,15,5,16,13,17,25,18,41,19,59,20,79,21,91,22,107,23,113,24,125,
            25,129,26,135];
        D:=[1,40,1,0,2,69/2,1,73/2,1,81/2,2,0,6,104/3,1,35,1,37,1,39,1,41,1,0,
            11,175/6,1,187/6,1,199/6,3,67/2,1,211/6,1,71/2,1,223/6,1,0,20,77/3,
            1,136/5,1,88/3,1,89/3,1,94/3,2,100/3,2,101/3,1,106/3,1,0,30,87/4,1,
            131/6,1,133/6,1,305/12,3,111/4,1,169/6,1,179/6,1,181/6,1,143/4,1,0,
            41,358/15,1,121/5,1,73/3,1,388/15,1,311/12,1,0,46,551/30,1,611/30,1,
            439/20,1,269/12,2,45/2,1,137/6,4,26,1,0,57,369/20,1,21,1,0,59,
            185/12,2,31/2,1,511/30,1,679/30,1,689/30,1,0,65,61/4,2,0,67,433/28,
            1,653/42,1,0,69,181/15,2,0];
        I:=[428318,428208,428214,428171,428219,428215,428314,428209,428192,
            428262,427851,428334,428159,428175,428226,428199,428203,428202,
            428206,428225,427831,428222,428326,428228,428335,428224,428371,
            428185,427906,428285,427890,427854,428188,428189,428223,428122,
            428200,427891,427801,427863,427861,427862,427855,427876,427858,
            428077,427857,428124,427864,428107,427902,427887,427895,427907,
            428100,427782,428134,427901,428268,428383,428183,427849,427623,
            427625,428180,428190,427670,427633,427910,427992];
    elif hash1 eq 2807 and hash2 eq 2060 and hash3 eq 3450 then
        V:=[23,1,25,5,26,11];
        D:=[1,45/2,1,0,2,103/6,6,35/2,5,0,13,47/3,2,16,1,0];
        I:=[671073,670827,670828,670997,671032,671046,671079,670442,670503,
            670934,671057,671090,670961,670962,670490];
    elif hash1 eq 316 and hash2 eq 2793 and hash3 eq 2346 then
        V:=[23,1,24,9,25,21,26,31,27,39,29,43];
        D:=[1,151/6,1,51/2,1,59/2,1,0,4,20,1,65/3,1,116/5,1,71/3,1,77/3,1,0,9,
            71/4,1,217/10,1,45/2,1,237/10,1,0,13,56/3,2,101/5,1,62/3,1,0,17,
            147/10,1,0,18,13,1,0];
        I:=[418197,418191,418163,418278,418271,417240,417421,418235,417532,
            417187,418360,417214,416365,417426,416673,418258,416372,416390];
    elif hash1 eq 3230 and hash2 eq 2903 and hash3 eq 1726 then
        V:=[16,1,17,7,18,13,19,17,20,27,22,39,23,45,24,49,25,53,26,57];
        D:=[1,24,1,36,1,0,3,53/2,2,61/2,1,0,6,39,1,0,7,187/6,1,199/6,1,139/4,1,
            211/6,2,0,12,53/3,1,76/3,1,94/3,1,95/3,1,106/3,1,0,17,16,1,49/3,1,0,
            19,551/30,1,0,20,275/12,1,0,21,67/4,1,0,22,187/12,1,0];
        I:=[516748,516530,516716,516773,516774,516480,516528,516707,516551,
            516162,516486,516698,516768,516529,516594,516210,516705,516723,
            516030,515705,516194,515702];
    elif hash1 eq 2619 and hash2 eq 1238 and hash3 eq 1236 then
        V:=[26,1,27,5,28,17,29,29,30,35,31,43,32,47];
        D:=[1,19,1,0,2,127/6,1,139/6,1,47/2,1,151/6,1,51/2,1,0,7,77/4,2,58/3,1,
            59/3,1,20,1,65/3,3,0,15,109/6,1,197/10,1,0,17,232/15,1,16,1,89/5,1,
            0,20,173/12,1,0,21,13,1,40/3,1,0];
        I:=[668855,668253,668621,668545,668620,668627,668161,668570,668149,
            668671,668739,668143,668625,668721,668702,668078,668075,668657,
            667939,668742,668763,668597];
    elif hash1 eq 1585 and hash2 eq 2158 and hash3 eq 1956 then
        V:=[20,1,21,11,22,23,23,31,24,41,25,53];
        D:=[1,92/3,3,31,2,98/3,1,35,2,0,9,39/2,1,99/4,1,151/6,1,55/2,1,175/6,1,
            0,14,65/3,1,70/3,1,77/3,2,0,18,71/4,1,37/2,1,119/6,1,121/6,3,0,24,
            73/4,1,55/3,1,56/3,1,62/3,1,89/4,1,0,29,193/12,1,101/6,1,221/12,1,
            113/6,1,0];
        I:=[424548,424769,424838,423901,424796,424421,424491,424814,424343,
            424521,424799,423906,424367,424414,424452,422889,424494,424525,
            424183,424395,424189,424366,424777,424353,424190,424828,423706,
            422535,423469,424357,422541,423461];
    elif hash1 eq 269 and hash2 eq 2482 and hash3 eq 1000 then
        V:=[19,1,20,9,21,19,22,27,23,35,24,41,25,47,26,53,28,65];
        D:=[1,49/2,1,65/2,1,85/2,1,0,4,74/3,1,27,1,92/3,1,116/3,1,0,8,127/6,1,
            139/6,1,247/6,1,0,11,77/3,1,83/3,1,95/3,1,0,14,119/6,1,133/6,1,0,16,
            56/3,1,62/3,1,0,18,167/10,1,41/2,1,0,20,72/5,2,223/15,1,248/15,1,
            278/15,1,223/12,1,0,26,196/15,2,0];
        I:=[512228,512219,512237,512291,512143,512093,511978,512230,512300,
            511883,511882,512131,512135,512236,511981,512117,512127,512108,
            511976,512188,512345,512293,511887,511975,511910,511952,512194];
    elif hash1 eq 739 and hash2 eq 2372 and hash3 eq 892 then
        V:=[18,1,20,5,21,9,22,13,23,25,24,37,25,41,27,45,28,49];
        D:=[1,51,1,0,2,113/3,1,0,3,191/6,1,0,4,359/12,1,478/15,1,383/12,1,32,1,
            455/12,1,0,9,479/20,1,731/30,1,791/30,1,419/14,1,63/2,1,0,14,307/12,
            1,0,15,493/24,1,0,16,725/42,1,0,17,1145/84,1,827/60,1,208/15,2,0];
        I:=[515879,515886,516250,515885,516219,516016,515882,515670,516193,
            516325,515643,515668,515691,515694,515641,515684,515761,516013,
            515884,516012];
    elif hash1 eq 2290 and hash2 eq 1524 and hash3 eq 224 then
        V:=[17,1,18,9,19,25,20,49,21,79,22,97,23,119,24,135,25,165,26,177,27,
            183];
        D:=[1,57/2,1,61/2,1,69/2,1,0,4,25,1,92/3,2,98/3,1,33,3,104/3,3,37,2,39,
            1,0,17,139/6,1,99/4,1,51/2,3,107/4,1,163/6,3,55/2,1,175/6,3,59/2,1,
            187/6,4,139/4,1,211/6,1,0,37,64/3,1,22,1,71/3,2,24,3,76/3,1,77/3,2,
            26,2,136/5,2,82/3,1,83/3,2,28,1,117/4,1,156/5,1,95/3,3,0,60,217/10,
            1,87/4,2,131/6,2,133/6,2,145/6,1,257/10,2,157/6,2,329/12,1,0,73,
            56/3,1,292/15,1,39/2,1,298/15,1,239/12,2,20,1,62/3,2,43/2,1,263/12,
            1,358/15,1,0,85,197/12,1,101/6,3,221/12,4,113/6,2,115/6,1,611/30,1,
            41/2,1,0,98,76/5,1,329/20,1,33/2,2,199/12,3,50/3,1,17,1,92/5,1,
            129/7,1,37/2,1,56/3,4,227/12,1,19,1,96/5,1,58/3,1,0,118,451/30,2,
            181/12,1,185/12,1,31/2,2,50/3,1,0,125,316/21,1,233/15,1,0,127,
            361/30,2,0];
        I:=[427524,427595,427525,427552,427364,427536,427592,426989,427012,
            427365,427030,427070,427473,427181,427385,427347,427466,427304,
            427027,427032,427561,427091,427196,427338,427593,427163,426999,
            427019,427430,427113,426993,427083,427151,427489,426836,427205,
            427207,427139,427081,427366,427089,427176,427391,427107,427417,
            427472,427009,427393,426832,426893,427018,427149,427409,427166,
            426672,426775,426621,426631,426795,426743,427140,427309,426607,
            427074,427117,427441,427086,426615,426616,426650,426723,426702,
            427545,426645,426808,426730,426860,427153,426584,427116,427451,
            426756,426669,426542,426839,426597,426620,426721,426481,426722,
            427011,427303,426748,427159,427141,426644,426944,426744,426664,
            426830,427020,426496,426589,427105,426494,427054,426746,426432,
            426509,426471,426489,426500,426524,426435,426476,426492,426656,
            426738,426747,426820,426720,426726,426770,426458,426459,426460,
            426701,426754];
    elif hash1 eq 1068 and hash2 eq 1319 and hash3 eq 1386 then
        V:=[25,1,26,11,27,23,28,35,29,43,30,47];
        D:=[1,35/2,1,115/6,2,39/2,3,43/2,1,0,8,52/3,1,53/3,4,77/4,1,59/3,1,20,1,
            0,16,29/2,1,157/10,1,95/6,1,97/6,2,33/2,1,0,22,167/12,1,43/3,4,44/3,
            2,0,29,147/10,1,0,30,35/3,2,0];
        I:=[241187,240742,240743,240903,241003,241113,240398,241204,239185,
            239600,239994,240692,240257,241073,239805,241395,240144,241261,
            240806,240920,239206,241384,239456,239735,240720,241365,238727,
            241007,238097,239783,239785];
    elif hash1 eq 3089 and hash2 eq 751 and hash3 eq 2633 then
        V:=[18,1,19,7,20,11,21,21,22,35,23,45,24,55,25,61,26,73,27,85,28,91];
        D:=[1,110/3,1,122/3,1,0,3,199/6,2,0,5,146/5,1,89/3,1,100/3,1,101/3,2,0,
            10,257/10,1,103/4,1,155/6,1,169/6,2,297/10,1,179/6,1,0,17,22,1,
            287/12,1,24,1,26,2,0,22,41/2,1,207/10,1,265/12,1,317/12,1,0,26,37/2,
            1,227/12,1,0,28,1117/60,1,56/3,1,571/30,2,229/12,1,581/30,2,0,35,
            89/6,1,91/6,3,77/5,1,233/15,2,47/3,2,0,44,46/3,1,563/30,1,0,46,
            707/60,2,71/6,1,0];
        I:=[427373,427132,426660,427491,426580,427322,427553,426759,427412,
            426783,426625,427559,426780,427450,426484,426806,427156,426815,
            427155,426727,427191,427044,426788,426719,426453,426523,426640,
            426758,426534,426564,426728,426776,426563,426810,426450,426735,
            426812,427046,426834,426677,426779,426842,427429,426461,426424,
            426703,426718,426755];
    elif hash1 eq 1068 and hash2 eq 2246 and hash3 eq 1960 then
        V:=[21,1,22,11,23,21,24,27,25,33];
        D:=[1,49/2,1,53/2,3,57/2,1,61/2,1,0,7,68/3,1,23,1,74/3,2,25,3,0,14,
            127/6,2,43/2,1,0,17,59/3,1,65/3,2,0,20,109/6,1,0];
        I:=[61917,61921,61922,61927,61928,61919,61925,61920,61901,61924,61904,
            61916,61918,61895,61923,61897,61903,61891,61902,61896];
    elif hash1 eq 1585 and hash2 eq 1791 and hash3 eq 1787 then
        V:=[10,1,11,7,13,13,14,19,15,25,16,39,17,43,18,59,19,75,20,81,21,85,22,
            103,23,113,24,121];
        D:=[1,46,1,62,1,0,3,101/2,1,109/2,1,0,5,247/6,1,259/6,2,0,8,176/5,1,
            206/5,1,0,10,297/10,1,135/4,2,425/12,1,215/6,1,417/10,1,263/6,1,0,
            17,568/15,1,0,18,671/30,1,339/14,1,731/30,1,63/2,1,679/20,1,409/12,
            1,433/12,1,0,25,470/21,1,392/15,1,561/20,1,433/15,1,95/3,1,34,1,
            681/20,1,0,32,619/30,1,691/30,1,0,34,400/21,1,0,35,917/60,1,373/24,
            1,937/60,1,63/4,1,61/3,1,1451/70,1,104/5,1,139/6,1,0,43,367/24,2,
            1481/84,1,436/21,1,104/5,1,0,48,1913/140,1,1327/84,1,319/20,1,0,51,
            184/15,1,0];
        I:=[544205,544212,544195,544192,544196,544163,544213,544183,544160,
            544188,544208,544209,544203,544193,544152,544151,544147,544148,
            544117,544111,544093,544167,544138,544164,544124,544139,544136,
            544171,544098,544141,544144,544137,544155,544123,544172,544089,
            544084,544206,544091,544106,544090,544110,544085,544095,544116,
            544097,544094,544099,544104,544101,544165];
    elif hash1 eq 1538 and hash2 eq 1978 and hash3 eq 951 then
        V:=[14,1,15,9,16,21,17,33,18,45,19,51,20,61,22,67];
        D:=[1,32,2,34,1,36,1,0,5,57/2,1,61/2,2,65/2,2,69/2,1,77/2,1,0,12,27,2,
            86/3,2,29,3,92/3,1,31,1,0,21,151/6,1,51/2,4,55/2,1,175/6,1,59/2,1,0,
            29,71/3,1,26,1,0,31,131/6,1,133/6,3,45/2,3,157/6,1,0,39,111/5,1,
            68/3,2,0,42,52/3,1,0];
        I:=[62053,62054,62055,62056,62050,62042,62049,62032,62045,62051,62026,
            62044,62052,62028,62031,62027,62040,62047,62025,62022,62029,62023,
            62024,62043,62046,62036,62033,62035,62037,62020,62021,62019,62034,
            62041,62030,62038,62048,62018,62015,62016,62017,62039];
    elif hash1 eq 1726 and hash2 eq 2153 and hash3 eq 2485 then
        V:=[22,1,23,5,24,9,25,17,26,21,28,29];
        D:=[1,19,1,0,2,43/2,1,0,3,52/3,1,59/3,1,64/3,2,0,7,33/2,2,0,9,31/2,1,
            49/3,2,298/15,1,0,13,40/3,1,72/5,1,223/15,1,0];
        I:=[413250,413224,413093,413216,413053,413214,413225,413226,413002,
            413033,413215,412718,412738,412885,412886];
    elif hash1 eq 1256 and hash2 eq 1466 and hash3 eq 2633 then
        V:=[24,1,25,11,26,19,27,25,28,33];
        D:=[1,74/3,1,80/3,1,27,1,29,1,0,5,43/2,1,151/6,1,163/6,1,0,8,65/3,3,
            76/3,1,0,12,79/4,1,119/6,1,121/6,2,0,16,55/3,1,0];
        I:=[238990,239868,239280,239149,239213,239042,240193,238986,239068,
            239865,239046,239056,239867,239904,240493,239243];
    elif hash1 eq 1679 and hash2 eq 2624 and hash3 eq 498 then
        V:=[21,1,22,5,23,11,24,25,25,39,26,53,27,63,28,67,29,71];
        D:=[1,65/2,1,0,2,74/3,1,104/3,1,0,4,139/6,1,151/6,1,163/6,1,55/2,1,
            175/6,1,67/2,1,0,10,106/5,1,65/3,3,93/4,2,71/3,1,77/3,1,83/3,1,0,19,
            197/10,1,257/12,1,217/10,1,131/6,1,133/6,1,145/6,1,0,25,239/12,1,
            101/5,1,81/4,2,61/3,1,0,30,491/30,1,0,31,253/15,1,0,32,451/30,1,
            461/30,1,0];
        I:=[250089,250423,250298,250419,250327,250148,250211,250264,250193,
            250657,250438,250458,250540,250277,250537,250210,249933,250303,
            250654,250146,250071,250267,250143,249929,250568,250366,250549,
            250563,250098,250544,249965,249872,249862];
    elif hash1 eq 1820 and hash2 eq 935 and hash3 eq 3333 then
        V:=[21,1,22,5,23,15,24,21,25,27,26,31,28,43];
        D:=[1,175/6,1,0,2,22,1,70/3,1,24,1,77/3,1,0,6,257/12,1,145/6,1,0,8,55/3,
            1,56/3,2,0,11,187/10,1,0,12,253/15,1,131/7,1,283/15,1,19,1,58/3,1,0,
            17,14,2,0];
        I:=[422720,422722,422785,422784,422743,422792,423112,423410,422795,
            422808,422744,422420,422475,422478,422408,422453,422758,422780];
    elif hash1 eq 269 and hash2 eq 2115 and hash3 eq 2222 then
        V:=[22,1,23,13,24,23,25,33,26,45];
        D:=[1,68/3,1,74/3,4,25,1,80/3,1,27,4,0,12,115/6,1,39/2,1,127/6,1,139/6,
            1,0,16,69/4,1,53/3,2,59/3,4,71/3,1,0,24,71/4,1,107/6,2,109/6,3,
            121/6,2,217/10,1,0,33,79/5,1,191/12,2,49/3,2,215/12,1,55/3,1,263/12,
            1,0];
        I:=[249547,248928,249491,249554,249575,249545,249424,249324,249346,
            249515,249687,249557,249542,248841,247779,248914,248788,249510,
            247786,248765,248772,249516,247610,248832,248854,249454,248818,
            249253,249605,247780,249252,247640,248109,247796,248879,248880,
            249325,247613,247600,247573];
    elif hash1 eq 2807 and hash2 eq 1991 and hash3 eq 2211 then
        V:=[28,1,29,5,30,15,31,23,32,33,33,37,34,45];
        D:=[1,101/4,1,0,2,109/6,1,87/4,1,143/6,1,145/6,1,0,6,55/3,1,20,1,61/3,1,
            0,9,167/10,1,67/4,1,103/6,1,221/12,1,0,13,40/3,1,0,14,51/4,1,157/12,
            1,79/6,2,0,18,239/21,1,12,1,0];
        I:=[232275,230690,230277,229925,233468,232486,229903,228991,229989,
            230411,231945,230025,234164,232199,233740,233469,233836,229886,
            229693];
    elif hash1 eq 1538 and hash2 eq 361 and hash3 eq 2694 then
        V:=[26,1,27,5,28,13,29,19,30,31,31,39];
        D:=[1,85/4,1,0,2,97/6,1,71/4,1,197/10,1,0,5,44/3,1,65/4,1,0,7,431/30,2,
            29/2,2,147/10,1,89/6,1,91/6,1,0,14,151/12,1,193/15,1,13,2,0,18,23/2,
            2,0];
        I:=[403841,403785,403893,403404,403316,403786,403319,403535,403411,
            403757,403458,403359,403619,403896,403674,403300,403538,403305,
            403326];
    elif hash1 eq 1303 and hash2 eq 2939 and hash3 eq 1386 then
        V:=[24,1,25,13,26,25,27,37,28,45,29,59,30,71,31,75];
        D:=[1,23,2,25,1,80/3,1,27,2,86/3,1,0,8,127/6,6,43/2,3,139/6,4,47/2,2,
            151/6,1,0,24,53/3,1,58/3,1,59/3,5,20,3,65/3,4,0,38,97/6,1,71/4,1,
            107/6,1,0,41,191/12,2,49/3,1,50/3,1,91/5,2,73/4,1,55/3,3,0,51,79/6,
            1,419/30,1,169/12,2,173/12,2,89/6,1,0,58,193/15,3,0,61,23/2,1,0];
        I:=[672262,672304,672416,672003,672001,672224,672000,671723,671751,
            671752,671805,672263,672346,672329,672418,672432,671221,671569,
            672182,672201,671621,672219,672177,671529,672166,671520,672002,
            672077,672084,672088,671524,672081,672264,671123,671126,671225,
            671239,671987,671583,671279,671157,671486,672223,671521,671116,
            671134,671106,671119,671132,671535,671551,671502,671340,671733,
            671542,671567,672350,671541,671632,671636,671640];
    elif hash1 eq 974 and hash2 eq 1597 and hash3 eq 1724 then
        V:=[23,1,24,5,25,9,26,15,27,23,28,27];
        D:=[1,115/6,1,0,2,20,3,0,5,177/10,2,121/6,2,0,9,44/3,2,17,1,91/5,1,0,13,
            147/10,1,0,14,13,1,103/7,1,46/3,1,0];
        I:=[417870,417604,418210,418581,416370,416746,416698,418250,416807,
            418380,416356,417683,415874,415788,415501,415338];
    elif hash1 eq 1538 and hash2 eq 3250 and hash3 eq 873 then
        V:=[19,1,20,7,21,13,23,19,24,25,25,29,26,33,27,39,28,45];
        D:=[1,211/6,1,155/4,1,0,3,196/5,1,157/4,1,0,5,377/12,1,159/4,1,0,7,
            197/12,1,217/12,1,0,9,247/12,1,0,10,877/60,1,0,11,310/21,1,226/15,1,
            0,13,913/60,1,1133/60,1,0,15,117/10,1,0];
        I:=[516737,515911,515817,515896,516195,515730,515970,516033,515981,
            516256,515663,516240,515631,515624,516303];
    elif hash1 eq 2807 and hash2 eq 2911 and hash3 eq 2293 then
        V:=[23,1,24,5,25,17,26,27,27,31,28,37];
        D:=[1,41/2,1,0,2,56/3,1,19,1,62/3,1,21,1,68/3,1,0,7,103/6,2,35/2,2,
            115/6,3,39/2,1,0,15,53/3,3,0,18,157/10,1,95/6,1,0,20,71/5,1,43/3,2,
            0];
        I:=[672016,672030,672013,671819,671939,671904,671404,671965,671308,
            671888,671889,671896,672014,671811,671470,671474,672053,671369,
            671959,671347,671405,671891];
    elif hash1 eq 2102 and hash2 eq 983 and hash3 eq 117 then
        V:=[23,1,24,5,25,13,26,19,27,31,28,35];
        D:=[1,45/2,2,0,3,56/3,1,19,4,68/3,1,0,9,103/6,3,115/6,1,0,13,61/4,1,
            46/3,1,16,3,86/5,1,77/4,1,0,20,95/6,1,0,21,43/3,1,0];
        I:=[244785,244982,244983,244751,244975,245098,245191,245237,242883,
            244986,245068,245227,245289,245027,243331,243806,244870,243808,
            242570,242390,242750];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 35 mod 47.
function cached_ids_35(P,hash1,hash2,hash3)
    if hash1 eq 1586 and hash2 eq 181 and hash3 eq 2194 then
        V:=[25,1,26,9,27,17,28,31,29,35,30,43];
        D:=[1,143/6,1,157/6,1,169/6,1,0,4,67/3,1,24,1,121/5,1,0,7,101/6,1,
            221/12,2,37/2,2,113/6,2,41/2,1,671/30,1,0,16,17,1,0,17,181/12,1,
            461/30,1,17,1,0,20,53/4,1,14,2,0];
        I:=[414638,414674,414677,414781,414782,414255,414800,414440,414728,
            414643,414763,414639,414789,414736,414439,413911,414635,414287,
            413662,414329,414811,414819];
    elif hash1 eq 834 and hash2 eq 878 and hash3 eq 2633 then
        V:=[16,1,17,11,18,21,19,39,20,59,21,73,22,101,23,117,24,135,25,153,26,
            161];
        D:=[1,31,1,104/3,1,35,1,41,1,0,5,187/6,2,199/6,3,67/2,1,75/2,1,0,12,
            77/3,1,82/3,1,83/3,3,89/3,1,125/4,1,133/4,1,100/3,1,101/3,1,0,22,
            133/6,1,305/12,1,155/6,2,277/10,1,167/6,3,169/6,1,353/12,1,297/10,1,
            179/6,2,0,35,358/15,1,73/3,1,388/15,2,335/12,1,86/3,1,92/3,1,0,42,
            187/10,1,113/6,1,207/10,1,125/6,3,265/12,1,671/30,2,269/12,1,45/2,1,
            227/10,1,247/10,1,149/6,1,779/30,1,26,1,0,58,131/7,1,283/15,1,20,1,
            401/20,1,313/15,1,106/5,1,64/3,1,0,65,181/12,1,91/6,1,185/12,1,31/2,
            1,521/30,2,35/2,1,565/28,1,45/2,1,0,74,76/5,1,61/4,3,46/3,1,475/28,
            1,1027/60,1,365/21,1,349/20,1,304/15,1,0,84,817/60,1,653/42,1,95/6,
            1,0,87,61/5,3,0];
        I:=[429390,429383,429387,429382,429371,429399,429356,429358,429377,
            429391,429392,429218,429367,429351,429386,429393,429379,429415,
            429355,429369,429364,429232,429360,429362,429381,429276,429182,
            429278,429368,429373,429398,429202,429185,429234,429242,429388,
            429240,429256,429184,429254,429089,429258,429216,429179,429272,
            429353,429372,429183,429192,429201,429178,429233,429259,429243,
            429204,429156,429147,429140,429095,429144,429148,429195,429203,
            429222,429238,429235,429403,429359,429180,429229,429271,429098,
            429119,429093,429363,429365,429366,429361,429110,429094,429112,
            429096,429090,429102,429108,429116,429241,429261,429273];
    elif hash1 eq 1727 and hash2 eq 2420 and hash3 eq 1241 then
        V:=[23,1,24,5,25,15,26,19,27,33,28,41,29,47,30,55];
        D:=[1,59/2,1,0,2,58/3,1,65/3,1,70/3,1,82/3,1,0,6,253/10,1,0,7,73/4,1,
            239/12,1,20,1,62/3,2,328/15,1,263/12,1,0,14,491/30,1,255/14,1,
            551/30,2,0,18,179/12,1,46/3,1,0,20,83/6,1,625/42,1,181/12,1,0,23,
            89/6,1,0];
        I:=[411041,410714,410281,411210,410303,409562,410806,410992,409765,
            410344,410686,409627,409619,410841,409059,409558,410869,410491,
            409871,410838,409089,409326,409265];
    elif hash1 eq 975 and hash2 eq 3151 and hash3 eq 1618 then
        V:=[21,1,22,11,23,21,24,27,25,39,26,45,27,55];
        D:=[1,175/6,1,59/2,2,187/6,1,63/2,1,0,6,71/3,2,24,1,77/3,2,89/3,1,0,12,
            237/10,1,145/6,1,0,14,20,1,61/3,1,62/3,1,22,1,111/5,1,0,19,37/2,2,
            113/6,1,0,22,329/20,1,253/15,1,278/15,1,223/12,2,0,27,59/4,1,181/12,
            1,461/30,1,31/2,1,0];
        I:=[253391,253216,253218,252867,253256,252955,253385,253215,253135,
            253402,252816,252705,252840,253131,252860,252686,252848,252507,
            252864,253136,253137,252548,252703,252530,252529,252563,252578,
            252821,252718,252746];
    elif hash1 eq 505 and hash2 eq 1134 and hash3 eq 400 then
        V:=[21,1,22,13,23,23,24,31,25,47,26,53,27,65,28,79,29,87];
        D:=[1,55/2,1,175/6,1,59/2,1,187/6,1,67/2,1,0,6,65/3,1,76/3,1,26,1,136/5,
            1,0,10,131/6,2,237/10,1,103/4,1,0,14,91/5,1,298/15,1,61/3,3,328/15,
            1,111/5,1,67/3,1,358/15,1,0,23,611/30,1,245/12,1,0,25,179/12,2,
            248/15,1,361/20,1,369/20,1,278/15,1,0,31,27/2,2,299/20,1,15,1,91/6,
            1,461/30,3,50/3,2,0,41,66/5,2,317/21,1,76/5,1,0,45,35/3,1,0];
        I:=[423894,424578,423949,422879,423897,423246,424584,423904,422880,
            424744,424753,423079,422605,423162,422881,424001,424068,424743,
            423142,422902,422874,423178,422960,423478,423194,423683,423159,
            422511,423631,423274,423241,423510,422640,422575,422644,422884,
            423192,423630,422417,422532,423163,423519,422483,422491,423451];
    elif hash1 eq 2526 and hash2 eq 2995 and hash3 eq 3171 then
        V:=[19,1,20,9,21,23,22,33,23,41,24,55,25,61,26,65];
        D:=[1,57/2,2,61/2,1,65/2,1,0,5,25,1,80/3,1,27,1,86/3,1,98/3,1,33,1,0,11,
            139/6,1,151/6,2,51/2,1,163/6,2,0,17,58/3,1,65/3,1,71/3,2,0,21,
            209/12,1,107/6,2,131/6,1,133/6,2,45/2,1,305/12,1,0,29,61/3,1,62/3,2,
            0,32,59/4,1,0,33,289/20,1,218/15,1,175/12,3,44/3,1,0];
        I:=[424117,424121,424004,424175,424027,424096,424008,423969,424031,
            423832,424022,424216,424245,424002,424009,424230,424169,424229,
            423932,424318,424086,423848,424156,423337,423402,424249,423372,
            422685,423083,423100,423118,424335,423399,423333,424073,424078,
            424092,423855];
    elif hash1 eq 1539 and hash2 eq 2013 and hash3 eq 2654 then
        V:=[24,1,25,7,26,13,27,23,29,31];
        D:=[1,23,1,25,1,0,3,139/6,1,151/6,1,0,5,59/3,1,64/3,1,65/3,1,22,2,0,10,
            109/6,3,37/2,1,119/6,2,0,16,173/12,2,29/2,1,89/6,5,167/10,1,101/6,1,
            0];
        I:=[670626,670568,670429,670582,670764,670374,670484,670474,670559,
            670411,670766,670874,670385,670655,670715,670149,670352,670143,
            670296,670558,670574,670575,670578,669926,669925];
    elif hash1 eq 2244 and hash2 eq 3398 and hash3 eq 152 then
        V:=[20,1,21,11,22,17,23,35,24,47,25,63,26,75,27,83,28,89];
        D:=[1,21,1,25,1,29,1,98/3,1,0,5,127/6,1,91/4,1,0,7,77/4,1,58/3,1,85/4,1,
            65/3,2,71/3,1,126/5,1,77/3,1,109/4,1,0,16,37/2,1,119/6,1,121/6,1,
            217/10,1,131/6,2,0,22,65/4,1,50/3,2,17,1,262/15,1,215/12,1,55/3,2,
            61/3,2,0,32,33/2,1,351/20,1,551/30,2,37/2,2,187/10,1,0,39,223/15,2,
            179/12,1,199/12,1,0,43,27/2,1,625/42,1,0,45,178/15,1,0];
        I:=[424810,424600,424555,424750,424508,424119,424351,424451,424141,
            424101,424815,424482,423141,424484,423575,424176,424486,423860,
            423563,422634,422636,423475,423527,424469,424167,423312,423616,
            422688,423999,423720,423733,422560,423454,422554,423682,422564,
            422686,423750,423289,423550,423523,422499,423517,422500,423448];
    elif hash1 eq 2056 and hash2 eq 1136 and hash3 eq 3136 then
        V:=[20,1,21,5,22,9,23,13,24,21,25,33,26,39,27,43,29,49,30,57];
        D:=[1,43,1,0,2,75/2,1,0,3,101/3,2,0,5,179/6,1,313/10,1,203/6,1,0,8,
            111/5,1,358/15,1,287/12,1,382/15,1,311/12,1,0,13,399/20,1,511/20,1,
            0,15,386/21,1,0,16,371/20,1,569/30,1,0,18,817/60,1,137/10,1,185/12,
            1,0,21,417/35,1,0];
        I:=[419898,419889,419875,419885,419887,419179,419891,418988,419635,
            419862,419180,419538,419676,419061,419011,419140,419177,419851,
            419836,419024,418967];
    elif hash1 eq 1351 and hash2 eq 832 and hash3 eq 1420 then
        V:=[21,1,22,7,23,11,24,17,25,27,26,45,27,59];
        D:=[1,305/12,1,103/4,1,0,3,22,1,0,4,719/30,2,197/6,1,0,7,247/12,1,23,1,
            373/15,1,33,1,0,11,865/42,2,83/4,1,419/20,1,257/12,1,265/12,1,133/6,
            1,199/8,1,751/30,1,0,20,256/15,1,94/5,1,115/6,1,389/20,1,443/21,1,
            1267/60,1,0,26,1411/70,1,1703/84,1,0];
        I:=[541947,541940,541944,541954,541960,541989,541948,541996,542022,
            541813,541938,541959,541973,541990,541994,542029,542030,541833,
            541814,541845,541987,542001,542027,541923,542020,542025,542018];
    elif hash1 eq 1398 and hash2 eq 2702 and hash3 eq 1002 then
        V:=[28,1,29,7,30,11,31,23,32,35];
        D:=[1,238/15,1,81/5,1,0,3,169/12,1,0,4,155/12,1,289/20,2,218/15,2,
            175/12,1,15,2,0,12,129/10,1,391/30,1,59/4,1,451/30,2,181/12,1,0,18,
            67/6,1,0];
        I:=[243642,243454,243442,242927,243498,244375,243436,243529,243444,
            243356,245602,243227,243431,241797,242301,242302,242393,244339];
    elif hash1 eq 3090 and hash2 eq 713 and hash3 eq 193 then
        V:=[21,1,22,7,23,21,24,29,25,43,26,55,27,61];
        D:=[1,49/2,3,57/2,1,0,5,62/3,2,21,2,68/3,2,74/3,1,25,1,27,1,0,14,115/6,
            1,39/2,1,43/2,1,0,17,86/5,1,69/4,1,53/3,7,96/5,1,59/3,3,20,1,0,31,
            97/6,1,177/10,1,107/6,1,109/6,1,233/12,2,0,37,238/15,1,16,1,0,39,
            431/30,1,173/12,1,0];
        I:=[251083,251101,251168,250959,251087,251179,250981,251028,250931,
            250991,251158,250939,250515,250934,250980,251254,250429,251077,
            250550,250932,250936,250937,250958,251080,251216,250551,250628,
            251010,251374,250567,250511,250510,250683,251078,249882,250421,
            250534,251256,250425,251163];
    elif hash1 eq 2573 and hash2 eq 2937 and hash3 eq 1257 then
        V:=[16,1,17,5,18,11,19,17,20,31,21,49,22,67,23,75,24,79,25,97,26,117,27,
            127];
        D:=[1,122/3,1,0,2,235/6,1,283/6,1,0,4,101/3,1,196/5,1,0,6,297/10,1,
            337/10,2,203/6,1,205/6,3,227/6,1,159/4,1,0,15,311/12,1,131/5,1,28,1,
            141/5,1,85/3,1,448/15,2,383/12,1,32,1,0,24,731/30,1,293/12,1,313/12,
            1,791/30,1,317/12,1,267/10,1,161/6,1,959/30,1,0,32,313/15,1,521/20,
            1,53/2,1,0,35,229/12,1,0,36,263/15,1,1123/60,2,1147/60,2,287/15,1,
            115/6,1,389/20,1,293/15,2,235/12,2,0,48,893/60,1,913/60,1,61/4,1,
            46/3,2,467/30,1,157/10,2,95/6,1,563/30,1,1133/60,1,0,59,1606/105,1,
            1289/84,1,461/30,1,94/5,1,0,63,709/60,1,178/15,1,0];
        I:=[428482,428253,427917,428101,427968,427978,427961,428072,428454,
            427945,428470,428472,427941,427784,428004,428095,428393,428089,
            427982,427737,427954,427962,428363,427947,427960,427973,428067,
            427757,428050,428112,427689,428096,427701,427665,427970,428083,
            427664,427667,428026,428033,427676,427771,427957,427708,427950,
            427787,428087,427698,427924,428048,428144,428244,428091,428007,
            428042,428518,427618,427617,427677,427663,427662,427608,428129,
            427995];
    elif hash1 eq 270 and hash2 eq 1306 and hash3 eq 2636 then
        V:=[16,1,17,11,18,21,19,39,20,51,21,61,22,69,23,73];
        D:=[1,28,1,30,2,32,2,36,1,0,7,49/2,3,53/2,4,57/2,3,61/2,6,0,23,68/3,1,
            23,2,74/3,1,25,5,80/3,1,27,7,86/3,1,31,1,0,42,127/6,1,43/2,1,139/6,
            5,47/2,1,51/2,4,0,54,59/3,2,64/3,1,65/3,4,22,2,0,63,109/6,1,133/6,1,
            45/2,4,0,69,55/3,1,0,70,101/6,2,0];
        I:=[254624,254615,254633,254619,254640,254626,254550,254616,254631,
            254574,254623,254625,254636,254506,254614,254627,254528,254554,
            254582,254584,254613,254621,254514,254527,254634,254519,254504,
            254539,254583,254617,254638,254587,254498,254505,254511,254524,
            254580,254585,254639,254586,254481,254552,254467,254510,254516,
            254518,254553,254559,254642,254457,254465,254470,254508,254525,
            254529,254444,254446,254449,254512,254565,254452,254523,254473,
            254406,254389,254391,254445,254456,254480,254399,254522];
    elif hash1 eq 2244 and hash2 eq 3274 and hash3 eq 1403 then
        V:=[21,1,22,5,23,11,24,21,25,29,26,41,27,51,28,63,29,69];
        D:=[1,139/6,1,0,2,117/4,1,89/3,2,0,5,281/12,1,257/10,1,317/10,1,193/6,1,
            0,9,215/12,1,55/3,1,56/3,2,0,13,479/30,1,16,1,551/30,1,207/10,1,
            125/6,1,0,18,248/15,1,203/12,1,278/15,1,131/7,1,0,22,625/42,1,
            451/30,1,973/60,1,331/20,1,997/60,1,0,27,79/6,1,163/10,1,0,29,
            697/60,1,55/4,1,0];
        I:=[510348,510168,510332,510402,509478,509906,508581,508848,510667,
            510486,510315,510396,509283,509292,510048,508819,508859,509967,
            509785,510008,508538,508457,510188,508445,509303,508678,508637,
            508305,509960,508420];
    elif hash1 eq 1022 and hash2 eq 3449 and hash3 eq 3407 then
        V:=[22,1,23,7,24,17,25,23,26,29,27,35,28,41,29,47];
        D:=[1,74/3,1,29,1,0,3,35/2,2,43/2,2,139/6,1,47/2,1,0,9,58/3,1,59/3,1,0,
            11,97/6,1,119/6,1,0,13,191/12,1,56/3,1,0,15,77/6,1,89/6,1,0,17,
            188/15,3,223/15,1,0,21,341/30,1,0];
        I:=[502497,502447,502230,502428,502499,502507,502619,502429,502334,
            502419,502347,502077,502440,502532,502338,502372,502177,502333,
            502342,501611,502364];
    elif hash1 eq 928 and hash2 eq 3210 and hash3 eq 2301 then
        V:=[18,1,19,7,20,13,21,19,22,37,23,49,24,59,25,63,26,69,27,85];
        D:=[1,39,2,134/3,1,0,4,175/6,1,199/6,1,0,6,82/3,1,107/3,1,0,8,103/4,1,
            155/6,1,157/6,1,167/6,1,179/6,1,61/2,1,191/6,1,357/10,1,0,16,68/3,1,
            388/15,1,311/12,1,418/15,1,113/4,1,0,21,245/12,1,439/20,1,139/6,1,
            731/30,1,0,25,247/12,2,0,27,521/30,1,1117/60,1,0,29,907/60,1,46/3,1,
            233/15,1,187/12,1,47/3,1,86/5,1,52/3,1,0,36,157/10,1,63/4,2,0];
        I:=[515512,515555,515520,515519,515477,515515,515432,514560,515468,
            515384,515528,515382,515452,514349,514406,515385,514742,514799,
            514343,514454,514530,514798,515457,514348,514347,514527,514238,
            514411,514758,515493,514408,514430,514232,514237,514228,514252,
            514235,514236];
    elif hash1 eq 2902 and hash2 eq 2897 and hash3 eq 76 then
        V:=[20,1,21,9,22,19,23,29,24,35,25,41,26,45,28,49];
        D:=[1,33,1,104/3,1,116/3,1,0,4,175/6,1,199/6,1,71/2,1,259/6,1,0,8,26,1,
            83/3,1,106/3,1,107/3,1,0,12,155/6,1,205/6,1,0,14,151/5,1,92/3,1,0,
            16,113/6,1,0,17,223/12,1,0,18,263/15,1,18,1,0];
        I:=[510953,511167,510956,511122,511103,510950,511312,510959,511171,
            511114,511177,511100,510991,511104,511181,510981,511314,511307,
            511105];
    elif hash1 eq 975 and hash2 eq 2046 and hash3 eq 1013 then
        V:=[23,1,24,5,25,9,26,15,27,21,28,29,29,41,30,47];
        D:=[1,123/4,1,0,2,59/3,1,0,3,197/10,1,111/4,1,0,5,215/12,1,24,1,0,7,
            491/30,1,125/6,1,99/4,1,0,10,321/20,1,344/21,1,17,1,69/4,1,21,1,0,
            15,193/12,1,973/60,1,0,17,163/12,1,0];
        I:=[502166,502462,502199,501150,502592,502390,502505,501133,501132,
            502265,501679,502329,502301,501194,502093,502090,501443];
    elif hash1 eq 2949 and hash2 eq 2412 and hash3 eq 540 then
        V:=[20,1,21,7,22,17,23,23,24,31,26,41];
        D:=[1,27,1,29,1,0,3,139/6,1,47/2,1,151/6,1,51/2,2,0,8,65/3,4,71/3,2,0,
            14,119/6,1,121/6,1,131/6,1,0,17,73/4,1,55/3,5,56/3,2,19,1,0,26,15,1,
            0];
        I:=[674327,674334,674319,674337,674322,674331,674365,674318,674332,
            674340,674341,674289,674313,674323,674290,674287,674286,674280,
            674297,674328,674329,674356,674335,674336,674342,674283];
    elif hash1 eq 505 and hash2 eq 3189 and hash3 eq 2361 then
        V:=[23,1,25,7,26,15,27,21,28,27,29,35,30,39];
        D:=[1,175/6,1,199/6,1,0,3,155/6,2,157/6,1,169/6,1,0,7,215/12,1,73/3,1,0,
            9,125/6,1,127/6,1,0,11,37/2,1,223/12,1,19,1,0,14,371/20,1,0,15,
            787/60,1,53/4,1,0];
        I:=[416641,416851,417133,417334,416487,417004,416637,416532,416548,
            417009,415523,415828,416625,414995,416512,416661];
    elif hash1 eq 1398 and hash2 eq 669 and hash3 eq 1920 then
        V:=[24,1,25,9,26,17,27,25,28,33,29,43,31,47];
        D:=[1,62/3,2,21,1,80/3,1,0,5,139/6,1,151/6,1,115/4,1,0,8,53/3,1,59/3,1,
            106/5,1,0,11,157/10,1,63/4,1,95/6,1,0,14,37/3,1,167/12,1,191/12,1,
            49/3,1,0,18,431/30,1,0,19,331/30,2,0];
        I:=[494171,494180,494207,494026,494165,494226,494241,494016,493933,
            494258,493921,494172,493931,494228,494037,493923,494020,493932,
            494022,494029];
    elif hash1 eq 3043 and hash2 eq 3143 and hash3 eq 799 then
        V:=[16,1,17,5,18,9,19,19,20,29,21,45,22,59,23,65,24,75,25,91,26,99,27,
            109];
        D:=[1,51,1,0,2,211/6,1,0,3,113/3,1,196/5,1,118/3,1,165/4,1,0,7,377/12,1,
            205/6,1,449/12,1,377/10,1,0,11,131/5,3,27,1,335/12,1,151/5,1,161/5,
            1,97/3,1,508/15,1,0,20,313/12,1,317/12,1,163/6,1,341/12,1,451/14,1,
            971/30,1,0,26,313/15,1,529/20,1,0,28,877/42,1,419/20,1,127/6,2,
            289/12,1,0,33,56/3,1,523/28,1,394/21,1,531/28,1,381/20,1,293/15,2,
            332/15,1,0,41,1271/84,1,683/42,1,163/10,1,0,44,827/60,1,31/2,1,
            736/45,1,394/21,1,0,48,709/60,2,527/42,1,377/30,1,1649/120,2,331/24,
            2,829/60,1,197/12,1,0];
        I:=[518072,518002,517948,517858,517874,517886,518071,518008,518038,
            517911,517854,517921,517938,517635,518076,517834,517949,517952,
            517826,517832,518020,517634,517939,517729,517708,517951,517761,
            517730,517717,517608,517620,517659,517752,517695,517694,517714,
            517793,517841,517869,517645,517664,517652,517647,517970,517724,
            517618,517617,517950,517954,517700,517691,517633,517663,517629,
            517654,517649,518177];
    elif hash1 eq 1351 and hash2 eq 2135 and hash3 eq 1765 then
        V:=[22,1,23,7,24,15,25,29,26,33,27,43];
        D:=[1,83/3,3,89/3,1,0,5,133/6,1,143/6,1,155/6,1,0,8,61/3,1,62/3,1,21,1,
            22,1,89/4,1,67/3,1,0,14,37/2,2,0,16,86/5,1,52/3,1,283/15,1,227/12,2,
            0,21,59/4,1,31/2,1,0];
        I:=[252238,252315,252326,252325,252117,252345,251989,252346,252321,
            252280,252234,251862,251968,251879,252114,251876,251887,251896,
            251893,251895,252040,252297];
    elif hash1 eq 787 and hash2 eq 1665 and hash3 eq 3052 then
        V:=[26,1,27,7,28,17,29,33,30,41,31,47,32,53];
        D:=[1,71/3,1,76/3,1,0,3,121/6,2,133/6,1,237/10,1,143/6,1,0,8,49/3,1,
            215/12,1,73/4,1,55/3,2,56/3,1,239/12,1,101/5,1,0,16,197/12,1,33/2,1,
            221/12,1,0,19,175/12,1,223/15,1,0,21,157/12,1,83/6,1,0,23,12,1,0];
        I:=[241411,241601,240535,241154,240446,240057,241231,241505,241361,
            241372,240650,240725,241227,238351,239123,241240,241471,238406,
            241123,239824,239973,241597,238140];
    elif hash1 eq 35 and hash2 eq 644 and hash3 eq 983 then
        V:=[25,1,26,5,27,13,28,21,29,35,30,41];
        D:=[1,119/6,2,0,3,49/3,2,17,1,55/3,1,0,7,91/6,1,31/2,1,101/6,3,0,12,
            175/12,1,44/3,1,74/5,1,179/12,1,46/3,2,329/20,1,0,19,157/12,1,
            181/12,2,0,22,37/3,1,0];
        I:=[414731,414752,414784,414879,414697,414899,414764,414698,414690,
            414772,414877,414429,414727,413930,414699,414745,414747,414391,
            413926,413681,413706,414351];
    elif hash1 eq 834 and hash2 eq 2585 and hash3 eq 144 then
        V:=[16,1,17,5,19,9,20,13,21,19,22,23,24,29,25,37,26,45,27,53];
        D:=[1,44,1,0,2,77/2,1,0,3,247/6,1,0,4,88/3,1,100/3,2,0,7,401/12,1,0,8,
            85/3,1,502/15,1,0,10,33/2,1,409/20,1,41/2,1,0,13,379/30,1,51/4,1,
            41/2,1,0,16,38/3,1,763/60,1,247/12,1,0,19,185/12,1,0];
        I:=[543364,543332,543145,543347,543349,543362,543246,543261,542896,
            543377,542969,542975,542987,542990,542707,542988,542961,542626,
            542838];
    elif hash1 eq 411 and hash2 eq 967 and hash3 eq 1863 then
        V:=[26,1,27,9,28,19,30,25,31,29,32,35];
        D:=[1,83/3,1,89/3,2,101/3,2,0,6,143/6,1,257/10,1,155/6,1,179/6,1,0,10,
            121/5,1,73/3,1,0,12,253/15,2,0,14,44/3,1,59/4,1,0,16,41/3,3,59/4,1,
            0];
        I:=[245583,244104,245599,242761,244094,245067,244190,245446,245497,
            242224,242223,244129,244497,242166,244478,242177,245472,245594,
            242132];
    elif hash1 eq 82 and hash2 eq 280 and hash3 eq 3442 then
        V:=[26,1,27,5,28,11,29,15,30,21,31,33];
        D:=[1,21,1,0,2,39/2,1,127/6,2,0,5,53/3,3,0,8,95/6,1,97/6,1,0,10,167/12,
            2,71/5,1,43/3,1,191/12,1,49/3,1,0,16,25/2,1,0];
        I:=[228451,228498,226886,228488,226347,228225,228496,226512,228432,
            226872,228519,227417,227721,225475,224429,228144];
    elif hash1 eq 834 and hash2 eq 2433 and hash3 eq 958 then
        V:=[18,1,19,9,20,17,21,25,23,31];
        D:=[1,24,1,28,1,30,1,0,4,49/2,1,65/2,1,69/2,1,0,7,21,1,86/3,1,29,1,0,10,
            127/6,1,47/2,2,0,13,87/4,1,131/6,1,133/6,1,0];
        I:=[510713,510714,510711,510535,510716,510730,510700,510724,510355,
            510364,510287,510528,509616,509996,509622];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 36 mod 47.
function cached_ids_36(P,hash1,hash2,hash3)
    if hash1 eq 1258 and hash2 eq 2526 and hash3 eq 2211 then
        V:=[24,1,25,5,26,13,27,21,28,25,29,37,30,45];
        D:=[1,64/3,1,0,2,109/6,2,49/2,1,103/4,1,0,6,17,1,73/4,1,328/15,1,0,9,
            101/6,2,0,11,218/15,1,15,1,46/3,1,17,1,53/3,1,0,16,181/12,1,67/4,1,
            169/10,1,0,19,203/15,1,0];
        I:=[501748,501711,501848,501747,501623,501793,501744,501434,501500,
            501766,501716,501845,501784,501498,501767,501472,501686,501738,
            501502];
    elif hash1 eq 2245 and hash2 eq 2495 and hash3 eq 2221 then
        V:=[24,1,25,7,26,17,27,29,28,35];
        D:=[1,62/3,1,74/3,1,0,3,103/6,3,115/6,1,43/2,1,139/6,1,0,9,16,1,53/3,2,
            18,2,59/3,2,106/5,1,0,17,153/10,1,185/12,2,0,20,208/15,1,43/3,1,
            44/3,2,16,1,0];
        I:=[245202,245208,245240,245385,245721,245218,243370,242448,243305,
            244731,245187,245183,245777,243821,244810,242441,243226,244811,
            244902,243386,245305,243338,243368,243823];
    elif hash1 eq 2010 and hash2 eq 134 and hash3 eq 342 then
        V:=[23,1,24,13,25,17,26,21,27,29,28,37];
        D:=[1,127/6,1,151/6,1,163/6,1,59/2,1,187/6,1,0,6,64/3,2,0,8,217/10,1,0,
            9,215/12,1,55/3,1,101/5,1,0,12,101/6,1,221/12,1,113/6,1,0,15,175/12,
            1,17,3,0];
        I:=[502597,502498,502542,502491,502200,502191,502607,502282,502068,
            502247,502601,502485,501521,502245,502579,502458,502533,502558];
    elif hash1 eq 2621 and hash2 eq 465 and hash3 eq 1229 then
        V:=[25,1,26,5,27,17,28,27,29,35,30,49,31,57];
        D:=[1,143/6,1,0,2,62/3,1,263/12,1,67/3,2,24,1,97/4,1,0,8,33/2,1,113/6,2,
            611/30,1,245/12,1,0,13,203/12,2,52/3,2,283/15,1,0,18,439/30,1,59/4,
            1,181/12,1,91/6,3,185/12,1,569/30,1,0,26,202/15,1,163/12,1,233/15,1,
            0,29,239/20,1,145/12,1,0];
        I:=[413918,414206,414209,413902,414234,414151,413868,414069,413715,
            413876,414095,414416,413863,414194,414197,414208,413915,414022,
            414199,414210,413806,413888,414140,413717,413487,414361,414077,
            413441,413490,413582];
    elif hash1 eq 2245 and hash2 eq 1982 and hash3 eq 2855 then
        V:=[11,1,12,5,13,9,15,13,16,17,17,25,18,31,19,39,20,43,22,47,24,53,26,
            61];
        D:=[1,133/2,1,0,2,134/3,1,0,3,319/6,1,0,4,275/6,1,0,5,201/5,1,334/7,1,
            241/5,1,0,8,457/12,1,563/14,1,0,10,162/5,2,488/15,1,518/15,1,0,14,
            1813/60,1,0,15,575/21,1,0,16,1649/84,1,114/5,1,0,18,1496/105,1,72/5,
            1,481/30,1,0,21,776/63,1,223/18,1,347/28,1,0];
        I:=[544191,544161,544162,544159,544158,544114,544125,544189,544112,
            544156,544175,544187,544166,544170,544113,544122,544108,544100,
            544157,544105,544083,544076,544082];
    elif hash1 eq 2715 and hash2 eq 2903 and hash3 eq 3194 then
        V:=[19,1,20,9,21,23,22,37,23,41,24,49,25,63,26,69];
        D:=[1,61/2,3,65/2,2,69/2,1,0,7,27,3,86/3,1,29,4,92/3,1,31,1,33,1,0,18,
            43/2,1,99/4,1,151/6,4,51/2,1,163/6,2,55/2,1,0,28,70/3,1,0,29,119/6,
            2,121/6,1,133/6,1,0,33,215/12,2,55/3,1,56/3,1,61/3,1,67/3,1,68/3,1,
            0,40,33/2,1,101/6,1,0,42,223/15,1,179/12,1,61/4,1,0];
        I:=[253209,253252,253365,253168,253372,253283,252951,253380,253404,
            252905,253169,253175,253226,253378,252973,253368,253411,253415,
            252931,252930,253234,253379,253421,252868,252855,253089,252895,
            253130,253146,253149,252927,252631,253055,253096,252969,253370,
            252628,252547,252553,252590,253099,252581,252588,252674];
    elif hash1 eq 224 and hash2 eq 3088 and hash3 eq 449 then
        V:=[14,1,15,5,17,9,18,13,19,25,21,33,22,39,23,43];
        D:=[1,22,1,0,2,45/2,1,0,3,115/6,1,0,4,64/3,1,24,1,30,2,36,1,42,1,0,10,
            109/6,1,145/6,1,203/6,1,0,13,255/14,1,399/20,2,0,16,386/21,1,0,17,
            119/8,1,62/3,1,0];
        I:=[543787,543788,543792,543777,543731,543678,543730,543682,543677,
            543725,543664,543742,543663,543756,543758,543670,543626,543622];
    elif hash1 eq 553 and hash2 eq 1376 and hash3 eq 3387 then
        V:=[21,1,22,15,23,29,24,39,25,51,26,77,27,83];
        D:=[1,139/6,1,151/6,1,51/2,1,163/6,2,55/2,3,59/2,1,0,10,59/3,2,22,1,24,
            3,126/5,1,101/4,1,77/3,1,0,19,109/6,2,37/2,1,119/6,1,121/6,4,0,27,
            50/3,2,91/5,1,55/3,2,61/3,1,62/3,1,0,34,91/6,2,163/10,1,197/12,3,
            33/2,4,101/6,4,103/6,1,551/30,3,221/12,1,37/2,2,187/10,3,75/4,1,
            659/30,1,0,60,223/15,1,50/3,1,0,62,161/12,1,27/2,1,0];
        I:=[422186,422235,422310,422164,422272,421925,421956,422290,421966,
            422172,422189,421805,421980,422080,422081,421660,421798,421809,
            421829,422319,421975,421822,421717,422043,422074,422170,421803,
            421819,421322,421816,422171,422041,422293,421312,421736,421696,
            421802,421960,422157,421821,421995,422005,422275,421917,422101,
            422142,422248,421943,421302,421305,421335,421290,421249,421481,
            421299,421475,421571,421257,421119,421747,421223,421329,421293];
    elif hash1 eq 1822 and hash2 eq 1055 and hash3 eq 769 then
        V:=[18,1,19,5,20,9,21,15,22,23,23,27,24,35];
        D:=[1,29,1,0,2,55/2,1,0,3,77/3,3,26,1,0,7,131/6,1,133/6,1,143/6,1,0,10,
            263/12,2,0,12,75/4,1,113/6,4,115/6,3,0,20,283/15,2,58/3,2,0];
        I:=[254086,254089,254060,254072,254085,254083,254061,254095,254065,
            254042,254043,254049,254037,254063,254094,254098,254070,254078,
            254080,254028,254031,254033,254034];
    elif hash1 eq 2668 and hash2 eq 67 and hash3 eq 1890 then
        V:=[24,1,25,5,26,9,27,15,28,33,29,47,30,55];
        D:=[1,53/3,1,0,2,107/6,1,0,3,49/3,1,50/3,2,0,6,173/12,1,59/4,1,89/6,3,
            91/6,1,491/30,1,33/2,1,359/20,2,539/30,1,0,17,44/3,1,223/15,3,15,1,
            46/3,2,248/15,1,199/12,2,0,27,157/12,2,451/30,2,91/6,1,0,32,23/2,1,
            0];
        I:=[418821,418863,418301,417936,418371,416463,416417,418309,418334,
            418605,418660,416386,417325,416576,416743,415496,418830,416522,
            416741,417315,416029,416626,418646,417437,415420,417402,416791,
            417504,415356,416186,415992,417708];
    elif hash1 eq 2433 and hash2 eq 1893 and hash3 eq 167 then
        V:=[17,1,18,13,19,27,20,35,21,47];
        D:=[1,53/2,1,57/2,4,61/2,1,65/2,2,69/2,1,0,10,23,1,74/3,1,25,3,80/3,1,
            27,3,29,1,0,20,139/6,3,47/2,3,163/6,1,0,27,59/3,1,64/3,1,65/3,6,22,
            1,70/3,1,0,37,119/6,2,217/10,1,131/6,1,133/6,2,0];
        I:=[254618,254567,254620,254628,254635,254538,254545,254578,254536,
            254568,254551,254542,254622,254641,254592,254534,254556,254569,
            254557,254461,254546,254548,254532,254561,254637,254451,254535,
            254544,254450,254466,254531,254541,254593,254594,254533,254474,
            254454,254540,254392,254475,254405,254453];
    elif hash1 eq 2480 and hash2 eq 2723 and hash3 eq 2675 then
        V:=[22,1,23,7,24,11,25,17,26,23,29,29];
        D:=[1,25,1,31,1,0,3,55/2,2,0,5,64/3,1,71/3,1,0,7,119/6,1,121/6,1,0,9,
            61/3,3,62/3,1,0,13,757/60,1,51/4,1,391/30,1,0];
        I:=[415865,415731,415872,415876,415909,415906,415889,415898,415802,
            415867,415877,415677,415787,415805,415727];
    elif hash1 eq 1352 and hash2 eq 2056 and hash3 eq 2862 then
        V:=[14,1,15,5,16,13,17,21,18,29,19,41,20,51,21,55,22,61,24,69];
        D:=[1,40,1,0,2,73/2,1,77/2,1,81/2,1,0,5,110/3,1,37,1,39,1,0,8,131/4,1,
            67/2,1,211/6,1,0,11,77/3,1,89/3,2,30,2,95/3,1,107/3,1,0,18,305/12,1,
            155/6,1,167/6,1,179/6,1,0,22,263/12,1,0,23,137/6,1,139/6,1,0,25,
            223/12,1,271/12,1,275/12,1,0,28,907/60,1,309/20,1,233/15,1,0];
        I:=[255667,255666,255670,255680,255661,255659,255664,255517,255543,
            255528,255544,255513,255514,255535,255683,255686,255532,255539,
            255546,255524,255429,255534,255428,255523,255518,255363,255362,
            255427,255420,255426];
    elif hash1 eq 1305 and hash2 eq 2452 and hash3 eq 67 then
        V:=[23,1,24,5,25,19,26,29,27,43,28,49,29,55];
        D:=[1,187/6,1,0,2,22,1,71/3,1,24,1,76/3,1,88/3,1,89/3,1,0,8,133/6,1,
            145/6,1,169/6,1,179/6,1,0,12,56/3,1,101/5,1,62/3,1,22,1,67/3,2,68/3,
            1,0,19,551/30,1,83/4,2,0,22,15,1,86/5,2,0,25,521/30,1,35/2,1,0];
        I:=[415149,415302,415089,415196,415169,415116,415163,415191,415082,
            415281,415334,415301,415087,415152,415118,415115,415161,415306,
            415187,415098,415099,415263,415110,415112,415305,415227];
    elif hash1 eq 788 and hash2 eq 3308 and hash3 eq 1496 then
        V:=[23,1,24,7,25,13,26,25,27,39,28,51,29,57];
        D:=[1,127/6,1,63/2,1,0,3,59/3,1,77/3,1,0,5,107/6,1,109/6,1,79/4,1,
            217/10,1,87/4,1,0,10,268/15,1,55/3,1,56/3,1,298/15,1,239/12,1,61/3,
            1,0,16,89/6,1,33/2,1,167/10,1,359/20,1,217/12,1,0,21,223/15,1,
            248/15,1,0,23,451/30,1,91/6,1,0];
        I:=[418889,418330,418896,416345,418888,417922,417339,416344,416726,
            416425,417921,417905,417761,416771,417825,418751,417328,416297,
            417782,416525,416735,416284,415976,416173];
    elif hash1 eq 2903 and hash2 eq 1560 and hash3 eq 1417 then
        V:=[10,1,12,5,13,13,15,17,18,25,19,33,21,39,24,43];
        D:=[1,46,1,0,2,134/3,1,45,1,158/3,1,0,5,223/6,1,0,6,377/10,1,397/10,1,
            521/12,1,0,9,145/6,1,169/6,1,193/6,1,0,12,1453/60,1,97/4,1,0,14,
            1583/84,1,0,15,739/60,1,0];
        I:=[520027,520033,520031,520009,520038,519975,519981,519977,520000,
            519963,519934,519958,519974,519943,519967];
    elif hash1 eq 271 and hash2 eq 3352 and hash3 eq 1517 then
        V:=[21,1,22,7,23,13,24,21,25,25];
        D:=[1,45/2,2,49/2,1,0,4,62/3,1,21,3,0,8,35/2,1,115/6,2,39/2,3,0,14,53/3,
            1,0,15,97/6,1,0];
        I:=[673659,673744,673726,673699,673428,673674,673721,673642,673455,
            673722,673654,673688,673745,673373,673656];
    elif hash1 eq 2151 and hash2 eq 1644 and hash3 eq 1994 then
        V:=[16,1,17,5,18,11,19,17,20,21,21,29,22,35,23,39,24,45];
        D:=[1,30,1,0,2,77/2,1,85/2,1,0,4,27,1,104/3,1,0,6,211/6,1,0,7,59/3,1,
            109/4,1,101/3,1,0,10,119/6,1,329/12,1,0,12,73/3,1,0,13,599/30,1,
            671/30,1,0,15,401/20,1,247/12,1,0];
        I:=[516714,516720,516721,516729,516718,516498,516702,516540,516224,
            516481,516177,516623,516085,515825,515800,515827];
    elif hash1 eq 553 and hash2 eq 503 and hash3 eq 2152 then
        V:=[12,1,13,9,14,19,15,27,16,35,17,39,18,43,19,47];
        D:=[1,36,2,38,2,42,1,0,6,69/2,1,73/2,1,77/2,1,85/2,1,0,10,27,1,29,1,
            104/3,1,0,13,151/6,1,163/6,2,211/6,2,0,18,83/3,1,0,19,329/12,1,0,20,
            68/3,1,0,21,241/12,1,137/6,1,0];
        I:=[430078,430082,430073,430080,430081,430079,430084,430059,430061,
            430065,430076,430043,430047,430009,430062,430025,430063,430019,
            430018,430021,430024,429982];
    elif hash1 eq 1963 and hash2 eq 2515 and hash3 eq 2336 then
        V:=[15,1,16,5,17,9,18,19,19,27,20,45,21,53,22,63,23,67,24,83,25,91];
        D:=[1,93/2,1,0,2,29,1,0,3,151/6,1,199/6,1,223/6,1,259/6,1,0,7,126/5,1,
            133/4,1,101/3,1,0,10,131/6,1,133/6,1,277/10,1,167/6,1,169/6,1,181/6,
            1,203/6,1,215/6,1,0,18,311/12,1,28,1,31,1,0,21,551/30,1,611/30,1,
            137/6,1,779/30,1,0,25,307/15,1,0,26,299/20,1,451/30,1,181/12,1,
            1213/60,1,62/3,1,1357/60,1,459/20,1,0,33,361/24,1,61/4,1,81/4,1,0,
            36,937/60,1,157/10,1,0];
        I:=[428179,428201,428196,428178,428174,427882,427859,428184,428210,
            428173,428167,427853,428195,427879,427911,427826,427804,427827,
            427840,427609,427893,427825,427905,427732,427725,427894,427896,
            427889,427637,427629,427655,427650,427611,428181,427613,427659,
            427653];
    elif hash1 eq 2339 and hash2 eq 2448 and hash3 eq 3202 then
        V:=[23,1,24,11,25,19,26,31,27,35,28,39,29,45];
        D:=[1,127/6,2,51/2,1,107/4,1,59/2,2,0,7,59/3,1,85/4,1,65/3,1,0,10,109/6,
            1,79/4,1,121/6,1,133/6,1,305/12,1,0,15,81/5,2,0,17,311/20,1,0,18,
            155/12,2,302/21,1,0,21,541/42,1,0];
        I:=[410703,410804,411846,410149,411848,411850,410307,411899,411933,
            411828,410340,410151,410559,409578,409146,410743,409963,411880,
            412248,409195,409192];
    elif hash1 eq 1775 and hash2 eq 2918 and hash3 eq 2987 then
        V:=[20,1,21,5,22,13,23,19,24,35,25,45,26,53];
        D:=[1,23,1,0,2,115/6,1,127/6,2,55/2,1,0,6,58/3,2,85/4,1,0,9,97/6,2,
            233/12,2,197/10,1,79/4,1,121/6,2,41/2,1,213/10,1,0,19,17,1,268/15,1,
            18,1,292/15,1,0,23,491/30,1,197/12,1,101/6,1,0,26,223/15,1,0];
        I:=[513933,513921,513928,513930,513890,513593,513925,513816,513736,
            513902,513826,513908,513744,513583,513802,513899,513855,513663,
            513585,513622,513589,513628,513715,513640,513592,513452];
    elif hash1 eq 459 and hash2 eq 1738 and hash3 eq 761 then
        V:=[22,1,23,9,24,19,25,27,26,31];
        D:=[1,62/3,1,21,1,23,1,0,4,115/6,4,83/4,1,127/6,1,43/2,1,0,11,53/3,2,
            96/5,1,58/3,1,0,15,97/6,1,0,16,44/3,1,0];
        I:=[249613,249725,248866,248873,249616,249678,249693,249307,247804,
            249419,247917,248726,247807,249240,247956,247789];
    elif hash1 eq 2151 and hash2 eq 3298 and hash3 eq 1159 then
        V:=[19,1,20,9,21,17,22,25,23,31,24,47,25,51,26,57,27,65,28,69];
        D:=[1,57/2,1,61/2,1,65/2,1,0,4,27,1,86/3,1,104/3,1,0,7,55/2,1,175/6,1,
            223/6,1,0,10,133/4,1,34,1,0,12,45/2,1,143/6,1,145/6,1,257/10,1,
            157/6,1,53/2,2,353/12,1,0,20,121/5,1,0,21,41/2,1,431/20,1,0,23,52/3,
            1,108/5,1,441/20,1,0,26,511/30,1,0,27,47/3,1,0];
        I:=[419711,419662,419710,419869,419233,419217,419712,419232,419597,
            419367,419596,419876,419720,419719,419151,419739,419574,419600,
            419648,419161,419733,419136,419126,419021,419019,419146,419046];
    elif hash1 eq 882 and hash2 eq 1091 and hash3 eq 1453 then
        V:=[18,1,20,9,22,19,23,23,24,27,25,35,26,43];
        D:=[1,106/3,1,36,1,216/5,1,0,4,328/15,1,74/3,1,478/15,1,199/5,1,0,8,
            464/21,1,0,9,1237/60,1,0,10,907/60,1,187/12,1,49/3,1,0,13,467/30,2,
            95/6,1,503/30,1,0,17,547/35,1,439/28,1,947/60,1,238/15,1,191/12,1,
            227/12,1,0];
        I:=[518167,518067,517904,517875,518031,517905,517709,517753,517866,
            517969,517777,518117,517867,517971,517783,517743,517731,517763,
            517712,517785,517787,517627];
    elif hash1 eq 976 and hash2 eq 250 and hash3 eq 1509 then
        V:=[21,1,22,5,23,17,24,35,25,41,26,47,27,55];
        D:=[1,65/2,1,0,2,19,1,23,1,74/3,1,80/3,1,86/3,1,0,7,115/6,2,83/4,1,
            127/6,1,43/2,2,139/6,2,47/2,1,51/2,3,163/6,1,0,20,18,2,58/3,2,0,24,
            107/6,2,109/6,2,0,28,232/15,1,16,1,268/15,1,0,31,59/4,1,0];
        I:=[251496,251506,251220,251307,250816,250810,250833,251355,251328,
            251555,251034,251561,250863,251341,251354,250312,250640,250717,
            250036,251205,251208,250131,250786,250233,251212,250634,250872,
            250392,250120,249972,250018];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 37 mod 47.
function cached_ids_37(P,hash1,hash2,hash3)
    if hash1 eq 2669 and hash2 eq 1826 and hash3 eq 1259 then
        V:=[26,1,27,9,28,17,29,27,30,41];
        D:=[1,25,1,129/4,1,98/3,1,0,4,91/4,1,317/12,1,161/6,1,0,7,19,1,77/4,1,
            59/3,1,295/12,1,0,11,559/30,1,115/6,1,581/30,1,39/2,1,197/10,1,
            641/30,1,0,17,50/3,1,1123/60,1,0];
        I:=[503273,502204,503271,503127,503126,502646,502057,503130,503272,
            501261,503114,502767,503253,502647,503381,503379,503097,502035];
    elif hash1 eq 37 and hash2 eq 330 and hash3 eq 1136 then
        V:=[13,1,14,7,16,13,17,23,18,29,19,37,20,43,21,47,22,55,24,63,26,69,27,
            73];
        D:=[1,73/2,1,77/2,1,0,3,86/3,1,134/3,1,0,5,196/5,1,119/3,1,173/4,1,
            137/3,1,0,9,193/6,1,289/6,1,0,11,222/7,1,103/3,1,622/15,1,0,14,53/2,
            1,451/14,1,0,16,275/12,1,0,17,189/10,2,571/30,1,581/30,1,0,21,
            358/21,1,662/35,1,531/28,1,0,24,218/15,1,2026/105,2,0,27,1033/84,1,
            0,28,1531/126,1,443/36,1,1733/140,1,0];
        I:=[543985,543981,543978,543982,543984,543959,543967,543975,543964,
            543948,543937,543966,543961,543950,543919,543987,543968,543977,
            543942,543960,543943,543946,543945,543979,543928,543941,543924,
            543884,543881,543892];
    elif hash1 eq 1024 and hash2 eq 2909 and hash3 eq 1723 then
        V:=[25,1,26,11,27,25,28,31,29,35,30,41];
        D:=[1,163/6,1,55/2,1,175/6,1,59/2,1,0,5,65/3,2,71/3,1,101/4,1,76/3,1,
            77/3,1,26,1,0,12,87/4,2,133/6,2,0,16,91/5,1,0,17,217/12,1,221/12,1,
            0,19,15,1,0];
        I:=[671252,672231,672233,672221,671305,671694,672230,671266,671292,
            671180,672115,671154,671544,671679,671765,671296,671152,671156,
            671269];
    elif hash1 eq 695 and hash2 eq 1077 and hash3 eq 1213 then
        V:=[10,1,11,13,12,23,13,29,14,41,15,55,16,59,17,71,18,79,19,83,20,87,22,
            93];
        D:=[1,38,1,40,1,44,1,46,1,56,1,0,6,69/2,1,77/2,1,81/2,1,85/2,1,0,10,
            104/3,1,41,1,0,12,163/6,1,175/6,1,211/6,2,71/2,1,247/6,1,0,18,77/3,
            1,83/3,1,28,1,88/3,1,100/3,1,44,1,0,24,305/12,1,0,25,335/12,1,85/3,
            1,151/5,1,431/12,1,109/3,1,0,30,45/2,1,137/6,1,851/30,1,0,33,409/20,
            1,0,34,521/30,1,0,35,1363/60,1,93/4,1,0,37,967/60,1,0];
        I:=[520049,520047,520051,520054,520041,520055,520056,520025,520048,
            520037,520029,520034,520005,520028,520032,520002,520011,520030,
            520043,519994,520023,519995,519964,519997,520014,520001,519979,
            519976,519968,519955,520010,519950,519970,519978,519942,519939,
            519940];
    elif hash1 eq 1071 and hash2 eq 62 and hash3 eq 446 then
        V:=[16,1,17,9,18,17,19,29,20,39,21,49,23,57];
        D:=[1,28,1,32,1,36,1,0,4,57/2,1,69/2,1,77/2,1,0,7,80/3,1,27,1,86/3,1,29,
            1,92/3,1,0,12,115/6,1,39/2,1,127/6,1,151/6,1,0,16,53/3,1,58/3,1,
            65/3,1,83/3,1,0,20,119/6,1,45/2,1,157/6,1,0,23,33/2,1,0];
        I:=[515147,515076,515083,515097,515230,515081,515165,515116,515061,
            515186,515240,515306,515154,515139,515019,515166,515318,514499,
            514261,514509,515234,514262,514744];
    elif hash1 eq 1776 and hash2 eq 1359 and hash3 eq 692 then
        V:=[21,1,22,7,23,17,24,21,26,31,27,35,28,39,29,45];
        D:=[1,187/6,1,199/6,1,0,3,65/3,1,109/4,1,146/5,1,94/3,1,0,7,257/10,1,0,
            8,73/4,1,239/12,1,101/5,1,111/5,1,0,12,278/15,1,0,13,91/6,1,0,14,
            66/5,1,53/4,1,0,16,117/10,1,0];
        I:=[424038,423765,423827,423385,423258,423257,422970,424214,424082,
            423054,423343,422412,422754,423245,423304,423341];
    elif hash1 eq 836 and hash2 eq 415 and hash3 eq 135 then
        V:=[24,1,25,9,26,19,27,27,28,35];
        D:=[1,21,1,68/3,1,23,1,0,4,115/6,2,39/2,3,127/6,3,43/2,2,0,14,58/3,1,
            59/3,2,20,1,0,18,29/2,1,95/6,2,109/6,3,0,24,167/12,2,14,1,71/5,1,
            43/3,1,0];
        I:=[240679,240673,241152,240647,241185,240779,241183,241408,239859,
            240682,241301,240445,240639,240094,239337,240069,240564,241402,
            240665,241337,238487,240071,240394,241382,241383,240669,239969,
            241255];
    elif hash1 eq 1259 and hash2 eq 1940 and hash3 eq 1660 then
        V:=[19,1,20,5,21,9,22,19,23,31,26,39,29,47];
        D:=[1,69/2,1,0,2,37,1,0,3,107/4,1,187/6,2,199/6,1,67/2,1,0,8,64/3,1,
            65/3,1,89/3,1,95/3,1,32,1,0,13,95/4,1,143/6,1,103/4,1,0,16,37/2,1,
            227/12,1,58/3,2,0,20,347/30,1,0];
        I:=[420294,419660,419768,419602,419658,420308,419806,419284,419287,
            419715,419639,419609,419780,419268,419740,419025,419026,419625,
            419631,419256];
    elif hash1 eq 836 and hash2 eq 1544 and hash3 eq 1380 then
        V:=[24,1,25,15,26,21,27,31,28,35];
        D:=[1,68/3,1,23,1,27,1,29,1,31,1,98/3,1,0,7,115/6,1,43/2,1,0,9,52/3,1,
            58/3,1,20,1,65/3,2,0,14,109/6,2,0,16,49/3,1,55/3,1,0];
        I:=[398360,399244,399018,399269,399020,399216,398694,398785,398720,
            398451,398993,398418,398784,399793,399800,398617,398851];
    elif hash1 eq 1306 and hash2 eq 2389 and hash3 eq 2939 then
        V:=[19,1,20,11,21,17,22,27,23,37,24,43,25,49,26,59];
        D:=[1,53/2,1,61/2,1,65/2,1,69/2,1,0,5,80/3,1,33,1,0,7,139/6,3,47/2,1,
            151/6,1,163/6,1,0,13,18,1,59/3,1,70/3,1,71/3,2,0,18,97/6,1,95/4,1,0,
            20,191/12,3,21,1,0,24,89/6,1,33/2,1,221/12,1,113/6,1,0,28,223/15,1,
            0];
        I:=[510593,510594,510261,510530,510527,510587,510546,510547,510570,
            510260,510683,510557,510588,510563,509493,510548,510562,510574,
            509352,510534,510549,510550,510560,510532,510568,509342,509656,
            509651];
    elif hash1 eq 883 and hash2 eq 241 and hash3 eq 1709 then
        V:=[12,1,13,5,14,9,15,15,16,21,17,25,18,35,19,45,20,51,21,55,22,67,23,
            75];
        D:=[1,44,1,0,2,93/2,1,0,3,37,1,41,1,0,5,223/6,1,259/6,1,0,7,101/3,3,0,
            10,181/6,1,401/12,1,203/6,1,215/6,1,0,14,311/12,1,28,2,31,2,169/5,1,
            0,20,779/30,1,313/12,1,0,22,70/3,1,0,23,1213/60,1,411/20,1,62/3,1,
            1357/60,1,459/20,1,0,28,81/4,1,203/10,1,62/3,1,0,31,917/60,1,937/60,
            1,157/10,1,0];
        I:=[429925,429924,429926,429831,429825,429746,429728,429734,429813,
            429736,429716,429717,429737,429727,429732,429733,429739,429745,
            429611,429630,429622,429723,429626,429628,429618,429602,429603,
            429600,429599,429604,429625,429620,429627];
    elif hash1 eq 3186 and hash2 eq 432 and hash3 eq 208 then
        V:=[25,1,27,7,29,21];
        D:=[1,35/2,2,39/2,1,0,4,63/4,1,95/6,3,97/6,2,33/2,1,71/4,1,109/6,3,0,15,
            79/6,1,0];
        I:=[241684,241759,241757,241565,241651,241663,241665,240775,240784,
            240790,238382,238380,239122,240374,239373];
    elif hash1 eq 1823 and hash2 eq 1657 and hash3 eq 2758 then
        V:=[23,1,24,7,25,13,26,19,27,27];
        D:=[1,45/2,1,57/2,3,0,5,25,1,80/3,1,0,7,139/6,1,151/6,1,0,9,58/3,1,59/3,
            1,64/3,1,0,12,185/12,1,95/6,1,107/6,3,0];
        I:=[244816,244815,244817,245336,244866,243070,244818,244966,244935,
            243074,244925,244943,244955,243805,244910,245349];
    elif hash1 eq 3280 and hash2 eq 3332 and hash3 eq 2967 then
        V:=[22,1,23,5,24,15,25,25,26,33,27,45,28,53,29,57];
        D:=[1,80/3,2,0,3,115/6,1,127/6,1,139/6,3,47/2,4,0,12,53/3,1,59/3,2,20,2,
            65/3,1,0,18,109/6,1,119/6,2,121/6,2,0,23,15,1,81/5,1,49/3,7,50/3,2,
            55/3,1,0,35,173/12,1,29/2,1,89/6,2,0,39,76/5,1,0,40,71/6,1,0];
        I:=[246970,246982,247110,247153,246941,247107,247145,247082,247087,
            247088,247114,247036,247120,247150,247089,247156,246562,247005,
            246598,246939,246605,246933,247112,246828,246591,246613,246826,
            246944,247095,247097,247177,246706,247157,246603,247096,247134,
            246992,247137,246471,246626];
    elif hash1 eq 1024 and hash2 eq 811 and hash3 eq 220 then
        V:=[14,1,15,5,16,11,17,21,18,31,19,39,20,51,21,63,22,71,23,81,24,87,25,
            91];
        D:=[1,38,1,0,2,77/2,1,81/2,1,0,4,92/3,1,33,1,110/3,1,116/3,1,0,8,59/2,1,
            67/2,1,211/6,1,83/2,1,0,12,126/5,1,26,1,196/5,1,0,15,79/4,1,133/6,1,
            257/10,1,157/6,1,425/12,1,0,20,73/4,1,20,1,22,1,358/15,1,86/3,1,0,
            25,167/10,1,217/12,1,671/30,1,0,28,242/15,1,18,1,361/20,1,313/15,1,
            0,32,877/60,1,45/2,1,0,34,568/35,1,0,35,63/4,1,0];
        I:=[518650,518622,518624,518646,518582,518633,518632,518400,518649,
            518697,518467,518507,518639,518564,518679,518401,518477,518466,
            518424,518601,518594,518434,518523,518428,518521,518370,518518,
            518371,518341,518361,518562,518532,518280,518226,518302];
    elif hash1 eq 1870 and hash2 eq 1131 and hash3 eq 1423 then
        V:=[26,1,27,5,28,17,30,21,32,29];
        D:=[1,89/3,1,0,2,281/12,1,95/4,1,143/6,1,155/6,1,181/6,1,0,7,358/15,1,0,
            8,84/5,1,86/5,1,52/3,1,0,11,763/60,1,275/21,2,787/60,1,53/4,2,0];
        I:=[243638,243158,244618,244084,243476,242838,242547,241919,244240,
            244411,243820,241940,241941,241924,243563,243564];
    elif hash1 eq 2575 and hash2 eq 1613 and hash3 eq 2898 then
        V:=[31,1,32,5,34,19];
        D:=[1,37/2,1,0,2,43/3,1,44/3,1,81/5,1,65/4,1,49/3,2,50/3,1,0,9,38/3,1,
            155/12,1,13,7,40/3,1,0];
        I:=[657021,656998,658435,657713,656975,656698,658133,658576,656635,
            656066,655673,656634,656766,658423,658663,658853,659203,656761];
    elif hash1 eq 3374 and hash2 eq 976 and hash3 eq 134 then
        V:=[27,1,28,7,29,15,30,23,31,31,33,35];
        D:=[1,233/12,1,111/4,1,0,3,18,1,73/4,1,61/3,1,0,6,89/6,1,91/6,1,221/12,
            1,0,9,218/15,1,223/15,1,84/5,2,0,13,44/3,1,0,14,301/30,2,61/6,1,0];
        I:=[500546,500539,500542,500544,500540,500548,500550,500545,500566,
            500547,500549,500554,500559,500556,500563,500555];
    elif hash1 eq 272 and hash2 eq 288 and hash3 eq 1179 then
        V:=[21,1,22,9,23,21,24,37,25,45,26,59,27,69,28,77,30,87];
        D:=[1,43/2,2,151/6,1,163/6,1,0,5,59/3,1,64/3,2,70/3,1,83/3,1,94/3,1,0,
            11,109/6,1,37/2,1,197/10,1,257/12,1,237/10,1,49/2,1,103/4,1,0,18,18,
            1,91/5,3,56/3,2,0,24,197/12,1,67/4,1,221/12,1,599/30,2,41/2,1,125/6,
            1,0,31,223/15,1,50/3,1,253/15,1,203/12,1,0,35,137/10,1,625/42,1,
            229/12,1,0,38,196/15,1,281/21,1,203/15,2,14,2,0,44,61/5,1,0];
        I:=[509834,510033,510193,510026,510056,509822,510160,510187,508830,
            509034,509965,510207,510131,510179,508567,509920,509235,509984,
            509848,509943,510164,510162,510192,509998,509918,509133,510088,
            510142,509061,509908,509839,509092,508545,509976,509803,508894,
            508809,509838,508814,508821,510180,508750,509964,508322];
    elif hash1 eq 2481 and hash2 eq 2150 and hash3 eq 672 then
        V:=[21,1,22,7,23,11,24,19,25,25];
        D:=[1,49/2,1,57/2,1,0,3,21,4,0,7,115/6,1,127/6,2,139/6,1,0,11,69/4,2,
            58/3,2,0,15,109/6,1,0];
        I:=[246210,246068,246064,246067,246134,246216,246211,245904,246117,
            246095,245972,246082,245913,246074,246150];
    elif hash1 eq 977 and hash2 eq 308 and hash3 eq 2071 then
        V:=[19,1,20,5,21,13,22,21,23,31,25,39,26,43];
        D:=[1,41/2,1,0,2,17,1,25,1,98/3,2,0,6,103/6,1,35/2,1,187/6,1,0,9,58/3,1,
            71/3,1,109/4,1,89/3,1,0,13,95/6,1,107/6,1,217/10,1,0,16,89/6,1,0,17,
            218/15,1,44/3,1,199/12,1,0];
        I:=[512651,512715,512728,512703,512746,512719,512572,512745,512637,
            511760,511481,511762,512662,511588,511763,512708,512118,512686,
            512382];
    elif hash1 eq 319 and hash2 eq 3385 and hash3 eq 2641 then
        V:=[23,1,24,9,25,25,26,35,27,49,29,59];
        D:=[1,151/6,2,55/2,1,175/6,1,0,5,58/3,1,59/3,1,65/3,2,93/4,1,24,1,77/3,
            1,26,1,0,13,37/2,1,197/10,1,121/6,2,143/6,1,0,18,49/3,1,215/12,2,
            91/5,1,55/3,2,56/3,1,239/12,3,0,28,491/30,1,33/2,1,539/30,1,659/30,
            1,0,32,401/30,2,27/2,1,83/6,1,0];
        I:=[248615,248619,248165,247721,248437,248380,248166,248539,248488,
            248311,248124,247842,248610,248433,248290,248536,247670,248287,
            248459,248542,248613,248152,248562,248200,248447,248544,248559,
            248401,248493,247689,247551,247527,248296,248440,248533];
    elif hash1 eq 1165 and hash2 eq 2438 and hash3 eq 484 then
        V:=[27,1,28,7,29,11,30,15,31,31];
        D:=[1,85/6,2,29/2,1,0,4,262/15,1,0,5,77/6,1,0,6,38/3,1,193/15,3,155/12,
            3,40/3,2,41/3,1,289/20,1,179/12,1,0,18,157/12,1,79/6,1,0];
        I:=[407159,407459,406091,405811,407961,407377,407390,407475,407757,
            407106,407423,407753,404489,406099,406120,404608,404449,405177,
            405754];
    elif hash1 eq 3186 and hash2 eq 644 and hash3 eq 3402 then
        V:=[17,1,18,5,19,11,20,25,21,35,22,43,23,49,24,63,25,71];
        D:=[1,69/2,2,0,3,104/3,1,35,1,0,5,163/6,1,175/6,1,59/2,1,187/6,1,131/4,
            1,199/6,1,0,11,59/3,1,77/3,1,82/3,3,83/3,1,0,17,119/6,1,121/6,1,
            155/6,1,0,20,215/12,1,239/12,2,0,23,599/30,1,20,1,203/10,1,245/12,1,
            671/30,1,227/10,1,0,29,17,1,401/20,1,121/6,1,0,32,877/60,1,451/30,1,
            181/12,1,185/12,1,0];
        I:=[254770,254773,427206,427354,427356,427419,427392,427023,427217,
            427216,427427,427426,427006,427128,427187,427007,427376,427358,
            426704,427215,426995,427036,426694,426849,426696,426697,426529,
            426537,427201,426532,426504,426690,426692,426693,426734];
    elif hash1 eq 2716 and hash2 eq 2881 and hash3 eq 1316 then
        V:=[11,1,12,9,13,19,14,27,15,33,16,45,17,63,18,85,19,97,20,111,21,125,
            23,143,24,157];
        D:=[1,89/2,1,101/2,2,105/2,1,0,5,110/3,1,122/3,1,134/3,1,152/3,1,0,9,
            147/4,1,179/4,1,283/6,1,0,12,113/3,2,119/3,1,0,15,377/10,1,229/6,1,
            159/4,1,239/6,2,175/4,1,0,21,359/12,1,30,1,91/3,1,478/15,1,383/12,1,
            32,3,98/3,1,34,1,0,31,791/30,1,365/12,1,32,1,385/12,2,971/30,1,
            327/10,1,197/6,1,36,1,433/12,1,481/12,1,0,42,373/15,1,157/6,1,
            398/15,1,319/12,1,674/21,1,0,47,1357/60,1,68/3,1,277/12,1,733/28,1,
            1117/42,1,80/3,1,0,53,1147/60,1,293/15,1,235/12,1,20,1,223/10,1,
            71/3,1,0,59,1311/70,1,1133/60,1,587/30,2,199/10,1,601/30,2,683/30,1,
            231/10,1,1393/60,1,0,69,110/7,1,1327/84,1,238/15,1,1697/90,1,
            1133/60,1,1141/60,1,0,75,739/60,2,0];
        I:=[519900,519857,519904,519881,519885,519798,519901,519872,519842,
            519839,519890,519868,519908,519865,519758,519759,519853,519787,
            519818,519747,519815,519879,519894,519763,519834,519863,519874,
            519897,519902,519875,519781,519782,519819,519806,519833,519770,
            519796,519767,519762,519735,519704,519766,519807,519737,519771,
            519691,519745,519727,519760,519677,519706,519678,519730,519790,
            519854,519898,519726,519676,519687,519682,519752,519776,519757,
            519769,519785,519686,519703,519702,519692,519700,519695,519674,
            519673,519675,519739,519773];
    elif hash1 eq 2105 and hash2 eq 950 and hash3 eq 54 then
        V:=[22,1,23,5,24,11,25,15,26,25,27,29,28,33,29,41];
        D:=[1,107/3,1,0,2,185/12,1,107/6,1,0,4,180/7,1,0,5,193/12,1,283/14,1,
            611/30,2,143/6,1,0,10,380/21,1,0,11,157/12,1,0,12,226/15,1,233/15,1,
            146/9,1,0,15,523/40,1,49/3,1,0];
        I:=[540322,540288,540308,539909,540287,539829,540245,540253,539794,
            539976,540321,540256,540227,539638,539797,540359];
    elif hash1 eq 1400 and hash2 eq 858 and hash3 eq 3311 then
        V:=[25,1,26,7,27,15,28,19];
        D:=[1,139/6,2,47/2,2,0,5,59/3,2,20,1,65/3,4,0,12,109/6,2,0,14,49/3,1,
            50/3,1,73/4,1,55/3,3,0];
        I:=[61479,61480,61451,61473,61468,61471,61475,61401,61410,61412,61413,
            61421,61422,61396,61394,61389,61391,61392,61406];
    elif hash1 eq 2669 and hash2 eq 1315 and hash3 eq 1283 then
        V:=[20,1,21,9,22,13,23,21,24,29,25,35,26,39,27,51,28,55,29,59];
        D:=[1,86/3,1,31,1,98/3,1,0,4,75/2,1,0,5,116/5,1,89/3,1,94/3,1,0,8,
            217/10,1,281/12,1,305/12,1,0,11,111/5,1,105/4,1,0,13,37/2,3,0,16,
            277/15,1,283/15,1,227/12,1,58/3,1,22,1,0,21,103/6,3,0,24,56/3,1,0,
            25,223/20,1,0];
        I:=[540068,540049,540161,540008,540165,540048,540067,540030,540113,
            540044,540032,539864,540054,540055,540116,539870,540099,540057,
            540145,539753,540020,540069,540109,539676,540038];
    elif hash1 eq 1353 and hash2 eq 1740 and hash3 eq 21 then
        V:=[20,1,21,5,22,13,23,19,24,29,25,35,26,41,28,45];
        D:=[1,39,1,0,2,151/6,1,187/6,1,211/6,1,0,5,76/3,2,83/3,1,0,8,133/6,1,
            157/6,1,53/2,1,169/6,1,0,12,67/3,1,73/3,1,0,14,113/6,3,125/6,1,0,18,
            19,1,0,19,187/12,1,0];
        I:=[422895,423244,423559,423666,423465,423496,423226,423229,423552,
            423560,423612,423494,423651,423492,423558,423570,423267,423676,
            422512];
    elif hash1 eq 742 and hash2 eq 2426 and hash3 eq 2234 then
        V:=[25,1,26,7,27,13,28,29,29,41,30,61,31,71];
        D:=[1,35/2,1,107/4,1,0,3,53/3,1,116/5,1,0,5,95/6,3,97/6,3,177/10,1,71/4,
            1,109/6,4,37/2,1,197/10,1,0,19,167/12,1,238/15,1,16,1,49/3,1,20,1,0,
            24,51/4,1,77/6,2,79/6,3,431/30,2,173/12,1,147/10,2,59/4,1,89/6,4,
            33/2,2,0,42,188/15,1,151/12,2,193/15,1,13,1,0,47,341/30,1,0];
        I:=[408630,407308,407122,408013,407680,408649,408651,407119,408585,
            408789,407960,407596,407248,407250,407315,408467,408703,407515,
            406785,407620,408579,407595,406167,408673,408460,408526,408528,
            408532,408533,407522,407619,405082,404543,407847,404138,404113,
            404774,407332,407784,405460,405773,406834,406821,408425,408061,
            404500,407232];
    elif hash1 eq 1870 and hash2 eq 1168 and hash3 eq 2131 then
        V:=[20,1,23,9,24,15,25,19,26,31,27,35,28,45];
        D:=[1,86/3,2,98/3,1,104/3,1,0,5,181/6,1,401/12,1,0,7,56/3,1,0,8,227/14,
            1,491/30,2,351/20,1,18,1,37/2,1,0,14,78/5,1,0,15,13,1,405/28,1,15,1,
            411/20,1,0,19,363/28,1,787/60,1,0];
        I:=[512142,512275,512057,512847,511989,511389,512562,511161,511108,
            512186,511343,512267,511936,511422,512620,511120,510866,510882,
            511111,512162];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 38 mod 47.
function cached_ids_38(P,hash1,hash2,hash3)
    if hash1 eq 2717 and hash2 eq 1260 and hash3 eq 3360 then
        V:=[21,1,25,5,26,13,27,25];
        D:=[1,127/6,2,0,3,20,1,245/12,2,125/6,3,0,9,422/21,2,102/5,1,143/7,1,
            247/12,2,313/15,2,0,17,1093/60,1,0];
        I:=[540186,540200,540201,540185,540187,540181,540182,540199,539837,
            539859,540184,539857,539836,539860,540180,540183,539684];
    elif hash1 eq 38 and hash2 eq 844 and hash3 eq 2576 then
        V:=[18,1,19,11,20,17,21,27,22,41,23,55,24,67,26,73,28,87];
        D:=[1,89/3,1,100/3,1,101/3,1,119/3,1,0,5,191/6,1,151/4,1,0,7,79/3,1,
            359/12,2,97/3,2,103/3,1,0,13,731/30,1,791/30,2,53/2,1,267/10,2,
            163/6,1,899/30,1,0,21,271/12,1,23,1,521/20,1,392/15,1,187/7,1,
            403/15,1,0,27,581/30,1,39/2,1,631/30,1,641/30,1,43/2,1,0,32,389/20,
            1,271/12,1,0,34,211/15,1,71/5,4,43/3,1,947/60,1,95/6,1,191/12,1,0,
            43,1033/84,1,0];
        I:=[429212,429219,429230,429249,429186,429251,429226,429193,429207,
            429196,429253,429223,429200,429198,429217,429227,429215,429247,
            429246,429255,429199,429181,429244,429265,429114,429106,429257,
            429245,429214,429225,429248,429097,429109,429260,429105,429188,
            429189,429194,429187,429099,429107,429115,429086];
    elif hash1 eq 1354 and hash2 eq 150 and hash3 eq 996 then
        V:=[26,1,27,13,28,31,30,41,31,47];
        D:=[1,65/3,1,71/3,1,124/5,1,76/3,2,109/4,1,0,7,233/12,1,197/10,1,79/4,1,
            121/6,2,87/4,1,131/6,1,133/6,1,237/10,1,0,16,191/12,1,49/3,2,215/12,
            1,55/3,2,0,22,289/20,1,175/12,1,0,24,259/20,1,79/6,1,27/2,1,59/4,1,
            0];
        I:=[243723,244182,242164,244397,244498,242242,243127,243725,244119,
            243728,244079,242228,242318,244082,242376,244016,244522,244531,
            244479,242296,243446,241903,243363,244394,244057,244435,242120];
    elif hash1 eq 3187 and hash2 eq 1545 and hash3 eq 328 then
        V:=[21,1,22,13,23,21,24,27,26,31];
        D:=[1,41/2,1,49/2,2,53/2,2,57/2,1,61/2,2,0,9,19,1,62/3,1,74/3,1,0,12,
            43/2,1,47/2,1,0,14,52/3,1,0,15,18,1,0];
        I:=[251475,251100,251502,250867,251547,250728,251516,251570,251474,
            251433,251551,251375,250105,251585,249871];
    elif hash1 eq 2294 and hash2 eq 1988 and hash3 eq 800 then
        V:=[25,1,26,9,28,15,29,21];
        D:=[1,139/6,1,47/2,1,51/2,1,0,4,20,2,64/3,1,0,7,49/3,3,50/3,2,0,12,89/6,
            1,91/6,3,0];
        I:=[670405,670899,670894,670434,670896,670865,670390,670407,670883,
            670254,670400,670389,670251,670397,670422];
    elif hash1 eq 2388 and hash2 eq 2604 and hash3 eq 1569 then
        V:=[29,1,30,11,31,19,32,33,33,41];
        D:=[1,97/6,1,33/2,1,71/4,1,107/6,1,0,5,191/12,1,16,1,49/3,2,0,9,169/12,
            1,173/12,1,29/2,1,59/4,1,89/6,1,491/30,2,0,16,155/12,1,13,1,53/4,1,
            0,19,137/12,1,71/6,1,0];
        I:=[226627,226875,227501,226742,227510,227313,226480,227320,227321,
            227495,226754,226680,226994,226193,227123,227119,227576,226486,
            224652,225035];
    elif hash1 eq 1871 and hash2 eq 553 and hash3 eq 2401 then
        V:=[17,1,18,9,19,21,20,33,21,47,22,63,23,71];
        D:=[1,57/2,4,61/2,2,65/2,2,0,9,23,1,25,4,80/3,1,27,2,29,4,0,21,43/2,2,
            139/6,3,47/2,1,151/6,5,59/2,1,0,33,20,1,64/3,2,65/3,1,22,2,76/3,1,
            77/3,1,0,41,37/2,1,119/6,2,121/6,2,217/10,1,133/6,1,45/2,1,155/6,1,
            0,50,18,1,55/3,1,56/3,2,0,54,167/10,1,101/6,4,0];
        I:=[254289,254343,254347,254358,254346,254361,254313,254360,254357,
            254267,254303,254349,254350,254244,254291,254348,254190,254247,
            254274,254305,254261,254314,254271,254275,254307,254242,254178,
            254215,254216,254252,254255,254123,254241,254189,254254,254209,
            254293,254311,254183,254194,254200,254256,254299,254152,254295,
            254129,254139,254188,254112,254203,254198,254177,254197,254132,
            254131,254248,254249,254251];
    elif hash1 eq 414 and hash2 eq 425 and hash3 eq 2944 then
        V:=[18,1,19,13,20,25,21,33,22,41,23,49,24,53];
        D:=[1,24,3,26,2,28,1,30,1,32,1,0,9,45/2,3,49/2,3,53/2,5,57/2,2,61/2,1,0,
            23,21,2,68/3,3,23,5,0,33,115/6,5,127/6,1,43/2,1,0,40,53/3,1,58/3,1,
            22,1,0,43,107/6,1,0,44,44/3,1,49/3,2,0];
        I:=[253914,253985,253986,253987,253990,253984,253947,253989,253943,
            253944,253999,253858,253991,253998,253823,253859,253899,253946,
            253966,253617,253948,253760,253712,253988,253619,253861,253884,
            253659,253749,253832,253902,253913,253621,253661,253826,253889,
            253938,253890,253906,253673,253616,253572,253660,253725,253569,
            253930];
    elif hash1 eq 1636 and hash2 eq 1073 and hash3 eq 3327 then
        V:=[23,1,24,5,25,15,26,25,27,35,28,51,29,59,30,63];
        D:=[1,55/2,1,0,2,71/3,2,24,1,76/3,1,77/3,1,0,7,37/2,1,233/12,1,119/6,1,
            145/6,1,0,11,55/3,1,56/3,1,61/3,1,109/5,1,0,15,59/4,1,491/30,1,
            197/12,1,101/6,3,103/6,2,35/2,1,221/12,1,0,25,44/3,1,46/3,1,17,1,0,
            28,157/12,2,0,30,139/12,1,0];
        I:=[414761,413939,414477,414566,414644,414518,414664,414271,414759,
            414279,414380,414600,414313,413470,414293,414042,414310,414088,
            414539,414652,414297,414420,414565,413474,414274,414051,413679,
            414294,414370,414295];
    elif hash1 eq 2905 and hash2 eq 1882 and hash3 eq 2060 then
        V:=[21,1,22,7,23,15,24,25,25,35,26,39,28,45];
        D:=[1,175/6,1,79/2,1,0,3,61/4,1,71/3,1,107/3,1,0,6,37/2,1,257/10,1,
            155/6,1,353/12,1,0,10,238/15,1,16,1,81/5,1,27,1,0,14,169/12,1,0,15,
            58/3,1,337/15,1,0,17,127/10,2,0];
        I:=[511369,511202,512247,511654,511380,511651,511558,511548,512333,
            511610,511618,511390,511201,512116,511165,511173,512113,512124];
    elif hash1 eq 1401 and hash2 eq 2844 and hash3 eq 377 then
        V:=[19,1,21,5,22,9,23,19,24,29,25,33,26,41,27,45,28,53,30,67];
        D:=[1,211/6,1,0,2,215/6,1,0,3,359/12,1,91/3,1,383/12,1,407/12,1,0,7,
            49/2,1,107/4,1,851/30,2,173/6,1,0,12,343/15,1,0,13,229/12,1,631/30,
            1,253/12,1,0,16,263/15,1,0,17,157/10,1,481/30,2,97/6,1,0,21,827/60,
            1,185/12,1,31/2,1,385/24,1,241/15,1,653/35,1,0,27,1279/105,1,0];
        I:=[517273,517153,517194,517206,517154,517152,517197,517151,517149,
            517156,517150,517205,517316,517135,517298,517363,517148,517169,
            517200,517162,517057,517040,517039,516998,517001,516977,516981];
    elif hash1 eq 2106 and hash2 eq 1929 and hash3 eq 280 then
        V:=[16,1,17,5,18,13,19,23,20,33,21,41,22,47,23,53,24,57,25,61];
        D:=[1,30,1,0,2,61/2,1,69/2,1,73/2,1,0,5,25,1,27,1,92/3,1,33,2,0,10,
            151/6,1,163/6,2,67/2,1,211/6,1,0,15,59/3,2,101/4,1,133/4,1,0,19,
            119/6,1,305/12,1,0,21,239/12,1,61/3,1,0,23,41/2,1,0,24,307/15,1,0,
            25,31/2,1,0];
        I:=[425395,425406,425404,425363,425409,425400,425357,425148,425301,
            425375,425077,425354,425151,425074,425101,425275,425149,424993,
            425290,424986,425138,425081,425003,424903,424959];
    elif hash1 eq 978 and hash2 eq 2603 and hash3 eq 1681 then
        V:=[21,1,22,7,23,21,24,27,25,33,26,41,27,47,28,51];
        D:=[1,151/6,1,123/4,1,0,3,65/3,1,126/5,1,26,1,109/4,1,83/3,1,88/3,1,0,9,
            157/6,1,181/6,1,0,11,55/3,1,101/5,1,0,13,491/30,3,33/2,1,599/30,1,0,
            18,175/12,1,44/3,1,0,20,13,1,0,21,787/60,1,53/4,1,0];
        I:=[509806,510306,509804,510036,510494,510487,510483,510171,509881,
            509865,510498,510042,509679,509684,509689,509653,509204,509670,
            509794,510635,508873,509678];
    elif hash1 eq 602 and hash2 eq 903 and hash3 eq 1832 then
        V:=[22,1,24,5,25,11,26,17,27,29,28,37,29,53,30,71,31,77];
        D:=[1,23,1,0,2,64/3,1,83/3,1,0,4,253/10,1,273/10,1,0,6,39/2,1,61/3,1,
            67/3,1,121/5,1,105/4,1,0,11,551/30,1,37/2,1,41/2,1,0,14,66/5,1,
            203/12,1,52/3,1,88/5,1,96/5,1,20,1,308/15,1,0,21,541/42,1,331/20,1,
            697/42,1,997/60,1,50/3,1,67/4,1,511/30,1,205/12,1,0,29,787/60,1,
            66/5,1,0,31,321/28,1,0];
        I:=[502750,502745,502738,502088,502490,502173,502386,502381,502097,
            502328,501832,501895,501598,501700,501491,502357,502147,502104,
            501422,502037,501655,502208,501497,501645,502363,502446,501458,
            501528,501660,501424,501039];
    elif hash1 eq 1589 and hash2 eq 3057 and hash3 eq 3178 then
        V:=[20,1,21,5,22,11,23,23,24,31,27,35,28,49];
        D:=[1,86/3,1,0,2,163/6,1,187/6,1,0,4,77/3,2,82/3,1,83/3,1,88/3,1,107/3,
            1,0,10,121/6,1,133/6,1,277/10,1,0,13,23,1,0,14,299/20,1,451/30,2,
            181/12,1,91/6,1,461/30,1,31/2,2,0,22,47/3,1,0];
        I:=[424681,424703,424682,424399,424724,424683,424546,424398,423578,
            424690,424400,423625,424441,422949,422863,422917,422953,424674,
            423577,423485,423626,422587];
    elif hash1 eq 2482 and hash2 eq 2488 and hash3 eq 439 then
        V:=[28,1,29,7,30,15,31,25];
        D:=[1,47/3,2,53/3,2,0,5,63/4,1,95/6,5,97/6,4,0,15,69/5,1,167/12,3,43/3,
            4,44/3,7,0,30,77/6,2,0];
        I:=[667084,667826,666712,667218,665229,666891,666917,667209,667300,
            667708,665173,666450,667246,667711,666756,666657,666908,666958,
            666516,667604,667605,667829,665569,665669,665700,667210,667220,
            667245,667707,666011,666111];
    elif hash1 eq 1824 and hash2 eq 1347 and hash3 eq 1429 then
        V:=[15,1,16,5,17,9,18,17,19,21,20,31,21,37,22,43,23,51,24,55];
        D:=[1,85/2,1,0,2,116/3,1,0,3,187/6,1,199/6,1,67/2,1,0,6,89/3,2,0,8,
            155/6,1,157/6,2,167/6,1,61/2,1,0,13,24,1,129/5,1,0,15,41/2,1,137/6,
            1,0,17,401/20,1,121/6,1,449/20,1,0,20,121/6,1,0,21,233/15,1,0];
        I:=[255350,255193,255205,255204,255202,255183,255200,255191,255184,
            255201,255185,255181,255186,255028,255192,255182,255027,255023,
            255019,255021,255025];
    elif hash1 eq 3140 and hash2 eq 2543 and hash3 eq 3198 then
        V:=[26,1,27,7,28,15,29,21,30,25,31,31];
        D:=[1,77/3,1,83/3,1,0,3,109/6,1,119/6,1,143/6,1,0,6,61/3,1,21,1,0,8,
            217/12,1,0,9,41/3,1,283/15,3,0,13,391/30,1,27/2,4,181/12,1,0];
        I:=[407422,406912,407247,407030,407434,406902,407255,407730,407236,
            404550,405055,405148,407839,407243,407246,407984,407985,405065];
    elif hash1 eq 1166 and hash2 eq 1967 and hash3 eq 1317 then
        V:=[22,1,24,5,25,9,26,17,27,31,28,37,30,41];
        D:=[1,23,1,0,2,117/4,1,0,3,79/4,1,131/6,1,95/4,1,0,6,55/3,1,298/15,1,
            61/3,1,109/5,1,67/3,1,287/12,1,0,12,113/6,1,279/14,1,0,14,211/12,1,
            0,15,787/60,1,0];
        I:=[410942,409888,410234,409799,411080,411129,410248,409804,410989,
            410389,410039,411157,409312,409134,411130];
    elif hash1 eq 1636 and hash2 eq 1063 and hash3 eq 1306 then
        V:=[22,1,23,11,24,15,25,21,26,27,27,33,28,37,29,43];
        D:=[1,68/3,1,86/3,1,92/3,1,110/3,1,0,5,51/2,1,0,6,52/3,1,53/3,1,0,8,
            107/6,2,109/6,1,0,11,18,1,101/5,1,0,13,25/2,1,0,14,188/15,1,155/12,
            1,0,16,59/4,1,0];
        I:=[506650,506158,505934,506186,506582,506591,506590,506090,506598,
            506568,506559,505601,505217,505351,506703,504477];
    elif hash1 eq 1871 and hash2 eq 3305 and hash3 eq 1260 then
        V:=[20,1,21,7,22,15,23,19,25,27,26,35,28,39];
        D:=[1,29,1,41,1,0,3,151/6,1,187/6,2,199/6,1,0,7,83/3,1,0,8,131/6,1,
            133/6,1,155/6,1,0,11,125/6,2,265/12,1,269/12,1,0,15,40/3,1,0,16,
            127/10,1,77/6,1,787/60,1,53/4,1,0];
        I:=[509104,509211,509130,509031,509137,509120,509129,509089,509059,
            509032,509047,509143,509153,509156,509325,509169,509033,509165,
            509040];
    elif hash1 eq 2341 and hash2 eq 2414 and hash3 eq 782 then
        V:=[18,1,19,7,20,19,21,35,22,49,23,59,24,81,25,89,27,93];
        D:=[1,27,1,35,3,0,5,151/6,1,51/2,1,175/6,1,59/2,1,187/6,1,0,10,65/3,1,
            71/3,2,76/3,1,77/3,2,26,2,82/3,1,95/3,1,0,20,121/6,1,131/6,1,95/4,1,
            103/4,1,155/6,3,157/6,1,0,28,55/3,3,56/3,1,263/12,2,89/4,1,0,35,
            33/2,1,551/30,1,37/2,1,187/10,1,113/6,4,115/6,4,241/12,1,245/12,1,
            22,1,265/12,2,0,52,223/12,1,58/3,5,59/3,1,0,59,451/30,1,0,60,145/12,
            1,0];
        I:=[426365,426124,426323,426328,426329,426357,426350,425996,426352,
            426358,426164,426349,426280,425847,426073,426117,426314,426337,
            425782,426159,426125,425766,426165,425707,426074,426167,426119,
            425883,426346,426368,426333,425675,426094,425850,425830,425760,
            425969,425925,425762,425841,425918,426136,425899,426101,426151,
            426363,425754,425774,425726,425656,425909,425633,425702,425708,
            425768,425813,425846,425935,425761,425855];
    elif hash1 eq 1636 and hash2 eq 857 and hash3 eq 3158 then
        V:=[27,1,28,7,29,11,30,19,31,25,32,29];
        D:=[1,83/4,1,91/4,1,0,3,86/5,2,0,5,161/12,1,83/6,2,85/6,2,0,10,37/3,2,
            167/12,1,0,13,149/12,2,0,15,56/5,1,0];
        I:=[485131,484859,484702,484820,484890,484888,484892,484863,485072,
            484625,484685,484883,484832,485000,484807];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 39 mod 47.
function cached_ids_39(P,hash1,hash2,hash3)
    if hash1 eq 1167 and hash2 eq 46 and hash3 eq 2499 then
        V:=[17,1,19,7,20,11,21,23,22,33,23,43,25,49,26,59,28,67];
        D:=[1,65/2,1,85/2,1,0,3,147/4,1,0,4,77/3,1,89/3,1,125/4,1,166/5,1,101/3,
            1,0,9,257/10,1,155/6,1,157/6,1,191/6,2,0,14,24,1,121/5,1,97/4,1,
            131/5,2,0,19,611/30,1,22,1,0,21,371/20,1,1117/60,1,571/30,2,229/12,
            2,0,27,411/28,1,883/60,3,89/6,2,0,33,176/15,1,707/60,1,0];
        I:=[425311,425278,424948,425085,425106,424981,424963,424969,425027,
            425334,425284,424966,425254,425093,424965,424954,424914,425045,
            425028,424957,424872,424895,424890,424975,424866,424888,424867,
            424868,424978,425051,424985,425245,424997,424956];
    elif hash1 eq 3282 and hash2 eq 769 and hash3 eq 1402 then
        V:=[22,1,23,11,24,29,25,37,26,45,27,51,29,59];
        D:=[1,23,1,74/3,1,29,1,92/3,1,0,5,39/2,1,127/6,2,43/2,1,51/2,1,115/4,1,
            187/6,1,131/4,1,67/2,1,0,14,126/5,1,76/3,1,117/4,1,0,17,107/6,1,
            109/6,1,257/12,1,0,20,91/5,1,56/3,1,0,22,89/6,1,18,1,659/30,1,0,25,
            401/30,1,83/6,1,31/2,1,0];
        I:=[411279,411213,411250,410410,411214,411226,411228,411227,411206,
            411357,409687,410459,411391,409677,411237,411430,411207,411215,
            411165,409556,411444,411395,411497,409388,411182,411436,409444];
    elif hash1 eq 1167 and hash2 eq 3401 and hash3 eq 3039 then
        V:=[28,1,29,7,30,13,32,25,33,35];
        D:=[1,59/3,1,85/4,1,0,3,95/6,1,107/6,1,0,5,44/3,2,16,1,49/3,2,50/3,1,
            73/4,1,0,12,193/15,3,155/12,1,13,3,40/3,3,0,22,137/12,1,0];
        I:=[227605,227229,227219,226596,225073,227541,227603,227210,227540,
            226980,225190,224836,226467,227492,224450,224805,226552,227543,
            226743,226927,226929,226750];
    elif hash1 eq 2859 and hash2 eq 214 and hash3 eq 986 then
        V:=[22,1,23,5,24,19,25,27,26,33,27,37,29,45,30,53];
        D:=[1,95/3,1,0,2,277/10,1,293/10,1,119/4,1,179/6,1,181/6,1,193/6,1,0,8,
            287/12,1,73/3,1,311/12,3,0,13,671/30,1,137/6,1,0,15,223/12,1,0,16,
            511/30,1,559/30,1,569/30,1,0,19,817/60,1,137/10,1,55/4,1,0,22,
            181/15,1,0];
        I:=[424531,422826,422591,423434,424552,424499,424440,422697,424533,
            424354,424365,424536,422457,424433,423924,423594,422580,422592,
            423477,423694,423696,422451];
    elif hash1 eq 1308 and hash2 eq 3135 and hash3 eq 796 then
        V:=[22,1,23,13,24,21,25,29,26,39,27,47,28,57,30,63];
        D:=[1,23,1,74/3,1,86/3,1,29,1,31,1,0,6,151/6,1,51/2,1,163/6,1,0,9,116/5,
            1,93/4,1,71/3,2,0,13,109/6,1,197/10,1,121/6,1,217/10,1,0,17,91/5,1,
            55/3,1,56/3,1,0,20,29/2,1,491/30,1,33/2,1,551/30,2,0,25,85/6,1,15,1,
            0,27,173/15,2,0];
        I:=[251161,250982,250940,251079,251230,250559,250424,250954,250406,
            250676,250592,250608,251242,250498,250967,250588,250501,250579,
            250279,250500,250677,250581,249836,250594,249830,251255,250560,
            250562];
    elif hash1 eq 697 and hash2 eq 1525 and hash3 eq 1377 then
        V:=[23,1,25,5,26,13,27,19,28,23,30,29,32,33];
        D:=[1,473/12,1,0,2,791/30,1,851/30,1,911/30,1,0,5,112/5,1,323/12,2,0,8,
            619/30,1,0,9,115/6,1,77/4,1,0,11,827/60,1,0,12,865/84,1,161/15,2,
            287/24,1,0];
        I:=[508854,509866,510130,510027,510117,508490,508878,509306,508833,
            508877,508845,508306,509710,509770,508300];
    elif hash1 eq 1684 and hash2 eq 1691 and hash3 eq 1920 then
        V:=[24,1,25,5,26,17,27,29,28,35,29,41,30,45,31,49];
        D:=[1,92/3,1,0,2,43/2,1,139/6,1,47/2,1,115/4,1,59/2,1,0,7,64/3,1,65/3,1,
            116/5,1,70/3,1,71/3,1,0,12,233/12,1,131/6,1,0,14,238/15,1,239/12,1,
            0,16,491/30,1,0,17,218/15,1,0,18,877/60,1,0];
        I:=[236161,236326,236204,236502,235477,235989,236368,236503,236095,
            236317,236151,235791,236176,235808,235958,236184,235590,235279];
    elif hash1 eq 39 and hash2 eq 2522 and hash3 eq 606 then
        V:=[20,1,21,11,22,17,23,27];
        D:=[1,21,1,74/3,1,31,1,33,1,0,5,151/6,1,163/6,4,0,10,116/5,1,93/4,1,
            101/4,1,76/3,1,0,14,97/6,1,257/12,1,87/4,1,0];
        I:=[424658,424259,424625,424623,424651,422723,423762,423768,424617,
            422735,423767,423763,423936,424653,423084,423012];
    elif hash1 eq 2859 and hash2 eq 2756 and hash3 eq 1498 then
        V:=[23,1,24,7,25,13,26,21,27,25,29,31];
        D:=[1,61/2,2,65/2,1,0,4,27,2,29,3,0,9,43/2,1,47/2,1,151/6,1,0,12,65/3,1,
            0,13,119/6,1,131/6,1,0,15,101/6,1,0];
        I:=[244011,244181,244045,244166,244236,244060,244061,244193,244212,
            244417,244076,244486,244125,242185,242183];
    elif hash1 eq 3282 and hash2 eq 909 and hash3 eq 2349 then
        V:=[25,1,26,9,28,13,29,23,30,29];
        D:=[1,91/6,1,31/2,1,491/30,2,0,5,22,1,0,6,316/21,1,76/5,2,323/21,1,
            233/15,2,0,12,431/30,3,29/2,2,0,17,73/6,1,0];
        I:=[539820,539731,539748,539819,539975,539815,539721,539814,539719,
            539717,539750,539751,539767,539768,539729,539742,539707];
    elif hash1 eq 1026 and hash2 eq 1753 and hash3 eq 3059 then
        V:=[24,1,25,13,26,17,27,25,28,29,29,33,30,41];
        D:=[1,59/3,1,20,1,65/3,1,24,1,77/3,1,0,6,119/6,1,0,7,50/3,2,55/3,1,62/3,
            1,0,11,37/2,1,0,12,14,1,0,13,401/30,1,27/2,1,83/6,1,0,16,35/3,1,0];
        I:=[417519,417181,417053,417312,417102,417412,416682,417603,416999,
            417069,415693,416520,417204,416959,416765,415282];
    elif hash1 eq 2154 and hash2 eq 1114 and hash3 eq 879 then
        V:=[23,1,24,9,25,15,26,21,27,31];
        D:=[1,115/6,2,39/2,1,47/2,1,0,5,53/3,1,59/3,2,0,8,71/4,2,107/6,2,0,12,
            191/12,1,49/3,1,215/12,1,55/3,1,0,16,59/4,1,89/6,1,0];
        I:=[249666,249669,249680,249335,249724,248737,248949,249342,249351,
            249272,249409,248438,249363,248163,247688,248341,248018];
    elif hash1 eq 1590 and hash2 eq 921 and hash3 eq 3435 then
        V:=[22,1,23,9,24,17,25,29,26,39,27,51];
        D:=[1,23,3,25,2,27,1,0,7,127/6,7,43/2,2,139/6,2,0,18,53/3,1,18,1,58/3,2,
            59/3,7,64/3,1,0,30,71/4,3,107/6,3,109/6,7,37/2,1,0,44,191/12,2,16,1,
            65/4,1,49/3,1,55/3,1,0,50,29/2,1,147/10,1,89/6,1,0];
        I:=[673558,673670,673685,673565,673667,673381,673542,673587,673589,
            673618,673634,673640,673725,673569,673672,673437,673439,673484,
            673605,673398,673418,673303,673405,673454,673465,673580,673606,
            673646,673447,673416,673420,673530,673339,673544,673698,673247,
            673264,673329,673359,673361,673543,673583,673631,673463,673474,
            673433,673263,673442,673228,673289,673276,673305];
    elif hash1 eq 3423 and hash2 eq 90 and hash3 eq 787 then
        V:=[21,1,22,11,23,17,24,37,25,45,26,57,27,63,28,69];
        D:=[1,47/2,1,151/6,2,107/4,1,163/6,1,0,6,101/4,1,83/3,1,0,8,217/10,2,
            133/6,1,95/4,1,143/6,1,253/10,1,305/12,1,257/10,1,103/4,1,157/6,1,0,
            18,62/3,1,352/15,1,74/3,1,0,21,167/10,1,245/12,1,41/2,1,125/6,2,
            471/20,1,0,27,344/21,1,64/3,1,0,29,265/14,1,229/12,1,0,31,203/15,1,
            907/60,1,309/20,1,0];
        I:=[509053,509070,509087,509136,509338,509117,509077,509062,509139,
            509051,509108,509041,509167,509100,509152,509171,509166,508599,
            509080,509076,508529,509027,508949,508645,509170,509149,508684,
            508707,508552,508585,508524,508624,508695];
    elif hash1 eq 274 and hash2 eq 3059 and hash3 eq 2748 then
        V:=[26,1,27,7,28,13,29,19];
        D:=[1,56/3,2,19,4,0,7,103/6,1,35/2,1,0,9,46/3,1,47/3,5,0,15,55/4,1,83/6,
            2,85/6,3,0];
        I:=[667292,667355,667116,667118,667174,667365,665519,665657,667391,
            665508,665630,667055,667109,667345,666886,665595,667049,665802,
            665832,667139];
    elif hash1 eq 1261 and hash2 eq 2175 and hash3 eq 2973 then
        V:=[21,1,22,7,23,17,24,23,25,31,26,39,29,57];
        D:=[1,175/6,1,187/6,1,0,3,76/3,2,88/3,1,89/3,1,30,1,0,8,281/12,1,257/10,
            1,0,10,73/4,1,68/3,2,74/3,1,0,14,221/12,1,113/6,1,659/30,1,0,17,
            203/12,1,17,3,277/15,1,223/12,1,283/15,1,19,3,77/4,1,58/3,1,0,29,
            47/4,1,0];
        I:=[421858,421860,421764,421885,421883,421601,422003,421739,421383,
            421732,421418,421599,421868,421400,421719,421179,421735,421577,
            421994,422015,421173,421157,421411,421470,421770,421774,421427,
            421711,421467];
    elif hash1 eq 2859 and hash2 eq 822 and hash3 eq 9 then
        V:=[16,1,17,13,18,23,19,41,20,51,21,61,22,69,23,81,24,89,25,93];
        D:=[1,26,1,28,1,30,1,32,1,34,1,0,6,57/2,1,61/2,1,65/2,1,73/2,1,0,10,21,
            1,68/3,1,74/3,1,25,1,86/3,1,33,1,35,2,110/3,1,0,19,43/2,1,139/6,1,
            47/2,1,175/6,4,0,26,71/3,3,136/5,1,83/3,1,28,1,0,32,107/6,1,119/6,2,
            87/4,1,0,36,49/3,1,268/15,1,263/12,1,89/4,1,67/3,1,0,41,33/2,2,45/2,
            1,137/6,1,0,45,52/3,2,0,47,63/4,1,0];
        I:=[515517,515587,515479,515566,515561,515513,515489,515570,515567,
            515465,515503,515082,515216,515586,515563,514777,515533,515049,
            515564,515172,515167,515197,515358,515485,515507,515101,515169,
            515211,514283,514705,514768,515134,514292,515571,515367,515319,
            514497,514927,514741,514532,515133,515323,514117,514053,514287,
            514304,514081];
    elif hash1 eq 462 and hash2 eq 699 and hash3 eq 1420 then
        V:=[25,1,26,7,27,15,28,25,29,35,30,51,31,59];
        D:=[1,119/6,1,273/10,1,0,3,55/3,1,24,1,97/4,1,0,6,113/6,1,245/12,1,47/2,
            1,293/12,1,0,10,329/20,1,199/12,1,77/4,1,143/7,1,0,14,391/30,1,
            401/30,1,451/30,1,697/42,1,997/60,2,499/30,1,50/3,1,0,22,787/60,1,
            361/24,1,226/15,1,0,25,347/30,1,27/2,1,0];
        I:=[507958,507972,507983,507922,507933,507992,507849,507978,507764,
            507817,507987,507921,507866,508034,508016,507957,507853,507809,
            507830,507986,507947,507893,507757,507746,507885,507709];
    elif hash1 eq 556 and hash2 eq 275 and hash3 eq 1761 then
        V:=[21,1,22,5,23,15,24,25,25,33,26,43,27,55,28,73,29,81];
        D:=[1,175/6,1,0,2,76/3,2,77/3,1,83/3,2,89/3,1,0,8,133/6,1,145/6,1,
            305/12,1,155/6,1,0,12,215/12,1,101/5,1,105/4,1,0,15,193/12,1,217/12,
            1,221/12,1,245/12,1,0,19,175/12,1,223/15,1,17,1,223/12,1,283/15,1,0,
            24,401/30,1,59/4,1,149/10,1,15,1,451/30,3,181/12,3,91/6,3,31/2,1,0,
            38,66/5,1,61/4,1,1123/60,1,0,41,697/60,2,35/3,1,0];
        I:=[424702,424564,424790,424590,423003,423644,423952,424566,423853,
            423384,423624,422920,423004,422392,423438,423601,423524,423705,
            423342,422979,423988,423543,422645,423240,422459,423374,423310,
            423348,423350,423497,422668,422843,424215,424554,424710,424759,
            423425,422977,422503,422334,422705,423463,423472];
    elif hash1 eq 932 and hash2 eq 1013 and hash3 eq 2852 then
        V:=[27,1,28,9,29,13,30,25,31,33,32,37];
        D:=[1,139/6,3,47/2,2,151/6,1,0,7,59/3,1,0,8,95/6,1,109/6,3,37/2,1,119/6,
            1,121/6,1,0,15,49/3,2,50/3,2,215/12,2,0,21,29/2,2,0,23,40/3,1,0];
        I:=[666793,667224,667786,667530,667638,667665,667596,666622,666826,
            666975,667810,667751,666578,667808,665448,667589,666463,667643,
            665445,666333,666022,666440,666397];
    elif hash1 eq 1731 and hash2 eq 2979 and hash3 eq 559 then
        V:=[18,1,19,9,20,21,21,29,22,39,23,47,24,53,26,57,27,61];
        D:=[1,35,1,122/3,1,140/3,1,0,4,59/2,1,199/6,1,67/2,1,211/6,1,87/2,1,0,9,
            28,1,107/3,1,130/3,1,0,12,157/6,1,167/6,1,217/6,1,253/6,1,0,16,
            388/15,1,151/5,1,137/4,1,0,19,911/30,1,185/6,1,0,21,227/12,1,0,22,
            76/5,1,0,23,527/30,1,541/30,1,0];
        I:=[542719,542720,542685,542715,542684,542693,542812,542681,542716,
            542669,542776,542689,542826,542700,542706,542686,542671,542666,
            542688,542813,542705,542697,542786,542751];
    elif hash1 eq 2859 and hash2 eq 1828 and hash3 eq 1731 then
        V:=[26,1,27,7,28,17,29,33,30,37,31,45];
        D:=[1,25,1,27,1,0,3,39/2,1,127/6,1,43/2,1,139/6,5,0,11,69/4,1,96/5,1,
            59/3,2,20,1,85/4,1,64/3,4,70/3,2,0,23,109/6,1,0,24,238/15,1,16,1,
            215/12,1,0,27,431/30,1,0];
        I:=[668949,669307,669189,669272,669112,668689,668710,668944,669188,
            669206,668643,668544,668629,668631,669163,668141,668492,668633,
            669201,669276,668547,668686,668617,668150,669208,668806,668866];
    elif hash1 eq 838 and hash2 eq 341 and hash3 eq 2141 then
        V:=[14,1,15,13,16,29,17,59,18,83,19,105,20,127,21,151,22,181,23,197,24,
            203];
        D:=[1,30,1,32,1,34,1,36,1,40,1,0,6,49/2,3,57/2,4,61/2,2,65/2,1,69/2,2,
            73/2,2,81/2,1,0,21,74/3,1,80/3,1,27,1,86/3,1,29,1,92/3,1,31,2,98/3,
            1,33,1,104/3,3,35,1,110/3,1,37,2,43,1,0,39,43/2,2,91/4,1,47/2,1,
            151/6,1,51/2,2,163/6,2,175/6,3,63/2,1,199/6,2,211/6,2,71/2,2,0,58,
            20,1,64/3,1,65/3,3,22,2,93/4,1,70/3,1,71/3,2,83/3,3,88/3,1,36,1,0,
            74,79/4,1,119/6,5,131/6,2,133/6,2,95/4,1,143/6,1,145/6,2,277/10,1,
            111/4,1,169/6,1,0,91,18,1,55/3,2,298/15,1,239/12,1,20,1,61/3,1,
            263/12,1,111/5,1,89/4,2,68/3,1,131/5,1,0,104,91/6,2,167/10,1,101/6,
            2,359/20,1,18,1,217/12,1,551/30,2,399/20,1,20,1,245/12,1,22,1,
            265/12,1,269/12,2,45/2,2,0,123,248/15,3,50/3,1,17,1,52/3,1,332/15,1,
            68/3,1,23,1,0,132,299/20,1,461/30,1,0,134,191/12,1,16,1,0];
        I:=[428561,428567,428553,428552,428554,428519,428569,428576,428265,
            428560,428570,428571,428486,428539,428520,428548,428549,428499,
            428579,428530,428545,428373,428353,428462,428491,428440,428493,
            428565,428333,428551,428438,428489,428516,428495,428528,428497,
            428504,428076,428492,428559,428303,428564,428321,428267,428355,
            428346,428481,428276,428283,428494,428338,428306,428396,428213,
            428282,428269,428317,428308,428514,428348,428465,428544,428288,
            428550,428297,428312,428301,428475,428081,428110,428392,428280,
            428127,428270,428328,428359,428474,428490,428511,428329,428386,
            427934,428426,428343,428005,428263,428325,427756,428119,428063,
            428512,428477,428535,428014,428351,427755,428457,428375,428131,
            428137,428221,428103,427792,428293,428299,427979,428008,428069,
            428132,428278,428337,428071,428145,427765,427773,427767,427939,
            427766,427867,428013,427778,428115,428017,428039,428060,428524,
            428018,428382,427692,427696,427694,428139,428138,427686,427681];
    elif hash1 eq 2483 and hash2 eq 3444 and hash3 eq 2265 then
        V:=[23,1,24,5,25,9,26,15,28,25,29,31,30,35];
        D:=[1,199/6,2,0,3,88/3,1,0,4,143/6,1,145/6,1,0,6,61/3,1,358/15,1,287/12,
            1,74/3,1,0,10,203/12,1,17,1,0,12,451/30,1,0,13,53/4,1,67/5,1,203/15,
            2,0];
        I:=[417901,417910,417971,417073,418181,417093,416900,416736,414983,
            416981,416336,417062,418262,416603,417137,417155];
    elif hash1 eq 2671 and hash2 eq 226 and hash3 eq 2777 then
        V:=[18,1,19,9,20,21,21,35,22,45,23,59,24,73,25,81,26,89,27,103];
        D:=[1,33,2,116/3,1,45,1,0,5,187/6,1,211/6,1,71/2,1,235/6,1,163/4,1,0,10,
            76/3,1,88/3,1,125/4,2,95/3,1,176/5,1,107/3,2,0,18,329/12,1,277/10,1,
            61/2,1,143/4,1,0,22,61/3,1,67/3,1,23,1,358/15,1,311/12,1,141/5,1,0,
            28,611/30,1,245/12,1,439/20,1,227/10,1,139/6,1,719/30,1,0,34,283/15,
            1,19,1,251/12,1,0,37,31/2,3,511/30,1,35/2,2,0,43,907/60,1,46/3,1,
            233/15,1,187/12,1,86/5,1,52/3,1,0,49,1051/70,2,1271/84,1,91/6,2,
            653/42,1,469/30,2,0];
        I:=[515421,515422,515119,515396,515427,515106,515398,515404,515195,
            515400,515403,515190,515389,515395,514715,515394,515399,515387,
            514712,515383,514450,515424,515206,515391,514360,514451,514469,
            514398,514390,514470,514476,515449,514224,514477,515442,515243,
            515392,515393,515397,514617,514226,514227,514399,515441,514445,
            514444,514244,514229,514006,514024,514033,514001,514025,514014,
            514011,514012];
    elif hash1 eq 1355 and hash2 eq 2798 and hash3 eq 3029 then
        V:=[24,1,25,9,26,21,27,27,29,31];
        D:=[1,23,3,74/3,1,25,3,0,8,127/6,2,43/2,2,139/6,4,47/2,1,51/2,1,0,18,
            65/3,2,22,1,0,21,107/6,2,0,23,431/30,1,147/10,1,89/6,4,101/6,1,0];
        I:=[671861,672024,672034,672019,671763,671806,672193,671974,672072,
            671841,672199,671994,672022,672033,672181,671690,671577,671230,
            671718,672200,671264,671750,671391,671291,671232,671626,671689,
            672071,671142];
    elif hash1 eq 3141 and hash2 eq 2783 and hash3 eq 3347 then
        V:=[20,1,21,5,22,17,23,25,24,41,25,55,26,67,27,73,28,85];
        D:=[1,92/3,1,0,2,151/6,1,175/6,1,187/6,1,63/2,1,211/6,1,0,7,71/3,1,77/3,
            1,109/4,1,0,10,121/6,1,131/6,2,45/2,1,237/10,1,95/4,1,277/10,1,
            169/6,1,0,18,101/5,1,328/15,1,111/5,1,68/3,1,23,1,358/15,1,0,24,
            37/2,1,187/10,1,113/6,1,611/30,1,245/12,1,0,29,46/3,2,248/15,1,0,32,
            299/20,1,181/12,1,91/6,1,31/2,2,511/30,2,0,39,419/28,1,907/60,1,
            323/21,1,233/15,2,187/12,1,0];
        I:=[424550,424551,424468,424532,424535,424537,424542,424437,423929,
            424436,424329,424471,424540,423542,423922,423646,423685,422834,
            423647,423557,423555,424438,423645,424490,423540,424475,423608,
            423926,424397,424541,422965,423526,423708,424434,423486,423707,
            422553,422572,422509,422516,422521,422508,422517,422510];
    elif hash1 eq 509 and hash2 eq 1467 and hash3 eq 431 then
        V:=[24,1,25,11,26,29,27,43,28,51,29,59,30,63];
        D:=[1,59/3,1,65/3,1,70/3,1,26,1,0,5,109/6,2,233/12,1,197/10,1,121/6,1,
            217/10,1,87/4,2,133/6,1,95/4,1,0,15,191/12,1,49/3,1,268/15,1,215/12,
            1,91/5,1,101/5,1,0,21,491/30,1,101/6,1,183/10,1,0,24,175/12,1,76/5,
            1,115/7,1,0,27,401/30,1,0,28,173/15,1,61/5,1,0];
        I:=[418658,418468,418291,418516,417620,417859,418493,417615,418091,
            417687,417854,418614,418382,417645,417400,417396,416317,418634,
            417386,417070,416566,415739,416246,417192,416316,415451,415949,
            417409,415529];
    elif hash1 eq 3188 and hash2 eq 1852 and hash3 eq 2924 then
        V:=[19,1,20,7,21,17,22,23,23,29];
        D:=[1,49/2,3,53/2,1,0,5,23,1,74/3,1,80/3,1,27,1,0,9,127/6,2,47/2,1,0,12,
            53/3,1,65/3,1,0,14,121/6,1,257/12,1,0];
        I:=[424593,424842,424845,424596,424597,424746,424714,424160,424177,
            424579,423547,424559,423520,423576,423443];
    elif hash1 eq 2812 and hash2 eq 2634 and hash3 eq 997 then
        V:=[18,1,19,9,20,21,21,35,22,49,23,71,24,89,25,111,26,121,27,125];
        D:=[1,27,1,31,1,33,1,0,4,151/6,1,163/6,3,59/2,1,199/6,1,211/6,1,0,11,
            65/3,2,71/3,1,101/4,1,76/3,2,26,1,82/3,1,0,19,79/4,1,133/6,2,145/6,
            1,155/6,1,157/6,3,111/4,1,0,28,55/3,1,19,1,239/12,2,20,1,61/3,3,
            62/3,1,263/12,2,111/5,1,67/3,3,311/12,1,0,44,217/12,1,221/12,1,37/2,
            1,187/10,1,113/6,6,115/6,2,611/30,1,125/6,1,0,58,46/3,1,199/12,1,
            361/20,2,272/15,1,109/6,1,56/3,1,283/15,1,227/12,1,19,2,96/5,4,0,73,
            451/30,2,181/12,1,91/6,1,185/12,1,0,78,907/60,1,0,79,145/12,2,0];
        I:=[426318,426309,426310,426315,426265,426321,426327,426174,426291,
            425878,426013,426389,426190,426197,426285,426324,426191,426192,
            426063,426312,426313,426186,425928,425957,426118,426187,425816,
            426380,426185,425815,426076,426008,426048,426253,426387,426195,
            425679,425810,425958,425829,425990,426002,425628,426046,425831,
            426239,425838,425987,426004,426020,426182,426246,426316,426188,
            426189,425965,425787,425784,425753,425822,425858,425747,425745,
            425668,425778,425670,425700,425811,425698,425805,425857,425870,
            425776,425794,425826,425832,425867,425654,425749,425859];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 40 mod 47.
function cached_ids_40(P,hash1,hash2,hash3)
    if hash1 eq 3142 and hash2 eq 900 and hash3 eq 3064 then
        V:=[27,1,28,9,29,15,30,19,31,25];
        D:=[1,43/2,1,139/6,2,47/2,1,0,5,59/3,2,20,2,0,9,107/6,1,0,10,49/3,3,
            50/3,1,0,14,29/2,1,89/6,3,0];
        I:=[60569,60642,60644,60810,60665,60871,60571,60638,60522,60427,60428,
            60599,60867,60655,60382,60624,60633];
    elif hash1 eq 839 and hash2 eq 2626 and hash3 eq 2916 then
        V:=[16,1,17,7,18,17,19,25,20,29,21,37,22,41];
        D:=[1,28,1,34,1,0,3,45/2,1,53/2,1,57/2,2,61/2,1,0,8,21,1,98/3,1,35,1,0,
            11,175/6,1,0,12,65/3,2,71/3,1,24,1,0,16,107/6,2,0,18,263/12,1,67/3,
            1,0];
        I:=[515609,515611,515616,515407,515516,515610,515557,515614,515523,
            515328,515426,515075,515527,514475,515417,515491,515602,514436,
            514690];
    elif hash1 eq 463 and hash2 eq 1202 and hash3 eq 825 then
        V:=[23,1,24,9,25,15,26,21,28,31,29,35,30,45];
        D:=[1,237/10,1,329/12,1,191/6,1,0,4,263/12,1,358/15,2,0,7,243/10,1,53/2,
            1,0,9,278/15,1,283/15,1,96/5,1,337/15,1,0,13,323/21,1,0,14,797/60,1,
            377/28,1,569/42,1,653/42,1,0,18,176/15,1,707/60,1,485/36,1,1145/84,
            1,0];
        I:=[509808,509861,510152,509859,509860,509863,508943,508470,509176,
            509857,508395,509321,508291,508810,508811,508483,508484,509874,
            510126,508293,508294];
    elif hash1 eq 1967 and hash2 eq 484 and hash3 eq 194 then
        V:=[22,1,23,5,24,9,25,15,26,23,27,29,28,39];
        D:=[1,21,1,0,2,43/2,1,0,3,53/3,1,58/3,1,0,5,107/6,1,37/2,1,119/6,1,0,8,
            65/4,1,50/3,1,0,10,29/2,1,89/6,1,91/6,1,103/6,1,0,14,15,1,46/3,1,0];
        I:=[414670,414586,414667,414661,414133,414596,413903,414300,414007,
            414702,414583,414055,413561,413506,413502];
    elif hash1 eq 792 and hash2 eq 1043 and hash3 eq 82 then
        V:=[19,1,20,5,21,11,22,17,23,21,24,25,25,29];
        D:=[1,45/2,1,0,2,21,3,68/3,2,0,7,127/6,3,139/6,1,0,11,59/3,1,0,12,109/6,
            3,0,15,49/3,1,0,16,91/6,1,0];
        I:=[253436,253202,253207,253264,253236,253294,252965,253059,253388,
            252645,252883,252678,253104,253106,253238,252673];
    elif hash1 eq 2954 and hash2 eq 1582 and hash3 eq 1292 then
        V:=[23,1,24,7,25,17,27,25,28,35,29,47];
        D:=[1,51/2,2,163/6,1,0,4,22,4,71/3,2,24,1,77/3,1,0,12,121/6,1,41/2,1,
            87/4,1,0,15,197/12,1,33/2,1,167/10,1,221/12,1,0,19,29/2,1,179/12,1,
            15,1,76/5,1,46/3,1,0,24,157/12,1,79/6,2,0];
        I:=[249492,249692,249524,248981,249378,249382,249657,249476,249588,
            248773,249527,249560,249339,247685,247589,248752,248094,247710,
            248932,247866,248681,247783,248927,247871,247874,247880];
    elif hash1 eq 2954 and hash2 eq 2645 and hash3 eq 2873 then
        V:=[23,1,24,5,25,13,27,25,28,31,29,35,30,39];
        D:=[1,151/6,1,0,2,65/3,1,93/4,1,70/3,1,0,5,233/12,1,197/10,1,79/4,1,
            121/6,1,217/10,2,0,11,193/12,1,491/30,1,0,13,44/3,1,0,14,157/12,1,0,
            15,35/3,1,0];
        I:=[249199,248262,249015,248509,249447,248385,248214,249191,248202,
            248543,249466,248409,248609,248586,248216];
    elif hash1 eq 557 and hash2 eq 2608 and hash3 eq 2138 then
        V:=[29,1,31,7,32,17,33,21];
        D:=[1,133/6,1,145/6,1,0,3,173/12,1,89/6,1,167/10,2,101/6,3,0,10,13,1,0,
            11,391/30,1,79/6,2,401/30,2,83/6,1,0];
        I:=[228650,228622,227586,227937,225578,226245,226223,227473,228619,
            227279,225547,228640,228660,226216,227142,227141];
    elif hash1 eq 2437 and hash2 eq 663 and hash3 eq 1273 then
        V:=[14,1,16,5,17,9,18,15,19,21,20,31,21,35,22,39,23,45,24,51];
        D:=[1,38,1,0,2,33,1,0,3,199/6,1,211/6,1,0,5,133/4,1,106/3,1,0,7,79/4,1,
            119/6,1,401/12,1,203/6,1,0,11,61/3,1,0,12,245/12,1,0,13,307/15,1,
            343/15,1,0,15,411/20,1,62/3,1,0,17,187/12,1,0];
        I:=[429062,429044,429031,428896,428878,428871,428949,429038,428784,
            428731,429025,428945,428775,428637,428653,428633,428718];
    elif hash1 eq 228 and hash2 eq 2256 and hash3 eq 1799 then
        V:=[27,1,28,5,29,17,30,29,31,37,32,41];
        D:=[1,59/2,1,0,2,65/3,2,70/3,1,71/3,1,101/4,1,77/3,4,0,11,197/10,1,79/4,
            1,119/6,2,121/6,1,217/10,1,0,17,91/5,1,55/3,1,56/3,1,0,20,33/2,1,0,
            21,85/6,1,0];
        I:=[669174,668557,668911,669205,669173,668692,668090,668103,668154,
            668159,668562,668684,669007,669213,668566,668098,668151,669172,
            669334,668147,668083];
    elif hash1 eq 3330 and hash2 eq 509 and hash3 eq 685 then
        V:=[16,1,17,7,18,13,19,19,20,25,21,31];
        D:=[1,30,1,34,1,0,3,53/2,2,57/2,1,0,6,25,2,29,1,0,9,47/2,1,51/2,1,0,11,
            65/3,3,22,5,0,19,133/6,1,0];
        I:=[674662,674658,674655,674659,674654,674656,674660,674647,674650,
            674649,674645,674651,674652,674646,674648,674653,674657,674661,
            674644];
    elif hash1 eq 745 and hash2 eq 303 and hash3 eq 917 then
        V:=[20,1,21,5,22,17,23,25,24,29,25,35,26,43,27,49];
        D:=[1,35,1,0,2,55/2,1,175/6,1,187/6,1,211/6,2,75/2,1,0,8,26,1,82/3,1,
            83/3,2,0,12,145/6,2,0,14,50/3,1,20,1,0,16,193/12,1,197/12,1,33/2,2,
            0,20,283/15,1,227/12,1,0,22,401/30,1,141/10,1,853/60,1,451/30,1,
            31/2,1,331/20,1,50/3,1,0];
        I:=[420442,419938,420249,419935,419937,420257,419583,419689,420245,
            419683,419685,419519,420305,419759,420231,419763,419871,419572,
            420253,419762,419624,419788,419696,419558,419614,420252,419029,
            419033];
    elif hash1 eq 3236 and hash2 eq 2922 and hash3 eq 1296 then
        V:=[23,1,24,9,25,15,26,19,28,29,29,35,30,41];
        D:=[1,95/6,1,277/10,1,293/10,1,0,4,328/15,1,74/3,1,0,6,29/2,1,0,7,
            218/15,1,175/12,1,386/21,1,22,1,0,11,787/60,2,61/4,1,0,14,1103/84,1,
            66/5,1,0,16,977/84,1,0];
        I:=[509837,508944,509671,509177,508942,509830,509836,509828,508564,
            508888,509795,509873,508791,508829,508828,508304];
    elif hash1 eq 1215 and hash2 eq 1062 and hash3 eq 3010 then
        V:=[19,1,21,7,22,15,23,23,24,29,25,33,27,37,28,41];
        D:=[1,53/2,1,65/2,1,0,3,139/6,1,151/6,1,175/6,1,0,6,47/3,1,70/3,1,109/4,
            1,0,9,79/4,1,257/12,1,0,11,71/5,1,0,12,221/12,2,0,14,391/30,1,0,15,
            173/15,2,0];
        I:=[509016,508998,508924,509021,508930,508934,508979,508933,509336,
            508926,508909,508922,509239,508935,508919,508929];
    elif hash1 eq 228 and hash2 eq 2344 and hash3 eq 1900 then
        V:=[24,1,25,5,26,13,27,17,28,25,29,31];
        D:=[1,31,1,0,2,139/6,1,151/6,2,163/6,2,0,7,106/5,1,0,8,97/6,2,71/4,1,
            107/6,1,0,12,73/4,1,55/3,1,0,14,419/30,1,14,1,169/12,1,29/2,2,89/6,
            2,0];
        I:=[245295,243039,244814,245391,242473,242591,242469,242590,244882,
            243944,245089,242019,242429,243132,243827,243388,243298,243357,
            245150,245181];
    elif hash1 eq 2860 and hash2 eq 2115 and hash3 eq 3202 then
        V:=[25,1,26,5,27,17,28,21,29,29,30,41,31,49];
        D:=[1,99/4,1,0,2,52/3,1,53/3,1,59/3,2,64/3,1,71/3,2,0,9,109/6,1,0,10,
            81/5,1,50/3,1,262/15,1,0,13,199/14,1,147/10,1,33/2,2,221/12,1,
            399/20,1,0,19,72/5,1,101/7,1,223/15,1,0,22,259/20,1,391/30,1,0];
        I:=[399184,399648,399759,398789,399624,398666,399006,399007,398809,
            398751,399166,397060,396251,399668,395873,398941,397902,397486,
            398045,395290,398150,395753,398455];
    elif hash1 eq 2437 and hash2 eq 3068 and hash3 eq 952 then
        V:=[15,1,16,11,17,21,18,37,19,51,20,61,21,69,22,87,23,105,26,119];
        D:=[1,57/2,1,65/2,1,73/2,1,77/2,1,0,5,86/3,1,92/3,1,31,1,98/3,1,0,9,
            175/6,1,59/2,1,187/6,1,63/2,1,75/2,1,235/6,2,79/2,1,0,17,24,1,88/3,
            1,156/5,1,125/4,1,95/3,3,100/3,1,0,25,127/4,1,191/6,4,193/6,2,65/2,
            1,0,33,287/12,2,121/5,1,26,1,0,37,101/6,1,551/30,1,37/2,2,113/6,1,
            241/12,1,269/12,1,289/12,1,313/12,1,0,46,248/15,2,253/15,1,109/6,1,
            278/15,1,283/15,3,19,1,247/12,1,133/6,1,0,57,781/42,1,75/4,1,233/12,
            1,39/2,1,119/6,3,1333/60,1,0,65,727/60,2,0];
        I:=[428412,428441,428313,428442,428444,428422,428411,428205,428423,
            428227,428417,428403,428102,427820,428207,428369,428217,428483,
            427822,427929,428093,428113,428324,428424,428198,427724,427788,
            428218,428323,427930,428186,427828,427783,428097,428094,427728,
            428141,427897,427886,428425,428407,428040,428099,427682,427700,
            427997,428041,428019,428092,427912,427852,427909,428136,427900,
            428084,427666,427636,427634,427726,427723,427714,427949,428022,
            427615,427998,428120];
    elif hash1 eq 604 and hash2 eq 590 and hash3 eq 2496 then
        V:=[19,1,20,7,21,19,22,27,24,41,25,49];
        D:=[1,49/2,1,53/2,2,0,4,21,1,23,1,27,1,86/3,2,31,1,0,10,103/6,1,139/6,2,
            163/6,1,0,14,53/3,1,59/3,3,20,1,106/5,1,85/4,1,65/3,1,0,22,16,1,
            49/3,3,55/3,1,0,27,89/6,1,103/6,2,0];
        I:=[424642,424629,424643,424661,423900,424152,423890,424034,423224,
            424662,424128,424146,422635,424088,422615,424628,424632,424646,
            423090,424159,423841,424634,424047,424205,424271,422653,422945,
            422650,422651];
    elif hash1 eq 1544 and hash2 eq 2575 and hash3 eq 158 then
        V:=[19,1,20,5,21,11,22,19,23,27,24,33,25,37,26,51,27,71];
        D:=[1,223/6,1,0,2,28,1,34,1,0,4,155/6,1,57/2,1,181/6,1,0,7,67/3,1,73/3,
            1,311/12,1,0,10,731/30,1,263/10,1,0,12,53/3,1,0,13,181/12,1,95/6,1,
            97/6,1,511/30,1,35/2,1,1117/60,1,0,19,61/4,1,46/3,1,309/20,1,31/2,1,
            233/15,4,187/12,1,47/3,1,16,1,1003/60,1,0,31,655/42,2,937/60,1,
            157/10,1,63/4,2,75/4,1,0];
        I:=[517575,517416,517420,517559,517418,517541,517550,517368,517253,
            517262,517045,517369,517250,517473,517419,517272,517565,517264,
            517122,517563,517100,517094,517125,517245,517265,517281,517417,
            517115,517405,517065,517002,517058,517042,517121,517093,517111,
            516982];
    elif hash1 eq 1027 and hash2 eq 1208 and hash3 eq 1131 then
        V:=[23,1,24,5,25,13,26,23,27,31,28,37];
        D:=[1,45/2,1,0,2,19,2,62/3,3,21,1,0,8,103/6,4,35/2,3,115/6,2,39/2,1,0,
            18,52/3,1,53/3,3,18,4,0,26,63/4,1,95/6,3,0,30,71/5,1,0];
        I:=[672417,672314,672366,671960,672050,672288,672385,671858,672082,
            672373,672409,671996,672380,672398,672118,672131,672128,671902,
            671408,671509,672127,671432,671585,672064,672094,671615,672155,
            672215,672374,671366];
    elif hash1 eq 510 and hash2 eq 1970 and hash3 eq 2327 then
        V:=[25,1,26,5,27,9,28,15,29,25,30,41,31,53,32,61];
        D:=[1,163/6,1,0,2,24,1,0,3,109/6,1,143/6,1,0,5,55/3,2,101/5,1,61/3,1,
            62/3,1,0,10,197/12,1,167/10,1,103/6,2,35/2,1,183/10,2,221/12,1,
            125/6,1,0,19,223/15,2,33/2,1,199/12,1,17,1,227/12,1,0,25,391/30,1,
            161/12,1,27/2,1,0,28,178/15,1,0];
        I:=[239050,238973,240275,238978,238049,240281,240035,239153,240354,
            240129,238560,239029,239912,238979,238362,238379,238398,239078,
            239861,240273,239134,238636,238036,238322,239077,239063,239079,
            238137];
    elif hash1 eq 3095 and hash2 eq 3036 and hash3 eq 1652 then
        V:=[27,1,28,5,29,9,30,17,31,23,32,35];
        D:=[1,79/4,1,0,2,49/3,1,0,3,91/6,1,193/12,1,183/10,1,0,6,40/3,2,199/12,
            1,0,9,391/30,1,157/12,1,79/6,2,401/30,1,27/2,1,0,15,173/15,1,0];
        I:=[403595,403610,402984,403589,403245,403429,403510,402857,403384,
            402985,403402,403446,402968,402923,403605];
    elif hash1 eq 886 and hash2 eq 1109 and hash3 eq 204 then
        V:=[22,1,24,9,27,19,29,27,30,35];
        D:=[1,98/3,1,33,1,41,1,0,4,83/3,1,94/3,1,95/3,1,100/3,1,0,8,491/30,1,
            241/12,1,127/6,1,0,11,83/6,1,59/4,1,571/30,3,0,16,41/3,1,0];
        I:=[505952,506152,506453,506335,506387,506183,506115,506480,505897,
            506219,506266,504172,504550,504984,505010,506468];
    elif hash1 eq 651 and hash2 eq 2921 and hash3 eq 3284 then
        V:=[20,1,21,5,22,11,23,19,24,31];
        D:=[1,62/3,2,0,3,35/2,2,47/2,1,0,6,52/3,1,53/3,1,96/5,1,0,9,29/2,1,97/6,
            1,107/6,1,79/4,1,121/6,1,0,14,43/3,3,238/15,1,0];
        I:=[510608,510616,510244,510250,510257,510615,510614,509515,510634,
            510668,509568,509391,509268,510572,510599,510656,509652];
    elif hash1 eq 1309 and hash2 eq 2907 and hash3 eq 1469 then
        V:=[24,1,25,11,26,19,27,35,28,47,29,51,30,57];
        D:=[1,69/4,1,24,1,76/3,1,83/3,1,0,5,37/2,1,119/6,1,233/10,1,0,8,167/12,
            1,14,1,44/3,1,16,1,49/3,1,101/5,1,61/3,2,0,16,169/12,1,103/6,1,
            217/12,1,37/2,1,75/4,1,0,21,155/12,1,0,22,27/2,1,31/2,1,0,24,37/3,1,
            0];
        I:=[503314,502494,502544,503491,502242,502270,501882,503503,503404,
            502272,502714,503558,502980,502978,502998,503320,502402,502053,
            501051,501858,503481,502484,502997,501826];
    elif hash1 eq 510 and hash2 eq 579 and hash3 eq 3312 then
        V:=[25,1,26,9,27,15,28,21];
        D:=[1,115/6,2,39/2,1,127/6,2,0,6,53/3,2,59/3,1,0,9,95/6,2,97/6,2,0,13,
            43/3,1,44/3,1,191/12,1,0];
        I:=[241674,241717,240993,240875,241698,241006,241022,240952,239734,
            240232,240939,241749,241729,241195,239428];
    elif hash1 eq 369 and hash2 eq 2302 and hash3 eq 3296 then
        V:=[21,1,22,5,23,17,24,21,25,27,26,31,27,39,28,43];
        D:=[1,45/2,1,0,2,21,1,68/3,1,25,1,80/3,2,27,1,0,8,163/6,1,0,9,53/3,3,20,
            2,0,14,107/6,2,0,16,43/3,1,44/3,1,328/15,1,0,19,91/6,1,0,20,46/3,1,
            0];
        I:=[417885,418477,418363,418040,416347,418563,416728,416715,418412,
            418441,418444,416648,416983,416773,417131,417476,418378,415346,
            416961,415063];
    elif hash1 eq 3236 and hash2 eq 198 and hash3 eq 209 then
        V:=[23,1,24,5,25,13,28,23,29,29,30,41];
        D:=[1,199/6,1,0,2,77/3,1,83/3,2,89/3,1,0,6,133/6,1,281/12,1,143/6,2,
            145/6,1,0,11,203/12,2,52/3,1,0,14,299/20,1,15,1,181/12,2,91/6,2,
            31/2,1,0,21,91/6,1,61/4,1,0];
        I:=[248853,247993,248800,248804,249632,247988,247824,248877,249555,
            249422,247903,247911,249602,247603,247620,247595,247900,247870,
            248027,248876,247552,247559];
    elif hash1 eq 839 and hash2 eq 1738 and hash3 eq 305 then
        V:=[25,1,27,7,28,19,29,27];
        D:=[1,39/2,1,127/6,2,0,4,185/12,2,177/10,1,107/6,1,197/10,1,79/4,1,0,10,
            57/4,1,81/5,1,49/3,1,0,13,77/6,1,223/14,1,33/2,1,0];
        I:=[400052,400074,400211,398475,398726,398838,398051,398839,399675,
            397964,399529,399035,397455,396031,395911];
    elif hash1 eq 228 and hash2 eq 1155 and hash3 eq 895 then
        V:=[20,1,21,9,22,23,23,33,24,55,25,63,26,75,27,83,28,87];
        D:=[1,25,1,92/3,1,33,1,0,4,43/2,1,139/6,1,47/2,1,151/6,3,163/6,2,175/6,
            2,0,14,70/3,1,71/3,3,24,2,77/3,1,0,21,107/6,1,109/6,1,121/6,4,41/2,
            1,257/12,1,217/10,2,87/4,1,131/6,2,133/6,1,49/2,1,0,36,61/3,3,62/3,
            2,21,1,0,42,91/6,1,491/30,1,33/2,2,101/6,3,113/6,2,0,51,175/12,2,
            179/12,1,15,1,0,55,213/14,1,0,56,178/15,1,0];
        I:=[421954,421964,422121,421929,422076,422126,421936,421944,422013,
            421938,421973,421907,421914,422001,421882,421924,421968,421880,
            421894,421311,421916,421610,421417,421920,421933,422097,421876,
            421468,421568,421703,421656,421753,421775,422059,421760,421472,
            421772,421934,421634,421992,421476,421962,421751,421733,422004,
            421370,421913,422008,421479,421718,422000,422011,421404,421941,
            421182,421572];
    elif hash1 eq 1920 and hash2 eq 525 and hash3 eq 3187 then
        V:=[12,1,13,11,14,23,15,33,16,45,17,53,18,57,19,61];
        D:=[1,34,1,36,3,38,4,42,2,0,11,65/2,2,69/2,2,73/2,1,81/2,1,85/2,1,0,18,
            29,2,98/3,1,33,2,35,3,0,26,55/2,1,175/6,1,59/2,1,211/6,1,71/2,1,0,
            31,77/3,1,82/3,1,83/3,4,0,37,167/6,1,0,38,67/3,2,0,40,671/30,1,
            137/6,2,0];
        I:=[255879,255871,255877,255878,255862,255869,255876,255880,255870,
            255875,255868,255872,255864,255873,255867,255860,255848,255863,
            255866,255846,255853,255857,255844,255861,255865,255874,255847,
            255854,255849,255856,255855,255843,255845,255850,255858,255859,
            255842,255851,255852,255840,255839,255841];
    elif hash1 eq 1920 and hash2 eq 1376 and hash3 eq 2550 then
        V:=[17,1,18,5,19,11,20,23,21,31,22,39,23,45,24,57,25,65];
        D:=[1,85/2,1,0,2,98/3,1,116/3,1,0,4,151/6,1,175/6,1,131/4,1,199/6,1,
            67/2,1,0,9,77/4,1,85/4,1,77/3,1,0,12,131/6,1,157/6,1,61/2,1,0,15,
            215/12,1,129/5,1,0,17,29/2,1,89/6,1,241/12,1,203/10,1,41/2,1,0,22,
            289/20,1,29/2,1,218/15,2,0,26,349/24,1,59/4,1,121/6,1,0];
        I:=[542484,542479,542462,542478,542476,542457,542456,542450,542117,
            542447,542461,542473,542468,541836,542448,542109,542081,542297,
            542078,542125,542083,542086,542077,542290,542294,541867,542481,
            541843];
    elif hash1 eq 2155 and hash2 eq 1907 and hash3 eq 1365 then
        V:=[25,1,26,5,27,11,28,15,29,35];
        D:=[1,121/6,1,0,2,239/12,1,119/5,1,0,4,33/2,1,0,5,115/7,1,253/15,1,17,1,
            143/7,1,307/15,1,247/12,1,104/5,1,64/3,1,452/21,1,0,14,877/60,1,
            397/24,1,17,1,0];
        I:=[502234,502212,501852,502220,501943,501345,502605,501465,501177,
            501878,501771,502263,501324,502275,501271,501171];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 41 mod 47.
function cached_ids_41(P,hash1,hash2,hash3)
    if hash1 eq 182 and hash2 eq 222 and hash3 eq 3075 then
        V:=[15,1,16,5,17,11,18,23,19,39,20,63,21,85,22,103,23,125,24,141,25,159,
            26,173,27,185,28,191];
        D:=[1,57/2,1,0,2,31,1,110/3,1,0,4,151/6,1,199/6,1,235/6,1,171/4,1,91/2,
            1,0,9,83/3,1,89/3,1,95/3,2,166/5,1,36,1,118/3,2,143/3,1,0,18,143/6,
            1,157/6,1,277/10,1,119/4,1,317/10,1,127/4,1,191/6,1,193/6,1,205/6,1,
            449/12,1,413/10,1,0,29,131/5,1,79/3,1,80/3,1,194/7,1,222/7,1,478/15,
            1,161/5,2,97/3,1,508/15,1,34,1,0,40,265/12,1,269/12,1,91/4,1,49/2,1,
            149/6,1,317/12,1,839/30,1,899/30,1,0,48,92/5,1,283/15,3,96/5,1,58/3,
            1,308/15,1,251/12,1,271/12,1,23,1,187/7,1,82/3,1,0,60,169/10,1,
            511/30,1,379/20,1,571/30,1,1213/60,1,877/42,1,127/6,1,0,67,592/35,1,
            475/28,1,211/12,1,1123/60,1,77/4,1,407/21,1,235/12,1,68/3,1,0,75,
            431/30,2,29/2,1,1439/84,1,1339/70,2,809/42,1,1157/60,2,0,84,206/15,
            2,295/21,1,71/5,6,461/30,1,947/60,1,0,95,171/14,1,737/60,1,0,97,
            437/36,1,85/7,1,110/9,1,1291/105,1,371/30,1,0];
        I:=[517933,517931,517937,517930,517936,517919,517859,517896,517888,
            517827,517857,517902,517932,517820,517822,517829,517710,517947,
            517819,517940,517856,517726,517833,517851,517861,517864,517907,
            517810,517923,517837,517835,517727,517678,517825,517828,517910,
            517862,517972,517836,517863,517917,517914,517770,517831,517789,
            517845,517956,517912,517889,517909,517981,517811,517852,517929,
            517855,517916,517776,517666,517719,517986,517754,517734,517844,
            517809,517728,517791,517736,517755,517915,517960,517890,517680,
            517720,517739,517935,517941,517840,517706,517696,517732,517735,
            517698,517733,517872,517876,517756,517759,517812,517830,517860,
            517945,517965,517748,517745,517692,517690,517615,517616,517624,
            517605,517606];
    elif hash1 eq 1639 and hash2 eq 1313 and hash3 eq 489 then
        V:=[14,1,15,9,16,25,17,43,18,57,19,75,20,89,21,103,22,109,23,113];
        D:=[1,30,1,32,1,34,2,0,5,53/2,1,57/2,3,61/2,4,65/2,5,69/2,1,73/2,2,81/2,
            1,0,22,74/3,1,80/3,2,27,2,29,2,92/3,3,31,1,104/3,3,35,4,0,40,139/6,
            1,47/2,1,151/6,1,51/2,1,163/6,5,211/6,1,0,50,65/3,1,22,1,70/3,2,
            71/3,3,77/3,1,109/4,1,82/3,1,89/3,1,0,61,119/6,2,131/6,2,133/6,4,
            45/2,1,281/12,1,143/6,2,0,73,239/12,1,20,1,328/15,1,111/5,1,67/3,4,
            68/3,3,0,84,551/30,1,269/12,1,0,86,17,2,0,88,461/30,1,0];
        I:=[255717,255715,255706,255711,255593,255688,255707,255716,255568,
            255613,255720,255724,255614,255623,255705,255712,255718,255708,
            255656,255690,255633,255561,255455,255601,255704,255723,255598,
            255721,255606,255622,255654,255652,255458,255501,255626,255589,
            255596,255609,255648,255587,255638,255486,255710,255483,255510,
            255584,255647,255649,255506,255644,255479,255440,255625,255456,
            255619,255643,255508,255470,255505,255369,255582,255585,255461,
            255628,255441,255465,255605,255645,255442,255511,255368,255477,
            255453,255454,255385,255384,255411,255413,255488,255497,255395,
            255409,255462,255365,255405,255392,255586,255412];
    elif hash1 eq 1451 and hash2 eq 2684 and hash3 eq 628 then
        V:=[31,1,32,7,33,11];
        D:=[1,71/4,1,107/6,1,0,3,49/3,6,0,9,173/12,3,29/2,1,59/4,1,89/6,3,0];
        I:=[58047,57978,57919,57971,57986,58000,58009,58065,57608,57614,58041,
            58017,57621,57641,57642,58022];
    elif hash1 eq 1216 and hash2 eq 2153 and hash3 eq 3455 then
        V:=[10,1,11,5,12,9,13,13,15,19,16,27,18,33,20,43,21,47,22,53];
        D:=[1,40,1,0,2,65/2,1,0,3,39,2,0,5,127/6,1,199/6,1,0,7,119/6,1,357/10,1,
            227/6,1,0,10,62/3,1,538/15,1,0,12,56/3,1,58/3,1,308/15,1,62/3,1,0,
            16,485/21,1,0,17,1037/60,1,109/6,1,0,19,1636/105,1,967/60,1,0];
        I:=[544228,544225,544215,544224,544223,544221,544216,544150,544220,
            544129,544169,544217,544142,544146,544130,544109,544154,544128,
            544103,544102];
    elif hash1 eq 746 and hash2 eq 2804 and hash3 eq 3287 then
        V:=[25,1,26,5,27,9,28,13,29,29,30,39,31,43];
        D:=[1,329/12,1,0,2,50/3,1,0,3,147/10,1,0,4,103/7,1,223/15,1,247/15,1,
            223/12,1,131/7,1,21,1,64/3,1,0,11,401/30,1,29/2,1,299/20,1,697/42,1,
            0,15,316/21,1,0,16,911/70,1,0];
        I:=[503392,501999,503395,501955,502670,502041,501998,502017,502642,
            503393,503059,502002,501079,501964,502023,501069];
    elif hash1 eq 1498 and hash2 eq 1827 and hash3 eq 1458 then
        V:=[14,1,15,15,16,25,17,37,18,47,19,59,20,65,21,71,22,77];
        D:=[1,28,1,30,1,32,2,34,2,36,1,38,1,0,9,49/2,3,53/2,1,69/2,1,73/2,1,0,
            15,68/3,1,74/3,2,86/3,1,29,1,116/3,1,0,21,139/6,1,99/4,1,151/6,1,
            51/2,1,0,25,65/3,1,70/3,1,71/3,1,30,1,107/3,1,0,30,217/10,1,133/6,2,
            0,33,239/12,1,61/3,1,0,35,91/6,1,167/10,1,0,37,248/15,1,271/12,1,0];
        I:=[429052,429045,429012,429040,429010,429041,429054,429039,429024,
            429046,429056,428996,428986,429023,429027,428933,428985,428982,
            428926,428923,429006,428894,428907,428946,428979,428947,429002,
            428795,428772,428726,428898,428911,428792,428800,428948,428797,
            428761,428606];
    elif hash1 eq 1310 and hash2 eq 2825 and hash3 eq 2094 then
        V:=[19,1,20,7,21,19,22,33,23,39,25,43,26,53];
        D:=[1,205/6,1,217/6,2,0,4,85/3,1,30,1,151/5,1,91/3,2,431/12,1,0,10,53/2,
            1,267/10,1,559/20,1,851/30,1,341/12,1,57/2,1,0,16,23,1,368/15,1,0,
            18,631/30,1,0,19,937/60,1,157/10,1,63/4,1,177/10,3,0,25,543/35,1,
            1313/84,1,47/3,2,947/60,1,0];
        I:=[427920,428394,428395,428305,428257,428078,428332,428367,428146,
            428341,428074,427963,427959,427928,428054,428377,427981,427964,
            427969,428154,428153,427735,427750,427799,427668,427674,427675,
            427704,427691];
    elif hash1 eq 276 and hash2 eq 443 and hash3 eq 679 then
        V:=[20,1,21,5,22,9,23,13,24,21,25,31,26,37];
        D:=[1,25,1,0,2,151/6,1,0,3,59/3,1,0,4,37/2,1,119/6,2,257/12,1,0,8,
            215/12,2,18,1,55/3,3,56/3,2,0,16,18,1,217/12,1,0,18,179/12,1,0];
        I:=[253305,253414,253062,253048,253065,253073,252697,253121,253142,
            253111,252599,253074,253079,252983,253041,252699,252814,252619];
    elif hash1 eq 2485 and hash2 eq 1178 and hash3 eq 2025 then
        V:=[27,1,28,9,29,17,31,27,32,35];
        D:=[1,107/6,1,109/6,1,197/10,1,0,4,49/3,3,55/3,2,322/15,1,0,10,431/30,1,
            197/12,1,167/10,1,183/10,1,0,14,391/30,1,27/2,1,625/42,1,0,17,
            229/20,1,173/15,1,0];
        I:=[407608,407345,407958,406943,407263,407623,406752,407274,406067,
            407226,407339,407609,406047,407671,407741,404919,407438,406815];
    elif hash1 eq 2344 and hash2 eq 2797 and hash3 eq 1265 then
        V:=[26,1,27,5,28,15,29,21,30,29];
        D:=[1,21,2,0,3,115/6,5,39/2,3,83/4,1,127/6,1,0,13,47/3,1,53/3,1,0,15,
            29/2,1,177/10,1,71/4,2,0,19,167/12,1,43/3,4,44/3,1,0];
        I:=[667062,667443,665622,665626,666121,667009,667441,667021,667130,
            667418,665695,666141,667152,666998,665632,665502,665225,666262,
            666159,665188,665223,665588,667314,665511];
    elif hash1 eq 3237 and hash2 eq 793 and hash3 eq 353 then
        V:=[27,1,28,5,29,13,30,29,31,37,32,45];
        D:=[1,151/6,1,0,2,53/3,2,59/3,1,85/4,1,0,6,95/6,2,71/4,1,107/6,3,109/6,
            2,79/4,1,119/6,1,121/6,1,0,17,167/12,1,57/4,1,44/3,3,0,22,173/12,1,
            29/2,3,491/30,1,0,27,13,1,175/12,1,0];
        I:=[234678,233133,234236,230569,232998,234348,234677,233620,232066,
            233033,233557,234240,234588,231983,233022,234586,233378,232172,
            232739,233564,233873,233496,231508,231794,233279,229287,231337,
            229258];
    elif hash1 eq 2673 and hash2 eq 2757 and hash3 eq 2032 then
        V:=[20,1,21,11,22,21,23,27,24,37,26,43,27,55];
        D:=[1,418/15,1,85/3,1,448/15,1,103/3,1,0,5,791/30,1,267/10,1,223/8,1,
            1031/30,1,0,9,373/15,1,554/21,2,0,12,631/30,1,589/24,1,199/8,1,
            751/30,1,0,16,150/7,1,259/12,1,0,18,487/35,1,295/21,1,211/15,1,
            313/20,1,947/60,1,0,23,1253/90,1,503/36,1,1187/84,1,0];
        I:=[517767,517766,517642,517764,517769,517687,517643,517685,517765,
            517684,517703,517768,517636,517637,517641,517686,517679,517689,
            517622,517683,517702,517699,517607,517611,517609];
    elif hash1 eq 746 and hash2 eq 3336 and hash3 eq 3259 then
        V:=[26,1,28,7,29,11,30,21,31,33];
        D:=[1,239/12,1,20,1,0,3,307/15,1,0,4,59/4,1,233/12,1,62/3,1,629/30,1,0,
            8,310/21,1,317/21,1,907/60,1,91/6,1,94/5,1,0,13,27/2,1,409/30,1,
            46/3,1,145/8,1,0];
        I:=[501919,501928,501930,501535,501421,501515,501772,501556,501929,
            501624,501566,501193,501506,501486,503730,501033];
    elif hash1 eq 3190 and hash2 eq 2653 and hash3 eq 2808 then
        V:=[21,1,22,5,23,9,25,13,26,19,27,29,28,41];
        D:=[1,163/6,1,0,2,28,1,0,3,41/2,1,0,4,217/12,1,221/12,1,0,6,15,1,17,1,
            369/20,1,277/15,1,0,10,451/30,3,91/6,1,185/12,2,31/2,2,95/6,1,0,19,
            187/12,1,47/3,1,0];
        I:=[422259,421984,421987,421337,421565,421769,422241,421759,421180,
            421471,421558,421615,422302,421271,421972,421296,421971,422150,
            421124,421272];
    elif hash1 eq 2438 and hash2 eq 2113 and hash3 eq 2103 then
        V:=[23,1,24,7,25,11,26,17];
        D:=[1,45/2,1,49/2,4,0,6,21,2,0,8,115/6,1,39/2,1,0,10,52/3,1,53/3,3,18,2,
            0];
        I:=[61676,61674,61679,61686,61689,61594,61716,61569,61563,61566,61574,
            61680,61684,61589,61673];
    elif hash1 eq 2438 and hash2 eq 264 and hash3 eq 648 then
        V:=[25,1,26,5,27,17,28,25,29,41,30,49,31,59];
        D:=[1,163/6,1,0,2,70/3,1,71/3,2,24,1,77/3,1,83/3,1,0,8,109/6,1,133/6,1,
            143/6,1,0,11,49/3,1,50/3,1,73/4,1,55/3,1,56/3,1,61/3,1,62/3,1,0,18,
            197/12,1,33/2,2,101/6,1,0,22,175/12,1,44/3,1,15,2,247/15,1,0,27,
            401/30,1,161/12,1,0];
        I:=[241426,240724,241289,241459,241162,241579,241264,240550,239053,
            241463,241350,241342,238639,239181,240800,241487,241588,239164,
            240462,240571,238924,240415,241117,239008,241563,238321,240523,
            239094];
    elif hash1 eq 1310 and hash2 eq 90 and hash3 eq 2450 then
        V:=[23,1,24,11,25,23,26,27,28,31,29,35,30,41,31,47];
        D:=[1,175/6,1,59/2,1,187/6,1,223/6,2,0,6,126/5,1,146/5,1,88/3,1,89/3,1,
            95/3,1,0,11,181/6,1,0,12,61/3,1,0,13,253/15,1,0,14,29/2,1,181/12,1,
            0,16,79/6,1,89/6,1,0,18,56/5,1,0];
        I:=[410645,411491,411883,410564,410917,410215,410212,412054,411879,
            410895,412223,410855,409581,409261,410943,410919,409240,410832];
    elif hash1 eq 1216 and hash2 eq 2590 and hash3 eq 552 then
        V:=[20,1,21,9,23,15,24,21,25,25,26,39,27,47];
        D:=[1,88/3,1,89/3,1,157/4,1,0,4,317/10,1,191/6,1,0,6,20,1,611/30,1,0,8,
            17,1,0,9,83/6,1,161/10,1,973/60,1,169/10,1,511/30,2,1213/60,1,0,16,
            203/15,2,242/15,1,163/10,1,0,20,797/60,1,1367/84,1,0];
        I:=[515444,515549,514314,514840,515599,514256,514372,515598,514403,
            514892,514893,514471,514352,514441,513997,514400,514872,514254,
            514330,514901,514016];
    elif hash1 eq 1451 and hash2 eq 2814 and hash3 eq 172 then
        V:=[21,1,22,7,23,15,24,25,25,33,26,39,27,45];
        D:=[1,103/6,1,115/6,1,0,3,59/3,2,126/5,1,146/5,1,0,7,95/6,2,97/6,1,
            107/6,1,95/4,1,0,12,81/5,2,49/3,1,268/15,1,0,16,89/6,1,491/30,1,0,
            18,44/3,2,50/3,1,0,21,625/42,1,0];
        I:=[510269,510324,510321,510467,509665,509083,510271,510393,510320,
            510390,509664,510028,510049,510400,509937,510652,509078,510385,
            510606,509699,508874];
    elif hash1 eq 276 and hash2 eq 3101 and hash3 eq 1003 then
        V:=[25,1,26,11,27,15,28,25,29,31,30,37,31,41];
        D:=[1,103/6,2,115/6,1,83/4,1,127/6,1,0,6,47/3,1,0,7,185/12,1,157/10,1,
            97/6,3,33/2,1,0,13,14,1,232/15,1,0,15,51/4,1,199/14,1,0,17,89/7,1,0,
            18,341/30,1,0];
        I:=[400364,400392,399210,398532,398987,398815,400402,398765,395805,
            399124,399387,400553,399874,397162,400477,396386,395880,398743];
    elif hash1 eq 276 and hash2 eq 1829 and hash3 eq 1066 then
        V:=[12,1,13,13,14,19,15,25,16,29,17,33,18,39,20,43,21,47];
        D:=[1,30,1,36,2,42,2,48,1,54,1,0,8,49/2,1,61/2,1,0,10,62/3,1,110/3,1,0,
            12,75/2,2,0,14,101/4,1,0,15,121/6,1,167/6,1,0,17,55/3,1,0,18,253/15,
            1,0,19,107/6,1,0];
        I:=[544026,544025,544028,544024,544034,544033,544032,544029,544006,
            544022,544030,543992,544011,544007,543955,543952,544010,543971,
            543954];
    elif hash1 eq 464 and hash2 eq 631 and hash3 eq 1905 then
        V:=[28,1,30,13,31,19,33,29,34,41];
        D:=[1,64/3,1,70/3,1,71/3,1,77/3,1,26,1,0,6,91/5,1,55/3,1,0,8,29/2,1,
            89/6,1,167/10,2,101/6,2,0,14,51/4,1,391/30,2,157/12,1,79/6,4,27/2,1,
            0,23,56/5,1,0];
        I:=[230894,231712,230752,234606,234131,232741,232987,231909,234937,
            230993,232279,231383,233823,232141,232047,232470,233817,230891,
            234497,234498,234601,232469,232051];
    elif hash1 eq 3237 and hash2 eq 1003 and hash3 eq 3172 then
        V:=[17,1,18,9,19,23,20,39,21,51,22,63,23,79,24,83,25,87];
        D:=[1,45/2,1,49/2,2,69/2,1,0,5,21,1,23,3,74/3,2,29,2,31,2,98/3,1,0,16,
            39/2,1,83/4,1,127/6,2,43/2,1,139/6,3,51/2,2,55/2,1,0,27,59/3,3,85/4,
            1,64/3,1,65/3,2,71/3,3,0,37,37/2,4,197/10,1,119/6,1,121/6,2,257/12,
            1,0,46,50/3,1,215/12,1,55/3,1,56/3,2,101/5,1,62/3,1,67/3,1,0,54,
            101/6,1,0,55,47/3,1,0,56,185/12,1,95/6,1,0];
        I:=[515618,515605,515617,515615,515619,515406,515423,515612,515433,
            515440,515189,515430,514387,515124,515526,515419,515420,515409,
            515606,515462,515443,515488,515601,515103,515138,515459,515408,
            515461,515539,515453,515522,514460,515524,515068,515171,515219,
            514354,515065,515196,515235,514385,514242,514377,515437,515213,
            514434,515401,514423,515194,515245,514243,515242,514225,515438,
            514422,514008,514009];
    elif hash1 eq 981 and hash2 eq 124 and hash3 eq 2856 then
        V:=[24,1,25,11,26,21,27,25,28,31,29,37,30,41,31,47];
        D:=[1,117/4,1,88/3,2,94/3,1,95/3,1,0,6,107/6,1,131/6,1,95/4,1,145/6,1,0,
            10,56/3,1,0,11,33/2,1,37/2,1,0,13,248/15,1,227/12,1,0,15,103/6,1,0,
            16,66/5,1,41/3,1,0,18,799/70,1,697/60,1,0];
        I:=[410803,409833,411334,410665,410186,410732,410615,409865,410198,
            410516,409803,410730,409574,411294,409436,410694,410939,409404,
            409007];
    elif hash1 eq 1592 and hash2 eq 3030 and hash3 eq 2945 then
        V:=[18,1,19,13,20,19,21,31,22,35,24,51,25,55];
        D:=[1,25,1,92/3,1,98/3,1,33,2,37,1,0,7,151/6,1,163/6,1,0,9,64/3,1,101/4,
            2,77/3,2,136/5,1,83/3,3,0,18,281/12,1,0,19,55/3,1,239/12,1,101/5,1,
            61/3,1,22,1,67/3,2,311/12,1,0,27,223/15,1,0,28,451/30,1,181/12,1,
            91/6,1,1213/60,1,0];
        I:=[427274,427279,427280,427234,427255,427282,426948,427269,427286,
            427231,427238,427236,427278,426922,426938,426960,427262,427261,
            426946,426898,426953,427289,426914,426888,426962,426559,426889,
            426428,427239,427263,426429];
    elif hash1 eq 1357 and hash2 eq 422 and hash3 eq 1979 then
        V:=[24,1,25,9,26,17,27,25,28,35,29,47,30,53,31,65];
        D:=[1,101/4,1,77/3,1,117/4,1,0,4,107/6,1,257/12,1,237/10,1,0,7,50/3,2,
            268/15,1,131/5,1,0,11,491/30,2,37/2,1,187/10,1,113/6,1,0,16,223/15,
            1,61/4,1,199/12,1,369/20,1,106/5,1,0,21,213/14,1,461/30,1,0,23,
            178/15,1,79/6,1,66/5,1,41/3,1,43/3,1,0,28,347/30,1,697/60,1,71/6,1,
            0];
        I:=[502946,503388,502885,503298,501951,503173,502889,503105,503183,
            502910,503171,503370,503299,503041,502834,502804,501107,502043,
            502504,502874,501786,502036,503290,502045,503079,503233,503128,
            503369,503265,501830];
    elif hash1 eq 1733 and hash2 eq 3188 and hash3 eq 988 then
        V:=[26,1,27,7,29,13,30,21,31,35,32,51];
        D:=[1,65/3,1,101/4,1,0,3,197/10,1,79/4,1,0,5,491/30,1,103/6,1,35/2,1,0,
            8,13,1,40/3,2,41/3,1,15,2,329/20,1,86/5,1,0,16,38/3,1,51/4,1,389/30,
            1,391/30,1,157/12,3,401/30,2,27/2,2,0,27,667/60,1,0];
        I:=[407576,407352,407491,407148,407517,408791,408795,407240,406792,
            406793,408670,408569,408581,406600,405114,408379,408530,405832,
            407739,405799,406036,407693,407238,407860,408606,408774,406771];
    elif hash1 eq 1310 and hash2 eq 567 and hash3 eq 1517 then
        V:=[24,1,25,13,26,19,27,29,28,35,29,43,30,47];
        D:=[1,23,1,74/3,2,25,1,29,4,92/3,1,0,10,115/6,1,39/2,2,0,13,18,2,58/3,1,
            20,1,106/5,1,0,18,157/10,2,107/6,1,0,21,49/3,1,55/3,1,56/3,1,0,24,
            33/2,2,0,26,249/20,1,188/15,1,0];
        I:=[400498,400167,400500,400386,399616,400391,400485,400547,400141,
            398701,399078,400062,400479,400536,400491,400199,397802,399812,
            399845,398233,399017,398995,398676,395912,396984,398872,398719];
    elif hash1 eq 2109 and hash2 eq 2032 and hash3 eq 1320 then
        V:=[26,1,27,5,28,9,29,15,30,23,31,29,32,35,33,41];
        D:=[1,93/4,1,0,2,107/6,1,0,3,268/15,1,61/3,1,0,5,431/30,1,103/6,1,
            551/30,1,0,8,289/20,1,50/3,1,0,10,391/30,1,27/2,2,0,13,229/20,1,
            173/15,1,0,15,499/42,1,0];
        I:=[407267,407157,407318,405340,407616,407617,405117,406057,405319,
            407508,407558,407641,407710,407755,403983];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 42 mod 47.
function cached_ids_42(P,hash1,hash2,hash3)
    if hash1 eq 1781 and hash2 eq 3248 and hash3 eq 3233 then
        V:=[22,1,23,9,24,25,25,33,26,41,27,47,28,51,30,55];
        D:=[1,27,1,29,2,33,1,0,5,47/2,1,151/6,4,107/4,1,163/6,3,55/2,1,175/6,1,
            123/4,1,0,17,85/4,1,116/5,2,71/3,1,0,21,233/12,1,119/6,1,121/6,1,0,
            24,268/15,1,55/3,1,0,26,75/4,1,0,27,15,1,0,28,173/15,1,0];
        I:=[251203,250912,251134,251207,251035,251119,251120,251155,251200,
            250638,250132,250158,250273,250900,250208,250403,250383,250642,
            250644,251144,250639,250530,251018,250641,250231,249887,251201,
            250391];
    elif hash1 eq 1687 and hash2 eq 2744 and hash3 eq 2998 then
        V:=[16,1,17,7,18,17,19,25,20,33,21,45,22,53,23,59,24,67,25,73,26,77,27,
            85];
        D:=[1,36,1,38,1,0,3,77/2,1,85/2,1,89/2,1,97/2,1,0,7,33,1,39,1,122/3,1,0,
            10,187/6,1,71/2,1,247/6,1,0,13,77/3,1,30,1,94/3,2,34,1,107/3,3,0,21,
            155/6,1,179/6,1,317/10,1,0,24,287/12,1,85/3,1,0,26,227/10,1,91/4,1,
            24,1,0,29,369/20,2,308/15,1,0,32,35/2,1,0,33,46/3,1,86/5,1,52/3,1,0,
            36,937/60,1,0];
        I:=[516600,516414,516616,516599,516387,516426,516595,516400,516585,
            516389,516362,516628,516626,516358,516356,516421,515979,516375,
            516376,516429,516221,516393,515927,515902,515948,515930,515986,
            515851,515929,516242,516352,515872,516372,515866,515849,515751];
    elif hash1 eq 371 and hash2 eq 95 and hash3 eq 116 then
        V:=[18,1,19,5,20,13,21,29,22,43,23,61,24,73,25,79,26,93,27,99,29,109];
        D:=[1,25,1,0,2,163/6,1,63/2,1,199/6,1,0,5,70/3,1,77/3,1,117/4,1,88/3,1,
            94/3,1,32,1,149/4,1,0,12,237/10,1,257/10,1,157/6,1,277/10,1,193/6,3,
            203/6,2,0,21,263/12,1,89/4,1,287/12,1,24,1,121/5,1,388/15,1,311/12,
            1,79/3,1,0,29,551/30,1,113/6,1,269/12,1,49/2,1,99/4,1,0,34,302/15,1,
            62/3,1,0,36,169/10,1,205/12,1,371/20,1,793/42,1,229/12,1,133/6,1,0,
            42,71/5,1,77/4,1,0,44,407/30,1,583/42,1,421/30,4,85/6,2,0,52,171/14,
            1,0];
        I:=[514633,514589,514808,514587,514622,514707,514946,514989,514949,
            514952,514593,514696,514852,514955,514156,514655,514980,514985,
            514684,514954,514951,514738,514652,514333,514986,514691,514291,
            514957,514488,514950,514739,514334,514335,514960,514277,514142,
            514860,514746,514220,514169,514074,514813,514160,514491,514099,
            514093,514146,514789,514953,514932,514959,513953];
    elif hash1 eq 371 and hash2 eq 825 and hash3 eq 2417 then
        V:=[17,1,18,9,19,21,20,35,21,51,22,63,23,79,24,89,25,97,27,103];
        D:=[1,61/2,1,65/2,1,69/2,1,0,4,86/3,1,29,1,98/3,1,33,3,35,1,0,11,163/6,
            4,55/2,2,175/6,1,187/6,2,199/6,2,67/2,1,0,23,65/3,3,70/3,1,76/3,4,
            77/3,1,26,1,109/4,1,89/3,1,0,35,257/12,1,143/6,1,145/6,2,305/12,1,
            111/4,1,0,41,91/5,1,298/15,1,101/5,1,263/12,1,67/3,2,51/2,1,311/12,
            1,0,49,359/20,1,551/30,1,41/2,1,125/6,1,0,53,46/3,1,248/15,1,253/15,
            1,0,56,181/12,1,91/6,1,0,58,361/30,2,0];
        I:=[427305,427248,427031,427292,427314,427256,427264,427276,427306,
            427285,426952,427075,427273,427310,427271,427284,427316,427000,
            427294,427180,427300,427241,426909,426926,427324,427320,427243,
            427312,427318,427326,427147,427287,426687,426881,427325,427259,
            426955,427275,426892,426918,426626,426891,426882,426919,426964,
            426967,426488,426541,426652,426878,426943,426961,427272,426913,
            426920,427233,427245,426684,426942];
    elif hash1 eq 2204 and hash2 eq 2310 and hash3 eq 1400 then
        V:=[21,1,22,7,23,15,25,27,26,37];
        D:=[1,49/2,1,53/2,1,0,3,21,1,68/3,1,27,1,0,6,115/6,1,39/2,4,83/4,1,
            127/6,3,151/6,1,0,16,95/6,2,97/6,3,177/10,1,109/6,1,0,23,44/3,1,0];
        I:=[249778,249781,249784,249111,248601,249780,249779,249785,249790,
            249791,249101,248265,249106,249114,247672,249299,249782,249148,
            249296,249357,247682,247649,248541];
    elif hash1 eq 606 and hash2 eq 978 and hash3 eq 2455 then
        V:=[21,1,22,13,23,27,24,53,25,71,26,83,27,99];
        D:=[1,43/2,1,139/6,1,151/6,1,115/4,1,175/6,1,0,6,59/3,2,20,1,64/3,1,
            65/3,2,93/4,2,71/3,2,0,16,107/6,2,109/6,2,37/2,1,197/10,1,119/6,2,
            121/6,2,41/2,2,217/10,2,87/4,1,131/6,1,145/6,1,253/10,1,0,34,50/3,1,
            17,1,215/12,1,18,1,298/15,1,101/5,1,81/4,1,62/3,1,0,42,89/6,1,
            491/30,1,33/2,1,101/6,2,103/6,2,0,49,175/12,1,44/3,1,179/12,1,15,2,
            46/3,1,329/20,1,253/15,1,0,57,161/12,1,27/2,1,31/2,1,0];
        I:=[422069,422105,422040,421957,421865,421989,422115,422025,422095,
            421849,422125,422027,422114,421443,421612,422057,422091,421582,
            422084,421525,421633,422061,422071,421908,422096,421861,421906,
            421592,421613,421986,421979,421553,421264,422014,421904,421598,
            422072,421542,421606,422129,421538,422050,421546,421540,421389,
            421866,421348,421513,422051,422029,421358,421624,422090,421900,
            421187,421168,421492,421202,421151];
    elif hash1 eq 2251 and hash2 eq 443 and hash3 eq 1045 then
        V:=[24,1,26,5,27,17,28,23,29,31,30,43];
        D:=[1,144/5,1,0,2,167/12,1,14,1,191/12,1,16,1,152/7,1,0,7,183/10,1,
            551/30,2,0,10,155/12,1,175/12,1,338/21,1,0,13,51/4,1,391/30,1,59/4,
            1,451/30,2,461/30,1,0,19,313/24,1,196/15,2,0];
        I:=[502707,503646,503645,503666,503492,501559,502697,503087,503267,
            503478,503661,501985,503490,503396,502021,502994,503017,503109,
            501387,501089,501382];
    elif hash1 eq 1123 and hash2 eq 2143 and hash3 eq 1262 then
        V:=[26,1,27,7,28,11,29,19,30,27,31,35,32,39];
        D:=[1,17,1,80/3,1,0,3,115/6,1,0,4,47/3,1,53/3,1,59/3,1,0,7,55/4,1,97/6,
            2,177/10,1,0,11,71/5,1,44/3,1,191/12,1,0,14,77/6,3,0,17,56/5,1,0];
        I:=[234444,233400,233069,233437,234448,232336,233722,232687,234479,
            231911,233413,233443,229029,232245,232247,232340,233156];
    elif hash1 eq 1640 and hash2 eq 1581 and hash3 eq 1477 then
        V:=[21,1,22,7,23,19,24,29,25,39,26,53,28,59,29,63];
        D:=[1,49/2,1,57/2,2,0,4,21,2,23,2,74/3,1,25,2,80/3,1,0,12,35/2,1,115/6,
            2,39/2,1,43/2,1,0,17,47/3,1,86/5,1,52/3,2,58/3,1,0,22,185/12,1,97/6,
            2,107/6,1,197/10,1,79/4,1,121/6,1,0,29,238/15,1,268/15,1,0,31,
            193/15,1,0,32,117/10,1,0];
        I:=[412110,412114,412295,410953,412201,410253,412293,412003,410675,
            411048,412224,411940,412053,412124,410370,410677,411748,410676,
            411942,412272,411114,412205,410562,411867,409679,410428,410050,
            411084,409667,410667,411361,411496];
    elif hash1 eq 183 and hash2 eq 1473 and hash3 eq 2317 then
        V:=[28,1,29,11,30,17,31,21,32,25,33,29,36,37];
        D:=[1,26,1,88/3,1,94/3,1,119/3,1,0,5,169/6,1,317/10,1,0,7,20,1,0,8,
            103/6,1,0,9,253/15,1,0,10,461/30,1,31/2,1,103/6,1,0,13,151/15,1,
            607/60,2,0];
        I:=[381883,383329,381858,383465,381895,383330,381295,382349,381841,
            381809,381528,381300,383280,382381,383289];
    elif hash1 eq 1170 and hash2 eq 1792 and hash3 eq 3078 then
        V:=[28,1,29,9,30,13,31,25,32,37];
        D:=[1,59/3,1,65/3,2,93/4,1,0,5,95/6,1,0,6,44/3,1,191/12,1,50/3,3,215/12,
            1,56/3,1,0,13,431/30,1,29/2,1,147/10,1,89/6,4,167/10,1,0,21,40/3,1,
            0];
        I:=[226967,226664,226895,226438,226781,227010,227381,226634,226663,
            226774,226198,226474,226821,226446,226894,226623,226942,227016,
            227122,227124,226948];
    elif hash1 eq 3144 and hash2 eq 1753 and hash3 eq 2328 then
        V:=[18,1,19,5,20,15,21,25,22,49,23,65,24,77,25,89,26,103];
        D:=[1,21,1,0,2,127/6,1,59/2,2,71/2,1,83/2,1,0,7,52/3,1,53/3,2,58/3,1,
            125/4,1,0,12,177/10,2,119/6,1,45/2,1,237/10,1,143/6,1,155/6,1,
            273/10,1,57/2,1,297/10,1,179/6,1,135/4,1,0,24,50/3,1,268/15,1,18,1,
            91/5,1,298/15,1,239/12,2,74/3,1,0,32,33/2,1,359/20,2,255/14,1,113/6,
            1,127/6,1,0,38,218/15,1,223/15,1,344/21,1,121/6,1,143/7,1,0,43,
            349/24,1,31/2,1,67/4,1,709/42,1,209/12,1,619/30,1,0,49,269/20,1,
            76/5,1,0];
        I:=[542394,542453,542053,542187,542220,542212,542390,542386,542392,
            542480,542420,542247,542250,542472,542054,542045,542050,542063,
            542062,542208,542190,542200,542056,542037,542040,542038,541884,
            542348,542296,542331,542194,542477,542285,542330,541907,541877,
            542174,542231,541859,541878,541858,541906,541861,542218,542279,
            541897,541999,541857,541896,541898];
    elif hash1 eq 2862 and hash2 eq 2460 and hash3 eq 2951 then
        V:=[16,1,17,5,18,13,19,23,20,31,23,35,24,41,25,51,26,57];
        D:=[1,143/3,1,0,2,151/4,1,473/12,1,397/10,1,0,5,171/5,2,568/15,1,598/15,
            1,479/12,1,0,10,851/30,1,389/12,1,40,1,0,13,641/20,1,0,14,1153/60,1,
            1177/60,1,0,16,1564/105,1,153/10,1,236/15,1,284/15,1,0,20,2153/140,
            1,569/30,1,0,22,119/10,1,0];
        I:=[519462,519454,519495,519470,519453,519473,519497,519395,519399,
            519450,519451,519359,519343,519356,519500,519354,519458,519501,
            519327,519360,519321,519441];
    elif hash1 eq 1358 and hash2 eq 1507 and hash3 eq 2491 then
        V:=[23,1,24,11,25,19,26,27,27,31,29,37];
        D:=[1,167/6,1,179/6,2,181/6,2,191/6,1,0,7,24,1,388/15,1,311/12,1,0,10,
            671/30,1,269/12,1,137/6,1,0,13,343/15,2,0,15,56/3,1,19,1,0,17,47/3,
            1,0];
        I:=[424676,423737,424829,422703,424831,424830,424675,422700,422694,
            422702,422693,424392,423731,423741,423695,422699,422578];
    elif hash1 eq 512 and hash2 eq 2956 and hash3 eq 2423 then
        V:=[14,1,16,5,17,9,18,13,19,17,20,21,21,27,23,31,24,37,25,45,26,57];
        D:=[1,140/3,1,0,2,236/5,1,0,3,337/10,1,0,4,103/3,1,0,5,1199/30,1,0,6,
            157/6,1,403/15,1,0,8,531/20,1,0,9,563/30,1,1177/60,1,0,11,236/15,1,
            95/6,1,16,2,0,15,3317/210,1,1331/84,1,238/15,1,113/6,1,569/30,1,0,
            20,119/10,1,0];
        I:=[519258,519035,519005,519282,518888,519028,519087,518871,518883,
            518986,519051,519168,519244,519295,518889,518878,518880,518845,
            518841,518989];
    elif hash1 eq 2204 and hash2 eq 439 and hash3 eq 1228 then
        V:=[23,1,24,7,25,21,26,31,27,49,29,55];
        D:=[1,151/6,1,163/6,1,0,3,59/3,1,64/3,1,65/3,1,71/3,3,24,1,77/3,1,0,11,
            109/6,1,79/4,1,121/6,2,131/6,1,0,16,49/3,1,268/15,1,91/5,1,55/3,4,
            56/3,1,298/15,1,239/12,1,61/3,2,0,28,101/6,2,221/12,2,0,32,401/30,1,
            0];
        I:=[248383,248659,248315,248655,248301,248377,248384,248428,248415,
            248478,248627,248400,248147,248668,248487,248295,248665,248306,
            248292,248416,248418,248639,248581,248431,248191,248490,248515,
            248133,248156,248242,248337,248300];
    elif hash1 eq 3285 and hash2 eq 2923 and hash3 eq 653 then
        V:=[21,1,22,7,23,19,24,29,25,35,26,43,27,57,28,65];
        D:=[1,139/6,1,123/4,1,0,3,20,1,24,1,77/3,1,125/4,1,164/5,1,0,8,257/12,1,
            145/6,1,167/6,1,293/10,1,0,12,298/15,1,67/3,1,0,14,91/6,1,391/20,1,
            125/6,1,0,17,15,1,199/12,1,253/15,1,223/12,1,227/12,1,22,1,0,23,
            461/28,1,781/42,2,1117/60,1,0,27,203/15,1,0];
        I:=[509443,510059,509444,509452,509451,509128,510094,509968,508803,
            509121,509527,509851,509392,509545,510110,510228,510214,509919,
            509409,509562,508901,508798,508765,508726,508727,508729,509459];
    elif hash1 eq 1123 and hash2 eq 834 and hash3 eq 94 then
        V:=[19,1,20,7,21,15,22,23,23,37,24,47,25,53,26,69,27,75,28,85];
        D:=[1,77/2,1,81/2,1,0,3,31,2,104/3,1,35,1,0,7,175/6,1,187/6,1,199/6,1,0,
            10,70/3,1,76/3,1,109/4,1,82/3,1,89/3,1,125/4,1,0,16,143/6,1,305/12,
            1,155/6,2,169/6,1,0,21,61/3,1,73/3,1,0,23,16,1,197/12,1,399/20,1,
            599/30,1,20,1,245/12,1,125/6,1,0,30,17,1,20,1,0,32,25/2,1,251/20,1,
            38/3,1,241/12,1,0,36,46/3,1,0];
        I:=[510513,510512,510350,510370,510343,510356,510375,510349,510367,
            510377,510436,510371,509457,510335,510369,510339,509985,510441,
            510444,510189,509792,510058,510460,510374,509151,509064,509154,
            509112,510011,510342,508618,509291,509072,509164,508377,508512];
    elif hash1 eq 2674 and hash2 eq 1607 and hash3 eq 835 then
        V:=[24,1,25,7,26,19,27,29,28,35,29,39,30,43];
        D:=[1,68/3,1,25,2,0,4,39/2,1,83/4,1,127/6,1,43/2,1,139/6,1,0,9,96/5,1,
            59/3,1,20,2,106/5,1,0,14,95/6,3,107/6,1,0,18,238/15,1,0,19,14,2,0,
            21,193/15,1,0];
        I:=[672769,672715,672858,672745,672508,672860,672895,672624,672573,
            672864,672564,672622,672456,672466,672518,672765,672650,672446,
            672512,672785,672489];
    elif hash1 eq 1546 and hash2 eq 2424 and hash3 eq 2428 then
        V:=[16,1,17,7,18,15,19,23,20,33,21,43,22,51,23,59,24,65,25,71];
        D:=[1,30,1,32,1,0,3,65/2,2,69/2,1,77/2,2,0,8,29,2,33,1,37,1,0,12,47/2,1,
            175/6,1,187/6,2,211/6,1,0,17,93/4,1,71/3,1,82/3,1,106/3,1,0,21,
            257/12,1,131/6,3,329/12,1,0,26,215/12,2,20,1,73/3,1,0,30,18,1,
            671/30,1,0,32,242/15,1,45/2,1,0,34,299/20,1,0];
        I:=[425329,425356,425327,425358,425309,425310,425322,425312,425326,
            425355,425188,425314,425179,425096,425174,425154,425169,425193,
            425157,424996,425182,425048,425158,425160,425008,425004,425181,
            424973,424902,425168,424905,425002,424871,425013];
    elif hash1 eq 1546 and hash2 eq 2604 and hash3 eq 1668 then
        V:=[18,1,19,17,20,35,21,53,22,75,23,95,24,115,25,139,26,151,27,155];
        D:=[1,27,1,86/3,1,29,1,31,2,33,2,104/3,1,110/3,1,0,10,151/6,1,163/6,3,
            55/2,1,175/6,3,59/2,1,187/6,2,63/2,2,199/6,1,0,24,22,1,70/3,1,101/4,
            1,76/3,3,77/3,3,26,1,82/3,2,83/3,4,0,40,41/2,1,257/12,1,133/6,2,
            237/10,1,143/6,2,145/6,2,305/12,1,257/10,1,155/6,2,157/6,1,0,54,
            73/4,1,55/3,2,61/3,3,62/3,3,43/2,1,328/15,1,111/5,1,89/4,1,67/3,3,0,
            70,551/30,1,221/12,2,75/4,1,113/6,7,115/6,1,207/10,1,125/6,2,269/12,
            1,91/4,1,0,87,46/3,1,247/15,1,199/12,3,50/3,2,17,1,223/12,2,283/15,
            6,227/12,1,19,2,77/4,1,441/20,1,0,108,451/30,1,181/12,2,461/30,1,
            185/12,1,31/2,1,0,114,323/21,1,0,115,361/30,1,0];
        I:=[426237,426356,426258,426235,426320,426201,426319,426284,426341,
            426359,426079,426297,426347,426250,426016,426307,426354,426301,
            426290,426293,425984,426257,426207,426374,426022,426153,426216,
            426283,426351,425978,426071,426366,426206,426024,426060,425932,
            426029,426236,426335,426051,426123,425751,426097,425721,426030,
            426078,425755,426077,426015,425929,425759,425922,426059,425999,
            426028,426376,426000,426081,426279,426233,426277,426371,425998,
            425921,425799,425800,425673,425695,426103,425785,426135,426161,
            426005,425677,425985,426064,426203,426219,426229,426360,426241,
            425719,425736,425974,425678,425685,425862,425727,425771,426039,
            426052,425854,426055,426010,425635,425691,425682,425729,425732,
            425744,425849,425863,425672,425676,425843,425665,425662,425786,
            425789,425834,425835,425833,425796,425658,425869];
    elif hash1 eq 2063 and hash2 eq 2748 and hash3 eq 837 then
        V:=[17,1,18,5,19,17,20,33,21,53,22,65,23,77,24,89,25,93,26,99,27,103];
        D:=[1,69/2,1,0,2,29,1,92/3,1,31,1,98/3,1,33,1,0,7,51/2,1,163/6,1,55/2,2,
            59/2,1,187/6,1,63/2,1,67/2,1,0,15,71/3,1,24,1,101/4,1,76/3,1,77/3,6,
            26,1,83/3,2,117/4,1,95/3,1,0,30,87/4,1,131/6,4,133/6,3,143/6,2,
            145/6,1,0,41,20,1,61/3,2,263/12,1,89/4,1,67/3,5,0,51,221/12,4,37/2,
            2,75/4,1,113/6,3,115/6,1,0,62,45/2,1,0,63,181/12,1,91/6,1,0,65,
            233/15,1,0,66,73/6,1,0];
        I:=[254319,254338,254272,254270,254301,254264,254281,254283,254240,
            254277,254282,254322,254268,254333,254324,254328,254195,254187,
            254210,254263,254273,254288,254300,254327,254329,254172,254173,
            254122,254151,254196,254259,254286,254287,254323,254120,254182,
            254318,254150,254222,254257,254233,254207,254211,254115,254138,
            254116,254156,254208,254212,254220,254174,254206,254229,254230,
            254149,254204,254218,254179,254284,254321,254265,254109,254147,
            254146,254108,254119];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 43 mod 47.
function cached_ids_43(P,hash1,hash2,hash3)
    if hash1 eq 1359 and hash2 eq 2283 and hash3 eq 3126 then
        V:=[23,1,24,5,25,9,26,13,27,17,28,23,29,31,31,39];
        D:=[1,223/6,1,0,2,88/3,1,0,3,155/6,1,0,4,43/3,1,0,5,551/30,1,221/12,1,0,
            7,193/15,1,223/12,1,19,1,0,10,511/30,1,205/12,2,37/2,1,0,14,347/30,
            1,117/10,1,0];
        I:=[502588,502214,502650,502124,502207,502729,502103,502219,502148,
            502481,502136,502636,501284,502161,502239];
    elif hash1 eq 2158 and hash2 eq 2084 and hash3 eq 1567 then
        V:=[20,1,21,5,22,13,23,27,24,43,25,49,26,53,27,57];
        D:=[1,80/3,1,0,2,127/6,2,47/2,1,99/4,1,0,6,18,1,77/4,1,58/3,1,59/3,1,20,
            2,65/3,1,0,13,107/6,2,109/6,1,37/2,1,233/12,1,197/10,1,121/6,1,
            213/10,1,0,21,17,1,55/3,1,0,23,91/6,1,0,24,76/5,1,0,25,73/6,1,0];
        I:=[422315,422166,422305,421814,422118,421808,421838,422187,422158,
            421817,422024,422058,421781,421788,421815,422036,422038,421778,
            422026,421602,422149,421291,421315,421217,421331];
    elif hash1 eq 3427 and hash2 eq 2000 and hash3 eq 2680 then
        V:=[17,1,18,5,19,9,20,15,21,23,22,27,23,31,24,35,25,43,27,49];
        D:=[1,83/2,1,0,2,157/4,1,0,3,205/6,1,377/10,1,0,5,311/12,1,472/15,1,
            508/15,1,0,8,471/20,1,0,9,512/21,1,0,10,573/28,1,0,11,309/20,1,
            1147/60,1,292/15,1,0,14,1271/84,1,1415/84,1,0,16,869/70,1,829/60,1,
            0];
        I:=[518665,518662,518501,518499,518589,518357,518450,518365,518248,
            518263,518200,518264,518343,518202,518276,518285,518206];
    elif hash1 eq 1970 and hash2 eq 3341 and hash3 eq 2189 then
        V:=[23,1,24,7,25,17,27,23];
        D:=[1,155/6,2,157/6,1,0,4,22,1,67/3,2,68/3,1,73/3,1,0,9,41/2,1,125/6,1,
            0,11,185/12,1,31/2,1,521/30,3,0];
        I:=[252120,252241,251875,252246,252122,252132,252256,251872,252037,
            252049,252051,252036,251842,251844,251846];
    elif hash1 eq 419 and hash2 eq 1857 and hash3 eq 2900 then
        V:=[21,1,22,7,23,15,24,19,25,27,26,31,27,39,28,43,29,47];
        D:=[1,61/2,1,65/2,1,0,3,27,1,92/3,1,104/3,1,0,6,151/6,1,0,7,65/3,1,22,1,
            71/3,2,0,11,281/12,1,0,12,268/15,1,239/12,1,20,1,0,15,551/30,1,0,16,
            179/12,1,0,17,461/30,1,0];
        I:=[251006,251372,251221,251211,251345,250663,251210,250894,250716,
            250878,249946,250459,250314,250669,250308,250109,249864];
    elif hash1 eq 1688 and hash2 eq 2469 and hash3 eq 256 then
        V:=[21,1,22,5,23,11,24,17,25,25];
        D:=[1,57/2,1,0,2,21,1,25,5,0,8,127/6,2,43/2,2,0,12,53/3,1,18,1,65/3,1,0,
            15,109/6,5,37/2,2,0];
        I:=[673679,673579,673429,673588,673678,673729,673731,673713,673720,
            673446,673696,673577,673576,673308,673450,673464,673524,673526,
            673648,673425,673639];
    elif hash1 eq 2816 and hash2 eq 1062 and hash3 eq 3460 then
        V:=[22,1,23,21,24,37,25,51,26,69,27,81,28,91,29,95];
        D:=[1,21,3,68/3,1,23,2,74/3,1,25,4,80/3,2,27,1,86/3,1,29,1,0,17,115/6,1,
            127/6,4,43/2,2,139/6,3,47/2,1,151/6,1,51/2,1,0,30,53/3,3,59/3,4,20,
            2,85/4,2,64/3,1,101/4,1,0,43,95/6,1,177/10,1,71/4,1,107/6,2,109/6,1,
            197/10,1,79/4,1,121/6,1,0,52,44/3,1,81/5,1,49/3,3,50/3,1,55/3,1,0,
            59,173/12,1,29/2,3,89/6,1,91/6,1,0,65,223/15,1,0,66,71/6,1,0];
        I:=[248965,249367,249428,249413,249052,249404,249196,249071,249095,
            249256,249316,249058,249203,249157,248933,249131,249337,249050,
            249084,249184,249368,249150,249401,248578,249104,249182,248647,
            248117,247862,248434,248962,249125,248351,248495,249295,249308,
            248618,249156,248256,248354,248597,247592,248436,248353,248263,
            248239,249355,248528,248516,247723,249147,249110,248492,248215,
            248605,249222,249261,248116,248944,248526,248959,249112,248612,
            248620,247520,248596];
    elif hash1 eq 466 and hash2 eq 3140 and hash3 eq 1267 then
        V:=[27,1,29,5,30,9,31,25,32,39];
        D:=[1,119/6,1,0,2,245/12,1,0,3,277/15,1,227/12,1,77/4,1,302/15,1,307/15,
            1,308/15,1,247/12,1,0,10,613/42,1,877/60,1,451/30,1,181/12,1,50/3,1,
            757/42,1,0,16,163/12,1,0];
        I:=[404694,404486,405343,404812,404873,404352,404767,405196,404777,
            404889,405122,404676,405032,404346,404172,405101];
    elif hash1 eq 419 and hash2 eq 3416 and hash3 eq 2803 then
        V:=[27,1,28,9,29,13,30,19,31,29];
        D:=[1,213/10,1,127/4,1,193/6,1,0,4,79/3,2,0,6,221/12,1,45/2,1,0,8,
            369/20,2,19,1,96/5,1,313/15,1,0,13,625/42,1,149/10,1,451/30,1,
            499/30,1,50/3,2,67/4,1,56/3,1,0];
        I:=[397869,396862,398377,396863,399963,399775,395620,396850,398358,
            398217,398403,396831,395913,395910,398928,397870,397820,398552,
            395618,395697];
    elif hash1 eq 560 and hash2 eq 3000 and hash3 eq 1601 then
        V:=[24,1,25,9,26,17,27,25,28,31];
        D:=[1,19,2,62/3,1,21,2,0,6,35/2,4,115/6,3,39/2,1,0,14,16,2,52/3,1,53/3,
            3,0,20,63/4,1,95/6,4,0,25,43/3,1,44/3,1,0];
        I:=[672401,672421,672408,671845,672415,672032,672392,672400,672413,
            671449,671851,671925,672074,672112,672160,672043,671348,671412,
            672141,671595,671504,671519,671983,672395,671517,671620];
    elif hash1 eq 2769 and hash2 eq 3333 and hash3 eq 71 then
        V:=[27,1,28,5,29,13,30,23,31,27];
        D:=[1,127/6,1,0,2,16,1,53/3,3,59/3,2,0,8,95/6,1,97/6,5,33/2,2,107/6,1,0,
            17,43/3,6,0,23,77/6,1,0];
        I:=[664163,664611,664475,664543,664614,663576,664434,664335,663679,
            663743,663760,664226,664232,663730,664330,663781,663586,663597,
            663660,663771,663883,664522,663798];
    elif hash1 eq 2440 and hash2 eq 3159 and hash3 eq 3085 then
        V:=[23,1,24,5,25,11,26,15,27,19,28,25,29,37];
        D:=[1,187/6,1,0,2,76/3,2,82/3,1,0,5,145/6,1,0,6,49/3,2,0,8,183/10,1,
            221/12,1,0,10,217/15,1,29/2,1,175/12,2,223/15,1,15,1,0,16,391/30,1,
            83/6,1,613/42,1,877/60,1,0];
        I:=[418858,418252,418494,416125,416221,416090,418827,415181,416178,
            417397,417379,416496,418277,416069,418877,417068,416083,415076,
            415078];
    elif hash1 eq 325 and hash2 eq 3091 and hash3 eq 2565 then
        V:=[19,1,20,9,21,19,23,31,24,39,26,49];
        D:=[1,45/2,1,57/2,1,65/2,1,0,4,86/3,1,29,1,92/3,1,98/3,1,0,8,115/6,1,
            151/6,2,51/2,1,187/6,1,63/2,1,0,14,95/6,1,97/6,1,145/6,1,0,17,
            215/12,1,91/5,1,55/3,1,67/3,1,0,21,218/15,2,50/3,1,0];
        I:=[420509,420186,420559,420524,420566,420066,420225,420560,420188,
            420550,420187,419354,420152,420551,420530,420153,420218,419305,
            420064,419080,419324,419328,419356];
    elif hash1 eq 1688 and hash2 eq 1236 and hash3 eq 2371 then
        V:=[19,1,20,7,21,15,22,27,23,35,24,49,25,59,27,69];
        D:=[1,59/2,1,235/6,1,0,3,71/3,1,32,1,119/3,1,0,6,237/10,1,143/6,1,145/6,
            1,317/10,2,191/6,1,0,12,263/12,2,121/5,1,73/3,1,0,16,37/2,1,187/10,
            1,113/6,1,115/6,1,267/10,1,163/6,1,0,22,253/15,1,278/15,1,302/15,1,
            307/15,1,0,26,709/42,1,269/14,1,581/30,1,39/2,1,0,30,569/42,1,
            407/30,1,137/10,1,421/30,2,0];
        I:=[514966,514920,514935,514940,514139,514975,514926,514943,514194,
            514973,514903,514938,515000,514844,514942,514787,514997,514500,
            514930,514199,514140,514508,514998,515010,514994,514172,514206,
            514197,514098,514039,514036,514516,514931,515001];
    elif hash1 eq 2158 and hash2 eq 418 and hash3 eq 1553 then
        V:=[25,1,26,7,27,13,28,19,29,25,30,31];
        D:=[1,151/6,1,51/2,1,0,3,53/3,2,58/3,2,0,7,109/6,1,233/12,1,0,9,44/3,1,
            81/5,1,0,11,29/2,2,89/6,1,0,14,40/3,1,223/15,2,179/12,1,0];
        I:=[239263,241605,241494,241515,240539,241124,240002,241539,241534,
            239933,240710,240768,239117,239305,238327,238336,238332];
    elif hash1 eq 560 and hash2 eq 2969 and hash3 eq 1019 then
        V:=[18,1,19,9,20,19,21,37,22,45,23,53,24,65,25,71,26,77];
        D:=[1,80/3,2,27,1,33,1,0,5,175/6,2,187/6,1,199/6,4,67/2,1,0,13,53/3,1,
            85/4,1,76/3,2,109/4,2,82/3,1,146/5,1,125/4,2,95/3,1,0,24,237/10,1,
            305/12,1,155/6,1,0,27,49/3,2,328/15,1,73/3,1,0,31,491/30,1,221/12,1,
            113/6,1,431/20,1,269/12,1,0,36,129/7,1,223/12,1,0,38,451/30,1,91/6,
            2,0,41,419/28,1,907/60,1,0];
        I:=[515463,515478,515569,515590,515208,515588,515501,515476,515510,
            515558,515577,515506,515534,515013,515464,515579,515280,515294,
            515490,514963,514857,515223,515541,514978,515174,514783,515494,
            515576,514562,515271,514506,515015,515070,514502,514329,514123,
            514034,514750,515470,515575,514089,514082];
    elif hash1 eq 2440 and hash2 eq 1644 and hash3 eq 153 then
        V:=[24,1,25,9,26,15,27,23,29,27];
        D:=[1,25,1,80/3,1,27,2,0,5,43/2,1,139/6,1,0,7,53/3,1,59/3,4,20,2,0,14,
            109/6,1,0,15,29/2,1,89/6,2,0];
        I:=[672005,672100,671975,672069,671344,671403,671323,671326,671399,
            671469,672060,671913,671966,671398,671421,671986,672006];
    elif hash1 eq 3333 and hash2 eq 2578 and hash3 eq 1865 then
        V:=[25,1,26,5,27,9,28,15,29,21,30,31,31,37,32,45,33,49];
        D:=[1,143/6,1,0,2,479/12,1,0,3,53/2,1,107/4,1,0,5,278/15,1,19,1,0,7,
            339/20,1,209/12,1,265/14,1,233/12,1,0,11,1027/60,1,394/21,1,0,13,
            31/2,1,469/30,1,1415/84,1,0,16,46/3,1,0,17,849/70,1,515/42,1,737/60,
            1,137/10,1,55/4,1,0];
        I:=[537850,537863,537866,537859,537839,537844,537837,537852,537855,
            537862,537853,537864,537872,537861,537841,537868,537840,537857,
            537851,537867,537871];
    elif hash1 eq 1265 and hash2 eq 724 and hash3 eq 1593 then
        V:=[22,1,23,7,24,13,25,19,26,27,27,33,28,39,29,45];
        D:=[1,88/3,1,124/3,1,0,3,119/4,1,425/12,1,0,5,67/3,1,141/5,1,0,7,221/12,
            1,37/2,1,49/2,1,0,10,50/3,1,21,1,0,12,451/30,3,181/12,1,0,16,76/5,1,
            61/4,1,0,18,797/60,1,817/60,1,0];
        I:=[509326,509373,510592,509735,509329,509110,509732,510639,509228,
            510688,509361,509560,509681,509696,509731,509102,509738,509739,
            509737];
    elif hash1 eq 2158 and hash2 eq 303 and hash3 eq 2238 then
        V:=[20,1,21,7,22,15,23,25,24,33,25,41,26,49,27,55,28,61];
        D:=[1,116/3,1,39,1,0,3,151/6,1,123/4,1,211/6,1,0,6,125/4,1,95/3,1,176/5,
            1,107/3,2,0,11,155/6,1,277/10,1,57/2,1,0,14,298/15,1,101/5,1,382/15,
            1,0,17,491/30,1,33/2,1,599/30,1,0,20,151/12,1,283/15,1,0,22,451/30,
            2,601/30,1,0,25,355/28,1,127/10,1,238/15,2,16,1,0];
        I:=[540005,540018,540010,540006,539999,540002,540012,540127,540014,
            540138,540009,539996,539724,540015,539995,540011,540136,540142,
            540019,540016,539997,539998,540141,539723,539652,539698,539993,
            539994,540017];
    elif hash1 eq 419 and hash2 eq 852 and hash3 eq 3374 then
        V:=[23,1,24,5,25,9,27,15,28,19,30,27];
        D:=[1,237/10,1,0,2,92/3,1,0,3,217/12,1,91/4,1,0,5,461/30,1,0,6,203/15,2,
            14,2,124/7,1,0,11,253/21,1,137/10,1,295/21,2,1714/105,1,0];
        I:=[540242,540292,540278,539774,540212,540232,540262,540324,540328,
            539939,539963,539782,539933,539974,540360];
    elif hash1 eq 701 and hash2 eq 278 and hash3 eq 2372 then
        V:=[22,1,23,5,24,17,25,29,26,41,27,53,29,59];
        D:=[1,23,1,0,2,47/2,1,151/6,1,51/2,1,59/2,1,199/6,1,0,7,59/3,1,64/3,2,
            22,1,71/3,1,76/3,1,0,13,109/6,2,121/6,1,257/12,1,87/4,1,133/6,1,0,
            19,50/3,1,91/5,1,55/3,1,56/3,1,20,1,0,24,101/6,1,221/12,1,0,26,
            401/30,1,27/2,1,0];
        I:=[416588,417156,416430,417229,417228,416860,417264,417285,417304,
            416821,417548,416854,416800,417311,417184,417621,416510,416819,
            417292,416447,417365,417011,416458,417215,416311,417046,417023];
    elif hash1 eq 1829 and hash2 eq 149 and hash3 eq 1845 then
        V:=[22,1,23,9,24,17,25,25,26,31,27,41];
        D:=[1,23,2,25,2,80/3,1,0,6,127/6,1,139/6,2,47/2,2,0,11,77/4,1,59/3,2,20,
            2,0,16,107/6,1,109/6,1,0,18,16,1,49/3,2,50/3,1,55/3,1,0,23,89/6,1,
            0];
        I:=[673732,673736,673625,673735,673357,673608,673367,673478,673363,
            673626,673391,673355,673400,673354,673621,673637,673607,673488,
            673473,673499,673500,673234,673246];
    elif hash1 eq 2863 and hash2 eq 3353 and hash3 eq 772 then
        V:=[22,1,24,5,25,13,26,21,27,25,28,31,29,37,30,43];
        D:=[1,106/3,1,0,2,335/12,1,85/3,1,151/5,1,0,5,551/30,1,221/12,1,267/10,
            1,0,8,313/15,1,0,9,181/12,1,33/2,1,0,11,61/4,1,46/3,1,0,13,797/60,1,
            817/60,1,0,15,1145/84,1,0];
        I:=[509091,509707,509237,509630,509647,509697,509706,509161,509746,
            508723,509774,509101,509775,509773,508434];
    elif hash1 eq 1688 and hash2 eq 657 and hash3 eq 3334 then
        V:=[20,1,21,9,22,21,23,37,24,49,25,53,26,59,27,69,29,77];
        D:=[1,29,1,33,1,104/3,1,0,4,175/6,1,59/2,1,187/6,1,199/6,1,67/2,1,0,9,
            76/3,1,77/3,2,26,1,83/3,3,89/3,1,125/4,1,95/3,1,0,19,131/6,1,237/10,
            2,143/6,2,103/4,1,155/6,1,0,26,67/3,2,0,28,221/12,1,113/6,4,0,33,
            199/12,1,227/12,1,19,2,58/3,1,0,38,57/4,1,44/3,1,181/12,1,0,41,
            117/10,1,0];
        I:=[252933,253214,252842,252952,252932,252949,252632,253159,252584,
            252862,253217,252924,252836,252926,253160,253157,252760,252845,
            252682,252827,252828,252676,252683,252562,252847,252524,252525,
            252583,252582,252856,252857,252858,252546,252565,252534,252570,
            252569,252587,252580,252579,252822];
    elif hash1 eq 2017 and hash2 eq 696 and hash3 eq 3310 then
        V:=[27,1,28,11,29,19,30,27,31,31,33,41];
        D:=[1,151/6,1,163/6,1,115/4,1,123/4,1,0,5,116/5,1,71/3,1,109/4,1,0,8,
            107/6,1,257/12,1,237/10,1,0,11,298/15,1,0,12,29/2,2,89/6,1,197/12,1,
            167/10,1,0,17,38/3,1,391/30,2,157/12,1,79/6,1,0];
        I:=[220587,220712,222193,221535,220522,221808,221113,221814,221145,
            219989,219991,221530,221643,222396,220862,221563,220845,220847,
            220848,220849,222368];
    elif hash1 eq 1594 and hash2 eq 648 and hash3 eq 1004 then
        V:=[22,1,23,11,24,19,25,29,26,39,27,43,28,49];
        D:=[1,77/3,1,109/4,1,82/3,1,117/4,1,0,5,133/6,2,155/6,1,273/10,1,0,9,
            111/5,1,67/3,2,382/15,1,79/3,1,0,14,241/12,1,611/30,1,245/12,1,83/4,
            1,0,18,253/15,1,0,19,511/30,1,229/12,1,0,21,14,2,61/4,1,0];
        I:=[421608,421381,421440,421702,421491,421655,421682,421437,421654,
            421556,421679,421466,421178,421730,421212,421749,421386,421519,
            421286,421154,421333,421684,421176];
    elif hash1 eq 1359 and hash2 eq 3245 and hash3 eq 2054 then
        V:=[22,1,23,7,24,13,25,19,26,31,27,45,28,49,29,55];
        D:=[1,83/3,1,107/3,1,0,3,155/6,1,179/6,2,0,6,287/12,2,24,1,0,9,41/2,1,
            265/12,2,45/2,1,227/10,2,289/12,1,0,16,278/15,1,19,1,121/6,1,247/12,
            1,313/15,2,21,1,0,23,511/30,1,0,24,323/21,1,233/15,2,0,27,55/4,1,
            61/4,1,559/30,1,0];
        I:=[422950,422897,422731,422739,423061,422709,422752,422955,422742,
            422810,422971,423238,422396,423011,422819,422815,422336,422934,
            422820,422759,422761,423363,422812,422361,422362,423022,422941,
            422458,422332];
    elif hash1 eq 372 and hash2 eq 2052 and hash3 eq 1774 then
        V:=[27,1,28,7,29,13,30,17,32,27,33,31];
        D:=[1,127/6,1,47/2,1,0,3,58/3,1,59/3,3,0,7,107/6,1,0,8,81/5,1,65/4,1,
            49/3,2,50/3,4,0,16,13,1,0,17,133/12,1,0];
        I:=[667428,667306,667290,666990,667341,667423,665751,665848,665665,
            666163,667310,665196,666870,666912,667416,666199,666016];
    elif hash1 eq 466 and hash2 eq 1316 and hash3 eq 2165 then
        V:=[21,1,22,7,23,11,24,15,25,19,26,29,27,35,28,47];
        D:=[1,175/6,1,187/6,1,0,3,95/3,1,0,4,107/6,1,0,5,61/3,1,0,6,193/12,1,
            33/2,1,101/6,1,113/6,1,0,10,321/20,1,283/15,2,0,13,149/10,1,161/10,
            1,97/6,1,973/60,1,781/42,1,0,18,66/5,1,0];
        I:=[424783,423981,423214,424296,423600,424272,424816,424801,424782,
            422990,422537,422678,423751,422544,422582,422677,422344,423415];
    elif hash1 eq 3192 and hash2 eq 3140 and hash3 eq 851 then
        V:=[24,1,25,11,26,21,27,35,28,41];
        D:=[1,19,3,62/3,3,21,2,68/3,1,0,10,103/6,6,35/2,1,115/6,4,39/2,1,0,22,
            46/3,1,47/3,2,86/5,1,52/3,1,96/5,1,59/3,1,0,29,85/6,3,95/6,2,0,34,
            208/15,1,167/12,1,0];
        I:=[244926,244934,244974,244834,244980,245081,244880,244893,244750,
            244789,245064,245127,245243,245335,245347,243961,242867,243077,
            243970,244843,243849,245072,243845,243866,243128,243327,243178,
            243315,243123,243996,245225,243219,243730,243125,245407];
    elif hash1 eq 2675 and hash2 eq 734 and hash3 eq 2110 then
        V:=[12,1,13,7,14,11,15,17,16,21,17,29,18,35,21,39,23,43];
        D:=[1,46,1,50,2,0,4,85/2,1,0,5,122/3,1,43,1,0,7,87/2,1,0,8,133/4,1,
            101/3,1,107/3,1,0,11,103/4,1,215/6,2,0,14,311/12,1,0,15,277/12,1,0,
            16,937/60,1,0];
        I:=[430038,430039,430040,430041,430036,430037,430006,430004,430008,
            430001,430005,429977,430002,430007,429965,429976];
    elif hash1 eq 795 and hash2 eq 1156 and hash3 eq 1633 then
        V:=[17,1,19,5,20,11,21,17,22,25,23,31,24,37,26,43,27,47,29,63];
        D:=[1,307/6,1,0,2,227/6,1,159/4,1,0,4,538/15,1,628/15,1,0,6,361/12,1,
            385/12,1,409/12,1,0,9,368/15,1,433/15,2,0,12,231/8,1,871/30,1,0,14,
            59/3,1,323/15,1,0,16,81/5,1,0,17,1159/84,1,187/12,1,643/40,1,389/24,
            1,487/30,1,785/42,1,3953/210,1,0,24,1097/90,1,1717/140,1,0];
        I:=[517901,517821,517898,517824,517849,517987,517920,517922,517924,
            517817,517846,517644,517638,517610,517815,517842,517722,517721,
            517630,517640,517632,517621,517619,517625,517623];
    elif hash1 eq 1077 and hash2 eq 2818 and hash3 eq 1984 then
        V:=[12,1,13,11,14,15,15,21,16,29,17,39,18,49,19,69,20,79,21,87,22,95,23,
            109];
        D:=[1,38,1,44,1,50,1,52,1,0,5,85/2,1,0,6,39,1,47,1,0,8,67/2,1,223/6,1,
            171/4,1,0,11,93/4,1,125/4,1,100/3,1,112/3,1,0,15,233/10,1,281/12,1,
            191/6,1,397/10,1,0,19,322/15,1,67/3,1,79/3,1,30,1,63/2,1,478/15,2,
            383/12,1,97/3,1,71/2,1,0,29,431/20,1,265/12,1,731/30,1,53/2,1,0,33,
            86/5,1,223/12,1,489/20,1,0,36,997/60,1,39/2,1,631/30,1,0,39,1003/60,
            1,523/28,1,401/21,1,77/4,2,97/5,1,113/5,1,0,46,1171/70,1,0];
        I:=[519634,519639,519623,519620,519611,519630,519545,519551,519616,
            519547,519546,519606,519561,519570,519425,519459,519403,519486,
            519489,519467,519516,519400,519576,519407,519487,519404,519553,
            519440,519426,519436,519490,519406,519485,519397,519387,519447,
            519405,519433,519449,519361,519368,519363,519398,519499,519352,
            519350];
    elif hash1 eq 278 and hash2 eq 77 and hash3 eq 148 then
        V:=[26,1,27,7,28,13,29,21,30,33,31,39];
        D:=[1,71/3,1,101/4,1,0,3,131/6,1,237/10,1,0,5,49/3,1,20,1,328/15,1,0,8,
            197/12,2,33/2,1,67/4,1,103/6,1,113/6,1,0,14,44/3,1,15,1,0,16,79/6,1,
            27/2,2,0];
        I:=[238993,239173,238355,238659,239048,238948,238381,238962,239180,
            239178,238364,240529,239067,240304,238365,239039,238960,239040];
    elif hash1 eq 3192 and hash2 eq 745 and hash3 eq 3087 then
        V:=[25,1,26,11,27,19,28,27,29,35,30,43];
        D:=[1,47/2,2,151/6,1,51/2,1,55/2,1,0,6,18,1,85/4,1,24,1,0,9,97/6,2,
            109/6,1,121/6,2,0,14,65/4,1,50/3,1,17,1,0,17,77/6,1,89/6,1,197/12,1,
            0,20,188/15,1,0];
        I:=[408455,408619,408226,408249,408281,408237,406449,407563,407211,
            407538,407125,407153,408542,406317,408459,408567,408117,407489,
            406318,406699];
    elif hash1 eq 2252 and hash2 eq 246 and hash3 eq 1776 then
        V:=[23,1,24,7,25,15,26,23,27,29,28,37];
        D:=[1,45/2,1,49/2,5,0,7,19,1,21,1,68/3,1,0,10,103/6,2,115/6,1,39/2,1,0,
            14,53/3,3,18,1,0,18,185/12,1,95/6,2,97/6,2,0,23,44/3,1,0];
        I:=[672325,672080,672145,672265,672281,672315,672271,672085,672161,
            672132,672146,672151,672134,671802,671894,672061,672276,672139,
            671588,671732,671677,671911,671203];
    elif hash1 eq 1970 and hash2 eq 1027 and hash3 eq 1159 then
        V:=[27,1,28,5,29,13,30,17,31,27,32,43];
        D:=[1,111/4,1,0,2,61/3,2,89/4,1,67/3,1,0,6,187/10,1,0,7,199/12,1,253/15,
            2,17,2,52/3,1,0,13,451/30,1,181/12,1,91/6,1,185/12,1,31/2,1,339/20,
            1,521/30,2,0,21,77/6,1,40/3,1,323/21,1,77/5,1,0];
        I:=[239914,239196,240003,238833,240413,238566,240169,239711,239712,
            239228,239846,240327,238602,238628,240464,239834,240408,238157,
            239356,239807,239989,240427,238009,238011];
    elif hash1 eq 2581 and hash2 eq 2778 and hash3 eq 3258 then
        V:=[25,1,26,15,27,29,28,39,29,49,30,59,31,65];
        D:=[1,163/6,2,55/2,1,175/6,1,187/6,2,63/2,2,199/6,1,0,10,65/3,1,71/3,1,
            77/3,1,26,1,109/4,1,83/3,1,0,16,197/10,1,257/12,1,131/6,1,143/6,1,0,
            20,49/3,1,55/3,1,56/3,2,61/3,1,0,25,33/2,3,101/6,1,217/12,1,221/12,
            2,0,32,175/12,2,97/6,1,0,35,27/2,1,451/30,1,0];
        I:=[242906,245545,243594,245649,244365,245536,245277,245469,245276,
            245228,244358,244148,245458,244316,242268,244172,244349,242267,
            244121,244318,245275,245177,245629,245547,242921,244429,245178,
            245564,244528,242253,242808,243949,244006,242193,244303,242139];
    elif hash1 eq 2205 and hash2 eq 2478 and hash3 eq 888 then
        V:=[27,1,28,5,29,9,30,15,31,23];
        D:=[1,47/2,1,0,2,59/3,7,0,9,107/6,2,109/6,1,0,12,65/4,1,49/3,1,50/3,1,0,
            15,29/2,1,89/6,1,0];
        I:=[60862,60647,60658,60826,60873,60876,60877,60883,60446,60699,60448,
            60419,60602,60848,60774,60593];
    elif hash1 eq 3286 and hash2 eq 1710 and hash3 eq 1560 then
        V:=[21,1,22,7,23,19,24,27,25,39,26,51,27,61,28,65];
        D:=[1,163/6,2,55/2,1,0,4,59/3,1,20,1,70/3,1,71/3,1,24,1,0,9,119/6,1,
            121/6,1,131/6,1,0,12,50/3,1,55/3,3,56/3,1,20,1,81/4,1,0,19,91/6,1,
            33/2,1,167/10,2,101/6,2,187/10,1,0,26,175/12,1,223/15,1,46/3,2,
            203/12,1,0,31,27/2,1,0,32,12,2,0];
        I:=[421047,421048,421035,421054,421044,421050,421052,421041,421007,
            421065,420986,421075,421056,421062,421067,421036,421022,421051,
            420985,421076,420991,420995,420998,421043,421004,421058,421003,
            421032,421038,420944,420993,421012,421015];
    elif hash1 eq 795 and hash2 eq 2769 and hash3 eq 1049 then
        V:=[26,1,27,7,28,13,29,23,30,29,31,35];
        D:=[1,62/3,1,23,3,0,5,115/6,2,39/2,3,0,10,46/3,1,47/3,3,52/3,1,18,1,0,
            16,33/2,1,177/10,1,0,18,38/3,1,238/15,1,0,20,25/2,1,29/2,1,0];
        I:=[234630,234933,234947,234957,231917,234610,233530,234877,234924,
            234900,231977,231986,234880,232566,233162,232901,231395,234593,
            231393,234699,229589];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 44 mod 47.
function cached_ids_44(P,hash1,hash2,hash3)
    if hash1 eq 843 and hash2 eq 54 and hash3 eq 977 then
        V:=[26,1,27,5,28,13,29,19,30,29,31,33,32,47,33,57];
        D:=[1,335/12,1,0,2,659/30,1,269/12,1,247/10,1,0,5,278/15,1,56/3,1,0,7,
            511/30,1,379/20,1,631/30,1,253/12,1,0,11,358/21,1,0,12,697/60,1,
            35/3,1,137/10,1,421/30,1,85/6,1,893/60,2,0,19,977/84,1,703/60,2,
            707/60,1,71/6,1,0,24,485/42,1,117/10,1,283/24,1,2477/210,1,167/14,1,
            0];
        I:=[503350,502675,502988,502629,503344,502727,502632,503372,503343,
            502271,501811,503057,503058,502643,502630,502672,501961,501970,
            501865,501968,501969,503373,503345,501067,501080,501376,501380,
            501403];
    elif hash1 eq 467 and hash2 eq 2729 and hash3 eq 783 then
        V:=[27,1,28,7,29,13,30,17,31,25,32,33];
        D:=[1,109/6,1,197/10,1,0,3,191/12,1,49/3,1,0,5,89/6,1,0,6,66/5,1,218/15,
            2,86/5,1,0,10,79/6,1,27/2,2,83/6,1,0,14,229/20,1,178/15,1,53/4,1,0];
        I:=[407989,407733,407479,407495,407686,407460,407302,407303,404751,
            408461,408449,408666,407798,404430,405382,404112];
    elif hash1 eq 2488 and hash2 eq 864 and hash3 eq 426 then
        V:=[25,1,26,7,27,15,28,25,29,29];
        D:=[1,35/2,2,115/6,4,0,7,16,2,53/3,3,18,2,0,14,63/4,1,95/6,3,97/6,3,
            33/2,1,0,22,44/3,4,0,26,79/6,1,0];
        I:=[241666,241720,241533,241669,241703,241785,240537,241493,241627,
            241652,241675,239135,241291,240592,241549,241643,241685,240674,
            241105,241683,241210,239131,239247,241485,241582,240482];
    elif hash1 eq 2441 and hash2 eq 1767 and hash3 eq 1597 then
        V:=[23,1,24,5,25,13,26,21,27,25,28,29];
        D:=[1,57/2,1,0,2,25,4,86/3,1,33,1,0,8,43/2,2,139/6,2,151/6,1,0,13,22,1,
            0,14,119/6,1,0,15,91/5,1,43/2,1,0];
        I:=[243555,243448,244098,244155,244177,244099,243015,242950,244164,
            244004,244201,243011,243014,243055,243049,241853];
    elif hash1 eq 2159 and hash2 eq 2535 and hash3 eq 3400 then
        V:=[16,1,17,7,18,21,19,31,20,45,22,53,23,67,24,77];
        D:=[1,86/3,1,37,1,0,3,151/6,1,51/2,1,59/2,1,187/6,1,211/6,1,223/6,1,0,9,
            83/3,1,88/3,1,89/3,1,156/5,1,0,13,109/6,1,145/6,2,49/2,1,277/10,1,
            167/6,1,169/6,1,0,20,55/3,2,19,1,61/3,1,0,24,17,3,283/15,2,19,2,
            157/7,1,271/12,1,275/12,1,0,34,339/20,1,511/30,1,205/12,2,107/6,3,0,
            41,77/5,1,191/12,1,0];
        I:=[429483,429475,429490,429500,429516,429480,429209,429494,429491,
            429479,429506,429205,429507,429501,429508,429517,429224,429190,
            429269,429486,429495,429157,429177,429488,429497,429509,429146,
            429161,429152,429160,429111,429101,429117,429239,429236,429237,
            429270,429151,429153,429154,429100,429118];
    elif hash1 eq 3052 and hash2 eq 2364 and hash3 eq 3080 then
        V:=[24,1,25,9,26,17,27,23,28,27,29,33,30,37,31,41,32,45];
        D:=[1,166/5,1,101/3,1,113/3,1,0,4,237/10,1,353/12,1,317/10,1,0,7,73/3,1,
            407/12,1,0,9,671/30,1,0,10,15,1,409/20,1,0,12,181/12,1,0,13,208/15,
            1,0,14,467/30,1,0,15,383/28,1,0];
        I:=[504626,504559,504724,504774,504863,504937,504752,504978,504936,
            504749,504604,504955,504897,504428,504748];
    elif hash1 eq 3240 and hash2 eq 2707 and hash3 eq 2224 then
        V:=[24,1,25,5,26,17,27,25,28,29,29,35];
        D:=[1,83/3,1,0,2,217/10,1,131/6,1,133/6,1,305/12,1,53/2,1,0,7,167/12,1,
            91/5,1,55/3,2,0,11,223/10,1,0,12,15,1,46/3,1,0,14,757/60,1,38/3,2,
            0];
        I:=[415981,416249,416253,415602,417455,415343,416984,415634,415612,
            415641,415539,416150,416162,416902,416901,416974];
    elif hash1 eq 2159 and hash2 eq 1211 and hash3 eq 1609 then
        V:=[22,1,23,7,24,13,25,27,27,31];
        D:=[1,17,1,74/3,1,0,3,139/6,1,107/4,1,0,5,47/3,3,52/3,1,53/3,1,96/5,1,
            106/5,1,101/4,1,0,13,157/10,1,0,14,29/2,2,0];
        I:=[412030,412212,409845,409965,411849,411853,411875,411857,409835,
            410873,409844,409916,410268,410173,411858];
    elif hash1 eq 185 and hash2 eq 2932 and hash3 eq 2809 then
        V:=[23,1,24,13,25,19,26,23,27,27,28,35];
        D:=[1,47/2,1,163/6,1,55/2,4,175/6,1,63/2,1,0,9,71/3,4,76/3,1,0,14,119/6,
            2,0,16,50/3,1,0,17,33/2,3,101/6,1,103/6,1,0,22,46/3,2,0];
        I:=[248826,249474,248787,248801,248843,249535,249473,249631,247823,
            248749,248838,249480,248792,248766,248786,248813,248746,248762,
            249481,249479,247822,247826,248881];
    elif hash1 eq 796 and hash2 eq 159 and hash3 eq 3021 then
        V:=[26,1,27,5,28,13,29,17,30,25,31,29,32,35];
        D:=[1,88/3,1,0,2,133/6,2,145/6,1,155/6,2,0,7,56/3,1,0,8,167/10,1,551/30,
            1,113/6,2,0,12,248/15,1,0,13,997/60,1,205/12,1,0,15,203/15,2,0];
        I:=[404284,404314,404316,404199,404125,404278,404397,404398,404336,
            404297,404299,404334,404154,404229,404212,404222];
    elif hash1 eq 1642 and hash2 eq 1513 and hash3 eq 2692 then
        V:=[22,1,23,5,24,17,25,21,26,31,27,35,28,45];
        D:=[1,92/3,1,0,2,127/6,1,163/6,3,55/2,1,175/6,1,187/6,1,0,9,126/5,1,0,
            10,71/4,1,119/6,2,131/6,1,133/6,1,0,15,215/12,1,0,16,173/12,1,
            539/30,1,37/2,1,187/10,1,0,20,15,3,0];
        I:=[249283,249206,248803,249152,249176,248986,248570,248584,248260,
            248593,248245,249115,248462,248588,248465,248456,247539,247511,
            248252,248670,249021,249136];
    elif hash1 eq 2817 and hash2 eq 3332 and hash3 eq 764 then
        V:=[19,1,20,7,21,15,22,23,23,29,24,37,25,41];
        D:=[1,65/2,1,73/2,1,0,3,80/3,1,86/3,2,33,1,0,7,139/6,1,55/2,1,175/6,1,0,
            10,71/3,1,24,1,0,12,87/4,1,131/6,1,133/6,1,0,15,62/3,1,0,16,33/2,1,
            659/30,1,0];
        I:=[424458,424498,423978,423995,424481,424455,423975,424460,423953,
            423641,424415,423532,423549,422708,423592,423070,422523];
    elif hash1 eq 749 and hash2 eq 297 and hash3 eq 3442 then
        V:=[26,1,27,7,28,11,29,19,30,23,31,29];
        D:=[1,62/3,1,21,1,0,3,115/6,2,0,5,47/3,1,52/3,2,53/3,3,0,11,95/6,1,0,12,
            167/12,3,14,2,0,17,127/10,1,0];
        I:=[61085,61260,60974,61162,61221,60970,61214,60965,60972,61127,61138,
            60957,60991,61119,61014,61158,60950];
    elif hash1 eq 373 and hash2 eq 2447 and hash3 eq 2830 then
        V:=[17,1,18,11,19,25,20,37,21,49,22,59,23,69,24,77,25,85];
        D:=[1,49/2,1,53/2,1,57/2,1,61/2,1,0,5,19,1,68/3,1,80/3,2,27,1,86/3,1,
            110/3,1,0,12,127/6,1,139/6,1,151/6,1,163/6,1,175/6,1,0,17,64/3,1,
            116/5,1,93/4,2,101/4,1,76/3,1,0,23,107/6,1,37/2,1,197/10,1,103/4,1,
            0,27,50/3,1,298/15,1,322/15,1,67/3,1,0,31,551/30,1,37/2,1,269/12,1,
            0,34,76/5,1,203/12,1,52/3,1,0,37,625/42,1,461/30,1,31/2,1,0];
        I:=[515537,515480,515500,515591,515495,515544,515483,515553,515499,
            515592,515264,515582,515529,514755,515552,515326,515547,514523,
            515178,515261,515336,514264,515048,514288,514303,514265,515277,
            514964,514948,514263,514309,514279,514174,514610,514858,514286,
            514040,514087,514046];
    elif hash1 eq 326 and hash2 eq 2749 and hash3 eq 1965 then
        V:=[23,1,24,7,25,19,26,23,27,35,28,49,29,53];
        D:=[1,43/2,1,115/4,1,0,3,59/3,2,116/5,1,101/4,1,26,1,82/3,1,0,9,197/10,
            1,0,10,215/12,1,73/4,1,99/5,1,62/3,1,22,1,0,15,193/12,1,491/30,1,
            197/12,2,33/2,1,539/30,1,217/12,1,0,22,344/21,1,0,23,401/30,3,
            451/30,1,0];
        I:=[416801,417635,416669,416811,416667,415900,417599,415924,416812,
            417183,415686,416891,417804,416039,417088,416806,416906,417109,
            415685,416503,417629,415074,415077,416814,416816,415457];
    elif hash1 eq 3099 and hash2 eq 2028 and hash3 eq 307 then
        V:=[24,1,25,5,27,11,28,15,29,19,30,29,31,37];
        D:=[1,71/3,1,0,2,233/12,1,61/2,1,0,4,197/12,1,0,5,46/3,1,0,6,401/30,1,
            83/6,2,451/30,2,177/10,1,0,12,41/3,3,208/15,1,65/4,1,0,17,677/60,1,
            499/42,1,817/60,1,41/3,1,0];
        I:=[503209,503313,503459,502977,502990,502971,503474,503475,502869,
            502870,501670,503560,503562,503563,503157,501072,503089,501912,
            501124,501344];
    elif hash1 eq 1642 and hash2 eq 3109 and hash3 eq 1231 then
        V:=[23,1,24,7,25,15,27,23];
        D:=[1,41/2,1,49/2,1,0,3,56/3,1,62/3,1,23,4,0,9,103/6,2,35/2,1,39/2,2,0,
            14,83/6,1,95/6,1,0];
        I:=[245268,245338,245329,245086,242936,243929,243997,245351,245214,
            245301,245080,243009,245369,243073,243983];
    elif hash1 eq 2676 and hash2 eq 829 and hash3 eq 287 then
        V:=[16,1,17,7,18,23,19,43,20,57,21,73,22,91,23,101,24,109,25,115];
        D:=[1,28,1,32,1,0,3,53/2,2,57/2,1,61/2,3,65/2,1,69/2,1,73/2,2,85/2,1,0,
            14,68/3,1,74/3,1,25,1,92/3,3,31,2,33,1,104/3,2,39,1,128/3,1,0,27,
            83/4,1,47/2,1,115/4,1,175/6,1,59/2,1,187/6,2,0,34,70/3,3,71/3,1,
            76/3,1,77/3,1,89/3,1,94/3,1,95/3,2,0,44,217/10,1,87/4,1,131/6,2,
            281/12,2,237/10,2,145/6,1,157/6,2,167/6,1,0,56,18,1,239/12,1,20,2,
            22,1,0,61,491/30,2,167/10,1,551/30,2,0,66,223/12,2,19,2,0,70,877/60,
            1,0];
        I:=[425350,425263,425305,425370,425372,425332,425349,425351,425105,
            425360,425281,425365,425102,425103,425352,425336,425072,425135,
            425243,425257,425282,425094,425110,425321,425152,425000,425132,
            425145,425150,425107,425014,425007,425175,425128,425242,425262,
            425130,425270,425084,425056,425155,424912,424952,424922,425057,
            425086,425159,425029,425190,424907,425026,425090,425052,425091,
            425009,424913,425055,424994,425075,424894,424992,425020,425017,
            424936,425053,424873,424910,424865,424891,424960];
    elif hash1 eq 1642 and hash2 eq 1227 and hash3 eq 2383 then
        V:=[23,1,24,7,25,17,26,29,27,39];
        D:=[1,39/2,2,139/6,2,0,5,53/3,1,58/3,1,59/3,3,20,2,0,12,33/2,1,107/6,1,
            109/6,3,37/2,1,79/4,1,0,19,49/3,5,50/3,1,17,1,91/5,1,0,27,29/2,1,
            89/6,1,91/6,3,101/6,1,0];
        I:=[247109,247133,246949,247076,247001,247143,246950,247004,247039,
            247049,247055,247063,247018,246742,247117,247121,247059,247160,
            246631,246758,247067,247069,247080,246723,246680,246640,247201,
            247010,247048,247060,247061,246480];
    elif hash1 eq 2770 and hash2 eq 1656 and hash3 eq 934 then
        V:=[21,1,22,5,23,11,24,19,25,31,27,37,30,45];
        D:=[1,61/2,1,0,2,98/3,1,33,1,0,4,51/2,1,163/6,1,187/6,1,0,7,64/3,1,70/3,
            1,71/3,1,24,1,26,1,0,12,79/4,1,87/4,1,0,14,221/12,2,113/6,1,115/6,1,
            0,18,173/15,1,0];
        I:=[250914,250919,250244,250165,250720,250700,250349,250538,250283,
            250186,250107,250156,250294,249869,249875,250121,250323,250576];
    elif hash1 eq 1501 and hash2 eq 107 and hash3 eq 2469 then
        V:=[21,1,22,5,23,9,24,19,25,29,26,41,27,47,28,51,29,55];
        D:=[1,259/6,1,0,2,38,1,0,3,143/6,1,127/4,1,193/6,1,401/12,1,0,7,55/3,1,
            61/3,1,79/3,1,94/3,1,0,11,359/20,1,551/30,1,599/30,1,307/14,1,191/8,
            1,0,16,52/3,1,223/12,1,0,18,529/30,1,0,19,74/5,1,0,20,603/40,1,0];
        I:=[512458,512477,511606,511865,512444,511995,512248,512252,510875,
            511199,512255,512309,511959,511269,510876,511482,511941,511208,
            512119,510870];
    elif hash1 eq 2582 and hash2 eq 3126 and hash3 eq 2548 then
        V:=[18,1,19,5,20,15,21,23,22,29,23,41,24,49,25,61,27,67];
        D:=[1,33,1,0,2,175/6,1,199/6,1,223/6,1,235/6,1,0,6,95/3,2,32,1,113/3,1,
            0,10,155/6,1,179/6,1,0,12,287/12,1,73/3,3,74/3,1,131/5,1,359/12,1,0,
            19,611/30,1,265/12,1,313/12,1,0,22,283/15,1,313/15,1,21,1,441/20,1,
            332/15,1,0,27,221/10,1,451/20,1,0,29,421/30,1,937/60,1,47/3,1,0];
        I:=[426947,426907,426774,426951,426707,426901,426925,426935,426973,
            426884,426859,426900,426880,426911,426971,426905,426904,426683,
            426897,426863,426700,426899,426883,426885,426567,426549,426472,
            426448,426970,426444,426434];
    elif hash1 eq 1736 and hash2 eq 897 and hash3 eq 64 then
        V:=[25,1,26,9,27,15,29,27];
        D:=[1,103/6,4,115/6,3,39/2,5,0,13,16,2,52/3,1,0,16,153/10,1,185/12,2,
            97/6,3,33/2,1,177/10,1,0,24,79/6,2,0];
        I:=[245735,245742,245781,245813,244014,245592,245611,244728,245211,
            245423,245754,245823,244781,245440,244100,244465,245156,245233,
            244764,244766,245544,245374,242198,243469,244460];
    elif hash1 eq 1642 and hash2 eq 3225 and hash3 eq 959 then
        V:=[14,1,15,7,17,19,18,23,20,37,21,43,24,51];
        D:=[1,43,1,134/3,1,0,3,99/4,1,107/4,1,75/2,1,235/6,1,79/2,1,0,8,145/6,1,
            0,9,89/4,1,358/15,1,287/12,1,24,1,74/3,1,40,1,0,15,481/20,1,145/6,1,
            0,17,75/4,1,229/12,1,677/28,1,0,20,213/14,1,1103/60,1,569/30,1,0];
        I:=[519567,519559,519564,519469,519595,519632,519586,519540,519457,
            519496,519593,519389,519539,519347,519484,519532,519413,519382,
            519325,519353,519322,519323];
    elif hash1 eq 890 and hash2 eq 1672 and hash3 eq 1334 then
        V:=[17,1,18,5,19,9,20,17,21,25,22,31,23,39,25,47];
        D:=[1,69/2,1,0,2,31,2,0,4,51/2,2,175/6,1,59/2,1,0,8,71/3,1,77/3,2,83/3,
            1,0,12,131/6,1,143/6,2,0,15,20,2,61/3,1,67/3,1,0,19,599/30,1,20,1,
            671/30,1,0,22,31/2,1,0];
        I:=[254364,254337,254363,254332,254339,254235,254336,254340,254167,
            254236,254165,254164,254162,254163,254161,254170,254168,254169,
            254114,254113,254111,254117];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 45 mod 47.
function cached_ids_45(P,hash1,hash2,hash3)
    if hash1 eq 1173 and hash2 eq 2883 and hash3 eq 2087 then
        V:=[14,1,15,7,16,15,17,21,18,25,19,31,20,39,21,47,22,51,23,55,24,59];
        D:=[1,38,1,46,2,0,4,69/2,1,77/2,1,85/2,1,0,7,104/3,1,43,1,0,9,211/6,2,0,
            11,83/3,1,106/3,1,0,13,79/4,1,121/6,1,167/6,1,0,16,73/4,1,239/12,2,
            335/12,1,0,20,137/6,1,0,21,275/12,1,0,22,181/12,1,0,23,907/60,1,0];
        I:=[518681,518644,518653,518664,518642,518682,518627,518638,518623,
            518680,518686,518645,518675,518648,518462,518676,518615,518655,
            518372,518465,518348,518685,518433];
    elif hash1 eq 2959 and hash2 eq 3238 and hash3 eq 597 then
        V:=[22,1,23,7,24,15,25,27,26,33,27,41];
        D:=[1,23,1,86/3,1,0,3,127/6,2,139/6,1,99/4,1,0,7,18,1,59/3,3,20,1,85/4,
            1,65/3,3,0,16,177/10,1,121/6,3,0,20,44/3,2,49/3,1,50/3,1,0,24,
            431/30,1,173/12,1,29/2,2,59/4,1,0];
        I:=[249315,249166,249281,249400,249241,248237,248580,248498,248654,
            248957,249170,248365,248258,249100,249303,248319,248255,248318,
            249263,248316,248963,248297,248644,248417,248951,248425,249070,
            248641];
    elif hash1 eq 1079 and hash2 eq 2525 and hash3 eq 2832 then
        V:=[25,1,26,5,27,15,28,19,29,33];
        D:=[1,31/2,1,0,2,47/3,2,16,2,53/3,1,77/4,1,0,8,85/6,1,0,9,14,2,71/5,2,
            43/3,1,44/3,2,191/12,1,16,1,0,18,431/30,1,0];
        I:=[408602,408593,408709,408524,408629,407117,407734,408564,408382,
            408663,407569,407573,407567,408426,408582,404578,407880,404542];
    elif hash1 eq 3100 and hash2 eq 1763 and hash3 eq 1415 then
        V:=[22,1,23,7,24,19,25,29,26,41,27,49,28,65,30,71];
        D:=[1,29,1,92/3,1,0,3,139/6,1,151/6,1,163/6,2,55/2,4,59/2,1,0,12,65/3,2,
            71/3,5,24,1,77/3,3,0,23,119/6,1,121/6,2,87/4,1,131/6,2,133/6,3,0,32,
            18,1,55/3,4,56/3,1,0,38,197/12,1,167/10,1,101/6,1,221/12,1,37/2,2,
            75/4,1,113/6,2,0,47,175/12,1,179/12,1,0,49,35/3,1,0];
        I:=[673702,673415,673506,673492,673344,673505,673343,673345,673552,
            673693,673553,673366,673620,673271,673282,673365,673554,673724,
            673737,673268,673351,673352,673280,673273,673371,673223,673346,
            673353,673230,673231,673270,673284,673243,673358,673369,673490,
            673489,673237,673274,673277,673217,673220,673221,673219,673238,
            673240,673311,673316,673314];
    elif hash1 eq 280 and hash2 eq 351 and hash3 eq 1662 then
        V:=[21,1,22,11,23,31,24,39,25,55,26,65,27,81,28,93,29,105];
        D:=[1,47/2,2,151/6,1,55/2,1,131/4,1,0,6,64/3,2,65/3,1,70/3,1,126/5,1,26,
            1,30,1,95/3,1,32,1,101/3,1,0,16,133/6,1,143/6,1,145/6,1,0,19,91/5,1,
            56/3,1,81/4,1,61/3,1,21,1,22,1,287/12,1,0,26,167/10,1,101/6,1,
            551/30,1,113/6,1,0,30,15,1,46/3,1,203/12,1,52/3,1,56/3,3,283/15,1,
            227/12,1,0,39,27/2,1,83/6,1,85/6,2,511/30,1,571/30,2,0,46,203/15,2,
            163/12,1,14,2,91/6,1,187/12,1,0,53,47/4,1,0];
        I:=[509980,510068,510216,509522,509573,509577,510230,509559,509607,
            510030,510218,510031,509970,510052,509904,509923,509546,509986,
            510053,509954,509995,510212,509981,509123,509344,509618,509575,
            509509,509194,509896,509914,509825,509956,509024,509274,509300,
            509388,509271,508592,510062,509464,509496,509506,508709,508879,
            509777,509982,508550,509504,509524,508802,508799,508643];
    elif hash1 eq 2160 and hash2 eq 2493 and hash3 eq 1709 then
        V:=[21,1,22,11,23,17,24,25,25,31,26,37,27,41,28,45,29,51];
        D:=[1,139/6,1,187/6,1,71/2,1,223/6,1,0,5,64/3,1,95/3,1,0,7,131/6,1,
            257/10,1,169/6,1,0,10,56/3,1,81/4,1,0,12,359/20,1,611/30,1,0,14,15,
            1,0,15,91/6,1,0,16,76/5,1,61/4,2,0,19,371/30,1,0];
        I:=[510740,510721,510540,510732,510728,510539,509521,509585,510674,
            510591,509465,509597,510231,510673,509358,509340,509357,509359,
            510222];
    elif hash1 eq 2583 and hash2 eq 2093 and hash3 eq 1331 then
        V:=[25,1,26,7,27,15,28,27,29,33,30,43];
        D:=[1,39/2,1,47/2,1,0,3,18,1,59/3,1,85/4,2,0,7,71/4,1,107/6,1,109/6,1,
            37/2,1,121/6,1,0,12,44/3,2,16,1,0,15,173/12,1,147/10,1,59/4,2,
            491/30,2,0,21,179/12,1,0];
        I:=[241554,241448,241456,240397,241106,241559,240649,239057,241499,
            241610,241270,241483,241508,241189,239031,239061,239080,239082,
            238375,238916,238326];
    elif hash1 eq 1079 and hash2 eq 2921 and hash3 eq 3139 then
        V:=[22,1,23,11,24,23,25,35,26,45,27,59,28,67];
        D:=[1,21,1,23,1,74/3,3,25,4,0,10,127/6,3,139/6,1,47/2,1,151/6,2,51/2,2,
            0,19,69/4,1,52/3,1,53/3,8,59/3,1,22,1,0,31,71/4,1,107/6,2,109/6,4,
            119/6,2,0,40,44/3,1,191/12,4,16,1,49/3,4,215/12,1,56/3,3,0,54,
            431/30,1,173/12,1,89/6,2,0,58,66/5,1,0];
        I:=[249748,249577,249074,249282,249628,248998,249570,249592,249705,
            248235,248894,249284,248796,248702,248008,248921,248521,248864,
            249365,248716,248916,248999,249030,249268,249641,249721,249755,
            249773,249578,247905,247666,249024,249123,247974,248860,248861,
            249769,248510,249025,248565,247754,248020,249108,249144,248004,
            248022,248824,249135,249774,247522,247695,247859,248057,248552,
            247970,247954,248101,247873];
    elif hash1 eq 1220 and hash2 eq 2801 and hash3 eq 1529 then
        V:=[23,1,24,9,25,13,26,19,27,29,28,39,29,47,30,65];
        D:=[1,145/6,1,119/4,1,191/6,1,0,4,418/15,1,0,5,719/30,1,267/10,1,0,7,
            223/12,1,77/4,1,449/20,1,512/21,1,0,11,213/14,1,461/30,1,379/20,1,
            229/12,1,0,15,110/7,1,101/6,1,103/6,1,0,18,361/30,1,145/12,1,40/3,1,
            569/42,1,817/60,1,137/10,2,83/6,2,653/42,1,0,28,707/60,4,253/21,1,
            181/15,1,473/35,1,0];
        I:=[513604,513602,513603,513605,513470,513598,513455,513331,513493,
            513419,513443,513448,513500,513352,513427,513353,513354,513597,
            513618,513456,513446,513447,513630,513705,513379,513682,513363,
            513600,513616,513619,513639,513457,513454,513291];
    elif hash1 eq 45 and hash2 eq 1957 and hash3 eq 3139 then
        V:=[21,1,22,5,23,15,24,27,25,35,26,39,28,43];
        D:=[1,65/2,1,0,2,25,1,86/3,1,29,1,31,1,0,6,139/6,1,47/2,1,51/2,2,55/2,1,
            187/6,1,0,12,64/3,1,65/3,1,71/3,2,0,16,121/6,1,0,17,55/3,1,0,18,
            179/12,1,15,1,0];
        I:=[674013,673866,673865,673858,673854,673852,673867,673872,674014,
            674008,673774,673779,673855,673776,673859,673871,673778,673769,
            673781];
    elif hash1 eq 1737 and hash2 eq 2707 and hash3 eq 1063 then
        V:=[23,1,24,9,25,19,26,29,27,39,28,47,29,53];
        D:=[1,151/6,1,163/6,1,187/6,1,0,4,71/3,2,24,1,77/3,1,82/3,1,0,9,237/10,
            1,143/6,1,103/4,1,57/2,1,0,13,56/3,1,101/5,1,62/3,1,97/4,1,0,17,
            551/30,1,83/4,1,127/6,1,0,20,203/12,1,86/5,1,0,22,67/4,1,0];
        I:=[501787,501788,502025,501426,501719,501736,501524,501423,501735,
            501754,501474,501706,501847,501428,501959,501476,501765,501622,
            502026,502038,501460,501720];
    elif hash1 eq 2912 and hash2 eq 2297 and hash3 eq 2911 then
        V:=[16,1,17,5,18,17,19,23,20,29,21,35,22,43];
        D:=[1,36,1,0,2,61/2,1,65/2,1,69/2,2,77/2,1,81/2,1,0,8,29,1,92/3,1,0,10,
            163/6,1,55/2,1,0,12,76/3,2,77/3,3,0,17,197/10,1,119/6,1,329/12,1,0,
            20,263/12,1,67/3,1,0];
        I:=[425123,425238,425212,425108,425244,425241,425198,425206,425205,
            425199,425240,425209,425228,425126,425203,425216,424926,424934,
            424909,425031,425030];
    elif hash1 eq 938 and hash2 eq 1874 and hash3 eq 3255 then
        V:=[25,1,26,15,27,29,28,39,30,47,31,53];
        D:=[1,47/2,1,151/6,1,107/4,1,163/6,1,59/2,1,187/6,1,0,7,65/3,1,22,2,
            70/3,2,71/3,2,77/3,1,26,1,0,16,121/6,2,257/12,1,87/4,1,133/6,1,0,21,
            191/12,1,55/3,1,298/15,1,0,24,223/15,1,199/12,2,0,27,259/20,1,0];
        I:=[244796,244689,244134,245484,245560,245451,245418,243494,243597,
            244092,245651,244749,245429,245661,244684,244790,245634,244297,
            244135,244669,243430,244466,242236,244068,242113,242304,243424];
    elif hash1 eq 1737 and hash2 eq 2006 and hash3 eq 2966 then
        V:=[22,1,23,9,24,13,25,19,26,29,27,35,28,39,29,45,30,51];
        D:=[1,89/3,1,95/3,1,176/5,1,0,4,215/6,1,0,5,388/15,1,141/5,1,0,7,167/10,
            1,551/30,2,221/12,1,37/2,1,0,12,329/20,1,278/15,1,0,14,625/42,1,0,
            15,76/5,1,1027/60,1,0,17,263/20,1,797/60,1,0,19,1216/105,1,0];
        I:=[512500,512520,511793,512325,511902,511797,511132,512201,512324,
            512490,512497,512411,511194,511264,511355,511274,511195,512422,
            511277];
    elif hash1 eq 2160 and hash2 eq 621 and hash3 eq 3002 then
        V:=[25,1,26,5,27,13,28,23,29,29,30,43,31,49];
        D:=[1,55/2,1,0,2,64/3,1,24,1,77/3,1,0,5,121/6,1,213/10,1,131/6,1,237/10,
            1,0,9,56/3,1,298/15,1,0,11,59/4,1,89/6,1,197/12,1,539/30,1,551/30,1,
            37/2,1,0,17,175/12,1,44/3,2,0,20,27/2,2,0];
        I:=[244058,243748,243785,243622,243745,242549,242447,242356,243784,
            242442,243657,242689,242391,242357,242340,242339,243347,242394,
            242395,243682,243757];
    elif hash1 eq 1549 and hash2 eq 1661 and hash3 eq 317 then
        V:=[20,1,21,11,22,29,23,37,24,45,25,55,26,61,27,67];
        D:=[1,29,2,31,1,98/3,1,33,1,0,6,35/2,1,115/6,1,39/2,1,127/6,1,151/6,4,
            107/4,1,163/6,2,187/6,1,0,18,64/3,1,76/3,1,83/3,2,0,22,97/6,2,
            197/10,1,257/12,1,0,26,73/4,1,55/3,2,56/3,1,43/2,1,0,31,269/12,1,
            91/4,1,0,33,15,2,199/12,1,0,36,185/12,1,0];
        I:=[423998,424614,424591,424694,424689,424737,424738,424641,424277,
            423997,424639,424648,424660,423985,424276,424733,424014,424630,
            423979,423055,423992,424647,424720,422770,424199,423826,422918,
            424315,423928,423167,422434,422432,424635,424680,423379,422410];
    elif hash1 eq 609 and hash2 eq 709 and hash3 eq 569 then
        V:=[17,1,18,7,19,13,20,33,21,47,22,53,23,61,24,77,25,91];
        D:=[1,61/2,1,73/2,2,0,4,33,3,104/3,1,0,8,51/2,1,163/6,1,115/4,1,175/6,4,
            187/6,2,63/2,2,199/6,1,67/2,1,211/6,1,0,22,59/3,1,71/3,1,76/3,1,
            77/3,2,82/3,1,83/3,1,0,29,143/6,1,145/6,1,0,31,89/5,1,263/12,1,
            287/12,1,0,34,18,1,221/12,1,113/6,1,125/6,1,439/20,1,22,1,269/12,1,
            0,41,321/20,1,97/6,1,50/3,1,17,1,223/12,1,441/20,1,0,47,461/30,1,0];
        I:=[427370,427371,427413,427421,427442,427476,426987,427038,427050,
            427005,427126,427346,427350,427401,427034,427437,427345,427447,
            427403,427498,427203,427060,427039,427399,427061,427119,427186,
            427221,426678,426686,426847,426740,426544,427037,427059,427428,
            426680,426535,426533,426546,426698,426768,426741,427056,426511,
            426451,426716];
    elif hash1 eq 1643 and hash2 eq 2676 and hash3 eq 1571 then
        V:=[10,1,11,11,12,21,13,31,14,45,15,57,16,69,17,85,18,105,20,121,22,133,
            23,143];
        D:=[1,40,1,42,1,44,1,46,1,0,5,73/2,1,77/2,2,81/2,1,85/2,2,0,11,110/3,1,
            37,1,39,1,122/3,1,0,15,199/6,1,67/2,1,211/6,1,71/2,2,235/6,1,79/2,1,
            0,22,89/3,2,30,2,95/3,1,107/3,4,131/3,1,0,32,157/6,1,167/6,1,179/6,
            2,181/6,2,357/10,1,0,39,24,1,79/3,1,448/15,1,151/5,2,91/3,1,92/3,1,
            109/3,1,0,47,137/6,1,139/6,1,47/2,1,313/12,1,317/12,1,53/2,1,851/30,
            1,341/12,1,57/2,1,0,56,449/20,1,271/12,1,343/15,1,275/12,1,23,1,
            116/5,1,70/3,1,0,63,53/3,3,635/28,1,1363/60,1,137/6,1,139/6,1,0,70,
            337/21,1,241/15,1,967/60,1,81/5,1,0,74,127/10,1,0];
        I:=[430515,430514,430513,430516,430512,430505,430511,430497,430508,
            430510,430485,430506,430507,430477,430484,430496,430478,430503,
            430504,430490,430475,430476,430487,430492,430509,430473,430491,
            430493,430494,430498,430468,430474,430495,430480,430481,430482,
            430488,430458,430483,430486,430453,430457,430464,430466,430456,
            430465,430489,430499,430502,430470,430471,430472,430454,430459,
            430467,430461,430452,430460,430463,430451,430462,430455,430479,
            430500,430501,430448,430445,430450,430444,430447,430443,430446,
            430449,430469];
    elif hash1 eq 3382 and hash2 eq 1898 and hash3 eq 714 then
        V:=[25,1,26,11,27,19,28,25];
        D:=[1,35/2,1,115/6,3,39/2,2,127/6,1,0,8,47/3,1,16,2,53/3,1,0,12,95/6,3,
            97/6,1,0,16,14,1,43/3,1,44/3,1,81/5,1,0];
        I:=[240997,240546,240820,240824,239563,241045,239578,241201,240844,
            240856,240332,240450,240811,241399,240185,241407,239632,240857,
            238107];
    elif hash1 eq 3241 and hash2 eq 2376 and hash3 eq 283 then
        V:=[22,1,23,5,24,11,25,23,26,27,27,39,28,51,29,59];
        D:=[1,23,1,0,2,51/2,1,115/4,1,0,4,59/3,1,64/3,1,22,1,71/3,1,77/3,1,0,9,
            95/4,1,0,10,39/2,1,239/12,1,62/3,1,109/5,1,111/5,1,0,15,89/6,1,33/2,
            1,101/6,1,221/12,1,37/2,2,0,21,44/3,1,248/15,1,19,1,0,24,27/2,1,
            83/6,2,181/12,1,0];
        I:=[417724,417723,416330,417711,417634,417717,417600,417801,416301,
            416580,417850,417750,416296,416303,417816,416363,416294,417771,
            416005,417772,417793,416287,415546,417739,417768,417827,415548];
    elif hash1 eq 233 and hash2 eq 322 and hash3 eq 766 then
        V:=[17,1,18,5,19,9,20,13,21,21,22,27,23,33,25,37,26,43];
        D:=[1,37/2,1,0,2,19,1,0,3,103/6,1,0,4,22,1,28,1,34,1,0,7,133/6,1,337/10,
            1,0,9,16,1,61/3,1,0,11,551/30,1,0,12,877/60,2,865/42,1,0,15,452/35,
            1,0];
        I:=[543286,543280,543283,542866,542880,542860,542771,543016,543346,
            543076,542789,543227,543229,542578,542843];
    elif hash1 eq 1596 and hash2 eq 2426 and hash3 eq 1179 then
        V:=[24,1,25,9,26,13,27,21,28,25,29,29,30,35,31,43];
        D:=[1,100/3,1,141/4,1,106/3,1,0,4,167/6,1,0,5,358/15,1,73/3,1,85/3,1,0,
            8,293/12,1,0,9,313/15,1,0,10,511/30,1,205/12,1,0,12,47/3,1,1027/60,
            1,268/15,1,0,15,653/42,1,467/30,1,0];
        I:=[410524,409498,410914,411258,410172,410511,410894,409127,408943,
            409997,410505,410854,409126,411037,408973,408972];
    elif hash1 eq 2724 and hash2 eq 1162 and hash3 eq 2760 then
        V:=[16,1,17,5,19,9,20,17,21,23,22,37,23,41,24,47,26,63];
        D:=[1,37,1,0,2,271/6,1,0,3,317/10,1,193/6,1,397/10,1,0,6,161/5,1,97/3,2,
            0,9,731/30,1,293/12,1,49/2,1,149/6,1,791/30,1,53/2,1,0,15,136/5,1,0,
            16,571/30,1,581/30,1,0,18,531/28,1,400/21,1,1147/60,1,407/21,1,
            293/15,1,332/15,1,89/4,1,0,25,295/21,1,211/15,1,71/5,1,0];
        I:=[427843,427841,427846,427845,427652,427844,427809,427866,427812,
            427838,427803,427802,427837,427847,427626,427818,427880,427643,
            427646,427642,427619,427622,427647,427656,427649,427660,427807];
    elif hash1 eq 1220 and hash2 eq 3304 and hash3 eq 1251 then
        V:=[16,1,17,7,18,13,19,19,20,29,21,41,22,53,23,59,24,63,25,67,26,73,27,
            77];
        D:=[1,34,1,40,1,0,3,65/2,1,73/2,1,0,5,104/3,1,45,1,0,7,55/2,1,211/6,1,
            71/2,1,223/6,1,0,11,30,1,133/4,1,34,1,107/3,1,112/3,1,0,16,353/12,1,
            179/6,1,401/12,1,203/6,1,205/6,1,0,21,131/5,1,59/2,1,0,23,45/2,1,0,
            24,441/20,1,0,25,35/2,1,221/10,1,0,27,86/5,1,0,28,63/4,1,0];
        I:=[516002,515974,516000,516003,516120,516336,515989,516132,516287,
            515910,516285,516228,516338,515917,516333,515940,515956,516198,
            515916,516326,515870,516298,515957,515850,515862,515754,515846,
            515758];
    elif hash1 eq 1878 and hash2 eq 769 and hash3 eq 577 then
        V:=[22,1,23,7,24,13,25,19,26,23,27,33,28,43,29,57];
        D:=[1,58/3,1,22,1,0,3,121/6,1,131/6,3,0,7,298/15,1,263/12,1,0,9,279/14,
            1,0,10,179/12,1,15,1,47/3,1,82/5,1,0,14,91/6,1,461/30,2,697/42,1,
            997/60,1,0,19,79/6,1,89/6,1,227/15,1,76/5,1,61/4,2,73/4,1,0,26,
            371/30,1,0];
        I:=[510643,510655,510621,509675,510596,510650,509636,508890,508849,
            509654,510658,510240,510220,509255,509501,509750,508755,509764,
            509518,508426,508425,509284,509259,509285,508299,509728];
    elif hash1 eq 1549 and hash2 eq 1172 and hash3 eq 1339 then
        V:=[24,1,25,5,26,11,27,21,28,31,29,43,30,49,32,59];
        D:=[1,83/3,1,0,2,131/6,1,145/6,1,0,4,268/15,1,239/12,1,81/4,1,328/15,1,
            0,8,491/30,1,33/2,1,359/20,1,241/12,1,0,12,179/12,1,199/12,1,17,1,
            52/3,1,18,1,0,17,451/30,1,181/12,1,0,19,163/12,1,41/3,1,233/15,1,
            47/3,1,0,23,727/60,1,0];
        I:=[417061,417165,417287,417135,417124,415747,417019,417558,415859,
            417045,415286,417199,415869,415679,417105,415030,417048,417096,
            416958,415421,414947,414953,414914];
    elif hash1 eq 1878 and hash2 eq 2775 and hash3 eq 1012 then
        V:=[23,1,24,5,25,9,26,17,27,23,28,31];
        D:=[1,59/2,1,0,2,65/3,1,0,3,87/4,1,45/2,1,145/6,1,0,6,49/3,2,55/3,2,0,
            10,101/6,3,75/4,1,113/6,3,0,17,223/15,1,15,1,0];
        I:=[414553,414472,413697,414100,414604,414332,414343,414277,414548,
            414090,414633,414723,413684,413683,413901,414104,414382,414381];
    elif hash1 eq 891 and hash2 eq 64 and hash3 eq 1244 then
        V:=[20,1,21,9,22,17,24,25,25,31];
        D:=[1,22,1,24,1,26,2,0,5,41/2,2,45/2,8,49/2,3,0,18,56/3,1,62/3,1,21,1,0,
            21,52/3,1,53/3,1,0,23,29/2,1,0];
        I:=[674146,674149,674080,674147,674048,674150,674054,674073,674082,
            674083,674100,674124,674125,674148,673882,673883,673891,674063,
            673878,673920,673880,673984,673904];
    elif hash1 eq 1643 and hash2 eq 3253 and hash3 eq 1436 then
        V:=[31,1,33,7,34,15,35,31];
        D:=[1,95/6,1,109/6,1,0,3,431/30,1,173/12,1,89/6,1,0,6,151/12,1,155/12,1,
            13,1,66/5,1,40/3,1,217/15,2,223/15,2,0,15,23/2,1,71/6,1,0];
        I:=[658051,655965,658188,658281,655837,656419,657951,656006,656914,
            658134,655232,655844,655984,656037,655013,658018];
    elif hash1 eq 1596 and hash2 eq 3062 and hash3 eq 408 then
        V:=[17,1,18,11,19,21,20,31,21,49,22,59,23,65,25,75];
        D:=[1,65/2,2,69/2,1,73/2,1,77/2,1,0,6,29,1,31,1,33,2,104/3,1,0,11,175/6,
            3,59/2,2,187/6,1,71/2,1,0,18,71/3,1,101/4,1,76/3,1,77/3,1,26,1,83/3,
            3,88/3,1,89/3,1,0,28,257/12,1,87/4,1,133/6,1,143/6,1,0,32,67/3,1,
            68/3,4,0,37,217/12,1,221/12,1,45/2,1,137/6,1,0,41,299/20,1,451/30,1,
            181/12,1,185/12,1,31/2,1,0];
        I:=[254596,254606,254496,254600,254492,254595,254491,254422,254605,
            254437,254436,254484,254495,254424,254490,254497,254415,254601,
            254433,254438,254418,254487,254419,254439,254486,254369,254375,
            254411,254435,254410,254420,254383,254372,254413,254427,254441,
            254483,254434,254367,254368,254408,254371,254378,254376,254377];
    elif hash1 eq 2066 and hash2 eq 2652 and hash3 eq 1626 then
        V:=[19,1,20,11,21,23,22,37,23,53,24,59,25,65];
        D:=[1,45/2,2,49/2,2,53/2,2,57/2,2,0,9,21,3,68/3,2,23,4,74/3,3,25,1,0,22,
            115/6,1,39/2,5,127/6,5,43/2,4,139/6,1,47/2,2,0,40,18,2,77/4,1,58/3,
            1,59/3,3,20,1,106/5,1,65/3,3,0,52,107/6,5,257/12,1,0,58,49/3,2,50/3,
            1,0,61,29/2,2,0];
        I:=[253435,253452,253199,253441,253200,253425,253271,253318,253374,
            253432,253442,253300,253317,253237,253265,253400,253443,252956,
            253075,253117,253080,253259,252885,253242,253260,253269,253390,
            252964,253071,253119,253313,253316,253064,253084,253203,253328,
            252882,252744,252938,252944,253124,252945,253332,252615,253087,
            253140,253105,252662,252726,252758,253040,252729,252937,252940,
            253086,253407,252671,252749,253275,252724,253114,253118];
    elif hash1 eq 750 and hash2 eq 770 and hash3 eq 2010 then
        V:=[25,1,26,11,27,21,28,29];
        D:=[1,31/2,1,127/6,1,139/6,1,47/2,1,0,5,47/3,1,16,2,96/5,1,59/3,1,0,10,
            85/6,1,157/10,1,63/4,1,0,13,208/15,1,14,2,57/4,1,43/3,2,191/12,1,0];
        I:=[408691,405529,405526,406914,408678,407957,408575,405527,408462,
            408344,407199,406342,407978,408343,408700,407371,408587,408747,
            407837];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// A cache of ids with high hash collision. This handles the case when
// hash1 eq 46 mod 47.
function cached_ids_46(P,hash1,hash2,hash3)
    if hash1 eq 3289 and hash2 eq 2419 and hash3 eq 213 then
        V:=[24,1,25,5,26,9,27,19,28,29,29,35];
        D:=[1,65/3,1,0,2,131/6,1,0,3,50/3,1,101/5,1,61/3,1,62/3,1,0,7,197/12,1,
            167/10,1,101/6,1,103/6,3,0,13,44/3,1,46/3,1,0,15,161/12,1,27/2,1,0];
        I:=[413411,413412,413413,413370,413409,413403,413393,413355,413402,
            413351,413415,413417,413354,413373,413377,413357];
    elif hash1 eq 2678 and hash2 eq 682 and hash3 eq 1325 then
        V:=[22,1,23,5,24,11,25,21,26,27,27,35];
        D:=[1,29,1,0,2,51/2,1,55/2,2,0,5,93/4,1,71/3,2,76/3,1,77/3,1,0,10,131/6,
            1,145/6,1,0,12,55/3,2,239/12,1,62/3,1,0,16,33/2,1,101/6,1,0];
        I:=[414547,414537,414538,414555,414543,414560,414578,414059,414605,
            414608,414128,414591,414601,413838,414166,414594,414597];
    elif hash1 eq 375 and hash2 eq 2672 and hash3 eq 3292 then
        V:=[23,1,24,11,25,15,26,29,27,43,28,59,29,69];
        D:=[1,127/6,2,43/2,1,139/6,2,163/6,1,0,7,59/3,4,0,11,95/6,1,97/6,2,33/2,
            1,109/6,1,79/4,1,119/6,1,0,18,44/3,2,16,1,81/5,2,49/3,3,50/3,1,
            268/15,1,0,28,169/12,1,431/30,1,173/12,1,29/2,2,89/6,7,167/10,1,
            101/6,1,0,42,40/3,1,41/3,1,223/15,1,15,1,0,46,71/6,2,0];
        I:=[414775,414776,414587,414624,414709,414122,414120,414252,414715,
            414791,414701,414595,414716,414519,414307,414471,414750,413905,
            414743,414718,414326,414397,414628,414696,414722,414129,414321,
            414720,413899,414164,414689,414770,414037,414071,414074,414515,
            414579,414582,414584,413507,413521,413636,413973,413471,413496,
            414317,414325];
    elif hash1 eq 2161 and hash2 eq 1675 and hash3 eq 401 then
        V:=[22,1,23,7,24,15,25,33,26,45,27,59,28,67];
        D:=[1,19,1,21,1,0,3,103/6,1,47/2,1,163/6,1,0,6,47/3,1,52/3,1,53/3,1,18,
            1,77/4,2,58/3,2,59/3,2,65/3,2,0,18,177/10,1,71/4,1,109/6,2,119/6,1,
            121/6,1,0,24,44/3,1,16,1,65/4,1,49/3,2,262/15,1,35/2,1,0,31,29/2,2,
            89/6,2,193/12,1,0,36,66/5,1,0];
        I:=[418765,418276,418806,416935,418620,418807,418260,418826,418824,
            416592,418415,418279,418548,418892,418897,417575,418895,416927,
            415687,418037,418046,417014,418239,417996,415574,416599,416824,
            418238,416402,417101,418019,418319,417471,417915,416995,416903];
    elif hash1 eq 1785 and hash2 eq 2887 and hash3 eq 1618 then
        V:=[20,1,21,5,22,11,23,15,26,23];
        D:=[1,107/3,2,0,3,179/6,1,181/6,2,0,6,79/3,1,0,7,439/20,1,671/30,1,
            269/12,2,0,11,76/5,1,61/4,1,53/3,3,0];
        I:=[427327,427331,427171,426990,427115,427133,426590,426595,426586,
            426588,426638,426809,426522,426530,426571];
    elif hash1 eq 610 and hash2 eq 2606 and hash3 eq 1160 then
        V:=[23,1,24,13,25,23,26,43,27,49,28,57,29,71,30,77];
        D:=[1,47/2,1,151/6,1,51/2,1,163/6,1,175/6,1,0,6,64/3,1,65/3,2,71/3,1,
            76/3,1,0,11,109/6,1,37/2,1,119/6,1,121/6,1,41/2,1,87/4,3,131/6,1,
            237/10,1,143/6,1,0,22,49/3,1,17,1,0,24,33/2,1,183/10,1,221/12,2,0,
            28,217/15,1,175/12,2,15,1,76/5,1,46/3,3,248/15,1,0,37,157/12,1,79/6,
            1,0,39,139/12,1,0];
        I:=[249395,249696,249390,248909,249727,248872,249697,249699,248443,
            248751,248221,248906,248938,248863,248956,247673,247920,248052,
            248270,247996,248012,249237,248929,248217,247519,247653,248056,
            248126,247720,248952,249427,248332,247788,247821,248987,247633,
            247915,248011,248060];
    elif hash1 eq 1879 and hash2 eq 210 and hash3 eq 1210 then
        V:=[25,1,26,9,27,17,28,27,29,33,30,39,31,43];
        D:=[1,175/6,1,59/2,1,63/2,1,0,4,65/3,1,71/3,1,77/3,5,0,11,119/6,3,
            217/10,1,131/6,4,143/6,1,0,20,18,1,20,1,0,22,217/12,1,221/12,2,0,25,
            179/12,2,0,27,451/30,1,0];
        I:=[671746,671762,672164,671668,672208,671186,671300,671670,671775,
            672178,671605,671993,672168,671256,671289,671712,671768,672102,
            671294,671988,671184,671516,671160,671631,671270,671592,671104];
    elif hash1 eq 2772 and hash2 eq 2315 and hash3 eq 524 then
        V:=[24,1,25,7,26,15,27,19];
        D:=[1,17,1,21,1,0,3,103/6,6,35/2,1,115/6,2,0,12,53/3,1,0,13,95/6,2,
            107/6,1,0];
        I:=[245778,243962,245000,245003,245763,245764,245811,245817,245193,
            244475,245477,243253,243978,245128,241874];
    elif hash1 eq 2772 and hash2 eq 1677 and hash3 eq 1996 then
        V:=[24,1,25,11,26,17,29,23,30,31];
        D:=[1,27,1,29,1,92/3,2,31,1,0,6,175/6,1,187/6,2,0,9,70/3,1,71/3,2,0,12,
            101/6,1,113/6,1,115/6,1,0,15,199/12,2,0];
        I:=[397006,397035,397040,397936,398167,398261,397096,398133,398255,
            398165,398259,398037,398117,398279,397813,398065];
    elif hash1 eq 3101 and hash2 eq 10 and hash3 eq 1650 then
        V:=[22,1,23,7,24,13,25,21,26,27,27,33,28,45];
        D:=[1,27,1,86/3,1,0,3,163/6,1,59/2,1,0,5,59/3,1,64/3,1,76/3,1,0,8,
            217/10,1,133/6,2,0,11,81/5,1,73/4,1,0,13,221/12,1,37/2,1,187/10,1,
            113/6,2,115/6,1,0,19,218/15,1,0];
        I:=[418102,418103,418065,417230,416445,416616,418139,415821,418124,
            418144,417531,416160,415800,415225,415816,415884,417226,416632,
            417543];
    elif hash1 eq 1503 and hash2 eq 798 and hash3 eq 44 then
        V:=[24,1,25,7,26,23,27,27,28,35,29,45,30,49];
        D:=[1,25,1,80/3,1,0,3,39/2,1,127/6,1,43/2,1,91/4,1,139/6,2,51/2,1,107/4,
            1,0,11,58/3,1,0,12,177/10,1,109/6,1,41/2,1,0,15,44/3,1,16,1,215/12,
            2,322/15,1,0,20,491/30,1,0,21,41/3,1,0];
        I:=[244611,243493,244080,243721,244141,243116,243499,243573,244085,
            243457,243574,243496,244503,244607,244506,243543,244491,244512,
            241830,242172,244408];
    elif hash1 eq 469 and hash2 eq 2770 and hash3 eq 1636 then
        V:=[21,1,22,9,23,17,24,25,25,29,26,37];
        D:=[1,139/6,1,47/2,1,51/2,1,0,4,65/3,2,93/4,1,71/3,2,0,9,109/6,1,119/6,
            2,121/6,2,0,14,55/3,2,0,16,197/12,1,33/2,1,101/6,1,0,19,15,1,46/3,1,
            0];
        I:=[252302,252462,252467,252305,252418,252116,252068,252107,252322,
            252249,252408,252100,252235,252264,252412,252228,252251,252229,
            251886,251890];
    elif hash1 eq 2725 and hash2 eq 2356 and hash3 eq 2376 then
        V:=[27,1,28,5,29,13,30,25,31,31,32,39];
        D:=[1,127/6,1,0,2,47/3,2,53/3,4,59/3,1,0,9,29/2,1,63/4,1,95/6,3,97/6,1,
            33/2,1,0,16,43/3,2,44/3,3,0,21,149/12,1,25/2,1,77/6,5,0,28,34/3,2,
            0];
        I:=[228121,228161,228536,227028,228146,228438,228611,227474,226466,
            227466,227027,228160,228587,227242,228228,226211,227176,227126,
            227230,227231,228089,228072,226597,226952,226966,228501,228521,
            226496,226554];
    elif hash1 eq 1644 and hash2 eq 1717 and hash3 eq 270 then
        V:=[10,1,12,11,14,21,16,31,18,39];
        D:=[1,30,1,38,1,42,1,50,1,0,5,27,2,31,1,116/3,1,41,1,0,10,101/4,1,77/3,
            1,109/4,1,113/3,1,0,14,111/5,1,68/3,1,85/3,1,0,17,47/3,1,283/15,1,
            247/12,1,295/12,1,0];
        I:=[544231,544232,544218,544230,544229,544233,544207,544219,544202,
            544198,544201,544199,544200,544177,544180,544135,544204,544176,
            544132,544133];
    elif hash1 eq 1785 and hash2 eq 2465 and hash3 eq 963 then
        V:=[22,1,23,7,24,15,27,19,28,23,29,35];
        D:=[1,89/3,1,101/3,1,0,3,257/10,1,169/6,1,179/6,3,0,8,24,1,0,9,107/6,1,
            0,10,883/60,1,907/60,1,76/5,2,61/4,2,47/3,1,0,17,1247/84,1,893/60,1,
            76/5,1,913/60,1,641/42,1,46/3,1,0];
        I:=[423677,424410,423308,423996,424570,424677,424770,424760,424827,
            422505,422684,423064,423602,423495,423591,424333,422484,422354,
            422504,422485,422493,422488];
    elif hash1 eq 3195 and hash2 eq 897 and hash3 eq 2894 then
        V:=[21,1,22,7,23,11,24,21,25,27,26,31];
        D:=[1,39/2,1,127/6,1,0,3,58/3,1,0,4,71/4,2,109/6,1,197/10,1,119/6,1,0,9,
            50/3,2,55/3,2,0,13,91/6,1,0,14,50/3,1,253/15,1,109/6,1,0];
        I:=[510553,510733,510744,509643,509722,510613,509516,509415,510289,
            510552,510554,510657,509572,508752,509494,508453];
    elif hash1 eq 328 and hash2 eq 462 and hash3 eq 955 then
        V:=[20,1,21,7,22,15,23,29,24,37,25,47];
        D:=[1,25,1,27,1,0,3,139/6,1,151/6,3,55/2,1,0,8,59/3,2,20,1,85/4,1,65/3,
            1,22,1,70/3,1,0,15,107/6,1,87/4,1,131/6,1,0,18,16,1,49/3,1,55/3,1,
            67/3,1,0,22,103/6,1,0];
        I:=[422207,422197,422196,421893,422258,422291,421642,422075,422216,
            422201,421886,422225,422252,421869,421609,421562,421705,422110,
            422094,421281,421161,421258];
    elif hash1 eq 187 and hash2 eq 2563 and hash3 eq 2451 then
        V:=[22,1,23,9,24,17,26,23,27,29,28,35,29,43,30,51];
        D:=[1,83/3,1,94/3,1,106/3,1,0,4,197/10,1,95/4,1,317/10,1,0,7,358/15,1,
            311/12,1,0,9,223/12,1,64/3,1,0,11,625/42,1,181/12,1,0,13,310/21,1,
            89/6,1,103/6,1,0,16,1103/84,1,66/5,1,137/10,1,0,19,707/60,1,0];
        I:=[512152,512241,512322,511134,511320,512319,512198,511361,511119,
            512182,511131,510860,510884,510859,511275,511166,511247,512384,
            512195];
    elif hash1 eq 3195 and hash2 eq 71 and hash3 eq 1550 then
        V:=[26,1,27,5,28,11,29,21,30,25];
        D:=[1,19,1,0,2,115/6,2,127/6,2,0,6,52/3,1,53/3,3,58/3,1,59/3,1,0,12,
            97/6,1,0,13,167/12,1,71/5,1,57/4,1,43/3,3,81/5,2,0];
        I:=[667302,667103,667522,667078,667584,667632,666197,667108,667249,
            667583,666587,666653,666051,666103,665946,667564,667692,667693,
            665347,665356];
    elif hash1 eq 845 and hash2 eq 630 and hash3 eq 187 then
        V:=[23,1,24,9,25,19,26,27,27,35,28,43,29,57,30,65];
        D:=[1,139/6,2,51/2,1,55/2,1,0,5,20,1,64/3,2,65/3,2,71/3,1,0,11,119/6,1,
            121/6,1,131/6,1,0,14,215/12,1,18,1,73/4,1,0,17,89/6,1,193/12,1,33/2,
            1,0,20,40/3,1,175/12,1,223/15,3,179/12,3,15,5,46/3,2,0,35,391/30,1,
            181/12,1,91/6,1,0,38,139/12,2,0];
        I:=[249685,249763,249143,249312,249327,249556,249708,249659,249726,
            249398,248441,249049,247851,247951,247753,247678,248574,248035,
            249345,248494,247949,248204,248218,248219,247661,248153,249418,
            247660,248193,249232,249682,249684,247815,248336,248127,247568,
            247548,247802,248059];
    elif hash1 eq 2208 and hash2 eq 1514 and hash3 eq 2549 then
        V:=[19,1,20,13,21,23,22,29,23,33,24,37];
        D:=[1,49/2,2,53/2,2,57/2,1,61/2,2,69/2,1,0,9,86/3,1,29,2,92/3,1,31,2,0,
            15,127/6,1,139/6,2,0,18,71/3,2,0,20,131/6,1,0,21,81/5,1,0];
        I:=[253845,253945,253846,253926,253609,253831,253962,253959,253615,
            253774,253956,253762,253740,253960,253830,253624,253818,253477,
            253570,253589,253571];
    elif hash1 eq 1503 and hash2 eq 2583 and hash3 eq 2196 then
        V:=[20,1,22,7,23,19,24,27,25,39,26,43,29,55];
        D:=[1,92/3,1,31,1,0,3,126/5,1,77/3,1,26,1,82/3,1,88/3,1,0,8,143/6,1,
            145/6,1,157/6,3,0,13,91/5,1,298/15,1,263/12,1,111/5,1,67/3,2,0,19,
            551/30,1,0,20,175/12,1,199/12,1,283/15,1,227/12,3,441/20,1,0,27,
            697/60,2,0];
        I:=[423776,424057,422802,422997,424021,423772,424241,424258,423206,
            422733,423830,423986,422816,422737,422788,422807,424065,424066,
            422732,423062,423239,422786,422639,422811,422818,422338,422923,
            423362];
    elif hash1 eq 1268 and hash2 eq 1402 and hash3 eq 624 then
        V:=[25,1,26,5,27,19,28,27,29,35,30,47,31,63];
        D:=[1,95/4,1,0,2,50/3,1,55/3,1,239/12,1,101/5,1,287/12,1,121/5,1,0,8,
            91/6,1,37/2,1,187/10,1,0,11,46/3,1,62/3,1,313/15,1,0,14,401/30,1,
            27/2,2,83/6,1,50/3,1,781/42,1,0,20,178/15,1,66/5,2,281/21,1,163/12,
            1,41/3,1,71/5,1,316/21,1,0,28,697/60,2,499/42,1,361/30,1,27/2,1,0];
        I:=[508068,508081,508089,508082,508046,508072,508142,508169,508071,
            508141,507776,507944,508101,508172,508158,508164,508094,507805,
            507904,508099,508120,508151,507906,507901,508069,508074,507825,
            508078,508161,507777,507775,507719];
    elif hash1 eq 3289 and hash2 eq 2323 and hash3 eq 1328 then
        V:=[23,1,24,5,25,11,26,19,27,25,28,37,29,41,30,49];
        D:=[1,163/6,1,0,2,65/3,1,71/3,3,0,6,121/6,2,131/6,1,145/6,1,0,10,18,1,
            55/3,1,0,12,197/12,2,167/10,2,67/4,1,101/6,1,103/6,1,0,19,175/12,1,
            0,20,401/30,1,161/12,1,27/2,1,0,23,35/3,1,0];
        I:=[247091,247115,246572,247078,247202,246595,247077,246806,247178,
            246566,246748,247167,247174,246579,246625,247094,246597,246621,
            246656,246630,246836,246837,246600];
    elif hash1 eq 1080 and hash2 eq 2238 and hash3 eq 1242 then
        V:=[28,1,30,7,31,15,32,23,33,29];
        D:=[1,58/3,1,71/3,1,0,3,191/12,1,55/3,2,101/5,1,0,7,29/2,1,193/12,1,
            197/12,1,0,10,155/12,1,53/4,1,0,12,331/30,1,341/30,2,23/2,1,0];
        I:=[392021,392113,393143,392023,392232,392709,393093,393144,393006,
            393157,393159,393019,392747,393174,392059];
    elif hash1 eq 2584 and hash2 eq 1382 and hash3 eq 2168 then
        V:=[17,1,18,7,19,19,20,31,21,39,22,55,23,65,24,81,25,85];
        D:=[1,61/2,2,69/2,3,0,6,80/3,1,29,2,92/3,1,98/3,1,33,1,0,12,151/6,2,
            163/6,1,175/6,3,59/2,1,63/2,3,0,22,71/3,3,77/3,10,89/3,1,0,36,
            217/10,1,131/6,5,133/6,1,237/10,1,143/6,1,145/6,1,167/6,1,0,47,20,1,
            22,1,89/4,1,67/3,1,0,51,18,1,217/12,1,551/30,3,221/12,3,37/2,4,
            187/10,1,113/6,6,0,70,223/12,1,0,71,451/30,1,0];
        I:=[254503,254607,254526,254537,254598,254500,254555,254579,254564,
            254421,254494,254502,254566,254520,254423,254479,254549,254599,
            254509,254543,254573,254513,254558,254589,254398,254463,254477,
            254493,254501,254507,254517,254547,254571,254577,254374,254394,
            254403,254476,254478,254581,254588,254471,254402,254448,254401,
            254370,254432,254385,254388,254382,254447,254469,254387,254397,
            254407,254404,254460,254472,254459,254464,254489,254602,254395,
            254386,254396,254400,254443,254468,254562,254365,254384];
    elif hash1 eq 939 and hash2 eq 1183 and hash3 eq 1816 then
        V:=[19,1,20,5,21,17,22,33,23,51,24,65,25,75,26,89,27,97,28,103];
        D:=[1,65/2,1,0,2,80/3,1,27,2,86/3,1,29,2,33,1,0,9,127/6,2,43/2,1,139/6,
            6,151/6,3,51/2,2,107/4,1,187/6,1,0,25,58/3,1,59/3,2,20,1,65/3,3,22,
            1,70/3,1,71/3,4,101/4,1,0,39,209/12,2,197/10,2,119/6,1,217/10,1,
            131/6,1,133/6,1,0,47,49/3,1,268/15,2,61/3,1,62/3,3,0,54,59/4,2,89/6,
            2,16,1,33/2,1,187/10,1,113/6,1,0,62,218/15,1,175/12,1,61/4,1,0,65,
            625/42,1,31/2,1,0,67,178/15,1,0];
        I:=[424098,424321,423795,423835,424322,423797,424028,423807,424044,
            424130,424325,423199,424003,424024,424029,424056,424324,424042,
            424154,424323,423401,423760,423181,422846,423851,423135,424313,
            424326,422844,423766,424052,424253,423387,422729,422837,424235,
            424477,422835,424084,424085,423071,423315,423132,422838,423839,
            423397,423154,422901,423086,423087,422936,423041,424250,424243,
            424445,423024,424091,422632,423000,423184,423204,423036,424072,
            422796,422403,422437,423176];
    elif hash1 eq 845 and hash2 eq 1708 and hash3 eq 2028 then
        V:=[14,1,15,7,16,11,17,15,18,23,19,31,20,37,21,43,22,51,23,59];
        D:=[1,42,1,46,1,0,3,81/2,1,0,4,37,1,0,5,199/6,1,67/2,1,223/6,1,0,8,
            117/4,1,94/3,1,106/3,2,0,12,79/4,1,425/12,1,0,14,22,1,73/3,1,0,16,
            539/30,1,217/12,1,293/12,1,0,19,37/2,1,449/20,1,45/2,1,0,22,973/60,
            1,451/20,1,0];
        I:=[429005,428988,428990,428999,428880,428902,428877,428901,428983,
            428869,428875,428903,428725,428702,428706,428730,428733,428639,
            428904,428636,428634,428714,428602];
    elif hash1 eq 516 and hash2 eq 1548 and hash3 eq 2966 then
        V:=[24,1,25,11,26,17,27,27,28,31,29,39];
        D:=[1,287/12,1,311/12,1,105/4,1,85/3,1,0,5,41/2,1,269/12,1,0,7,227/12,1,
            302/15,1,308/15,1,247/12,1,0,11,511/30,1,0,12,233/15,1,238/15,1,16,
            1,0,15,817/60,1,481/30,1,0];
        I:=[421442,421463,421455,421456,421451,421453,421457,421488,421364,
            421548,421584,421359,421351,421458,421156,421118];
    elif hash1 eq 2490 and hash2 eq 2877 and hash3 eq 1574 then
        V:=[14,1,17,5,18,11,19,15,20,21,22,25,23,33,25,41,26,49];
        D:=[1,55,1,0,2,473/12,1,397/10,1,0,4,171/5,1,0,5,55/2,1,799/20,1,0,7,
            680/21,1,0,8,587/28,1,64/3,1,362/15,1,0,11,433/28,1,49/3,1,1583/84,
            1,0,14,295/18,1,493/30,1,265/14,1,0,17,1321/105,1,553/40,2,1034/63,
            1,1381/84,1,0];
        I:=[543993,544012,543965,543983,543906,543933,543944,543940,543885,
            543897,543899,543901,543917,543889,544038,543890,543926,543898,
            543914,544036,544037];
    else
        return false,_;
    end if;
    bool,ids:=unpack_ids(P,V,D,I);
    if bool then
        return true,ids;
    else
        return false,_;
    end if;
end function;

// Given the hash key for a canonical Fano 3-tope, returns the block id, and the
// range of lines you need to consult in the block file.
function block_and_offset_data(hash1)
    // The reverse block to use
    block:=hash1 mod 135;
    // Recover the offset data
    offset:=case <block |
        0: [Integers() | 0,203,135,175,270,274,405,174,540,172,675,235,810,216,
            945,213,1080,224,1215,211,1350,192,1485,267,1620,184,1755,179,1890,
            167,2025,148,2160,222,2295,177,2430,222,2565,197,2700,166,2835,284,
            2970,161,3105,169,3240,206,3375,191],
        1: [Integers() | 1,191,136,135,271,179,406,171,541,168,676,147,811,254,
            946,217,1081,227,1216,227,1351,251,1486,192,1621,152,1756,228,1891,
            176,2026,251,2161,222,2296,157,2431,166,2566,181,2701,230,2836,193,
            2971,196,3106,200,3241,205,3376,203],
        2: [Integers() | 2,174,137,167,272,235,407,208,542,194,677,164,812,200,
            947,182,1082,250,1217,190,1352,249,1487,181,1622,206,1757,200,1892,
            219,2027,179,2162,208,2297,206,2432,189,2567,202,2702,154,2837,204,
            2972,214,3107,190,3242,205,3377,200],
        3: [Integers() | 3,159,138,149,273,175,408,190,543,186,678,249,813,188,
            948,177,1083,206,1218,189,1353,186,1488,167,1623,194,1758,167,1893,
            198,2028,200,2163,170,2298,192,2433,233,2568,176,2703,242,2838,175,
            2973,203,3108,178,3243,190,3378,183],
        4: [Integers() | 4,200,139,188,274,241,409,191,544,226,679,185,814,172,
            949,171,1084,174,1219,180,1354,230,1489,221,1624,182,1759,216,1894,
            202,2029,184,2164,167,2299,202,2434,172,2569,224,2704,247,2839,181,
            2974,202,3109,190,3244,181,3379,195],
        5: [Integers() | 5,179,140,176,275,186,410,203,545,215,680,166,815,206,
            950,194,1085,214,1220,249,1355,201,1490,196,1625,194,1760,168,1895,
            188,2030,192,2165,166,2300,171,2435,218,2570,188,2705,314,2840,181,
            2975,219,3110,167,3245,160,3380,174],
        6: [Integers() | 6,191,141,322,276,232,411,218,546,237,681,181,816,185,
            951,187,1086,155,1221,183,1356,163,1491,197,1626,160,1761,192,1896,
            199,2031,182,2166,239,2301,184,2436,173,2571,236,2706,176,2841,209,
            2976,213,3111,189,3246,225,3381,192],
        7: [Integers() | 7,184,142,165,277,174,412,137,547,180,682,177,817,231,
            952,153,1087,228,1222,164,1357,202,1492,180,1627,155,1762,196,1897,
            163,2032,153,2167,194,2302,178,2437,262,2572,183,2707,202,2842,201,
            2977,233,3112,180,3247,193,3382,213],
        8: [Integers() | 8,170,143,216,278,213,413,182,548,205,683,189,818,202,
            953,160,1088,235,1223,175,1358,193,1493,162,1628,216,1763,189,1898,
            154,2033,220,2168,176,2303,170,2438,212,2573,272,2708,212,2843,246,
            2978,200,3113,154,3248,208,3383,213],
        9: [Integers() | 9,198,144,246,279,169,414,219,549,244,684,174,819,166,
            954,233,1089,192,1224,194,1359,221,1494,161,1629,218,1764,230,1899,
            212,2034,173,2169,210,2304,207,2439,203,2574,158,2709,256,2844,191,
            2979,183,3114,174,3249,184,3384,202],
        10: [Integers() | 10,309,145,204,280,253,415,201,550,221,685,208,820,
            206,955,172,1090,175,1225,152,1360,157,1495,188,1630,186,1765,212,
            1900,180,2035,171,2170,216,2305,169,2440,222,2575,201,2710,263,2845,
            192,2980,223,3115,161,3250,160,3385,186],
        11: [Integers() | 11,182,146,174,281,194,416,189,551,168,686,186,821,
            182,956,209,1091,196,1226,165,1361,156,1496,191,1631,192,1766,203,
            1901,161,2036,204,2171,180,2306,179,2441,192,2576,173,2711,155,2846,
            164,2981,193,3116,217,3251,173,3386,203],
        12: [Integers() | 12,186,147,175,282,198,417,186,552,182,687,176,822,
            203,957,184,1092,193,1227,190,1362,171,1497,191,1632,178,1767,194,
            1902,170,2037,260,2172,197,2307,186,2442,181,2577,196,2712,171,2847,
            177,2982,226,3117,147,3252,211,3387,199],
        13: [Integers() | 13,184,148,172,283,181,418,214,553,271,688,130,823,
            236,958,205,1093,173,1228,187,1363,218,1498,242,1633,175,1768,178,
            1903,182,2038,204,2173,199,2308,213,2443,190,2578,172,2713,160,2848,
            218,2983,197,3118,221,3253,190,3388,217],
        14: [Integers() | 14,186,149,174,284,254,419,209,554,198,689,190,824,
            204,959,159,1094,206,1229,174,1364,190,1499,210,1634,158,1769,158,
            1904,183,2039,198,2174,163,2309,166,2444,193,2579,174,2714,199,2849,
            183,2984,233,3119,168,3254,199,3389,196],
        15: [Integers() | 15,177,150,200,285,204,420,178,555,195,690,187,825,
            195,960,157,1095,245,1230,264,1365,163,1500,194,1635,205,1770,233,
            1905,161,2040,205,2175,187,2310,187,2445,190,2580,173,2715,218,2850,
            273,2985,171,3120,225,3255,194,3390,160],
        16: [Integers() | 16,176,151,210,286,249,421,184,556,244,691,209,826,
            214,961,199,1096,218,1231,221,1366,169,1501,224,1636,254,1771,158,
            1906,196,2041,186,2176,188,2311,163,2446,161,2581,237,2716,238,2851,
            155,2986,170,3121,196,3256,196,3391,177],
        17: [Integers() | 17,167,152,184,287,217,422,155,557,184,692,169,827,
            151,962,198,1097,363,1232,213,1367,258,1502,186,1637,174,1772,197,
            1907,183,2042,200,2177,186,2312,242,2447,191,2582,198,2717,198,2852,
            267,2987,300,3122,182,3257,190,3392,259],
        18: [Integers() | 18,216,153,202,288,172,423,229,558,167,693,182,828,
            203,963,204,1098,183,1233,194,1368,210,1503,219,1638,173,1773,217,
            1908,204,2043,217,2178,201,2313,224,2448,179,2583,246,2718,200,2853,
            179,2988,230,3123,178,3258,182,3393,170],
        19: [Integers() | 19,175,154,241,289,182,424,165,559,156,694,209,829,
            166,964,195,1099,148,1234,197,1369,176,1504,257,1639,254,1774,181,
            1909,212,2044,212,2179,222,2314,194,2449,174,2584,240,2719,208,2854,
            166,2989,161,3124,240,3259,168,3394,206],
        20: [Integers() | 20,200,155,229,290,189,425,225,560,263,695,234,830,
            221,965,196,1100,201,1235,185,1370,192,1505,168,1640,209,1775,175,
            1910,198,2045,172,2180,185,2315,207,2450,217,2585,193,2720,153,2855,
            205,2990,214,3125,237,3260,193,3395,187],
        21: [Integers() | 21,200,156,190,291,151,426,301,561,178,696,180,831,
            228,966,186,1101,189,1236,184,1371,180,1506,197,1641,173,1776,238,
            1911,212,2046,234,2181,197,2316,207,2451,204,2586,196,2721,182,2856,
            172,2991,185,3126,148,3261,162,3396,268],
        22: [Integers() | 22,198,157,236,292,196,427,222,562,171,697,222,832,
            182,967,188,1102,238,1237,182,1372,183,1507,192,1642,271,1777,164,
            1912,198,2047,252,2182,173,2317,214,2452,171,2587,193,2722,207,2857,
            180,2992,191,3127,176,3262,174,3397,214],
        23: [Integers() | 23,185,158,179,293,197,428,165,563,191,698,218,833,
            174,968,225,1103,183,1238,296,1373,193,1508,218,1643,260,1778,173,
            1913,159,2048,152,2183,222,2318,156,2453,222,2588,142,2723,162,2858,
            188,2993,160,3128,162,3263,210,3398,205],
        24: [Integers() | 24,168,159,182,294,149,429,180,564,173,699,155,834,
            298,969,234,1104,230,1239,165,1374,187,1509,315,1644,192,1779,188,
            1914,221,2049,154,2184,240,2319,172,2454,200,2589,207,2724,225,2859,
            312,2994,196,3129,175,3264,202,3399,210],
        25: [Integers() | 25,201,160,205,295,213,430,228,565,208,700,219,835,
            183,970,171,1105,154,1240,204,1375,188,1510,158,1645,189,1780,143,
            1915,195,2050,210,2185,181,2320,204,2455,215,2590,184,2725,203,2860,
            213,2995,165,3130,175,3265,191,3400,172],
        26: [Integers() | 26,230,161,211,296,160,431,191,566,223,701,212,836,
            211,971,189,1106,158,1241,178,1376,165,1511,222,1646,152,1781,203,
            1916,171,2051,209,2186,174,2321,241,2456,162,2591,161,2726,161,2861,
            176,2996,218,3131,193,3266,164,3401,203],
        27: [Integers() | 27,208,162,212,297,187,432,189,567,214,702,175,837,
            200,972,160,1107,238,1242,171,1377,152,1512,166,1647,170,1782,217,
            1917,161,2052,219,2187,219,2322,183,2457,156,2592,172,2727,199,2862,
            160,2997,159,3132,184,3267,163,3402,215],
        28: [Integers() | 28,275,163,177,298,214,433,159,568,181,703,177,838,
            295,973,185,1108,200,1243,212,1378,213,1513,154,1648,205,1783,159,
            1918,201,2053,183,2188,232,2323,205,2458,141,2593,167,2728,194,2863,
            181,2998,184,3133,202,3268,201,3403,160],
        29: [Integers() | 29,202,164,206,299,214,434,210,569,252,704,144,839,
            243,974,208,1109,213,1244,202,1379,211,1514,200,1649,176,1784,201,
            1919,212,2054,195,2189,207,2324,218,2459,189,2594,169,2729,163,2864,
            180,2999,210,3134,185,3269,179,3404,192],
        30: [Integers() | 30,192,165,200,300,197,435,260,570,170,705,182,840,
            200,975,212,1110,213,1245,197,1380,256,1515,219,1650,202,1785,207,
            1920,271,2055,169,2190,160,2325,242,2460,196,2595,221,2730,204,2865,
            183,3000,200,3135,194,3270,164,3405,306],
        31: [Integers() | 31,195,166,208,301,173,436,187,571,164,706,192,841,
            165,976,246,1111,237,1246,221,1381,164,1516,194,1651,163,1786,238,
            1921,167,2056,209,2191,239,2326,166,2461,192,2596,193,2731,190,2866,
            198,3001,191,3136,218,3271,229,3406,244],
        32: [Integers() | 32,255,167,188,302,174,437,205,572,173,707,210,842,
            152,977,212,1112,175,1247,189,1382,214,1517,196,1652,241,1787,170,
            1922,175,2057,166,2192,163,2327,182,2462,191,2597,183,2732,161,2867,
            220,3002,163,3137,196,3272,212,3407,229],
        33: [Integers() | 33,201,168,225,303,192,438,191,573,213,708,198,843,
            211,978,222,1113,279,1248,132,1383,198,1518,207,1653,220,1788,200,
            1923,140,2058,199,2193,204,2328,221,2463,210,2598,302,2733,180,2868,
            189,3003,193,3138,153,3273,166,3408,188],
        34: [Integers() | 34,182,169,156,304,207,439,174,574,184,709,150,844,
            210,979,185,1114,201,1249,186,1384,203,1519,250,1654,212,1789,218,
            1924,178,2059,220,2194,167,2329,173,2464,190,2599,178,2734,183,2869,
            188,3004,182,3139,154,3274,286,3409,178],
        35: [Integers() | 35,208,170,146,305,168,440,176,575,271,710,208,845,
            238,980,187,1115,185,1250,194,1385,173,1520,194,1655,232,1790,183,
            1925,188,2060,216,2195,216,2330,217,2465,198,2600,185,2735,168,2870,
            198,3005,148,3140,196,3275,193,3410,202],
        36: [Integers() | 36,173,171,158,306,180,441,171,576,190,711,185,846,
            182,981,209,1116,191,1251,175,1386,190,1521,250,1656,227,1791,151,
            1926,185,2061,134,2196,183,2331,163,2466,216,2601,224,2736,219,2871,
            179,3006,172,3141,195,3276,190,3411,172],
        37: [Integers() | 37,209,172,219,307,150,442,187,577,243,712,186,847,
            161,982,138,1117,166,1252,176,1387,182,1522,173,1657,195,1792,194,
            1927,161,2062,150,2197,175,2332,177,2467,212,2602,209,2737,224,2872,
            222,3007,228,3142,198,3277,224,3412,239],
        38: [Integers() | 38,227,173,254,308,200,443,227,578,165,713,208,848,
            189,983,172,1118,200,1253,201,1388,197,1523,205,1658,242,1793,228,
            1928,248,2063,237,2198,209,2333,210,2468,222,2603,168,2738,188,2873,
            187,3008,198,3143,217,3278,163,3413,229],
        39: [Integers() | 39,208,174,196,309,211,444,197,579,218,714,192,849,
            200,984,187,1119,174,1254,193,1389,190,1524,168,1659,197,1794,196,
            1929,192,2064,145,2199,150,2334,172,2469,217,2604,176,2739,301,2874,
            167,3009,189,3144,249,3279,177,3414,186],
        40: [Integers() | 40,188,175,218,310,201,445,179,580,171,715,164,850,
            164,985,206,1120,208,1255,209,1390,244,1525,165,1660,201,1795,185,
            1930,172,2065,170,2200,182,2335,175,2470,158,2605,194,2740,194,2875,
            169,3010,200,3145,214,3280,239,3415,214],
        41: [Integers() | 41,195,176,195,311,179,446,178,581,214,716,166,851,
            232,986,206,1121,193,1256,202,1391,243,1526,190,1661,190,1796,201,
            1931,174,2066,251,2201,191,2336,179,2471,179,2606,251,2741,189,2876,
            191,3011,222,3146,180,3281,189,3416,219],
        42: [Integers() | 42,212,177,198,312,211,447,179,582,165,717,194,852,
            177,987,201,1122,183,1257,200,1392,185,1527,189,1662,223,1797,174,
            1932,214,2067,190,2202,210,2337,212,2472,184,2607,203,2742,192,2877,
            216,3012,179,3147,157,3282,220,3417,180],
        43: [Integers() | 43,174,178,164,313,209,448,201,583,203,718,222,853,
            200,988,175,1123,255,1258,199,1393,194,1528,143,1663,218,1798,227,
            1933,235,2068,200,2203,175,2338,159,2473,187,2608,187,2743,193,2878,
            184,3013,207,3148,154,3283,184,3418,179],
        44: [Integers() | 44,164,179,179,314,202,449,170,584,210,719,170,854,
            369,989,161,1124,153,1259,195,1394,208,1529,182,1664,186,1799,185,
            1934,216,2069,183,2204,206,2339,200,2474,209,2609,201,2744,188,2879,
            252,3014,227,3149,173,3284,197,3419,178],
        45: [Integers() | 45,171,180,191,315,172,450,178,585,233,720,154,855,
            197,990,171,1125,204,1260,183,1395,244,1530,169,1665,231,1800,226,
            1935,150,2070,188,2205,181,2340,153,2475,204,2610,199,2745,148,2880,
            175,3015,187,3150,179,3285,207,3420,199],
        46: [Integers() | 46,195,181,199,316,180,451,201,586,187,721,213,856,
            169,991,182,1126,183,1261,193,1396,169,1531,266,1666,177,1801,180,
            1936,189,2071,227,2206,203,2341,242,2476,182,2611,203,2746,214,2881,
            175,3016,162,3151,173,3286,229,3421,206],
        47: [Integers() | 47,189,182,267,317,153,452,183,587,172,722,193,857,
            183,992,221,1127,157,1262,169,1397,206,1532,174,1667,175,1802,193,
            1937,227,2072,193,2207,222,2342,181,2477,202,2612,276,2747,190,2882,
            196,3017,184,3152,165,3287,189,3422,165],
        48: [Integers() | 48,206,183,220,318,201,453,245,588,209,723,179,858,
            192,993,191,1128,190,1263,189,1398,192,1533,197,1668,173,1803,197,
            1938,179,2073,185,2208,217,2343,185,2478,209,2613,168,2748,201,2883,
            228,3018,167,3153,192,3288,194,3423,205],
        49: [Integers() | 49,208,184,195,319,252,454,186,589,176,724,195,859,
            207,994,185,1129,166,1264,186,1399,176,1534,164,1669,229,1804,159,
            1939,267,2074,167,2209,160,2344,258,2479,153,2614,182,2749,194,2884,
            219,3019,168,3154,170,3289,207,3424,182],
        50: [Integers() | 50,190,185,176,320,169,455,208,590,183,725,173,860,
            158,995,220,1130,178,1265,209,1400,210,1535,164,1670,183,1805,239,
            1940,200,2075,213,2210,198,2345,206,2480,191,2615,223,2750,191,2885,
            208,3020,196,3155,208,3290,198,3425,181],
        51: [Integers() | 51,165,186,224,321,192,456,186,591,189,726,213,861,
            208,996,180,1131,176,1266,149,1401,230,1536,280,1671,204,1806,186,
            1941,189,2076,193,2211,147,2346,204,2481,191,2616,225,2751,202,2886,
            142,3021,172,3156,206,3291,219,3426,195],
        52: [Integers() | 52,169,187,199,322,209,457,178,592,173,727,165,862,
            172,997,190,1132,219,1267,190,1402,203,1537,185,1672,188,1807,185,
            1942,166,2077,168,2212,180,2347,175,2482,234,2617,202,2752,226,2887,
            164,3022,172,3157,181,3292,170,3427,211],
        53: [Integers() | 53,237,188,151,323,170,458,223,593,231,728,206,863,
            196,998,180,1133,197,1268,237,1403,236,1538,260,1673,189,1808,222,
            1943,247,2078,242,2213,203,2348,229,2483,253,2618,195,2753,163,2888,
            213,3023,190,3158,167,3293,198,3428,170],
        54: [Integers() | 54,186,189,216,324,162,459,224,594,168,729,217,864,
            223,999,293,1134,188,1269,180,1404,145,1539,213,1674,215,1809,200,
            1944,203,2079,173,2214,219,2349,144,2484,171,2619,228,2754,187,2889,
            236,3024,223,3159,189,3294,203,3429,202],
        55: [Integers() | 55,167,190,169,325,177,460,182,595,168,730,200,865,
            205,1000,201,1135,193,1270,200,1405,182,1540,207,1675,185,1810,155,
            1945,167,2080,241,2215,187,2350,214,2485,201,2620,188,2755,282,2890,
            179,3025,220,3160,180,3295,179,3430,208],
        56: [Integers() | 56,195,191,227,326,217,461,219,596,189,731,169,866,
            242,1001,153,1136,167,1271,231,1406,183,1541,216,1676,216,1811,267,
            1946,183,2081,187,2216,177,2351,186,2486,215,2621,195,2756,176,2891,
            250,3026,196,3161,178,3296,189,3431,209],
        57: [Integers() | 57,204,192,206,327,191,462,218,597,189,732,196,867,
            147,1002,206,1137,160,1272,190,1407,172,1542,191,1677,200,1812,206,
            1947,227,2082,188,2217,176,2352,143,2487,176,2622,196,2757,294,2892,
            195,3027,171,3162,157,3297,181,3432,222],
        58: [Integers() | 58,225,193,176,328,199,463,219,598,186,733,230,868,
            161,1003,166,1138,243,1273,230,1408,198,1543,172,1678,179,1813,193,
            1948,174,2083,236,2218,182,2353,193,2488,171,2623,159,2758,200,2893,
            227,3028,231,3163,208,3298,159,3433,199],
        59: [Integers() | 59,160,194,210,329,162,464,220,599,195,734,151,869,
            173,1004,214,1139,187,1274,174,1409,195,1544,218,1679,179,1814,191,
            1949,176,2084,187,2219,198,2354,234,2489,194,2624,189,2759,196,2894,
            201,3029,233,3164,179,3299,174,3434,231],
        60: [Integers() | 60,233,195,221,330,217,465,183,600,206,735,197,870,
            162,1005,220,1140,164,1275,161,1410,165,1545,146,1680,165,1815,227,
            1950,159,2085,196,2220,183,2355,221,2490,210,2625,183,2760,146,2895,
            209,3030,244,3165,168,3300,160,3435,196],
        61: [Integers() | 61,195,196,205,331,179,466,230,601,183,736,199,871,
            167,1006,220,1141,203,1276,194,1411,164,1546,328,1681,163,1816,252,
            1951,171,2086,170,2221,190,2356,324,2491,190,2626,197,2761,181,2896,
            169,3031,234,3166,183,3301,211,3436,188],
        62: [Integers() | 62,223,197,219,332,191,467,219,602,234,737,178,872,
            224,1007,189,1142,205,1277,190,1412,174,1547,149,1682,201,1817,179,
            1952,185,2087,207,2222,197,2357,203,2492,206,2627,167,2762,170,2897,
            179,3032,219,3167,221,3302,217,3437,186],
        63: [Integers() | 63,231,198,179,333,222,468,222,603,169,738,189,873,
            185,1008,149,1143,209,1278,249,1413,163,1548,210,1683,194,1818,194,
            1953,186,2088,232,2223,181,2358,175,2493,167,2628,192,2763,176,2898,
            202,3033,186,3168,166,3303,175,3438,177],
        64: [Integers() | 64,196,199,189,334,231,469,198,604,202,739,218,874,
            154,1009,186,1144,197,1279,192,1414,182,1549,208,1684,219,1819,228,
            1954,216,2089,200,2224,188,2359,185,2494,178,2629,151,2764,187,2899,
            247,3034,153,3169,219,3304,155,3439,196],
        65: [Integers() | 65,187,200,193,335,158,470,184,605,230,740,184,875,
            204,1010,193,1145,182,1280,205,1415,198,1550,176,1685,169,1820,211,
            1955,203,2090,187,2225,180,2360,209,2495,186,2630,182,2765,170,2900,
            153,3035,207,3170,196,3305,235,3440,221],
        66: [Integers() | 66,185,201,251,336,169,471,225,606,198,741,193,876,
            203,1011,185,1146,190,1281,210,1416,185,1551,204,1686,199,1821,180,
            1956,186,2091,169,2226,201,2361,202,2496,202,2631,185,2766,178,2901,
            185,3036,196,3171,189,3306,150,3441,218],
        67: [Integers() | 67,176,202,147,337,177,472,199,607,171,742,229,877,
            156,1012,308,1147,178,1282,182,1417,205,1552,176,1687,256,1822,220,
            1957,194,2092,209,2227,171,2362,218,2497,171,2632,193,2767,214,2902,
            209,3037,181,3172,185,3307,180,3442,193],
        68: [Integers() | 68,192,203,197,338,193,473,227,608,204,743,179,878,
            197,1013,191,1148,221,1283,195,1418,202,1553,226,1688,292,1823,190,
            1958,200,2093,189,2228,150,2363,171,2498,188,2633,219,2768,164,2903,
            192,3038,206,3173,184,3308,165,3443,185],
        69: [Integers() | 69,211,204,164,339,167,474,170,609,211,744,204,879,
            171,1014,197,1149,151,1284,166,1419,161,1554,180,1689,158,1824,186,
            1959,202,2094,169,2229,197,2364,210,2499,179,2634,226,2769,203,2904,
            165,3039,203,3174,211,3309,179,3444,164],
        70: [Integers() | 70,188,205,218,340,161,475,169,610,229,745,220,880,
            145,1015,226,1150,175,1285,214,1420,183,1555,200,1690,179,1825,207,
            1960,167,2095,198,2230,183,2365,158,2500,203,2635,210,2770,172,2905,
            177,3040,232,3175,169,3310,207,3445,164],
        71: [Integers() | 71,178,206,195,341,158,476,230,611,197,746,201,881,
            180,1016,181,1151,193,1286,221,1421,246,1556,201,1691,181,1826,177,
            1961,189,2096,285,2231,223,2366,178,2501,210,2636,164,2771,159,2906,
            154,3041,152,3176,222,3311,213,3446,206],
        72: [Integers() | 72,174,207,238,342,228,477,177,612,185,747,161,882,
            222,1017,256,1152,175,1287,169,1422,232,1557,216,1692,170,1827,176,
            1962,168,2097,182,2232,167,2367,214,2502,199,2637,167,2772,207,2907,
            210,3042,204,3177,177,3312,183,3447,182],
        73: [Integers() | 73,181,208,211,343,163,478,198,613,200,748,177,883,
            218,1018,196,1153,201,1288,188,1423,186,1558,156,1693,175,1828,136,
            1963,221,2098,221,2233,177,2368,198,2503,152,2638,219,2773,201,2908,
            188,3043,242,3178,202,3313,152,3448,179],
        74: [Integers() | 74,168,209,189,344,183,479,190,614,199,749,203,884,
            170,1019,168,1154,172,1289,185,1424,187,1559,347,1694,182,1829,206,
            1964,209,2099,215,2234,325,2369,213,2504,149,2639,166,2774,150,2909,
            188,3044,192,3179,181,3314,241,3449,187],
        75: [Integers() | 75,217,210,197,345,147,480,177,615,202,750,227,885,
            219,1020,203,1155,187,1290,225,1425,166,1560,213,1695,200,1830,207,
            1965,215,2100,178,2235,220,2370,177,2505,197,2640,202,2775,149,2910,
            160,3045,181,3180,208,3315,197,3450,200],
        76: [Integers() | 76,181,211,173,346,153,481,210,616,174,751,181,886,
            192,1021,168,1156,189,1291,195,1426,195,1561,163,1696,147,1831,162,
            1966,140,2101,219,2236,233,2371,207,2506,163,2641,172,2776,230,2911,
            172,3046,174,3181,208,3316,186,3451,158],
        77: [Integers() | 77,204,212,184,347,169,482,203,617,155,752,184,887,
            155,1022,204,1157,183,1292,175,1427,158,1562,226,1697,208,1832,174,
            1967,211,2102,230,2237,172,2372,162,2507,201,2642,223,2777,184,2912,
            232,3047,189,3182,193,3317,190,3452,184],
        78: [Integers() | 78,223,213,200,348,163,483,160,618,215,753,202,888,
            234,1023,203,1158,154,1293,243,1428,205,1563,213,1698,194,1833,178,
            1968,186,2103,186,2238,166,2373,161,2508,160,2643,208,2778,292,2913,
            207,3048,193,3183,212,3318,172,3453,191],
        79: [Integers() | 79,171,214,201,349,179,484,151,619,189,754,207,889,
            225,1024,233,1159,152,1294,161,1429,194,1564,228,1699,196,1834,179,
            1969,221,2104,164,2239,141,2374,174,2509,163,2644,149,2779,187,2914,
            220,3049,151,3184,184,3319,159,3454,179],
        80: [Integers() | 80,180,215,168,350,183,485,241,620,170,755,201,890,
            222,1025,177,1160,186,1295,170,1430,192,1565,175,1700,183,1835,223,
            1970,232,2105,214,2240,233,2375,164,2510,199,2645,171,2780,227,2915,
            216,3050,189,3185,214,3320,194,3455,187],
        81: [Integers() | 81,143,216,158,351,178,486,148,621,165,756,235,891,
            238,1026,234,1161,201,1296,183,1431,210,1566,237,1701,181,1836,246,
            1971,172,2106,213,2241,179,2376,190,2511,155,2646,188,2781,193,2916,
            193,3051,147,3186,241,3321,143,3456,167],
        82: [Integers() | 82,206,217,163,352,174,487,174,622,179,757,184,892,
            166,1027,234,1162,193,1297,176,1432,193,1567,268,1702,166,1837,199,
            1972,197,2107,172,2242,215,2377,210,2512,197,2647,177,2782,206,2917,
            163,3052,187,3187,211,3322,166],
        83: [Integers() | 83,192,218,184,353,198,488,187,623,183,758,194,893,
            193,1028,168,1163,178,1298,190,1433,169,1568,169,1703,211,1838,187,
            1973,223,2108,184,2243,204,2378,230,2513,193,2648,203,2783,188,2918,
            185,3053,147,3188,192,3323,196],
        84: [Integers() | 84,207,219,202,354,238,489,174,624,167,759,175,894,
            167,1029,224,1164,219,1299,186,1434,170,1569,154,1704,221,1839,183,
            1974,187,2109,173,2244,231,2379,177,2514,226,2649,174,2784,179,2919,
            188,3054,189,3189,180,3324,196],
        85: [Integers() | 85,193,220,185,355,213,490,193,625,148,760,167,895,
            170,1030,173,1165,183,1300,155,1435,179,1570,160,1705,271,1840,186,
            1975,178,2110,172,2245,232,2380,225,2515,165,2650,175,2785,227,2920,
            219,3055,215,3190,195,3325,159],
        86: [Integers() | 86,208,221,185,356,182,491,176,626,162,761,272,896,
            190,1031,172,1166,201,1301,163,1436,164,1571,178,1706,219,1841,205,
            1976,191,2111,174,2246,178,2381,225,2516,198,2651,241,2786,192,2921,
            169,3056,222,3191,179,3326,189],
        87: [Integers() | 87,229,222,216,357,228,492,222,627,187,762,198,897,
            196,1032,151,1167,253,1302,164,1437,188,1572,208,1707,191,1842,192,
            1977,172,2112,170,2247,209,2382,223,2517,194,2652,185,2787,164,2922,
            179,3057,179,3192,219,3327,163],
        88: [Integers() | 88,181,223,156,358,159,493,208,628,182,763,183,898,
            141,1033,201,1168,174,1303,265,1438,166,1573,164,1708,160,1843,163,
            1978,166,2113,217,2248,177,2383,211,2518,188,2653,200,2788,170,2923,
            186,3058,205,3193,202,3328,183],
        89: [Integers() | 89,184,224,219,359,157,494,171,629,209,764,199,899,
            196,1034,206,1169,191,1304,183,1439,181,1574,227,1709,206,1844,211,
            1979,172,2114,172,2249,176,2384,197,2519,179,2654,212,2789,238,2924,
            207,3059,185,3194,156,3329,174],
        90: [Integers() | 90,163,225,169,360,215,495,206,630,192,765,196,900,
            275,1035,164,1170,194,1305,229,1440,152,1575,282,1710,191,1845,172,
            1980,193,2115,164,2250,173,2385,179,2520,222,2655,154,2790,212,2925,
            210,3060,244,3195,190,3330,203],
        91: [Integers() | 91,160,226,193,361,222,496,236,631,208,766,238,901,
            188,1036,215,1171,192,1306,198,1441,197,1576,162,1711,203,1846,156,
            1981,200,2116,189,2251,250,2386,182,2521,192,2656,163,2791,165,2926,
            207,3061,198,3196,193,3331,174],
        92: [Integers() | 92,168,227,196,362,215,497,182,632,170,767,195,902,
            185,1037,162,1172,175,1307,204,1442,210,1577,209,1712,191,1847,214,
            1982,183,2117,215,2252,171,2387,151,2522,239,2657,221,2792,168,2927,
            221,3062,221,3197,171,3332,173],
        93: [Integers() | 93,180,228,286,363,197,498,195,633,186,768,199,903,
            216,1038,223,1173,183,1308,200,1443,184,1578,195,1713,165,1848,173,
            1983,211,2118,216,2253,172,2388,201,2523,138,2658,232,2793,219,2928,
            290,3063,182,3198,178,3333,222],
        94: [Integers() | 94,240,229,172,364,173,499,178,634,206,769,224,904,
            419,1039,193,1174,206,1309,209,1444,174,1579,216,1714,142,1849,168,
            1984,154,2119,207,2254,191,2389,178,2524,164,2659,190,2794,208,2929,
            187,3064,202,3199,230,3334,206],
        95: [Integers() | 95,237,230,210,365,196,500,173,635,151,770,186,905,
            214,1040,173,1175,224,1310,263,1445,202,1580,184,1715,183,1850,188,
            1985,226,2120,227,2255,242,2390,215,2525,215,2660,200,2795,266,2930,
            180,3065,198,3200,163,3335,185],
        96: [Integers() | 96,173,231,190,366,196,501,220,636,163,771,179,906,
            212,1041,192,1176,235,1311,168,1446,179,1581,200,1716,281,1851,208,
            1986,204,2121,178,2256,212,2391,196,2526,234,2661,173,2796,207,2931,
            198,3066,193,3201,209,3336,214],
        97: [Integers() | 97,270,232,160,367,149,502,194,637,205,772,230,907,
            212,1042,285,1177,174,1312,199,1447,195,1582,131,1717,186,1852,179,
            1987,197,2122,212,2257,128,2392,228,2527,199,2662,202,2797,200,2932,
            165,3067,203,3202,182,3337,161],
        98: [Integers() | 98,171,233,173,368,219,503,207,638,187,773,198,908,
            266,1043,194,1178,219,1313,231,1448,173,1583,216,1718,178,1853,225,
            1988,207,2123,209,2258,212,2393,206,2528,163,2663,170,2798,181,2933,
            180,3068,195,3203,194,3338,185],
        99: [Integers() | 99,184,234,162,369,206,504,184,639,195,774,172,909,
            204,1044,199,1179,181,1314,207,1449,182,1584,198,1719,263,1854,188,
            1989,193,2124,207,2259,229,2394,245,2529,177,2664,131,2799,167,2934,
            200,3069,163,3204,164,3339,187],
        100: [Integers() | 100,166,235,207,370,158,505,214,640,194,775,194,910,
            164,1045,171,1180,221,1315,137,1450,176,1585,328,1720,177,1855,234,
            1990,194,2125,189,2260,170,2395,148,2530,168,2665,195,2800,177,2935,
            227,3070,179,3205,245,3340,344],
        101: [Integers() | 101,187,236,184,371,279,506,198,641,158,776,214,911,
            181,1046,180,1181,211,1316,154,1451,225,1586,163,1721,190,1856,187,
            1991,192,2126,163,2261,240,2396,182,2531,189,2666,187,2801,211,2936,
            212,3071,195,3206,199,3341,172],
        102: [Integers() | 102,191,237,186,372,207,507,167,642,158,777,158,912,
            186,1047,224,1182,194,1317,222,1452,177,1587,189,1722,198,1857,171,
            1992,150,2127,217,2262,206,2397,204,2532,195,2667,159,2802,200,2937,
            175,3072,220,3207,199,3342,195],
        103: [Integers() | 103,182,238,178,373,237,508,213,643,210,778,157,913,
            175,1048,219,1183,188,1318,181,1453,150,1588,182,1723,182,1858,171,
            1993,195,2128,164,2263,175,2398,197,2533,183,2668,234,2803,239,2938,
            200,3073,267,3208,188,3343,212],
        104: [Integers() | 104,209,239,187,374,168,509,222,644,177,779,153,914,
            202,1049,192,1184,199,1319,199,1454,177,1589,168,1724,170,1859,211,
            1994,232,2129,189,2264,201,2399,183,2534,222,2669,239,2804,158,2939,
            319,3074,200,3209,199,3344,193],
        105: [Integers() | 105,156,240,202,375,219,510,212,645,191,780,167,915,
            207,1050,177,1185,209,1320,174,1455,176,1590,267,1725,206,1860,163,
            1995,181,2130,188,2265,242,2400,220,2535,166,2670,166,2805,219,2940,
            232,3075,187,3210,227,3345,201],
        106: [Integers() | 106,189,241,228,376,191,511,213,646,167,781,181,916,
            178,1051,165,1186,159,1321,183,1456,153,1591,186,1726,215,1861,184,
            1996,250,2131,160,2266,152,2401,170,2536,160,2671,233,2806,188,2941,
            202,3076,176,3211,200,3346,304],
        107: [Integers() | 107,181,242,199,377,228,512,194,647,204,782,200,917,
            126,1052,185,1187,146,1322,199,1457,179,1592,200,1727,244,1862,166,
            1997,217,2132,170,2267,180,2402,171,2537,155,2672,182,2807,232,2942,
            158,3077,177,3212,188,3347,167],
        108: [Integers() | 108,195,243,161,378,205,513,199,648,182,783,220,918,
            173,1053,184,1188,187,1323,232,1458,240,1593,164,1728,225,1863,240,
            1998,154,2133,179,2268,373,2403,173,2538,179,2673,215,2808,178,2943,
            210,3078,241,3213,239,3348,161],
        109: [Integers() | 109,180,244,214,379,188,514,204,649,175,784,151,919,
            192,1054,189,1189,179,1324,201,1459,199,1594,201,1729,155,1864,199,
            1999,197,2134,215,2269,195,2404,224,2539,223,2674,189,2809,158,2944,
            187,3079,219,3214,173,3349,199],
        110: [Integers() | 110,240,245,203,380,183,515,186,650,186,785,213,920,
            222,1055,168,1190,158,1325,200,1460,222,1595,188,1730,178,1865,229,
            2000,169,2135,180,2270,191,2405,197,2540,179,2675,254,2810,207,2945,
            189,3080,172,3215,191,3350,200],
        111: [Integers() | 111,202,246,179,381,172,516,214,651,246,786,168,921,
            232,1056,184,1191,213,1326,240,1461,213,1596,236,1731,221,1866,204,
            2001,184,2136,199,2271,213,2406,202,2541,188,2676,232,2811,171,2946,
            197,3081,209,3216,186,3351,230],
        112: [Integers() | 112,216,247,178,382,214,517,167,652,166,787,256,922,
            184,1057,203,1192,268,1327,222,1462,217,1597,180,1732,138,1867,163,
            2002,153,2137,200,2272,168,2407,184,2542,195,2677,235,2812,299,2947,
            165,3082,153,3217,196,3352,198],
        113: [Integers() | 113,187,248,196,383,186,518,192,653,180,788,251,923,
            168,1058,204,1193,208,1328,201,1463,224,1598,190,1733,209,1868,165,
            2003,186,2138,194,2273,176,2408,217,2543,181,2678,191,2813,194,2948,
            169,3083,263,3218,157,3353,189],
        114: [Integers() | 114,240,249,218,384,196,519,152,654,182,789,179,924,
            224,1059,228,1194,162,1329,202,1464,192,1599,189,1734,168,1869,189,
            2004,189,2139,235,2274,171,2409,180,2544,189,2679,169,2814,164,2949,
            186,3084,201,3219,207,3354,208],
        115: [Integers() | 115,164,250,167,385,204,520,169,655,181,790,188,925,
            223,1060,170,1195,179,1330,168,1465,200,1600,170,1735,160,1870,218,
            2005,190,2140,205,2275,191,2410,198,2545,192,2680,207,2815,201,2950,
            170,3085,220,3220,169,3355,183],
        116: [Integers() | 116,206,251,190,386,174,521,179,656,154,791,197,926,
            221,1061,162,1196,165,1331,193,1466,210,1601,191,1736,177,1871,244,
            2006,156,2141,185,2276,207,2411,221,2546,175,2681,179,2816,261,2951,
            193,3086,246,3221,190,3356,191],
        117: [Integers() | 117,171,252,144,387,182,522,204,657,189,792,210,927,
            198,1062,193,1197,164,1332,182,1467,372,1602,180,1737,198,1872,185,
            2007,176,2142,181,2277,158,2412,210,2547,219,2682,175,2817,216,2952,
            200,3087,230,3222,156,3357,193],
        118: [Integers() | 118,259,253,220,388,239,523,198,658,188,793,186,928,
            220,1063,188,1198,172,1333,188,1468,176,1603,183,1738,195,1873,182,
            2008,180,2143,197,2278,196,2413,147,2548,162,2683,175,2818,176,2953,
            182,3088,154,3223,210,3358,238],
        119: [Integers() | 119,213,254,196,389,236,524,169,659,193,794,196,929,
            189,1064,183,1199,241,1334,186,1469,192,1604,198,1739,229,1874,159,
            2009,189,2144,193,2279,165,2414,189,2549,185,2684,178,2819,187,2954,
            250,3089,213,3224,276,3359,188],
        120: [Integers() | 120,210,255,175,390,170,525,152,660,209,795,209,930,
            178,1065,200,1200,159,1335,196,1470,198,1605,269,1740,245,1875,180,
            2010,170,2145,187,2280,154,2415,202,2550,187,2685,199,2820,164,2955,
            177,3090,247,3225,216,3360,161],
        121: [Integers() | 121,177,256,176,391,197,526,154,661,170,796,185,931,
            193,1066,168,1201,181,1336,194,1471,177,1606,162,1741,224,1876,167,
            2011,187,2146,189,2281,207,2416,199,2551,168,2686,262,2821,195,2956,
            195,3091,203,3226,210,3361,187],
        122: [Integers() | 122,170,257,166,392,173,527,159,662,176,797,150,932,
            219,1067,152,1202,216,1337,190,1472,221,1607,163,1742,205,1877,190,
            2012,174,2147,173,2282,188,2417,165,2552,208,2687,182,2822,206,2957,
            190,3092,177,3227,177,3362,188],
        123: [Integers() | 123,225,258,219,393,195,528,170,663,222,798,202,933,
            180,1068,233,1203,166,1338,181,1473,167,1608,171,1743,180,1878,193,
            2013,182,2148,206,2283,217,2418,278,2553,169,2688,186,2823,171,2958,
            184,3093,177,3228,210,3363,177],
        124: [Integers() | 124,206,259,209,394,187,529,156,664,190,799,198,934,
            215,1069,209,1204,161,1339,177,1474,171,1609,216,1744,163,1879,186,
            2014,203,2149,180,2284,187,2419,166,2554,224,2689,170,2824,191,2959,
            207,3094,188,3229,167,3364,192],
        125: [Integers() | 125,207,260,169,395,216,530,222,665,189,800,196,935,
            210,1070,159,1205,283,1340,167,1475,186,1610,222,1745,159,1880,196,
            2015,214,2150,195,2285,137,2420,152,2555,182,2690,183,2825,211,2960,
            246,3095,201,3230,219,3365,177],
        126: [Integers() | 126,246,261,273,396,190,531,262,666,216,801,201,936,
            178,1071,206,1206,191,1341,164,1476,166,1611,149,1746,189,1881,344,
            2016,190,2151,236,2286,186,2421,192,2556,170,2691,223,2826,178,2961,
            211,3096,220,3231,183,3366,173],
        127: [Integers() | 127,216,262,285,397,218,532,198,667,196,802,178,937,
            202,1072,192,1207,213,1342,147,1477,202,1612,203,1747,145,1882,173,
            2017,236,2152,216,2287,312,2422,206,2557,178,2692,163,2827,196,2962,
            179,3097,213,3232,187,3367,237],
        128: [Integers() | 128,178,263,265,398,156,533,184,668,166,803,234,938,
            201,1073,194,1208,157,1343,165,1478,206,1613,175,1748,199,1883,193,
            2018,185,2153,204,2288,211,2423,166,2558,183,2693,151,2828,228,2963,
            202,3098,171,3233,239,3368,192],
        129: [Integers() | 129,130,264,190,399,190,534,194,669,165,804,166,939,
            256,1074,186,1209,184,1344,165,1479,165,1614,167,1749,215,1884,170,
            2019,192,2154,200,2289,224,2424,196,2559,173,2694,224,2829,178,2964,
            202,3099,200,3234,219,3369,167],
        130: [Integers() | 130,196,265,188,400,187,535,194,670,157,805,174,940,
            202,1075,137,1210,191,1345,196,1480,269,1615,155,1750,205,1885,186,
            2020,201,2155,212,2290,315,2425,191,2560,164,2695,219,2830,202,2965,
            242,3100,203,3235,229,3370,180],
        131: [Integers() | 131,162,266,199,401,193,536,272,671,162,806,213,941,
            237,1076,184,1211,181,1346,220,1481,156,1616,243,1751,198,1886,199,
            2021,154,2156,167,2291,177,2426,253,2561,187,2696,226,2831,163,2966,
            168,3101,222,3236,204,3371,199],
        132: [Integers() | 132,171,267,178,402,215,537,172,672,248,807,188,942,
            228,1077,248,1212,197,1347,203,1482,197,1617,192,1752,196,1887,252,
            2022,210,2157,189,2292,162,2427,169,2562,190,2697,178,2832,158,2967,
            217,3102,254,3237,261,3372,210],
        133: [Integers() | 133,184,268,226,403,185,538,239,673,168,808,218,943,
            217,1078,169,1213,138,1348,156,1483,199,1618,206,1753,221,1888,163,
            2023,165,2158,254,2293,225,2428,198,2563,190,2698,166,2833,195,2968,
            232,3103,166,3238,169,3373,183],
        134: [Integers() | 134,157,269,276,404,174,539,197,674,213,809,152,944,
            159,1079,317,1214,184,1349,193,1484,227,1619,188,1754,183,1889,215,
            2024,173,2159,249,2294,250,2429,166,2564,204,2699,161,2834,246,2969,
            211,3104,214,3239,179,3374,193],
      default: 0 >;
    // Return the block and offset
    start:=1;
    for i in [1..#offset by 2] do
        if offset[i] eq hash1 then
            return block,start,start + offset[i + 1] - 1;
        else
            start +:= offset[i + 1];
        end if;
    end for;
    error "The polytope is not a canonical Fano 3-tope";
end function;

/////////////////////////////////////////////////////////////////////////
// Local functions
/////////////////////////////////////////////////////////////////////////

// Returns three integers in the range 0..3462. Argument should the the facet
// ids of the canonical Fano 3-fold, sorted in increasing order.
function hash_fids(I)
    F1:=Integers(3457);
    F2:=Integers(3461);
    F3:=Integers(3463);
    hash1:=F1 ! 0;
    hash2:=F2 ! 0;
    hash3:=F3 ! 0;
    p:=2;
    for i in Reverse(I) do
        p1:=F1 ! p;
        hash1:=(hash1 + 1) * p1^i;
        p2:=F2 ! p;
        hash2:=(hash2 + 1) * p2^i;
        p3:=F3 ! p;
        hash3:=(hash3 + 1) * p3^i;
        p:=NextPrime(p);
    end for;
    return Integers() ! hash1, Integers() ! hash2, Integers() ! hash3;
end function;

// Returns a true followed by a sequence of possible database ids for the given
// canonical Fano 3-tope, false followed by an error message otherwise.
function candidate_ids(P)
    // Compute the facet ids and calculate the hash
    fids:=Sort([canonical_dim3_facet_to_id(F) : F in Facets(P)]);
    hash1,hash2,hash3:=hash_fids(fids);
    // It's possible the result is in the hard-coded cache functions above
    case hash1 mod 47:
        when 0:
            bool,ids:=cached_ids_0(P,hash1,hash2,hash3);
        when 1:
            bool,ids:=cached_ids_1(P,hash1,hash2,hash3);
        when 2:
            bool,ids:=cached_ids_2(P,hash1,hash2,hash3);
        when 3:
            bool,ids:=cached_ids_3(P,hash1,hash2,hash3);
        when 4:
            bool,ids:=cached_ids_4(P,hash1,hash2,hash3);
        when 5:
            bool,ids:=cached_ids_5(P,hash1,hash2,hash3);
        when 6:
            bool,ids:=cached_ids_6(P,hash1,hash2,hash3);
        when 7:
            bool,ids:=cached_ids_7(P,hash1,hash2,hash3);
        when 8:
            bool,ids:=cached_ids_8(P,hash1,hash2,hash3);
        when 9:
            bool,ids:=cached_ids_9(P,hash1,hash2,hash3);
        when 10:
            bool,ids:=cached_ids_10(P,hash1,hash2,hash3);
        when 11:
            bool,ids:=cached_ids_11(P,hash1,hash2,hash3);
        when 12:
            bool,ids:=cached_ids_12(P,hash1,hash2,hash3);
        when 13:
            bool,ids:=cached_ids_13(P,hash1,hash2,hash3);
        when 14:
            bool,ids:=cached_ids_14(P,hash1,hash2,hash3);
        when 15:
            bool,ids:=cached_ids_15(P,hash1,hash2,hash3);
        when 16:
            bool,ids:=cached_ids_16(P,hash1,hash2,hash3);
        when 17:
            bool,ids:=cached_ids_17(P,hash1,hash2,hash3);
        when 18:
            bool,ids:=cached_ids_18(P,hash1,hash2,hash3);
        when 19:
            bool,ids:=cached_ids_19(P,hash1,hash2,hash3);
        when 20:
            bool,ids:=cached_ids_20(P,hash1,hash2,hash3);
        when 21:
            bool,ids:=cached_ids_21(P,hash1,hash2,hash3);
        when 22:
            bool,ids:=cached_ids_22(P,hash1,hash2,hash3);
        when 23:
            bool,ids:=cached_ids_23(P,hash1,hash2,hash3);
        when 24:
            bool,ids:=cached_ids_24(P,hash1,hash2,hash3);
        when 25:
            bool,ids:=cached_ids_25(P,hash1,hash2,hash3);
        when 26:
            bool,ids:=cached_ids_26(P,hash1,hash2,hash3);
        when 27:
            bool,ids:=cached_ids_27(P,hash1,hash2,hash3);
        when 28:
            bool,ids:=cached_ids_28(P,hash1,hash2,hash3);
        when 29:
            bool,ids:=cached_ids_29(P,hash1,hash2,hash3);
        when 30:
            bool,ids:=cached_ids_30(P,hash1,hash2,hash3);
        when 31:
            bool,ids:=cached_ids_31(P,hash1,hash2,hash3);
        when 32:
            bool,ids:=cached_ids_32(P,hash1,hash2,hash3);
        when 33:
            bool,ids:=cached_ids_33(P,hash1,hash2,hash3);
        when 34:
            bool,ids:=cached_ids_34(P,hash1,hash2,hash3);
        when 35:
            bool,ids:=cached_ids_35(P,hash1,hash2,hash3);
        when 36:
            bool,ids:=cached_ids_36(P,hash1,hash2,hash3);
        when 37:
            bool,ids:=cached_ids_37(P,hash1,hash2,hash3);
        when 38:
            bool,ids:=cached_ids_38(P,hash1,hash2,hash3);
        when 39:
            bool,ids:=cached_ids_39(P,hash1,hash2,hash3);
        when 40:
            bool,ids:=cached_ids_40(P,hash1,hash2,hash3);
        when 41:
            bool,ids:=cached_ids_41(P,hash1,hash2,hash3);
        when 42:
            bool,ids:=cached_ids_42(P,hash1,hash2,hash3);
        when 43:
            bool,ids:=cached_ids_43(P,hash1,hash2,hash3);
        when 44:
            bool,ids:=cached_ids_44(P,hash1,hash2,hash3);
        when 45:
            bool,ids:=cached_ids_45(P,hash1,hash2,hash3);
        when 46:
            bool,ids:=cached_ids_46(P,hash1,hash2,hash3);
    end case;
    if bool then
        return true,ids;
    end if;
    // Recover the block file and line offsets
    block,start,finish:=block_and_offset_data(hash1);
    // Open the block file
    bool,F:=open_database_block("canonical3/reverse",block :
                                name:="canonical3");
    if not bool then return false,F; end if;
    // Look for the matching lines
    hash:=IntegerToString(hash2) cat IntegerToString(hash3);
    line:=1;
    S:=Gets(F);
    ids:=[Integers()|];
    while line le finish and not IsEof(S) do
        if line ge start then
            idx:=Index(S,":");
            if idx eq 0 then
                return false,"The database \"canonical3\" is corrupted";
            end if;
            if S[1..idx - 1] eq hash then
                Append(~ids,StringToInteger(S[idx + 1..#S]));
            end if;
        end if;
        S:=Gets(F);
        line +:= 1;
    end while;
    // Handle any problems
    if #ids eq 0 then
        return false,"The database \"canonical3\" is corrupted";
    end if;
    // Return what we have
    return true,ids;
end function;

// Assuming P is a canonical Fano 3-tope, returns true followed by the
// corresponding ID in the database. May also return false followed by an error
// message.
function canonical_dim3_to_id(P)
    // Fetch the candidate ids
    bool,ids:=candidate_ids(P);
    if not bool then return false,ids; end if;
    // If there's a unique id, this is easy...
    if #ids eq 1 then return true,ids[1]; end if;
    // ...otherwise we need to hunt the candidates for a match
    for id in ids do
        nf:=NormalForm(P);
        if nf eq NormalForm(PolytopeCanonicalFanoDim3(id)) then
            return true,id;
        end if;
    end for;
    // We should never get here
    return false,"Argument must be a canonical Fano 3-tope";
end function;
